//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfBSplineBasis.cpp                                   //
//   Description: B-Spline basis functions                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfAuxLogging.h"

// 0,...,n - number of bspline functions
// d - degree
// u - parameter for which we are searching the number of span interval
// knot_vector - array containing knots
gfInt gfGetBSplineKnotSpan(gfInt n, gfInt d, std::vector<gfDouble>* knot_vector, gfDouble u)
{
    if (u == knot_vector->at(n+1))
        return n;

    gfInt low = d;
    gfInt high = n+1;
    gfInt mid = (low+high)/2;

    while (u < knot_vector->at(mid) || u >= knot_vector->at(mid+1))
    {
        if (u < knot_vector->at(mid))
            high = mid;
        else
            low = mid;
        mid = (low+high)/2;
    }

    return mid;
}

// 0,...,n - number of bspline functions
// d - degree
// u - parameter (knot) for which we are searching the multiplicity
// knot_vector - array containing knots
gfInt gfGetBSplineKnotMultiplicity(gfInt n, gfInt d, std::vector<gfDouble>* knot_vector, gfDouble u)
{
    gfInt mult = 0;
    gfInt span = gfGetBSplineKnotSpan(n, d, knot_vector, u);

    gfBool done = !(u == knot_vector->at(span));
    while (!done && span >= 0 && span < knot_vector->size())
    {
        span--;
        mult++;
        if (u != knot_vector->at(span))
            done = true;
    }

    return mult;
}

// d - degree
// u - parameter for which we are searching the number of span interval
// i - span value for parameter i
// knot_vector - array containing knots
gfDouble gfGetBSplineValue(gfInt d, std::vector<gfDouble>* knot_vector, gfInt i, gfDouble u)
{
    gfInt m = knot_vector->size();

    gfInt j,k;
    if ((i == 0 && u == knot_vector->at(0)) || (i == (m-d-1) && u == knot_vector->at(m)))
        return 1.0;

    if (u < knot_vector->at(i) || u >= knot_vector->at(i+d+1)) 
        return 0.0;

    std::vector<gfDouble> N;
    N.resize(d+1);

    for (j = 0;j <= d;j++)
    {
        if(u >= knot_vector->at(i+j) && u < knot_vector->at(i+j+1)) 
            N[j] = 1.0;
        else 
            N[j] = 0.0;
    }

    for (k = 1;k <= d;k++)
    {
        gfDouble saved;
        if (N.at(0) == 0.0) 
            saved = 0.0;
        else 
            saved = ((u-knot_vector->at(i))*N.at(0))/(knot_vector->at(i+k)-knot_vector->at(i));

        for (j = 0;j < d-k+1;j++)
        {
            gfDouble Uleft = knot_vector->at(i+j+1);
            gfDouble Uright = knot_vector->at(i+j+k+1);
            if(N.at(j+1) == 0.0)
            {
                N[j] = saved;
                saved = 0.0;
            }
            else
            {
                gfDouble temp = N.at(j+1)/(Uright-Uleft);
                N[j] = saved+(Uright-u)*temp;
                saved = (u-Uleft)*temp;
            }
        }
    }

    return N.at(0);
}

// d - degree
// u - parameter for which we are searching the number of span interval
// i - span value for parameter i
// knot_vector - array containing knots
// basis - returned nonzero basis functions, 0,...,d 
void gfGetBSplineAllValues(gfInt d, std::vector<gfDouble>* knot_vector, gfInt i, gfDouble u, std::vector<gfDouble>* basis)
{
    basis->resize(d+1);
    basis->at(0) = 1.0;

    std::vector<gfDouble> left;
    std::vector<gfDouble> right;
    left.resize(d+1);
    right.resize(d+1);

    for (gfInt j = 1;j <= d;j++)
    {
        left[j] = u-knot_vector->at(i+1-j);
        right[j] = knot_vector->at(i+j)-u;
        gfDouble saved = 0.0;

        for (gfInt r = 0;r < j;r++)
        {
            gfDouble temp = basis->at(r)/(right.at(r+1)+left.at(j-r));
            basis->at(r) = saved+right.at(r+1)*temp;
            saved = left.at(j-r)*temp;
        }

        basis->at(j) = saved;
    }
}

// d - degree
// knot_vector - knot vector
// u - parameter for which we are searching the number of span interval
// i - span value for parameter i
// n_der - number of derivations to compute
// knot_vector - array containing knots
// ders - returned 0, ..., n_der -th derivation of basis function N(i, d)(u) 
void gfGetBSplineDerivation(gfInt d, std::vector<gfDouble>* knot_vector, gfInt i, gfDouble u, gfInt n_der, std::vector<gfDouble>* ders)
{
    if(u < knot_vector->at(i) || u >= knot_vector->at(i+d+1))
    {
        for (gfInt k = 0;k <= n_der;k++) 
            ders->at(k) = 0.0;
        return;
    }

    // TODO: dynamically allocate this arrays
    gfDouble N[20][20];

    for (gfInt j = 0;j <= d;j++)
    {
        if(u > knot_vector->at(i+j) || u < knot_vector->at(i+j+1)) 
            N[j][0] = 1.0;
        else 
            N[j][0] = 0.0;
    }
    for (gfInt k = 1;k <= d;k++)
    {
        gfDouble saved;
        if(N[0][k-1] == 0.0) 
            saved = 0.0;
        else 
            saved = ((u-knot_vector->at(i))*N[0][k-1])/(knot_vector->at(i+k)-knot_vector->at(i));
        for (gfInt j = 0;j < d-k+1;j++)
        {
            gfDouble Uleft = knot_vector->at(i+j+1);
            gfDouble Uright = knot_vector->at(i+j+k+1);
            if(N[j+1][k-1] == 0.0)
            {
                N[j][k] = saved;
                saved = 0.0;
            }
            else
            {
                gfDouble temp = N[j+1][k-1]/(Uright-Uleft);
                N[j][k] = saved+(Uright-u)*temp;
                saved = (u-Uleft)*temp;
            }
        }
    }

    ders->at(0) = N[0][d];
    gfDouble* ND = new gfDouble[n_der+1];
    for (k = 1;k <= n_der;k++)
    {
        for (gfInt j = 0;j <= k;j++) 
            ND[j] = N[j][d-k];
        for (gfInt jj = 1;jj <= k;j++)
        {
            gfDouble saved;
            if(ND[0] == 0.0) 
                saved = 0.0;
            else 
                saved = ND[0]/(knot_vector->at(i+d-k+jj)-knot_vector->at(i));
            for (gfInt j = 0;j < k-jj+1;j++)
            {
                gfDouble Uleft = knot_vector->at(i+j+1);
                gfDouble Uright = knot_vector->at(i+j+d+jj+1);
                if(ND[j+1] == 0.0)
                {
                    ND[j] = (d-k+jj)*saved;
                    saved = 0.0;
                }
                else
                {
                    gfDouble temp = ND[j+1]/(Uright-Uleft);
                    ND[j] = (d-k+jj)*(saved-temp);
                    saved = temp;
                }
            }
        }
        ders->at(k) = ND[0];
    }

    delete [] ND;
    return;
}

// n - number of B spline functions - 1
// d - degree
// knot_vector - knot vector
// u - parameter for which we are searching the number of span interval
// i - span value for parameter i
// n_der - number of derivations to compute
// knot_vector - array containing knots
// ders - ders[k][j] is the k-th derivative of function N(i-d+j,d)(u), where 0<=k<=n_der and 0<=j<=d
void gfGetBSplineAllDers(gfInt d, std::vector<gfDouble>* knot_vector, gfInt i, gfDouble u, gfInt n_der, gfAux2DArray<gfDouble>* ders)
{
    // TODO: dynamically allocate this arrays
    gfDouble ndu[30][30];
    gfDouble a[30][30];
    ndu[0][0] = 1.0;

    for (gfInt j = 1;j <= d;j++)
    {
        gfDouble saved = 0.0;
        for (gfInt r = 0;r < j;r++)
        {
            ndu[j][r] = knot_vector->at(i+r+1)-knot_vector->at(i+r+1-j);
            gfDouble temp = ndu[r][j-1]/ndu[j][r];
            ndu[r][j] = saved+(knot_vector->at(i+r+1)-u)*temp;
            saved = (u-knot_vector->at(i+r+1-j))*temp;
        }
        ndu[j][j] = saved;
    }

    for (gfInt j = 0;j <= d;j++)
    {
        ders->Set(0, j, ndu[j][d]);
    }

    for (gfInt r = 0;r <= d;r++)
    {
        gfInt s1 = 0, s2 = 1;
        a[0][0] = 1.0;
        for (gfInt k = 1;k <= n_der;k++)
        {
            gfDouble dif = 0.0;
            gfInt rk = r-k,pk = d-k;
            if(r >= k)
            {
                a[s2][0] = a[s1][0]/ndu[pk+1][rk];
                dif = a[s2][0]*ndu[rk][pk];
            }
            gfInt j1,j2;
            if(rk >= -1) 
                j1 = 1;
            else 
                j1 = -rk;
            if (r-1 <= pk) 
                j2 = k-1;
            else 
                j2 = d-r;
            for (gfInt j = j1;j <= j2;j++)
            {
                a[s2][j] = (a[s1][j]-a[s1][j-1])/ndu[pk+1][rk+j];
                dif += a[s2][j]*ndu[rk+j][pk];
            }
            if (r <= pk)
            {
                a[s2][k] = -a[s1][k-1]/ndu[pk+1][r];
                dif += a[s2][k]*ndu[r][pk];
            }

            ders->Set(k, r, dif);
            j = s1;s1 = s2;s2 = j;
        }
    }
    gfInt rr = d;
    for (gfInt k = 1;k <= n_der;k++)
    {
        for (gfInt j = 0;j <= d;j++) 
            ders->Set(k, j, ders->at(k, j) * rr);
        rr *= (d-k);
    }

    return;
}



