//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfAuxBoundCone.h                                     //
//   Description: Class for cones bounding set of vectors       //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "gfAuxBoundCone.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"


gfAuxBoundCone::gfAuxBoundCone(void)
{
    m_angle = 0.0;
    m_axis.Set(0.0, 0.0, 0.0);
}

gfAuxBoundCone::~gfAuxBoundCone(void)
{
}

GFRESULT gfAuxBoundCone::CreateFromVectors(std::vector<gfVector3>* vectors)
{
    // find axis
    m_axis.Set(0.0, 0.0, 0.0);
    gfInt size = (gfInt)vectors->size();
    for (gfInt i = 0;i < size;i++)
    {
        gfVector3 vec = vectors->at(i);
        vec.Normalize();
        gfCombineVectors3(1.0, &m_axis, 1.0, &vec, &m_axis);
    }
    m_axis.Normalize();

    //find angle
    gfDouble min = GF_MAX_DOUBLE;
    for (gfInt i = 0;i < size;i++)
    {
        gfVector3 vec = vectors->at(i);
        vec.Normalize();
        gfDouble scp = gfScalarProduct3(&m_axis, &vec);
        if (scp < min)
            min = scp;
    }
    m_angle = gfArccos(min);

    return GF_S_OK;
}

GFRESULT gfAuxBoundCone::CrossProduct(gfAuxBoundCone c1, gfAuxBoundCone  c2)
{
    gfVectorProduct3(&c1.m_axis, &c2.m_axis, &m_axis);
    m_axis.Normalize();
    /*gfDouble beta = gfScalarProduct3(&c1.m_axis, &c2.m_axis);

    gfDouble s1 = gfSin(c1.m_angle);
    gfDouble s2 = gfSin(c2.m_angle);
    m_angle = gfArcsin(gfSquareRoot(s1*s1+2*s1*s2*gfCos(beta)+s2*s2)/gfSin(beta));*/
    m_angle = (c1.m_angle > c2.m_angle) ? c1.m_angle : c2.m_angle;

    return GF_S_OK;
}

GFRESULT gfAuxBoundCone::ScalarProduct(gfAuxBoundCone second, gfDouble* min, gfDouble* max)
{
    // get four boundary vectors
    gfVector3 v1, v2, v3, v4;

    gfDouble SC = gfScalarProduct3(&m_axis, &second.m_axis);
    gfDouble beta = gfCos(m_angle);
    gfDouble p, q;

    if (gfIsZero(1.0-SC))
    {
        *min = gfCos(m_angle+second.m_angle);
        *max = gfCos(m_angle-second.m_angle);

        return GF_S_OK;
    }

    q = gfSquareRoot((1.0-beta*beta)/(1.0-SC*SC));
    p = beta-q*SC;
    gfCombineVectors3(p, &m_axis, q, &second.m_axis, &v1);

    q = -q;
    p = beta-q*SC;
    gfCombineVectors3(p, &m_axis, q, &second.m_axis, &v2);

    beta = gfCos(second.m_angle);
    
    p = gfSquareRoot((1.0-beta*beta)/(1.0-SC*SC));
    q = beta-p*SC;
    gfCombineVectors3(p, &m_axis, q, &second.m_axis, &v3);

    p = -p;
    q = beta-p*SC;
    gfCombineVectors3(p, &m_axis, q, &second.m_axis, &v4);

    // find minimal and maximal values of scalar products
    *min = GF_MAX_DOUBLE;
    *max = GF_MIN_DOUBLE;

    gfDouble prod = gfScalarProduct3(&v1, &v2);
    if (prod < *min)
        *min = prod;
    if (prod > *max)
        *max = prod;

    prod = gfScalarProduct3(&v1, &v3);
    if (prod < *min)
        *min = prod;
    if (prod > *max)
        *max = prod;

    prod = gfScalarProduct3(&v1, &v4);
    if (prod < *min)
        *min = prod;
    if (prod > *max)
        *max = prod;

    prod = gfScalarProduct3(&v2, &v3);
    if (prod < *min)
        *min = prod;
    if (prod > *max)
        *max = prod;

    prod = gfScalarProduct3(&v2, &v4);
    if (prod < *min)
        *min = prod;
    if (prod > *max)
        *max = prod;

    prod = gfScalarProduct3(&v3, &v4);
    if (prod < *min)
        *min = prod;
    if (prod > *max)
        *max = prod;

    return GF_S_OK;
}