//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfAuxBoundingVolume.cpp                              //
//   Description: Class for representing bounding volumes       //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "gfAuxBoundingVolume.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfObjectFactory.h"
#include "gfAuxItersectionUtils.h"

#include "./OGL/glut.h"


gfAuxBoundingShere::gfAuxBoundingShere(void)
{
    m_middle_x = 0.0;
    m_middle_y = 0.0;
    m_middle_z = 0.0;

    m_radius = 0.0;
}

gfBoundVolID gfAuxBoundingShere::GetID(void)
{
    return GF_BOUND_VOL_SPHERE;
}

GFRESULT gfAuxBoundingShere::RayIntersection(gfPoint3* ray_begin, gfVector3* ray_dir, gfDouble* tmin, gfDouble* tmax)
{
    gfDouble a = ray_dir->X()*ray_dir->X()+ray_dir->Y()*ray_dir->Y()+ray_dir->Z()*ray_dir->Z();
    gfDouble b = 2*(ray_dir->X()*(ray_begin->X()-m_middle_x)+ray_dir->Y()*(ray_begin->Y()-m_middle_y)+
                 ray_dir->Z()*(ray_begin->Z()-m_middle_z));
    gfDouble c = (ray_begin->X()-m_middle_x)*(ray_begin->X()-m_middle_x)+
                 (ray_begin->Y()-m_middle_y)*(ray_begin->Y()-m_middle_y)+
                 (ray_begin->Z()-m_middle_z)*(ray_begin->Z()-m_middle_z);

    gfDouble D = b*b-4*a*c;
    if (D<0)
        return GF_E_FAIL;
    *tmin = (-b-D)/2*a;
    *tmax = (-b+D)/2*a;

    if (*tmin > *tmax)
    {
        gfDouble temp = *tmin;
        *tmin = *tmax;
        *tmax = temp;
    }

    return GF_S_OK;
}

GFRESULT gfAuxBoundingShere::OGLDraw(gfLong flags)
{    
    gfInt i, j;
    gfDouble r, x, y, z;
    gfInt slices = 16, stacks = 16;

    for (i = 1; i < stacks; i++)
    {
        z = gfCos(GF_PI * i / stacks);
        r = gfSin(GF_PI * i / stacks);

        glBegin(GL_LINE_LOOP);

            for (j = 0; j <= slices; j++)
            {
                x = gfCos(2 * GF_PI * j / slices);
                y = gfSin(2 * GF_PI * j / slices);

                glVertex3d(m_middle_x + x * r * m_radius, m_middle_y + y * r * m_radius, m_middle_z + z * m_radius);
            }

        glEnd();
    }

    gfDouble s, c;
    for (i = 0; i < slices; i++)
    {
        s = gfSin(2 * GF_PI * i / slices);
        c = gfCos(2 * GF_PI * i / slices);

        glBegin(GL_LINE_LOOP);

            for (j = 0; j <= stacks; j++)
            {
                x = c * gfSin(GF_PI * j / stacks);
                y = s * gfSin(GF_PI * j / stacks);
                z = gfCos(GF_PI * j / stacks);

                glVertex3d(m_middle_x + x * m_radius, m_middle_y + y * m_radius, m_middle_z + z * m_radius);
            }

        glEnd();
    }

    return GF_E_FAIL;
}

GFRESULT gfAuxBoundingShere::Transform(gfMatrix4* matrix)
{
    gfPoint3 pt(m_middle_x, m_middle_y, m_middle_z);
    gfTransformPoint3(&pt, matrix);
    m_middle_x = pt.X();
    m_middle_y = pt.Y();
    m_middle_z = pt.Z();

    return GF_S_OK;
}

gfPoint3* gfAuxBoundingShere::GetCenter(void)
{
    gfPoint3* ret = gfCreatePoint3();
    ret->Set(m_middle_x, m_middle_y, m_middle_z);
    return ret;
}

GFRESULT gfAuxBoundingShere::InitFromPoints(gfInt num_points, gfPoint3* points)
{
    if (num_points < 1)
        return GF_E_BADARG;

    // compute mean value of points (barycenter)
    m_middle_x = 0.0;
    m_middle_y = 0.0;
    m_middle_z = 0.0;

    for (gfInt i = 0; i < num_points; i++)
    {
        m_middle_x += points[i].X();
        m_middle_y += points[i].Y();
        m_middle_z += points[i].Z();
    }

    m_middle_x /= num_points;
    m_middle_y /= num_points;
    m_middle_z /= num_points;

    // get point with largest distance from barycenter
    gfDouble tx, ty, tz;
    gfDouble d, dist = 0.0;
    for (gfInt i = 0; i < num_points; i++)
    {
        tx = points[i].X();
        ty = points[i].Y();
        tz = points[i].Z();

        d = (tx - m_middle_x) * (tx - m_middle_x) + (ty - m_middle_y) * (ty - m_middle_y) + (tz - m_middle_z) * (tz - m_middle_z);
        if (d > dist)
            dist = d;
    }

    m_radius = gfSquareRoot(dist);

    return GF_S_OK;
}

gfAuxAABB::gfAuxAABB(void)
{
    m_min_x = m_max_x = m_min_y = m_max_y = m_min_z = m_max_z = 0.0;
}

gfBoundVolID gfAuxAABB::GetID(void)
{
    return GF_BOUND_VOL_AABB;
}

GFRESULT gfAuxAABB::RayIntersection(gfPoint3* ray_begin, gfVector3* ray_dir, gfDouble* tmin, gfDouble* tmax)
{
    gfDouble t, auxx, auxy, auxz;
    *tmin = GF_MAX_DOUBLE;
    *tmax = GF_MIN_DOUBLE;
    gfInt inter = 0;

    if (!gfIsZero(ray_dir->X()))
    {
        t = (m_min_x - ray_begin->X()) / ray_dir->X();
        auxy = ray_begin->Y() + t * ray_dir->Y();
        auxz = ray_begin->Z() + t * ray_dir->Z();
        if (auxy >= m_min_y && auxy <= m_max_y && auxz >= m_min_z && auxz <= m_max_z)
        {
            if (t < *tmin)
                *tmin = t;
            if (t > *tmax)
                *tmax = t;
            inter++;
        }

        t = (m_max_x - ray_begin->X()) / ray_dir->X();
        auxy = ray_begin->Y() + t * ray_dir->Y();
        auxz = ray_begin->Z() + t * ray_dir->Z();
        if (auxy >= m_min_y && auxy <= m_max_y && auxz >= m_min_z && auxz <= m_max_z)
        {
            if (t < *tmin)
                *tmin = t;
            if (t > *tmax)
                *tmax = t;
            inter++;
            if (inter == 2)
                return GF_S_OK;
        }
    }

    if (!gfIsZero(ray_dir->Y()))
    {
        t = (m_min_y - ray_begin->Y()) / ray_dir->Y();
        auxx = ray_begin->X() + t * ray_dir->X();
        auxz = ray_begin->Z() + t * ray_dir->Z();
        if (auxx >= m_min_x && auxx <= m_max_x && auxz >= m_min_z && auxz <= m_max_z)
        {
            if (t < *tmin)
                *tmin = t;
            if (t > *tmax)
                *tmax = t;
            inter++;
            if (inter == 2)
                return GF_S_OK;
        }

        t = (m_max_y - ray_begin->Y()) / ray_dir->Y();
        auxx = ray_begin->X() + t * ray_dir->X();
        auxz = ray_begin->Z() + t * ray_dir->Z();
        if (auxx >= m_min_x && auxx <= m_max_x && auxz >= m_min_z && auxz <= m_max_z)
        {
            if (t < *tmin)
                *tmin = t;
            if (t > *tmax)
                *tmax = t;
            inter++;
            if (inter == 2)
                return GF_S_OK;
        }
    }

    if (!gfIsZero(ray_dir->Z()))
    {
        t = (m_min_z - ray_begin->Z()) / ray_dir->Z();
        auxy = ray_begin->Y() + t * ray_dir->Y();
        auxx = ray_begin->X() + t * ray_dir->X();
        if (auxy >= m_min_y && auxy <= m_max_y && auxx >= m_min_x && auxx <= m_max_x)
        {
            if (t < *tmin)
                *tmin = t;
            if (t > *tmax)
                *tmax = t;
            inter++;
            if (inter == 2)
                return GF_S_OK;
        }

        t = (m_max_z - ray_begin->Z()) / ray_dir->Z();
        auxy = ray_begin->Y() + t * ray_dir->Y();
        auxx = ray_begin->X() + t * ray_dir->X();
        if (auxy >= m_min_y && auxy <= m_max_y && auxx >= m_min_x && auxx <= m_max_x)
        {
            if (t < *tmin)
                *tmin = t;
            if (t > *tmax)
                *tmax = t;
            inter++;
            if (inter == 2)
                return GF_S_OK;
        }
    }

    return GF_E_FAIL;
}

GFRESULT gfAuxAABB::OGLDraw(gfLong flags)
{
    glBegin(GL_LINES);

    glVertex3f(m_min_x, m_min_y, m_min_z);
    glVertex3f(m_max_x, m_min_y, m_min_z);

    glVertex3f(m_min_x, m_max_y, m_min_z);
    glVertex3f(m_max_x, m_max_y, m_min_z);

    glVertex3f(m_min_x, m_min_y, m_max_z);
    glVertex3f(m_max_x, m_min_y, m_max_z);

    glVertex3f(m_min_x, m_max_y, m_max_z);
    glVertex3f(m_max_x, m_max_y, m_max_z);

    glVertex3f(m_min_x, m_min_y, m_min_z);
    glVertex3f(m_min_x, m_max_y, m_min_z);

    glVertex3f(m_max_x, m_min_y, m_min_z);
    glVertex3f(m_max_x, m_max_y, m_min_z);

    glVertex3f(m_min_x, m_min_y, m_max_z);
    glVertex3f(m_min_x, m_max_y, m_max_z);

    glVertex3f(m_max_x, m_min_y, m_max_z);
    glVertex3f(m_max_x, m_max_y, m_max_z);

    glVertex3f(m_min_x, m_min_y, m_min_z);
    glVertex3f(m_min_x, m_min_y, m_max_z);

    glVertex3f(m_max_x, m_min_y, m_min_z);
    glVertex3f(m_max_x, m_min_y, m_max_z);

    glVertex3f(m_min_x, m_max_y, m_min_z);
    glVertex3f(m_min_x, m_max_y, m_max_z);

    glVertex3f(m_max_x, m_max_y, m_min_z);
    glVertex3f(m_max_x, m_max_y, m_max_z);

    glEnd();

    return GF_S_OK;
}

GFRESULT gfAuxAABB::Transform(gfMatrix4* matrix)
{
    // must be rocomputed from object
    return GF_E_FAIL;
}

gfPoint3* gfAuxAABB::GetCenter(void)
{
    gfPoint3* ret = gfCreatePoint3();
    ret->Set((m_min_x + m_max_x) / 2.0, (m_min_y + m_max_y) / 2.0, (m_min_z + m_max_z) / 2.0);
    return ret;
}

GFRESULT gfAuxAABB::InitFromPoints(gfInt num_points, gfPoint3* points)
{
    if (num_points < 1)
        return GF_E_BADARG;

    m_min_x = GF_MAX_DOUBLE;
    m_max_x = GF_MIN_DOUBLE;
    m_min_y = GF_MAX_DOUBLE;
    m_max_y = GF_MIN_DOUBLE;
    m_min_z = GF_MAX_DOUBLE;
    m_max_z = GF_MIN_DOUBLE;

    gfDouble tx, ty, tz;

    for (gfInt i = 0;i < num_points;i++)
    {
        tx = points[i].X();
        ty = points[i].Y();
        tz = points[i].Z();

        if (tx < m_min_x)
            m_min_x = tx;
        if (tx > m_max_x)
            m_max_x = tx;

        if (ty < m_min_y)
            m_min_y = ty;
        if (ty > m_max_y)
            m_max_y = ty;

        if (tz < m_min_z)
            m_min_z = tz;
        if (tz > m_max_z)
            m_max_z = tz;
    }

    return GF_S_OK;
}

gfAuxOBB::gfAuxOBB(void)
{
    m_middle_x = 0.0;
    m_middle_y = 0.0;
    m_middle_z = 0.0;

    m_axis1_x = 0.0;
    m_axis1_y = 0.0;
    m_axis1_z = 0.0;

    m_axis2_x = 0.0;
    m_axis2_y = 0.0;
    m_axis2_z = 0.0;

    m_axis3_x = 0.0;
    m_axis3_y = 0.0;
    m_axis3_z = 0.0;
}

gfBoundVolID gfAuxOBB::GetID(void)
{
    return GF_BOUND_VOL_OBB;
}

GFRESULT gfAuxOBB::RayIntersection(gfPoint3* ray_begin, gfVector3* ray_dir, gfDouble* tmin, gfDouble* tmax)
{
    // find intersections with six boundary planes
    gfPoint3* pt = gfCreatePoint3();
    gfVector3* base1 = gfCreateVector3();
    gfVector3* base2 = gfCreateVector3();
    gfDouble u, v, t;

    pt->Set(m_middle_x - m_axis1_x - m_axis2_x - m_axis3_x,
            m_middle_y - m_axis1_y - m_axis2_y - m_axis3_y,
            m_middle_z - m_axis1_z - m_axis2_z - m_axis3_z);
    base1->Set(2.0 * m_axis1_x, 2.0 * m_axis1_y, 2.0 * m_axis1_z);
    base2->Set(2.0 * m_axis2_x, 2.0 * m_axis2_y, 2.0 * m_axis2_z);
    gfRayPlaneIntersection(ray_begin, ray_dir, pt, base1, base2, &u, &v, &t);
    if (u >= 0.0 && u <= 1.0 && v >= 0.0 && v <= 1.0)
        return GF_S_OK;

    //base1->Set(2.0 * m_axis1_x, 2.0 * m_axis1_y, 2.0 * m_axis1_z);
    base2->Set(2.0 * m_axis3_x, 2.0 * m_axis3_y, 2.0 * m_axis3_z);
    gfRayPlaneIntersection(ray_begin, ray_dir, pt, base1, base2, &u, &v, &t);
    if (u >= 0.0 && u <= 1.0 && v >= 0.0 && v <= 1.0)
        return GF_S_OK;

    base1->Set(2.0 * m_axis2_x, 2.0 * m_axis2_y, 2.0 * m_axis2_z);
    //base2->Set(2.0 * m_axis2_x, 2.0 * m_axis2_y, 2.0 * m_axis2_z);
    gfRayPlaneIntersection(ray_begin, ray_dir, pt, base1, base2, &u, &v, &t);
    if (u >= 0.0 && u <= 1.0 && v >= 0.0 && v <= 1.0)
        return GF_S_OK;

    pt->Set(m_middle_x + m_axis1_x - m_axis2_x - m_axis3_x,
            m_middle_y + m_axis1_y - m_axis2_y - m_axis3_y,
            m_middle_z + m_axis1_z - m_axis2_z - m_axis3_z);
    base1->Set(2.0 * m_axis3_x, 2.0 * m_axis3_y, 2.0 * m_axis3_z);
    base2->Set(2.0 * m_axis2_x, 2.0 * m_axis2_y, 2.0 * m_axis2_z);
    gfRayPlaneIntersection(ray_begin, ray_dir, pt, base1, base2, &u, &v, &t);
    if (u >= 0.0 && u <= 1.0 && v >= 0.0 && v <= 1.0)
        return GF_S_OK;

    pt->Set(m_middle_x - m_axis1_x - m_axis2_x + m_axis3_x,
            m_middle_y - m_axis1_y - m_axis2_y + m_axis3_y,
            m_middle_z - m_axis1_z - m_axis2_z + m_axis3_z);
    base1->Set(2.0 * m_axis1_x, 2.0 * m_axis1_y, 2.0 * m_axis1_z);
    base2->Set(2.0 * m_axis2_x, 2.0 * m_axis2_y, 2.0 * m_axis2_z);
    gfRayPlaneIntersection(ray_begin, ray_dir, pt, base1, base2, &u, &v, &t);
    if (u >= 0.0 && u <= 1.0 && v >= 0.0 && v <= 1.0)
        return GF_S_OK;

    pt->Set(m_middle_x - m_axis1_x + m_axis2_x - m_axis3_x,
            m_middle_y - m_axis1_y + m_axis2_y - m_axis3_y,
            m_middle_z - m_axis1_z + m_axis2_z - m_axis3_z);
    base1->Set(2.0 * m_axis1_x, 2.0 * m_axis1_y, 2.0 * m_axis1_z);
    base2->Set(2.0 * m_axis3_x, 2.0 * m_axis3_y, 2.0 * m_axis3_z);
    gfRayPlaneIntersection(ray_begin, ray_dir, pt, base1, base2, &u, &v, &t);
    if (u >= 0.0 && u <= 1.0 && v >= 0.0 && v <= 1.0)
        return GF_S_OK;


    gfDeletePoint3(pt);
    gfDeleteVector3(base1);
    gfDeleteVector3(base2);

    return GF_E_FAIL;
}

GFRESULT gfAuxOBB::OGLDraw(gfLong flags)
{
    gfDouble v1x = m_middle_x + m_axis1_x + m_axis2_x + m_axis3_x;
    gfDouble v1y = m_middle_y + m_axis1_y + m_axis2_y + m_axis3_y;
    gfDouble v1z = m_middle_z + m_axis1_z + m_axis2_z + m_axis3_z;

    gfDouble v2x = m_middle_x - m_axis1_x + m_axis2_x + m_axis3_x;
    gfDouble v2y = m_middle_y - m_axis1_y + m_axis2_y + m_axis3_y;
    gfDouble v2z = m_middle_z - m_axis1_z + m_axis2_z + m_axis3_z;

    gfDouble v3x = m_middle_x + m_axis1_x - m_axis2_x + m_axis3_x;
    gfDouble v3y = m_middle_y + m_axis1_y - m_axis2_y + m_axis3_y;
    gfDouble v3z = m_middle_z + m_axis1_z - m_axis2_z + m_axis3_z;

    gfDouble v4x = m_middle_x + m_axis1_x + m_axis2_x - m_axis3_x;
    gfDouble v4y = m_middle_y + m_axis1_y + m_axis2_y - m_axis3_y;
    gfDouble v4z = m_middle_z + m_axis1_z + m_axis2_z - m_axis3_z;

    gfDouble v5x = m_middle_x - m_axis1_x - m_axis2_x + m_axis3_x;
    gfDouble v5y = m_middle_y - m_axis1_y - m_axis2_y + m_axis3_y;
    gfDouble v5z = m_middle_z - m_axis1_z - m_axis2_z + m_axis3_z;

    gfDouble v6x = m_middle_x + m_axis1_x - m_axis2_x - m_axis3_x;
    gfDouble v6y = m_middle_y + m_axis1_y - m_axis2_y - m_axis3_y;
    gfDouble v6z = m_middle_z + m_axis1_z - m_axis2_z - m_axis3_z;

    gfDouble v7x = m_middle_x - m_axis1_x + m_axis2_x - m_axis3_x;
    gfDouble v7y = m_middle_y - m_axis1_y + m_axis2_y - m_axis3_y;
    gfDouble v7z = m_middle_z - m_axis1_z + m_axis2_z - m_axis3_z;

    gfDouble v8x = m_middle_x - m_axis1_x - m_axis2_x - m_axis3_x;
    gfDouble v8y = m_middle_y - m_axis1_y - m_axis2_y - m_axis3_y;
    gfDouble v8z = m_middle_z - m_axis1_z - m_axis2_z - m_axis3_z;

    glBegin(GL_LINES);

    glVertex3f(v1x, v1y, v1z);
    glVertex3f(v2x, v2y, v2z);

    glVertex3f(v1x, v1y, v1z);
    glVertex3f(v3x, v3y, v3z);

    glVertex3f(v1x, v1y, v1z);
    glVertex3f(v4x, v4y, v4z);

    glVertex3f(v5x, v5y, v5z);
    glVertex3f(v2x, v2y, v2z);

    glVertex3f(v7x, v7y, v7z);
    glVertex3f(v2x, v2y, v2z);

    glVertex3f(v3x, v3y, v3z);
    glVertex3f(v5x, v5y, v5z);

    glVertex3f(v3x, v3y, v3z);
    glVertex3f(v6x, v6y, v6z);

    glVertex3f(v4x, v4y, v4z);
    glVertex3f(v6x, v6y, v6z);

    glVertex3f(v4x, v4y, v4z);
    glVertex3f(v7x, v7y, v7z);

    glVertex3f(v5x, v5y, v5z);
    glVertex3f(v8x, v8y, v8z);

    glVertex3f(v6x, v6y, v6z);
    glVertex3f(v8x, v8y, v8z);

    glVertex3f(v7x, v7y, v7z);
    glVertex3f(v8x, v8y, v8z);

    glEnd();

    return GF_S_OK;
}

GFRESULT gfAuxOBB::Transform(gfMatrix4* matrix)
{
    gfPoint3 pt(m_middle_x, m_middle_y, m_middle_z);
    gfTransformPoint3(&pt, matrix);
    m_middle_x = pt.X();
    m_middle_y = pt.Y();
    m_middle_z = pt.Z();

    gfVector3 vector(m_axis1_x, m_axis1_y, m_axis1_z);
    gfTransformVector3(&vector, matrix);
    m_axis1_x = vector.X();
    m_axis1_y = vector.Y();
    m_axis1_z = vector.Z();

    vector.Set(m_axis2_x, m_axis2_y, m_axis2_z);
    gfTransformVector3(&vector, matrix);
    m_axis2_x = vector.X();
    m_axis2_y = vector.Y();
    m_axis2_z = vector.Z();

    vector.Set(m_axis3_x, m_axis3_y, m_axis3_z);
    gfTransformVector3(&vector, matrix);
    m_axis3_x = vector.X();
    m_axis3_y = vector.Y();
    m_axis3_z = vector.Z();

    return GF_S_OK;
}

gfPoint3* gfAuxOBB::GetCenter(void)
{
    gfPoint3* ret = gfCreatePoint3();
    ret->Set(m_middle_x, m_middle_y, m_middle_z);
    return ret;
}

GFRESULT gfAuxOBB::InitFromPoints(gfInt num_points, gfPoint3* points)
{
    if (num_points < 1)
        return GF_E_BADARG;

    // compute mean value of points
    gfDouble nx = 0.0;
    gfDouble ny = 0.0;
    gfDouble nz = 0.0;

    for (gfInt i = 0; i < num_points; i++)
    {
        nx += points[i].X();
        ny += points[i].Y();
        nz += points[i].Z();
    }

    nx /= num_points;
    ny /= num_points;
    nz /= num_points;

    // compute covariance matrix of points
    gfDouble c11 = 0.0;
    gfDouble c12 = 0.0;
    gfDouble c13 = 0.0;
    gfDouble c21 = 0.0;
    gfDouble c22 = 0.0;
    gfDouble c23 = 0.0;
    gfDouble c31 = 0.0;
    gfDouble c32 = 0.0;
    gfDouble c33 = 0.0;

    for (gfInt i = 0; i < num_points; i++)
    {
        c11 += (points[i].X() - nx) * (points[i].X() - nx);
        c12 += (points[i].X() - nx) * (points[i].Y() - ny);
        c13 += (points[i].X() - nx) * (points[i].Z() - nz);
        c21 += (points[i].Y() - ny) * (points[i].X() - nx);
        c22 += (points[i].Y() - ny) * (points[i].Y() - ny);
        c23 += (points[i].Y() - ny) * (points[i].Z() - nz);
        c31 += (points[i].Z() - nz) * (points[i].X() - nx);
        c32 += (points[i].Z() - nz) * (points[i].Y() - ny);
        c33 += (points[i].Z() - nz) * (points[i].Z() - nz);
    }

    c11 /= num_points;
    c12 /= num_points;
    c13 /= num_points;
    c21 /= num_points;
    c22 /= num_points;
    c23 /= num_points;
    c31 /= num_points;
    c32 /= num_points;
    c33 /= num_points;

    gfMatrix3* cov_matrix = gfCreateMatrix3();
    cov_matrix->Set(c11, c12, c13,
        c21, c22, c23,
        c31, c32, c33);
    gfMatrix3* e_vectors = gfCreateMatrix3();
    gfVector3* e_values = gfCreateVector3();
    gfInt rot;
    gfGetEigensSym3x3(cov_matrix, e_values, e_vectors, &rot);

    // we have three orthogonal eigenvectors as columns in e_vectors matrix
    // find min max points in that directions

    gfDouble a1_min_t = GF_MAX_DOUBLE;
    gfDouble a1_max_t = GF_MIN_DOUBLE;
    gfDouble a2_min_t = GF_MAX_DOUBLE;
    gfDouble a2_max_t = GF_MIN_DOUBLE;
    gfDouble a3_min_t = GF_MAX_DOUBLE;
    gfDouble a3_max_t = GF_MIN_DOUBLE;

    gfDouble t;
    for (gfInt i = 0; i < num_points; i++)
    {
        // projection of point on axis1
        t = (points[i].X() - nx) * e_vectors->at(1, 1) +
            (points[i].Y() - ny) * e_vectors->at(2, 1) + 
            (points[i].Z() - nz) * e_vectors->at(3, 1);
        if (t < a1_min_t)
            a1_min_t = t;
        if (t > a1_max_t)
            a1_max_t = t;

        // projection of point on axis2
        t = (points[i].X() - nx) * e_vectors->at(1, 2) +
            (points[i].Y() - ny) * e_vectors->at(2, 2) + 
            (points[i].Z() - nz) * e_vectors->at(3, 2);
        if (t < a2_min_t)
            a2_min_t = t;
        if (t > a2_max_t)
            a2_max_t = t;

        // projection of point on axis3
        t = (points[i].X() - nx) * e_vectors->at(1, 3) +
            (points[i].Y() - ny) * e_vectors->at(2, 3) + 
            (points[i].Z() - nz) * e_vectors->at(3, 3);
        if (t < a3_min_t)
            a3_min_t = t;
        if (t > a3_max_t)
            a3_max_t = t;
    }

    // compute middle of OBB and scale axis 1
    m_middle_x = nx + e_vectors->at(1, 1) * (a1_min_t + a1_max_t) / 2.0;
    m_middle_y = ny + e_vectors->at(2, 1) * (a1_min_t + a1_max_t) / 2.0;
    m_middle_z = nz + e_vectors->at(3, 1) * (a1_min_t + a1_max_t) / 2.0;

    m_axis1_x = e_vectors->at(1, 1) * (a1_max_t - a1_min_t) / 2.0;
    m_axis1_y = e_vectors->at(2, 1) * (a1_max_t - a1_min_t) / 2.0;
    m_axis1_z = e_vectors->at(3, 1) * (a1_max_t - a1_min_t) / 2.0;

    // by axis 2
    m_middle_x += e_vectors->at(1, 2) * (a2_min_t + a2_max_t) / 2.0;
    m_middle_y += e_vectors->at(2, 2) * (a2_min_t + a2_max_t) / 2.0;
    m_middle_z += e_vectors->at(3, 2) * (a2_min_t + a2_max_t) / 2.0;

    m_axis2_x = e_vectors->at(1, 2) * (a2_max_t - a2_min_t) / 2.0;
    m_axis2_y = e_vectors->at(2, 2) * (a2_max_t - a2_min_t) / 2.0;
    m_axis2_z = e_vectors->at(3, 2) * (a2_max_t - a2_min_t) / 2.0;

    // by axis 3
    m_middle_x += e_vectors->at(1, 3) * (a3_min_t + a3_max_t) / 2.0;
    m_middle_y += e_vectors->at(2, 3) * (a3_min_t + a3_max_t) / 2.0;
    m_middle_z += e_vectors->at(3, 3) * (a3_min_t + a3_max_t) / 2.0;

    m_axis3_x = e_vectors->at(1, 3) * (a3_max_t - a3_min_t) / 2.0;
    m_axis3_y = e_vectors->at(2, 3) * (a3_max_t - a3_min_t) / 2.0;
    m_axis3_z = e_vectors->at(3, 3) * (a3_max_t - a3_min_t) / 2.0;

    // delete used objects
    gfDeleteMatrix3(cov_matrix);
    gfDeleteMatrix3(e_vectors);
    gfDeleteVector3(e_values);

    return GF_S_OK;
}
