//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfHalfEdge.cpp                                       //
//   Description: Half-edge data structure implementation       //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "StdAfx.h"
#include "gfAuxHalfEdge.h"
#include "gfAuxNumUtils.h"
#include "gfObjectFactory.h"


gfHalfEdge* gfCreateHalfEdge(void)
{
    gfHalfEdge* edge = new gfHalfEdge;
    if (edge)
        memset(edge, 0, sizeof(gfHalfEdge));

    return edge;
}

GFRESULT gfDeleteHalfEdge(gfHalfEdge* edge)
{
    if (edge)
        delete edge;

    return GF_S_OK;
}

gfHEFace* gfCreateHEFace(void)
{
    gfHEFace* face = new gfHEFace;
    if (face)
        memset(face, 0, sizeof(gfHEFace));

    return face;
}

GFRESULT gfDeleteHEFace(gfHEFace* face)
{
    if (face)
        delete face;

    return GF_S_OK;
}

gfHEPoint* gfCreateHEPoint(void)
{
    gfHEPoint* point = new gfHEPoint;
    if (point)
        memset(point, 0, sizeof(gfHEPoint));

    return point;
}

GFRESULT gfDeleteHEPoint(gfHEPoint* point)
{
    if (point)
        delete point;

    return GF_S_OK;
}

gfHENormal* gfCreateHENormal(void)
{
    gfHENormal* normal = new gfHENormal;
    if (normal)
        memset(normal, 0, sizeof(gfHENormal));

    return normal;
}

GFRESULT gfDeleteHENormal(gfHENormal* normal)
{
    if (normal)
        delete normal;

    return GF_S_OK;
}

gfHETexCoords* gfCreateHETexCoords(void)
{
    gfHETexCoords* tex_coords = new gfHETexCoords;
    if (tex_coords)
        memset(tex_coords, 0, sizeof(gfHETexCoords));

    return tex_coords;
}

GFRESULT gfDeleteHETexCoords(gfHETexCoords* tex_coords)
{
    if (tex_coords)
        delete tex_coords;

    return GF_S_OK;
}

gfHalfEdge* gfGetPreviousEdge(gfHalfEdge* edge)
{
    gfHalfEdge* next_edge = edge->next;
    gfHalfEdge* prev_edge = edge;
    while (edge != next_edge)
    {
        prev_edge = next_edge;
        next_edge = next_edge->next;
    }

    return prev_edge;
}

gfPoint3* gfGetEdgeMiddlePoint(gfHalfEdge* edge)
{
    gfPoint3* point = gfCreatePoint3();

    // end point
    gfDouble ax = edge->end->x;
    gfDouble ay = edge->end->y;
    gfDouble az = edge->end->z;

    // find start point
    gfHalfEdge* prev_edge = gfGetPreviousEdge(edge);

    ax += prev_edge->end->x;
    ay += prev_edge->end->y;
    az += prev_edge->end->z;

    point->Set(ax / 2.0, ay / 2.0, az / 2.0);
    return point;
}

gfHENormal* gfGetHEFaceNormal(gfHEFace* face)
{
    if (!face)
        return NULL;
    gfHalfEdge* edge = face->edge;
    if (!edge)
        return NULL;
    gfHalfEdge* next_edge = edge->next;

    gfHENormal* normal = gfCreateHENormal();

    normal->x = -(next_edge->end->y - edge->end->y) * (next_edge->end->z + edge->end->z);
    normal->y = -(next_edge->end->z - edge->end->z) * (next_edge->end->x + edge->end->x);
    normal->z = -(next_edge->end->x - edge->end->x) * (next_edge->end->y + edge->end->y);

    // find all edges of face
    while (next_edge != edge)
    {   
        gfHEPoint *first = next_edge->end;
        gfHEPoint *second = next_edge->next->end;
        
        normal->x += -(second->y - first->y) * (second->z + first->z);
        normal->y += -(second->z - first->z) * (second->x + first->x);
        normal->z += -(second->x - first->x) * (second->y + first->y);

        next_edge = next_edge->next;
    }

    gfHENormalNormalize(normal);
    return normal;
}

GFRESULT gfHENormalNormalize(gfHENormal* normal)
{
    if (!normal)
        return GF_E_FAIL;
    gfDouble d = gfSquareRoot(normal->x*normal->x + normal->y*normal->y + normal->z*normal->z);
    if (!gfIsZero(d))
    {
        normal->x /= d;
        normal->y /= d;
        normal->z /= d;
    }

    return GF_S_OK;
}

gfHENormal* gfGetHEVertexNormal(gfHEPoint* vertex)
{
    std::vector<gfHEFace*> neigh_faces;
    gfGetAllHEPointFaces(vertex, &neigh_faces);

    gfHENormal* normal = gfCreateHENormal();
    normal->x = normal->y = normal->z = 0.0;

    gfInt num = (gfInt)neigh_faces.size();
    for (gfInt i = 0;i < num;i++)
    {
        gfHEFace* face = neigh_faces.at(i);
        gfHENormal* n = gfGetHEFaceNormal(face);
        normal->x += n->x;
        normal->y += n->y;
        normal->z += n->z;
        gfDeleteHENormal(n);
    }

    gfHENormalNormalize(normal);
    return normal;
}

gfHEPoint* gfGetCCFacePoint(gfHEFace* face)
{
    // face point is average of all points of face
    // create HE point
    gfHEPoint* new_point = gfCreateHEPoint();

    gfInt num_edges = 1;
    gfHalfEdge* start_edge = face->edge;
    gfHalfEdge* next_edge = start_edge->next;
    gfPoint3 aux_pt(start_edge->end->x, start_edge->end->y, start_edge->end->z);

    // sum all points of face
    while (start_edge != next_edge)
    {
        aux_pt.Set(aux_pt.X() + next_edge->end->x, aux_pt.Y() + next_edge->end->y, aux_pt.Z() + next_edge->end->z);
        next_edge = next_edge->next;
        num_edges++;
    }

    new_point->x = aux_pt.X() / num_edges;
    new_point->y = aux_pt.Y() / num_edges;
    new_point->z = aux_pt.Z() / num_edges;

    return new_point;
}

gfHEPoint* gfGetCCEdgePoint(gfHalfEdge* edge)
{
    // edge point is average neighboured HE points and face points
    // create HE point
    gfHEPoint* new_point = gfCreateHEPoint();

    // end point
    gfDouble ax = edge->end->x;
    gfDouble ay = edge->end->y;
    gfDouble az = edge->end->z;

    // find start point
    gfHalfEdge* next_edge = edge->next;
    gfHalfEdge* prev_edge = edge;
    while (edge != next_edge)
    {
        prev_edge = next_edge;
        next_edge = next_edge->next;
    }

    ax += prev_edge->end->x;
    ay += prev_edge->end->y;
    az += prev_edge->end->z;

    // is it boundary edge ?
    if (edge->opp)
    {
        gfHEPoint* face1_point = (gfHEPoint*) edge->face->data;
        ax += face1_point->x;
        ay += face1_point->y;
        az += face1_point->z;

        gfHEPoint* face2_point = (gfHEPoint*) edge->opp->face->data;
        ax += face2_point->x;
        ay += face2_point->y;
        az += face2_point->z;

        ax /= 4.0;
        ay /= 4.0;
        az /= 4.0;
    }
    else
    {
        ax /= 2.0;
        ay /= 2.0;
        az /= 2.0;
    }
    
    new_point->x = ax;
    new_point->y = ay;
    new_point->z = az;
    return new_point;
}

gfHEPoint* gfGetCCMovedPoint(gfHEPoint* point)
{
    gfHEPoint* new_point = gfCreateHEPoint();

    gfHalfEdge* first_edge = point->edge;
    if (!first_edge)
    {
        new_point->x = point->x;
        new_point->y = point->y;
        new_point->z = point->z;
        return new_point;
    }
    gfHalfEdge* next_edge;
    gfBool boundary = false;

    // for average of face points
    gfInt num_faces = 1;
    gfHEPoint* face_point = (gfHEPoint*) first_edge->face->data; 
    gfDouble qx = face_point->x;
    gfDouble qy = face_point->y;
    gfDouble qz = face_point->z;

    // for average of edge middles
    gfInt num_edges = 1;
    gfPoint3 * middle = gfGetEdgeMiddlePoint(first_edge);
    gfDouble rx = middle->X();
    gfDouble ry = middle->Y();
    gfDouble rz = middle->Z();
    gfDeletePoint3(middle);

    if (first_edge->next->opp)
    {
        next_edge = first_edge->next->opp;
    }
    else
    {
        next_edge = first_edge;
        boundary = true;
    }

    // first pass in one direction
    while (first_edge != next_edge && !boundary)
    {
        //num_edges++;
        num_faces++;

        // find middle of this edge
        gfPoint3 * middle = gfGetEdgeMiddlePoint(next_edge);
        rx += middle->X();
        ry += middle->Y();
        rz += middle->Z();
        gfDeletePoint3(middle);

        // add face point
        gfHEPoint* face_point = (gfHEPoint*) next_edge->face->data; 
        qx += face_point->x;
        qy += face_point->y;
        qz += face_point->z;

        if (next_edge->next->opp)
        {
            next_edge = next_edge->next->opp;
        }
        else
        {
            boundary = true;
        }

        // hack, for meshes with damaged topology
        if (num_faces > 40)
            break;
    }

    // is point on bounary?
    if (boundary && first_edge->opp)
    {
        next_edge = gfGetPreviousEdge(first_edge->opp);
        gfBool end = false;

        while (first_edge != next_edge && !end)
        {
            num_faces++;

            // find middle of this edge
            gfPoint3 * middle = gfGetEdgeMiddlePoint(next_edge);
            rx += middle->X();
            ry += middle->Y();
            rz += middle->Z();
            gfDeletePoint3(middle);

            // add face point
            gfHEPoint* face_point = (gfHEPoint*) next_edge->face->data; 
            qx += face_point->x;
            qy += face_point->y;
            qz += face_point->z;

            if (next_edge->opp)
                next_edge = gfGetPreviousEdge(next_edge->opp);
            else
                end = true;
        }
    }

    // set number of incident edges 
    if (boundary)
        num_edges = num_faces + 1;
    else
        num_edges = num_faces;

    // compute new position
    gfDouble ax = qx / num_faces / num_edges + 2.0 * rx / num_edges / num_edges + (num_edges - 3) * point->x / num_edges;
    gfDouble ay = qy / num_faces / num_edges + 2.0 * ry / num_edges / num_edges + (num_edges - 3) * point->y / num_edges;
    gfDouble az = qz / num_faces / num_edges + 2.0 * rz / num_edges / num_edges + (num_edges - 3) * point->z / num_edges;

    new_point->x = ax;
    new_point->y = ay;
    new_point->z = az;

    return new_point;
}

GFRESULT gfGetAllHEPointFaces(gfHEPoint* point, std::vector<gfHEFace*>* faces)
{

    gfHalfEdge* in_edge = point->edge;
    std::vector<gfHEFace*> work_faces;
    while (work_faces.size() > 0)
    {
        gfHEFace* face = work_faces.at(0);
        faces->push_back(face);
        work_faces.erase(0);

        // look in neighbouring faces for our given point
        gfHalfEdge* first_edge = face->edge;
        gfHalfEdge* edge = first_edge;
        gfBool first = true;
        while (edge != first_edge || first)
        {
            first = false;
            if (edge->opp)
            {
                gfHEFace* neigh_face = edge->opp->face;

                // search for point in neighbouring face
                gfBool face_has_point = false;
                gfHalfEdge* first_edge_neigh = neigh_face->edge;
                if (first_edge_neigh->end == point)
                    face_has_point = true;
                gfHalfEdge* edge_neigh = first_edge_neigh->next;
                
                while ((edge_neigh != first_edge_neigh) && !face_has_point)
                {
                    if (edge_neigh->end == point)
                        face_has_point = true;
                    edge_neigh = edge_neigh->next;
                }

                // if point was found in face, check if face wasn't already processed 
                if (face_has_point)
                {
                    gfInt num = (gfInt)faces->size();
                    gfBool add = true;
                    for (gfInt i = 0;i < num;i++)
                        if (faces->at(i) == neigh_face)
                            add = false;
                    if (add)
                        work_faces.push_back(neigh_face);
                }
            }
            edge = edge->next;
        }
    }

    return GF_S_OK;
}

void gfSetHEPointCoords(gfHEPoint* point, gfDouble x, gfDouble y, gfDouble z)
{
    point->x = x;
    point->y = y;
    point->z = z;
}
