//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfHalfEdge.h                                         //
//   Description: Half-edge data structure implementation       //
//                                                              //
//////////////////////////////////////////////////////////////////


#ifndef _GF_HALF_EDGE_H_
#define _GF_HALF_EDGE_H_


#include "gfDefines.h"
#include "gfPoint.h"
#include "gfVector.h"
#include <vector>

struct gfHEPoint;
struct gfHEFace;
struct gfHENormal;
struct gfHETexCoords;

struct gfHalfEdge
{
    // end vertex of half-edge
    gfHEPoint* end;
    // normal in end vertex
    gfHENormal* normal;
    // texture coordinates in end vertex
    gfHETexCoords* tex_coord;

    // opposite half-edge
    gfHalfEdge* opp;
    // next half-edge binded to  face
    gfHalfEdge* next;
    // face that half-edge is binded to
    gfHEFace* face;

    // auxiliary data, for inner computings
    void* data;
};

struct gfHEFace
{
    // one half edge around face
    gfHalfEdge* edge;
    // auxiliary data, for inner computings
    void* data;
};

struct gfHEPoint
{
    // coordinates of point
    gfDouble x, y, z;
    // one half edge going to point
    gfHalfEdge* edge;
    // curvature in point
    gfDouble curvature;
    // auxiliary data, for inner computings
    void* data;
};

struct gfHENormal
{
    // coordinates of point
    gfDouble x, y, z;
    // auxiliary data, for inner computings
    void* data;
};

struct gfHETexCoords
{
    // coordinates of point
    gfDouble u, v, w;
    // auxiliary data, for inner computings
    void* data;
};


gfHalfEdge* gfCreateHalfEdge(void);
GFRESULT gfDeleteHalfEdge(gfHalfEdge* edge);
gfHEFace* gfCreateHEFace(void);
GFRESULT gfDeleteHEFace(gfHEFace* face);
gfHEPoint* gfCreateHEPoint(void);
GFRESULT gfDeleteHEPoint(gfHEPoint* edge);
gfHENormal* gfCreateHENormal(void);
GFRESULT gfDeleteHENormal(gfHENormal* normal);
gfHETexCoords* gfCreateHETexCoords(void);
GFRESULT gfDeleteHETexCoords(gfHETexCoords* tex_coords);

gfHalfEdge* gfGetPreviousEdge(gfHalfEdge* edge);
gfPoint3 * gfGetEdgeMiddlePoint(gfHalfEdge* edge);
gfHENormal* gfGetHEFaceNormal(gfHEFace* face);
gfHENormal* gfGetHEVertexNormal(gfHEPoint* vertex);
gfHEPoint* gfGetCCFacePoint(gfHEFace* face);
gfHEPoint* gfGetCCEdgePoint(gfHalfEdge* edge);
gfHEPoint* gfGetCCMovedPoint(gfHEPoint* point);
GFRESULT gfHENormalNormalize(gfHENormal* normal);
GFRESULT gfGetAllHEPointFaces(gfHEPoint* point, std::vector<gfHEFace*>* faces);

void gfSetHEPointCoords(gfHEPoint* point, gfDouble x, gfDouble y, gfDouble z);


#endif