//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfAuxIntersection.cpp                                //
//   Description: Utilities for computing intersections         //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "gfAuxItersectionUtils.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfObjectFactory.h"


gfPoint3* gfLinePlaneIntersection(gfPoint3* line_pt1, gfPoint3* line_pt2, gfPoint3* plane_pt, gfVector3* plane_normal)
{
    gfDouble d = - (plane_normal->X() * plane_pt->X() + plane_normal->Y() * plane_pt->Y() + plane_normal->Z() * plane_pt->Z());
    gfDouble denom = (line_pt2->X() - line_pt1->X()) * plane_normal->X() +
                     (line_pt2->Y() - line_pt1->Y()) * plane_normal->Y() +
                     (line_pt2->Z() - line_pt1->Z()) * plane_normal->Z();
    if (gfIsZero(denom))
    {
        // line is paraller with plane
        return 0;
    }

    gfPoint3* inter = gfCreatePoint3();
    gfDouble t = ( - d - line_pt1->X() * plane_normal->X() -
                         line_pt1->Y() * plane_normal->Y() - 
                         line_pt1->Z() * plane_normal->Z()) / denom;
    inter->Set(t * line_pt2->X() + (1 - t) * line_pt1->X(),
               t * line_pt2->Y() + (1 - t) * line_pt1->Y(),
               t * line_pt2->Z() + (1 - t) * line_pt1->Z());

    return inter;
}

GFRESULT gfRayPlaneIntersection(gfPoint3* ray_start, gfVector3* ray_dir, gfPoint3* plane_pt, gfVector3* plane_normal, gfDouble* t)
{
    gfDouble d = - (plane_normal->X() * plane_pt->X() + plane_normal->Y() * plane_pt->Y() + plane_normal->Z() * plane_pt->Z());
    gfDouble denom = ray_dir->X() * plane_normal->X() +
                     ray_dir->Y() * plane_normal->Y() +
                     ray_dir->Z() * plane_normal->Z();
    if (gfIsZero(denom))
    {
        return GF_E_FAIL;
    }

    *t = (- d - ray_start->X() * plane_normal->X() -
                ray_start->Y() * plane_normal->Y() - 
                ray_start->Z() * plane_normal->Z()) / denom;
    if (*t < GF_EPS)
    {
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfRayPlaneIntersection(gfPoint3* ray_start, gfVector3* ray_dir, 
                                gfPoint3* plane_pt, gfVector3* plane_base1, gfVector3* plane_base2,
                                gfDouble* ret_u, gfDouble* ret_v, gfDouble* ret_t)
{
    gfDouble det;

    gfDouble pvectx = ray_dir->Y() * plane_base2->Z() - ray_dir->Z() * plane_base2->Y();
    gfDouble pvecty = - ray_dir->X() * plane_base2->Z() + ray_dir->Z() * plane_base2->X();
    gfDouble pvectz = ray_dir->X() * plane_base2->Y() - ray_dir->Y() * plane_base2->X();

    det = pvectx * plane_base1->X() + pvecty * plane_base1->Y() + pvectz * plane_base1->Z();
    if (gfIsZero(det))
        return GF_E_FAIL;

    gfDouble tvectx = ray_start->X() - plane_pt->X();
    gfDouble tvecty = ray_start->Y() - plane_pt->Y();
    gfDouble tvectz = ray_start->Z() - plane_pt->Z();

    *ret_u = (tvectx * pvectx + tvecty * pvecty + tvectz * pvectz) / det;

    gfDouble qvectx = tvecty * plane_base1->Z() - tvectz * plane_base1->Y();
    gfDouble qvecty = - tvectx * plane_base1->Z() + tvectz * plane_base1->X();
    gfDouble qvectz = tvectx * plane_base1->Y() - tvecty * plane_base1->X();

    *ret_v = (ray_dir->X() * qvectx + ray_dir->Y() * qvecty + ray_dir->Z() * qvectz) / det;

    // check if intersection is on good side of ray
    *ret_t = (plane_pt->X() + (*ret_u) *plane_base1->X() + (*ret_v) * plane_base2->X() - ray_start->X()) * ray_dir->X() +
             (plane_pt->Y() + (*ret_u) *plane_base1->Y() + (*ret_v) * plane_base2->Y() - ray_start->Y()) * ray_dir->Y() +
             (plane_pt->Z() + (*ret_u) *plane_base1->Z() + (*ret_v) * plane_base2->Z() - ray_start->Z()) * ray_dir->Z();

    if (*ret_t > GF_EPS)
        return GF_S_OK;
    else
        return GF_E_FAIL;
}

GFRESULT gfRayTriangleInter(gfPoint3* ray_begin, gfVector3* ray_dir, gfPoint3* pt1, gfPoint3* pt2, gfPoint3* pt3, gfDouble* t)
{
    gfVector3 base1, base2;
    gfSubstractPoints3(pt1, pt3, &base1);
    gfSubstractPoints3(pt2, pt3, &base2);

    gfDouble u, v;
    GFRESULT res = gfRayPlaneIntersection(ray_begin, ray_dir, pt3, &base1, &base2, &u, &v, t);

    if (GF_FAIL(res) || u < -GF_EPS || v < -GF_EPS || (u + v) > 1.0+GF_EPS)
    { 
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfRayTriangleInter(gfPoint3* ray_begin, gfVector3* ray_dir, gfPoint3* pt1, gfPoint3* pt2, gfPoint3* pt3,
                            gfDouble* u, gfDouble* v, gfDouble* t)
{
    gfVector3 base1, base2;
    gfSubstractPoints3(pt1, pt3, &base1);
    gfSubstractPoints3(pt2, pt3, &base2);

    GFRESULT res = gfRayPlaneIntersection(ray_begin, ray_dir, pt3, &base1, &base2, u, v, t);

    if (GF_FAIL(res) || *u < -GF_EPS || *v < -GF_EPS || (*u + *v) > 1.0+GF_EPS)
    {
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfRaySegmentInter(gfPoint3* ray_begin, gfVector3* ray_dir, gfPoint3* seg1, gfPoint3* seg2, gfDouble* t)
{
    gfDouble ux = ray_dir->X();
    gfDouble uy = ray_dir->Y();
    gfDouble uz = ray_dir->Z();
    gfDouble vx = seg2->X() - seg1->X();
    gfDouble vy = seg2->Y() - seg1->Y();
    gfDouble vz = seg2->Z() - seg1->Z();

    gfDouble ax = ray_begin->X();
    gfDouble ay = ray_begin->Y();
    gfDouble az = ray_begin->Z();
    gfDouble bx = seg1->X();
    gfDouble by = seg1->Y();
    gfDouble bz = seg1->Z();

    gfDouble A = (ax-bx)*ux+(ay-by)*uy+(az-bz)*uz;
    gfDouble B = ux*ux+uy*uy+uz*uz;
    gfDouble C = ux*vx+uy*vy+uz*vz;
    gfDouble D = (ax-bx)*vx+(ay-by)*vy+(az-bz)*vz;
    gfDouble E = vx*vx+vy*vy+vz*vz;

    gfDouble s = (A*C-B*D)/(C*C-E*B);
    gfDouble tt = (A*E-C*D)/(C*C-E*B);

    if (tt <= 0.0)
        return GF_E_FAIL;

    gfDouble res1x = ax+tt*ux;
    gfDouble res1y = ay+tt*uy;
    gfDouble res1z = az+tt*uz;
    gfDouble res2x = bx+s*vx;
    gfDouble res2y = by+s*vy;
    gfDouble res2z = bz+s*vz;

    gfDouble dist = (res1x-res2x)*(res1x-res2x)+(res1y-res2y)*(res1y-res2y)+(res1z-res2z)*(res1z-res2z);
    if (dist > 0.001)
        return GF_E_FAIL;

    *t = tt;

    return GF_S_OK;
}

GFRESULT gfAxisXConvexHullInter2D(std::vector<gfPoint2>* points, gfDouble* minx, gfDouble* maxx)
{
    if (!points)
        return GF_E_BADARG;
    std::vector<gfDouble> intersections;

    gfInt num = (gfInt)points->size();

    *minx = GF_MAX_DOUBLE;
    *maxx = GF_MIN_DOUBLE;
    gfDouble x;
    gfBool inter = false;
    for (gfInt i = 0;i < num;i++)
    {
        gfPoint2 pt1 = points->at(i);
        if (gfAbs(points->at(i).Y()) < GF_EPS)
        {
            x = points->at(i).X();
            inter = true;
            if (x < *minx)
                *minx = x;
            if (x > *maxx)
                *maxx = x;
            continue;
        }
        for (gfInt j = i+1;j < num;j++)
        {
            gfPoint2 pt2 = points->at(j);
            if (gfAbs(points->at(j).Y()) < GF_EPS)
            {
                x = points->at(j).X();
                inter = true;
                if (x < *minx)
                    *minx = x;
                if (x > *maxx)
                    *maxx = x;
                continue;
            }

            if (points->at(i).Y()*points->at(j).Y() < 0)
            {
                x = points->at(i).X()-points->at(i).Y()*(points->at(j).X()-points->at(i).X())/(points->at(j).Y()-points->at(i).Y());
                inter = true;
                if (x < *minx)
                    *minx = x;
                if (x > *maxx)
                    *maxx = x;
            }
        }
    }

    if (!inter)
        return GF_E_FAIL;

    return GF_S_OK;
}

GFRESULT gfLinesIntersection(gfPoint3* pt_line1, gfVector3* dir_line1, gfPoint3* pt_line2, gfVector3* dir_line2, gfPoint3* result)
{
    gfDouble rx = pt_line1->X() - pt_line2->X();
    gfDouble ry = pt_line1->Y() - pt_line2->Y();
    gfDouble p1x = -dir_line1->X();
    gfDouble p1y = -dir_line1->Y();
    gfDouble p2x = dir_line2->X();
    gfDouble p2y = dir_line2->Y();

    gfDouble denom = p1x*p2y - p1y*p2x;
    if (gfIsZero(denom))
        return GF_E_FAIL;

    gfDouble s = (rx*p2y-ry*p2x) / denom;
    gfDouble t = (rx*p1y-ry*p1x) / denom;

    if (!gfIsZero(pt_line1->Z()-pt_line2->Z()+s*dir_line1->Z()-t*dir_line2->Z()))
        return GF_E_FAIL;

    result->SetX(pt_line1->X()+s*dir_line1->X());
    result->SetY(pt_line1->Y()+s*dir_line1->Y());
    result->SetZ(pt_line1->Z()+s*dir_line1->Z());    

    return GF_S_OK;
}

