//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfAuxRaytracer.cpp                                   //
//   Description: Class for computing raytraced scenes          //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "gfAuxRaytracer.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfAuxLogging.h"
#include "gfWinFPSCounter.h"
#include "gfObjectFactory.h"
#include "gfObjNURBSPatch.h"


gfAuxRaytracer::gfAuxRaytracer(void)
{
    m_pic_data = NULL;
    m_width = 0;
    m_height = 0;
    m_shadows = true;
    m_interleaved = true;
    m_aa = GF_AA_ADAPTIVE;
}

gfAuxRaytracer::~gfAuxRaytracer(void)
{
    if (m_pic_data)
        delete [] m_pic_data;
    m_pic_data = NULL;
}

GFRESULT gfAuxRaytracer::GetLocalLighting(std::vector<gfScnLight*>* vis_lights, gfScnMaterial* material, gfPoint3* pos, 
                          gfVector3* view_dir, gfVector2* uv, gfVector3* normal, gfVector3* res_color)
{
    // phong shader

    // set global ambient
    gfDouble Ir = material->m_ambient_r * m_scene->m_global_ambient_r;
    gfDouble Ig = material->m_ambient_g * m_scene->m_global_ambient_g;
    gfDouble Ib = material->m_ambient_b * m_scene->m_global_ambient_b;

    if (gfScalarProduct3(view_dir, normal) < 0.0)
    {
        //return GF_S_OK;
    }

    // get ambient from all lights
    std::vector<gfScnLight*> all_lights;
    m_scene->GetSceneLights(&all_lights);
    gfInt num_all_lights = (gfInt)all_lights.size();
    for (gfInt i = 0;i < num_all_lights;i++)
    {
        gfScnLight* light = all_lights.at(i);
        Ir += material->m_ambient_r * light->m_ambient_r;
        Ig += material->m_ambient_g * light->m_ambient_g;
        Ib += material->m_ambient_b * light->m_ambient_b;
    }

    // go throught all visible lights
    gfInt num_lights = (gfInt)vis_lights->size();
    for (gfInt i = 0;i < num_lights;i++)
    {
        gfScnLight* light = vis_lights->at(i);
        
        // compute light vector
        gfVector3 L(light->m_pos_x - pos->X(), light->m_pos_y - pos->Y(), light->m_pos_z - pos->Z());
        gfDouble dist = L.Length();
        L.Normalize();
        // compute (normal.L)
        gfDouble NdotL = gfScalarProduct3(normal, &L);

        // add specular part
        if (NdotL > 0.0)
        {
            // compute attenuation and spotlight factor
            gfVector3 spot(light->m_spot_dir_x, light->m_spot_dir_y, light->m_spot_dir_z);
            spot.Normalize();
            gfDouble att = gfScalarProduct3(&spot, &L);
            if (att < gfCos(light->m_spot_cutoff))
                att = 0.0;
            att = gfAbs(att);
            gfDouble denom = dist*dist*light->m_att_quad+dist*light->m_att_lin+light->m_att_const;
            if (!gfIsZero(denom))
                att /= denom;
            att = 1.0;

            // add diffuse part
            Ir += att * NdotL * material->m_diffuse_r * light->m_diffuse_r;
            Ig += att * NdotL * material->m_diffuse_g * light->m_diffuse_g;
            Ib += att * NdotL * material->m_diffuse_b * light->m_diffuse_b;

            // compute reflection vector
            gfVector3 R(2.0*NdotL*normal->X()-L.X(), 2.0*NdotL*normal->Y()-L.Y(), 2.0*NdotL*normal->Z()-L.Z());
            R.Normalize();
            // compute specular coefficient, (R.V)^n
            gfDouble spec = gfPower(gfAbs(gfScalarProduct3(&R, view_dir)), material->m_shininess);

            Ir += att * spec * material->m_specular_r * light->m_specular_r;
            Ig += att * spec * material->m_specular_g * light->m_specular_g;
            Ib += att * spec * material->m_specular_b * light->m_specular_b;
        }
    }

    // clamp channels to <0,1>
    Ir = gfClamp(0.0, 1.0, Ir);
    Ig = gfClamp(0.0, 1.0, Ig);
    Ib = gfClamp(0.0, 1.0, Ib);

    res_color->Set(Ir, Ig, Ib);
    return GF_S_OK;
}

GFRESULT gfAuxRaytracer::ComputeColorByRay(gfPoint3* ray_begin, gfVector3* ray_dir, gfInt depth, gfVector3* res_color)
{
    // find intersection in scene
    std::vector<gfScnNode*> nodes;
    gfDouble min_t = GF_MAX_DOUBLE;
    gfScnNode* nearest_node = NULL;
    gfRayInterInfo* nearest_info = NULL;
    res_color->Set(0.0, 0.0, 0.0);

    nodes.push_back(m_scene->m_root);
    gfInt size = (gfInt)nodes.size();
    while (size != 0)
    {
        gfScnNode* node = nodes.at(0);
        nodes.erase(nodes.begin());
        size = (gfInt)nodes.size();
        if (!node)
            continue;

        if (node->m_bound_vol)
        {
            gfDouble tmin, tmax;
            if (GF_SUCC(node->m_bound_vol->RayIntersection(ray_begin, ray_dir, &tmin, &tmax)))
            {
                if (node->m_id == GF_SCENE_GEOM_OBJECT)                        
                {
                    gfObjGeomObject* obj = (gfObjGeomObject*) node->m_data;
                    if (obj->GetID() == GF_NURBS_PATCH)
                    {
                        gfObjNURBSPatch* patch = (gfObjNURBSPatch*) node->m_data;
                        if (patch)
                            patch->EnableBezierClipping(m_bezier_clipping);
                    }
                    gfRayInterInfo* info = obj->RayIntersection(node->m_node_matrix, ray_begin, ray_dir);

                    if (info)
                    {
                        // now check only first intersection
                        if (info->m_inters.at(0) < min_t)
                        {
                            m_hit_count++;
                            min_t = info->m_inters.at(0);
                            nearest_node = node;
                            if (nearest_info)
                                gfDeleteRayInterInfo(nearest_info);
                            nearest_info = info;
                        }
                        else
                        {
                            gfDeleteRayInterInfo(info);
                        }
                    }
                }
            }
        }

        // add child to queue
        //if (node->m_id != GF_SCENE_GROUP && node->m_id != GF_SCENE_CONTROL_OBJECT_GROUP)
        {
            for (gfScnNode* child = node->GetFirstChild(); child; child = node->GetNextChild(child))
            {
                nodes.push_back(child);
            }
        }

        size = (gfInt)nodes.size();
    }

    if (nearest_node && nearest_info)
    {
        // intersection was found
        // point of intersection
        gfPoint3 inter(ray_begin->X()+nearest_info->m_inters.at(0)*ray_dir->X(),
                       ray_begin->Y()+nearest_info->m_inters.at(0)*ray_dir->Y(),
                       ray_begin->Z()+nearest_info->m_inters.at(0)*ray_dir->Z());
        // normal in point of intersection
        gfVector3 normal(nearest_info->m_normals.at(0).X(), 
                         nearest_info->m_normals.at(0).Y(), 
                         nearest_info->m_normals.at(0).Z());

        // get all visible lights from point
        std::vector<gfScnLight*> vis_lights;
        if (m_shadows)
            m_scene->GetAllVisibleLights(&inter, &vis_lights);
        else
            m_scene->GetSceneLights(&vis_lights);

        // get material of node
        std::vector<gfScnMaterial*> materials;
        m_scene->GetNodeMaterials(nearest_node, &materials);
        gfScnMaterial* material = NULL;
        if (materials.size() > 0)
            material = materials.at(0);

        // get local illumination color
        gfVector3 local_color(0.0, 0.0, 0.0);
        gfVector3 view_dir(-ray_dir->X(), -ray_dir->Y(), -ray_dir->Z());
        if (material)
            GetLocalLighting(&vis_lights, material, &inter, &view_dir, NULL, &normal, &local_color);

        depth--;

        // get color from reflection ray
        gfVector3 reflected_color(0.0, 0.0, 0.0);
        if (depth > 0 && material && !gfIsZero(material->m_reflection))
        {
            // get reflection vector
            gfVector3 V(-ray_dir->X(), -ray_dir->Y(), -ray_dir->Z());
            gfDouble NdotV = gfScalarProduct3(&normal, &V);
            gfVector3 R(2.0*NdotV*normal.X()-V.X(), 2.0*NdotV*normal.Y()-V.Y(), 2.0*NdotV*normal.Z()-V.Z());
            R.Normalize();
            ComputeColorByRay(&inter, &R, depth, &reflected_color);
            reflected_color.Set(reflected_color.X()*material->m_specular_r,
                                reflected_color.Y()*material->m_specular_g,
                                reflected_color.Z()*material->m_specular_b);
        }

        // get color from reflection ray
        gfVector3 refracted_color(0.0, 0.0, 0.0);
        if (depth > 0 && material && !gfIsZero(material->m_refraction))
        {
            gfBool ray_in = (gfScalarProduct3(&normal, ray_dir) < 0.0);
            gfDouble k = material->m_refraction;
            if (ray_in)
                k = 1.0 / k;

            // get refracted vector
            gfVector3 R;
            if (ray_in)
            {
                R.Set(-(k - 1.0)*normal.X()+k*ray_dir->X(), 
                   -(k - 1.0)*normal.Y()+k*ray_dir->Y(), 
                   -(k - 1.0)*normal.Z()+k*ray_dir->Z());
            }
            else
            {
                R.Set((k - 1.0)*normal.X()+k*ray_dir->X(), 
                (k - 1.0)*normal.Y()+k*ray_dir->Y(), 
                (k - 1.0)*normal.Z()+k*ray_dir->Z());
            }
            R.Normalize();
            ComputeColorByRay(&inter, &R, depth, &refracted_color);
            gfDouble att = gfClamp(0.5, 1.0, 1.0-inter.Distance(ray_begin));
            refracted_color.Set(att*refracted_color.X(), att*refracted_color.Y(), att*refracted_color.Z());
        }

        if (material && depth > 0)
        {
            gfDouble d = 1.0 + material->m_reflection + material->m_refraction;
            res_color->Set(gfClamp(0.0, 1.0, (local_color.X()+material->m_reflection*reflected_color.X()+material->m_refraction*refracted_color.X())/d),
                        gfClamp(0.0, 1.0, (local_color.Y()+material->m_reflection*reflected_color.Y()+material->m_refraction*refracted_color.Y())/d),
                        gfClamp(0.0, 1.0, (local_color.Z()+material->m_reflection*reflected_color.Z()+material->m_refraction*refracted_color.Z())/d));
        }
        else
        {
            res_color->Set(gfClamp(0.0, 1.0, local_color.X()),
                           gfClamp(0.0, 1.0, local_color.Y()),
                           gfClamp(0.0, 1.0, local_color.Z()));
        }
    }
    else //if (depth == m_depth)
    {
        // return color of background or just black color
        res_color->Set(m_scene->m_background_r, m_scene->m_background_g, m_scene->m_background_b);
    }

    if (nearest_info)
        gfDeleteRayInterInfo(nearest_info);

    return GF_S_OK;
}

GFRESULT gfAuxRaytracer::SetResolution(gfInt width, gfInt height)
{
    if (m_pic_data)
        delete [] m_pic_data;
    m_pic_data = NULL;
    m_width = width;
    m_height = height;

    // allocate memory for pixels
    m_pic_data = new gfDouble[3 * width * height];

    return GF_S_OK;
}

GFRESULT gfAuxRaytracer::SetScene(gfScnGraph* scene)
{
    if (!scene)
        return GF_E_BADARG;
    m_scene = scene;
    return GF_S_OK;
}

GFRESULT gfAuxRaytracer::SetCamera(gfScnCamera* camera)
{
    if (!camera)
        return GF_E_BADARG;
    m_camera = camera;
    return GF_S_OK;
}

gfInt gfAuxRaytracer::GetWidth(void)
{
    return m_width;
}

gfInt gfAuxRaytracer::GetHeight(void)
{
    return m_height;
}

GFRESULT gfAuxRaytracer::Raytrace(gfDouble depth)
{
    if (!m_camera || !m_scene)
        return GF_E_BADARG;
    m_hit_count = 0;
    m_depth = depth;

    gfWinFPSCounter count;
    count.BeginMSCount();

    // camera position
    gfPoint3 pos;
    m_camera->GetCameraPosition(&pos);
    m_ex = pos.X();
    m_ey = pos.Y();
    m_ez = pos.Z();

    // camera coordinate system
    gfVector3 dir;
    gfVector3 up;
    gfVector3 base;
    m_camera->GetCameraFrame(&dir, &up, &base);

    // distance of camera position and projection plane
    gfDouble plane_dist = m_camera->m_near;

    // middle of projection plane
    gfPoint3 center;
    center.Set(m_ex+plane_dist*dir.X(), m_ey+plane_dist*dir.Y(), m_ez+plane_dist*dir.Z());

    gfDouble left, right, bottom, top;
    if (m_camera->m_type == GF_CAMERA_PERSPECTIVE_DEF || m_camera->m_type == GF_CAMERA_PERSPECTIVE_DEF)
    {
        left = m_camera->m_left;
        right = m_camera->m_right;
        top = m_camera->m_top;
        bottom = m_camera->m_bottom;
    }
    else if (m_camera->m_type == GF_CAMERA_ORTHO_FRONT || m_camera->m_type == GF_CAMERA_ORTHO_RIGHT ||
        m_camera->m_type == GF_CAMERA_ORTHO_TOP || m_camera->m_type == GF_CAMERA_ORTHO_CUST)
    {
        left = m_camera->m_dist*m_camera->m_left;
        right = m_camera->m_dist*m_camera->m_right;
        top = m_camera->m_dist*m_camera->m_top;
        bottom = m_camera->m_dist*m_camera->m_bottom;
    }
    gfDouble aax = (right - left) / (gfDouble)(m_width - 1);
    gfDouble aay = (top - bottom) / (gfDouble)(m_height - 1);
    for (gfInt i = 0;i < m_width;i++)
    {
        for (gfInt j = 0;j < m_height;j++)
        {
            if (m_interleaved && (i % 2 == 1 || j % 2 == 1))
                continue;

            gfVector3 color(0.0, 0.0, 0.0);
            gfVector3 color1(0.0, 0.0, 0.0);
            gfVector3 color2(0.0, 0.0, 0.0);
            gfVector3 color3(0.0, 0.0, 0.0);
            gfVector3 color4(0.0, 0.0, 0.0);

            // point in projection plane in 3D
            gfDouble rat_i = left + i * aax;
            gfDouble rat_j = bottom + j * aay;
            gfDouble pt_x = center.X() + rat_i * base.X() + rat_j * up.X();
            gfDouble pt_y = center.Y() + rat_i * base.Y() + rat_j * up.Y();
            gfDouble pt_z = center.Z() + rat_i * base.Z() + rat_j * up.Z();

            if (m_camera->m_type == GF_CAMERA_PERSPECTIVE_DEF || m_camera->m_type == GF_CAMERA_PERSPECTIVE_DEF)
            {
                // perspective camera
                gfVector3 ray(pt_x-m_ex, pt_y-m_ey, pt_z-m_ez);
                ray.Normalize();
                ComputeColorByRay(&center, &ray, depth, &color);
            }
            else if (m_camera->m_type == GF_CAMERA_ORTHO_FRONT || m_camera->m_type == GF_CAMERA_ORTHO_RIGHT ||
                     m_camera->m_type == GF_CAMERA_ORTHO_TOP || m_camera->m_type == GF_CAMERA_ORTHO_CUST)
            {
                // ortho camera
                gfPoint3 start(pt_x-plane_dist*dir.X(), pt_y-plane_dist*dir.Y(), pt_z-plane_dist*dir.Z());
                ComputeColorByRay(&start, &dir, depth, &color);
            }

            gfDouble d = color.Length();
            if (d > gfSquareRoot(3.0) - 0.1)
            {
                gfInt i = 0;
            }
            m_pic_data[3*(i+j*m_width)] = color.X();
            m_pic_data[3*(i+j*m_width)+1] = color.Y();
            m_pic_data[3*(i+j*m_width)+2] = color.Z();
        }
    }

    // now go throught all pixels, and if there is high frequency, sample point in grandient direction
    if (m_aa == GF_AA_ADAPTIVE && !m_interleaved)
    {
        gfDouble* aux = new gfDouble[m_width*m_height*3];
        memcpy(aux, m_pic_data, m_width*m_height*3*sizeof(gfDouble));

        for (gfInt i = 0;i < m_width;i++)
        {
            for (gfInt j = 0;j < m_height;j++)
            {
                if ((i < 1) || (j < 1) || (i > m_width-2) || (j > m_height-2))
                    continue;

                gfDouble d_r = gfAbs(aux[3*(i+j*m_width)] - aux[3*(i+1+j*m_width)])+
                            gfAbs(aux[3*(i+j*m_width)] - aux[3*(i-1+j*m_width)])+
                            gfAbs(aux[3*(i+j*m_width)] - aux[3*(i+(j+1)*m_width)])+
                            gfAbs(aux[3*(i+j*m_width)] - aux[3*(i+(j-1)*m_width)]);
                gfDouble d_g = gfAbs(aux[3*(i+j*m_width)+1] - aux[3*(i+1+j*m_width)+1])+
                            gfAbs(aux[3*(i+j*m_width)+1] - aux[3*(i-1+j*m_width)+1])+
                            gfAbs(aux[3*(i+j*m_width)+1] - aux[3*(i+(j+1)*m_width)+1])+
                            gfAbs(aux[3*(i+j*m_width)+1] - aux[3*(i+(j-1)*m_width)+1]);
                gfDouble d_b = gfAbs(aux[3*(i+j*m_width)+2] - aux[3*(i+1+j*m_width)+2])+
                            gfAbs(aux[3*(i+j*m_width)+2] - aux[3*(i-1+j*m_width)+2])+
                            gfAbs(aux[3*(i+j*m_width)+2] - aux[3*(i+(j+1)*m_width)+2])+
                            gfAbs(aux[3*(i+j*m_width)+2] - aux[3*(i+(j-1)*m_width)+2]);

                // check if we are in locally homogenous region
                if (d_r*d_g*d_b < 0.001)
                    continue;

                // gradient
                gfVector3 g1, g2;
                g1.Set(aux[3*(i+1+j*m_width)]-aux[3*(i-1+j*m_width)],
                    aux[3*(i+1+j*m_width)+1]-aux[3*(i-1+j*m_width)+1],
                    aux[3*(i+1+j*m_width)+2]-aux[3*(i-1+j*m_width)+2]);
                g2.Set(aux[3*(i+(j+1)*m_width)]-aux[3*(i+(j-1)*m_width)],
                    aux[3*(i+(j+1)*m_width)+1]-aux[3*(i+(j-1)*m_width)+1],
                    aux[3*(i+(j+1)*m_width)+2]-aux[3*(i+(j-1)*m_width)+2]);
                gfVector2 gradient;
                gradient.Set(g1.Length(), g2.Length());
                gradient.Normalize();

                gfVector3 color1(0.0, 0.0, 0.0);
                gfDouble rat_i = (0.5)*aax*gradient.X() + left + i * aax;
                gfDouble rat_j = (0.5)*aay*gradient.Y() + bottom + j * aay;
                gfDouble pt_x = center.X() + rat_i * base.X() + rat_j * up.X();
                gfDouble pt_y = center.Y() + rat_i * base.Y() + rat_j * up.Y();
                gfDouble pt_z = center.Z() + rat_i * base.Z() + rat_j * up.Z();
                if (m_camera->m_type == GF_CAMERA_PERSPECTIVE_DEF || m_camera->m_type == GF_CAMERA_PERSPECTIVE_DEF)
                {
                    // perspective camera
                    gfVector3 ray(pt_x-m_ex, pt_y-m_ey, pt_z-m_ez);
                    ray.Normalize();
                    ComputeColorByRay(&center, &ray, depth, &color1);
                }
                else if (m_camera->m_type == GF_CAMERA_ORTHO_FRONT || m_camera->m_type == GF_CAMERA_ORTHO_RIGHT ||
                    m_camera->m_type == GF_CAMERA_ORTHO_TOP || m_camera->m_type == GF_CAMERA_ORTHO_CUST)
                {
                    // ortho camera
                    gfPoint3 start(pt_x-plane_dist*dir.X(), pt_y-plane_dist*dir.Y(), pt_z-plane_dist*dir.Z());
                    ComputeColorByRay(&start, &dir, depth, &color1);
                }

                gfVector3 color2(0.0, 0.0, 0.0);
                rat_i = (-0.5)*aax*gradient.X() + left + i * aax;
                rat_j = (-0.5)*aay*gradient.Y() + bottom + j * aay;
                pt_x = center.X() + rat_i * base.X() + rat_j * up.X();
                pt_y = center.Y() + rat_i * base.Y() + rat_j * up.Y();
                pt_z = center.Z() + rat_i * base.Z() + rat_j * up.Z();
                if (m_camera->m_type == GF_CAMERA_PERSPECTIVE_DEF || m_camera->m_type == GF_CAMERA_PERSPECTIVE_DEF)
                {
                    // perspective camera
                    gfVector3 ray(pt_x-m_ex, pt_y-m_ey, pt_z-m_ez);
                    ray.Normalize();
                    ComputeColorByRay(&center, &ray, depth, &color2);
                }
                else if (m_camera->m_type == GF_CAMERA_ORTHO_FRONT || m_camera->m_type == GF_CAMERA_ORTHO_RIGHT ||
                    m_camera->m_type == GF_CAMERA_ORTHO_TOP || m_camera->m_type == GF_CAMERA_ORTHO_CUST)
                {
                    // ortho camera
                    gfPoint3 start(pt_x-plane_dist*dir.X(), pt_y-plane_dist*dir.Y(), pt_z-plane_dist*dir.Z());
                    ComputeColorByRay(&start, &dir, depth, &color2);
                }

                m_pic_data[3*(i+j*m_width)] = (aux[3*(i+j*m_width)]+color1.X()+color2.X())/3.0;
                m_pic_data[3*(i+j*m_width)+1] = (aux[3*(i+j*m_width)+1]+color1.Y()+color2.Y())/3.0;
                m_pic_data[3*(i+j*m_width)+2] = (aux[3*(i+j*m_width)+2]+color1.Z()+color2.Z())/3.0;
            }
        }
        delete [] aux;
    }

    gfLog("Scene raytraced, width = %d, height = %d, depth = %d, casted rays = %d, time = %d ms.\n", m_width, m_height, m_depth, m_hit_count, count.EndMSCount());

    return GF_S_OK;
}

GFRESULT gfAuxRaytracer::GetPixel(gfInt i, gfInt j, gfVector3* color)
{
    if (i < 0 || i >= m_width)
        return GF_E_BADARG;
    if (j < 0 || j >= m_height)
        return GF_E_BADARG;

    color->Set(m_pic_data[3*(i+j*m_width)], m_pic_data[3*(i+j*m_width)+1], m_pic_data[3*(i+j*m_width)+2]);

    return GF_S_OK;
}

gfDouble* gfAuxRaytracer::GetPixels(void)
{
    return m_pic_data;
}

gfDouble gfAuxRaytracer::GetCameraRatio(void)
{
    if (!m_camera)
        return 0.0;

    gfDouble nom = (m_camera->m_right - m_camera->m_left);
    gfDouble denom = (m_camera->m_top - m_camera->m_bottom);

    if (gfIsZero(denom))
        return 0.0;

    return (nom / denom);
}

GFRESULT gfAuxRaytracer::SetShadowsRendering(gfBool state)
{
    m_shadows = state;
    return GF_S_OK;
}

GFRESULT gfAuxRaytracer::SetInterleaved(gfBool state)
{
    m_interleaved = state;
    return GF_S_OK;
}

GFRESULT gfAuxRaytracer::SetBezierClipping(gfBool state)
{
    m_bezier_clipping = state;
    return GF_S_OK;
}

GFRESULT gfAuxRaytracer::SetAAType(gfAntialiasingType aa)
{
    m_aa = aa;
    return GF_S_OK;
}
