//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfMatrix.cpp                                         //
//   Description: Class for work with 4x4 matrix                //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "gfMatrix.h"
#include "gfAuxNumUtils.h"
#include "gfObjectFactory.h"


// ----------------------- 2x2 matrix ------------------------
gfMatrix2::gfMatrix2()
{
    m_data[0][0] = 0.0;
    m_data[1][0] = 0.0;
    m_data[0][1] = 0.0;
    m_data[1][1] = 0.0;
}

gfMatrix2::~gfMatrix2()
{
}

gfMatrix2* gfMatrix2::Clone(void)
{
    gfMatrix2* new_matrix = gfCreateMatrix2();
    new_matrix->Set(m_data[0][0], m_data[0][1], m_data[1][0], m_data[1][1]);
    return new_matrix;
}

GFRESULT gfMatrix2::Set(gfDouble a11, gfDouble a12, gfDouble a21, gfDouble a22)
{
    m_data[0][0] = a11;
    m_data[1][0] = a12;
    m_data[0][1] = a21;
    m_data[1][1] = a22;

    return GF_S_OK;
}

GFRESULT gfMatrix2::Set(gfInt i,gfInt j, gfDouble value)
{
    if (i < 1 || j > 2 || j < 1 || j > 2)
        return GF_E_BADARG;

    m_data[i - 1][j - 1] = value;
    return GF_S_OK;
}

gfDouble gfMatrix2::at(gfInt i, gfInt j)
{
    if (i < 1 || j > 2 || j < 1 || j > 2)
        return GF_E_BADARG;

    return m_data[i - 1][j - 1];
}

gfDouble* gfMatrix2::GetData(void)
{
    return (gfDouble*) m_data;
}

GFRESULT gfMatrix2::LoadIdentity(void)
{
    for (gfInt i = 0;i < 2;i++)
        for (gfInt j = 0;j < 2;j++)
        {
            if (i == j)
                m_data[i][j] = 1.0;
            else
                m_data[i][j] = 0.0;
        }

    return GF_S_OK;
}

GFRESULT gfMatrix2::LoadScale(gfDouble sx, gfDouble sy)
{
    m_data[0][0] = sx;
    m_data[1][0] = 0.0;
    m_data[0][1] = 0.0;
    m_data[1][1] = sy;

    return GF_S_OK;
}

GFRESULT gfMatrix2::LoadRotation(gfDouble angle)
{
    m_data[0][0] = gfCos(angle);
    m_data[0][1] = gfSin(angle);
    m_data[1][0] = -gfSin(angle);
    m_data[1][1] = gfCos(angle);

    return GF_S_OK;
}

GFRESULT gfMatrix2::AddScale(gfDouble sx, gfDouble sy)
{
    gfMatrix2* m = gfCreateMatrix2();
    m->LoadScale(sx, sy);
    Multiply(m);
    gfDeleteMatrix2(m);

    return GF_S_OK;
}

GFRESULT gfMatrix2::AddRotation(gfDouble angle)
{
    gfMatrix2* m = gfCreateMatrix2();
    m->LoadRotation(angle);
    Multiply(m);
    gfDeleteMatrix2(m);

    return GF_S_OK;
}

GFRESULT gfMatrix2::PreMultiply(gfMatrix2* matrix) 
{
    gfMatrix2* m = Clone();

    for (gfInt i = 0;i < 2;i++)
    {
        for (gfInt j = 0;j < 2;j++)
        {
            gfDouble aux = 0.0;
            for (gfInt k = 0;k < 2;k++)
            {
                aux += matrix->at(i+1, k+1)* m->at(k+1, j+1);
            }

            Set(i+1, j+1, aux);
        }
    }

    gfDeleteMatrix2(m);

    return GF_S_OK;
}

GFRESULT gfMatrix2::Multiply(gfMatrix2* param) 
{
    gfMatrix2* m = Clone();

    for (gfInt i = 0;i < 2;i++)
    {
        for (gfInt j = 0;j < 2;j++)
        {
            gfDouble aux = 0.0;
            for (gfInt k = 0;k < 2;k++)
            {
                aux += m->at(i+1, k+1)* param->at(k+1, j+1);
            }

            Set(i+1, j+1, aux);
        }
    }

    gfDeleteMatrix2(m);

    return GF_S_OK;
}

gfDouble gfMatrix2::Determinant(void)
{
    return m_data[0][0] * m_data[1][1] - m_data[1][0] * m_data[1][1];
}

GFRESULT gfMatrix2::Inverse(void)
{
    return GF_S_OK;
}


// ----------------------- 3x3 matrix ------------------------
gfMatrix3::gfMatrix3()
{
    m_data[0][0] = 0.0;
    m_data[0][1] = 0.0;
    m_data[0][2] = 0.0;
    m_data[1][0] = 0.0;
    m_data[1][1] = 0.0;
    m_data[1][2] = 0.0;
    m_data[2][0] = 0.0;
    m_data[2][1] = 0.0;
    m_data[2][2] = 0.0;
}

gfMatrix3::~gfMatrix3()
{
}

gfMatrix3* gfMatrix3::Clone(void)
{
    gfMatrix3* new_matrix = gfCreateMatrix3();
    new_matrix->Set(m_data[0][0], m_data[0][1], m_data[0][2], 
                    m_data[1][0], m_data[1][1], m_data[1][2],
                    m_data[2][0], m_data[2][1], m_data[2][2]);
    return new_matrix;
}

GFRESULT gfMatrix3::Set(gfDouble a11, gfDouble a12, gfDouble a13,
                        gfDouble a21, gfDouble a22, gfDouble a23,
                        gfDouble a31, gfDouble a32, gfDouble a33)
{
    m_data[0][0] = a11;
    m_data[0][1] = a12;
    m_data[0][2] = a13;
    m_data[1][0] = a21;
    m_data[1][1] = a22;
    m_data[1][2] = a23;
    m_data[2][0] = a31;
    m_data[2][1] = a32;
    m_data[2][2] = a33;

    return GF_S_OK;
}

GFRESULT gfMatrix3::Set(gfInt i,gfInt j, gfDouble value)
{
    if (i < 1 || j > 3 || j < 1 || j > 3)
        return GF_E_BADARG;

    m_data[i - 1][j - 1] = value;
    return GF_S_OK;
}

gfDouble gfMatrix3::at(gfInt i, gfInt j)
{
    if (i < 1 || j > 3 || j < 1 || j > 3)
        return GF_E_BADARG;

    return m_data[i - 1][j - 1];
}

gfDouble* gfMatrix3::GetData(void)
{
    return (gfDouble*) m_data;
}

GFRESULT gfMatrix3::LoadIdentity(void)
{
    for (gfInt i = 0;i < 3;i++)
        for (gfInt j = 0;j < 3;j++)
        {
            if (i == j)
                m_data[i][j] = 1.0;
            else
                m_data[i][j] = 0.0;
        }

    return GF_S_OK;
}

GFRESULT gfMatrix3::LoadScale(gfDouble sx, gfDouble sy, gfDouble sz)
{
    m_data[0][0] = sx;
    m_data[0][1] = 0.0;
    m_data[0][2] = 0.0;
    m_data[1][0] = 0.0;
    m_data[1][1] = sy;
    m_data[1][2] = 0.0;
    m_data[2][0] = 0.0;
    m_data[2][1] = 0.0;
    m_data[2][2] = sz;

    return GF_S_OK;
}

GFRESULT gfMatrix3::LoadRotationX(gfDouble angle)
{
    LoadIdentity();

    m_data[1][1] = gfCos(angle);
    m_data[1][2] = gfSin(angle);
    m_data[2][1] = -gfSin(angle);
    m_data[2][2] = gfCos(angle);

    return GF_S_OK;
}

GFRESULT gfMatrix3::LoadRotationY(gfDouble angle)
{
    LoadIdentity();

    m_data[0][0] = gfCos(angle);
    m_data[0][2] = gfSin(angle);
    m_data[2][0] = -gfSin(angle);
    m_data[2][2] = gfCos(angle);

    return GF_S_OK;
}

GFRESULT gfMatrix3::LoadRotationZ(gfDouble angle)
{
    LoadIdentity();

    m_data[0][0] = gfCos(angle);
    m_data[0][1] = gfSin(angle);
    m_data[1][0] = -gfSin(angle);
    m_data[1][1] = gfCos(angle);

    return GF_S_OK;
}

GFRESULT gfMatrix3::LoadTranslation(gfDouble tx, gfDouble ty)
{
    LoadIdentity();

    m_data[0][2] = tx;
    m_data[1][2] = ty;

    return GF_S_OK;
}

GFRESULT gfMatrix3::AddScale(gfDouble sx, gfDouble sy, gfDouble sz)
{
    gfMatrix3* m = gfCreateMatrix3();
    m->LoadTranslation(sx, sy);
    Multiply(m);
    gfDeleteMatrix3(m);

    return GF_S_OK;
}

GFRESULT gfMatrix3::AddRotationX(gfDouble angle)
{
    gfMatrix3* m = gfCreateMatrix3();
    m->LoadRotationX(angle);
    Multiply(m);
    gfDeleteMatrix3(m);

    return GF_S_OK;
}

GFRESULT gfMatrix3::AddRotationY(gfDouble angle)
{
    gfMatrix3* m = gfCreateMatrix3();
    m->LoadRotationY(angle);
    Multiply(m);
    gfDeleteMatrix3(m);

    return GF_S_OK;
}

GFRESULT gfMatrix3::AddRotationZ(gfDouble angle)
{
    gfMatrix3* m = gfCreateMatrix3();
    m->LoadRotationZ(angle);
    Multiply(m);
    gfDeleteMatrix3(m);

    return GF_S_OK;
}

GFRESULT gfMatrix3::PreMultiply(gfMatrix3* matrix) 
{
    gfMatrix3* m = Clone();

    for (gfInt i = 0;i < 3;i++)
    {
        for (gfInt j = 0;j < 3;j++)
        {
            gfDouble aux = 0.0;
            for (gfInt k = 0;k < 3;k++)
            {
                aux += matrix->at(i+1, k+1)* m->at(k+1, j+1);
            }

            Set(i+1, j+1, aux);
        }
    }

    gfDeleteMatrix3(m);

    return GF_S_OK;
}

GFRESULT gfMatrix3::Multiply(gfMatrix3* param) 
{
    gfMatrix3* m = Clone();

    for (gfInt i = 0;i < 3;i++)
    {
        for (gfInt j = 0;j < 3;j++)
        {
            gfDouble aux = 0.0;
            for (gfInt k = 0;k < 3;k++)
            {
                aux += m->at(i+1, k+1)* param->at(k+1, j+1);
            }

            Set(i+1, j+1, aux);
        }
    }

    gfDeleteMatrix3(m);

    return GF_S_OK;
}

gfDouble gfMatrix3::Determinant(void)
{
    gfDouble result = m_data[0][0] * m_data[1][1] * m_data[2][2] +
                      m_data[1][0] * m_data[2][1] * m_data[0][2] +
                      m_data[2][0] * m_data[0][1] * m_data[1][2] -
                      m_data[2][0] * m_data[1][1] * m_data[0][2] -
                      m_data[0][0] * m_data[2][1] * m_data[1][2] -
                      m_data[1][0] * m_data[0][1] * m_data[2][2];

    return result;
}

GFRESULT gfMatrix3::Inverse(void)
{
    return GF_S_OK;
}


// ----------------------- 3x3 matrix ------------------------
gfMatrix4::gfMatrix4()
{
    m_data[0][0] = 0.0;
    m_data[0][1] = 0.0;
    m_data[0][2] = 0.0;
    m_data[0][3] = 0.0;

    m_data[1][0] = 0.0;
    m_data[1][1] = 0.0;
    m_data[1][2] = 0.0;
    m_data[1][3] = 0.0;

    m_data[2][0] = 0.0;
    m_data[2][1] = 0.0;
    m_data[2][2] = 0.0;
    m_data[2][3] = 0.0;

    m_data[3][0] = 0.0;
    m_data[3][1] = 0.0;
    m_data[3][2] = 0.0;
    m_data[3][3] = 0.0;
}

gfMatrix4::~gfMatrix4()
{
}

gfMatrix4* gfMatrix4::Clone(void)
{
    gfMatrix4* new_matrix = gfCreateMatrix4();
    new_matrix->Set(m_data[0][0], m_data[0][1], m_data[0][2], m_data[0][3],
                    m_data[1][0], m_data[1][1], m_data[1][2], m_data[1][3],
                    m_data[2][0], m_data[2][1], m_data[2][2], m_data[2][3],
                    m_data[3][0], m_data[3][1], m_data[3][2], m_data[3][3]);
    return new_matrix;
}

GFRESULT gfMatrix4::Set(gfDouble a11, gfDouble a12, gfDouble a13, gfDouble a14,
                        gfDouble a21, gfDouble a22, gfDouble a23, gfDouble a24,
                        gfDouble a31, gfDouble a32, gfDouble a33, gfDouble a34,
                        gfDouble a41, gfDouble a42, gfDouble a43, gfDouble a44)
{
    m_data[0][0] = a11;
    m_data[0][1] = a12;
    m_data[0][2] = a13;
    m_data[0][3] = a14;

    m_data[1][0] = a21;
    m_data[1][1] = a22;
    m_data[1][2] = a23;
    m_data[1][3] = a24;

    m_data[2][0] = a31;
    m_data[2][1] = a32;
    m_data[2][2] = a33;
    m_data[2][3] = a34;

    m_data[3][0] = a41;
    m_data[3][1] = a42;
    m_data[3][2] = a43;
    m_data[3][3] = a44;

    return GF_S_OK;
}

GFRESULT gfMatrix4::Set(gfInt i,gfInt j, gfDouble value)
{
    if (i < 1 || j > 4 || j < 1 || j > 4)
        return GF_E_BADARG;

    m_data[i - 1][j - 1] = value;
    return GF_S_OK;
}

gfDouble gfMatrix4::at(gfInt i, gfInt j)
{
    if (i < 1 || j > 4 || j < 1 || j > 4)
        return GF_E_BADARG;

    return m_data[i - 1][j - 1];
}

gfDouble* gfMatrix4::GetData(void)
{
    return (gfDouble*) m_data;
}

GFRESULT gfMatrix4::LoadIdentity(void)
{
    for (gfInt i = 0;i < 4;i++)
        for (gfInt j = 0;j < 4;j++)
        {
            if (i == j)
                m_data[i][j] = 1.0;
            else
                m_data[i][j] = 0.0;
        }

    return GF_S_OK;
}

GFRESULT gfMatrix4::LoadScale(gfDouble sx, gfDouble sy, gfDouble sz, gfDouble sw)
{
    m_data[0][0] = sx;
    m_data[0][1] = 0.0;
    m_data[0][2] = 0.0;
    m_data[0][3] = 0.0;

    m_data[1][0] = 0.0;
    m_data[1][1] = sy;
    m_data[1][2] = 0.0;
    m_data[1][3] = 0.0;

    m_data[2][0] = 0.0;
    m_data[2][1] = 0.0;
    m_data[2][2] = sz;
    m_data[2][3] = 0.0;

    m_data[3][0] = 0.0;
    m_data[3][1] = 0.0;
    m_data[3][2] = 0.0;
    m_data[3][3] = sw;

    return GF_S_OK;
}

GFRESULT gfMatrix4::LoadRotationX(gfDouble angle)
{
    LoadIdentity();

    m_data[1][1] = gfCos(angle);
    m_data[1][2] = gfSin(angle);
    m_data[2][1] = -gfSin(angle);
    m_data[2][2] = gfCos(angle);

    return GF_S_OK;
}

GFRESULT gfMatrix4::LoadRotationY(gfDouble angle)
{
    LoadIdentity();

    m_data[0][0] = gfCos(angle);
    m_data[0][2] = gfSin(angle);
    m_data[2][0] = -gfSin(angle);
    m_data[2][2] = gfCos(angle);

    return GF_S_OK;
}

GFRESULT gfMatrix4::LoadRotationZ(gfDouble angle)
{
    LoadIdentity();

    m_data[0][0] = gfCos(angle);
    m_data[0][1] = gfSin(angle);
    m_data[1][0] = -gfSin(angle);
    m_data[1][1] = gfCos(angle);

    return GF_S_OK;
}

GFRESULT gfMatrix4::LoadTranslation(gfDouble tx, gfDouble ty, gfDouble tz)
{
    LoadIdentity();

    m_data[0][3] = tx;
    m_data[1][3] = ty;
    m_data[2][3] = tz;

    return GF_S_OK;
}

GFRESULT gfMatrix4::AddScale(gfDouble sx, gfDouble sy, gfDouble sz, gfDouble sw)
{
    gfMatrix4* m = gfCreateMatrix4();
    m->LoadScale(sx, sy, sz, sw);
    Multiply(m);
    gfDeleteMatrix4(m);

    return GF_S_OK;
}

GFRESULT gfMatrix4::AddRotationX(gfDouble angle)
{
    gfMatrix4* m = gfCreateMatrix4();
    m->LoadRotationX(angle);
    Multiply(m);
    gfDeleteMatrix4(m);

    return GF_S_OK;
}

GFRESULT gfMatrix4::AddRotationY(gfDouble angle)
{
    gfMatrix4* m = gfCreateMatrix4();
    m->LoadRotationY(angle);
    Multiply(m);
    gfDeleteMatrix4(m);

    return GF_S_OK;
}

GFRESULT gfMatrix4::AddRotationZ(gfDouble angle)
{
    gfMatrix4* m = gfCreateMatrix4();
    m->LoadRotationZ(angle);
    Multiply(m);
    gfDeleteMatrix4(m);

    return GF_S_OK;
}

GFRESULT gfMatrix4::PreMultiply(gfMatrix4* matrix) 
{
    gfMatrix4* m = Clone();

    for (gfInt i = 0;i < 4;i++)
    {
        for (gfInt j = 0;j < 4;j++)
        {
            gfDouble aux = 0.0;
            for (gfInt k = 0;k < 4;k++)
            {
                aux += matrix->at(i+1, k+1)* m->at(k+1, j+1);
            }

            Set(i+1, j+1, aux);
        }
    }

    gfDeleteMatrix4(m);

    return GF_S_OK;
}

GFRESULT gfMatrix4::Multiply(gfMatrix4* param) 
{
    gfMatrix4* m = Clone();

    for (gfInt i = 0;i < 4;i++)
    {
        for (gfInt j = 0;j < 4;j++)
        {
            gfDouble aux = 0.0;
            for (gfInt k = 0;k < 4;k++)
            {
                aux += m->at(i+1, k+1)* param->at(k+1, j+1);
            }

            Set(i+1, j+1, aux);
        }
    }

    gfDeleteMatrix4(m);

    return GF_S_OK;
}

gfDouble gfMatrix4::Determinant(void)
{
    gfDouble a11 = at(1,1);gfDouble a12 = at(1,2);gfDouble a13 = at(1,3);gfDouble a14 = at(1,4);
    gfDouble a21 = at(2,1);gfDouble a22 = at(2,2);gfDouble a23 = at(2,3);gfDouble a24 = at(2,4);
    gfDouble a31 = at(3,1);gfDouble a32 = at(3,2);gfDouble a33 = at(3,3);gfDouble a34 = at(3,4);
    gfDouble a41 = at(4,1);gfDouble a42 = at(4,2);gfDouble a43 = at(4,3);gfDouble a44 = at(4,4);

    gfDouble result =   a11*a22*a33*a44 - a11*a22*a34*a43 + a11*a23*a34*a42 - a11*a23*a32*a44
        + a11*a24*a32*a43 - a11*a24*a33*a42 - a12*a23*a34*a41 + a12*a23*a31*a44
        - a12*a24*a31*a43 + a12*a24*a33*a41 - a12*a21*a33*a44 + a12*a21*a34*a43
        + a13*a24*a31*a42 - a13*a24*a32*a41 + a13*a21*a32*a44 - a13*a21*a34*a42
        + a13*a22*a34*a41 - a13*a22*a31*a44 - a14*a21*a32*a43 + a14*a21*a33*a42
        - a14*a22*a33*a41 + a14*a22*a31*a43 - a14*a23*a31*a42 + a14*a23*a32*a41;

    return result;
}

GFRESULT gfMatrix4::Inverse(void)
{
    return GF_S_OK;
}