//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObj3DBoxLattice.cpp                                    //
//   Description: NURBS box class                               //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfObj3DBoxLattice.h"
#include "gfObjectFactory.h"
#include "gfAuxNumUtils.h"
#include "gfAuxLogging.h"
#include "gfAuxGeomUtils.h"
#include "gfWinFPSCounter.h"
#include "gfAux4DArray.h"
#include "gfAuxItersectionUtils.h"
#include "gfAuxBoundCone.h"

#include "./OGL/glut.h"


gfObj3DBoxLattice::gfObj3DBoxLattice(void)
{
    m_time_stamp = 0;
    m_control_points.resize(0);
    m_num_u_control_points = 0;
    m_num_v_control_points = 0;
    m_num_w_control_points = 0;
}

gfObj3DBoxLattice::~gfObj3DBoxLattice(void)
{
    Clear();
}

GFRESULT gfObj3DBoxLattice::OGLDraw(gfLong flags)
{
    if (m_time_stamp == 0)
        return GF_S_OK;

    gfInt approx = GF_VIS_PREC;

    glNormal3f(0.0, 0.0, 0.0);
    glColor3f(0.0, 0.0, 0.0);
    glPointSize(3.0);
    for (gfInt i = 0;i < m_num_u_control_points;i++)
    {
        for (gfInt j = 0;j < m_num_v_control_points;j++)
        {
            for (gfInt k = 0;k < m_num_w_control_points;k++)
            {
                gfPoint3*  pt = GetControlPoint(i, j, k);
                glBegin(GL_POINTS);
                glVertex3f(pt->X(), pt->Y(), pt->Z());
                glEnd();

                glBegin(GL_LINES);

                if ((i + 1) < m_num_u_control_points)
                {
                    gfPoint3*  pt1 = GetControlPoint(i + 1, j, k);
                    glVertex3f(pt->X(), pt->Y(), pt->Z());
                    glVertex3f(pt1->X(), pt1->Y(), pt1->Z());
                }
                if ((j + 1) < m_num_v_control_points)
                {
                    gfPoint3*  pt1 = GetControlPoint(i, j + 1, k);
                    glVertex3f(pt->X(), pt->Y(), pt->Z());
                    glVertex3f(pt1->X(), pt1->Y(), pt1->Z());
                }
                if ((k + 1) < m_num_w_control_points)
                {
                    gfPoint3*  pt1 = GetControlPoint(i, j, k + 1);
                    glVertex3f(pt->X(), pt->Y(), pt->Z());
                    glVertex3f(pt1->X(), pt1->Y(), pt1->Z());
                }

                glEnd();
            }
        }
    }
    glPointSize(1.0);

    return GF_S_OK;
}

gfRayInterInfo* gfObj3DBoxLattice::RayIntersection(gfMatrix4* obj_matrix, gfPoint3 * ray_begin, gfVector3* ray_dir)
{
    gfRayInterInfo* info = gfCreateRayInterInfo();
    info->m_inters.push_back(0.5);
    info->m_normals.push_back(gfVector3(0.0, 0.0, 0.0));
    info->m_uvw.push_back(gfPoint3(0.0, 0.0, 0.0));
    return info;
}

gfGeomObjID gfObj3DBoxLattice::GetID(void)
{
    return GF_3D_BOX_LATTICE;
}

GFRESULT gfObj3DBoxLattice::Transform(gfMatrix4* matrix)
{
    // transform control points
    for (gfInt i = 0;i < m_num_u_control_points;i++)
    {
        for (gfInt j = 0;j < m_num_v_control_points;j++)
        {
            for (gfInt k = 0;k < m_num_w_control_points;k++)
            {
                gfPoint3* pt = GetControlPoint(i, j, k);
                gfTransformPoint3(pt, matrix);
            }
        }
    }

    return GF_S_OK;
}

gfObjGeomObject* gfObj3DBoxLattice::Clone(void)
{
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());
    gfObj3DBoxLattice* box = (gfObj3DBoxLattice*) clone;

    // set parameters of patch
    gfInt num_u_control_points = GetNumUControlPoints();
    gfInt num_v_control_points = GetNumVControlPoints();
    gfInt num_w_control_points = GetNumWControlPoints();
    box->Init(num_u_control_points, num_v_control_points, num_w_control_points);

    // copy control points and weights
    for (gfInt i = 0;i < num_u_control_points;i++)
        for (gfInt j = 0;j < num_v_control_points;j++)
            for (gfInt k = 0;k < num_w_control_points;k++)
            {
                gfPoint3* new_point = box->GetControlPoint(i, j, k);
                gfPoint3* old_point = GetControlPoint(i, j, k);
                new_point->Set(old_point->X(), old_point->Y(), old_point->Z());
            }

    return clone;
}

GFRESULT gfObj3DBoxLattice::Clear()
{
    Init(0, 0, 0);

    return GF_S_OK;
}

GFRESULT gfObj3DBoxLattice::Init(gfInt num_u_control_points, gfInt num_v_control_points, gfInt num_w_control_points)
{
    // create auxiliary arrays and fill them with old 
    std::vector<gfPoint3 *> aux_control_points;
    aux_control_points.resize(num_u_control_points * num_v_control_points * num_w_control_points);

    gfInt umax = (m_num_u_control_points > num_u_control_points) ? m_num_u_control_points : num_u_control_points;
    gfInt vmax = (m_num_v_control_points > num_v_control_points) ? m_num_v_control_points : num_v_control_points;
    gfInt wmax = (m_num_w_control_points > num_w_control_points) ? m_num_w_control_points : num_w_control_points;

    for (gfInt i = 0;i < umax;i++)
    {
        for (gfInt j = 0;j < vmax;j++)
        {
            for (gfInt k = 0;k < wmax;k++)
            {
                // point in new & old
                if ((i < num_u_control_points && j < num_v_control_points && k < num_w_control_points) && (i < m_num_u_control_points && j < m_num_v_control_points && k < m_num_w_control_points))
                {
                    aux_control_points[i + num_u_control_points * j + num_u_control_points * num_v_control_points * k] = GetControlPoint(i, j, k);
                }

                // point in new & not in old
                if ((i < num_u_control_points && j < num_v_control_points && k < num_w_control_points) && !(i < m_num_u_control_points && j < m_num_v_control_points && k < m_num_w_control_points))
                {
                    gfPoint3 * new_pt = gfCreatePoint3();
                    aux_control_points[i + num_u_control_points * j + num_u_control_points * num_v_control_points * k] = new_pt;
                }

                // point in old & not in new
                if (!(i < num_u_control_points && j < num_v_control_points && k < num_w_control_points) && (i < m_num_u_control_points && j < m_num_v_control_points && k < m_num_w_control_points))
                {
                    gfPoint3 * old_pt = GetControlPoint(i, j, k);
                    gfDeletePoint3(old_pt);
                }
            }
        }
    }

    m_num_u_control_points = num_u_control_points;
    m_num_v_control_points = num_v_control_points;
    m_num_w_control_points = num_w_control_points;

    m_control_points.resize(m_num_u_control_points * m_num_v_control_points * m_num_w_control_points);

    // copy from auxiliary arrays
    for (gfInt i = 0;i < m_num_u_control_points * m_num_v_control_points * m_num_w_control_points;i++)
    {
        m_control_points[i] = aux_control_points.at(i);
    }

    m_time_stamp++;

    return GF_S_OK;
}

GFRESULT gfObj3DBoxLattice::InitUniformNet(gfInt num_u_control_points, gfInt num_v_control_points, gfInt num_w_control_points)
{
    Clear();
    Init(num_u_control_points, num_v_control_points, num_w_control_points);

    gfDouble invu = 1.0 / (num_u_control_points-1);
    gfDouble invv = 1.0 / (num_v_control_points-1);
    gfDouble invw = 1.0 / (num_w_control_points-1);
    for (gfInt i = 0;i < num_u_control_points;i++)
    {
        for (gfInt j = 0;j < num_v_control_points;j++)
        {
            for (gfInt k = 0;k < num_w_control_points;k++)
            {
                gfPoint3* pt = GetControlPoint(i, j, k);
                pt->Set(i*invu, j*invv, k*invw);
            }
        }
    }

    return GF_S_OK;
}

gfInt gfObj3DBoxLattice::GetNumUControlPoints(void)
{
    return m_num_u_control_points;
}

gfInt gfObj3DBoxLattice::GetNumVControlPoints(void)
{
    return m_num_v_control_points;
}

gfInt gfObj3DBoxLattice::GetNumWControlPoints(void)
{
    return m_num_w_control_points;
}

gfPoint3 * gfObj3DBoxLattice::GetControlPoint(gfInt u_index, gfInt v_index, gfInt w_index)
{
    gfInt u = u_index;
    if (u < 0) u = 0;
    if (u > (m_num_u_control_points-1)) u = m_num_u_control_points-1;

    gfInt v = v_index;
    if (v < 0) v = 0;
    if (v > (m_num_v_control_points-1)) v = m_num_v_control_points-1;

    gfInt w = w_index;
    if (w < 0) w = 0;
    if (w > (m_num_w_control_points-1)) w = m_num_w_control_points-1;

    return m_control_points.at(u + v * m_num_u_control_points + w * m_num_u_control_points * m_num_v_control_points);
}

gfAuxBoundingVolume* gfObj3DBoxLattice::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    // create auxiliary array of mesh points
    gfInt num_points = (gfInt)m_control_points.size();
    if (num_points < 1)
        return NULL;

    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfInt i = 0; i < num_points; i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        curr_point->Set(pt->X(), pt->Y(), pt->Z());
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    //delete [] points;
    return result;
}

GFRESULT gfObj3DBoxLattice::DooSabinSubdivisionStep()
{
    gfInt new_num_u = 2*(m_num_u_control_points-1);
    gfInt new_num_v = 2*(m_num_v_control_points-1);
    gfInt new_num_w = 2*(m_num_w_control_points-1);

    std::vector<gfPoint3> new_points;
    new_points.resize(new_num_u*new_num_v*new_num_w);

    for (gfInt alpha = 0;alpha < 2;alpha++)
    {
        for (gfInt beta = 0;beta < 2;beta++)
        {
            for (gfInt gama = 0;gama < 2;gama++)
            {
                for (gfInt i = 0;i < m_num_u_control_points-1;i++)
                {
                    for (gfInt j = 0;j < m_num_v_control_points-1;j++)
                    {
                        for (gfInt k = 0;k < m_num_w_control_points-1;k++)
                        {
                            gfInt new_index = (2*i+alpha)+(2*j+beta)*new_num_u+(2*k+gama)*new_num_u*new_num_v;

                            gfDouble x = 27.0/64.0*GetControlPoint(i+alpha,j+beta,k+gama)->X()+
                                         9.0/64.0*GetControlPoint(i+((alpha+1)%2),j+beta,k+gama)->X()+
                                         9.0/64.0*GetControlPoint(i+alpha,j+((beta+1)%2),k+gama)->X()+
                                         9.0/64.0*GetControlPoint(i+alpha,j+beta,k+((gama+1)%2))->X()+
                                         3.0/64.0*GetControlPoint(i+((alpha+1)%2),j+((beta+1)%2),k+gama)->X()+
                                         3.0/64.0*GetControlPoint(i+((alpha+1)%2),j+beta,k+((gama+1)%2))->X()+
                                         3.0/64.0*GetControlPoint(i+alpha,j+((beta+1)%2),k+((gama+1)%2))->X()+
                                         1.0/64.0*GetControlPoint(i+((alpha+1)%2),j+((beta+1)%2),k+((gama+1)%2))->X();

                            gfDouble y = 27.0/64.0*GetControlPoint(i+alpha,j+beta,k+gama)->Y()+
                                         9.0/64.0*GetControlPoint(i+((alpha+1)%2),j+beta,k+gama)->Y()+
                                         9.0/64.0*GetControlPoint(i+alpha,j+((beta+1)%2),k+gama)->Y()+
                                         9.0/64.0*GetControlPoint(i+alpha,j+beta,k+((gama+1)%2))->Y()+
                                         3.0/64.0*GetControlPoint(i+((alpha+1)%2),j+((beta+1)%2),k+gama)->Y()+
                                         3.0/64.0*GetControlPoint(i+((alpha+1)%2),j+beta,k+((gama+1)%2))->Y()+
                                         3.0/64.0*GetControlPoint(i+alpha,j+((beta+1)%2),k+((gama+1)%2))->Y()+
                                         1.0/64.0*GetControlPoint(i+((alpha+1)%2),j+((beta+1)%2),k+((gama+1)%2))->Y();

                            gfDouble z = 27.0/64.0*GetControlPoint(i+alpha,j+beta,k+gama)->Z()+
                                         9.0/64.0*GetControlPoint(i+((alpha+1)%2),j+beta,k+gama)->Z()+
                                         9.0/64.0*GetControlPoint(i+alpha,j+((beta+1)%2),k+gama)->Z()+
                                         9.0/64.0*GetControlPoint(i+alpha,j+beta,k+((gama+1)%2))->Z()+
                                         3.0/64.0*GetControlPoint(i+((alpha+1)%2),j+((beta+1)%2),k+gama)->Z()+
                                         3.0/64.0*GetControlPoint(i+((alpha+1)%2),j+beta,k+((gama+1)%2))->Z()+
                                         3.0/64.0*GetControlPoint(i+alpha,j+((beta+1)%2),k+((gama+1)%2))->Z()+
                                         1.0/64.0*GetControlPoint(i+((alpha+1)%2),j+((beta+1)%2),k+((gama+1)%2))->Z();

                            new_points[new_index].Set(x,y,z);
                        }

                    }
                }
            }
        }
    }

    Init(new_num_u, new_num_v, new_num_w);
    for (gfInt i = 0;i < new_num_u;i++)
    {
        for (gfInt j = 0;j < new_num_v;j++)
        {
            for (gfInt k = 0;k < new_num_w;k++)
            {
                gfPoint3 pt = new_points[i+j*new_num_u+k*new_num_u*new_num_v];
                GetControlPoint(i, j, k)->Set(pt.X(), pt.Y(), pt.Z());
            }
        }
    }

    return GF_S_OK;
}

GFRESULT gfObj3DBoxLattice::CatmullClarkSubdivisionStep()
{
    gfInt nu = m_num_u_control_points-1;
    gfInt nv = m_num_v_control_points-1;
    gfInt nw = m_num_w_control_points-1;
    gfInt new_num_u = 2 * nu + 1;
    gfInt new_num_v = 2 * nv + 1;
    gfInt new_num_w = 2 * nw + 1;

    std::vector<gfPoint3> new_points;
    new_points.resize(new_num_u*new_num_v*new_num_w);

    // cell points
    for (gfInt i = 0;i < nu;i++)
    {
        for (gfInt j = 0;j < nv;j++)
        {
            for (gfInt k = 0;k < nw;k++)
            {
                gfInt new_index = (2*i+1)+(2*j+1)*new_num_u+(2*k+1)*new_num_u*new_num_v;

                gfDouble x = 1.0/8.0*GetControlPoint(i,j,k)->X()+
                    1.0/8.0*GetControlPoint(i+1,j,k)->X()+
                    1.0/8.0*GetControlPoint(i,j+1,k)->X()+
                    1.0/8.0*GetControlPoint(i,j,k+1)->X()+
                    1.0/8.0*GetControlPoint(i+1,j+1,k)->X()+
                    1.0/8.0*GetControlPoint(i+1,j,k+1)->X()+
                    1.0/8.0*GetControlPoint(i,j+1,k+1)->X()+
                    1.0/8.0*GetControlPoint(i+1,j+1,k+1)->X();

                gfDouble y = 1.0/8.0*GetControlPoint(i,j,k)->Y()+
                    1.0/8.0*GetControlPoint(i+1,j,k)->Y()+
                    1.0/8.0*GetControlPoint(i,j+1,k)->Y()+
                    1.0/8.0*GetControlPoint(i,j,k+1)->Y()+
                    1.0/8.0*GetControlPoint(i+1,j+1,k)->Y()+
                    1.0/8.0*GetControlPoint(i+1,j,k+1)->Y()+
                    1.0/8.0*GetControlPoint(i,j+1,k+1)->Y()+
                    1.0/8.0*GetControlPoint(i+1,j+1,k+1)->Y();

                gfDouble z = 1.0/8.0*GetControlPoint(i,j,k)->Z()+
                    1.0/8.0*GetControlPoint(i+1,j,k)->Z()+
                    1.0/8.0*GetControlPoint(i,j+1,k)->Z()+
                    1.0/8.0*GetControlPoint(i,j,k+1)->Z()+
                    1.0/8.0*GetControlPoint(i+1,j+1,k)->Z()+
                    1.0/8.0*GetControlPoint(i+1,j,k+1)->Z()+
                    1.0/8.0*GetControlPoint(i,j+1,k+1)->Z()+
                    1.0/8.0*GetControlPoint(i+1,j+1,k+1)->Z();

                new_points[new_index].Set(x,y,z);
            }
        }
    }

    // face points 1
    for (gfInt i = 0;i <= nu;i++)
    {
        for (gfInt j = 0;j < nv;j++)
        {
            for (gfInt k = 0;k < nw;k++)
            {
                gfInt new_index = (2*i)+(2*j+1)*new_num_u+(2*k+1)*new_num_u*new_num_v;

                gfDouble x = 3.0/16.0*GetControlPoint(i,j,k)->X()+
                    3.0/16.0*GetControlPoint(i,j+1,k)->X()+
                    3.0/16.0*GetControlPoint(i,j,k+1)->X()+
                    3.0/16.0*GetControlPoint(i,j+1,k+1)->X()+
                    1.0/32.0*GetControlPoint(i-1,j,k)->X()+
                    1.0/32.0*GetControlPoint(i-1,j+1,k)->X()+
                    1.0/32.0*GetControlPoint(i-1,j,k+1)->X()+
                    1.0/32.0*GetControlPoint(i-1,j+1,k+1)->X()+
                    1.0/32.0*GetControlPoint(i+1,j,k)->X()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k)->X()+
                    1.0/32.0*GetControlPoint(i+1,j,k+1)->X()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k+1)->X();

                gfDouble y = 3.0/16.0*GetControlPoint(i,j,k)->Y()+
                    3.0/16.0*GetControlPoint(i,j+1,k)->Y()+
                    3.0/16.0*GetControlPoint(i,j,k+1)->Y()+
                    3.0/16.0*GetControlPoint(i,j+1,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i-1,j,k)->Y()+
                    1.0/32.0*GetControlPoint(i-1,j+1,k)->Y()+
                    1.0/32.0*GetControlPoint(i-1,j,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i-1,j+1,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j,k)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k+1)->Y();

                gfDouble z = 3.0/16.0*GetControlPoint(i,j,k)->Z()+
                    3.0/16.0*GetControlPoint(i,j+1,k)->Z()+
                    3.0/16.0*GetControlPoint(i,j,k+1)->Z()+
                    3.0/16.0*GetControlPoint(i,j+1,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i-1,j,k)->Z()+
                    1.0/32.0*GetControlPoint(i-1,j+1,k)->Z()+
                    1.0/32.0*GetControlPoint(i-1,j,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i-1,j+1,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j,k)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k+1)->Z();

                new_points[new_index].Set(x,y,z);
            }
        }
    }

    // face points 2
    for (gfInt i = 0;i < nu;i++)
    {
        for (gfInt j = 0;j <= nv;j++)
        {
            for (gfInt k = 0;k < nw;k++)
            {
                gfInt new_index = (2*i+1)+(2*j)*new_num_u+(2*k+1)*new_num_u*new_num_v;

                gfDouble x = 3.0/16.0*GetControlPoint(i,j,k)->X()+
                    3.0/16.0*GetControlPoint(i+1,j,k)->X()+
                    3.0/16.0*GetControlPoint(i,j,k+1)->X()+
                    3.0/16.0*GetControlPoint(i+1,j,k+1)->X()+
                    1.0/32.0*GetControlPoint(i,j-2,k)->X()+
                    1.0/32.0*GetControlPoint(i+1,j-1,k)->X()+
                    1.0/32.0*GetControlPoint(i,j-1,k+1)->X()+
                    1.0/32.0*GetControlPoint(i+1,j-1,k+1)->X()+
                    1.0/32.0*GetControlPoint(i,j+1,k)->X()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k)->X()+
                    1.0/32.0*GetControlPoint(i,j+1,k+1)->X()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k+1)->X();

                gfDouble y = 3.0/16.0*GetControlPoint(i,j,k)->Y()+
                    3.0/16.0*GetControlPoint(i+1,j,k)->Y()+
                    3.0/16.0*GetControlPoint(i,j,k+1)->Y()+
                    3.0/16.0*GetControlPoint(i+1,j,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i,j-2,k)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j-1,k)->Y()+
                    1.0/32.0*GetControlPoint(i,j-1,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j-1,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i,j+1,k)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k)->Y()+
                    1.0/32.0*GetControlPoint(i,j+1,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k+1)->Y();

                gfDouble z = 3.0/16.0*GetControlPoint(i,j,k)->Z()+
                    3.0/16.0*GetControlPoint(i+1,j,k)->Z()+
                    3.0/16.0*GetControlPoint(i,j,k+1)->Z()+
                    3.0/16.0*GetControlPoint(i+1,j,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i,j-2,k)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j-1,k)->Z()+
                    1.0/32.0*GetControlPoint(i,j-1,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j-1,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i,j+1,k)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k)->Z()+
                    1.0/32.0*GetControlPoint(i,j+1,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k+1)->Z();

                new_points[new_index].Set(x,y,z);
            }
        }
    }

    // face points 3
    for (gfInt i = 0;i < nu;i++)
    {
        for (gfInt j = 0;j < nv;j++)
        {
            for (gfInt k = 0;k <= nw;k++)
            {
                gfInt new_index = (2*i+1)+(2*j+1)*new_num_u+(2*k)*new_num_u*new_num_v;

                gfDouble x = 3.0/16.0*GetControlPoint(i,j,k)->X()+
                    3.0/16.0*GetControlPoint(i,j+1,k)->X()+
                    3.0/16.0*GetControlPoint(i+1,j,k)->X()+
                    3.0/16.0*GetControlPoint(i+1,j+1,k)->X()+
                    1.0/32.0*GetControlPoint(i,j,k-1)->X()+
                    1.0/32.0*GetControlPoint(i,j+1,k-1)->X()+
                    1.0/32.0*GetControlPoint(i+1,j,k-1)->X()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k-1)->X()+
                    1.0/32.0*GetControlPoint(i,j,k+1)->X()+
                    1.0/32.0*GetControlPoint(i,j+1,k+1)->X()+
                    1.0/32.0*GetControlPoint(i+1,j,k+1)->X()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k+1)->X();

                gfDouble y = 3.0/16.0*GetControlPoint(i,j,k)->Y()+
                    3.0/16.0*GetControlPoint(i,j+1,k)->Y()+
                    3.0/16.0*GetControlPoint(i+1,j,k)->Y()+
                    3.0/16.0*GetControlPoint(i+1,j+1,k)->Y()+
                    1.0/32.0*GetControlPoint(i,j,k-1)->Y()+
                    1.0/32.0*GetControlPoint(i,j+1,k-1)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j,k-1)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k-1)->Y()+
                    1.0/32.0*GetControlPoint(i,j,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i,j+1,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j,k+1)->Y()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k+1)->Y();

                gfDouble z = 3.0/16.0*GetControlPoint(i,j,k)->Z()+
                    3.0/16.0*GetControlPoint(i,j+1,k)->Z()+
                    3.0/16.0*GetControlPoint(i+1,j,k)->Z()+
                    3.0/16.0*GetControlPoint(i+1,j+1,k)->Z()+
                    1.0/32.0*GetControlPoint(i,j,k-1)->Z()+
                    1.0/32.0*GetControlPoint(i,j+1,k-1)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j,k-1)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k-1)->Z()+
                    1.0/32.0*GetControlPoint(i,j,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i,j+1,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j,k+1)->Z()+
                    1.0/32.0*GetControlPoint(i+1,j+1,k+1)->Z();

                new_points[new_index].Set(x,y,z);
            }
        }
    }

    // edge points 1
    for (gfInt i = 0;i < nu;i++)
    {
        for (gfInt j = 0;j <= nv;j++)
        {
            for (gfInt k = 0;k <= nw;k++)
            {
                gfInt new_index = (2*i+1)+(2*j)*new_num_u+(2*k)*new_num_u*new_num_v;

                gfDouble x = 9.0/32.0*GetControlPoint(i,j,k)->X()+
                    9.0/32.0*GetControlPoint(i+1,j,k)->X()+
                    3.0/64.0*GetControlPoint(i,j,k-1)->X()+
                    3.0/64.0*GetControlPoint(i,j-1,k)->X()+
                    3.0/64.0*GetControlPoint(i,j,k+1)->X()+
                    3.0/64.0*GetControlPoint(i,j+1,k)->X()+
                    3.0/64.0*GetControlPoint(i+1,j,k-1)->X()+
                    3.0/64.0*GetControlPoint(i+1,j-1,k)->X()+
                    3.0/64.0*GetControlPoint(i+1,j,k+1)->X()+
                    3.0/64.0*GetControlPoint(i+1,j+1,k)->X()+
                    1.0/128.0*GetControlPoint(i,j-1,k-1)->X()+
                    1.0/128.0*GetControlPoint(i,j+1,k-1)->X()+
                    1.0/128.0*GetControlPoint(i,j-1,k+1)->X()+
                    1.0/128.0*GetControlPoint(i,j+1,k+1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k-1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k-1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k+1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k+1)->X();

                gfDouble y = 9.0/32.0*GetControlPoint(i,j,k)->Y()+
                    9.0/32.0*GetControlPoint(i+1,j,k)->Y()+
                    3.0/64.0*GetControlPoint(i,j,k-1)->Y()+
                    3.0/64.0*GetControlPoint(i,j-1,k)->Y()+
                    3.0/64.0*GetControlPoint(i,j,k+1)->Y()+
                    3.0/64.0*GetControlPoint(i,j+1,k)->Y()+
                    3.0/64.0*GetControlPoint(i+1,j,k-1)->Y()+
                    3.0/64.0*GetControlPoint(i+1,j-1,k)->Y()+
                    3.0/64.0*GetControlPoint(i+1,j,k+1)->Y()+
                    3.0/64.0*GetControlPoint(i+1,j+1,k)->Y()+
                    1.0/128.0*GetControlPoint(i,j-1,k-1)->Y()+
                    1.0/128.0*GetControlPoint(i,j+1,k-1)->Y()+
                    1.0/128.0*GetControlPoint(i,j-1,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i,j+1,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k-1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k-1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k+1)->Y();

                gfDouble z = 9.0/32.0*GetControlPoint(i,j,k)->Z()+
                    9.0/32.0*GetControlPoint(i+1,j,k)->Z()+
                    3.0/64.0*GetControlPoint(i,j,k-1)->Z()+
                    3.0/64.0*GetControlPoint(i,j-1,k)->Z()+
                    3.0/64.0*GetControlPoint(i,j,k+1)->Z()+
                    3.0/64.0*GetControlPoint(i,j+1,k)->Z()+
                    3.0/64.0*GetControlPoint(i+1,j,k-1)->Z()+
                    3.0/64.0*GetControlPoint(i+1,j-1,k)->Z()+
                    3.0/64.0*GetControlPoint(i+1,j,k+1)->Z()+
                    3.0/64.0*GetControlPoint(i+1,j+1,k)->Z()+
                    1.0/128.0*GetControlPoint(i,j-1,k-1)->Z()+
                    1.0/128.0*GetControlPoint(i,j+1,k-1)->Z()+
                    1.0/128.0*GetControlPoint(i,j-1,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i,j+1,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k-1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k-1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k+1)->Z();

                new_points[new_index].Set(x,y,z);
            }
        }
    }

    // edge points 2
    for (gfInt i = 0;i <= nu;i++)
    {
        for (gfInt j = 0;j < nv;j++)
        {
            for (gfInt k = 0;k <= nw;k++)
            {
                gfInt new_index = (2*i)+(2*j+1)*new_num_u+(2*k)*new_num_u*new_num_v;

                gfDouble x = 9.0/32.0*GetControlPoint(i,j,k)->X()+
                    9.0/32.0*GetControlPoint(i,j+1,k)->X()+
                    3.0/64.0*GetControlPoint(i,j,k-1)->X()+
                    3.0/64.0*GetControlPoint(i-1,j,k)->X()+
                    3.0/64.0*GetControlPoint(i,j,k+1)->X()+
                    3.0/64.0*GetControlPoint(i+1,j,k)->X()+
                    3.0/64.0*GetControlPoint(i,j+1,k-1)->X()+
                    3.0/64.0*GetControlPoint(i-1,j+1,k)->X()+
                    3.0/64.0*GetControlPoint(i,j+1,k+1)->X()+
                    3.0/64.0*GetControlPoint(i+1,j+1,k)->X()+
                    1.0/128.0*GetControlPoint(i-1,j,k-1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j,k-1)->X()+
                    1.0/128.0*GetControlPoint(i-1,j,k+1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j,k+1)->X()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k-1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k-1)->X()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k+1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k+1)->X();

                gfDouble y = 9.0/32.0*GetControlPoint(i,j,k)->Y()+
                    9.0/32.0*GetControlPoint(i,j+1,k)->Y()+
                    3.0/64.0*GetControlPoint(i,j,k-1)->Y()+
                    3.0/64.0*GetControlPoint(i-1,j,k)->Y()+
                    3.0/64.0*GetControlPoint(i,j,k+1)->Y()+
                    3.0/64.0*GetControlPoint(i+1,j,k)->Y()+
                    3.0/64.0*GetControlPoint(i,j+1,k-1)->Y()+
                    3.0/64.0*GetControlPoint(i-1,j+1,k)->Y()+
                    3.0/64.0*GetControlPoint(i,j+1,k+1)->Y()+
                    3.0/64.0*GetControlPoint(i+1,j+1,k)->Y()+
                    1.0/128.0*GetControlPoint(i-1,j,k-1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j,k-1)->Y()+
                    1.0/128.0*GetControlPoint(i-1,j,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k-1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k-1)->Y()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k+1)->Y();

                gfDouble z = 9.0/32.0*GetControlPoint(i,j,k)->Z()+
                    9.0/32.0*GetControlPoint(i,j+1,k)->Z()+
                    3.0/64.0*GetControlPoint(i,j,k-1)->Z()+
                    3.0/64.0*GetControlPoint(i-1,j,k)->Z()+
                    3.0/64.0*GetControlPoint(i,j,k+1)->Z()+
                    3.0/64.0*GetControlPoint(i+1,j,k)->Z()+
                    3.0/64.0*GetControlPoint(i,j+1,k-1)->Z()+
                    3.0/64.0*GetControlPoint(i-1,j+1,k)->Z()+
                    3.0/64.0*GetControlPoint(i,j+1,k+1)->Z()+
                    3.0/64.0*GetControlPoint(i+1,j+1,k)->Z()+
                    1.0/128.0*GetControlPoint(i-1,j,k-1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j,k-1)->Z()+
                    1.0/128.0*GetControlPoint(i-1,j,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k-1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k-1)->Z()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k+1)->Z();

                new_points[new_index].Set(x,y,z);
            }
        }
    }

    // edge points 3
    for (gfInt i = 0;i <= nu;i++)
    {
        for (gfInt j = 0;j <= nv;j++)
        {
            for (gfInt k = 0;k < nw;k++)
            {
                gfInt new_index = (2*i)+(2*j)*new_num_u+(2*k+1)*new_num_u*new_num_v;

                gfDouble x = 9.0/32.0*GetControlPoint(i,j,k)->X()+
                    9.0/32.0*GetControlPoint(i,j,k+1)->X()+
                    3.0/64.0*GetControlPoint(i,j-1,k)->X()+
                    3.0/64.0*GetControlPoint(i-1,j,k)->X()+
                    3.0/64.0*GetControlPoint(i,j+1,k)->X()+
                    3.0/64.0*GetControlPoint(i+1,j,k)->X()+
                    3.0/64.0*GetControlPoint(i,j-1,k+1)->X()+
                    3.0/64.0*GetControlPoint(i-1,j,k+1)->X()+
                    3.0/64.0*GetControlPoint(i,j+1,k+1)->X()+
                    3.0/64.0*GetControlPoint(i+1,j,k+1)->X()+
                    1.0/128.0*GetControlPoint(i-1,j-1,k)->X()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k)->X()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k)->X()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k)->X()+
                    1.0/128.0*GetControlPoint(i-1,j-1,k+1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k+1)->X()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k+1)->X()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k+1)->X();

                gfDouble y = 9.0/32.0*GetControlPoint(i,j,k)->Y()+
                    9.0/32.0*GetControlPoint(i,j,k+1)->Y()+
                    3.0/64.0*GetControlPoint(i,j-1,k)->Y()+
                    3.0/64.0*GetControlPoint(i-1,j,k)->Y()+
                    3.0/64.0*GetControlPoint(i,j+1,k)->Y()+
                    3.0/64.0*GetControlPoint(i+1,j,k)->Y()+
                    3.0/64.0*GetControlPoint(i,j-1,k+1)->Y()+
                    3.0/64.0*GetControlPoint(i-1,j,k+1)->Y()+
                    3.0/64.0*GetControlPoint(i,j+1,k+1)->Y()+
                    3.0/64.0*GetControlPoint(i+1,j,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i-1,j-1,k)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k)->Y()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k)->Y()+
                    1.0/128.0*GetControlPoint(i-1,j-1,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k+1)->Y()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k+1)->Y();

                gfDouble z = 9.0/32.0*GetControlPoint(i,j,k)->Z()+
                    9.0/32.0*GetControlPoint(i,j,k+1)->Z()+
                    3.0/64.0*GetControlPoint(i,j-1,k)->Z()+
                    3.0/64.0*GetControlPoint(i-1,j,k)->Z()+
                    3.0/64.0*GetControlPoint(i,j+1,k)->Z()+
                    3.0/64.0*GetControlPoint(i+1,j,k)->Z()+
                    3.0/64.0*GetControlPoint(i,j-1,k+1)->Z()+
                    3.0/64.0*GetControlPoint(i-1,j,k+1)->Z()+
                    3.0/64.0*GetControlPoint(i,j+1,k+1)->Z()+
                    3.0/64.0*GetControlPoint(i+1,j,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i-1,j-1,k)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k)->Z()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k)->Z()+
                    1.0/128.0*GetControlPoint(i-1,j-1,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j-1,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i-1,j+1,k+1)->Z()+
                    1.0/128.0*GetControlPoint(i+1,j+1,k+1)->Z();

                new_points[new_index].Set(x,y,z);
            }
        }
    }

    // vertex points
    for (gfInt i = 0;i <= nu;i++)
    {
        for (gfInt j = 0;j <= nv;j++)
        {
            for (gfInt k = 0;k <= nw;k++)
            {
                gfInt new_index = (2*i)+(2*j)*new_num_u+(2*k)*new_num_u*new_num_v;

                gfDouble x = 27.0/64.0*GetControlPoint(i,j,k)->X()+
                    9.0/128.0*GetControlPoint(i-1,j,k)->X()+
                    9.0/128.0*GetControlPoint(i+1,j,k)->X()+
                    9.0/128.0*GetControlPoint(i,j-1,k)->X()+
                    9.0/128.0*GetControlPoint(i,j+1,k)->X()+
                    9.0/128.0*GetControlPoint(i,j,k-1)->X()+
                    9.0/128.0*GetControlPoint(i,j,k+1)->X()+
                    3.0/256.0*GetControlPoint(i,j-1,k-1)->X()+
                    3.0/256.0*GetControlPoint(i,j+1,k-1)->X()+
                    3.0/256.0*GetControlPoint(i,j-1,k+1)->X()+
                    3.0/256.0*GetControlPoint(i,j+1,k+1)->X()+
                    3.0/256.0*GetControlPoint(i-1,j,k-1)->X()+
                    3.0/256.0*GetControlPoint(i+1,j,k-1)->X()+
                    3.0/256.0*GetControlPoint(i-1,j,k+1)->X()+
                    3.0/256.0*GetControlPoint(i+1,j,k+1)->X()+
                    3.0/256.0*GetControlPoint(i-1,j-1,k)->X()+
                    3.0/256.0*GetControlPoint(i+1,j-1,k)->X()+
                    3.0/256.0*GetControlPoint(i-1,j+1,k)->X()+
                    3.0/256.0*GetControlPoint(i+1,j+1,k)->X()+
                    1.0/512.0*GetControlPoint(i-1,j-1,k-1)->X()+
                    1.0/512.0*GetControlPoint(i-1,j+1,k-1)->X()+
                    1.0/512.0*GetControlPoint(i-1,j-1,k+1)->X()+
                    1.0/512.0*GetControlPoint(i-1,j+1,k+1)->X()+
                    1.0/512.0*GetControlPoint(i+1,j-1,k-1)->X()+
                    1.0/512.0*GetControlPoint(i+1,j+1,k-1)->X()+
                    1.0/512.0*GetControlPoint(i+1,j-1,k+1)->X()+
                    1.0/512.0*GetControlPoint(i+1,j+1,k+1)->X();

                gfDouble y = 27.0/64.0*GetControlPoint(i,j,k)->Y()+
                    9.0/128.0*GetControlPoint(i-1,j,k)->Y()+
                    9.0/128.0*GetControlPoint(i+1,j,k)->Y()+
                    9.0/128.0*GetControlPoint(i,j-1,k)->Y()+
                    9.0/128.0*GetControlPoint(i,j+1,k)->Y()+
                    9.0/128.0*GetControlPoint(i,j,k-1)->Y()+
                    9.0/128.0*GetControlPoint(i,j,k+1)->Y()+
                    3.0/256.0*GetControlPoint(i,j-1,k-1)->Y()+
                    3.0/256.0*GetControlPoint(i,j+1,k-1)->Y()+
                    3.0/256.0*GetControlPoint(i,j-1,k+1)->Y()+
                    3.0/256.0*GetControlPoint(i,j+1,k+1)->Y()+
                    3.0/256.0*GetControlPoint(i-1,j,k-1)->Y()+
                    3.0/256.0*GetControlPoint(i+1,j,k-1)->Y()+
                    3.0/256.0*GetControlPoint(i-1,j,k+1)->Y()+
                    3.0/256.0*GetControlPoint(i+1,j,k+1)->Y()+
                    3.0/256.0*GetControlPoint(i-1,j-1,k)->Y()+
                    3.0/256.0*GetControlPoint(i+1,j-1,k)->Y()+
                    3.0/256.0*GetControlPoint(i-1,j+1,k)->Y()+
                    3.0/256.0*GetControlPoint(i+1,j+1,k)->Y()+
                    1.0/512.0*GetControlPoint(i-1,j-1,k-1)->Y()+
                    1.0/512.0*GetControlPoint(i-1,j+1,k-1)->Y()+
                    1.0/512.0*GetControlPoint(i-1,j-1,k+1)->Y()+
                    1.0/512.0*GetControlPoint(i-1,j+1,k+1)->Y()+
                    1.0/512.0*GetControlPoint(i+1,j-1,k-1)->Y()+
                    1.0/512.0*GetControlPoint(i+1,j+1,k-1)->Y()+
                    1.0/512.0*GetControlPoint(i+1,j-1,k+1)->Y()+
                    1.0/512.0*GetControlPoint(i+1,j+1,k+1)->Y();

                gfDouble z = 27.0/64.0*GetControlPoint(i,j,k)->Z()+
                    9.0/128.0*GetControlPoint(i-1,j,k)->Z()+
                    9.0/128.0*GetControlPoint(i+1,j,k)->Z()+
                    9.0/128.0*GetControlPoint(i,j-1,k)->Z()+
                    9.0/128.0*GetControlPoint(i,j+1,k)->Z()+
                    9.0/128.0*GetControlPoint(i,j,k-1)->Z()+
                    9.0/128.0*GetControlPoint(i,j,k+1)->Z()+
                    3.0/256.0*GetControlPoint(i,j-1,k-1)->Z()+
                    3.0/256.0*GetControlPoint(i,j+1,k-1)->Z()+
                    3.0/256.0*GetControlPoint(i,j-1,k+1)->Z()+
                    3.0/256.0*GetControlPoint(i,j+1,k+1)->Z()+
                    3.0/256.0*GetControlPoint(i-1,j,k-1)->Z()+
                    3.0/256.0*GetControlPoint(i+1,j,k-1)->Z()+
                    3.0/256.0*GetControlPoint(i-1,j,k+1)->Z()+
                    3.0/256.0*GetControlPoint(i+1,j,k+1)->Z()+
                    3.0/256.0*GetControlPoint(i-1,j-1,k)->Z()+
                    3.0/256.0*GetControlPoint(i+1,j-1,k)->Z()+
                    3.0/256.0*GetControlPoint(i-1,j+1,k)->Z()+
                    3.0/256.0*GetControlPoint(i+1,j+1,k)->Z()+
                    1.0/512.0*GetControlPoint(i-1,j-1,k-1)->Z()+
                    1.0/512.0*GetControlPoint(i-1,j+1,k-1)->Z()+
                    1.0/512.0*GetControlPoint(i-1,j-1,k+1)->Z()+
                    1.0/512.0*GetControlPoint(i-1,j+1,k+1)->Z()+
                    1.0/512.0*GetControlPoint(i+1,j-1,k-1)->Z()+
                    1.0/512.0*GetControlPoint(i+1,j+1,k-1)->Z()+
                    1.0/512.0*GetControlPoint(i+1,j-1,k+1)->Z()+
                    1.0/512.0*GetControlPoint(i+1,j+1,k+1)->Z();

                new_points[new_index].Set(x,y,z);
            }
        }
    }

    Init(new_num_u, new_num_v, new_num_w);
    for (gfInt i = 0;i < new_num_u;i++)
    {
        for (gfInt j = 0;j < new_num_v;j++)
        {
            for (gfInt k = 0;k < new_num_w;k++)
            {
                gfPoint3 pt = new_points[i+j*new_num_u+k*new_num_u*new_num_v];
                GetControlPoint(i, j, k)->Set(pt.X(), pt.Y(), pt.Z());
            }
        }
    }

    return GF_S_OK;
}