//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjBezierSPatch.cpp                              //
//   Description: Rational Bezier triangle class                //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfObjBezierSPatch.h"
#include "gfObjectFactory.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfWinFPSCounter.h"
#include "gfAuxLogging.h"
#include <vector>

#include "./OGL/glut.h"
#include "./LinearSolver/DirectSolver.h"


gfInt GetIndexFromMultiindex(gfInt degree, std::vector<gfInt>& multi_index)
{
    gfInt multi = (gfInt)multi_index.size();
    gfInt index = 0;
    for (int i = 0;i < multi;i++)
    {
        gfInt ii = multi_index.at(i);
        index += multi_index.at(i) * pow(degree+1, i);
    }
    return index;
}

struct gfAuxTriangle2
{
    gfPoint2 pt1;
    gfPoint2 pt2;
    gfPoint2 pt3;
};

void SampleDomainTriangle(gfInt prec, gfPoint2 pt1, gfPoint2 pt2, gfPoint2 pt3, std::vector<gfAuxTriangle2>& triangles)
{
    // sampling using barycentric coordinates
    for (gfInt i = 1;i < prec - 0;i++)
    {
        for (gfInt j = 0;j < prec - i;j++)
        {
            gfAuxTriangle2 tri1;
            gfAuxTriangle2 tri2;

            // compute sample triangles
            gfDouble u = (gfDouble) i / (prec - 1);
            gfDouble v = (gfDouble) j / (prec - 1);
            gfDouble w = 1.0 - u - v;
            tri1.pt1.SetX(u*pt1.X()+v*pt2.X()+w*pt3.X());
            tri1.pt1.SetY(u*pt1.Y()+v*pt2.Y()+w*pt3.Y());
            tri2.pt1.SetX(u*pt1.X()+v*pt2.X()+w*pt3.X());
            tri2.pt1.SetY(u*pt1.Y()+v*pt2.Y()+w*pt3.Y());

            u = (gfDouble) (i - 1)/ (prec - 1);
            v = (gfDouble) j / (prec - 1);
            w = 1.0 - u - v;
            tri1.pt3.SetX(u*pt1.X()+v*pt2.X()+w*pt3.X());
            tri1.pt3.SetY(u*pt1.Y()+v*pt2.Y()+w*pt3.Y());
            tri2.pt2.SetX(u*pt1.X()+v*pt2.X()+w*pt3.X());
            tri2.pt2.SetY(u*pt1.Y()+v*pt2.Y()+w*pt3.Y());

            u = (gfDouble) (i - 1)/ (prec - 1);
            v = (gfDouble) (j + 1) / (prec - 1);
            w = 1.0 - u - v;
            tri1.pt2.SetX(u*pt1.X()+v*pt2.X()+w*pt3.X());
            tri1.pt2.SetY(u*pt1.Y()+v*pt2.Y()+w*pt3.Y());

            triangles.push_back(tri1);

            if (j > 0)
            {
                u = (gfDouble) i / (prec - 1);
                v = (gfDouble) (j - 1) / (prec - 1);
                w = 1.0 - u - v;
                tri2.pt3.SetX(u*pt1.X()+v*pt2.X()+w*pt3.X());
                tri2.pt3.SetY(u*pt1.Y()+v*pt2.Y()+w*pt3.Y());

                triangles.push_back(tri2);
            }
        }
    }
}

gfObjBezierSPatch::gfObjBezierSPatch(void)
{
    m_time_stamp = 0;
    m_weights.resize(0);
    m_control_points.resize(0);
    m_degree = 0;
    m_multi = 0;
    m_old_stamp = 0;
    m_wachspress = true;
}

gfObjBezierSPatch::~gfObjBezierSPatch(void)
{
    SetParams(0, 0);
}

GFRESULT gfObjBezierSPatch::OGLDraw(gfLong flags)
{
    if (m_time_stamp == 0)
        return GF_S_OK;

    gfInt num_indices = (gfInt)m_multiindices.size();

    // draw possibly control points, it isn't cached
    if (flags & GF_FLG_CONTROL_NET)
    {
        GLfloat ambient[4];
        ambient[0] = ambient[1] = ambient[2] = ambient[3] = 0.0;
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, ambient);
        glColor3f(0.0, 0.0, 0.0);

        std::vector<gfInt> aux_multi_index;
        glBegin(GL_LINES);
        for (gfInt i = 0;i < num_indices;i++)
        {
            std::vector<gfInt> multi_index = m_multiindices[i];
            gfPoint3* pt = GetControlPoint(multi_index);

            // in this multi index, create all +1,-1 combinations
            for (gfInt k = 0;k < m_multi;k++)
            {
                if ((multi_index[k] + 1) > m_degree)
                    continue;

                for (gfInt l = 0;l < m_multi;l++)
                {
                    if (k == l)
                        continue;

                    if ((multi_index[l] - 1) < 0)
                        continue;

                    glVertex3f(pt->X(), pt->Y(), pt->Z());
                    aux_multi_index = multi_index;
                    aux_multi_index[k]++;
                    aux_multi_index[l]--;
                    gfPoint3* aux_pt = GetControlPoint(aux_multi_index);
                    glVertex3f(aux_pt->X(), aux_pt->Y(), aux_pt->Z());
                }
            }
        }
        glEnd();

        glPointSize(8.0);
        glBegin(GL_POINTS);
        for (gfInt i = 0;i < num_indices;i++)
        {
            std::vector<gfInt> multi_index(m_multi, 0);
            multi_index = m_multiindices[i];

            gfPoint3* pt = GetControlPoint(multi_index);
            glVertex3f(pt->X(), pt->Y(), pt->Z());
        }
        glEnd();
        glPointSize(1.0);
    }

    // here domain is regular m-polygon with equal sides
    // sampling domain
    if (m_time_stamp != m_old_stamp)
    {
        gfInt prec = GF_VIS_PREC;

        m_cached_points.clear();
        m_cached_normals.clear();

        gfPoint2 center(0.0, 0.0);
        for (gfInt i = 0;i < m_multi;i++)
        {
            std::vector<gfAuxTriangle2> triangles;
            SampleDomainTriangle(prec, center, m_domain_points[i], m_domain_points[(i+1)%m_multi], triangles);
            gfInt num_tris = (gfInt)triangles.size();
            for (gfInt k = 0;k < num_tris;k++)
            {
                gfAuxTriangle2 tri = triangles[k];
                // have element triangle, compute surface points
                // first compute barycentric coordinates

                std::vector<gfDouble> coordinates1;
                std::vector<gfDouble> coordinates2;
                std::vector<gfDouble> coordinates3;

                if (!m_wachspress)
                {
                    gfCompGenBaryCoordsMean2(m_domain_points, tri.pt1, coordinates1);
                    gfCompGenBaryCoordsMean2(m_domain_points, tri.pt2, coordinates2);
                    gfCompGenBaryCoordsMean2(m_domain_points, tri.pt3, coordinates3);
                }
                else
                {
                    gfCompGenBaryCoordsWachspress2(m_domain_points, tri.pt1, coordinates1);
                    gfCompGenBaryCoordsWachspress2(m_domain_points, tri.pt2, coordinates2);
                    gfCompGenBaryCoordsWachspress2(m_domain_points, tri.pt3, coordinates3);
                }

                gfPoint3 result1(0.0, 0.0, 0.0);
                gfPoint3 result2(0.0, 0.0, 0.0);
                gfPoint3 result3(0.0, 0.0, 0.0);
                gfDouble weight1 = 0.0;
                gfDouble weight2 = 0.0;
                gfDouble weight3 = 0.0;
                for (gfInt j = 0;j < num_indices;j++)
                {
                    gfDouble w = GetWeight(m_multiindices[j]);
                    gfDouble coef1 = w * gfBernsteinPolyGeneral(m_multiindices[j], coordinates1);
                    gfDouble coef2 = w * gfBernsteinPolyGeneral(m_multiindices[j], coordinates2);
                    gfDouble coef3 = w * gfBernsteinPolyGeneral(m_multiindices[j], coordinates3);
                    gfPoint3* pt = GetControlPoint(m_multiindices[j]);
                    result1.Set(result1.X() + pt->X()*coef1, result1.Y() + pt->Y()*coef1, result1.Z() + pt->Z()*coef1);
                    result2.Set(result2.X() + pt->X()*coef2, result2.Y() + pt->Y()*coef2, result2.Z() + pt->Z()*coef2);
                    result3.Set(result3.X() + pt->X()*coef3, result3.Y() + pt->Y()*coef3, result3.Z() + pt->Z()*coef3);
                    weight1 += coef1;
                    weight2 += coef2;
                    weight3 += coef3;
                }
                
                gfDouble invW = 1.0 / weight1;
                result1.Set(result1.X() * invW, result1.Y() * invW, result1.Z() * invW);
                m_cached_points.push_back(result1);
                invW = 1.0 / weight2;
                result2.Set(result2.X() * invW, result2.Y() * invW, result2.Z() * invW);
                m_cached_points.push_back(result2);
                invW = 1.0 / weight3;
                result3.Set(result3.X() * invW, result3.Y() * invW, result3.Z() * invW);
                m_cached_points.push_back(result3);

                // normals
                m_cached_normals.push_back(GetNormal(coordinates1));
                m_cached_normals.push_back(GetNormal(coordinates2));
                m_cached_normals.push_back(GetNormal(coordinates3));

                if (result1.X() == 0 || result1.Y() == 0 && result1.Z() == 0)
                    int k = 00;
                if (result2.X() == 0 && result2.Y() == 0 && result2.Z() == 0)
                    int k = 00;
                if (result3.X() == 0 && result3.Y() == 0 && result3.Z() == 0)
                    int k = 00;
            }
        }
    }
    
    if (flags & GF_FLG_SOLID)
    {
        glBegin(GL_TRIANGLES);
        gfInt num_samples = (gfInt)m_cached_points.size();
        for (gfInt k = 0;k < num_samples;k++)
        {
            // we have coordinate, compute point on surface
            // based on bezier bernstein representation
            gfVector3 normal = m_cached_normals[k];
            if (flags & GF_FLG_NORMALS)
                glNormal3f(normal.X(), normal.Y(), normal.Z());
            gfPoint3 result = m_cached_points[k];
            glVertex3f(result.X(), result.Y(), result.Z());
        }
        glEnd();
    }

    if (flags & GF_FLG_WIRE)
    {
        GLfloat ambient[4];
        ambient[0] = ambient[1] = ambient[2] = ambient[3] = 0.0;
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, ambient);
        glColor3f(0.0, 0.0, 0.0);

        glBegin(GL_LINES);
        gfInt num_samples = (gfInt)m_cached_points.size() / 3;
        for (gfInt k = 0;k < num_samples;k++)
        {
            glVertex3f(m_cached_points[3*k].X(), m_cached_points[3*k].Y(), m_cached_points[3*k].Z());
            glVertex3f(m_cached_points[3*k+1].X(), m_cached_points[3*k+1].Y(), m_cached_points[3*k+1].Z());
            glVertex3f(m_cached_points[3*k].X(), m_cached_points[3*k].Y(), m_cached_points[3*k].Z());
            glVertex3f(m_cached_points[3*k+2].X(), m_cached_points[3*k+2].Y(), m_cached_points[3*k+2].Z());
            glVertex3f(m_cached_points[3*k+1].X(), m_cached_points[3*k+1].Y(), m_cached_points[3*k+1].Z());
            glVertex3f(m_cached_points[3*k+2].X(), m_cached_points[3*k+2].Y(), m_cached_points[3*k+2].Z());
        }
        glEnd();
    }

    m_old_stamp = m_time_stamp;

    return GF_S_OK;
}

gfRayInterInfo* gfObjBezierSPatch::RayIntersection(gfMatrix4* obj_matrix, gfPoint3 * ray_begin, gfVector3* ray_dir)
{
    if (m_cached_triangles.size() == 0)
    {
        m_cached_triangles = DecomposeToBezierTriangles();
    }

    gfRayInterInfo* info = NULL;
    for (gfInt i = 0;i < (int)m_cached_triangles.size();i++)
    {
        gfObjBezierTriangle* tri = m_cached_triangles.at(i);
        gfRayInterInfo* info_temp = tri->RayIntersection(obj_matrix, ray_begin, ray_dir);

        if (info_temp)
        {
            if (!info)
            {
                info = gfCreateRayInterInfo();
            }

            // copy normal if necessary
            for (gfInt i = 0;i < (gfInt)info_temp->m_inters.size();i++)
            {
                info->m_inters.push_back(info_temp->m_inters.at(i));
                info->m_normals.push_back(info_temp->m_normals.at(i));
            }
            gfDeleteRayInterInfo(info_temp);
            gfSortRayInterInfo(info);
        }
    }

    return info;
}

gfGeomObjID gfObjBezierSPatch::GetID(void)
{
    return GF_BEZIER_SPATCH;
}

GFRESULT gfObjBezierSPatch::Transform(gfMatrix4* matrix)
{
    m_cached_triangles.clear();

    for (gfInt k = 0;k < (gfInt)m_control_points.size();k++)
    {
        gfPoint3* point = m_control_points.at(k);
        if (point)
        {
            gfTransformPoint3(point, matrix);
        }
    }

    m_time_stamp++;

    return GF_S_OK;
}

gfObjGeomObject* gfObjBezierSPatch::Clone(void)
{
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());

    gfObjBezierSPatch* patch = (gfObjBezierSPatch*) clone;
    patch->SetParams(m_degree, m_multi);
    gfInt num = (gfInt)m_control_points.size();

    for (gfInt i = 0;i <= num;i++)
    {
        patch->m_control_points[i] = m_control_points.at(i);
        patch->m_weights[i] = m_weights.at(i);
    }

    return clone;
}

GFRESULT gfObjBezierSPatch::SetParams(gfInt degree, gfInt multi)
{
    m_cached_triangles.clear();

    // clear previously allocated points
    gfInt num = (gfInt)m_control_points.size();
    for (gfInt i = 0;i < num;i++)
    {
        gfPoint3* old_pt = m_control_points.at(i);
        gfDeletePoint3(old_pt);
    }
    m_control_points.clear();
    m_weights.clear();

    // for given parameters, compute number of control points
    std::vector< std::vector<gfInt> > multiindices;
    gfGetMultiindices(multi, degree, multiindices);

    gfInt new_size = pow(degree+1, multi);
    m_control_points.resize(new_size);
    m_weights.resize(new_size);
    m_degree = degree;
    m_multi = multi;

    m_multiindices.clear();
    gfGetMultiindices(m_multi, m_degree, m_multiindices);

    // prepare uniform polygon
    m_domain_points.clear();
    for (gfInt i = 0;i < m_multi;i++)
    {
        gfPoint2 pt;
        pt.SetX(cos(2*GF_PI*i/m_multi));
        pt.SetY(sin(2*GF_PI*i/m_multi));
        m_domain_points.push_back(pt);
    }

    for (gfInt i = 0;i < new_size;i++)
    {
        gfPoint3* pt = gfCreatePoint3();
        pt->Set(0.0, 0.0, 0.0);
        m_control_points[i] = pt;
        m_weights[i] = 1.0;
    }

    m_time_stamp++;

    return GF_S_OK;
}

gfInt gfObjBezierSPatch::GetDegree(void)
{
    return m_degree;
}

gfInt gfObjBezierSPatch::GetMulti(void)
{
    return m_multi;
}

gfInt gfObjBezierSPatch::GetNumControlPoints(void)
{
    return (gfInt)m_multiindices.size();
}

gfPoint3 * gfObjBezierSPatch::GetControlPoint(std::vector<gfInt> multi_index)
{
    return m_control_points.at(GetIndexFromMultiindex(m_degree, multi_index));
}

gfDouble gfObjBezierSPatch::GetWeight(std::vector<gfInt> multi_index)
{
    return m_weights.at(GetIndexFromMultiindex(m_degree, multi_index));
}

GFRESULT gfObjBezierSPatch::SetWeight(std::vector<gfInt> multi_index, gfDouble weight)
{
    m_time_stamp++;
    m_weights[GetIndexFromMultiindex(m_degree, multi_index)] = weight;
    return GF_S_OK;
}

std::vector<gfPoint2> gfObjBezierSPatch::GetDomain()
{
    return m_domain_points;
}

void gfObjBezierSPatch::SetDomainPoint(gfInt index, gfPoint2 pt)
{
    if (index >=0 && index < (gfInt)m_domain_points.size())
        m_domain_points[index] = pt;
}

gfAuxBoundingVolume* gfObjBezierSPatch::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    // create auxiliary array of mesh points
    gfInt num_points = (gfInt)m_control_points.size();
    if (num_points < 1)
        return NULL;

    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfInt i = 0; i < num_points; i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        curr_point->Set(pt->X(), pt->Y(), pt->Z());
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    delete [] points;
    return result;
}

GFRESULT gfObjBezierSPatch::InitRandom(gfInt degree, gfInt multi)
{
    if (degree < 1 || multi < 1)
        return GF_S_OK;
    SetParams(degree, multi);

    // get multiindices for this SPatch
    std::vector< std::vector<gfInt> > multiindices;
    gfGetMultiindices(multi, degree, multiindices);

    // modify
    gfDouble inv_degree = 1.0 / degree;
    gfInt num_indices = (gfInt)multiindices.size();
    for (gfInt k = 0;k < num_indices;k++)
    {
        std::vector<gfInt> multi_index;
        multi_index.resize(multi);
        for (gfInt i = 0;i < multi;i++)
            multi_index[i] = multiindices[k][i];

        gfPoint3* pt = GetControlPoint(multi_index);
        pt->Set(0.0, 0.0, 0.0);
        for (gfInt i = 0;i < multi;i++)
        {
            pt->Translate(inv_degree*multi_index.at(i)*m_domain_points[i].X(),
                          inv_degree*multi_index.at(i)*m_domain_points[i].Y(),
                          //inv_degree*multi_index.at(i)*polygon[i].Z());
                          0.0);
        }
        pt->Translate(gfRand(-0.2, 0.2), gfRand(-0.2, 0.2), gfRand(-1.0, 1.0));
        SetWeight(multi_index, 1.0);
    }

    return GF_S_OK;
}

void gfObjBezierSPatch::ResetBaryCoordinatesComp()
{
    m_wachspress = !m_wachspress;
    m_time_stamp++;
}

gfVector3 gfObjBezierSPatch::GetNormal(std::vector<gfDouble> coordinates)
{
    // using deCasteljau algorithm
    // fill control points
    gfInt num_points = (gfInt) m_multiindices.size();
    std::vector<gfPoint4> aux_points;
    aux_points.resize(pow(m_degree+1, m_multi));
    for (gfInt i = 0;i < num_points;i++)
    {
        gfPoint3* pt = GetControlPoint(m_multiindices[i]);
        gfDouble w = GetWeight(m_multiindices[i]);
        aux_points[GetIndexFromMultiindex(m_degree, m_multiindices[i])].Set(w*pt->X(), w*pt->Y(), w*pt->Z(), w);
    }

    for (gfInt k = 1;k <= m_degree;k++)
    {
        std::vector< std::vector<gfInt> > aux_indices;
        gfGetMultiindices(m_multi, m_degree - k, aux_indices);

        std::vector<gfPoint4> points;
        for (gfInt index = 0;index < (gfInt)aux_indices.size();index++)
        {
            gfPoint4 res(0.0, 0.0, 0.0, 0.0);
            for (gfInt j = 0;j < m_multi;j++)
            {
                std::vector<gfInt> new_index = aux_indices[index];
                new_index[j]++;
                gfPoint4 old_point = aux_points[GetIndexFromMultiindex(m_degree - k + 1, new_index)];
                res.Translate(coordinates[j] * old_point.X(),
                              coordinates[j] * old_point.Y(),
                              coordinates[j] * old_point.Z(),
                              coordinates[j] * old_point.W());
            }
            points.push_back(res);
        }

        for (gfInt index = 0;index < (gfInt)aux_indices.size();index++)
        {
            aux_points[GetIndexFromMultiindex(m_degree - k, aux_indices[index])] = points[index];
        }
    }

    // now compute last step of casteljau
    gfDouble w1 = aux_points[0].W();
    gfPoint3 p1(aux_points[0].X(), aux_points[0].Y(), aux_points[0].Z());
    if (gfAbs(w1) > GF_EPS)
        p1.Set(aux_points[0].X() / w1, aux_points[0].Y() / w1, aux_points[0].Z() / w1);

    gfDouble w2 = aux_points[1].W();
    gfPoint3 p2(aux_points[1].X(), aux_points[1].Y(), aux_points[1].Z());
    if (gfAbs(w2) > GF_EPS)
        p2.Set(aux_points[1].X() / w2, aux_points[1].Y() / w2, aux_points[1].Z() / w2);

    gfDouble w3 = aux_points[2].W();
    gfPoint3 p3(aux_points[2].X(), aux_points[2].Y(), aux_points[2].Z());
    if (gfAbs(w3) > GF_EPS)
        p3.Set(aux_points[2].X() / w3, aux_points[2].Y() / w3, aux_points[2].Z() / w3);


    gfVector3 v1(p2.X() - p1.X(), p2.Y() - p1.Y(), p2.Z() - p1.Z());
    gfVector3 v2(p3.X() - p1.X(), p3.Y() - p1.Y(), p3.Z() - p1.Z());
    gfVector3 normal;
    gfVectorProduct3(&v1, &v2, &normal);
    normal.Normalize();
    return normal;
}

gfObjBezierSPatch* gfObjBezierSPatch::GetSubpatch(std::vector<gfPoint2> domain_points)
{
    // start counting time
    gfLog("Creating subpatch of %i-sided S-patch with degree %i.\n", m_multi, m_degree);
    gfWinFPSCounter count;
    count.BeginMSCount();

    gfObjBezierSPatch* result = (gfObjBezierSPatch*)gfCreateGeomObject(GF_BEZIER_SPATCH);
    if (!result)
        return NULL;

    // set degree of patch
    gfInt subpatch_multi = (gfInt)domain_points.size();
    gfInt new_degree = m_degree * (m_multi - 2);
    result->SetParams(new_degree, subpatch_multi);
    for (gfInt i = 0;i < subpatch_multi;i++)
        result->SetDomainPoint(i, domain_points[i]);

    // get indices for triangle
    std::vector< std::vector<gfInt> > tri_indices;
    gfGetMultiindices(subpatch_multi, new_degree, tri_indices);

    // prepare variables for linear system
    gfInt num_samples = (gfInt)tri_indices.size();
    Matrix A(num_samples, num_samples);
    Vector bx(num_samples);
    Vector by(num_samples);
    Vector bz(num_samples);
    Vector bw(num_samples);

    // now we have to do tri_indices.size() samples to compute control points of the new bezier triangle
    gfDouble scale = 10000000;
    for (gfInt i = 0;i < num_samples;i++)
    {
        // get not-random sample inside patch
        std::vector<gfDouble> sample_coords;
        for (gfInt j = 0;j < subpatch_multi;j++)
            sample_coords.push_back(tri_indices[i][j] / gfDouble(new_degree)); 

        // convert it to affine point
        gfPoint2 aff(0.0, 0.0);
        for (gfInt j = 0;j < subpatch_multi;j++)
            aff.Translate(sample_coords[j]*domain_points[j].X(), sample_coords[j]*domain_points[j].Y());

        // compute generalized barycentric coordinates for this point
        std::vector<gfDouble> coordinates;
        if (!m_wachspress)
            gfCompGenBaryCoordsMean2(m_domain_points, aff, coordinates);
        else
            gfCompGenBaryCoordsWachspress2(m_domain_points, aff, coordinates);

        //now compute point on Spatch for this coordinates
        gfPoint3 point(0.0, 0.0, 0.0);
        gfDouble weight = 0.0;
        gfInt num_indices = (gfInt)m_multiindices.size();
        for (gfInt j = 0;j < num_indices;j++)
        {
            gfDouble w = GetWeight(m_multiindices[j]);
            gfDouble coef = w * gfBernsteinPolyGeneral(m_multiindices[j], coordinates);
            gfPoint3* pt = GetControlPoint(m_multiindices[j]);
            point.Set(point.X() + pt->X()*coef, point.Y() + pt->Y()*coef, point.Z() + pt->Z()*coef);
            weight += coef;
        }

        point.Set(point.X() * weight, point.Y() * weight, point.Z() * weight);
        bx.setI(i, scale*point.X());
        by.setI(i, scale*point.Y());
        bz.setI(i, scale*point.Z());
        bw.setI(i, scale*weight);

        // get coefficients of matrix in this row by computing generalized bernstein polynomial for triangle
        for (gfInt j = 0;j < num_samples;j++)
        {
            gfDouble bdiu = gfBernsteinPolyGeneral(tri_indices[j], sample_coords);
            A.setIJ(i, j, scale*bdiu);
        }
    }

    // now we have linear system, solve it!
    DirectSolver dsx(A, bx);
    DirectSolver dsy(A, by);
    DirectSolver dsz(A, bz);
    DirectSolver dsw(A, bw);

    Vector resx, resy, resz, resw;
    resx = dsx.pivotGaussElimination();
    resy = dsy.pivotGaussElimination();
    resz = dsz.pivotGaussElimination();
    resw = dsw.pivotGaussElimination();

    // fill bezier triangle
    gfPoint3* pt;
    for (gfInt i = 0;i < num_samples;i++)
    {
        gfDouble w = resw.getI(i);
        result->SetWeight(tri_indices[i], w);

        gfDouble inv;
        if (gfAbs(w) < GF_EPS)
            inv = 0.0;
        else
            inv = 1.0 / w;
        pt = result->GetControlPoint(tri_indices[i]);
        pt->Set(inv * resx.getI(i), inv * resy.getI(i), inv * resz.getI(i));
    }

    gfLog("Subpatching finished, subpatch has %i sides and %i degree, time = %d ms.\n", subpatch_multi, new_degree, count.EndMSCount());

    return result;
}

gfObjBezierTriangle* gfObjBezierSPatch::GetSubtriangle(gfPoint2* pt1, gfPoint2* pt2, gfPoint2* pt3)
{
    gfObjBezierTriangle* result = (gfObjBezierTriangle*)gfCreateGeomObject(GF_BEZIER_TRIANGLE);
    if (!result)
        return NULL;

    // set degree of patch
    gfInt new_degree = m_degree * (m_multi - 2);
    result->SetParams(new_degree);

    // get indices for triangle
    std::vector< std::vector<gfInt> > tri_indices;
    gfGetMultiindices(3, new_degree, tri_indices);

    // prepare variables for linear system
    gfInt num_samples = (gfInt)tri_indices.size();
    Matrix A(num_samples, num_samples);
    Vector bx(num_samples);
    Vector by(num_samples);
    Vector bz(num_samples);
    Vector bw(num_samples);

    // now we have to do (new_degree + 1) * (new_degree + 2) samples to compute control points of the new bezier triangle
    gfDouble scale = 10000000;
    for (gfInt i = 0;i < num_samples;i++)
    {
        // get random sample inside triangle
        gfDouble u = tri_indices[i][0] / gfDouble(new_degree);
        gfDouble v = tri_indices[i][1] / gfDouble(new_degree);
        gfDouble w = tri_indices[i][2] / gfDouble(new_degree);

        // convert it to affine point
        gfPoint2 aff;
        aff.SetX(u*pt1->X()+v*pt2->X()+w*pt3->X());
        aff.SetY(u*pt1->Y()+v*pt2->Y()+w*pt3->Y());

        // compute generalized barycentric coordinates for this point
        std::vector<gfDouble> coordinates;
        if (!m_wachspress)
            gfCompGenBaryCoordsMean2(m_domain_points, aff, coordinates);
        else
            gfCompGenBaryCoordsWachspress2(m_domain_points, aff, coordinates);

        //now compute point on Spatch for this coordinates
        gfPoint3 point(0.0, 0.0, 0.0);
        gfDouble weight = 0.0;
        gfInt num_indices = (gfInt)m_multiindices.size();
        for (gfInt j = 0;j < num_indices;j++)
        {
            gfDouble w = GetWeight(m_multiindices[j]);
            gfDouble coef = w * gfBernsteinPolyGeneral(m_multiindices[j], coordinates);
            gfPoint3* pt = GetControlPoint(m_multiindices[j]);
            point.Set(point.X() + pt->X()*coef, point.Y() + pt->Y()*coef, point.Z() + pt->Z()*coef);
            weight += coef;
        }

        point.Set(point.X() * weight, point.Y() * weight, point.Z() * weight);
        bx.setI(i, scale*point.X());
        by.setI(i, scale*point.Y());
        bz.setI(i, scale*point.Z());
        bw.setI(i, scale*weight);

        // get coefficients of matrix in this row by computing generalized bernstein polynomial for triangle
        for (gfInt j = 0;j < num_samples;j++)
        {
            gfDouble bdiu = gfBernsteinPoly3(new_degree, tri_indices[j][0], tri_indices[j][1], u, v);
            A.setIJ(i, j, scale*bdiu);
        }
    }

    /*gfDouble det = A.getDeterminant();
    if (gfAbs(det) < GF_EPS)
        return result;*/

    // now we have linear system, solve it!
    DirectSolver dsx(A, bx);
    DirectSolver dsy(A, by);
    DirectSolver dsz(A, bz);
    DirectSolver dsw(A, bw);

    Vector resx, resy, resz, resw;
    resx = dsx.pivotGaussElimination();
    resy = dsy.pivotGaussElimination();
    resz = dsz.pivotGaussElimination();
    resw = dsw.pivotGaussElimination();

    // fill bezier triangle
    gfPoint3* pt;
    for (gfInt i = 0;i < num_samples;i++)
    {
        gfDouble w = resw.getI(i);
        result->SetWeight(tri_indices[i][0], tri_indices[i][1], tri_indices[i][2], w);

        gfDouble inv;
        if (gfAbs(w) < GF_EPS)
            inv = 0.0;
        else
            inv = 1.0 / w;
        pt = result->GetControlPoint(tri_indices[i][0], tri_indices[i][1], tri_indices[i][2]);
        pt->Set(inv * resx.getI(i), inv * resy.getI(i), inv * resz.getI(i));
    }

    return result;
}

std::vector<gfObjBezierTriangle*> gfObjBezierSPatch::DecomposeToBezierTriangles()
{
    // starting
    gfLog("Decomposing %i-sided S-patch with degree %i into triangles.\n", m_multi, m_degree);
    gfWinFPSCounter count;
    count.BeginMSCount();

    // for this path we have simple decompostion
    std::vector<gfObjBezierTriangle*> result;

    gfPoint2 mid(0.0, 0.0);
    for (gfInt i = 0;i < m_multi;i++)
        mid.Translate(m_domain_points[i].X() / m_multi, m_domain_points[i].Y() / m_multi);

    for (gfInt i = 0;i < m_multi;i++)
    {
        gfPoint2 pt2 = m_domain_points[i];
        gfPoint2 pt3 = m_domain_points[(i + 1) % m_multi];
        gfObjBezierTriangle* tri = GetSubtriangle(&mid, &pt2, &pt3);
        result.push_back(tri);
    }

    gfLog("Decomposition finished, each triangle has degree %i, time = %d ms.\n", m_degree * (m_multi - 2), count.EndMSCount());

    return result;
}