//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjBezierSVolume.cpp                              //
//   Description: Rational Bezier triangle class                //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfObjBezierSVolume.h"
#include "gfObjectFactory.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfAuxLogging.h"
#include "gfWinFPSCounter.h"
#include <vector>

#include "./OGL/glut.h"
#include "./LinearSolver/DirectSolver.h"


gfInt GetIndexFromMultiindexVol(gfInt degree, std::vector<gfInt>& multi_index)
{
    gfInt multi = (gfInt)multi_index.size();
    gfInt index = 0;
    for (int i = 0;i < multi;i++)
    {
        gfInt ii = multi_index.at(i);
        index += multi_index.at(i) * pow(degree+1, i);
    }
    return index;
}

struct gfAuxTriangle3
{
    gfPoint3 pt1;
    gfPoint3 pt2;
    gfPoint3 pt3;
};

void CreateRegularDomain(gfInt multi, std::vector<gfPoint3>& points, std::vector<gfTripleIndex>& triangles)
{
    gfDouble cosxi = cos(2 * GF_PI / 3);
    gfDouble sinxi = sin(2 * GF_PI / 3);

    gfInt num_top = multi / 2;
    gfInt num_bottom = multi - num_top;

    // prepare points on top
    for (gfInt i = 0;i < num_top;i++)
    {
        gfDouble fi = 2 * i * GF_PI / num_bottom;
        if (num_top == num_bottom && num_bottom == 2)
            fi += GF_PI / 2.0;
        gfPoint3 pt;
        pt.SetX(cos(fi) * cosxi);
        pt.SetY(sin(fi) * cosxi);
        pt.SetZ(sinxi);
        points.push_back(pt);
    }

    // prepare points on bottom
    for (gfInt i = 0;i < num_bottom;i++)
    {
        gfDouble fi = 2 * i * GF_PI / num_bottom;
        gfPoint3 pt;
        pt.SetX(cos(fi) * cosxi);
        pt.SetY(sin(fi) * cosxi);
        pt.SetZ(-sinxi);
        points.push_back(pt);
    }

    // triangulate top plane
    for (gfInt i = 1;i < num_top - 1;i++)
    {
        gfTripleIndex tri;
        tri.i = 0;
        tri.j = i;
        tri.k = i+1;
        triangles.push_back(tri);
    }

    // triangulate bottom plane
    for (gfInt i = 1;i < num_bottom - 1;i++)
    {
        gfTripleIndex tri;
        tri.i = num_top;
        tri.j = num_top + i;
        tri.k = num_top + i + 1;
        triangles.push_back(tri);
    }

    // triangulate sides
    for (gfInt i = 0;i < num_top - 1;i++)
    {
        gfTripleIndex tri;
        tri.i = i;
        tri.j = i + num_top;
        tri.k = ((i + 1) % num_top) + num_top;
        triangles.push_back(tri);

        tri.i = i;
        tri.j = ((i + 1) % num_top) + num_top;
        tri.k = (i + 1) % num_top;
        triangles.push_back(tri);
    }

    if (num_top == num_bottom)
    {
        gfTripleIndex tri;
        tri.i = num_top - 1;
        tri.j = num_top - 1 + num_top;
        tri.k = num_top;
        triangles.push_back(tri);

        tri.i = num_top - 1;
        tri.j = num_top;
        tri.k = 0;
        triangles.push_back(tri);
    }
    else
    {
        gfTripleIndex tri;

        tri.i = num_top - 1;
        tri.j = num_bottom - 1 + num_top;
        tri.k = num_top - 1 + num_top;
        triangles.push_back(tri);

        tri.i = 0;
        tri.j = num_bottom - 1 + num_top;
        tri.k = num_top - 1;
        triangles.push_back(tri);

        tri.i = num_top;
        tri.j = num_bottom - 1 + num_top;
        tri.k = 0;
        triangles.push_back(tri);
    }

}

void SampleDomainTetrahedron(gfInt prec, gfPoint3 pt1, gfPoint3 pt2, gfPoint3 pt3, gfPoint3 pt4, std::vector<gfAuxTriangle3>& triangles)
{
    // sampling using barycentric coordinates
    // first direction
    for (gfInt i = 0;i <= prec - 2;i++)
    {     
        for (gfInt j = 1;j <= prec - 1 - i;j++)
        {
            for (gfInt k = 0;k <= prec - 1 - i - j;k++)
            {
                gfAuxTriangle3 tri1;
                gfAuxTriangle3 tri2;

                gfDouble u = (gfDouble) i / (prec - 1);
                gfDouble v = (gfDouble) j / (prec - 1);
                gfDouble w = (gfDouble) k / (prec - 1);
                gfDouble q = 1.0 - u - v - w;
                tri1.pt1.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());
                tri2.pt1.Set(tri1.pt1.X(), tri1.pt1.Y(), tri1.pt1.Z());

                u = (gfDouble) i / (prec - 1);
                v = (gfDouble) (j - 1) / (prec - 1);
                w = (gfDouble) (k + 1) / (prec - 1);
                q = 1.0 - u - v - w;
                tri1.pt3.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());

                u = (gfDouble) i / (prec - 1);
                v = (gfDouble) (j - 1) / (prec - 1);
                w = (gfDouble) k / (prec - 1);
                q = 1.0 - u - v - w;
                tri1.pt2.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());
                tri2.pt3.Set(tri1.pt2.X(), tri1.pt2.Y(), tri1.pt2.Z());

                triangles.push_back(tri1);

                if (k > 0)
                {
                    u = (gfDouble) i / (prec - 1);
                    v = (gfDouble) j / (prec - 1);
                    w = (gfDouble) (k - 1) / (prec - 1);
                    q = 1.0 - u - v - w;
                    tri2.pt2.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                        u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                        u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());

                    triangles.push_back(tri2);
                }
            }
        }
    }

    // second direction
    for (gfInt j = 0;j <= prec - 2;j++)
    {     
        for (gfInt i = 1;i <= prec - 1 - j;i++)
        {
            for (gfInt k = 0;k <= prec - 1 - i - j;k++)
            {
                gfAuxTriangle3 tri1;
                gfAuxTriangle3 tri2;

                gfDouble u = (gfDouble) i / (prec - 1);
                gfDouble v = (gfDouble) j / (prec - 1);
                gfDouble w = (gfDouble) k / (prec - 1);
                gfDouble q = 1.0 - u - v - w;
                tri1.pt1.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());
                tri2.pt1 = tri1.pt1;

                u = (gfDouble) (i - 1) / (prec - 1);
                v = (gfDouble) j / (prec - 1);
                w = (gfDouble) (k + 1) / (prec - 1);
                q = 1.0 - u - v - w;
                tri1.pt3.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());

                u = (gfDouble) (i - 1) / (prec - 1);
                v = (gfDouble) j / (prec - 1);
                w = (gfDouble) k / (prec - 1);
                q = 1.0 - u - v - w;
                tri1.pt2.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());
                tri2.pt3 = tri1.pt2;

                triangles.push_back(tri1);

                if (k > 0)
                {
                    u = (gfDouble) i / (prec - 1);
                    v = (gfDouble) j / (prec - 1);
                    w = (gfDouble) (k - 1) / (prec - 1);
                    q = 1.0 - u - v - w;
                    tri2.pt2.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                        u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                        u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());

                    triangles.push_back(tri2);
                }
            }
        }
    }

    // third direction
    for (gfInt k = 0;k <= prec - 2;k++)
    {     
        for (gfInt j = 1;j <= prec - 1 - k;j++)
        {
            for (gfInt i = 0;i <= prec - 1 - k - j;i++)
            {
                gfAuxTriangle3 tri1;
                gfAuxTriangle3 tri2;

                gfDouble u = (gfDouble) i / (prec - 1);
                gfDouble v = (gfDouble) j / (prec - 1);
                gfDouble w = (gfDouble) k / (prec - 1);
                gfDouble q = 1.0 - u - v - w;
                tri1.pt1.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());
                tri2.pt1 = tri1.pt1;

                u = (gfDouble) (i + 1) / (prec - 1);
                v = (gfDouble) (j - 1) / (prec - 1);
                w = (gfDouble) k / (prec - 1);
                q = 1.0 - u - v - w;
                tri1.pt3.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());

                u = (gfDouble) i / (prec - 1);
                v = (gfDouble) (j - 1) / (prec - 1);
                w = (gfDouble) k / (prec - 1);
                q = 1.0 - u - v - w;
                tri1.pt2.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());
                tri2.pt3 = tri1.pt2;

                triangles.push_back(tri1);

                if (i > 0)
                {
                    u = (gfDouble) (i - 1) / (prec - 1);
                    v = (gfDouble) j / (prec - 1);
                    w = (gfDouble) k / (prec - 1);
                    q = 1.0 - u - v - w;
                    tri2.pt2.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                        u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                        u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());

                    triangles.push_back(tri2);
                }
            }
        }
    }

    // fourth direction
    for (gfInt l = 0;l <= prec - 2;l++)
    {     
        for (gfInt j = 1;j <= prec - 1 - l;j++)
        {
            for (gfInt k = 0;k <= prec - 1 - l - j;k++)
            {
                gfAuxTriangle3 tri1;
                gfAuxTriangle3 tri2;

                gfInt i = prec - 1 - l - j - k;
                gfDouble u = (gfDouble) i / (prec - 1);
                gfDouble v = (gfDouble) j / (prec - 1);
                gfDouble w = (gfDouble) k / (prec - 1);
                gfDouble q = 1.0 - u - v - w;
                tri1.pt1.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());
                tri2.pt1 = tri1.pt1;

                u = (gfDouble) (i - 1) / (prec - 1);
                v = (gfDouble) (j + 1) / (prec - 1);
                w = (gfDouble) k / (prec - 1);
                q = 1.0 - u - v - w;
                tri1.pt3.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());

                u = (gfDouble) (i - 1) / (prec - 1);
                v = (gfDouble) j / (prec - 1);
                w = (gfDouble) k / (prec - 1);
                q = 1.0 - u - v - w;
                tri1.pt2.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                    u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                    u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());
                tri2.pt3 = tri1.pt2;

                triangles.push_back(tri1);

                if (j > 0)
                {
                    u = (gfDouble) i / (prec - 1);
                    v = (gfDouble) (j - 1) / (prec - 1);
                    w = (gfDouble) k / (prec - 1);
                    q = 1.0 - u - v - w;
                    tri2.pt2.Set(u*pt1.X()+v*pt2.X()+w*pt3.X()+q*pt4.X(),
                        u*pt1.Y()+v*pt2.Y()+w*pt3.Y()+q*pt4.Y(),
                        u*pt1.Z()+v*pt2.Z()+w*pt3.Z()+q*pt4.Z());

                    triangles.push_back(tri2);
                }
            }
        }
    }
}

gfObjBezierSVolume::gfObjBezierSVolume(void)
{
    m_time_stamp = 0;
    m_weights.resize(0);
    m_control_points.resize(0);
    m_degree = 0;
    m_multi = 0;
    m_old_stamp = 0;
}

gfObjBezierSVolume::~gfObjBezierSVolume(void)
{
    SetParams(0, 0);
}

GFRESULT gfObjBezierSVolume::OGLDraw(gfLong flags)
{
    if (m_time_stamp == 0)
        return GF_S_OK;

    // this draws domain
    /*std::vector<gfPoint3> points;
    std::vector<gfTripleIndex> triangles;
    CreateRegularDomain(m_multi, points, triangles);
    gfInt ii = (gfInt)triangles.size();

    if (flags & GF_FLG_SOLID)
    {
        glBegin(GL_TRIANGLES);
        for (gfInt i = 0;i < (gfInt)triangles.size();i++)
        {
            gfPoint3 result = points[triangles[i].i];
            glVertex3f(result.X(), result.Y(), result.Z());
            result = points[triangles[i].j];
            glVertex3f(result.X(), result.Y(), result.Z());
            result = points[triangles[i].k];
            glVertex3f(result.X(), result.Y(), result.Z());
        }
        glEnd();
    }
    
    if (flags & GF_FLG_WIRE)
    {
        glColor3f(0.0, 0.0, 0.0);
        for (gfInt i = 0;i < (gfInt)triangles.size();i++)
        {
            glBegin(GL_LINE_LOOP);
            gfPoint3 result = points[triangles[i].i];
            glVertex3f(result.X(), result.Y(), result.Z());
            result = points[triangles[i].j];
            glVertex3f(result.X(), result.Y(), result.Z());
            result = points[triangles[i].k];
            glVertex3f(result.X(), result.Y(), result.Z());
            glEnd();
        }
    }
    return GF_S_OK;*/

    std::vector< std::vector<gfInt> > multiindices;
    gfGetMultiindices(m_multi, m_degree, multiindices);
    gfInt num_indices = (gfInt)multiindices.size();

    // draw possibly control points, it isn't cached
    if (flags & GF_FLG_CONTROL_NET)
    {
        glColor3f(0.0, 0.0, 0.0);
        std::vector<gfInt> aux_multi_index;
        glBegin(GL_LINES);
        for (gfInt i = 0;i < num_indices;i++)
        {
            std::vector<gfInt> multi_index = multiindices[i];
            gfPoint3* pt = GetControlPoint(multi_index);

            // in this multi index, create all +1,-1 combinations
            for (gfInt k = 0;k < m_multi;k++)
            {
                if ((multi_index[k] + 1) > m_degree)
                    continue;

                for (gfInt l = 0;l < m_multi;l++)
                {
                    if (k == l)
                        continue;

                    if ((multi_index[l] - 1) < 0)
                        continue;

                    glVertex3f(pt->X(), pt->Y(), pt->Z());
                    aux_multi_index = multi_index;
                    aux_multi_index[k]++;
                    aux_multi_index[l]--;
                    gfPoint3* aux_pt = GetControlPoint(aux_multi_index);
                    glVertex3f(aux_pt->X(), aux_pt->Y(), aux_pt->Z());
                }
            }
        }
        glEnd();

        glPointSize(8.0);
        glBegin(GL_POINTS);
        for (gfInt i = 0;i < num_indices;i++)
        {
            std::vector<gfInt> multi_index(m_multi, 0);
            multi_index = multiindices[i];

            gfPoint3* pt = GetControlPoint(multi_index);
            glVertex3f(pt->X(), pt->Y(), pt->Z());
        }
        glEnd();
        glPointSize(1.0);
    }

    // here domain is polytope with m vertices with triangular faces
    // vertices of domain are on surface of sphere
    // sampling domain
    if (m_time_stamp != m_old_stamp)
    {
        gfInt prec = GF_VIS_PREC / 2;
        std::vector<gfPoint3> domain_points;
        std::vector<gfTripleIndex> domain_triangles;
        CreateRegularDomain(m_multi, domain_points, domain_triangles);

        m_cached_points.clear();

        gfPoint3 center(0.0, 0.0, 0.0);
        gfInt num_domain_triangles = (gfInt)domain_triangles.size();
        for (gfInt i = 0;i < num_domain_triangles;i++)
        {
            gfTripleIndex dtri = domain_triangles[i];
            std::vector<gfAuxTriangle3> triangles;
            SampleDomainTetrahedron(prec, center, domain_points[dtri.i], domain_points[dtri.j], domain_points[dtri.k], triangles);

            gfInt num_tris = (gfInt)triangles.size();
            for (gfInt k = 0;k < num_tris;k++)
            {
                gfAuxTriangle3 tri = triangles[k];
                // have element triangle, compute surface points
                // first compute barycentric coordinates

                std::vector<gfDouble> coordinates1;
                gfCompGenBaryCoordsMean3(domain_points, domain_triangles, tri.pt1, coordinates1);
                std::vector<gfDouble> coordinates2;
                gfCompGenBaryCoordsMean3(domain_points, domain_triangles, tri.pt2, coordinates2);
                std::vector<gfDouble> coordinates3;
                gfCompGenBaryCoordsMean3(domain_points, domain_triangles, tri.pt3, coordinates3);

                gfPoint3 result1(0.0, 0.0, 0.0);
                gfPoint3 result2(0.0, 0.0, 0.0);
                gfPoint3 result3(0.0, 0.0, 0.0);
                gfDouble weight1 = 0.0;
                gfDouble weight2 = 0.0;
                gfDouble weight3 = 0.0;
                for (gfInt j = 0;j < num_indices;j++)
                {
                    gfDouble w = GetWeight(multiindices[j]);
                    gfDouble coef1 = w * gfBernsteinPolyGeneral(multiindices[j], coordinates1);
                    if (coef1 > 0.0)
                    {
                        gfPoint3* pt = GetControlPoint(multiindices[j]);
                        //m_cached_points.push_back(*pt);
                    }
                    gfDouble coef2 = w * gfBernsteinPolyGeneral(multiindices[j], coordinates2);
                    if (coef2 > 0.0)
                    {
                        gfPoint3* pt = GetControlPoint(multiindices[j]);
                        //m_cached_points.push_back(*pt);
                    }
                    gfDouble coef3 = w * gfBernsteinPolyGeneral(multiindices[j], coordinates3);
                    gfPoint3* pt = GetControlPoint(multiindices[j]);
                    result1.Set(result1.X() + pt->X()*coef1, result1.Y() + pt->Y()*coef1, result1.Z() + pt->Z()*coef1);
                    result2.Set(result2.X() + pt->X()*coef2, result2.Y() + pt->Y()*coef2, result2.Z() + pt->Z()*coef2);
                    result3.Set(result3.X() + pt->X()*coef3, result3.Y() + pt->Y()*coef3, result3.Z() + pt->Z()*coef3);
                    weight1 += coef1;
                    weight2 += coef2;
                    weight3 += coef3;
                }

                gfDouble invW = 1.0 / weight1;
                result1.Set(result1.X() * invW, result1.Y() * invW, result1.Z() * invW);
                m_cached_points.push_back(result1);
                invW = 1.0 / weight2;
                result2.Set(result2.X() * invW, result2.Y() * invW, result2.Z() * invW);
                m_cached_points.push_back(result2);
                invW = 1.0 / weight3;
                result3.Set(result3.X() * invW, result3.Y() * invW, result3.Z() * invW);
                m_cached_points.push_back(result3);
            }
        }
    }

    if (flags & GF_FLG_SOLID)
    {
        glBegin(GL_TRIANGLES);
        glPointSize(10);
        //glBegin(GL_POINTS);
        gfInt num_samples = (gfInt)m_cached_points.size();
        for (gfInt k = 0;k < num_samples;k++)
        {
            // we have coordinate, compute point on surface
            // based on bezier bernstein representation
            gfPoint3 result = m_cached_points[k];
            glVertex3f(result.X(), result.Y(), result.Z());
        }
        glEnd();
    }

    if (flags & GF_FLG_WIRE)
    {
        glColor3f(0.0, 0.0, 0.0);
        glBegin(GL_LINES);
        gfInt num_samples = (gfInt)m_cached_points.size() / 3;
        for (gfInt k = 0;k < num_samples;k++)
        {
            glVertex3f(m_cached_points[3*k].X(), m_cached_points[3*k].Y(), m_cached_points[3*k].Z());
            glVertex3f(m_cached_points[3*k+1].X(), m_cached_points[3*k+1].Y(), m_cached_points[3*k+1].Z());
            glVertex3f(m_cached_points[3*k].X(), m_cached_points[3*k].Y(), m_cached_points[3*k].Z());
            glVertex3f(m_cached_points[3*k+2].X(), m_cached_points[3*k+2].Y(), m_cached_points[3*k+2].Z());
            glVertex3f(m_cached_points[3*k+1].X(), m_cached_points[3*k+1].Y(), m_cached_points[3*k+1].Z());
            glVertex3f(m_cached_points[3*k+2].X(), m_cached_points[3*k+2].Y(), m_cached_points[3*k+2].Z());
        }
        glEnd();
    }

    m_old_stamp = m_time_stamp;

    return GF_S_OK;
}

gfRayInterInfo* gfObjBezierSVolume::RayIntersection(gfMatrix4* obj_matrix, gfPoint3 * ray_begin, gfVector3* ray_dir)
{
    gfRayInterInfo* info = gfCreateRayInterInfo();
    info->m_inters.push_back(0.5);
    info->m_normals.push_back(gfVector3(0.0, 0.0, 0.0));
    info->m_uvw.push_back(gfPoint3(0.0, 0.0, 0.0));
    return info;
}

gfGeomObjID gfObjBezierSVolume::GetID(void)
{
    return GF_BEZIER_SVOLUME;
}

GFRESULT gfObjBezierSVolume::Transform(gfMatrix4* matrix)
{
    for (gfInt k = 0;k < (gfInt)m_control_points.size();k++)
    {
        gfPoint3* point = m_control_points.at(k);
        if (point)
        {
            gfTransformPoint3(point, matrix);
        }
    }

    m_time_stamp++;

    return GF_S_OK;
}

gfObjGeomObject* gfObjBezierSVolume::Clone(void)
{
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());

    gfObjBezierSVolume* patch = (gfObjBezierSVolume*) clone;
    patch->SetParams(m_degree, m_multi);
    gfInt num = (gfInt)m_control_points.size();

    for (gfInt i = 0;i <= num;i++)
    {
        patch->m_control_points[i] = m_control_points.at(i);
        patch->m_weights[i] = m_weights.at(i);
    }

    return clone;
}

GFRESULT gfObjBezierSVolume::SetParams(gfInt degree, gfInt multi)
{
    // clear previously allocated points
    gfInt num = (gfInt)m_control_points.size();
    for (gfInt i = 0;i < num;i++)
    {
        gfPoint3* old_pt = m_control_points.at(i);
        gfDeletePoint3(old_pt);
    }
    m_control_points.clear();
    m_weights.clear();

    if (multi == 0 && multi == 0)
        return GF_S_OK;

    // for given parameters, compute number of control points
    std::vector< std::vector<gfInt> > multiindices;
    gfGetMultiindices(multi, degree, multiindices);
    m_num_control_points = (gfInt)multiindices.size();

    gfInt new_size = pow(degree+1, multi+1);
    m_control_points.resize(new_size);
    m_weights.resize(new_size);
    m_degree = degree;
    m_multi = multi;

    for (gfInt i = 0;i < new_size;i++)
    {
        gfPoint3* pt = gfCreatePoint3();
        pt->Set(0.0, 0.0, 0.0);
        m_control_points[i] = pt;
        m_weights[i] = 1.0;
    }

    m_time_stamp++;

    return GF_S_OK;
}

gfInt gfObjBezierSVolume::GetDegree(void)
{
    return m_degree;
}

gfInt gfObjBezierSVolume::GetMulti(void)
{
    return m_multi;
}

gfInt gfObjBezierSVolume::GetNumControlPoints(void)
{
    return m_num_control_points;
}

gfPoint3 * gfObjBezierSVolume::GetControlPoint(std::vector<gfInt> multi_index)
{
    return m_control_points.at(GetIndexFromMultiindexVol(m_degree, multi_index));
}

gfDouble gfObjBezierSVolume::GetWeight(std::vector<gfInt> multi_index)
{
    return m_weights.at(GetIndexFromMultiindexVol(m_degree, multi_index));
}

GFRESULT gfObjBezierSVolume::SetWeight(std::vector<gfInt> multi_index, gfDouble weight)
{
    m_time_stamp++;
    m_weights[GetIndexFromMultiindexVol(m_degree, multi_index)] = weight;
    return GF_S_OK;
}

gfAuxBoundingVolume* gfObjBezierSVolume::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    // create auxiliary array of mesh points
    gfInt num_points = (gfInt)m_control_points.size();
    if (num_points < 1)
        return NULL;

    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfInt i = 0; i < num_points; i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        curr_point->Set(pt->X(), pt->Y(), pt->Z());
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    delete [] points;
    return result;
}

GFRESULT gfObjBezierSVolume::InitRandom(gfInt degree, gfInt multi)
{
    if (degree < 1 || multi < 1)
        return GF_S_OK;
    SetParams(degree, multi);

    // get multiindices for this SVolume
    std::vector< std::vector<gfInt> > multiindices;
    gfGetMultiindices(multi, degree, multiindices);

    // prepare uniform polygon
    std::vector<gfPoint3> points;
    std::vector<gfTripleIndex> triangles;
    CreateRegularDomain(m_multi, points, triangles);

    // modify
    gfDouble inv_degree = 1.0 / degree;
    gfInt num_indices = (gfInt)multiindices.size();
    for (gfInt k = 0;k < num_indices;k++)
    {
        std::vector<gfInt> multi_index;
        multi_index.resize(multi);
        for (gfInt i = 0;i < multi;i++)
            multi_index[i] = multiindices[k][i];

        gfPoint3* pt = GetControlPoint(multi_index);
        pt->Set(0.0, 0.0, 0.0);
        for (gfInt i = 0;i < multi;i++)
        {
            pt->Translate(inv_degree*multi_index.at(i)*points[i].X(),
                          inv_degree*multi_index.at(i)*points[i].Y(),
                          inv_degree*multi_index.at(i)*points[i].Z());
        }
        pt->Translate(gfRand(-0.2, 0.2), gfRand(-0.2, 0.2), gfRand(-0.2, 0.2));
        pt->Set(5*pt->X(), 5*pt->Y(), 2*pt->Z());
        SetWeight(multi_index, 1.0);
    }

    return GF_S_OK;
}

gfObjBezierTetrahedra* gfObjBezierSVolume::GetSubtetrahedron(gfPoint3* pt1, gfPoint3* pt2, gfPoint3* pt3, gfPoint3* pt4)
{
    gfObjBezierTetrahedra* result = (gfObjBezierTetrahedra*)gfCreateGeomObject(GF_BEZIER_TETRAHEDRA);
    if (!result)
        return NULL;

    // set degree of patch
    gfInt new_degree = m_degree * (m_multi - 2);
    result->SetParams(new_degree);

    // get indices for new tetrahedron
    std::vector< std::vector<gfInt> > tetra_indices;
    gfGetMultiindices(4, new_degree, tetra_indices);

    // prepare variables for linear system
    gfInt num_samples = (gfInt)tetra_indices.size();
    Matrix A(num_samples, num_samples);
    Vector bx(num_samples);
    Vector by(num_samples);
    Vector bz(num_samples);
    Vector bw(num_samples);

    // create domain, for now it is generated, not stored
    std::vector<gfPoint3> domain_points;
    std::vector<gfTripleIndex> domain_triangles;
    CreateRegularDomain(m_multi, domain_points, domain_triangles);

    // get indices for old tetrahedron
    std::vector< std::vector<gfInt> > old_tetra_indices;
    gfGetMultiindices(m_multi, m_degree, old_tetra_indices);

    // now we have to do (new_degree + 1) * (new_degree + 2) samples to compute control points of the new bezier tetrahedra
    gfDouble scale = 10000000;
    for (gfInt i = 0;i < num_samples;i++)
    {
        // get random sample inside triangle
        gfDouble u = tetra_indices[i][0] / gfDouble(new_degree);
        gfDouble v = tetra_indices[i][1] / gfDouble(new_degree);
        gfDouble w = tetra_indices[i][2] / gfDouble(new_degree);
        gfDouble t = tetra_indices[i][3] / gfDouble(new_degree);

        // convert it to affine point
        gfPoint3 aff;
        aff.SetX(u*pt1->X()+v*pt2->X()+w*pt3->X()+t*pt4->X());
        aff.SetY(u*pt1->Y()+v*pt2->Y()+w*pt3->Y()+t*pt4->Y());
        aff.SetZ(u*pt1->Z()+v*pt2->Z()+w*pt3->Z()+t*pt4->Z());

        // compute generalized barycentric coordinates for this point
        std::vector<gfDouble> coordinates;
        gfCompGenBaryCoordsMean3(domain_points, domain_triangles, aff, coordinates);

        //now compute point of Bezier tetrahedra for this coordinates
        gfPoint3 point(0.0, 0.0, 0.0);
        gfDouble weight = 0.0;
        gfInt num_indices = (gfInt)old_tetra_indices.size();
        for (gfInt j = 0;j < num_indices;j++)
        {
            gfDouble ww = GetWeight(old_tetra_indices[j]);
            gfDouble coef = ww * gfBernsteinPolyGeneral(old_tetra_indices[j], coordinates);
            gfPoint3* pt = GetControlPoint(old_tetra_indices[j]);
            point.Set(point.X() + pt->X()*coef, point.Y() + pt->Y()*coef, point.Z() + pt->Z()*coef);
            weight += coef;
        }

        point.Set(point.X() * weight, point.Y() * weight, point.Z() * weight);
        bx.setI(i, scale*point.X());
        by.setI(i, scale*point.Y());
        bz.setI(i, scale*point.Z());
        bw.setI(i, scale*weight);

        // get coefficients of matrix in this row by computing generalized bernstein polynomial for triangle
        for (gfInt j = 0;j < num_samples;j++)
        {
            gfDouble bdiu = gfBernsteinPoly4(new_degree, tetra_indices[j][0], tetra_indices[j][1], tetra_indices[j][2], u, v, w);
            A.setIJ(i, j, scale*bdiu);
        }
    }

    /*gfDouble det = A.getDeterminant();
    if (gfAbs(det) < GF_EPS)
        return result;*/

    // now we have linear system, solve it!
    DirectSolver dsx(A, bx);
    DirectSolver dsy(A, by);
    DirectSolver dsz(A, bz);
    DirectSolver dsw(A, bw);

    Vector resx, resy, resz, resw;
    resx = dsx.pivotGaussElimination();
    resy = dsy.pivotGaussElimination();
    resz = dsz.pivotGaussElimination();
    resw = dsw.pivotGaussElimination();

    // fill bezier tetrahedron
    gfPoint3* pt;
    for (gfInt i = 0;i < num_samples;i++)
    {
        gfDouble ww = resw.getI(i);
        result->SetWeight(tetra_indices[i][0], tetra_indices[i][1], tetra_indices[i][2], tetra_indices[i][3], ww);

        gfDouble inv;
        if (gfAbs(ww) < GF_EPS)
            inv = 1.0;
        else
            inv = 1.0 / ww;
        pt = result->GetControlPoint(tetra_indices[i][0], tetra_indices[i][1], tetra_indices[i][2], tetra_indices[i][3]);
        pt->Set(inv * resx.getI(i), inv * resy.getI(i), inv * resz.getI(i));

        //gfLog(" Tetra point: %f, %f, %f, %f \n", pt->X(), pt->Y(), pt->Z(), ww);
    }

    return result;
}

std::vector<gfObjBezierTetrahedra*> gfObjBezierSVolume::DecomposeToBezierTetrahedras()
{
    // starting
    gfWinFPSCounter count;
    count.BeginMSCount();

    // for this path we have simple decompostion
    std::vector<gfObjBezierTetrahedra*> result;

    std::vector<gfPoint3> domain_points;
    std::vector<gfTripleIndex> domain_triangles;
    CreateRegularDomain(m_multi, domain_points, domain_triangles);

    gfLog("Decomposing %i-sided Bezier S-volume with degree %i and %i domain triangles into tetrahedrons.\n", m_multi, m_degree, domain_triangles.size());

    gfPoint3 mid(0.0, 0.0, 0.0);
    for (gfInt i = 0;i < m_multi;i++)
        mid.Translate(domain_points[i].X() / m_multi, domain_points[i].Y() / m_multi, domain_points[i].Z() / m_multi);

    for (gfInt i = 0;i < domain_triangles.size();i++)
    {
        gfPoint3 pt2 = domain_points[domain_triangles[i].i];
        gfPoint3 pt3 = domain_points[domain_triangles[i].j];
        gfPoint3 pt4 = domain_points[domain_triangles[i].k];
        gfObjBezierTetrahedra* tri = GetSubtetrahedron(&mid, &pt2, &pt3, &pt4);
        result.push_back(tri);
    }

    gfLog("Decomposition finished, number of tetra is %i, each tetrahedron has degree %i, time = %d ms.\n", result.size(), m_degree * (m_multi - 2), count.EndMSCount());

    return result;
}

GFRESULT gfObjBezierSVolume::RaiseDegree()
{
    // copy old indices
    std::vector<gfDouble> old_weights;
    std::vector<gfPoint3> old_control_points;
    gfInt num_old = m_control_points.size();
    for (gfInt i = 0;i < num_old;i++)
    {
        old_weights.push_back(m_weights[i]);
        gfPoint3 pt;
        pt.Set(m_control_points[i]->X(), m_control_points[i]->Y(), m_control_points[i]->Z());
        old_control_points.push_back(pt);
    }

    SetParams(m_degree + 1, m_multi);

    // compute new control points and weights
    // get multiindices for new SVolume
    std::vector< std::vector<gfInt> > multiindices;
    gfGetMultiindices(m_multi, m_degree, multiindices);

    // modify
    gfDouble inv_degree = 1.0 / m_degree;
    gfInt num_indices = (gfInt)multiindices.size();
    for (gfInt k = 0;k < num_indices;k++)
    {
        std::vector<gfInt> multi_index;
        multi_index.resize(m_multi);
        for (gfInt i = 0;i < m_multi;i++)
            multi_index[i] = multiindices[k][i];

        gfDouble x = 0.0;
        gfDouble y = 0.0;
        gfDouble z = 0.0;
        gfDouble w = 0.0;
        for (gfInt i = 0;i < m_multi;i++)
        {
            if (multi_index[i] == 0)
                continue;
            multi_index[i]--;
            gfInt old_index = GetIndexFromMultiindexVol(m_degree-1, multi_index);
            multi_index[i]++;
            gfDouble old_weight = old_weights[old_index];
            gfPoint3 old_pt = old_control_points[old_index];
            x += inv_degree * multi_index[i] * old_weight * old_pt.X();
            y += inv_degree * multi_index[i] * old_weight * old_pt.Y();
            z += inv_degree * multi_index[i] * old_weight * old_pt.Z();
            w += inv_degree * multi_index[i] * old_weight;
        }
        GetControlPoint(multiindices[k])->Set(x / w, y / w, z / w);
        SetWeight(multiindices[k], w);
    }

    m_time_stamp++;

    return GF_S_OK;
}

std::vector<gfObjBezierSPatch*> gfObjBezierSVolume::DecomposeToBoundarySPatches()
{
    std::vector<gfObjBezierSPatch*> result;

    std::vector<gfPoint3> domain_points;
    std::vector<gfTripleIndex> domain_triangles;
    CreateRegularDomain(m_multi, domain_points, domain_triangles);

    for (gfInt i = 0;i < (gfInt)domain_triangles.size();i++)
    {
        // it is bezier triangle
        gfObjBezierSPatch* tri = (gfObjBezierSPatch*)gfCreateGeomObject(GF_BEZIER_SPATCH);
        tri->SetParams(m_degree, 3);

        gfTripleIndex triple = domain_triangles[i];

        // get all multi-indices where only on three positions are non-zero indices
        for (gfInt k = 0;k <= m_degree;k++)
        {
            for (gfInt l = 0;l <= (m_degree-k);l++)
            {
                std::vector<gfInt> multiindex;
                multiindex.resize(m_multi, 0);
                multiindex[triple.i]=k;
                multiindex[triple.j]=l;
                multiindex[triple.k]=m_degree-k-l;

                // get point of S-volume
                gfPoint3* pt = GetControlPoint(multiindex);

                std::vector<gfInt> multiindex2;
                multiindex2.resize(3, 0);
                multiindex2[0]=k;
                multiindex2[1]=l;
                multiindex2[2]=m_degree-k-l;
                tri->GetControlPoint(multiindex2)->Set(pt->X(), pt->Y(), pt->Z());
                tri->SetWeight(multiindex2, GetWeight(multiindex));
            }
        }
        
        result.push_back(tri);
    }

    return result;
}

std::vector<gfObjBezierSVolume*> gfObjBezierSVolume::DecomposeToSVolumes(std::vector<gfDouble> value)
{
    std::vector<gfObjBezierSVolume*> result;

    // multiindices
    std::vector< std::vector<gfInt> > multiindices;
    gfGetMultiindices(m_multi, m_degree, multiindices);

    // create m_multi S-volumes
    for (gfInt j = 0;j < m_multi;j++)
    {
        gfObjBezierSVolume* vol = (gfObjBezierSVolume*)gfCreateGeomObject(GF_BEZIER_SVOLUME);

        vol->SetParams(m_degree, m_multi);

        // compute control points of new S-volume
        for (gfInt ind = 0;ind < (gfInt)multiindices.size();ind++)
        {
            std::vector<gfInt> index = multiindices[ind];

            // compute from auxiliary de Casteljau points
            gfDouble x = 0.0;
            gfDouble y = 0.0;
            gfDouble z = 0.0;
            gfDouble w = 0.0;

            std::vector< std::vector<gfInt> > aux_multi;
            gfGetMultiindices(m_multi, index[j], aux_multi);

            for (gfInt l = 0;l < (gfInt)aux_multi.size();l++)
            {
                std::vector<gfInt> aux_index;
                aux_index.resize(m_multi, 0);
                for (gfInt kkk = 0; kkk < m_multi;kkk++)
                    aux_index[kkk] = index[kkk];
                aux_index[j] = 0;
                for (gfInt kkk = 0; kkk < m_multi;kkk++)
                    aux_index[kkk] = aux_index[kkk] + aux_multi[l][kkk];

                gfDouble aux = GetWeight(aux_index) * gfBernsteinPolyGeneral(aux_multi[l], value);
                gfPoint3* pt = GetControlPoint(aux_index);

                x += aux * pt->X();
                y += aux * pt->Y();
                z += aux * pt->Z();
                w += aux;
            }

            vol->GetControlPoint(index)->Set(x / w, y / w, z / w);
            vol->SetWeight(index, w);
        }

        result.push_back(vol);
    }

    return result;
}
