//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjBezierTetrahedra.cpp                            //
//   Description: Rational Bezier tetrahedra class              //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfObjBezierTetrahedra.h"
#include "gfObjectFactory.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"

#include "./OGL/glut.h"

gfObjBezierTetrahedra::gfObjBezierTetrahedra(void)
{
    m_time_stamp = 0;
    m_weights.resize(0);
    m_control_points.resize(0);
    m_degree = 0;
}

gfObjBezierTetrahedra::~gfObjBezierTetrahedra(void)
{
    SetParams(0);
}

GFRESULT gfObjBezierTetrahedra::OGLDraw(gfLong flags)
{
    if (m_time_stamp == 0)
        return GF_S_OK;

    gfInt degree = GF_VIS_PREC / 2;
    gfPoint3 * aux_points = new gfPoint3[(degree*(degree+1)*(2*degree+1)/6+3*degree*(degree+1)/2+2*(degree+1))/2];

    for (gfInt i = 0;i < degree;i++)
    {
        // current u parameter
        gfDouble u = (gfDouble) i / (degree - 1);
        for (gfInt j = 0;j < degree - i;j++)
        {
            // current v parameter
            gfDouble v = (gfDouble) j / (degree - 1);
            for (gfInt k = 0;k < degree - i - j;k++)
            {
                // current w parameter     
                gfDouble w = (gfDouble) k / (degree - 1);
                gfDouble t = 1.0 - u - v - w;

                // compute point of bezier triangle for parameter (u, v, w, t)
                gfDouble ax = 0;
                gfDouble ay = 0;
                gfDouble az = 0;
                gfDouble aw = 0;

                for (gfInt l1 = 0;l1 <= m_degree;l1++)
                {
                    for (gfInt l2 = 0;l2 <= m_degree - l1;l2++)
                    {
                        for (gfInt l3 = 0;l3 <= m_degree - l1 - l2;l3++)
                        {
                            gfInt l4 = m_degree - l1 - l2 - l3;

                            gfDouble aux = GetWeight(l1, l2, l3, l4) * gfBernsteinPoly4(m_degree, l1, l2, l3, u, v, w);
                            ax += aux * GetControlPoint(l1, l2, l3, l4)->X();
                            ay += aux * GetControlPoint(l1, l2, l3, l4)->Y();
                            az += aux * GetControlPoint(l1, l2, l3, l4)->Z();
                            aw += aux;
                        }
                    }
                }

                gfInt n = degree;
                gfInt offset = GetPtsOffset(degree, i, j, k);
                aux_points[offset].Set(ax / aw, ay / aw, az / aw);
            }


        }
    }

    // fill triangle loops
    if (flags & GF_FLG_SOLID)
    {
        gfPoint3  ver;

        for (int i = 0;i <= degree - 1;i++)
        {
            for (int j = 0;j <= degree - 1 - i - 1;j++)
            {
                glBegin(GL_TRIANGLE_STRIP);
                for (int k = 0;k <= degree - 1 - i - j;k++)
                {
                    int l = degree - 1 - i - j - k;
                    ver = aux_points[GetPtsOffset(degree,i,j,k)];
                    glColor3f(float(i)/float(degree - 1),0.5+float(j)/float(2*degree - 1),float(k)/float(degree - 1));
                    glVertex3f(ver.X(),ver.Y(),ver.Z());
                    if (k > 0)
                    {
                        glColor3f(float(i)/float(degree - 1),0.5+float(j)/float(2*degree - 1),float(k)/float(degree - 1));
                        ver = aux_points[GetPtsOffset(degree,i,j +1,k - 1)];
                        glVertex3f(ver.X(),ver.Y(),ver.Z());
                    }

                }
                glEnd();
            }
        }

        for (int j = 0;j <= degree - 1;j++)
        {
            for (int i = 0;i <= degree - 1 - j - 1;i++)
            {
                glBegin(GL_TRIANGLE_STRIP);
                for (int k = 0;k <= degree - 1 - i - j;k++)
                {
                    int l = degree - 1 - i - j - k;
                    ver = aux_points[GetPtsOffset(degree,i,j,k)];
                    glColor3f(float(i)/float(degree - 1),0.5+float(j)/float(2*degree - 1),float(k)/float(degree - 1));
                    glVertex3f(ver.X(),ver.Y(),ver.Z());
                    if (k > 0)
                    {
                        glColor3f(float(i)/float(degree - 1),0.5+float(j)/float(2*degree - 1),float(k)/float(degree - 1));
                        ver = aux_points[GetPtsOffset(degree,i+1,j,k - 1)];
                        glVertex3f(ver.X(),ver.Y(),ver.Z());
                    }

                }
                glEnd();
            }
        }

        for (int k = 0;k <= degree - 1;k++)
        {
            for (int i = 0;i <= degree - 1 - k - 1;i++)
            {
                glBegin(GL_TRIANGLE_STRIP);
                for (int j = 0;j <= degree - 1 - i - k;j++)
                {
                    int l = degree - 1 - i - j - k;
                    ver = aux_points[GetPtsOffset(degree,i,j,k)];
                    glColor3f(float(i)/float(degree - 1),0.5+float(j)/float(2*degree - 1),float(k)/float(degree - 1));
                    glVertex3f(ver.X(),ver.Y(),ver.Z());
                    if (j > 0)
                    {
                        glColor3f(float(i)/float(degree - 1),0.5+float(j)/float(2*degree - 1),float(k)/float(degree - 1));
                        ver = aux_points[GetPtsOffset(degree,i+1,j-1,k)];
                        glVertex3f(ver.X(),ver.Y(),ver.Z());
                    }

                }
                glEnd();
            }
        }

        for (int l = 0;l <= degree - 1;l++)
        {
            for (int i = 0;i <= degree - 1 - l - 1;i++)
            {
                glBegin(GL_TRIANGLE_STRIP);
                for (int j = 0;j <= degree- 1 - i - l;j++)
                {
                    int k = degree - 1 - i - j - l;
                    ver = aux_points[GetPtsOffset(degree,i,j,k)];
                    glColor3f(float(i)/float(degree - 1),0.5+float(j)/float(2*degree - 1),float(k)/float(degree - 1));
                    glVertex3f(ver.X(),ver.Y(),ver.Z());
                    if (j > 0)
                    {
                        glColor3f(float(i)/float(degree - 1),0.5+float(j)/float(2*degree - 1),float(k)/float(degree - 1));
                        ver = aux_points[GetPtsOffset(degree,i+1,j-1,k)];
                        glVertex3f(ver.X(),ver.Y(),ver.Z());
                    }

                }
                glEnd();
            }
        }
    }

    if (flags & GF_FLG_WIRE)
    {
        glColor3f(0,0,0);
        for (int i = 1;i <= degree - 1;i++)
        {
            for (int j = 0;j <= degree - 1 - i;j++)
            {
                for (int k=0;k <= degree - 1 - i - j;k++)
                {
                    int l = degree - 1 - i - j - k;
                    gfPoint3  ver1 = aux_points[GetPtsOffset(degree,i,j,k)];
                    gfPoint3  ver2 = aux_points[GetPtsOffset(degree,i-1,j+1,k)];
                    gfPoint3  ver3 = aux_points[GetPtsOffset(degree,i-1,j,k+1)];
                    gfPoint3  ver4 = aux_points[GetPtsOffset(degree,i-1,j,k)];

                    glBegin(GL_LINES);
                    glVertex3f(ver1.X(),ver1.Y(),ver1.Z());
                    glVertex3f(ver2.X(),ver2.Y(),ver2.Z());

                    glVertex3f(ver1.X(),ver1.Y(),ver1.Z());
                    glVertex3f(ver3.X(),ver3.Y(),ver3.Z());

                    glVertex3f(ver1.X(),ver1.Y(),ver1.Z());
                    glVertex3f(ver4.X(),ver4.Y(),ver4.Z());

                    glVertex3f(ver3.X(),ver3.Y(),ver3.Z());
                    glVertex3f(ver2.X(),ver2.Y(),ver2.Z());

                    glVertex3f(ver4.X(),ver4.Y(),ver4.Z());
                    glVertex3f(ver2.X(),ver2.Y(),ver2.Z());

                    glVertex3f(ver4.X(),ver4.Y(),ver4.Z());
                    glVertex3f(ver3.X(),ver3.Y(),ver3.Z());
                    glEnd();
                }
            }
        }
    }

    // draw control net
    if (flags & GF_FLG_CONTROL_NET)
    {
        glNormal3f(0.0, 0.0, 0.0);
        glColor3f(0.0, 0.0, 0.0);
        DrawControlNet();
    }

    delete [] aux_points;         

    return GF_S_OK;
}

gfRayInterInfo* gfObjBezierTetrahedra::RayIntersection(gfMatrix4* obj_matrix, gfPoint3 * ray_begin, gfVector3* ray_dir)
{
    gfRayInterInfo* result = NULL;
    result = gfCreateRayInterInfo();

    result->m_inters.push_back(1.0);
    result->m_normals.push_back(gfVector3(0.0, 0.0, 0.0));

    return result;
}

gfGeomObjID gfObjBezierTetrahedra::GetID(void)
{
    return GF_BEZIER_TETRAHEDRA;
}

GFRESULT gfObjBezierTetrahedra::Transform(gfMatrix4* matrix)
{
    // transform control points
    for (int i = 0;i <= m_degree;i++)
    {
        for (int j = 0;j <= m_degree - i;j++)
        {
            for (int k = 0;k <= m_degree - i - j;k++)
            {
                gfPoint3* pt = GetControlPoint(i, j, k, m_degree - i - j - k);
                gfTransformPoint3(pt, matrix);
            }
        }
    }

    return GF_S_OK;
}

gfObjGeomObject* gfObjBezierTetrahedra::Clone(void)
{
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());

    gfObjBezierTetrahedra* tetra = (gfObjBezierTetrahedra*) clone;
    tetra->SetParams(m_degree);

    for (gfInt i = 0;i <=  m_degree;i++)
    {
        for (gfInt j = 0;j <= m_degree - i;j++)
        {
            for (gfInt k = 0;k <= m_degree - i - j;k++)
            {
                gfPoint3* old_pt = GetControlPoint(i, j, k, m_degree - i - j - k);
                gfPoint3* new_pt = tetra->GetControlPoint(i, j, k, m_degree - i - j - k);
                new_pt->Set(old_pt->X(), old_pt->Y(), old_pt->Z());

                tetra->SetWeight(i, j, k, m_degree - i - j - k, GetWeight(i, j, k, m_degree - i - j - k));
            }
        }
    }

    return clone;
}

GFRESULT gfObjBezierTetrahedra::InitTetrahedron(gfPoint3* center, gfDouble size)
{
    SetParams(1);
    gfDouble sq = gfSquareRoot(3.0);

    gfPoint3* pt = GetControlPoint(1, 0, 0, 0);
    pt->Set(sq*size/3.0, 0.0, 0.0);
    SetWeight(1, 0, 0, 0, 1.0);

    pt = GetControlPoint(0, 1, 0, 0);
    pt->Set(-sq*size/6.0, -size/2.0, 0.0);
    SetWeight(0, 1, 0, 0, 1.0);

    pt = GetControlPoint(0, 0, 1, 0);
    pt->Set(-sq*size/6.0, size/2.0, 0.0);
    SetWeight(0, 0, 1, 0, 1.0);

    pt = GetControlPoint(0, 0, 0, 1);
    pt->Set(0.0, 0.0, size*gfSquareRoot(2.0)/sq);
    SetWeight(0, 0, 0, 1, 1.0);

    return GF_S_OK;
}

GFRESULT gfObjBezierTetrahedra::InitCone(gfPoint3* center, gfDouble radius, gfDouble height)
{
	SetParams(2);
    
    gfPoint3* aux;
    gfDouble sq = gfSquareRoot(3.0);

    aux = GetControlPoint(2, 0, 0, 0);
    aux->Set(0.0, -radius, 0.0);
    SetWeight(2, 0, 0, 0, 1.0);

    aux = GetControlPoint(1, 0, 1, 0);
    aux->Set(-sq*radius, -radius, 0.0);
    SetWeight(1, 0, 1, 0, 0.5);

    aux = GetControlPoint(1, 1, 0, 0);
    aux->Set(sq*radius, -radius, 0.0);
    SetWeight(1, 1, 0, 0, 0.5);

    aux = GetControlPoint(0, 0, 2, 0);
    aux->Set(-sq*radius/2.0, radius/2.0, 0.0);
    SetWeight(0, 0, 2, 0, 1.0);

    aux = GetControlPoint(0, 2, 0, 0);
    aux->Set(sq*radius/2.0, radius/2.0, 0.0);
    SetWeight(0, 2, 0, 0, 1.0);

    aux = GetControlPoint(0, 1, 1, 0);
    aux->Set(0.0, 2.0*radius, 0.0);
    SetWeight(0, 1, 1, 0, 0.5);

    aux = GetControlPoint(0, 1, 0, 1);
	aux->Set(sq*radius/4.0, radius/4.0, height/2.0);
    SetWeight(0, 1, 0, 1, 0.0);

    aux = GetControlPoint(1, 0, 0, 1);
    aux->Set(0.0, -radius/2.0, height/2.0);
    SetWeight(1, 0, 0, 1, 0.0);

    aux = GetControlPoint(0, 0, 1, 1);
    aux->Set(-sq*radius/4.0, radius/4.0, height/2.0);
    SetWeight(0, 0, 1, 1, 0.0);
 
    aux = GetControlPoint(0, 0, 0, 2);
    aux->Set(0.0, 0.0, height);
    SetWeight(0, 0, 0, 2, 1.0);

    return GF_S_OK;
}

GFRESULT gfObjBezierTetrahedra::InitSphereSector(gfPoint3* center, gfDouble radius)
{
	SetParams(2);
    
    gfPoint3* aux;
    gfDouble sq = gfSquareRoot(3.0)/2.0;

    aux = GetControlPoint(2, 0, 0, 0);
    aux->Set(radius, 0.0, 0.0);
    SetWeight(2, 0, 0, 0, 1.0);

    aux = GetControlPoint(1, 0, 1, 0);
    aux->Set(radius, 0.0, radius);
    SetWeight(1, 0, 1, 0, sq);

    aux = GetControlPoint(1, 1, 0, 0);
    aux->Set(radius, radius, 0.0);
    SetWeight(1, 1, 0, 0, sq);

    aux = GetControlPoint(0, 0, 2, 0);
    aux->Set(0.0, 0.0, radius);
    SetWeight(0, 0, 2, 0, sq);

    aux = GetControlPoint(0, 2, 0, 0);
    aux->Set(0.0, radius, 0.0);
    SetWeight(0, 2, 0, 0, 1.0);

    aux = GetControlPoint(0, 1, 1, 0);
    aux->Set(0.0, radius, radius);
    SetWeight(0, 1, 1, 0, sq);

    aux = GetControlPoint(0, 1, 0, 1);
	aux->Set(0.0, radius/2.0, 0.0);
    SetWeight(0, 1, 0, 1, 0.0);

    aux = GetControlPoint(1, 0, 0, 1);
    aux->Set(radius/2.0, 0.0, 0.0);
    SetWeight(1, 0, 0, 1, 0.0);

    aux = GetControlPoint(0, 0, 1, 1);
    aux->Set(0.0, 0.0, radius/2.0);
    SetWeight(0, 0, 1, 1, 0.0);
 
    aux = GetControlPoint(0, 0, 0, 2);
    aux->Set(0.0, 0.0, 0.0);
    SetWeight(0, 0, 0, 2, 1.0);

    return GF_S_OK;
}

GFRESULT gfObjBezierTetrahedra::InitRandom(gfPoint3* center, gfDouble tx, gfDouble ty, gfDouble tz)
{
    SetParams(4);

    for (int i = 0;i <= m_degree;i++)
    {
        for (int j = 0;j <= m_degree - i;j++)
        {
            for (int k = 0;k <= m_degree - i - j;k++)
            {
                gfDouble ax = (- tx + 2 * tx * i / 2.0 + /*gfRand(0.8)*/ - 0.2) / 2.0;
                gfDouble ay = (- ty + 2 * ty * j / 2.0 + /*gfRand(0.8)*/ - 0.2) / 2.0;
                gfDouble az = (- tz + 2 * tz * k / 2.0 + /*gfRand(0.8)*/ - 0.2) / 2.0;
                GetControlPoint(i, j, k, m_degree - i - j - k)->Set(center->X() + ax, center->Y() + ay, center->Z() + az);
                SetWeight(i, j, k, m_degree - i - j - k, 1.0);
            }
        }
    }

    return GF_S_OK;
}

GFRESULT gfObjBezierTetrahedra::SetParams(gfInt degree)
{
    // create auxiliary arrays and fill them with old 
    std::vector<gfDouble> aux_weights;
    std::vector<gfPoint3 *> aux_control_points;
    gfInt num_control_points = (degree*(degree+1)*(2*degree+1)/6+3*degree*(degree+1)/2+2*(degree+1))/2;
    aux_weights.resize(num_control_points);
    aux_control_points.resize(num_control_points);

    gfInt max = m_degree > degree ? m_degree : degree;

    gfInt n = degree;
    for (gfInt i = 0;i <=  max;i++)
    {
        for (gfInt j = 0;j <= max - i;j++)
        {
            for (gfInt k = 0;k <= max - i - j;k++)
            {
                gfInt l = max - i - j - k;

                gfInt offset = GetPtsOffset(degree + 1, i, j, k);
                // point in new & old
                if ((i + j + k + l <= m_degree) && (i + j + k + l <= degree))
                {
                    aux_weights[offset] = GetWeight(i, j, k, l);
                    aux_control_points[offset] = GetControlPoint(i, j, k, l);
                }

                // point in new & not in old
                if ((i + j + k + l > m_degree) && (i + j + k + l <= degree))
                {
                    aux_weights[offset] = 1.0;
                    gfPoint3* new_pt = gfCreatePoint3();
                    new_pt->Set(i /*+ rand()/double(RAND_MAX)*/, j /*+ rand()/double(RAND_MAX)*/, k /*+ rand()/double(RAND_MAX)*/);
                    aux_control_points[offset] = new_pt;
                }

                // point in old & not in new
                if ((i + j + k + l <= m_degree) && (i + j + k + l > degree))
                {
                    gfPoint3* old_pt = GetControlPoint(i, j, k, l);
                    gfDeletePoint3(old_pt);
                }
            }
        }
    }

    m_control_points.resize(num_control_points);
    m_weights.resize(num_control_points);
    m_degree = degree;

    // copy from auxiliary arrays
    for (gfInt i = 0;i < num_control_points;i++)
    {
        m_control_points[i] = aux_control_points.at(i);
        m_weights[i] = aux_weights.at(i);
    }

    m_time_stamp++;

    return GF_S_OK;
}

gfInt gfObjBezierTetrahedra::GetDegree(void)
{
    return m_degree;
}

gfInt gfObjBezierTetrahedra::GetNumControlPoints(void)
{
    return (gfInt)m_control_points.size();
}

gfPoint3 * gfObjBezierTetrahedra::GetControlPoint(gfInt u_index, gfInt v_index, gfInt w_index, gfInt t_index)
{
    gfInt offset = GetPtsOffset(m_degree + 1, u_index, v_index, w_index);
    return m_control_points.at(offset);
}

gfDouble gfObjBezierTetrahedra::GetWeight(gfInt u_index, gfInt v_index, gfInt w_index, gfInt t_index)
{
    gfInt offset = GetPtsOffset(m_degree + 1, u_index, v_index, w_index);
    return m_weights.at(offset);
}

GFRESULT gfObjBezierTetrahedra::SetWeight(gfInt u_index, gfInt v_index, gfInt w_index, gfInt t_index, gfDouble weight)
{
    m_time_stamp++;
    gfInt offset = GetPtsOffset(m_degree + 1, u_index, v_index, w_index);
    return m_weights[offset] = weight;
}

gfInt gfObjBezierTetrahedra::GetPtsOffset(gfInt n, gfInt u_index, gfInt v_index, gfInt w_index)
{
    gfInt i = u_index;
    gfInt j = v_index;
    gfInt k = w_index;

    return (i*(n-k)-i*(i-1)/2+j+((n-k-2)*(n-k-1)*(2*n-2*k-3)/6+3*(n-k-1)*(n-k-2)/2+2*(n-k-1))/2);
}

GFRESULT gfObjBezierTetrahedra::DrawControlNet(void)
{
    glColor3f(0,0,0);
    for (int i = 1;i <= m_degree;i++)
    {
        for (int j = 0;j <= m_degree - i;j++)
        {
            for (int k=0;k <= m_degree - i - j;k++)
            {
                int l = m_degree - i - j - k;
                gfPoint3 * ver1 = m_control_points.at(GetPtsOffset(m_degree + 1,i,j,k));
                gfPoint3 * ver2 = m_control_points.at(GetPtsOffset(m_degree + 1,i-1,j+1,k));
                gfPoint3 * ver3 = m_control_points.at(GetPtsOffset(m_degree + 1,i-1,j,k+1));
                gfPoint3 * ver4 = m_control_points.at(GetPtsOffset(m_degree + 1,i-1,j,k));

                glBegin(GL_LINES);
                glVertex3f(ver1->X(),ver1->Y(),ver1->Z());
                glVertex3f(ver2->X(),ver2->Y(),ver2->Z());

                glVertex3f(ver1->X(),ver1->Y(),ver1->Z());
                glVertex3f(ver3->X(),ver3->Y(),ver3->Z());

                glVertex3f(ver1->X(),ver1->Y(),ver1->Z());
                glVertex3f(ver4->X(),ver4->Y(),ver4->Z());

                glVertex3f(ver3->X(),ver3->Y(),ver3->Z());
                glVertex3f(ver2->X(),ver2->Y(),ver2->Z());

                glVertex3f(ver4->X(),ver4->Y(),ver4->Z());
                glVertex3f(ver2->X(),ver2->Y(),ver2->Z());

                glVertex3f(ver4->X(),ver4->Y(),ver4->Z());
                glVertex3f(ver3->X(),ver3->Y(),ver3->Z());
                glEnd();

                glPointSize(8.0);
                glBegin(GL_POINTS);
                glVertex3f(ver1->X(),ver1->Y(),ver1->Z());
                glEnd();
                glPointSize(1.0);
            }
        }
    }

    glPointSize(8.0);
    glBegin(GL_POINTS);
    for (int i = 0;i <= m_degree;i++)
    {
        for (int j = 0;j <= m_degree - i;j++)
        {
            for (int k=0;k <= m_degree - i - j;k++)
            {
                int l = m_degree - i - j - k;
                gfPoint3 * ver1 = m_control_points.at(GetPtsOffset(m_degree + 1,i,j,k));
                glVertex3f(ver1->X(),ver1->Y(),ver1->Z());
            }
        }
    }
    glEnd();
    glPointSize(1.0);

    return GF_S_OK;
}

gfAuxBoundingVolume* gfObjBezierTetrahedra::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    // create auxiliary array of mesh points
    gfInt num_points = (gfInt)m_control_points.size();
    if (num_points < 1)
        return NULL;

    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfInt i = 0; i < num_points; i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        curr_point->Set(pt->X(), pt->Y(), pt->Z());
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    delete [] points;
    return result;
}

gfObjBezierTetrahedra* gfObjBezierTetrahedra::GetContinuousTetrahedronU(gfInt order)
{
    gfObjBezierTetrahedra* result = (gfObjBezierTetrahedra*)gfCreateGeomObject(GF_BEZIER_TETRAHEDRA);

    result->SetParams(m_degree);

    // fill vertices
    for (gfInt i = 0;i <= m_degree;i++)
    {
        for (gfInt j = 0;j <= m_degree-i;j++)
        {
            for (gfInt k = 0;k <= m_degree-i-j;k++)
            {
                
            }
        }
    }
    return result;
}