//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjBezierTriangle.cpp                              //
//   Description: Rational Bezier triangle class                //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfObjBezierTriangle.h"
#include "gfObjectFactory.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfAuxItersectionUtils.h"
#include "gfWinFPSCounter.h"
#include "gfAuxLogging.h"

#include "./OGL/glut.h"

gfObjBezierTriangle::gfObjBezierTriangle(void)
{
    m_time_stamp = 0;
    m_weights.resize(0);
    m_control_points.resize(0);
    m_degree = 0;
    m_old_stamp = 0;
}

gfObjBezierTriangle::~gfObjBezierTriangle(void)
{
    SetParams(0);
}

GFRESULT gfObjBezierTriangle::OGLDraw(gfLong flags)
{
    if (m_time_stamp == 0)
        return GF_S_OK;

    // draw possibly control points, it isn't cached
    if (flags & GF_FLG_CONTROL_NET)
    {
        glColor3f(0.0, 0.0, 0.0);
        for (gfInt i = 0;i < m_degree;i++)
        {
            for (gfInt j = 0;j < m_degree-i;j++)
            {
                gfPoint3* pt1 = GetControlPoint(i, j, m_degree-i-j);
                gfPoint3* pt2 = GetControlPoint(i+1, j, m_degree-i-j-1);
                gfPoint3* pt3 = GetControlPoint(i, j+1, m_degree-i-j-1);

                glBegin(GL_LINE_LOOP);
                glVertex3f(pt1->X(), pt1->Y(), pt1->Z());
                glVertex3f(pt2->X(), pt2->Y(), pt2->Z());
                glVertex3f(pt3->X(), pt3->Y(), pt3->Z());
                glEnd();
            }
        }

        glPointSize(8.0);
        glBegin(GL_POINTS);
        for (gfInt i = 0;i <= m_degree;i++)
        {
            for (gfInt j = 0;j <= m_degree-i;j++)
            {
                gfPoint3* pt = GetControlPoint(i, j, m_degree-i-j);
                glVertex3f(pt->X(), pt->Y(), pt->Z());
            }
        }
        glEnd();
        glPointSize(1.0);
    }

    if (m_time_stamp != m_old_stamp)
    {
        // fill cached points
        m_cached_points.resize(GF_VIS_PREC * (GF_VIS_PREC + 1) / 2);
        m_cached_normals.resize(GF_VIS_PREC * (GF_VIS_PREC + 1) / 2);
        for (gfInt i = 0;i < GF_VIS_PREC;i++)
        {
            // current u parameter
            gfDouble u = (gfDouble) i / (GF_VIS_PREC - 1);

            for (gfInt j = 0;j < GF_VIS_PREC - i;j++)
            {
                // current v parameter     
                gfDouble v = (gfDouble) j / (GF_VIS_PREC - 1);
                gfDouble w = 1.0 - u - v;

                // compute point of bezier triangle for parameter (u, v, w)
                gfDouble ax = 0;
                gfDouble ay = 0;
                gfDouble az = 0;
                gfDouble aw = 0;

                for (gfInt k = 0;k <= m_degree;k++)
                {
                    for (gfInt l = 0;l <= m_degree - k;l++)
                    {
                        gfInt m = m_degree - k - l;

                        gfDouble aux = GetWeight(k, l, m) * gfBernsteinPoly3(m_degree, k, l, u, v);
                        ax += aux * GetControlPoint(k, l, m)->X();
                        ay += aux * GetControlPoint(k, l, m)->Y();
                        az += aux * GetControlPoint(k, l, m)->Z();
                        aw += aux;
                    }
                }

                m_cached_points[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j].Set(ax / aw, ay / aw, az / aw);

                // compute normal
                gfVector3* normal = GetNormal(u, v, w);
                m_cached_normals[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j].SetX(normal->X());
                m_cached_normals[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j].SetY(normal->Y());
                m_cached_normals[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j].SetZ(normal->Z());
                gfDeleteVector3(normal);
            }
        }
    }
    

    // fill triangle loops
    if (flags & GF_FLG_SOLID)
    {
        for (gfInt i = 0;i < GF_VIS_PREC - 1;i++)
        {
            glBegin(GL_TRIANGLE_STRIP);
            for (gfInt j = 0;j < GF_VIS_PREC - i;j++)
            {
                gfPoint3 aux = m_cached_points[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j];
                if (flags & GF_FLG_NORMALS)
                {
                    gfVector3 normal = m_cached_normals[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j];
                    glNormal3f(normal.X(), normal.Y(), normal.Z());
                }
                glVertex3f(aux.X(), aux.Y(), aux.Z());

                if (j < GF_VIS_PREC - i - 1)
                {
                    aux = m_cached_points[(GF_VIS_PREC) * (i + 1) - (i + 1) * (i) / 2 + j];
                    if (flags & GF_FLG_NORMALS)
                    {
                        gfVector3 normal = m_cached_normals[(GF_VIS_PREC) * (i + 1) - (i + 1) * (i) / 2 + j];
                        glNormal3f(normal.X(), normal.Y(), normal.Z());
                    }
                    glVertex3f(aux.X(), aux.Y(), aux.Z());
                }
            }
            glEnd();
        }
    }

    if (flags & GF_FLG_WIRE)
    {
        glColor3f(0.0, 0.0, 0.0);
        for (gfInt i = 0;i < GF_VIS_PREC - 1;i++)
        {
            for (gfInt j = 0;j < GF_VIS_PREC - i - 1;j++)
            {
                glBegin(GL_LINE_LOOP);
                gfPoint3 aux = m_cached_points[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j];
                glVertex3f(aux.X(), aux.Y(), aux.Z());
                aux = m_cached_points[(GF_VIS_PREC) * (i + 1) - (i + 1) * (i) / 2 + j];
                glVertex3f(aux.X(), aux.Y(), aux.Z());
                aux = m_cached_points[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j + 1];
                glVertex3f(aux.X(), aux.Y(), aux.Z());
                glEnd();
            }
        }
    }

    m_old_stamp = m_time_stamp;         

    return GF_S_OK;
}

gfRayInterInfo* gfObjBezierTriangle::RayIntersection(gfMatrix4* obj_matrix, gfPoint3 * ray_begin, gfVector3* ray_dir)
{
    //return RayControlMeshIntersection(obj_matrix, ray_begin, ray_dir);
    //return RayApproxMeshIntersection(obj_matrix, ray_begin, ray_dir);

    // using AABB tree of nested patches
    gfInt num_trees = (gfInt)m_bezier_aabb_trees.size();
    if (num_trees == 0)
        ComputeWholeAABBTree();
    num_trees = (gfInt)m_bezier_aabb_trees.size();

    gfDisableLogging();

    gfRayInterInfo* info = NULL;
    for (gfInt i = 0;i < num_trees;i++)
    {
        gfBezierTriangleAABBTreeNode* root = m_bezier_aabb_trees.at(i);
        gfRayInterInfo* info_temp = RayAABBIntersection(root, ray_begin, ray_dir);

        if (!info)
        {
            info = gfCreateRayInterInfo();
        }

        if (info_temp)
        {

            // copy normal if necessary
            for (gfInt i = 0;i < (gfInt)info_temp->m_inters.size();i++)
            {
                info->m_inters.push_back(info_temp->m_inters.at(i));
                info->m_normals.push_back(info_temp->m_normals.at(i));
            }
            gfDeleteRayInterInfo(info_temp);
            gfSortRayInterInfo(info);
        }
    }

    gfEnableLogging();
    if (info && (info->m_inters.size() == 0))
    {
        gfDeleteRayInterInfo(info);
        info = NULL;
    }

    return info;
}

gfGeomObjID gfObjBezierTriangle::GetID(void)
{
    return GF_BEZIER_TRIANGLE;
}

GFRESULT gfObjBezierTriangle::Transform(gfMatrix4* matrix)
{
    // delete AABB tree
    DeleteWholeAABBTree();

    for (gfInt k = 0;k <= m_degree;k++)
    {
        for (gfInt l = 0;l <= m_degree - k;l++)
        {
            gfInt m = m_degree - k - l;

            gfPoint3* point = GetControlPoint(k, l, m);
            if (point)
            {
                gfTransformPoint3(point, matrix);
            }
        }
    }

    return GF_S_OK;
}

gfObjGeomObject* gfObjBezierTriangle::Clone(void)
{
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());

    gfObjBezierTriangle* triangle = (gfObjBezierTriangle*) clone;
    triangle->SetParams(m_degree);

    for (gfInt i = 0;i <=  m_degree;i++)
    {
        for (gfInt j = 0;j <= m_degree - i;j++)
        {
                gfPoint3* old_pt = GetControlPoint(i, j, m_degree - i - j);
                gfPoint3* new_pt = triangle->GetControlPoint(i, j, m_degree - i - j);
                new_pt->Set(old_pt->X(), old_pt->Y(), old_pt->Z());

                triangle->SetWeight(i, j, m_degree - i - j , GetWeight(i, j, m_degree - i - j));
        }
    }

    return clone;
}

GFRESULT gfObjBezierTriangle::SetParams(gfInt degree)
{
    DeleteWholeAABBTree();

    // create auxiliary arrays and fill them with old 
    std::vector<gfDouble> aux_weights;
    std::vector<gfPoint3 *> aux_control_points;
    gfInt num_control_points = (degree + 1) * (degree + 2) / 2;
    aux_weights.resize(num_control_points);
    aux_control_points.resize(num_control_points);

    gfInt max = m_degree > degree ? m_degree : degree;

    for (gfInt i = 0;i <=  max;i++)
    {
        for (gfInt j = 0;j <= max - i;j++)
        {
            gfInt k = max - i - j;

            // point in new & old
            if ((i + j + k <= m_degree) && (i + j + k <= degree))
            {
                aux_weights[i * (degree + 1) - (i) * (i - 1) / 2 + j] = GetWeight(i, j, k);
                aux_control_points[i * (degree + 1) - (i) * (i - 1) / 2 + j] = GetControlPoint(i, j, k);
            }

            // point in new & not in old
            if ((i + j + k > m_degree) && (i + j + k <= degree))
            {
                aux_weights[i * (degree + 1) - (i) * (i - 1) / 2 + j] = 1.0;
                gfPoint3 * new_pt = gfCreatePoint3();
                aux_control_points[i * (degree + 1) - (i) * (i - 1) / 2 + j] = new_pt;
            }

            // point in old & not in new
            if ((i + j + k <= m_degree) && (i + j + k > degree))
            {
                gfPoint3 * old_pt = GetControlPoint(i, j, k);
                gfDeletePoint3(old_pt);
            }
        }
    }

    m_control_points.resize(num_control_points);
    m_weights.resize(num_control_points);
    m_degree = degree;

    // copy from auxiliary arrays
    for (gfInt i = 0;i < num_control_points;i++)
    {
        m_control_points[i] = aux_control_points.at(i);
        m_weights[i] = aux_weights.at(i);
    }

    m_time_stamp++;

    return GF_S_OK;
}

gfInt gfObjBezierTriangle::GetDegree(void)
{
    return m_degree;
}

gfInt gfObjBezierTriangle::GetNumControlPoints(void)
{
    return (gfInt)m_control_points.size();
}

gfPoint3 * gfObjBezierTriangle::GetControlPoint(gfInt u_index, gfInt v_index, gfInt w_index)
{
    return m_control_points.at(u_index * (m_degree + 1) - (u_index) * (u_index - 1) / 2 + v_index);
}

gfDouble gfObjBezierTriangle::GetWeight(gfInt u_index, gfInt v_index, gfInt w_index)
{
    return m_weights.at(u_index * (m_degree + 1) - (u_index) * (u_index - 1) / 2 + v_index);
}

GFRESULT gfObjBezierTriangle::SetWeight(gfInt u_index, gfInt v_index, gfInt w_index, gfDouble weight)
{
    m_time_stamp++;
    return m_weights[u_index * (m_degree + 1) - (u_index) * (u_index - 1) / 2 + v_index] = weight;
}

gfVector3* gfObjBezierTriangle::GetNormal(gfDouble u, gfDouble v, gfDouble w)
{
    gfVector3* result = gfCreateVector3();

    // do the deCasteljau algorithm
    std::vector<gfPoint4> aux_points((m_degree+1)*(m_degree+1)*(m_degree+1));
    // fill points with control points
    for (gfInt i = 0;i <= m_degree;i++)
    {
        for (gfInt j = 0;j <= m_degree-i;j++)
        {
            gfPoint3* pt = GetControlPoint(i, j, m_degree - i - j);
            gfDouble ww = GetWeight(i, j, m_degree - i - j);

            aux_points[i + (m_degree+1)*j + 0*(m_degree+1)*(m_degree+1)].Set(ww*pt->X(), ww*pt->Y(), ww*pt->Z(), ww);
        }
    }

    for (gfInt p = 1;p <= m_degree;p++)
    {
        // p is top index
        for (gfInt i = 0;i <= m_degree-p;i++)
        {
            for (gfInt j = 0;j <= m_degree-p-i;j++)
            {
                gfPoint4 v1 = aux_points[i+1+(m_degree+1)*j+(m_degree+1)*(m_degree+1)*(p-1)];
                gfPoint4 v2 = aux_points[i+(m_degree+1)*(j+1)+(m_degree+1)*(m_degree+1)*(p-1)];
                gfPoint4 v3 = aux_points[i+(m_degree+1)*j+(m_degree+1)*(m_degree+1)*(p-1)];
                aux_points[i+(m_degree+1)*j+(m_degree+1)*(m_degree+1)*p].Set(u*v1.X()+v*v2.X()+w*v3.X(), u*v1.Y()+v*v2.Y()+w*v3.Y(), u*v1.Z()+v*v2.Z()+w*v3.Z(), u*v1.W()+v*v2.W()+w*v3.W());
            }
        }
    }

    // now compute last step of casteljau
    gfPoint4 pt = aux_points[0 + (m_degree+1)*0 + (m_degree+1)*(m_degree+1)*(m_degree-1)];
    gfDouble w1 = pt.W();
    gfPoint3 p1(pt.X(), pt.Y(), pt.Z());
    if (gfAbs(w1) > GF_EPS)
        p1.Set(pt.X() / w1, pt.Y() / w1, pt.Z() / w1);

    pt = aux_points[1 + (m_degree+1)*0 + (m_degree+1)*(m_degree+1)*(m_degree-1)];
    gfDouble w2 = pt.W();
    gfPoint3 p2(pt.X(), pt.Y(), pt.Z());
    if (gfAbs(w2) > GF_EPS)
        p2.Set(pt.X() / w2, pt.Y() / w2, pt.Z() / w2);

    pt = aux_points[0 + (m_degree+1)*1 + (m_degree+1)*(m_degree+1)*(m_degree-1)];
    gfDouble w3 = pt.W();
    gfPoint3 p3(pt.X(), pt.Y(), pt.Z());
    if (gfAbs(w3) > GF_EPS)
        p3.Set(pt.X() / w3, pt.Y() / w3, pt.Z() / w3);


    gfVector3 v2(p2.X() - p1.X(), p2.Y() - p1.Y(), p2.Z() - p1.Z());
    gfVector3 v1(p3.X() - p1.X(), p3.Y() - p1.Y(), p3.Z() - p1.Z());
    gfVectorProduct3(&v1, &v2, result);
    result->Normalize();

    return result;
}

gfObjHalfEdgeMesh* gfObjBezierTriangle::GetControlMesh(void)
{
    gfObjHalfEdgeMesh* ret = (gfObjHalfEdgeMesh*) gfCreateGeomObject(GF_HALF_EDGE_MESH);

    return ret;
}

gfObjHalfEdgeMesh* gfObjBezierTriangle::GetApproxMesh(gfInt approx)
{
    gfObjHalfEdgeMesh* ret = (gfObjHalfEdgeMesh*) gfCreateGeomObject(GF_HALF_EDGE_MESH);

    gfHalfEdge** edges = new gfHalfEdge*[3 * (approx) * (approx - 1)];
    gfHEPoint** points = new gfHEPoint*[approx * (approx + 1) / 2];

    for (gfInt i = 0;i < approx;i++)
    {
        // current u parameter
        gfDouble u = (gfDouble) i / (approx - 1);

        for (gfInt j = 0;j < approx - i;j++)
        {
            // current v parameter     
            gfDouble v = (gfDouble) j / (approx - 1);
            gfDouble w = 1.0 - u - v;

            // compute point of bezier triangle for parameter (u, v, w)
            gfDouble ax = 0;
            gfDouble ay = 0;
            gfDouble az = 0;
            gfDouble aw = 0;

            for (gfInt k = 0;k <= m_degree;k++)
            {
                for (gfInt l = 0;l <= m_degree - k;l++)
                {
                    gfInt m = m_degree - k - l;

                    gfDouble aux = GetWeight(k, l, m) * gfBernsteinPoly3(m_degree, k, l, u, v);
                    ax += aux * GetControlPoint(k, l, m)->X();
                    ay += aux * GetControlPoint(k, l, m)->Y();
                    az += aux * GetControlPoint(k, l, m)->Z();
                    aw += aux;
                }
            }

            gfHEPoint* point = gfCreateHEPoint();
            gfSetHEPointCoords(point, ax / aw, ay / aw, az / aw);
            point->edge = NULL;
            points[(approx) * i - (i) * (i - 1) / 2 + j] = point;
            ret->AddHEPoint(point);
        }
    }

    for (gfInt i = 0;i < (approx - 1);i++)
    {
        for (gfInt j = 0;j < 2 * i + 1;j++)
        {
            // create face
            //gfHEFace* face = gfCreateHEFace();

            // create three edges of face
            //gfHalfEdge* edge1 = gfCreateHalfEdge();
            //edge1->face = face;
            //edge1->end = points;
        }
    }

    return ret;
}

gfAuxBoundingVolume* gfObjBezierTriangle::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    // create auxiliary array of mesh points
    gfInt num_points = (gfInt)m_control_points.size();
    if (num_points < 1)
        return NULL;

    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfInt i = 0; i < num_points; i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        curr_point->Set(pt->X(), pt->Y(), pt->Z());
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    delete [] points;
    return result;
}

GFRESULT gfObjBezierTriangle::InitRandom(gfInt degree)
{
    SetParams(degree);

    for (gfInt k = 0;k <= m_degree;k++)
    {
        for (gfInt l = 0;l <= m_degree - k;l++)
        {
            gfInt m = m_degree - k - l;

            SetWeight(k, l, m, 1.0);
            gfPoint3* pt = GetControlPoint(k, l, m);
            pt->Set(-0.5 + k / (gfDouble)m_degree + gfRand(-0.1, 0.1), -0.5 + l / (gfDouble)m_degree + gfRand(-0.1, 0.1), gfRand(-1.0, 1.0));
        }
    }

    return GF_S_OK;
}

vector<gfObjBezierTriangle*> gfObjBezierTriangle::Subdivide(gfDouble u, gfDouble v, gfDouble w)
{
    // do the deCasteljau algorithm
    std::vector<gfPoint4> aux_points((m_degree+1)*(m_degree+1)*(m_degree+1));
    // fill points with control points
    for (gfInt i = 0;i <= m_degree;i++)
    {
        for (gfInt j = 0;j <= m_degree-i;j++)
        {
            gfPoint3* pt = GetControlPoint(i, j, m_degree - i - j);
            gfDouble ww = GetWeight(i, j, m_degree - i - j);

            aux_points[i + (m_degree+1)*j + 0*(m_degree+1)*(m_degree+1)].Set(ww*pt->X(), ww*pt->Y(), ww*pt->Z(), ww);
        }
    }

    for (gfInt p = 1;p <= m_degree;p++)
    {
        // p is top index
        for (gfInt i = 0;i <= m_degree-p;i++)
        {
            for (gfInt j = 0;j <= m_degree-p-i;j++)
            {
                gfPoint4 v1 = aux_points[i+1+(m_degree+1)*j+(m_degree+1)*(m_degree+1)*(p-1)];
                gfPoint4 v2 = aux_points[i+(m_degree+1)*(j+1)+(m_degree+1)*(m_degree+1)*(p-1)];
                gfPoint4 v3 = aux_points[i+(m_degree+1)*j+(m_degree+1)*(m_degree+1)*(p-1)];
                aux_points[i+(m_degree+1)*j+(m_degree+1)*(m_degree+1)*p].Set(u*v1.X()+v*v2.X()+w*v3.X(), u*v1.Y()+v*v2.Y()+w*v3.Y(), u*v1.Z()+v*v2.Z()+w*v3.Z(), u*v1.W()+v*v2.W()+w*v3.W());
            }
        }
    }

    // have deCasteljau points, fill three new Bezier triangles
    gfObjBezierTriangle* t1 = (gfObjBezierTriangle*)gfCreateGeomObject(GF_BEZIER_TRIANGLE);
    gfObjBezierTriangle* t2 = (gfObjBezierTriangle*)gfCreateGeomObject(GF_BEZIER_TRIANGLE);
    gfObjBezierTriangle* t3 = (gfObjBezierTriangle*)gfCreateGeomObject(GF_BEZIER_TRIANGLE);
    t1->SetParams(m_degree);
    t2->SetParams(m_degree);
    t3->SetParams(m_degree);

    for (gfInt p = 0;p <= m_degree;p++)
    {
        // p is top index
        for (gfInt i = 0;i <= m_degree-p;i++)
        {
            for (gfInt j = 0;j <= m_degree-p-i;j++)
            {
                gfInt k = m_degree - i - j - p;
                gfPoint4 aux = aux_points[i+(m_degree+1)*j+(m_degree+1)*(m_degree+1)*p];
                gfDouble ww = aux.W();
                gfPoint3 pt;
                if (gfAbs(ww) < GF_EPS)
                    pt.Set(aux.X(), aux.Y(), aux.Z());
                else
                    pt.Set(aux.X() / ww, aux.Y() / ww, aux.Z() / ww);

                // first triangle
                if (i == 0)
                {
                    t1->SetWeight(p, j, k, ww);
                    t1->GetControlPoint(p, j, k)->Set(pt.X(), pt.Y(), pt.Z());
                }

                // second triangle
                if (j == 0)
                {
                    t2->SetWeight(i, p, k, ww);
                    t2->GetControlPoint(i, p, k)->Set(pt.X(), pt.Y(), pt.Z());
                }

                // third triangle
                if (k == 0)
                {
                    t3->SetWeight(i, j, p, ww);
                    t3->GetControlPoint(i, j, p)->Set(pt.X(), pt.Y(), pt.Z());
                }
            }
        }
    }

    std::vector<gfObjBezierTriangle*> result;

    if (gfAbs(u) < GF_EPS)
        gfDeleteGeomObject(t1);
    else
        result.push_back(t1);
    if (gfAbs(v) < GF_EPS)
        gfDeleteGeomObject(t2);
    else
        result.push_back(t2);
    if (gfAbs(w) < GF_EPS)
        gfDeleteGeomObject(t3);
    else
        result.push_back(t3);

    return result;
}

gfRayInterInfo* gfObjBezierTriangle::RayControlMeshIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    gfDouble t_min = GF_MAX_DOUBLE;
    gfVector3 normal;

    for (gfInt i = 1;i <= m_degree;i++)
    {
        for (gfInt j = 0;j <= m_degree - i;j++)
        {
            int k = m_degree - i - j;
            gfPoint3 v1 = *GetControlPoint(i, j, k);
            gfPoint3 v2 = *GetControlPoint(i-1, j+1, k);
            gfPoint3 v3 = *GetControlPoint(i-1, j, k+1);

            // one triangle intersection
            if (obj_matrix)
            {
                gfTransformPoint3(&v1, obj_matrix);
                gfTransformPoint3(&v2, obj_matrix);
                gfTransformPoint3(&v3, obj_matrix);
            }

            gfDouble u, v, t;
            GFRESULT res = gfRayTriangleInter(ray_begin, ray_dir, &v1, &v2, &v3, &u, &v, &t);

            if (GF_SUCC(res))
            {
                if (t > 0 && t < t_min)
                {
                    gfVector3 vec1(v2.X()-v1.X(), v2.Y()-v1.Y(), v2.Z()-v1.Z());
                    gfVector3 vec2(v3.X()-v1.X(), v3.Y()-v1.Y(), v3.Z()-v1.Z());
                    gfVectorProduct3(&vec1, &vec2, &normal);

                    // normalize
                    normal.Normalize();
                    t_min = t;
                }
            }

            if (j > 0)
            {
                gfPoint3 v4 = *GetControlPoint(i, j-1, k+1);
                if (obj_matrix)
                    gfTransformPoint3(&v4, obj_matrix);

                res = gfRayTriangleInter(ray_begin, ray_dir, &v1, &v3, &v4, &u, &v, &t);

                if (GF_SUCC(res))
                {
                    if (t > 0 && t < t_min)
                    {
                        gfVector3 vec2(v3.X()-v1.X(), v3.Y()-v1.Y(), v3.Z()-v1.Z());
                        gfVector3 vec3(v4.X()-v1.X(), v4.Y()-v1.Y(), v4.Z()-v1.Z());
                        gfVectorProduct3(&vec2, &vec3, &normal);

                        // normalize
                        normal.Normalize();
                        t_min = t;
                    }
                }
            }
        }
    }

    if (t_min < GF_MAX_DOUBLE)
    {
        gfRayInterInfo* info = gfCreateRayInterInfo();
        info->m_inters.push_back(t_min);
        info->m_normals.push_back(gfVector3(normal.X(), normal.Y(), normal.Z()));
        return info;
    }

    return NULL;
}

gfRayInterInfo* gfObjBezierTriangle::RayApproxMeshIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    // update cache
    //m_time_stamp++;
    OGLDraw(NULL);

    gfDouble t_min = GF_MAX_DOUBLE;
    gfVector3 normal;
    for (gfInt i = 1;i < GF_VIS_PREC;i++)
    {
        for (gfInt j = 0;j < GF_VIS_PREC - i - 1;j++)
        {
            gfPoint3 v1 = m_cached_points[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j];
            gfPoint3 v2 = m_cached_points[(GF_VIS_PREC) * (i - 1) - (i - 1) * (i - 2) / 2 + j + 1];
            gfPoint3 v3 = m_cached_points[(GF_VIS_PREC) * (i - 1) - (i - 1) * (i - 2) / 2 + j];

            // one triangle intersection
            if (obj_matrix)
            {
                gfTransformPoint3(&v1, obj_matrix);
                gfTransformPoint3(&v2, obj_matrix);
                gfTransformPoint3(&v3, obj_matrix);
            }

            gfDouble u, v, t;
            GFRESULT res = gfRayTriangleInter(ray_begin, ray_dir, &v1, &v2, &v3, &u, &v, &t);

            if (GF_SUCC(res))
            {
                if (t > 0 && t < t_min)
                {
                    gfVector3 vec1(v2.X()-v1.X(), v2.Y()-v1.Y(), v2.Z()-v1.Z());
                    gfVector3 vec2(v3.X()-v1.X(), v3.Y()-v1.Y(), v3.Z()-v1.Z());
                    gfVectorProduct3(&vec1, &vec2, &normal);

                    // normalize
                    normal.Normalize();
                    t_min = t;
                }
            }

            if (j > 0)
            {
                gfPoint3 v4 = m_cached_points[(GF_VIS_PREC) * i - (i) * (i - 1) / 2 + j - 1];
                if (obj_matrix)
                    gfTransformPoint3(&v4, obj_matrix);

                res = gfRayTriangleInter(ray_begin, ray_dir, &v1, &v3, &v4, &u, &v, &t);

                if (GF_SUCC(res))
                {
                    if (t > 0 && t < t_min)
                    {
                        gfVector3 vec2(v3.X()-v1.X(), v3.Y()-v1.Y(), v3.Z()-v1.Z());
                        gfVector3 vec3(v4.X()-v1.X(), v4.Y()-v1.Y(), v4.Z()-v1.Z());
                        gfVectorProduct3(&vec2, &vec3, &normal);

                        // normalize
                        normal.Normalize();
                        t_min = t;
                    }
                }
            }
        }
    }

    if (t_min < GF_MAX_DOUBLE)
    {
        gfRayInterInfo* info = gfCreateRayInterInfo();
        info->m_inters.push_back(t_min);
        info->m_normals.push_back(gfVector3(normal.X(), normal.Y(), normal.Z()));
        return info;
    }

    return NULL;
}


GFRESULT gfObjBezierTriangle::DeleteAABBTree(gfBezierTriangleAABBTreeNode* tree_root)
{
    if (!tree_root)
        return GF_S_OK;
    if (tree_root->aabb)
        gfDeleteBoundVolume(tree_root->aabb);

    if (tree_root->child1)
        DeleteAABBTree(tree_root->child1);
    if (tree_root->child2)
        DeleteAABBTree(tree_root->child2);
    if (tree_root->child3)
        DeleteAABBTree(tree_root->child3);

    if (tree_root->triangle)
        gfDeleteGeomObject(tree_root->triangle);

    delete tree_root;
    return GF_S_OK;
}

GFRESULT gfObjBezierTriangle::DeleteWholeAABBTree(void)
{
    gfInt num_trees = (gfInt)m_bezier_aabb_trees.size();
    for (gfInt i = 0;i < num_trees;i++)
    {
        gfBezierTriangleAABBTreeNode* root = m_bezier_aabb_trees.at(i);
        DeleteAABBTree(root->child1);
        DeleteAABBTree(root->child2);
        DeleteAABBTree(root->child3);
    }
    m_bezier_aabb_trees.clear();

    return GF_S_OK;
}

GFRESULT gfObjBezierTriangle::ComputeWholeAABBTree(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    DeleteWholeAABBTree();

    m_num_nodes = 0;

    gfBezierTriangleAABBTreeNode* root = CreateAABBTree(0, this);
    m_bezier_aabb_trees.push_back(root);

    gfLog("Computed AABB tree of Bezier patches for NURBS patch, nodes = %d, time = %d ms.\n", m_num_nodes, count.EndMSCount());

    return GF_S_OK;
}

gfBezierTriangleAABBTreeNode* gfObjBezierTriangle::CreateAABBTree(gfDouble depth, gfObjBezierTriangle* triangle)
{
    gfBezierTriangleAABBTreeNode* root = new gfBezierTriangleAABBTreeNode;
    root->child1 = root->child2 = root->child3 = NULL;
    root->aabb = (gfAuxAABB*) triangle->GetBoundVolume(GF_BOUND_VOL_AABB, NULL);
    root->triangle = NULL;
    m_num_nodes++;

    // compute also OBB
    gfAuxOBB* obb = (gfAuxOBB*) triangle->GetBoundVolume(GF_BOUND_VOL_OBB, NULL);

    // check if AABB needs to be splitted
    gfDouble x_dir = 2.0*gfSquareRoot(obb->m_axis1_x*obb->m_axis1_x+
        obb->m_axis1_y*obb->m_axis1_y+
        obb->m_axis1_z*obb->m_axis1_z);
    gfDouble y_dir = 2.0*gfSquareRoot(obb->m_axis2_x*obb->m_axis2_x+
        obb->m_axis2_y*obb->m_axis2_y+
        obb->m_axis2_z*obb->m_axis2_z);
    gfDouble z_dir = 2.0*gfSquareRoot(obb->m_axis3_x*obb->m_axis3_x+
        obb->m_axis3_y*obb->m_axis3_y+
        obb->m_axis3_z*obb->m_axis3_z);

    int degree = triangle->GetDegree();
    gfPoint3* pt1 = triangle->GetControlPoint(degree, 0, 0);
    gfPoint3* pt2 = triangle->GetControlPoint(0, degree, 0);
    gfPoint3* pt3 = triangle->GetControlPoint(0, 0, degree);
    gfDouble d12 = pt1->Distance(pt2);
    gfDouble d13 = pt1->Distance(pt3);
    gfDouble d23 = pt2->Distance(pt3);
    gfDouble dmax = d12;
    if (d13 > dmax)
        dmax = d13;
    if (d23 > dmax)
        dmax = d23;

    // find shortest dimension
    gfDouble min = 0.001;;
    gfDouble space = x_dir * y_dir * z_dir;

    // delete OBB
    gfDeleteBoundVolume(obb);

    // check if OBB is flat enough
    if (space < 0.000001 || depth > GF_MAX_BEZIER_AABB_TREE_DEPTH)
    {
        root->triangle = triangle;
        //root->control_mesh = patch->GetControlMesh();
        //root->control_mesh->ComputeSmoothNormals();
    }
    else
    {
        // split patch
        vector<gfObjBezierTriangle*> triangles;
        if (dmax == d12)
            triangles = triangle->Subdivide(0.5, 0.5, 0.0);
        else if (dmax == d13)
            triangles = triangle->Subdivide(0.5, 0.0, 0.5);
        else if (dmax == d23)
            triangles = triangle->Subdivide(0.0, 0.5, 0.5);
        else
            triangles = triangle->Subdivide(0.33333333, 0.33333333, 0.33333333);
        /*if ((int)depth % 3 == 0)
            triangles = triangle->Subdivide(0.5, 0.5, 0.0);
        else if ((int)depth % 3 == 1)
            triangles = triangle->Subdivide(0.5, 0.0, 0.5);
        else if ((int)depth % 3 == 2)
            triangles = triangle->Subdivide(0.0, 0.5, 0.5);*/


        if (triangles.size() > 0)
            root->child1 = CreateAABBTree(depth+1, triangles[0]);
        if (triangles.size() > 1)
            root->child2 = CreateAABBTree(depth+1, triangles[1]);
        if (triangles.size() > 2)
            root->child3 = CreateAABBTree(depth+1, triangles[2]);
    }

    return root;
}

gfRayInterInfo* gfObjBezierTriangle::RayAABBIntersection(gfBezierTriangleAABBTreeNode* root, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    if (!root)
        return NULL;
    gfDouble tmin, tmax;
    if (GF_FAIL(root->aabb->RayIntersection(ray_begin, ray_dir, &tmin, &tmax)))
        return NULL;

    gfRayInterInfo* info = NULL;
    if (root->child1 || root->child2 || root->child3)
    {
        // get intersection from childs
        gfRayInterInfo* info1 = RayAABBIntersection(root->child1, ray_begin, ray_dir);
        gfRayInterInfo* info2 = RayAABBIntersection(root->child2, ray_begin, ray_dir);
        gfRayInterInfo* info3 = RayAABBIntersection(root->child3, ray_begin, ray_dir);

        info = gfCreateRayInterInfo();
        if (info1)
        {
            for (gfInt i = 0;i < (gfInt)info1->m_inters.size();i++)
            {
                info->m_inters.push_back(info1->m_inters.at(i));
                info->m_normals.push_back(info1->m_normals.at(i));
                info->m_uvw.push_back(info1->m_uvw.at(i));
            }
            gfDeleteRayInterInfo(info1);
        }
        if (info2)
        {
            for (gfInt i = 0;i < (gfInt)info2->m_inters.size();i++)
            {
                info->m_inters.push_back(info2->m_inters.at(i));
                info->m_normals.push_back(info2->m_normals.at(i));
                info->m_uvw.push_back(info2->m_uvw.at(i));
            }
            gfDeleteRayInterInfo(info2);
        }
        if (info3)
        {
            for (gfInt i = 0;i < (gfInt)info3->m_inters.size();i++)
            {
                info->m_inters.push_back(info3->m_inters.at(i));
                info->m_normals.push_back(info3->m_normals.at(i));
                info->m_uvw.push_back(info3->m_uvw.at(i));
            }
            gfDeleteRayInterInfo(info3);
        }
    }
    else
    {
        if (!root->triangle || !root->aabb)
            return NULL;

        //if (!m_bezier_clipping)
        {
            // return intersection of ray and control mesh

            //info = root->control_mesh->RayIntersection(NULL, ray_begin, ray_dir);
            info = root->triangle->RayControlMeshIntersection(NULL, ray_begin, ray_dir);
            /*gfVector3* normal = root->triangle->GetNormal(0.3333333, 0.3333333, 0.3333333);
            if (info && info->m_normals.size() > 0)
                info->m_normals[0] = *normal;*/

            if (info)
            {
                // add dummy uvw coordinates
                for (gfInt i = 0;i < (gfInt)info->m_inters.size();i++)
                    info->m_uvw.push_back(gfPoint3(0.0, 0.0, 0.0));
            }
        }
        /*else
        {
            // use bezier clipping
            gfVector3 up;
            if (gfAbs(ray_dir->X()) > gfAbs(ray_dir->Y()) && gfAbs(ray_dir->X()) > gfAbs(ray_dir->Z()))
                up.Set(ray_dir->Y(), -ray_dir->X(), 0.0);
            else
                up.Set(0.0, ray_dir->Z(), -ray_dir->Y());
            gfVector3 right;
            gfVectorProduct3(ray_dir, &up, &right);
            right.Normalize();
            up.Normalize();
            gfDouble d1 = -(up.X()*ray_begin->X()+up.Y()*ray_begin->Y()+up.Z()*ray_begin->Z());
            gfDouble d2 = -(right.X()*ray_begin->X()+right.Y()*ray_begin->Y()+right.Z()*ray_begin->Z());

            gfRayBezierPatchInfo* res = root->patch->BezierClipping(NULL, 0, up.X(), up.Y(), up.Z(), d1, right.X(), right.Y(), right.Z(), d2);
            if (res)
            {
                gfInt size = (gfInt)res->m_inters.size();
                if (size > 0)
                    info = gfCreateRayInterInfo();
                for (gfInt i = 0;i < size;i++)
                {
                    gfPoint3* pt = root->patch->GetPatchPoint(res->m_inters.at(i).X(), res->m_inters.at(i).Y());
                    gfVector3* nm = root->patch->GetPatchNormal(res->m_inters.at(i).X(), res->m_inters.at(i).Y());
                    gfVector3 vec(pt->X()-ray_begin->X(), pt->Y()-ray_begin->Y(), pt->Z()-ray_begin->Z());

                    info->m_inters.push_back(gfScalarProduct3(ray_dir, &vec));
                    info->m_normals.push_back(gfVector3(nm->X(), nm->Y(), nm->Z()));
                    info->m_uvw.push_back(gfPoint3(res->m_inters.at(i).X(), res->m_inters.at(i).Y(), 0.0));

                    gfDeletePoint3(pt);
                    gfDeleteVector3(nm);
                }

                delete res;
            }
        }*/
    }

    // sort intersections
    if (info)
    {
        gfSortRayInterInfo(info);
    }

    return info;
}

