//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjBezierTriangle.h                                //
//   Description: Rational Bezier triangle class                //
//                                                              //
//////////////////////////////////////////////////////////////////


#ifndef _GF_C_BEZIER_TRIANGLE_H_
#define _GF_C_BEZIER_TRIANGLE_H_


#include "gfDefines.h"
#include <vector>
#include "gfPoint.h"
#include "gfVector.h"
#include "gfMatrix.h"
#include "gfObjGeomObject.h"
#include "gfObjNURBSCurve.h"
#include "gfObjHalfEdgeMesh.h"


class gfObjBezierTriangle;

struct gfBezierTriangleAABBTreeNode
{
    gfAuxAABB* aabb;
    gfBezierTriangleAABBTreeNode* child1;
    gfBezierTriangleAABBTreeNode* child2;
    gfBezierTriangleAABBTreeNode* child3;
    gfObjBezierTriangle* triangle;
};


class gfObjBezierTriangle : public gfObjGeomObject
{
private:

    std::vector<gfDouble> m_weights;
    std::vector<gfPoint3 *> m_control_points;
    gfInt m_degree;
    gfInt m_old_stamp;
    std::vector<gfPoint3> m_cached_points;
    std::vector<gfVector3> m_cached_normals;

    // auxiliary functions for ray intersection
    gfRayInterInfo* RayControlMeshIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfRayInterInfo* RayApproxMeshIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir);

    // Bezier AABB tree functions
    std::vector<gfBezierTriangleAABBTreeNode*> m_bezier_aabb_trees;
    gfRayInterInfo* RayAABBIntersection(gfBezierTriangleAABBTreeNode* root, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfBezierTriangleAABBTreeNode* CreateAABBTree(gfDouble depth, gfObjBezierTriangle* triangle);
    GFRESULT ComputeWholeAABBTree(void);
    GFRESULT DeleteWholeAABBTree(void);
    GFRESULT DeleteAABBTree(gfBezierTriangleAABBTreeNode* tree_root);
    gfInt m_num_nodes;

public:

    gfObjBezierTriangle(void);
    ~gfObjBezierTriangle(void);

    // gfObjGeomObject functions
    GFRESULT OGLDraw(gfLong flags);
    gfRayInterInfo* RayIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfGeomObjID GetID(void);
    GFRESULT Transform(gfMatrix4* matrix);
    gfObjGeomObject* Clone(void);
    gfAuxBoundingVolume* GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix);

    GFRESULT InitRandom(gfInt degree);

    // working with parameters of patch
    gfInt SetParams(gfInt degree);
    gfInt GetDegree(void);
    gfInt GetNumControlPoints(void);
    gfPoint3* GetControlPoint(gfInt u_index, gfInt v_index, gfInt w_index);
    gfDouble GetWeight(gfInt u_index, gfInt v_index, gfInt w_index);
    GFRESULT SetWeight(gfInt u_index, gfInt v_index, gfInt w_index, gfDouble weight);

    // modelling features
    gfVector3* GetNormal(gfDouble u, gfDouble v, gfDouble w);
    gfObjHalfEdgeMesh* GetControlMesh(void);
    gfObjHalfEdgeMesh* GetApproxMesh(gfInt approx);
    vector<gfObjBezierTriangle*> Subdivide(gfDouble u, gfDouble v, gfDouble w);
} ;


#endif