//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjHalfEdgeMesh.cpp                                //
//   Description: Implementation of mesh based on half edge     //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfObjHalfEdgeMesh.h"
#include "gfObjectFactory.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfAuxLogging.h"
#include "gfAuxItersectionUtils.h"
#include "gfWinFPSCounter.h"
#include "gfWinOGLExtensions.h"
#include "gfObjNURBSBox.h"

#include "./OGL/glut.h"


#define GF_NUM_AABB_ELEMENTS 50


std::list<gfInt>* gfObjHalfEdgeMesh::load_faces;
gfHEPoint** gfObjHalfEdgeMesh::load_vertices;
gfInt gfObjHalfEdgeMesh::load_num_faces;
gfInt gfObjHalfEdgeMesh::load_num_vertices;
gfDouble gfObjHalfEdgeMesh::minx;
gfDouble gfObjHalfEdgeMesh::maxx;
gfDouble gfObjHalfEdgeMesh::miny;
gfDouble gfObjHalfEdgeMesh::maxy;
gfDouble gfObjHalfEdgeMesh::minz;
gfDouble gfObjHalfEdgeMesh::maxz;


gfObjHalfEdgeMesh::gfObjHalfEdgeMesh(void)
{
    load_faces = NULL;
    load_vertices = NULL;
    load_num_faces = 0;
    load_num_vertices = 0;

    m_time_stamp = 0;
    m_aabb_tree_root = NULL;

    m_cache_vertices = NULL;
    m_cache_normals = NULL;
    m_cache_wire_vertices = NULL;
    m_num_cached_triangles = 0;
    m_num_cached_wire_lines = 0;
}

gfObjHalfEdgeMesh::~gfObjHalfEdgeMesh(void)
{
    Clear();
}

GFRESULT gfObjHalfEdgeMesh::OGLDraw(gfLong flags)
{
    // check if we are caching
    if (flags & GF_FLG_VERTEX_ARRAY_CACHE)
    {
        if (!m_cache_vertices)
            CreateVACache();
        return DrawVACache(flags);
    }
    DeleteVACache();

    //gfLog("Drawing half-edge mesh...\r");
    for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
    {
        //gfHEFace* face = *iter;
        gfHalfEdge* first_edge = face->edge;

        gfHalfEdge* edge = first_edge->next;

        if (flags & GF_FLG_SOLID)
        {
            // TODO: use tesellation
            glBegin(GL_POLYGON);
            if ((flags & GF_FLG_NORMALS))
                glNormal3f(first_edge->normal->x, first_edge->normal->y, first_edge->normal->z);

            //glColor3f(1.0 - abs(first_edge->end->curvature) * 2.0 * GF_PI, 0.0, 0.0);
            glVertex3f(first_edge->end->x, first_edge->end->y, first_edge->end->z);

            while (first_edge != edge)
            {
                if ((flags & GF_FLG_NORMALS))
                    glNormal3f(edge->normal->x, edge->normal->y, edge->normal->z);

                //glColor3f(1.0 - abs(first_edge->end->curvature) * 2.0 * GF_PI, 0.0, 0.0);
                glVertex3f(edge->end->x, edge->end->y, edge->end->z);
                edge = edge->next;
            }
            glEnd();
        }

        glNormal3f(0.0, 0.0, 0.0);

        if (flags & GF_FLG_WIRE)
        {
            glBegin(GL_LINE_LOOP);
            if ((flags & GF_FLG_NORMALS))
                glNormal3f(first_edge->normal->x, first_edge->normal->y, first_edge->normal->z);
            glVertex3f(first_edge->end->x, first_edge->end->y, first_edge->end->z);

            while (first_edge != edge)
            {
                if ((flags & GF_FLG_NORMALS))
                    glNormal3f(edge->normal->x, edge->normal->y, edge->normal->z);
                glVertex3f(edge->end->x, edge->end->y, edge->end->z);
                edge = edge->next;
            }
            glEnd();
        }
    }

    //gfLog("Drawing half-edge mesh...Done.\n");
    return GF_S_OK;
}

gfRayInterInfo* gfObjHalfEdgeMesh::RayIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    if (!m_aabb_tree_root)
        ComputeWholeAABBTree();

    // fill queue
    RayAABBIntersection(m_aabb_tree_root, ray_begin, ray_dir);

    //go throught queue and find exact intersection
    gfDouble t_min = GF_MAX_DOUBLE;
    gfVector3 normal;
    gfBool inter = false;

    for (; !m_aabb_queue.empty(); m_aabb_queue.pop())
    {
        gfAABBQueueUnit unit = m_aabb_queue.top();
        if (unit.tmin>t_min)
            continue;

        gfDouble t, u, v;
        gfPoint3 pt1, pt2, pt3;

        for (gfHEFace* face = unit.node->faces.GetFirst(); face; face = unit.node->faces.GetNext())
        {
            gfHalfEdge* first_edge = face->edge;
            gfHalfEdge* second_edge = first_edge->next;
            gfHalfEdge* edge = second_edge->next;

            while (first_edge != edge)
            {
                pt1.Set(first_edge->end->x, first_edge->end->y, first_edge->end->z);
                pt2.Set(second_edge->end->x, second_edge->end->y, second_edge->end->z);
                pt3.Set(edge->end->x, edge->end->y, edge->end->z);

                GFRESULT res = gfRayTriangleInter(ray_begin, ray_dir, &pt1, &pt2, &pt3, &u, &v, &t);

                if (GF_SUCC(res))
                {
                    if (t > 0 && t < t_min)
                    {
                        if (first_edge->normal && second_edge->normal && edge->normal)
                        {
                            normal.SetX(u*first_edge->normal->x+v*second_edge->normal->x+(1-u-v)*edge->normal->x);
                            normal.SetY(u*first_edge->normal->y+v*second_edge->normal->y+(1-u-v)*edge->normal->y);
                            normal.SetZ(u*first_edge->normal->z+v*second_edge->normal->z+(1-u-v)*edge->normal->z);

                            // normalize
                            normal.Normalize();
                        }
                        else
                        {
                            normal.Set(0.0, 0.0, 0.0);
                        }
                        t_min = t;
                        inter = true;
                    }
                }

                edge = edge->next;
                second_edge = second_edge->next;
            }
        }
    }

    if (inter)
    {
        gfRayInterInfo* info = gfCreateRayInterInfo();
        info->m_inters.push_back(t_min);
        info->m_normals.push_back(gfVector3(normal.X(), normal.Y(), normal.Z()));
        return info;
    }

    return NULL;
    //return RayAABBIntersection(m_aabb_tree_root, ray_begin, ray_dir);
}

gfGeomObjID gfObjHalfEdgeMesh::GetID(void)
{
    return GF_HALF_EDGE_MESH;
}

GFRESULT gfObjHalfEdgeMesh::Transform(gfMatrix4* matrix)
{
    if (!m_point_list.Empty())
    {
        for (gfHEPoint* point = m_point_list.GetFirst(); point; point = m_point_list.GetNext())
        {
            if (point)
            {
                gfPoint3 pt(point->x, point->y, point->z);
                gfTransformPoint3(&pt, matrix);
                point->x = pt.X();
                point->y = pt.Y();
                point->z = pt.Z();
            }
        }
    }

    // transform normals;
    if (!m_normal_list.Empty())
    {
        for (gfHENormal* normal = m_normal_list.GetFirst(); normal; normal = m_normal_list.GetNext())
        {
            if (normal)
            {
                gfVector3 vec(normal->x, normal->y, normal->z);
                gfTransformVector3(&vec, matrix);
                normal->x = vec.X();
                normal->y = vec.Y();
                normal->z = vec.Z();
            }
        }
    }

    // delete AABB tree, next time must be recomputed
    if (m_aabb_tree_root)
        DeleteAABBTree(m_aabb_tree_root);
    m_aabb_tree_root = NULL;

    // delete cache
    DeleteVACache();

    return GF_S_OK;
}

gfObjGeomObject* gfObjHalfEdgeMesh::Clone(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    gfObjGeomObject* clone = gfCreateGeomObject(GetID());
    gfObjHalfEdgeMesh* new_mesh = (gfObjHalfEdgeMesh*) clone;

    gfInt num_faces = (gfInt) m_face_list.size();
    gfInt num_points = (gfInt) m_point_list.size();
    gfInt num_edges = (gfInt) m_edge_list.size();

    // clone list of faces
    if (!m_face_list.Empty())
    {
        for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
        {
            //gfHEFace* face = *iter;
            gfHEFace* new_face = gfCreateHEFace();
            
            new_face->data = face;
            face->data = new_face;

            new_mesh->AddHEFace(new_face);
        }
    }
    
    // clone list of vertices
    if (!m_point_list.Empty())
    {
        for (gfHEPoint* point = m_point_list.GetFirst(); point; point = m_point_list.GetNext())
        {
            //gfHEPoint* point = *iter;
            gfHEPoint* new_point = gfCreateHEPoint();

            new_point->data = point;
            point->data = new_point;

            if (point)
            {
                new_point->x = point->x;
                new_point->y = point->y;
                new_point->z = point->z;
            }
            new_point->curvature = point->curvature;

            new_mesh->AddHEPoint(new_point);
        }
    }

    // clone list of normals
    if (!m_normal_list.Empty())
    {
        for (gfHENormal* normal = m_normal_list.GetFirst(); normal; normal = m_normal_list.GetNext())
        {
            gfHENormal* new_normal = gfCreateHENormal();

            new_normal->data = normal;
            normal->data = new_normal;

            if (normal)
            {
                new_normal->x = normal->x;
                new_normal->y = normal->y;
                new_normal->z = normal->z;
            }

            new_mesh->AddHENormal(new_normal);
        }
    }

    // clone list of texture coordinates
    if (!m_tex_coord_list.Empty())
    {
        for (gfHETexCoords* tex_coord = m_tex_coord_list.GetFirst(); tex_coord; tex_coord = m_tex_coord_list.GetNext())
        {
            gfHETexCoords* new_tex_coord = gfCreateHETexCoords();

            new_tex_coord->data = tex_coord;
            tex_coord->data = new_tex_coord;

            if (tex_coord)
            {
                new_tex_coord->u = tex_coord->u;
                new_tex_coord->v = tex_coord->v;
                new_tex_coord->w = tex_coord->w;
            }

            new_mesh->AddHETexCoords(tex_coord);
        }
    }

    // clone list of edges
    if (!m_edge_list.Empty())
    {
        for (gfHalfEdge* edge = m_edge_list.GetFirst(); edge; edge = m_edge_list.GetNext())
        {
            //gfHalfEdge* edge = *iter;
            gfHalfEdge* new_edge = gfCreateHalfEdge();

            edge->data = new_edge;
            new_edge->data = edge;

            new_mesh->AddHalfEdge(new_edge);
        }
    }

    // now clone topology
    if (!m_edge_list.Empty())
    {
        for (gfHalfEdge* edge = m_edge_list.GetFirst(); edge; edge = m_edge_list.GetNext())
        {
            //gfHalfEdge* edge = *iter;
            gfHalfEdge* new_edge = (gfHalfEdge*) edge->data;

            new_edge->end = (gfHEPoint*) edge->end->data;
            new_edge->normal = (gfHENormal*) edge->normal->data;
            new_edge->tex_coord = (gfHETexCoords*) edge->normal->data;

            new_edge->face = (gfHEFace*) edge->face->data;
            if (edge->opp)
                new_edge->opp = (gfHalfEdge*) edge->opp->data;
            new_edge->next = (gfHalfEdge*) edge->next->data;
            
            new_edge->end->edge = new_edge;
            new_edge->face->edge = new_edge;
        }
    }

    gfLog("Cloned half-edge mesh, time = %d ms.\n", count.EndMSCount());

    return clone;
}

GFRESULT gfObjHalfEdgeMesh::Clear(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    // clear faces
    if (!m_face_list.Empty())
    {
        for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
        {
            //gfHEFace* face = *iter;
            gfDeleteHEFace(face);
        }

        m_face_list.Clear();
    }
    
    // clear edges
    if (!m_edge_list.Empty())
    {
        for (gfHalfEdge* edge = m_edge_list.GetFirst(); edge; edge = m_edge_list.GetNext())
        {
            //gfHalfEdge* edge = *iter;
            gfDeleteHalfEdge(edge);
        }

        m_edge_list.Clear();
    }

    // clear vertices
    if (!m_point_list.Empty())
    {
        for (gfHEPoint* point = m_point_list.GetFirst(); point; point = m_point_list.GetNext())
        {
            //gfHEPoint* point = *iter;
            gfDeleteHEPoint(point);
        }

        m_point_list.Clear();
    }

    // clear normals
    if (!m_normal_list.Empty())
    {
        for (gfHENormal* normal = m_normal_list.GetFirst(); normal; normal = m_normal_list.GetNext())
        {
            gfDeleteHENormal(normal);
        }

        m_normal_list.Clear();
    }

    // clear texture coordinates
    if (!m_tex_coord_list.Empty())
    {
        for (gfHETexCoords* tex_coord = m_tex_coord_list.GetFirst(); tex_coord; tex_coord = m_tex_coord_list.GetNext())
        {
            gfDeleteHETexCoords(tex_coord);
        }

        m_tex_coord_list.Clear();
    }

    m_time_stamp++;

    // delete AABB tree
    if (m_aabb_tree_root)
        DeleteAABBTree(m_aabb_tree_root);
    m_aabb_tree_root = NULL;

    // clear vertex array cache
    DeleteVACache();

    //gfLog("Cleared half-edge mesh, time = %d ms.\n", count.EndMSCount());

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::AddHalfEdge(gfHalfEdge* edge)
{
    m_edge_list.PushBack(edge);

    m_time_stamp++;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::RemoveHalfEdge(gfHalfEdge* edge)
{
    m_edge_list.Remove(edge);

    m_time_stamp++;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::AddHEFace(gfHEFace* face)
{
    m_face_list.PushBack(face);

    m_time_stamp++;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::RemoveHEFace(gfHEFace* face)
{
    m_face_list.Remove(face);

    m_time_stamp++;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::AddHEPoint(gfHEPoint* point)
{
    m_point_list.PushBack(point);

    m_time_stamp++;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::RemoveHEPoint(gfHEPoint* point)
{
    m_point_list.Remove(point);

    m_time_stamp++;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::AddHENormal(gfHENormal* normal)
{
    m_normal_list.PushBack(normal);
    m_time_stamp++;
    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::RemoveHENormal(gfHENormal* normal)
{
    m_normal_list.Remove(normal);
    m_time_stamp++;
    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::AddHETexCoords(gfHETexCoords* tex_coords)
{
    m_tex_coord_list.PushBack(tex_coords);
    m_time_stamp++;
    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::RemoveHETexCoords(gfHETexCoords* tex_coords)
{
    m_tex_coord_list.Remove(tex_coords);
    m_time_stamp++;
    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::InitCube(gfPoint3 * center, gfDouble sizex, gfDouble sizey, gfDouble sizez)
{
    Clear();

    gfHEPoint* pt1 = gfCreateHEPoint();AddHEPoint(pt1);
    gfHEPoint* pt2 = gfCreateHEPoint();AddHEPoint(pt2);
    gfHEPoint* pt3 = gfCreateHEPoint();AddHEPoint(pt3);
    gfHEPoint* pt4 = gfCreateHEPoint();AddHEPoint(pt4);
    gfHEPoint* pt5 = gfCreateHEPoint();AddHEPoint(pt5);
    gfHEPoint* pt6 = gfCreateHEPoint();AddHEPoint(pt6);
    gfHEPoint* pt7 = gfCreateHEPoint();AddHEPoint(pt7);
    gfHEPoint* pt8 = gfCreateHEPoint();AddHEPoint(pt8);

    gfHEFace* fac1 = gfCreateHEFace();AddHEFace(fac1);
    gfHEFace* fac2 = gfCreateHEFace();AddHEFace(fac2);
    gfHEFace* fac3 = gfCreateHEFace();AddHEFace(fac3);
    gfHEFace* fac4 = gfCreateHEFace();AddHEFace(fac4);
    gfHEFace* fac5 = gfCreateHEFace();AddHEFace(fac5);
    gfHEFace* fac6 = gfCreateHEFace();AddHEFace(fac6);

    gfHalfEdge* edg1 = gfCreateHalfEdge();AddHalfEdge(edg1);
    gfHalfEdge* edg2 = gfCreateHalfEdge();AddHalfEdge(edg2);
    gfHalfEdge* edg3 = gfCreateHalfEdge();AddHalfEdge(edg3);
    gfHalfEdge* edg4 = gfCreateHalfEdge();AddHalfEdge(edg4);
    gfHalfEdge* edg5 = gfCreateHalfEdge();AddHalfEdge(edg5);
    gfHalfEdge* edg6 = gfCreateHalfEdge();AddHalfEdge(edg6);
    gfHalfEdge* edg7 = gfCreateHalfEdge();AddHalfEdge(edg7);
    gfHalfEdge* edg8 = gfCreateHalfEdge();AddHalfEdge(edg8);
    gfHalfEdge* edg9 = gfCreateHalfEdge();AddHalfEdge(edg9);
    gfHalfEdge* edg10 = gfCreateHalfEdge();AddHalfEdge(edg10);
    gfHalfEdge* edg11 = gfCreateHalfEdge();AddHalfEdge(edg11);
    gfHalfEdge* edg12 = gfCreateHalfEdge();AddHalfEdge(edg12);
    gfHalfEdge* edg13 = gfCreateHalfEdge();AddHalfEdge(edg13);
    gfHalfEdge* edg14 = gfCreateHalfEdge();AddHalfEdge(edg14);
    gfHalfEdge* edg15 = gfCreateHalfEdge();AddHalfEdge(edg15);
    gfHalfEdge* edg16 = gfCreateHalfEdge();AddHalfEdge(edg16);
    gfHalfEdge* edg17 = gfCreateHalfEdge();AddHalfEdge(edg17);
    gfHalfEdge* edg18 = gfCreateHalfEdge();AddHalfEdge(edg18);
    gfHalfEdge* edg19 = gfCreateHalfEdge();AddHalfEdge(edg19);
    gfHalfEdge* edg20 = gfCreateHalfEdge();AddHalfEdge(edg20);
    gfHalfEdge* edg21 = gfCreateHalfEdge();AddHalfEdge(edg21);
    gfHalfEdge* edg22 = gfCreateHalfEdge();AddHalfEdge(edg22);
    gfHalfEdge* edg23 = gfCreateHalfEdge();AddHalfEdge(edg23);
    gfHalfEdge* edg24 = gfCreateHalfEdge();AddHalfEdge(edg24);

    // fill edges
    edg1->end = pt2;edg1->face = fac2;edg1->opp = edg2;edg1->next = edg9;
    edg2->end = pt1;edg2->face = fac1;edg2->opp = edg1;edg2->next = edg8;

    edg3->end = pt3;edg3->face = fac3;edg3->opp = edg1;edg3->next = edg11;
    edg4->end = pt2;edg4->face = fac1;edg4->opp = edg1;edg4->next = edg2;

    edg5->end = pt4;edg5->face = fac4;edg5->opp = edg6;edg5->next = edg13;
    edg6->end = pt3;edg6->face = fac1;edg6->opp = edg5;edg6->next = edg4;

    edg7->end = pt1;edg7->face = fac5;edg7->opp = edg8;edg7->next = edg15;
    edg8->end = pt4;edg8->face = fac1;edg8->opp = edg7;edg8->next = edg6;

    edg9->end = pt6;edg9->face = fac2;edg9->opp = edg10;edg9->next = edg18;
    edg10->end = pt2;edg10->face = fac3;edg10->opp = edg9;edg10->next = edg3;

    edg11->end = pt7;edg11->face = fac3;edg11->opp = edg12;edg11->next = edg20;
    edg12->end = pt3;edg12->face = fac4;edg12->opp = edg11;edg12->next = edg5;

    edg13->end = pt8;edg13->face = fac4;edg13->opp = edg14;edg13->next = edg22;
    edg14->end = pt4;edg14->face = fac5;edg14->opp = edg13;edg14->next = edg7;

    edg15->end = pt5;edg15->face = fac5;edg15->opp = edg16;edg15->next = edg24;
    edg16->end = pt1;edg16->face = fac2;edg16->opp = edg15;edg16->next = edg1;

    edg17->end = pt6;edg17->face = fac6;edg17->opp = edg18;edg17->next = edg19;
    edg18->end = pt5;edg18->face = fac2;edg18->opp = edg17;edg18->next = edg16;

    edg19->end = pt7;edg19->face = fac6;edg19->opp = edg20;edg19->next = edg21;
    edg20->end = pt6;edg20->face = fac3;edg20->opp = edg19;edg20->next = edg10;

    edg21->end = pt8;edg21->face = fac6;edg21->opp = edg22;edg21->next = edg23;
    edg22->end = pt7;edg22->face = fac4;edg22->opp = edg21;edg22->next = edg12;

    edg23->end = pt5;edg23->face = fac6;edg23->opp = edg24;edg23->next = edg17;
    edg24->end = pt8;edg24->face = fac5;edg24->opp = edg23;edg24->next = edg14;

    // fill faces
    fac1->edge = edg2;
    fac2->edge = edg1;
    fac3->edge = edg3;
    fac4->edge = edg5;
    fac5->edge = edg7;
    fac6->edge = edg17;

    // fill vertices
    gfSetHEPointCoords(pt1, center->X() + sizex / 2, center->Y() - sizey / 2, center->Z() - sizez / 2);pt1->edge = edg7;
    gfSetHEPointCoords(pt2, center->X() + sizex / 2, center->Y() + sizey / 2, center->Z() - sizez / 2);pt2->edge = edg1;
    gfSetHEPointCoords(pt3, center->X() - sizex / 2, center->Y() + sizey / 2, center->Z() - sizez / 2);pt3->edge = edg3;
    gfSetHEPointCoords(pt4, center->X() - sizex / 2, center->Y() - sizey / 2, center->Z() - sizez / 2);pt4->edge = edg5;
    gfSetHEPointCoords(pt5, center->X() + sizex / 2, center->Y() - sizey / 2, center->Z() + sizez / 2);pt5->edge = edg15;
    gfSetHEPointCoords(pt6, center->X() + sizex / 2, center->Y() + sizey / 2, center->Z() + sizez / 2);pt6->edge = edg17;
    gfSetHEPointCoords(pt7, center->X() - sizex / 2, center->Y() + sizey / 2, center->Z() + sizez / 2);pt7->edge = edg19;
    gfSetHEPointCoords(pt8, center->X() - sizex / 2, center->Y() - sizey / 2, center->Z() + sizez / 2);pt8->edge = edg21;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::InitPlane(gfPoint3* begin, gfVector3* base1, gfVector3* base2, gfInt slices1, gfInt slices2)
{
    Clear();

    gfHalfEdge** edges = new gfHalfEdge*[4 * (slices1 - 1) * (slices2 - 1)];
    gfHEPoint** points = new gfHEPoint*[slices1 * slices2];

    // create vertices
    for (gfInt i = 0;i < slices1;i++)
    {
        for (gfInt j = 0;j < slices2;j++)
        {
            gfHEPoint* point = gfCreateHEPoint();

            point->x = begin->X() + base1->X() * i / (gfDouble)(slices1 - 1) + base2->X() * j / (gfDouble)(slices2 - 1);
            point->y = begin->Y() + base1->Y() * i / (gfDouble)(slices1 - 1) + base2->Y() * j / (gfDouble)(slices2 - 1);
            point->z = begin->Z() + base1->Z() * i / (gfDouble)(slices1 - 1) + base2->Z() * j / (gfDouble)(slices2 - 1);
            point->edge = NULL;
            points[i + slices1 * j] = point;
            AddHEPoint(point);
        }
    }

    // create faces and edges
    for (gfInt i = 0;i < slices1 - 1;i++)
    {
        for (gfInt j = 0;j < slices2 - 1;j++)
        {
            gfHEFace* face = gfCreateHEFace();
            gfHalfEdge* edge1 = gfCreateHalfEdge();
            gfHalfEdge* edge2 = gfCreateHalfEdge();
            gfHalfEdge* edge3 = gfCreateHalfEdge();
            gfHalfEdge* edge4 = gfCreateHalfEdge();

            edge1->end = points[i + 1 + slices1 * j];
            if (edge1->end->edge == NULL)
                edge1->end->edge = edge1;
            edge1->face = face;
            edge1->next = edge2;

            if (j > 0)
            {
                edges[4 * (i + (slices1 - 1) * (j - 1)) + 2]->opp = edge1;
                edge1->opp = edges[4 * (i + (slices1 - 1) * (j - 1)) + 2];
            }
            else
            {
                edge1->opp = NULL;
            }
            edges[4 * (i + (slices1 - 1) * j) + 0] = edge1;

            edge2->end = points[i + 1 + slices1 * (j + 1)];
            if (edge2->end->edge == NULL)
                edge2->end->edge = edge2;
            edge2->face = face;
            edge2->next = edge3;
            edge2->opp = NULL;
            edges[4 * (i + (slices1 - 1) * j) + 1] = edge2;

            edge3->end = points[i + slices1 * (j + 1)];
            if (edge3->end->edge == NULL)
                edge3->end->edge = edge3;
            edge3->face = face;
            edge3->next = edge4;
            edge3->opp = NULL;
            edges[4 * (i + (slices1 - 1) * j) + 2] = edge3;

            edge4->end = points[i + slices1 * j];
            if (edge4->end->edge == NULL)
                edge4->end->edge = edge4;
            edge4->face = face;
            edge4->next = edge1;
            if (i > 0)
            {
                edges[4 * (i - 1 + (slices1 - 1) * j) + 1]->opp = edge4;
                edge4->opp = edges[4 * (i - 1 + (slices1 - 1) * j) + 1];
            }
            else
            {
                edge4->opp = NULL;
            }
            edges[4 * (i + (slices1 - 1) * j) + 3] = edge4;
            
            face->edge = edge1;

            // add all to mesh
            AddHEFace(face);
            AddHalfEdge(edge1);
            AddHalfEdge(edge2);
            AddHalfEdge(edge3);
            AddHalfEdge(edge4);
        }
    }

    delete [] edges;
    delete [] points;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::InitSphere(gfPoint3* center, gfDouble radius, gfInt slices1, gfInt slices2)
{
    Clear();

    gfHalfEdge** edges = new gfHalfEdge*[4 * (slices1 - 1) * (slices2 - 1)];
    gfHEPoint** points = new gfHEPoint*[slices1 * slices2];

    // create vertices
    for (gfInt i = 0;i < slices1;i++)
    {
        for (gfInt j = 0;j < slices2;j++)
        {
            gfHEPoint* point = gfCreateHEPoint();

            gfDouble fi = 2 * GF_PI * i / (gfDouble)(slices1 - 1);
            gfDouble xi = - GF_PI / 2.0 + GF_PI * j / (gfDouble)(slices2 - 1);
            point->x = center->X() + radius*gfCos(fi)*gfCos(xi);
            point->y = center->Y() + radius*gfSin(fi)*gfCos(xi);
            point->z = center->Z() + radius*gfSin(xi);
            point->edge = NULL;
            points[i + slices1 * j] = point;
            AddHEPoint(point);
        }
    }

    // create faces and edges
    for (gfInt i = 0;i < slices1 - 1;i++)
    {
        for (gfInt j = 0;j < slices2 - 1;j++)
        {
            gfHEFace* face = gfCreateHEFace();
            gfHalfEdge* edge1 = gfCreateHalfEdge();
            gfHalfEdge* edge2 = gfCreateHalfEdge();
            gfHalfEdge* edge3 = gfCreateHalfEdge();
            gfHalfEdge* edge4 = gfCreateHalfEdge();

            edge1->end = points[i + 1 + slices1 * j];
            if (edge1->end->edge == NULL)
                edge1->end->edge = edge1;
            edge1->face = face;
            edge1->next = edge2;

            if (j > 0)
            {
                edges[4 * (i + (slices1 - 1) * (j - 1)) + 2]->opp = edge1;
                edge1->opp = edges[4 * (i + (slices1 - 1) * (j - 1)) + 2];
            }
            else
            {
                edge1->opp = NULL;
            }
            edges[4 * (i + (slices1 - 1) * j) + 0] = edge1;

            edge2->end = points[i + 1 + slices1 * (j + 1)];
            if (edge2->end->edge == NULL)
                edge2->end->edge = edge2;
            edge2->face = face;
            edge2->next = edge3;
            edge2->opp = NULL;
            edges[4 * (i + (slices1 - 1) * j) + 1] = edge2;

            edge3->end = points[i + slices1 * (j + 1)];
            if (edge3->end->edge == NULL)
                edge3->end->edge = edge3;
            edge3->face = face;
            edge3->next = edge4;
            edge3->opp = NULL;
            edges[4 * (i + (slices1 - 1) * j) + 2] = edge3;

            edge4->end = points[i + slices1 * j];
            if (edge4->end->edge == NULL)
                edge4->end->edge = edge4;
            edge4->face = face;
            edge4->next = edge1;
            if (i > 0)
            {
                edges[4 * (i - 1 + (slices1 - 1) * j) + 1]->opp = edge4;
                edge4->opp = edges[4 * (i - 1 + (slices1 - 1) * j) + 1];
            }
            else
            {
                edge4->opp = NULL;
            }
            edges[4 * (i + (slices1 - 1) * j) + 3] = edge4;
            
            face->edge = edge1;

            // add all to mesh
            AddHEFace(face);
            AddHalfEdge(edge1);
            AddHalfEdge(edge2);
            AddHalfEdge(edge3);
            AddHalfEdge(edge4);
        }
    }

    delete [] edges;
    delete [] points;

    return GF_S_OK;
}

int gfObjHalfEdgeMesh::vertex_cb(p_ply_argument argument) 
{
    long eol;
    ply_get_argument_user_data(argument, NULL, &eol);

    long inst;
    ply_get_argument_element(argument, NULL, &inst);
    p_ply_property property;
    ply_get_argument_property(argument, &property, NULL, NULL);
    char* ch;
    ply_get_property_info(property, (const char**) &ch, NULL, NULL, NULL);

    if (!strcmp(ch,"x"))
    {
        gfDouble x = ply_get_argument_value(argument);
        load_vertices[inst]->x = x;
        if (x < minx)
            minx = x;
        if (x > maxx)
            maxx = x;
    }
    if (!strcmp(ch,"y"))
    {
        gfDouble y = ply_get_argument_value(argument);
        load_vertices[inst]->y = y;
        if (y < miny)
            miny = y;
        if (y > maxy)
            maxy = y;
    }
    if (!strcmp(ch,"z"))
    {
        gfDouble z = ply_get_argument_value(argument);
        load_vertices[inst]->z = z;
        if (z < minz)
            minz = z;
        if (z > maxz)
            maxz = z;
    }

    return 1;
}

int gfObjHalfEdgeMesh::face_cb(p_ply_argument argument) 
{
    long length, value_index;
    long inst;
    ply_get_argument_element(argument, NULL, &inst);
    ply_get_argument_property(argument, NULL, &length, &value_index);

    /*switch (value_index) 
    {
        case 0:
        case 1:
            load_faces[inst].push_back(ply_get_argument_value(argument));
            break;
        case 2:
            load_faces[inst].push_back(ply_get_argument_value(argument));
            break;
        default: 
            break;
    }*/
    if (value_index >= 0 && value_index < length)
        load_faces[inst].push_back(ply_get_argument_value(argument));

    return 1;
}

GFRESULT gfObjHalfEdgeMesh::InitFromPly(std::string strFile)
{
    gfLog("Loading .ply file %s...\r", strFile.c_str());
    p_ply ply = ply_open(strFile.c_str(), NULL);
    if (!ply) 
    {
        gfLog("Cannot open ply file.\n");
        return GF_E_FAIL;
    }
    if (!ply_read_header(ply))
    {
        gfLog("Wrong header in ply file.\n");
        return GF_E_FAIL;
    }

    gfWinFPSCounter count;
    count.BeginMSCount();

    // clear previous content
    Clear();

    // vertices
    load_num_vertices = ply_set_read_cb(ply, "vertex", "x", vertex_cb, NULL, 0);
    ply_set_read_cb(ply, "vertex", "y", vertex_cb, NULL, 0);
    ply_set_read_cb(ply, "vertex", "z", vertex_cb, NULL, 1);

    // faces
    load_num_faces = ply_set_read_cb(ply, "face", "vertex_indices", face_cb, NULL, 0);

    // allocate auxiliary arrays
    load_faces = (std::list<gfInt>*) new list<gfInt>[load_num_faces];

    load_vertices = (gfHEPoint**) new gfHEPoint*[load_num_vertices];
    for (gfInt i = 0;i < load_num_vertices;i++)
    {
        gfHEPoint* point = gfCreateHEPoint();
        AddHEPoint(point);
        load_vertices[i] = point;
    }

    minx = miny = minz = GF_MAX_DOUBLE;
    maxx = maxy = maxz = GF_MIN_DOUBLE;

    ply_read(ply);
    ply_close(ply);

    // scale vertices so they fit in unit cube
    gfDouble s = maxx - minx;
    if (s < (maxy - miny))
        s = maxy - miny;
    if (s < (maxz - minz))
        s = maxz - minz;
    if (!gfIsZero(s))
    {
        for (gfInt i = 0;i < load_num_vertices;i++)
        {
            gfHEPoint* point = load_vertices[i];
            point->x = point->x / s - (maxx + minx) / 2.0 / s;
            point->y = point->y / s - (maxy + miny) / 2.0 / s;
            point->z = point->z / s - (maxz + minz) / 2.0 / s;
        }
    }

    // array of faces
    gfHEFace** created_faces = (gfHEFace**) new gfHEFace*[load_num_faces];
    for (gfInt i = 0;i < load_num_faces;i++)
    {
        gfHEFace* face = gfCreateHEFace();
        AddHEFace(face);
        created_faces[i] = face;
    }

    // for each vertex, find its neighbouring faces
    std::list<gfInt>* vert_neighs = (std::list<gfInt>*) new list<gfInt>[load_num_vertices];
    for (gfInt i = 0;i < load_num_faces;i++)
    {
        gfHalfEdge* first_edge = NULL;
        gfHalfEdge* prev_edge = NULL;
        for (std::list<gfInt>::iterator iter = load_faces[i].begin(); iter != load_faces[i].end();iter++)
        {
            gfInt point1 = *iter;
            vert_neighs[point1].push_back(i);
        }
    }

    // go throught all faces and fill edges
    for (gfInt i = 0;i < load_num_faces;i++)
    {
        //gfLog("Converting .ply file....%.0f%%\r", 100 * gfDouble(i) / gfDouble(load_num_faces));

        gfHalfEdge* first_edge = NULL;
        gfHalfEdge* prev_edge = NULL;
        for (std::list<gfInt>::iterator iter = load_faces[i].begin(); iter != load_faces[i].end();iter++)
        {
            gfInt point1 = *iter;
            gfInt point2;
            iter++;
            if (iter == load_faces[i].end())
            {
                point2 = *(load_faces[i].begin());
            }
            else
            {
                point2 = *(iter);
            }
            iter--;

            // find faces with (point1, point2) edge
            gfInt neigh_face = -1;
            gfInt opp_orient;
            for (std::list<gfInt>::iterator iter_neigh = vert_neighs[point1].begin(); iter_neigh != vert_neighs[point1].end();iter_neigh++)
            {
                gfInt j = *iter_neigh;
                if (i == j)
                    continue;

                for (std::list<gfInt>::iterator iter2 = load_faces[j].begin(); iter2 != load_faces[j].end();iter2++)
                {
                    gfInt aux_point1 = *iter2;
                    gfInt aux_point2;
                    iter2++;
                    if (iter2 == load_faces[j].end())
                    {
                        aux_point2 = *(load_faces[j].begin());
                    }
                    else
                    {
                        aux_point2 = *(iter2);
                    }
                    iter2--;

                    // compare edges (point1,point2) and (aux_point1,aux_point2)
                    if ((point1 == aux_point1) && (point2 == aux_point2))
                    {
                        neigh_face = j;
                        opp_orient = 1;
                        break;
                    }
                    if ((point1 == aux_point2) && (point2 == aux_point1))
                    {
                        neigh_face = j;
                        opp_orient = 2;
                        break;
                    }
                }

                if (neigh_face != -1)
                    break;
            }

            // create edge 
            gfHalfEdge* curr_edge = gfCreateHalfEdge();
            // set end vertex
            curr_edge->end = load_vertices[point2];
            // set adjanced face
            curr_edge->face = created_faces[i];
            // set opposite edge from neigh
            if (neigh_face > -1 && neigh_face < i)
            {
                gfHalfEdge* beg_edge = created_faces[neigh_face]->edge;
                gfHalfEdge* next_edge = beg_edge->next;

                if (beg_edge->end == load_vertices[point1] && opp_orient == 2)
                {
                    curr_edge->opp = beg_edge;
                    beg_edge->opp = curr_edge;
                }
                else if (beg_edge->end == load_vertices[point2] && opp_orient == 1)
                {
                    curr_edge->opp = beg_edge;
                    beg_edge->opp = curr_edge;
                }
                else
                {
                    while (next_edge != beg_edge)
                    {
                        if (next_edge->end == load_vertices[point1] && opp_orient == 2)
                        {
                            curr_edge->opp = next_edge;
                            next_edge->opp = curr_edge;
                            break;
                        }
                        else if (next_edge->end == load_vertices[point2] && opp_orient == 1)
                        {
                            curr_edge->opp = next_edge;
                            next_edge->opp = curr_edge;
                            break;
                        }
                        next_edge = next_edge->next;
                    }
                }
            }
            else
            {
                int ii = 1;
            }

            if (iter == load_faces[i].begin())
                first_edge = curr_edge;
            if (prev_edge)
                prev_edge->next = curr_edge;
            prev_edge = curr_edge;

            if (load_vertices[point2]->edge == NULL)
                load_vertices[point2]->edge = curr_edge;
            if (curr_edge->opp)
                load_vertices[point1]->edge = curr_edge->opp;

            // add half edge to mesh
            AddHalfEdge(curr_edge);
        }

        prev_edge->next = first_edge;
        created_faces[i]->edge = first_edge;
    }


    // free auxiliary arrays
    for (gfInt i = 0;i < load_num_faces;i++)
        load_faces[i].clear();
    for (gfInt i = 0;i < load_num_vertices;i++)
        vert_neighs[i].clear();
    load_num_vertices = 0;
    load_num_faces = 0;
    delete [] load_faces;
    load_faces = NULL;
    delete [] load_vertices;
    load_vertices = NULL;
    delete [] created_faces;
    delete [] vert_neighs;

    gfLog("Loaded & converted half-edge mesh, vertices = %d, faces = %d,time = %d ms.\n", GetNumVertices(), GetNumFaces(), count.EndMSCount());

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::ComputeSmoothNormals(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    m_normal_list.Clear();

    if (!m_face_list.Empty() && !m_point_list.Empty())
    {
        for (gfHEPoint* point = m_point_list.GetFirst(); point; point = m_point_list.GetNext())
        {
            if (point)
                point->data = NULL;
        }

        for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
        {
            gfHalfEdge* edge = face->edge;
            if (!edge)
                return GF_E_FAIL;
            gfHalfEdge* next_edge = edge;

            gfHENormal* n = gfGetHEFaceNormal(face);

            // find all vertices of face
            gfBool first = true;
            while (next_edge != edge || first)
            {   
                first = false;

                if (next_edge->end->data)
                {
                    gfHENormal* normal = (gfHENormal*) next_edge->end->data;
                    normal->x += n->x; 
                    normal->y += n->y;
                    normal->z += n->z;
                }
                else
                {
                    gfHENormal* normal = gfCreateHENormal();
                    normal->x = n->x; 
                    normal->y = n->y;
                    normal->z = n->z;
                    next_edge->end->data = (void*) normal;
                }

                next_edge->normal = (gfHENormal*) next_edge->end->data;
                next_edge = next_edge->next;
            }

            gfDeleteHENormal(n);
        }

        // normalize all computed normals
        for (gfHEPoint* point = m_point_list.GetFirst(); point; point = m_point_list.GetNext())
        {
            if (point->data)
            {
                gfHENormal* normal = (gfHENormal*) point->data;
                gfHENormalNormalize(normal);
                AddHENormal(normal);
                point->data = NULL;
            }
        }

        DeleteVACache();
    }

    //gfLog("Computed vertex normals for half-edge mesh, time = %d ms.\n", count.EndMSCount());

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::ComputeFacesNormals(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    m_normal_list.Clear();

    if (!m_face_list.Empty() && !m_point_list.Empty())
    {
        for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
        {
            //gfLog("Computing mesh normals....%.0f%%\r", 100 * gfDouble(i) / gfDouble(num_count));
            //gfHEFace* face = *iter;

            gfHalfEdge* edge = face->edge;
            if (!edge)
                return GF_E_FAIL;
            gfHalfEdge* next_edge = edge;

            // get face normal
            gfHENormal* n = gfGetHEFaceNormal(face);
            AddHENormal(n);

            // find all vertices of face
            gfBool first = true;
            while (next_edge != edge || first)
            {   
                first = false;
                next_edge->normal = n;
                next_edge = next_edge->next;
            }

        }

        DeleteVACache();
        gfLog("Computed face normals for half-edge mesh, time = %d ms.\n", count.EndMSCount());
    }

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::ComputeGaussianCurvature(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    if (m_face_list.Empty() || m_point_list.Empty())
        return GF_E_FAIL;

    // check all vertices of mesh and empty curvature
    for (gfHEPoint* point = m_point_list.GetFirst(); point; point = m_point_list.GetNext())
    {
        //gfHEPoint* point = *iter;
        point->curvature = 2.0 * GF_PI;
    }

    gfInt num_faces = (gfInt) m_face_list.size(); 
    gfInt i = 0;

    gfVector3* v1 = gfCreateVector3();
    gfVector3* v2 = gfCreateVector3();

    for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
    {
        //gfLog("Computing Gaussian curvature....%.0f%%\r", 100 * gfDouble(i) / gfDouble(num_faces));
        //gfHEFace* face = *iter;

        gfHalfEdge* prev_edge = face->edge;
        gfHalfEdge* edge = prev_edge->next;
        gfHalfEdge* first_edge = prev_edge;
        if (!edge)
            return GF_E_FAIL;
        gfBool first = true;

        while(prev_edge != first_edge || first)
        {
            first = false;

            v1->Set(prev_edge->end->x - edge->end->x, prev_edge->end->y - edge->end->y, prev_edge->end->z - edge->end->z);
            v2->Set(edge->next->end->x - edge->end->x, edge->next->end->y - edge->end->y, edge->next->end->z - edge->end->z);

            // compute angle of v1 and v2
            gfDouble alfa = gfVectorsAngle3(v1, v2);
            
            edge->end->curvature = edge->end->curvature - alfa;

            prev_edge = edge;
            edge = edge->next;
        }

        i++;
    }

    gfDeleteVector3(v1);
    gfDeleteVector3(v2);

    gfLog("Computed Gaussian curvature for half-edge mesh, time = %d ms.\n", count.EndMSCount());

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::ReverseNormals(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    if (m_normal_list.Empty())
        return GF_S_OK;

    // reverse each normal
    for (gfHENormal* normal = m_normal_list.GetFirst(); normal; normal = m_normal_list.GetNext())
    {
        normal->x *= -1;
        normal->y *= -1;
        normal->z *= -1;
    }

    //TODO: reverse also each face

    DeleteVACache();
    gfLog("Reversed normals and faces in half-edge mesh, time = %d ms.\n", count.EndMSCount());

    return GF_S_OK;
}

gfInt gfObjHalfEdgeMesh::GetNumVertices()
{
    if (m_point_list.Empty())
        return 0;
    else
        return (gfInt) m_point_list.size();
}

gfInt gfObjHalfEdgeMesh::GetNumFaces()
{
    if (m_face_list.Empty())
        return 0;
    else
        return (gfInt) m_face_list.size();
}

gfInt gfObjHalfEdgeMesh::GetNumHalfEdges()
{
    if (m_edge_list.Empty())
        return 0;
    else
        return (gfInt) m_edge_list.size();
}

void gfObjHalfEdgeMesh::PrintStatus(void)
{
    gfLog("================== Half-Edge Mesh Status ================================\n");
    gfLog("Vertices:  %d\n", GetNumVertices());
    gfLog("Half edges:  %d\n", GetNumHalfEdges());
    gfLog("Faces:  %d\n", GetNumFaces());
    gfLog("Memory used:  %d\n", GetNumVertices() * sizeof(gfHEPoint) + 
                                GetNumFaces() * sizeof(gfHEFace) + 
                                GetNumHalfEdges() * sizeof(gfHalfEdge) +
                                m_point_list.GetAllocMemorySize() +
                                m_face_list.GetAllocMemorySize() +
                                m_edge_list.GetAllocMemorySize());
    gfLog("=========================================================================\n");
}

gfObjHalfEdgeMesh* gfObjHalfEdgeMesh::CatmullClarkSubdivision(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    gfObjGeomObject* clone = gfCreateGeomObject(GetID());
    gfObjHalfEdgeMesh* new_mesh = (gfObjHalfEdgeMesh*) clone;

    // get number of elements
    gfInt num_faces = (gfInt) m_face_list.size();
    gfInt num_points = (gfInt) m_point_list.size();
    gfInt num_edges = (gfInt) m_edge_list.size();

    gfInt counter_end = num_faces + num_points + num_edges + num_edges + num_faces + num_edges;
    gfInt counter = 0;

    // create face points
    if (!m_face_list.Empty())
    {
        for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
        {
            //gfHEFace* face = *iter;
            gfHEPoint* face_point = gfGetCCFacePoint(face);
            face_point->data = face;
            face->data = face_point;
            new_mesh->AddHEPoint(face_point);

            //gfLog("Computing Catmull-Clark subdivision.......%.0f%%\r", 100 * gfDouble(counter) / gfDouble(counter_end));
            counter++;
        }
    }

    //gfLog("CC-created face points.\n");

    // move old points
    if (!m_point_list.Empty())
    {
        for (gfHEPoint* point = m_point_list.GetFirst(); point; point = m_point_list.GetNext())
        {
            //gfHEPoint* point = *iter;
            gfHEPoint* moved_point = gfGetCCMovedPoint(point);
            moved_point->data = point;
            point->data = moved_point;
            new_mesh->AddHEPoint(moved_point);

            //gfLog("Computing Catmull-Clark subdivision.......%.0f%%\r", 100 * gfDouble(counter) / gfDouble(counter_end));
            counter++;
        }
    }

    //gfLog("CC-moved old points.\n");

    // create edge points, fill also part of topology
    if (!m_edge_list.Empty())
    {
        for (gfHalfEdge* edge = m_edge_list.GetFirst(); edge; edge = m_edge_list.GetNext())
        {
            //gfHalfEdge* edge = *iter;
            edge->data = NULL;

            //gfLog("Computing Catmull-Clark subdivision.......%.0f%%\r", 100 * gfDouble(counter) / gfDouble(counter_end));
            counter++;
        }
    }
    if (!m_edge_list.Empty())
    {
        for (gfHalfEdge* edge = m_edge_list.GetFirst(); edge; edge = m_edge_list.GetNext())
        {
            //gfHalfEdge* edge = *iter;
            gfAuxExtHE* aux_edge = new gfAuxExtHE;

            // fill auxiliary data
            if (edge->opp)
            {
                if (edge->opp->data)
                {
                    gfAuxExtHE* aux_opp_edge = (gfAuxExtHE*) edge->opp->data;
                    aux_edge->edge_point = aux_opp_edge->edge_point;
                }
                else
                {
                    aux_edge->edge_point = gfGetCCEdgePoint(edge);
                    new_mesh->AddHEPoint(aux_edge->edge_point);
                }
            }
            else
            {
                aux_edge->edge_point = gfGetCCEdgePoint(edge);
                new_mesh->AddHEPoint(aux_edge->edge_point);
            }
            
            aux_edge->new_edge_1 = gfCreateHalfEdge();
            aux_edge->new_edge_2 = gfCreateHalfEdge();
            aux_edge->new_edge_3 = gfCreateHalfEdge();
            aux_edge->new_edge_4 = gfCreateHalfEdge();

            // add this to new mesh
            new_mesh->AddHalfEdge(aux_edge->new_edge_1);
            new_mesh->AddHalfEdge(aux_edge->new_edge_2);
            new_mesh->AddHalfEdge(aux_edge->new_edge_3);
            new_mesh->AddHalfEdge(aux_edge->new_edge_4);

            // connect it with old halfedge
            aux_edge->old_edge = edge;
            edge->data = aux_edge;

            // topology
            aux_edge->new_edge_1->next = aux_edge->new_edge_2;
            aux_edge->new_edge_2->opp = aux_edge->new_edge_3;
            aux_edge->new_edge_3->next = aux_edge->new_edge_4;
            aux_edge->new_edge_3->opp = aux_edge->new_edge_2;

            aux_edge->new_edge_1->end = aux_edge->edge_point;
            aux_edge->new_edge_2->end = (gfHEPoint*) edge->face->data;
            aux_edge->new_edge_3->end = aux_edge->edge_point;
            aux_edge->new_edge_4->end = (gfHEPoint*) edge->end->data;

            aux_edge->edge_point->edge = aux_edge->new_edge_1;
            aux_edge->new_edge_2->end->edge = aux_edge->new_edge_2;
            aux_edge->new_edge_4->end->edge = aux_edge->new_edge_4;

            //gfLog("Computing Catmull-Clark subdivision.......%.0f%%\r", 100 * gfDouble(counter) / gfDouble(counter_end));
            counter++;
        }
    }

    //gfLog("CC-created auxiliary data.\n");

    // finish creating of topology, create faces
    if (!m_face_list.Empty())
    {
        for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
        {
            //gfHEFace* face = *iter;
            // go throught all half-edges of face

            gfHalfEdge* first_edge = face->edge;
            gfHalfEdge* next_edge = first_edge;

            gfBool first = true;
            while (first_edge != next_edge || first)
            {
                first = false;
                gfAuxExtHE* aux_edge = (gfAuxExtHE*) next_edge->data;
                gfHEFace* new_face = gfCreateHEFace();
                new_mesh->AddHEFace(new_face);

                // get next auxiliary edge
                gfAuxExtHE* next_aux_edge = (gfAuxExtHE*) next_edge->next->data;

                new_face->edge = aux_edge->new_edge_4;

                aux_edge->new_edge_3->face = new_face;
                aux_edge->new_edge_4->face = new_face;
                next_aux_edge->new_edge_1->face = new_face;
                next_aux_edge->new_edge_2->face = new_face;

                aux_edge->new_edge_4->next = next_aux_edge->new_edge_1;
                next_aux_edge->new_edge_2->next = aux_edge->new_edge_3;

                if (next_edge->opp)
                {
                    gfAuxExtHE* opp_aux_edge = (gfAuxExtHE*) next_edge->opp->data;
                    aux_edge->new_edge_1->opp = opp_aux_edge->new_edge_4;
                    aux_edge->new_edge_4->opp = opp_aux_edge->new_edge_1;
                }

                next_edge = next_edge->next;
            }

            //gfLog("Computing Catmull-Clark subdivision.......%.0f%%\r", 100 * gfDouble(counter) / gfDouble(counter_end));
            counter++;
        }
    }

    //gfLog("CC-finished topology.\n");

    // clear auxiliary edges
    if (!m_edge_list.Empty())
    {
        for (gfHalfEdge* edge = m_edge_list.GetFirst(); edge; edge = m_edge_list.GetNext())
        {
            //gfHalfEdge* edge = *iter;
            gfAuxExtHE* aux_edge = (gfAuxExtHE*) edge->data;
            delete aux_edge;
            edge->data = NULL;

            //gfLog("Computing Catmull-Clark subdivision.......%.0f%%\r", 100 * gfDouble(counter) / gfDouble(counter_end));
            counter++;
        }
    }

    //gfLog("CC-cleared auxiliary data.\n");

    gfLog("Computed Catmull-Clark subdivision for half-edge mesh, time = %d ms.\n", count.EndMSCount());

    return new_mesh;
}

gfAuxBoundingVolume* gfObjHalfEdgeMesh::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    if (m_point_list.Empty())
        return NULL;

    // create auxiliary array of mesh points
    gfInt num_points = m_point_list.size();
    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfHEPoint* point = m_point_list.GetFirst(); point; point = m_point_list.GetNext())
    {
        curr_point->Set(point->x, point->y, point->z);
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    delete [] points;
    return result;
}

GFRESULT gfObjHalfEdgeMesh::DeleteAABBTree(gfAABBTreeNode* tree_root)
{
    if (tree_root->aabb)
        gfDeleteBoundVolume(tree_root->aabb);
    tree_root->faces.Clear();
    if (tree_root->left)
        DeleteAABBTree(tree_root->left);
    if (tree_root->right)
        DeleteAABBTree(tree_root->right);
    delete tree_root;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::ComputeWholeAABBTree(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    if (m_aabb_tree_root)
        DeleteAABBTree(m_aabb_tree_root);

    load_num_faces = 0;
    m_aabb_tree_root = CreateAABBTree(0, &m_face_list);
    gfLog("Computed AABB tree for half-edge mesh, nodes = %d, time = %d ms.\n", load_num_faces, count.EndMSCount());

    return GF_S_OK;
}

gfAABBTreeNode* gfObjHalfEdgeMesh::CreateAABBTree(gfInt depth, gfAuxList<gfHEFace*>* faces)
{
    load_num_faces++;
    gfInt num_faces = faces->size();
    if (num_faces == 0)
        return NULL;

    gfAABBTreeNode* root = new gfAABBTreeNode;
    root->faces.Clear();
    root->left = root->right = NULL;
    root->aabb = (gfAuxAABB*) gfCreateBoundVolume(GF_BOUND_VOL_AABB);
    root->aabb->m_min_x = root->aabb->m_min_y = root->aabb->m_min_z = GF_MAX_DOUBLE;
    root->aabb->m_max_x = root->aabb->m_max_y = root->aabb->m_max_z = GF_MIN_DOUBLE;

    // go throught all faces, compute AABB
    for (gfHEFace* face = faces->GetFirst(); face; face = faces->GetNext())
    {
        gfHalfEdge* first_edge = face->edge;
        gfHalfEdge* edge = first_edge->next;

        if (first_edge->end->x < root->aabb->m_min_x)
            root->aabb->m_min_x = first_edge->end->x;
        if (first_edge->end->y < root->aabb->m_min_y)
            root->aabb->m_min_y = first_edge->end->y;
        if (first_edge->end->z < root->aabb->m_min_z)
            root->aabb->m_min_z = first_edge->end->z;

        if (first_edge->end->x > root->aabb->m_max_x)
            root->aabb->m_max_x = first_edge->end->x;
        if (first_edge->end->y > root->aabb->m_max_y)
            root->aabb->m_max_y = first_edge->end->y;
        if (first_edge->end->z > root->aabb->m_max_z)
            root->aabb->m_max_z = first_edge->end->z;

        while (first_edge != edge)
        {
            if (edge->end->x < root->aabb->m_min_x)
                root->aabb->m_min_x = edge->end->x;
            if (edge->end->y < root->aabb->m_min_y)
                root->aabb->m_min_y = edge->end->y;
            if (edge->end->z < root->aabb->m_min_z)
                root->aabb->m_min_z = edge->end->z;

            if (edge->end->x > root->aabb->m_max_x)
                root->aabb->m_max_x = edge->end->x;
            if (edge->end->y > root->aabb->m_max_y)
                root->aabb->m_max_y = edge->end->y;
            if (edge->end->z > root->aabb->m_max_z)
                root->aabb->m_max_z = edge->end->z;

            edge = edge->next;
        }
    }

    // check if AABB needs to be splitted
    gfDouble x_dir = root->aabb->m_max_x - root->aabb->m_min_x;
    gfDouble y_dir = root->aabb->m_max_y - root->aabb->m_min_y;
    gfDouble z_dir = root->aabb->m_max_z - root->aabb->m_min_z;
    if (num_faces < GF_NUM_AABB_ELEMENTS || depth > GF_MAX_MESH_AABB_TREE_DEPTH)
    {
        for (gfHEFace* face = faces->GetFirst(); face; face = faces->GetNext())
        {
            root->faces.PushBack(face);
        }
        return root;
    }

    // cut volume in half of longest direction
    gfInt longest;
    if (x_dir > y_dir)
    {
        if (x_dir > z_dir)
            longest = 1;
        else
            longest = 3;
    }
    else
    {
        if (y_dir > z_dir)
            longest = 2;
        else
            longest = 3;
    }

    gfAuxList<gfHEFace*> faces1;
    gfAuxList<gfHEFace*> faces2;
    gfDouble bx, by, bz;
    gfInt n;
    for (gfHEFace* face = faces->GetFirst(); face; face = faces->GetNext())
    {
        // compute barycenter of face
        gfHalfEdge* first_edge = face->edge;
        gfHalfEdge* edge = first_edge->next;

        bx = first_edge->end->x;
        by = first_edge->end->y;
        bz = first_edge->end->z;
        n = 1;

        while (first_edge != edge)
        {
            bx += edge->end->x;
            by += edge->end->y;
            bz += edge->end->z;
            n++;
            edge = edge->next;
        }
        bx /= n;by /= n;bz /= n;

        if (longest == 1)
        {
            if (2*bx < (root->aabb->m_max_x + root->aabb->m_min_x))
                faces1.PushBack(face);
            else
                faces2.PushBack(face);
        }
        if (longest == 2)
        {
            if (2*by < (root->aabb->m_max_y + root->aabb->m_min_y))
                faces1.PushBack(face);
            else
                faces2.PushBack(face);
        }
        if (longest == 3)
        {
            if (2*bz < (root->aabb->m_max_z + root->aabb->m_min_z))
                faces1.PushBack(face);
            else
                faces2.PushBack(face);
        }
    }

    root->left = CreateAABBTree(depth+1, &faces1);
    root->right = CreateAABBTree(depth+1, &faces2);
    return root;
}

gfRayInterInfo* gfObjHalfEdgeMesh::RayAABBIntersection(gfAABBTreeNode* root, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    if (!root)
        return NULL;
    gfDouble tmin, tmax;
    if (GF_FAIL(root->aabb->RayIntersection(ray_begin, ray_dir, &tmin, &tmax)))
        return NULL;

    if (root->left && root->right)
    {
        RayAABBIntersection(root->left, ray_begin, ray_dir);
        RayAABBIntersection(root->right, ray_begin, ray_dir);
    }
    else
    {
        gfAABBQueueUnit unit;
        unit.tmin = tmin;
        unit.tmax = tmax;
        unit.node = root;
        m_aabb_queue.push(unit);
    }

    return NULL;

    /*gfInt num_faces = root->faces.size();
    gfRayInterInfo* info = NULL;
    if (num_faces == 0 && root->left && root->right)
    {
        // get intersection from childs
        gfRayInterInfo* info1 = RayAABBIntersection(root->left, ray_begin, ray_dir);
        gfRayInterInfo* info2 = RayAABBIntersection(root->right, ray_begin, ray_dir);
        if (info1 && info2)
        {
            if (info1->m_inters.at(0) < info2->m_inters.at(0))
            {
                info = info1;
                for (gfInt i = 0;i < info2->m_inters.size();i++)
                    info->m_inters.push_back(info2->m_inters.at(i));
                gfDeleteRayInterInfo(info2);
            }
            else
            {
                info = info2;
                for (gfInt i = 0;i < info1->m_inters.size();i++)
                    info->m_inters.push_back(info1->m_inters.at(i));
                gfDeleteRayInterInfo(info1);
            }
        }
        else if (info1)
        {
            info = info1;
        }
        else if (info2)
        {
            info = info2;
        }
    }
    else
    {
        gfVector3 normal;
        gfDouble t_min = GF_MAX_DOUBLE;
        gfDouble t_max = GF_MIN_DOUBLE;
        info = gfCreateRayInterInfo();
        gfDouble t, u, v;
        gfPoint3 pt1, pt2, pt3;

        for (gfHEFace* face = root->faces.GetFirst(); face; face = root->faces.GetNext())
        {
            gfHalfEdge* first_edge = face->edge;
            gfHalfEdge* second_edge = first_edge->next;
            gfHalfEdge* edge = second_edge->next;

            while (first_edge != edge)
            {
                pt1.Set(first_edge->end->x, first_edge->end->y, first_edge->end->z);
                pt2.Set(second_edge->end->x, second_edge->end->y, second_edge->end->z);
                pt3.Set(edge->end->x, edge->end->y, edge->end->z);

                GFRESULT res = gfRayTriangleInter(ray_begin, ray_dir, &pt1, &pt2, &pt3, &u, &v, &t);

                if (GF_SUCC(res))
                {
                    gfVector3 normal;
                    if (t < t_min)
                    {
                        if (first_edge->normal && second_edge->normal && edge->normal)
                        {
                            normal.SetX(u*first_edge->normal->x+v*second_edge->normal->x+(1-u-v)*edge->normal->x);
                            normal.SetY(u*first_edge->normal->y+v*second_edge->normal->y+(1-u-v)*edge->normal->y);
                            normal.SetZ(u*first_edge->normal->z+v*second_edge->normal->z+(1-u-v)*edge->normal->z);

                            // normalize
                            normal.Normalize();
                        }
                        else
                        {
                            normal.Set(0.0, 0.0, 0.0);
                        }
                        t_min = t;
                    }

                    info->m_inters.push_back(t);
                    info->m_normals.push_back(normal);
                }

                edge = edge->next;
                second_edge = second_edge->next;
            }
        }

        // check if there is intersection
        if (info->m_inters.size() < 1)
        {
            gfDeleteRayInterInfo(info);
            info = NULL;
        }
    }

    // sort intersections
    if (info)
    {
        gfSortRayInterInfo(info);
    }

    return info;*/
}

GFRESULT gfObjHalfEdgeMesh::CreateVACache(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    DeleteVACache();

    // get number of triangles
    m_num_cached_triangles = 0;
    m_num_cached_wire_lines = 0;
    for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
    {
        //gfHEFace* face = *iter;
        gfHalfEdge* first_edge = face->edge;
        gfHalfEdge* edge = first_edge->next;
        m_num_cached_triangles++;
        m_num_cached_wire_lines++;

        while (first_edge != edge)
        {
            m_num_cached_triangles++;
            m_num_cached_wire_lines++;
            edge = edge->next;
        }

        m_num_cached_triangles = m_num_cached_triangles-2;
    }

    // allocate streams
    m_cache_vertices = new gfDouble[9*m_num_cached_triangles];
    m_cache_normals = new gfDouble[9*m_num_cached_triangles];
    m_cache_wire_vertices = new gfDouble[6*m_num_cached_wire_lines];

    // fill streams
    gfInt i = 0;
    gfInt j = 0;
    for (gfHEFace* face = m_face_list.GetFirst(); face; face = m_face_list.GetNext())
    {
        gfHalfEdge* first_edge = face->edge;
        gfHalfEdge* second_edge = first_edge->next;
        gfHalfEdge* edge = second_edge->next;

        // fill edge (first_edge, second_edge)
        m_cache_wire_vertices[j] = first_edge->end->x;
        j++;
        m_cache_wire_vertices[j] = first_edge->end->y;
        j++;
        m_cache_wire_vertices[j] = first_edge->end->z;
        j++;
        m_cache_wire_vertices[j] = second_edge->end->x;
        j++;
        m_cache_wire_vertices[j] = second_edge->end->y;
        j++;
        m_cache_wire_vertices[j] = second_edge->end->z;
        j++;

        while (first_edge != edge)
        {
            // fill edge (second_edge, edge)
            m_cache_wire_vertices[j] = second_edge->end->x;
            j++;
            m_cache_wire_vertices[j] = second_edge->end->y;
            j++;
            m_cache_wire_vertices[j] = second_edge->end->z;
            j++;
            m_cache_wire_vertices[j] = edge->end->x;
            j++;
            m_cache_wire_vertices[j] = edge->end->y;
            j++;
            m_cache_wire_vertices[j] = edge->end->z;
            j++;

            // fill triangle (first_edge, second_edge, edge)
            m_cache_vertices[i] = first_edge->end->x;
            if (first_edge->normal)
                m_cache_normals[i] = first_edge->normal->x;
            i++;

            m_cache_vertices[i] = first_edge->end->y;
            if (first_edge->normal)
                m_cache_normals[i] = first_edge->normal->y;
            i++;

            m_cache_vertices[i] = first_edge->end->z;
            if (first_edge->normal)
                m_cache_normals[i] = first_edge->normal->z;
            i++;

            m_cache_vertices[i] = second_edge->end->x;
            if (second_edge->normal)
                m_cache_normals[i] = second_edge->normal->x;
            i++;

            m_cache_vertices[i] = second_edge->end->y;
            if (second_edge->normal)
                m_cache_normals[i] = second_edge->normal->y;
            i++;

            m_cache_vertices[i] = second_edge->end->z;
            if (second_edge->normal)
                m_cache_normals[i] = second_edge->normal->z;
            i++;

            m_cache_vertices[i] = edge->end->x;
            if (edge->normal)
                m_cache_normals[i] = edge->normal->x;
            i++;

            m_cache_vertices[i] = edge->end->y;
            if (edge->normal)
                m_cache_normals[i] = edge->normal->y;
            i++;

            m_cache_vertices[i] = edge->end->z;
            if (edge->normal)
                m_cache_normals[i] = edge->normal->z;
            i++;

            edge = edge->next;
            second_edge = second_edge->next;
        }

        // fill edge (second_edge, first_edge)
        m_cache_wire_vertices[j] = second_edge->end->x;
        j++;
        m_cache_wire_vertices[j] = second_edge->end->y;
        j++;
        m_cache_wire_vertices[j] = second_edge->end->z;
        j++;
        m_cache_wire_vertices[j] = first_edge->end->x;
        j++;
        m_cache_wire_vertices[j] = first_edge->end->y;
        j++;
        m_cache_wire_vertices[j] = first_edge->end->z;
        j++;
    }

    gfLog("Computed VertexArray cache for half-edge mesh, time = %d ms.\n", count.EndMSCount());

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::DeleteVACache(void)
{
    if (m_cache_vertices)
        delete [] m_cache_vertices;
    m_cache_vertices = NULL;
    if (m_cache_normals)
        delete [] m_cache_normals;
    m_cache_normals = NULL;
    if (m_cache_wire_vertices)
        delete [] m_cache_wire_vertices;
    m_cache_wire_vertices = NULL;
    m_num_cached_triangles = 0;
    m_num_cached_wire_lines = 0;

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::DrawVACache(gfLong flags)
{
    if (flags & GF_FLG_SOLID)
    {
        glEnableClientState(GL_VERTEX_ARRAY);
        if (flags & GF_FLG_NORMALS)
            glEnableClientState(GL_NORMAL_ARRAY);

        if (m_cache_vertices)
            glVertexPointer(3, GL_DOUBLE, 0, m_cache_vertices);
        if (m_cache_normals && (flags & GF_FLG_NORMALS))
            glNormalPointer(GL_DOUBLE, 0, m_cache_normals);

        glLockArraysEXT(0, 3*m_num_cached_triangles);
        glDrawArrays(GL_TRIANGLES, 0, 3*m_num_cached_triangles);
        glUnlockArraysEXT();

        if (flags & GF_FLG_NORMALS)
            glDisableClientState(GL_NORMAL_ARRAY);
        glDisableClientState(GL_VERTEX_ARRAY);
    }

    if (flags & GF_FLG_WIRE)
    {
        glNormal3f(0.0, 0.0, 0.0);
        glEnableClientState(GL_VERTEX_ARRAY);

        if (m_cache_wire_vertices)
            glVertexPointer(3, GL_DOUBLE, 0, m_cache_wire_vertices);

        glLockArraysEXT(0, 2*m_num_cached_wire_lines);
        glDrawArrays(GL_LINES, 0, 2*m_num_cached_wire_lines);
        glUnlockArraysEXT();

        glDisableClientState(GL_VERTEX_ARRAY);
    }

    return GF_S_OK;
}

GFRESULT gfObjHalfEdgeMesh::DeformByNURBSLattice(gfObjGeomObject* lattice)
{
    if (!lattice)
        return GF_E_BADARG;
    if (lattice->GetID() != GF_NURBS_BOX)
        return GF_E_BADARG;
    gfObjNURBSBox* box = (gfObjNURBSBox*)lattice;

    // get domain of NURBS Box
    gfDouble u_min, u_max, v_min, v_max, w_min, w_max;
    box->GetDomain(&u_min, &u_max, &v_min, &v_max, &w_min, &w_max);

    // get bounding box of mesh
    if (!m_aabb_tree_root)
        ComputeWholeAABBTree();

    // get bounding box of NURBS box
    gfAuxAABB* box_aabb = (gfAuxAABB*)box->GetBoundVolume(GF_BOUND_VOL_AABB, NULL);

    if (!m_point_list.Empty())
    {
        for (gfHEPoint* point = m_point_list.GetFirst(); point; point = m_point_list.GetNext())
        {
            if (point)
            {
                // for point, compute domain parameters
                gfDouble w = w_min+(w_max-w_min)*(point->x-m_aabb_tree_root->aabb->m_min_x)/(m_aabb_tree_root->aabb->m_max_x-m_aabb_tree_root->aabb->m_min_x);
                gfDouble v = v_min+(v_max-v_min)*(point->y-m_aabb_tree_root->aabb->m_min_y)/(m_aabb_tree_root->aabb->m_max_y-m_aabb_tree_root->aabb->m_min_y);
                gfDouble u = u_min+(u_max-u_min)*(point->z-m_aabb_tree_root->aabb->m_min_z)/(m_aabb_tree_root->aabb->m_max_z-m_aabb_tree_root->aabb->m_min_z);

                // convert point to aabb of lattice
                gfPoint3* pt = box->GetVolumePoint(u, v, w);

                // convert computed point from lattice point to mesh point
                point->x = m_aabb_tree_root->aabb->m_min_x + (m_aabb_tree_root->aabb->m_max_x-m_aabb_tree_root->aabb->m_min_x) * (pt->X()-box_aabb->m_min_x) / (box_aabb->m_max_x-box_aabb->m_min_x);
                point->y = m_aabb_tree_root->aabb->m_min_y + (m_aabb_tree_root->aabb->m_max_y-m_aabb_tree_root->aabb->m_min_y) * (pt->Y()-box_aabb->m_min_y) / (box_aabb->m_max_y-box_aabb->m_min_y);
                point->z = m_aabb_tree_root->aabb->m_min_z + (m_aabb_tree_root->aabb->m_max_z-m_aabb_tree_root->aabb->m_min_z) * (pt->Z()-box_aabb->m_min_z) / (box_aabb->m_max_z-box_aabb->m_min_z);
                point->x = pt->X() - 0.5 * (box_aabb->m_max_x+box_aabb->m_min_x) + 0.5 * (m_aabb_tree_root->aabb->m_max_x+m_aabb_tree_root->aabb->m_min_x);
                point->y = pt->Y() - 0.5 * (box_aabb->m_max_y+box_aabb->m_min_y) + 0.5 * (m_aabb_tree_root->aabb->m_max_y+m_aabb_tree_root->aabb->m_min_y);
                point->z = pt->Z() - 0.5 * (box_aabb->m_max_z+box_aabb->m_min_z) + 0.5 * (m_aabb_tree_root->aabb->m_max_z+m_aabb_tree_root->aabb->m_min_z);
                gfDeletePoint3(pt);
            }
        }
    }

    gfDeleteBoundVolume(box_aabb);

    // delete AABB tree, next time must be recomputed
    if (m_aabb_tree_root)
        DeleteAABBTree(m_aabb_tree_root);
    m_aabb_tree_root = NULL;
    if (!m_aabb_tree_root)
        ComputeWholeAABBTree();

    // delete cache
    DeleteVACache();
    ComputeSmoothNormals();

    return GF_S_OK;
}