//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjHalfEdgeMesh.h                                  //
//   Description: Implementation of mesh based on half edge     //
//                                                              //
//////////////////////////////////////////////////////////////////


#ifndef _GF_C_HALF_EDGE_MESH_H_
#define _GF_C_HALF_EDGE_MESH_H_

#include <list>
#include <queue>
#include "gfAuxList.h"
#include "gfDefines.h"
#include <vector>
#include "gfPoint.h"
#include "gfVector.h"
#include "gfMatrix.h"
#include "gfObjGeomObject.h"
#include "gfObjPolyLine.h"
#include "gfAuxHalfEdge.h"
#include "gfAuxBoundingVolume.h"
#include "Ply/rply.h"
#include <string>

using namespace std;

// auxiliary structures
struct gfAuxExtHE
{
    gfHalfEdge* new_edge_1;
    gfHalfEdge* new_edge_2;
    gfHalfEdge* new_edge_3;
    gfHalfEdge* new_edge_4;
    gfHEPoint* edge_point;
    gfHalfEdge* old_edge;
};

struct gfAABBTreeNode
{
    gfAuxAABB* aabb;
    gfAABBTreeNode* left;
    gfAABBTreeNode* right;
    gfAuxList<gfHEFace*> faces;
};

struct gfAABBQueueUnit
{
    gfAABBTreeNode* node;
    gfDouble tmin;
    gfDouble tmax;
};

class gfObjHalfEdgeMesh : public gfObjGeomObject
{
private:

    gfAuxList<gfHalfEdge*> m_edge_list;
    gfAuxList<gfHEFace*> m_face_list;
    gfAuxList<gfHEPoint*> m_point_list;
    gfAuxList<gfHENormal*> m_normal_list;
    gfAuxList<gfHETexCoords*> m_tex_coord_list;

    // auxiliary for mesh load from ply
    static std::list<gfInt>* load_faces;
    static gfHEPoint** load_vertices;
    static gfInt load_num_faces;
    static gfInt load_num_vertices;
    static gfDouble minx, maxx, miny, maxy, minz, maxz;
    static int vertex_cb(p_ply_argument argument);
    static int face_cb(p_ply_argument argument);

    // AABB tree
    gfAABBTreeNode* m_aabb_tree_root;
    //std::vector<gfAABBQueueUnit> m_aabb_queue;
    struct compare_queue_unit
    {
        bool operator() (const gfAABBQueueUnit & x, const gfAABBQueueUnit & y)
        {
            return x.tmin > y.tmin;
        }
    };
    std::priority_queue<gfAABBQueueUnit, std::vector<gfAABBQueueUnit>, compare_queue_unit> m_aabb_queue;

    // vertex array cache
    gfDouble* m_cache_vertices;
    gfDouble* m_cache_normals;
    gfDouble* m_cache_wire_vertices;
    gfInt m_num_cached_triangles;
    gfInt m_num_cached_wire_lines;
    GFRESULT CreateVACache(void);
    GFRESULT DeleteVACache(void);
    GFRESULT DrawVACache(gfLong flags);

public:

    gfObjHalfEdgeMesh(void);
    ~gfObjHalfEdgeMesh(void);

    // gfObjGeomObject functions
    GFRESULT OGLDraw(gfLong flags);
    gfRayInterInfo* RayIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfGeomObjID GetID(void);
    GFRESULT Transform(gfMatrix4* matrix);
    gfObjGeomObject* Clone(void);
    gfAuxBoundingVolume* GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix);

    // work with content
    GFRESULT Clear(void);
    GFRESULT AddHalfEdge(gfHalfEdge* edge);
    GFRESULT RemoveHalfEdge(gfHalfEdge* edge);
    GFRESULT AddHEFace(gfHEFace* face);
    GFRESULT RemoveHEFace(gfHEFace* face);
    GFRESULT AddHEPoint(gfHEPoint* point);
    GFRESULT RemoveHEPoint(gfHEPoint* point);
    GFRESULT AddHENormal(gfHENormal* normal);
    GFRESULT RemoveHENormal(gfHENormal* normal);
    GFRESULT AddHETexCoords(gfHETexCoords* tex_coords);
    GFRESULT RemoveHETexCoords(gfHETexCoords* tex_coords);

    // get informations about mesh
    gfInt GetNumVertices();
    gfInt GetNumFaces();
    gfInt GetNumHalfEdges();
    void PrintStatus(void);

    // initialize as basic objects
    GFRESULT InitCube(gfPoint3* center, gfDouble sizex, gfDouble sizey, gfDouble sizez);
    GFRESULT InitPlane(gfPoint3* begin, gfVector3* base1, gfVector3* base2, gfInt slices1, gfInt slices2);
    GFRESULT InitSphere(gfPoint3* center, gfDouble radius, gfInt slices1, gfInt slices2);

    // for loading from ply
    GFRESULT InitFromPly(std::string strFile);

    // support functions
    GFRESULT ComputeSmoothNormals(void);
    GFRESULT ComputeFacesNormals(void);
    GFRESULT ComputeGaussianCurvature(void);
    GFRESULT ReverseNormals(void);

    // subdivision
    gfObjHalfEdgeMesh* CatmullClarkSubdivision(void);

    // deformation
    GFRESULT DeformByNURBSLattice(gfObjGeomObject* lattice);

    // work with AABB tree
    GFRESULT DeleteAABBTree(gfAABBTreeNode* tree_root);
    GFRESULT ComputeWholeAABBTree(void);
    gfAABBTreeNode* CreateAABBTree(gfInt depth, gfAuxList<gfHEFace*>* faces);
    gfRayInterInfo* RayAABBIntersection(gfAABBTreeNode* root, gfPoint3 * ray_begin, gfVector3* ray_dir);
};

#endif