//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjNURBSBox.cpp                                    //
//   Description: NURBS box class                               //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfObjNURBSBox.h"
#include "gfObjectFactory.h"
#include "gfAuxNumUtils.h"
#include "gfAuxLogging.h"
#include "gfAuxGeomUtils.h"
#include "gfWinFPSCounter.h"
#include "gfAux4DArray.h"
#include "gfAuxItersectionUtils.h"
#include "gfAuxBoundCone.h"

#include "./OGL/glut.h"


gfObjNURBSBox::gfObjNURBSBox(void)
{
    m_time_stamp = 0;
    m_weights.resize(0);
    m_control_points.resize(0);
    m_u_knot_vector.resize(0);
    m_v_knot_vector.resize(0);
    m_w_knot_vector.resize(0);
    m_num_u_control_points = 0;
    m_num_v_control_points = 0;
    m_num_w_control_points = 0;
    m_u_degree = 0;
    m_v_degree = 0;
    m_w_degree = 0;

    m_computed_boundaries = false;

    m_cache_vertices = NULL;
    m_cache_normals = NULL;
    m_cache_wire_vertices = NULL;
}

gfObjNURBSBox::~gfObjNURBSBox(void)
{
    Clear();
}

GFRESULT gfObjNURBSBox::OGLDraw(gfLong flags)
{
    if (m_time_stamp == 0)
        return GF_S_OK;

    gfInt approx = GF_VIS_PREC;

    if (flags & GF_FLG_CONTROL_NET)
    {
        glNormal3f(0.0, 0.0, 0.0);
        glColor3f(0.0, 0.0, 0.0);
        glPointSize(5.0);
        for (gfInt i = 0;i < m_num_u_control_points;i++)
        {
            for (gfInt j = 0;j < m_num_v_control_points;j++)
            {
                for (gfInt k = 0;k < m_num_w_control_points;k++)
                {
                    gfPoint3*  pt = GetControlPoint(i, j, k);
                    glBegin(GL_POINTS);
                    glVertex3f(pt->X(), pt->Y(), pt->Z());
                    glEnd();

                    glBegin(GL_LINES);

                    if ((i + 1) < m_num_u_control_points)
                    {
                        gfPoint3*  pt1 = GetControlPoint(i + 1, j, k);
                        glVertex3f(pt->X(), pt->Y(), pt->Z());
                        glVertex3f(pt1->X(), pt1->Y(), pt1->Z());
                    }
                    if ((j + 1) < m_num_v_control_points)
                    {
                        gfPoint3*  pt1 = GetControlPoint(i, j + 1, k);
                        glVertex3f(pt->X(), pt->Y(), pt->Z());
                        glVertex3f(pt1->X(), pt1->Y(), pt1->Z());
                    }
                    if ((k + 1) < m_num_w_control_points)
                    {
                        gfPoint3*  pt1 = GetControlPoint(i, j, k + 1);
                        glVertex3f(pt->X(), pt->Y(), pt->Z());
                        glVertex3f(pt1->X(), pt1->Y(), pt1->Z());
                    }

                    glEnd();
                }
            }
        }
        glPointSize(1.0);
    }

    // check if we are caching
    if (flags & GF_FLG_VERTEX_ARRAY_CACHE)
    {
        if (!m_cache_vertices)
            CreateVACache(approx);
        return DrawVACache(flags);
    }
    DeleteVACache();

    // interval of definition
    gfDouble a, b, c, d, e, f;
    GetDomain(&a, &b, &c, &d, &e, &f);
    if (gfIsZero(b - a) || gfIsZero(d - c) || gfIsZero(f - e))
        return GF_S_OK;

    gfInt n = GetNumUControlPoints() - 1;
    gfInt m = GetNumVControlPoints() - 1;
    gfInt o = GetNumWControlPoints() - 1;

    gfPoint3 * aux_points = new gfPoint3 [approx * approx * approx];

    for (gfInt i = 0;i < approx;i++)
    {
        // current u parameter
        gfDouble u = a + i * (b - a) / (approx - 1);
        gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
        std::vector<gfDouble> ubasis;
        gfGetBSplineAllValues(m_u_degree, &m_u_knot_vector, uspan, u, &ubasis);

        for (gfInt j = 0;j < approx;j++)
        {
            // current v parameter     
            gfDouble v = c + j * (d - c) / (approx - 1);
            gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
            std::vector<gfDouble> vbasis;
            gfGetBSplineAllValues(m_v_degree, &m_v_knot_vector, vspan, v, &vbasis);

            for (gfInt k = 0;k < approx;k++)
            {
                // current w parameter     
                gfDouble w = e + k * (f - e) / (approx - 1);
                gfInt wspan = gfGetBSplineKnotSpan(o, m_w_degree, &m_w_knot_vector, w);
                std::vector<gfDouble> wbasis;
                gfGetBSplineAllValues(m_w_degree, &m_w_knot_vector, wspan, w, &wbasis);

                gfDouble ax = 0;
                gfDouble ay = 0;
                gfDouble az = 0;
                gfDouble aw = 0;

                for (gfInt iter1 = 0;iter1 <= m_u_degree;iter1++)
                {
                    gfInt uindex = uspan - m_u_degree + iter1;

                    for (gfInt iter2 = 0;iter2 <= m_v_degree;iter2++)
                    {
                        gfInt vindex = vspan - m_v_degree + iter2;

                        for (gfInt iter3 = 0;iter3 <= m_w_degree;iter3++)
                        {
                            gfInt windex = wspan - m_w_degree + iter3;

                            gfDouble aux = ubasis.at(iter1) * vbasis.at(iter2) * wbasis.at(iter3)* GetWeight(uindex, vindex, windex);
                            ax += aux * GetControlPoint(uindex, vindex, windex)->X();
                            ay += aux * GetControlPoint(uindex, vindex, windex)->Y();
                            az += aux * GetControlPoint(uindex, vindex, windex)->Z();
                            aw += aux;
                        }
                    }
                }

                // set computed point to array
                aux_points[i + j * approx + k * approx * approx].Set(ax / aw, ay / aw, az / aw);
            }
        }
    }

    if (flags & GF_FLG_WIRE)
    {
        for (gfInt k = 0;k < approx;k++)
        {
            for (gfInt i = 0;i < approx;i++)
            {
                for (gfInt j = 0;j < approx;j++)
                {
                    gfPoint3  pt = aux_points[i + j * approx + k * approx * approx];
                    glBegin(GL_LINES);

                    if ((i + 1) < approx)
                    {
                        gfPoint3  pt1 = aux_points[i + 1 + j * approx + k * approx * approx];
                        glVertex3f(pt.X(), pt.Y(), pt.Z());
                        glVertex3f(pt1.X(), pt1.Y(), pt1.Z());
                    }
                    if ((j + 1) < approx)
                    {
                        gfPoint3  pt1 = aux_points[i + (j + 1) * approx + k * approx * approx];
                        glVertex3f(pt.X(), pt.Y(), pt.Z());
                        glVertex3f(pt1.X(), pt1.Y(), pt1.Z());
                    }
                    if ((k + 1) < approx)
                    {
                        gfPoint3  pt1 = aux_points[i + j * approx + (k + 1) * approx * approx];
                        glVertex3f(pt.X(), pt.Y(), pt.Z());
                        glVertex3f(pt1.X(), pt1.Y(), pt1.Z());
                    }

                    glEnd();
                }
            }
        }    
    }

    if (flags & GF_FLG_SOLID)
    {
        // fill triangle loops
        for (gfInt k = 0;k < approx;k++)
        {
            for (gfInt i = 0;i < approx - 1;i++)
            {
                glBegin(GL_TRIANGLE_STRIP);
                for (gfInt j = 0;j < approx;j++)
                {
                    // color
                    glColor3f( 0.25 + i / (2.0 * approx), 0.25 + j / (2.0 * approx), 0.25 + k / (2.0 * approx));

                    gfPoint3  pt1 = aux_points[i + j * approx + k * approx * approx];
                    gfPoint3  pt2 = aux_points[(i + 1) + j * approx + k * approx * approx];
                    glVertex3f(pt1.X(), pt1.Y(), pt1.Z());
                    glVertex3f(pt2.X(), pt2.Y(), pt2.Z());
                }
                glEnd();
            }
        }

        for (gfInt j = 0;j < approx;j++)
        {
            for (gfInt i = 0;i < approx - 1;i++)
            {
                glBegin(GL_TRIANGLE_STRIP);
                for (gfInt k = 0;k < approx;k++)
                {
                    // color
                    glColor3f( 0.25 + i / (2.0 * approx), 0.25 + j / (2.0 * approx), 0.25 + k / (2.0 * approx));

                    gfPoint3  pt1 = aux_points[i + j * approx + k * approx * approx];
                    gfPoint3  pt2 = aux_points[(i + 1) + j * approx + k * approx * approx];
                    glVertex3f(pt1.X(), pt1.Y(), pt1.Z());
                    glVertex3f(pt2.X(), pt2.Y(), pt2.Z());
                }
                glEnd();
            }
        }

        for (gfInt i = 0;i < approx;i++)
        {
            for (gfInt j = 0;j < approx - 1;j++)
            {
                glBegin(GL_TRIANGLE_STRIP);
                for (gfInt k = 0;k < approx;k++)
                {
                    // color
                    glColor3f( 0.25 + i / (2.0 * approx), 0.25 + j / (2.0 * approx), 0.25 + k / (2.0 * approx));

                    gfPoint3  pt1 = aux_points[i + j * approx + k * approx * approx];
                    gfPoint3  pt2 = aux_points[i + (j + 1) * approx + k * approx * approx];
                    glVertex3f(pt1.X(), pt1.Y(), pt1.Z());
                    glVertex3f(pt2.X(), pt2.Y(), pt2.Z());
                }
                glEnd();
            }
        }
    }

    delete [] aux_points;            

    return GF_S_OK;
}

gfRayInterInfo* gfObjNURBSBox::RayIntersection(gfMatrix4* obj_matrix, gfPoint3 * ray_begin, gfVector3* ray_dir)
{
    /*gfInt num_trees = m_bezier_aabb_trees.size();
    if (num_trees == 0)
        ComputeWholeAABBTree();
    num_trees = m_bezier_aabb_trees.size();*/

    gfRayInterInfo* result = NULL;
    //gfDisableLogging();

    if (!m_computed_boundaries)
    {
        gfDouble umin, umax, vmin, vmax, wmin, wmax;
        GetDomain(&umin, &umax, &vmin, &vmax, &wmin, &wmax);
        m_boundary_patches[0] = GetUVIsopatch(wmin);
        m_boundary_patches[1] = GetUVIsopatch(wmax-GF_EPS);
        m_boundary_patches[2] = GetUWIsopatch(vmin);
        m_boundary_patches[3] = GetUWIsopatch(vmax-GF_EPS);
        m_boundary_patches[4] = GetVWIsopatch(umin);
        m_boundary_patches[5] = GetVWIsopatch(umax-GF_EPS);
        m_computed_boundaries = true;
    }

    // get inttersection with boundary patches
    gfRayInterInfo* info[6];
    info[0] = m_boundary_patches[0]->RayIntersection(obj_matrix, ray_begin, ray_dir);
    info[1] = m_boundary_patches[1]->RayIntersection(obj_matrix, ray_begin, ray_dir);
    info[2] = m_boundary_patches[2]->RayIntersection(obj_matrix, ray_begin, ray_dir);
    info[3] = m_boundary_patches[3]->RayIntersection(obj_matrix, ray_begin, ray_dir);
    info[4] = m_boundary_patches[4]->RayIntersection(obj_matrix, ray_begin, ray_dir);
    info[5] = m_boundary_patches[5]->RayIntersection(obj_matrix, ray_begin, ray_dir);

    gfDouble min = GF_MAX_DOUBLE;
    gfInt index = -1;
    for (gfInt i = 0;i <6;i++)
    {
        if (info[i])
        {
            if (info[i]->m_inters.at(0) < min)
            {
                min = info[i]->m_inters.at(0);
                index = i;
            }
        }
    }
    for (gfInt i = 0;i <6;i++)
        if (i != index && info[i])
            gfDeleteRayInterInfo(info[i]);
    if (index > -1)
        return result = info[index];

    // clear queue
    /*for (; !m_aabb_queue.empty(); m_aabb_queue.pop())
        ;

    // fill queue
    for (gfInt i = 0;i < num_trees;i++)
        RayAABBIntersection(m_bezier_aabb_trees.at(i), ray_begin, ray_dir);

    //go throught queue and find exact intersection
    gfDouble t_min = GF_MAX_DOUBLE;
    gfDouble u_min, v_min, w_min;
    gfVector3 normal;
    gfBool inter = false;

    gfVector3 up;
    up.Set(-ray_dir->Y()*ray_dir->Z(), ray_dir->X()*ray_dir->Z()/2.0, ray_dir->X()*ray_dir->Y()/2.0);
    gfVector3 right;
    gfVectorProduct3(ray_dir, &up, &right);
    right.Normalize();
    up.Normalize();
    gfDouble d1 = -(up.X()*ray_begin->X()+up.Y()*ray_begin->Y()+up.Z()*ray_begin->Z());
    gfDouble d2 = -(right.X()*ray_begin->X()+right.Y()*ray_begin->Y()+right.Z()*ray_begin->Z());

    gfInt num = 0;
    gfInt num_in = 0;
    gfInt num_out = 0;
    for (; !m_aabb_queue.empty(); m_aabb_queue.pop())
    {
        num++;
        gfBezier3AABBQueueUnit unit = m_aabb_queue.top();
        if (unit.tmin>t_min)
        {
            num_out++;
            continue;
        }

        num_in++;

        gfRayBezierBoxInfo* res = unit.node->box->BezierClipping(NULL, 0, 
                                                    up.X(), up.Y(), up.Z(), d1, right.X(), right.Y(), right.Z(), d2);
        gfDouble umin, umax, vmin, vmax, wmin, wmax;
        unit.node->box->GetDomain(&umin, &umax, &vmin, &vmax, &wmin, &wmax);

        //res->m_inters.push_back(gfPoint3((umin+umax)/2.0, (vmin+vmax)/2.0, (wmin+wmax)/2.0));
        if (res)
        {
            gfInt size = res->m_inters.size();
            if (size < 1)
                continue;

            gfPoint3* pt = unit.node->box->GetVolumePoint(res->m_inters.at(i).X(), res->m_inters.at(i).Y(), res->m_inters.at(i).Z());
            gfVector3 vec(pt->X()-ray_begin->X(), pt->Y()-ray_begin->Y(), pt->Z()-ray_begin->Z());
            gfDouble t = gfScalarProduct3(ray_dir, &vec);
            t = (unit.tmin+unit.tmax)/2.0;
            if (t > 0 && t < t_min)
            {
                inter = true;
                t_min = t;
                u_min = res->m_inters.at(i).X();
                v_min = res->m_inters.at(i).Y();
                w_min = res->m_inters.at(i).Z();

                gfVector3* nm = unit.node->box->GetNormal(res->m_inters.at(i).X(), res->m_inters.at(i).Y(), res->m_inters.at(i).Z());
                normal.Set(nm->X(), nm->Y(), nm->Z());
                gfDeleteVector3(nm);
            }
            gfDeletePoint3(pt);

            delete res;
        }
    }

    if (inter)
    {
        if (result)
        {
            if (result->m_inters.at(0) > t_min)
            {
                result->m_inters.clear();
                result->m_normals.clear();
                result->m_uvw.clear();

                result->m_inters.push_back(t_min);
                result->m_normals.push_back(gfVector3(normal.X(), normal.Y(), normal.Z()));
                result->m_uvw.push_back(gfPoint3(u_min, v_min, w_min));
            }
        }
        else
        {
            result = gfCreateRayInterInfo();

            result->m_inters.push_back(t_min);
            result->m_normals.push_back(gfVector3(normal.X(), normal.Y(), normal.Z()));
            result->m_uvw.push_back(gfPoint3(u_min, v_min, w_min));
        }
    }*/

    gfEnableLogging();

    return result;
}

gfGeomObjID gfObjNURBSBox::GetID(void)
{
    return GF_NURBS_BOX;
}

GFRESULT gfObjNURBSBox::Transform(gfMatrix4* matrix)
{
    // transform control points
    for (gfInt i = 0;i < m_num_u_control_points;i++)
    {
        for (gfInt j = 0;j < m_num_v_control_points;j++)
        {
            for (gfInt k = 0;k < m_num_w_control_points;k++)
            {
                gfPoint3* pt = GetControlPoint(i, j, k);
                gfTransformPoint3(pt, matrix);
            }
        }
    }

    DeleteVACache();
    if (m_computed_boundaries)
    {
        gfDeleteGeomObject(m_boundary_patches[0]);
        gfDeleteGeomObject(m_boundary_patches[1]);
        gfDeleteGeomObject(m_boundary_patches[2]);
        gfDeleteGeomObject(m_boundary_patches[3]);
        gfDeleteGeomObject(m_boundary_patches[4]);
        gfDeleteGeomObject(m_boundary_patches[5]);
        m_computed_boundaries = false;
    }

    return GF_S_OK;
}

gfObjGeomObject* gfObjNURBSBox::Clone(void)
{
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());
    gfObjNURBSBox* box = (gfObjNURBSBox*) clone;

    // set parameters of patch
    gfInt num_u_control_points = GetNumUControlPoints();
    gfInt num_v_control_points = GetNumVControlPoints();
    gfInt num_w_control_points = GetNumWControlPoints();
    box->SetParams(m_u_degree, m_v_degree, m_w_degree, num_u_control_points, num_v_control_points, num_w_control_points);

    // copy control points and weights
    for (gfInt i = 0;i < num_u_control_points;i++)
        for (gfInt j = 0;j < num_v_control_points;j++)
            for (gfInt k = 0;k < num_w_control_points;k++)
            {
                gfPoint3* new_point = box->GetControlPoint(i, j, k);
                gfPoint3* old_point = GetControlPoint(i, j, k);
                new_point->Set(old_point->X(), old_point->Y(), old_point->Z());
                box->SetWeight(i, j, k, GetWeight(i, j, k));
            }

    // copy knot vectors
    gfInt num_u_knots = GetNumUKnots();
    for (gfInt i = 0;i < num_u_knots;i++)
    {
        box->SetUKnot(i, GetUKnot(i));
    }
    gfInt num_v_knots = GetNumVKnots();
    for (gfInt i = 0;i < num_v_knots;i++)
    {
        box->SetVKnot(i, GetVKnot(i));
    }
    gfInt num_w_knots = GetNumWKnots();
    for (gfInt i = 0;i < num_w_knots;i++)
    {
        box->SetWKnot(i, GetWKnot(i));
    }

    return clone;
}

GFRESULT gfObjNURBSBox::Clear()
{
    SetParams(0, 0, 0, 0, 0, 0);
    DeleteVACache();
    DeleteWholeAABBTree();

    if (m_computed_boundaries)
    {
        gfDeleteGeomObject(m_boundary_patches[0]);
        gfDeleteGeomObject(m_boundary_patches[1]);
        gfDeleteGeomObject(m_boundary_patches[2]);
        gfDeleteGeomObject(m_boundary_patches[3]);
        gfDeleteGeomObject(m_boundary_patches[4]);
        gfDeleteGeomObject(m_boundary_patches[5]);
        m_computed_boundaries = false;
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitBezier(gfInt num_u_points, gfInt num_v_points, gfInt num_w_points)
{
    Clear();

    gfInt u_degree = num_u_points - 1;
    gfInt v_degree = num_v_points - 1;
    gfInt w_degree = num_w_points - 1;
    SetParams(u_degree, v_degree, w_degree, num_u_points, num_v_points, num_w_points);

    gfInt num_u_knots = GetNumUKnots();
    gfInt num_v_knots = GetNumVKnots();
    gfInt num_w_knots = GetNumWKnots();
    for (gfInt i = 0;i < num_u_knots;i++)
    {
        SetUKnot(i, (i < num_u_knots / 2) ? 0.0 : 1.0);
    }
    for (gfInt i = 0;i < num_v_knots;i++)
    {
        SetVKnot(i, (i < num_v_knots / 2) ? 0.0 : 1.0);
    }
    for (gfInt i = 0;i < num_w_knots;i++)
    {
        SetWKnot(i, (i < num_w_knots / 2) ? 0.0 : 1.0);
    }

    for (gfInt i = 0;i < num_u_points;i++)
        for (gfInt j = 0;j < num_v_points;j++)
            for (gfInt k = 0;k < num_w_points;k++)
            {
                SetWeight(i, j, k, 1.0);
            }

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitBezier(gfInt num_u_points, gfInt num_v_points, gfInt num_w_points,
                    gfDouble u_min, gfDouble u_max, gfDouble v_min, gfDouble v_max, gfDouble w_min, gfDouble w_max)
{
    /*if (m_num_u_control_points > 0)
        Clear();*/

    gfInt u_degree = num_u_points - 1;
    gfInt v_degree = num_v_points - 1;
    gfInt w_degree = num_w_points - 1;
    SetParams(u_degree, v_degree, w_degree, num_u_points, num_v_points, num_w_points);

    gfInt num_u_knots = GetNumUKnots();
    gfInt num_v_knots = GetNumVKnots();
    gfInt num_w_knots = GetNumWKnots();
    for (gfInt i = 0;i < num_u_knots;i++)
    {
        SetUKnot(i, (i < num_u_knots / 2) ? u_min : u_max);
    }
    for (gfInt i = 0;i < num_v_knots;i++)
    {
        SetVKnot(i, (i < num_v_knots / 2) ? v_min : v_max);
    }
    for (gfInt i = 0;i < num_w_knots;i++)
    {
        SetWKnot(i, (i < num_w_knots / 2) ? w_min : w_max);
    }

    for (gfInt i = 0;i < num_u_points;i++)
    {
        for (gfInt j = 0;j < num_v_points;j++)
        {
            for (gfInt k = 0;k < num_w_points;k++)
            {
                SetWeight(i, j, k, 1.0);
            }
        }
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitBox(gfPoint3* center, gfDouble tx, gfDouble ty, gfDouble tz)
{
    Clear();
    SetParams(1, 1, 1, 2, 2, 2);

    SetUKnot(0, 0.0);
    SetUKnot(1, 0.0);
    SetUKnot(2, 1.0);
    SetUKnot(3, 1.0);

    SetVKnot(0, 0.0);
    SetVKnot(1, 0.0);
    SetVKnot(2, 1.0);
    SetVKnot(3, 1.0);

    SetWKnot(0, 0.0);
    SetWKnot(1, 0.0);
    SetWKnot(2, 1.0);
    SetWKnot(3, 1.0);

    GetControlPoint(0, 0, 0)->Set(center->X() - tx, center->Y() - ty, center->Z() - tz);
    SetWeight(0, 0, 0, 1.0);

    GetControlPoint(1, 0, 0)->Set(center->X() + tx, center->Y() - ty, center->Z() - tz);
    SetWeight(1, 0, 0, 1.0);

    GetControlPoint(0, 1, 0)->Set(center->X() - tx, center->Y() + ty, center->Z() - tz);
    SetWeight(0, 1, 0, 1.0);

    GetControlPoint(0, 0, 1)->Set(center->X() - tx, center->Y() - ty, center->Z() + tz);
    SetWeight(0, 0, 1, 1.0);

    GetControlPoint(1, 1, 0)->Set(center->X() + tx, center->Y() + ty, center->Z() - tz);
    SetWeight(1, 1, 0, 1.0);

    GetControlPoint(1, 0, 1)->Set(center->X() + tx, center->Y() - ty, center->Z() + tz);
    SetWeight(1, 0, 1, 1.0);

    GetControlPoint(0, 1, 1)->Set(center->X() - tx, center->Y() + ty, center->Z() + tz);
    SetWeight(0, 1, 1, 1.0);

    GetControlPoint(1, 1, 1)->Set(center->X() + tx, center->Y() + ty, center->Z() + tz);
    SetWeight(1, 1, 1, 1.0);

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitSphere(gfPoint3* center, gfDouble radius)
{
    Clear();
    gfObjNURBSPatch* patch = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
    patch->InitBezier(4, 2);

    patch->GetControlPoint(0, 0)->Set(center->X(), center->Y() + radius, center->Z());
    patch->SetWeight(0, 0, 1.0);
    patch->GetControlPoint(1, 0)->Set(center->X() + 2*radius, center->Y() + radius, center->Z());
    patch->SetWeight(1, 0, 1.0 / 3.0);
    patch->GetControlPoint(2, 0)->Set(center->X() + 2*radius, center->Y() - radius, center->Z());
    patch->SetWeight(2, 0, 1.0 / 3.0);
    patch->GetControlPoint(3, 0)->Set(center->X(), center->Y() - radius, center->Z());
    patch->SetWeight(3, 0, 1.0);
    patch->GetControlPoint(0, 1)->Set(center->X(), center->Y() + radius, center->Z());
    patch->SetWeight(0, 1, 1.0);
    patch->GetControlPoint(1, 1)->Set(center->X(), center->Y() + radius / 3.0, center->Z());
    patch->SetWeight(1, 1, 1.0);
    patch->GetControlPoint(2, 1)->Set(center->X(), center->Y() - radius / 3.0, center->Z());
    patch->SetWeight(2, 1, 1.0);
    patch->GetControlPoint(3, 1)->Set(center->X(), center->Y() - radius, center->Z());
    patch->SetWeight(3, 1, 1.0);

    gfVector3 axis(0.0, 1.0, 0.0);
    InitRotational(center, &axis, patch);
    gfDeleteGeomObject(patch);

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitTorus(gfPoint3* center, gfDouble radius1, gfDouble radius2)
{
    Clear();
    gfObjNURBSPatch* patch = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
    
    //init patch as full circle
    patch->InitBezier(3, 3);

    gfPoint3 * pt = patch->GetControlPoint(0, 0);
    pt->Set(center->X(), center->Y() + radius1, center->Z() - radius2);
    patch->SetWeight(0, 0, 1.0);

    pt = patch->GetControlPoint(1, 0);
    pt->Set(center->X(), center->Y() + radius1 - radius2, center->Z() - radius2);
    patch->SetWeight(1, 0, 1.0);

    pt = patch->GetControlPoint(2, 0);
    pt->Set(center->X(), center->Y() + radius1 - radius2, center->Z());
    patch->SetWeight(2, 0, 2.0);

    pt = patch->GetControlPoint(1, 1);
    pt->Set(center->X() , center->Y()+ radius1, center->Z());
    patch->SetWeight(1, 1, 1.0);

    pt = patch->GetControlPoint(0, 1);
    pt->Set(center->X(), center->Y() + radius1 + radius2, center->Z() - radius2);
    patch->SetWeight(0, 1, 1.0);

    pt = patch->GetControlPoint(0, 2);
    pt->Set(center->X(), center->Y() + radius1 + radius2, center->Z());
    patch->SetWeight(0, 2, 2.0);

    pt = patch->GetControlPoint(2, 1);
    pt->Set(center->X(), center->Y() + radius1 - radius2, center->Z() + radius2);
    patch->SetWeight(2, 1, 2.0);

    pt = patch->GetControlPoint(1, 2);
    pt->Set(center->X(), center->Y() + radius1 + radius2, center->Z() + radius2);
    patch->SetWeight(1, 2, 2.0);

    pt = patch->GetControlPoint(2, 2);
    pt->Set(center->X(), center->Y() + radius1, center->Z() + radius2);
    patch->SetWeight(2, 2, 4.0);

    gfVector3 axis(0.0, 0.0, 1.0);
    InitRotational(center, &axis, GF_PI / 3, patch);
    gfDeleteGeomObject(patch);

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitCone(gfPoint3* center, gfDouble radius1, gfDouble radius2, gfDouble height)
{
    Clear();
    gfObjNURBSCurve* curve = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
    curve->InitPolyline(2);

    gfPoint3* pt = curve->GetControlPoint(0);
    pt->Set(center->X(), center->Y() + radius1, center->Z());

    pt = curve->GetControlPoint(1);
    pt->Set(center->X(), center->Y() + radius2, center->Z() + height);

    gfVector3 axis(0.0, 0.0, 1.0);
    InitRotational(center, &axis, curve);
    gfDeleteGeomObject(curve);

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitRandom(gfPoint3* center, gfDouble tx, gfDouble ty, gfDouble tz)
{
    Clear();
    InitBezier(3, 4, 5);

    for (gfInt i = 0;i < 3;i++)
        for (gfInt j = 0;j < 4;j++)
            for (gfInt k = 0;k < 5;k++)
            {
                gfDouble ax = (- tx + 2 * tx * i / 2.0 + /*gfRand(0.8)*/ - 0.2) / 1.0;
                gfDouble ay = (- ty + 2 * ty * j / 3.0 + /*gfRand(0.8)*/ - 0.2) / 1.0;
                gfDouble az = (- tz + 2 * tz * k / 4.0 + /*gfRand(0.8)*/ - 0.2) / 2.0;

                // move interior control point
                /*if (i == 2 && j == 2 && k == 2)
                    ax += 3.0;*/



                GetControlPoint(i, j, k)->Set(center->X() + ax, center->Y() + ay, center->Z() + az);
                SetWeight(i, j, k, 1.0);
            }

            // move interior control point


    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitTransitional(gfObjNURBSCurve* nurbs_curve, gfObjNURBSPatch* nurbs_patch)
{
    Clear();

    gfPoint3* pt1 = nurbs_curve->GetControlPoint(0);
    gfPoint3* pt2 = nurbs_patch->GetControlPoint(0, 0);
    /*if (!gfArePointsEqual3(pt1, pt2))
        return GF_E_BADARG;*/

    // retrieve data from given curve and patch
    gfInt u_degree = nurbs_patch->GetUDegree();
    gfInt num_u_points = nurbs_patch->GetNumUControlPoints();
    gfInt num_u_knots = nurbs_patch->GetNumUKnots();

    gfInt v_degree = nurbs_patch->GetVDegree();
    gfInt num_v_points = nurbs_patch->GetNumVControlPoints();
    gfInt num_v_knots = nurbs_patch->GetNumVKnots();

    gfInt w_degree = nurbs_curve->GetDegree();
    gfInt num_w_points = nurbs_curve->GetNumControlPoints();
    gfInt num_w_knots = nurbs_curve->GetNumKnots();

    // set parameters of patch
    SetParams(u_degree, v_degree, w_degree, num_u_points, num_v_points, num_w_points);

    // set u and v knots of patch
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, nurbs_patch->GetUKnot(i));
    for (gfInt j = 0;j < num_v_knots;j++)
        SetVKnot(j, nurbs_patch->GetVKnot(j));
    for (gfInt k = 0;k < num_w_knots;k++)
        SetWKnot(k, nurbs_curve->GetKnot(k));

    for (gfInt k = 0;k < num_w_points;k++)
    {
        // control point and weight of curve
        gfPoint3 * curr_cpoint = nurbs_curve->GetControlPoint(k);
        gfDouble curr_cweight = nurbs_curve->GetWeight(k);

        for (gfInt i = 0; i < num_u_points;i++)
            for (gfInt j = 0; j < num_v_points;j++)
            {
                gfPoint3 * pt = nurbs_patch->GetControlPoint(i, j);
                gfPoint3 * new_pt = GetControlPoint(i, j, k);
                new_pt->Set(pt->X() + curr_cpoint->X() - pt1->X(), pt->Y() + curr_cpoint->Y() - pt1->Y(), pt->Z() + curr_cpoint->Z() - pt1->Z());
                SetWeight(i, j, k, nurbs_curve->GetWeight(k) * nurbs_patch->GetWeight(i, j));
            }
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitRotational(gfPoint3* center, gfVector3* axis, gfObjNURBSPatch* nurbs_patch)
{
    Clear();

    // retrieve data from given curve and patch
    gfInt u_degree = nurbs_patch->GetUDegree();
    gfInt num_u_points = nurbs_patch->GetNumUControlPoints();
    gfInt num_u_knots = nurbs_patch->GetNumUKnots();

    gfInt v_degree = nurbs_patch->GetVDegree();
    gfInt num_v_points = nurbs_patch->GetNumVControlPoints();
    gfInt num_v_knots = nurbs_patch->GetNumVKnots();

    gfInt w_degree = 2;
    gfInt num_w_points = 9;

    // set parameters of box
    SetParams(u_degree, v_degree, w_degree, num_u_points, num_v_points, num_w_points);

    // set u and v and w knots of box
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, nurbs_patch->GetUKnot(i));
    for (gfInt j = 0;j < num_v_knots;j++)
        SetVKnot(j, nurbs_patch->GetVKnot(j));
    SetWKnot(0, 0.0);SetWKnot(1, 0.0);SetWKnot(2, 0.0);
    SetWKnot(3, 0.25);SetWKnot(4, 0.25);
    SetWKnot(5, 0.5);SetWKnot(6, 0.5);
    SetWKnot(7, 0.75);SetWKnot(8, 0.75);
    SetWKnot(9, 1.0);SetWKnot(10, 1.0);SetWKnot(11, 1.0);

    
    gfPoint3* curr_center = gfCreatePoint3();
    gfVector3* aux_vector = gfCreateVector3();

    for (gfInt i = 0;i < num_u_points;i++)
    {
        for (gfInt j = 0; j < num_v_points;j++)
        {
            // here find projection of current point on axis given by point and vector
            gfPoint3* curr_point = nurbs_patch->GetControlPoint(i, j);

            gfSubstractPoints3(curr_point, center, aux_vector);
            gfDouble scl_prod = gfScalarProduct3(axis, aux_vector);
            gfCombinePointVector3(center, scl_prod, axis, curr_center);

            // init auxiliary curve as circle and get its data
            gfObjNURBSCurve* aux_curve = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
            aux_curve->InitCircle(curr_center, curr_point, axis);

            // fill patch with data from auxiliary curve
            
            for (gfInt k = 0; k < 9;k++)
            {
                gfPoint3 * pt = GetControlPoint(i, j, k);
                gfPoint3 * pt2 = aux_curve->GetControlPoint(k);
                pt->Set(pt2->X(), pt2->Y(), pt2->Z());
                SetWeight(i, j, k, aux_curve->GetWeight(k) * nurbs_patch->GetWeight(i, j));
            }

            gfDeleteGeomObject((gfObjGeomObject*) aux_curve);
        }
    }

    gfDeletePoint3(curr_center);
    gfDeleteVector3(aux_vector);

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitRotational(gfPoint3* center, gfVector3* axis, gfObjNURBSCurve* nurbs_curve)
{
    Clear();
    axis->Normalize();

    // retrieve data from given curve and patch
    gfInt u_degree = nurbs_curve->GetDegree();
    gfInt num_u_points = nurbs_curve->GetNumControlPoints();
    gfInt num_u_knots = nurbs_curve->GetNumKnots();

    gfInt v_degree = 2;
    gfInt num_v_points = 3;

    gfInt w_degree = 2;
    gfInt num_w_points = 3;

    // set parameters of box
    SetParams(u_degree, v_degree, w_degree, num_u_points, num_v_points, num_w_points);

    // set u and v and w knots of box
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, nurbs_curve->GetKnot(i));
    SetVKnot(0, 0.0);SetVKnot(1, 0.0);SetVKnot(2, 0.0);
    SetVKnot(3, 1.0);SetVKnot(4, 1.0);SetVKnot(5, 1.0);
    SetWKnot(0, 0.0);SetWKnot(1, 0.0);SetWKnot(2, 0.0);
    SetWKnot(3, 1.0);SetWKnot(4, 1.0);SetWKnot(5, 1.0);

    gfVector3* aux_vector = gfCreateVector3();
    gfVector3* e1 = gfCreateVector3();
    gfVector3* e2 = gfCreateVector3();
    gfPoint3* curr_center = gfCreatePoint3();

    gfDouble a = gfSquareRoot(2.0) / 2.0;
    for (gfInt i = 0;i < num_u_points;i++)
    {
        // here find projection of current point on axis given by point and vector
        gfPoint3* curr_point = nurbs_curve->GetControlPoint(i);

        gfSubstractPoints3(curr_point, center, aux_vector);
        gfDouble scl_prod = gfScalarProduct3(axis, aux_vector);
        gfCombinePointVector3(center, scl_prod, axis, curr_center);

        gfSubstractPoints3(curr_point, curr_center, e1);
        gfVectorProduct3(e1, axis, e2);
        e2->SetLength(e1->Length());

        gfPoint3 * pt = GetControlPoint(i, 0, 0);
        pt->Set(curr_center->X() + e1->X(), curr_center->Y() + e1->Y(), curr_center->Z() + e1->Z());
        SetWeight(i, 0, 0, 1.0 * nurbs_curve->GetWeight(i));

        pt = GetControlPoint(i, 1, 0);
        pt->Set(curr_center->X() + e1->X() - e2->X(), curr_center->Y() + e1->Y() - e2->Y(), curr_center->Z() + e1->Z() - e2->Z());
        SetWeight(i, 1, 0, 1.0 * nurbs_curve->GetWeight(i));

        pt = GetControlPoint(i, 2, 0);
        pt->Set(curr_center->X() - e2->X(), curr_center->Y() - e2->Y(), curr_center->Z() - e2->Z());
        SetWeight(i, 2, 0, 2.0 * nurbs_curve->GetWeight(i));

        pt = GetControlPoint(i, 1, 1);
        pt->Set(curr_center->X(), curr_center->Y(), curr_center->Z());
        SetWeight(i, 1, 1, 1.0 * nurbs_curve->GetWeight(i));

        pt = GetControlPoint(i, 0, 1);
        pt->Set(curr_center->X() + e1->X() + e2->X(), curr_center->Y() + e1->Y() + e2->Y(), curr_center->Z() + e1->Z() + e2->Z());
        SetWeight(i, 0, 1, 1.0 * nurbs_curve->GetWeight(i));

        pt = GetControlPoint(i, 0, 2);
        pt->Set(curr_center->X() + e2->X(), curr_center->Y() + e2->Y(), curr_center->Z() + e2->Z());
        SetWeight(i, 0, 2, 2.0 * nurbs_curve->GetWeight(i));

        pt = GetControlPoint(i, 2, 1);
        pt->Set(curr_center->X() - e1->X() - e2->X(), curr_center->Y() - e1->Y() - e2->Y(), curr_center->Z() - e1->Z() - e2->Z());
        SetWeight(i, 2, 1, 2.0 * nurbs_curve->GetWeight(i));

        pt = GetControlPoint(i, 1, 2);
        pt->Set(curr_center->X() - e1->X() + e2->X(), curr_center->Y() - e1->Y() + e2->Y(), curr_center->Z() - e1->Z() + e2->Z());
        SetWeight(i, 1, 2, 2.0 * nurbs_curve->GetWeight(i));

        pt = GetControlPoint(i, 2, 2);
        pt->Set(curr_center->X() - e1->X(), curr_center->Y() - e1->Y(), curr_center->Z() - e1->Z());
        SetWeight(i, 2, 2, 4.0 * nurbs_curve->GetWeight(i));
    }

    gfDeleteVector3(aux_vector);
    gfDeleteVector3(e1);
    gfDeleteVector3(e2);
    gfDeletePoint3(curr_center);

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitRotational(gfPoint3* center, gfVector3* axis, gfDouble angle, gfObjNURBSPatch* nurbs_patch)
{
    Clear();

    // retrieve data from given curve and patch
    gfInt u_degree = nurbs_patch->GetUDegree();
    gfInt num_u_points = nurbs_patch->GetNumUControlPoints();
    gfInt num_u_knots = nurbs_patch->GetNumUKnots();

    gfInt v_degree = nurbs_patch->GetVDegree();
    gfInt num_v_points = nurbs_patch->GetNumVControlPoints();
    gfInt num_v_knots = nurbs_patch->GetNumVKnots();

    gfObjNURBSCurve* aux_curve = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
    gfPoint3 some(102.5, 201.4, 301.3);
    aux_curve->InitCircularArc(center, &some, axis, angle);
    gfInt w_degree = aux_curve->GetDegree();
    gfInt num_w_points = aux_curve->GetNumControlPoints();
    gfInt num_w_knots = aux_curve->GetNumKnots();

    // set parameters of box
    SetParams(u_degree, v_degree, w_degree, num_u_points, num_v_points, num_w_points);

    // set u and v and w knots of box
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, nurbs_patch->GetUKnot(i));
    for (gfInt j = 0;j < num_v_knots;j++)
        SetVKnot(j, nurbs_patch->GetVKnot(j));
    // w knots based on knots of circular arc
    for (gfInt k = 0;k < num_w_knots;k++)
        SetWKnot(k, aux_curve->GetKnot(k));

    gfPoint3* curr_center = gfCreatePoint3();
    gfVector3* aux_vector = gfCreateVector3();

    for (gfInt i = 0;i < num_u_points;i++)
    {
        for (gfInt j = 0; j < num_v_points;j++)
        {
            // here find projection of current point on axis given by point and vector
            gfPoint3* curr_point = nurbs_patch->GetControlPoint(i, j);

            gfSubstractPoints3(curr_point, center, aux_vector);
            gfDouble scl_prod = gfScalarProduct3(axis, aux_vector);
            gfCombinePointVector3(center, scl_prod, axis, curr_center);

            // init auxiliary curve as circle and get its data
            aux_curve->InitCircularArc(curr_center, curr_point, axis, angle);

            // fill patch with data from auxiliary curve
            
            for (gfInt k = 0; k < aux_curve->GetNumControlPoints();k++)
            {
                gfPoint3 * pt = GetControlPoint(i, j, k);
                gfPoint3 * pt2 = aux_curve->GetControlPoint(k);
                pt->Set(pt2->X(), pt2->Y(), pt2->Z());
                SetWeight(i, j, k, aux_curve->GetWeight(k) * nurbs_patch->GetWeight(i, j));
            }
        }
    }

    gfDeleteGeomObject((gfObjGeomObject*) aux_curve);

    gfDeletePoint3(curr_center);
    gfDeleteVector3(aux_vector);

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitRotational(gfPoint3* center, gfVector3* axis, gfDouble angle, gfObjNURBSCurve* nurbs_curve)
{
    gfObjNURBSPatch* aux_patch = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
    aux_patch->InitRuledCurveToLine(nurbs_curve, center, axis);
    
    GFRESULT res = InitRotational(center, axis, angle, aux_patch);
    gfDeleteGeomObject(aux_patch);

    return res;
}

GFRESULT gfObjNURBSBox::InitProfile(gfDouble xmin, gfDouble xmax, gfInt num_u_points, 
                                      gfDouble ymin, gfDouble ymax, gfInt num_v_points,
                                      gfDouble zmin, gfDouble zmax, gfInt num_w_points)
{
    SetParams(3, 3, 3, num_u_points, num_v_points, num_w_points);

    gfInt num_u_knots = GetNumUKnots();
    gfInt num_v_knots = GetNumVKnots();
    gfInt num_w_knots = GetNumWKnots();

    for (gfInt i = 0;i <= m_u_degree;i++)
    {
        SetUKnot(i, 0);
    }
    for (gfInt i = m_u_degree+1;i < num_u_knots-m_u_degree-1;i++)
    {
        SetUKnot(i, (gfDouble) i-m_u_degree);
    }
    for (gfInt i = num_u_knots-m_u_degree-1;i < num_u_knots;i++)
    {
        SetUKnot(i, (gfDouble) num_u_knots-m_u_degree-2);
    }

    for (gfInt i = 0;i <= m_v_degree;i++)
    {
        SetVKnot(i, 0);
    }
    for (gfInt i = m_v_degree+1;i < num_v_knots-m_v_degree-1;i++)
    {
        SetVKnot(i, (gfDouble) i-m_v_degree);
    }
    for (gfInt i = num_v_knots-m_v_degree-1;i < num_v_knots;i++)
    {
        SetVKnot(i, (gfDouble) num_v_knots-m_v_degree-2);
    }

    for (gfInt i = 0;i <= m_w_degree;i++)
    {
        SetWKnot(i, 0);
    }
    for (gfInt i = m_w_degree+1;i < num_w_knots-m_w_degree-1;i++)
    {
        SetWKnot(i, (gfDouble) i-m_w_degree);
    }
    for (gfInt i = num_w_knots-m_w_degree-1;i < num_w_knots;i++)
    {
        SetWKnot(i, (gfDouble) num_w_knots-m_w_degree-2);
    }

    for (gfInt i = 0;i < num_u_points;i++)
        for (gfInt j = 0;j < num_v_points;j++)
            for (gfInt k = 0;k < num_w_points;k++)
            {
                gfPoint3* pt = GetControlPoint(i, j, k);

                pt->Set(xmin+(xmax-xmin)*i/(num_u_points-1)+gfRand(-0.5, 0.5),
                        ymin+(ymax-ymin)*j/(num_v_points-1)+gfRand(-0.5, 0.5),
                        zmin+(zmax-zmin)*k/(num_w_points-1)+gfRand(-0.5, 0.5));

                SetWeight(i, j, k, 1.0);
            }

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::InitRuledVolume(gfObjNURBSPatch* nurbs_patch1, gfObjNURBSPatch* nurbs_patch2)
{
    Clear();


    // retrieve data from given curve and patch
    gfInt u_degree = nurbs_patch1->GetUDegree();
    gfInt num_u_points = nurbs_patch1->GetNumUControlPoints();
    gfInt num_u_knots = nurbs_patch1->GetNumUKnots();

    gfInt v_degree = nurbs_patch2->GetVDegree();
    gfInt num_v_points = nurbs_patch2->GetNumVControlPoints();
    gfInt num_v_knots = nurbs_patch2->GetNumVKnots();

    // check if parameters are equal
    if (u_degree != v_degree)
        return GF_E_BADARG;
    if (num_u_points != num_v_points)
        return GF_E_BADARG;
    if (num_u_knots != num_v_knots)
        return GF_E_BADARG;

    // set parameters of patch
    SetParams(u_degree, v_degree, 1, num_u_points, num_v_points, 2);

    // set u and v knots of patch
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, nurbs_patch1->GetUKnot(i));
    for (gfInt j = 0;j < num_v_knots;j++)
        SetVKnot(j, nurbs_patch2->GetVKnot(j));
    SetWKnot(0, 0);
    SetWKnot(1, 0);
    SetWKnot(2, 1);
    SetWKnot(3, 1);

    for (gfInt k = 0;k < 2;k++)
    {
        for (gfInt i = 0; i < num_u_points;i++)
            for (gfInt j = 0; j < num_v_points;j++)
            {
                gfPoint3* pt;
                gfDouble weight;
                if (k == 0)
                {
                    pt = nurbs_patch1->GetControlPoint(i, j);
                    weight = nurbs_patch1->GetWeight(i, j);
                }
                else
                {
                    pt = nurbs_patch2->GetControlPoint(i, j);
                    weight = nurbs_patch2->GetWeight(i, j);
                }
                gfPoint3* new_pt = GetControlPoint(i, j, k);
                new_pt->Set(pt->X(), pt->Y(), pt->Z());
                SetWeight(i, j, k, weight);
            }
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::SetParams(gfInt u_degree, gfInt v_degree, gfInt w_degree, gfInt num_u_control_points, gfInt num_v_control_points, gfInt num_w_control_points)
{
    // create auxiliary arrays and fill them with old 
    std::vector<gfDouble> aux_weights;
    std::vector<gfPoint3 *> aux_control_points;
    aux_weights.resize(num_u_control_points * num_v_control_points * num_w_control_points);
    aux_control_points.resize(num_u_control_points * num_v_control_points * num_w_control_points);

    m_u_knot_vector.resize(u_degree + num_u_control_points + 1);
    m_v_knot_vector.resize(v_degree + num_v_control_points + 1);
    m_w_knot_vector.resize(w_degree + num_w_control_points + 1);
    m_u_degree = u_degree;
    m_v_degree = v_degree;
    m_w_degree = w_degree;

    gfInt umax = (m_num_u_control_points > num_u_control_points) ? m_num_u_control_points : num_u_control_points;
    gfInt vmax = (m_num_v_control_points > num_v_control_points) ? m_num_v_control_points : num_v_control_points;
    gfInt wmax = (m_num_w_control_points > num_w_control_points) ? m_num_w_control_points : num_w_control_points;

    for (gfInt i = 0;i < umax;i++)
    {
        for (gfInt j = 0;j < vmax;j++)
        {
            for (gfInt k = 0;k < wmax;k++)
            {
                // point in new & old
                if ((i < num_u_control_points && j < num_v_control_points && k < num_w_control_points) && (i < m_num_u_control_points && j < m_num_v_control_points && k < m_num_w_control_points))
                {
                    aux_weights[i + num_u_control_points * j + num_u_control_points * num_v_control_points * k] = GetWeight(i, j, k);
                    aux_control_points[i + num_u_control_points * j + num_u_control_points * num_v_control_points * k] = GetControlPoint(i, j, k);
                }

                // point in new & not in old
                if ((i < num_u_control_points && j < num_v_control_points && k < num_w_control_points) && !(i < m_num_u_control_points && j < m_num_v_control_points && k < m_num_w_control_points))
                {
                    aux_weights[i + num_u_control_points * j + num_u_control_points * num_v_control_points * k] = 1.0;
                    gfPoint3 * new_pt = gfCreatePoint3();
                    aux_control_points[i + num_u_control_points * j + num_u_control_points * num_v_control_points * k] = new_pt;
                }

                // point in old & not in new
                if (!(i < num_u_control_points && j < num_v_control_points && k < num_w_control_points) && (i < m_num_u_control_points && j < m_num_v_control_points && k < m_num_w_control_points))
                {
                    gfPoint3 * old_pt = GetControlPoint(i, j, k);
                    gfDeletePoint3(old_pt);
                }
            }
        }
    }

    m_num_u_control_points = num_u_control_points;
    m_num_v_control_points = num_v_control_points;
    m_num_w_control_points = num_w_control_points;

    m_control_points.resize(m_num_u_control_points * m_num_v_control_points * m_num_w_control_points);
    m_weights.resize(m_num_u_control_points * m_num_v_control_points * m_num_w_control_points);

    // copy from auxiliary arrays
    for (gfInt i = 0;i < m_num_u_control_points * m_num_v_control_points * m_num_w_control_points;i++)
    {
        m_control_points[i] = aux_control_points.at(i);
        m_weights[i] = aux_weights.at(i);
    }

    m_time_stamp++;

    return GF_S_OK;
}


gfInt gfObjNURBSBox::GetUDegree(void)
{
    return m_u_degree;
}

gfInt gfObjNURBSBox::GetVDegree(void)
{
    return m_v_degree;
}

gfInt gfObjNURBSBox::GetWDegree(void)
{
    return m_w_degree;
}

gfInt gfObjNURBSBox::GetNumUKnots(void)
{
    return (gfInt)m_u_knot_vector.size();
}

gfInt gfObjNURBSBox::GetNumVKnots(void)
{
    return (gfInt)m_v_knot_vector.size();
}

gfInt gfObjNURBSBox::GetNumWKnots(void)
{
    return (gfInt)m_w_knot_vector.size();
}

gfDouble gfObjNURBSBox::GetUKnot(gfInt index)
{
    return m_u_knot_vector.at(index);
}

gfDouble gfObjNURBSBox::GetVKnot(gfInt index)
{
    return m_v_knot_vector.at(index);
}

gfDouble gfObjNURBSBox::GetWKnot(gfInt index)
{
    return m_w_knot_vector.at(index);
}

GFRESULT gfObjNURBSBox::SetUKnot(gfInt index, gfDouble knot)
{
    m_time_stamp++;
    return m_u_knot_vector[index] = knot;
}

GFRESULT gfObjNURBSBox::SetVKnot(gfInt index, gfDouble knot)
{
    m_time_stamp++;
    return m_v_knot_vector[index] = knot;
}

GFRESULT gfObjNURBSBox::SetWKnot(gfInt index, gfDouble knot)
{
    m_time_stamp++;
    return m_w_knot_vector[index] = knot;
}

gfInt gfObjNURBSBox::GetNumUControlPoints(void)
{
    return m_num_u_control_points;
}

gfInt gfObjNURBSBox::GetNumVControlPoints(void)
{
    return m_num_v_control_points;
}

gfInt gfObjNURBSBox::GetNumWControlPoints(void)
{
    return m_num_w_control_points;
}

gfPoint3 * gfObjNURBSBox::GetControlPoint(gfInt u_index, gfInt v_index, gfInt w_index)
{
    return m_control_points.at(u_index + v_index * m_num_u_control_points + w_index * m_num_u_control_points * m_num_v_control_points);
}

gfDouble gfObjNURBSBox::GetWeight(gfInt u_index, gfInt v_index, gfInt w_index)
{
    return m_weights.at(u_index + v_index * m_num_u_control_points + w_index * m_num_u_control_points * m_num_v_control_points);
}

GFRESULT gfObjNURBSBox::SetWeight(gfInt u_index, gfInt v_index, gfInt w_index, gfDouble weight)
{
    m_time_stamp++;
    return m_weights[u_index + v_index * m_num_u_control_points + w_index * m_num_u_control_points * m_num_v_control_points] = weight;
}

GFRESULT gfObjNURBSBox::GetDomain(gfDouble* u_min, gfDouble* u_max, gfDouble* v_min, gfDouble* v_max, gfDouble* w_min, gfDouble* w_max)
{
    *u_min = GetUKnot(GetUDegree());
    *u_max = GetUKnot(GetNumUControlPoints());
    *v_min = GetVKnot(GetVDegree());
    *v_max = GetVKnot(GetNumVControlPoints());
    *w_min = GetWKnot(GetWDegree());
    *w_max = GetWKnot(GetNumWControlPoints());

    return GF_S_OK;
}

gfObjNURBSPatch* gfObjNURBSBox::GetUVIsopatch(gfDouble w_param)
{
    gfDouble a = GetWKnot(m_w_degree);
    gfDouble b = GetWKnot(GetNumWControlPoints());
    if (gfIsZero(b - a))
        return NULL;
    if (w_param >= b || w_param < a)
        return NULL;

    gfObjNURBSPatch* ret = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
    ret->SetParams(m_u_degree, m_v_degree, m_num_u_control_points, m_num_v_control_points);

    // fill knots
    gfInt num_u_knots = GetNumUKnots();
    for (gfInt i = 0;i < num_u_knots;i++)
    {
        ret->SetUKnot(i, GetUKnot(i));
    }
    gfInt num_v_knots = GetNumVKnots();
    for (gfInt i = 0;i < num_v_knots;i++)
    {
        ret->SetVKnot(i, GetVKnot(i));
    }

    // prepare basis
    gfInt o = GetNumWControlPoints() - 1;
    gfInt span = gfGetBSplineKnotSpan(o, m_w_degree, &m_w_knot_vector, w_param);
    std::vector<gfDouble> basis;
    gfGetBSplineAllValues(m_w_degree, &m_w_knot_vector, span, w_param, &basis);

    for (gfInt i = 0;i < m_num_u_control_points;i++)
        for (gfInt j = 0;j < m_num_v_control_points;j++)
        {
            gfPoint3 * pt = ret->GetControlPoint(i, j);

            gfDouble ax = 0;
            gfDouble ay = 0;
            gfDouble az = 0;
            gfDouble aw = 0;
            for (gfInt k = 0;k <= m_w_degree;k++)
            {
                gfInt index = span - m_w_degree + k;
                ax += basis.at(k) * GetWeight(i, j, index) * GetControlPoint(i, j, index)->X();
                ay += basis.at(k) * GetWeight(i, j, index) * GetControlPoint(i, j, index)->Y();
                az += basis.at(k) * GetWeight(i, j, index) * GetControlPoint(i, j, index)->Z();
                aw += basis.at(k) * GetWeight(i, j, index);
            }

            pt->Set(ax / aw, ay / aw, az / aw);
            ret->SetWeight(i, j, aw);
        }

    return ret;
}

gfBool gfObjNURBSBox::IsBezier(void)
{
    // check parameters of patch
    if (GetUDegree() != (GetNumUControlPoints() - 1))
        return false;
    if (GetVDegree() != (GetNumVControlPoints() - 1))
        return false;
    if (GetWDegree() != (GetNumWControlPoints() - 1))
        return false;

    // get num of knots
    gfInt num_u_knots = GetNumUKnots();
    gfInt num_v_knots = GetNumVKnots();
    gfInt num_w_knots = GetNumWKnots();
    if (num_u_knots == 0)
        return true;
    if (num_v_knots == 0)
        return true;
    if (num_w_knots == 0)
        return true;

    // check u knots
    for (gfInt i = 0;i < num_u_knots / 2;i++)
    {
        if (GetUKnot(i) != GetUKnot(0))
            return false;
    }
    for (gfInt i = num_u_knots / 2;i < num_u_knots ;i++)
    {
        if (GetUKnot(i) != GetUKnot(num_u_knots - 1))
            return false;
    }

    // check v knots
    for (gfInt i = 0;i < num_v_knots / 2;i++)
    {
        if (GetVKnot(i) != GetVKnot(0))
            return false;
    }
    for (gfInt i = num_v_knots / 2;i < num_v_knots ;i++)
    {
        if (GetVKnot(i) != GetVKnot(num_v_knots - 1))
            return false;
    }

    // check w knots
    for (gfInt i = 0;i < num_w_knots / 2;i++)
    {
        if (GetWKnot(i) != GetWKnot(0))
            return false;
    }
    for (gfInt i = num_w_knots / 2;i < num_w_knots ;i++)
    {
        if (GetWKnot(i) != GetWKnot(num_w_knots - 1))
            return false;
    }

    return true;
}

gfPoint3* gfObjNURBSBox::GetVolumePoint(gfDouble u, gfDouble v, gfDouble w)
{

    gfInt n = GetNumUControlPoints() - 1;
    gfInt m = GetNumVControlPoints() - 1;
    gfInt o = GetNumWControlPoints() - 1;

    // current u parameter
    gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
    std::vector<gfDouble> ubasis;
    gfGetBSplineAllValues(m_u_degree, &m_u_knot_vector, uspan, u, &ubasis);

    // current v parameter     
    gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
    std::vector<gfDouble> vbasis;
    gfGetBSplineAllValues(m_v_degree, &m_v_knot_vector, vspan, v, &vbasis);

    // current v parameter     
    gfInt wspan = gfGetBSplineKnotSpan(o, m_w_degree, &m_w_knot_vector, w);
    std::vector<gfDouble> wbasis;
    gfGetBSplineAllValues(m_w_degree, &m_w_knot_vector, wspan, w, &wbasis);

    gfDouble ax = 0;
    gfDouble ay = 0;
    gfDouble az = 0;
    gfDouble aw = 0;

    for (gfInt i = 0;i <= m_u_degree;i++)
    {
        gfInt uindex = uspan - m_u_degree + i;

        for (gfInt j = 0;j <= m_v_degree;j++)
        {
            gfInt vindex = vspan - m_v_degree + j;

            for (gfInt k = 0;k <= m_w_degree;k++)
            {
                gfInt windex = wspan - m_w_degree + k;

                gfDouble aux = ubasis.at(i) * vbasis.at(j) * wbasis.at(k) * GetWeight(uindex, vindex, windex);
                ax += aux * GetControlPoint(uindex, vindex, windex)->X();
                ay += aux * GetControlPoint(uindex, vindex, windex)->Y();
                az += aux * GetControlPoint(uindex, vindex, windex)->Z();
                aw += aux;
            }
        }
    }

    gfPoint3* result = gfCreatePoint3();
    result->Set(ax / aw, ay / aw, az / aw);

    return result;
}

gfVector3* gfObjNURBSBox::GetPatchUDer(gfDouble u, gfDouble v, gfDouble w)
{
    gfInt n = GetNumUControlPoints()  - 1;
    gfInt m = GetNumVControlPoints()  - 1;
    gfInt o = GetNumWControlPoints()  - 1;

    // current u basis functions
    gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
    gfAux2DArray<gfDouble> ubasis(2, m_u_degree + 1);
    gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 1, &ubasis);

    // current v basis functions    
    gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
    gfAux2DArray<gfDouble> vbasis(1, m_v_degree + 1);
    gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 0, &vbasis);

    // current w basis functions    
    gfInt wspan = gfGetBSplineKnotSpan(o, m_w_degree, &m_w_knot_vector, w);
    gfAux2DArray<gfDouble> wbasis(1, m_w_degree + 1);
    gfGetBSplineAllDers(m_w_degree, &m_w_knot_vector, wspan, w, 0, &wbasis);

    gfDouble ax = 0.0;
    gfDouble ay = 0.0;
    gfDouble az = 0.0;
    gfDouble aw = 0.0;

    gfDouble dux = 0.0;
    gfDouble duy = 0.0;
    gfDouble duz = 0.0;
    gfDouble duw = 0.0;

    for (gfInt i = 0;i <= m_u_degree;i++)
    {
        gfInt uindex = uspan - m_u_degree + i;

        for (gfInt j = 0;j <= m_v_degree;j++)
        {
            gfInt vindex = vspan - m_v_degree + j;

            for (gfInt k = 0;k <= m_w_degree;k++)
            {
                gfInt windex = wspan - m_w_degree + k;

                gfDouble aux = ubasis.at(0, i) * vbasis.at(0, j) * wbasis.at(0, k) * GetWeight(uindex, vindex, windex);
                ax += aux * GetControlPoint(uindex, vindex, windex)->X();
                ay += aux * GetControlPoint(uindex, vindex, windex)->Y();
                az += aux * GetControlPoint(uindex, vindex, windex)->Z();
                aw += aux;

                gfDouble aux1 = ubasis.at(1, i) * vbasis.at(0, j) * wbasis.at(0, k) * GetWeight(uindex, vindex, windex);
                dux += aux1 * GetControlPoint(uindex, vindex, windex)->X();
                duy += aux1 * GetControlPoint(uindex, vindex, windex)->Y();
                duz += aux1 * GetControlPoint(uindex, vindex, windex)->Z();
                duw += aux1;
            }
        }
    }

    gfDouble fdux = (dux * aw - ax * duw) / aw / aw;
    gfDouble fduy = (duy * aw - ay * duw) / aw / aw;
    gfDouble fduz = (duz * aw - az * duw) / aw / aw;

    gfVector3* result = gfCreateVector3();
    result->Set(fdux, fduy, fduz);

    return result;
}

gfVector3* gfObjNURBSBox::GetPatchVDer(gfDouble u, gfDouble v, gfDouble w)
{
    gfInt n = GetNumUControlPoints()  - 1;
    gfInt m = GetNumVControlPoints()  - 1;
    gfInt o = GetNumWControlPoints()  - 1;

    // current u basis functions
    gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
    gfAux2DArray<gfDouble> ubasis(1, m_u_degree + 1);
    gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 0, &ubasis);

    // current v basis functions    
    gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
    gfAux2DArray<gfDouble> vbasis(2, m_v_degree + 1);
    gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 1, &vbasis);

    // current w basis functions    
    gfInt wspan = gfGetBSplineKnotSpan(o, m_w_degree, &m_w_knot_vector, w);
    gfAux2DArray<gfDouble> wbasis(1, m_w_degree + 1);
    gfGetBSplineAllDers(m_w_degree, &m_w_knot_vector, wspan, w, 0, &wbasis);

    gfDouble ax = 0.0;
    gfDouble ay = 0.0;
    gfDouble az = 0.0;
    gfDouble aw = 0.0;

    gfDouble dvx = 0.0;
    gfDouble dvy = 0.0;
    gfDouble dvz = 0.0;
    gfDouble dvw = 0.0;

    for (gfInt i = 0;i <= m_u_degree;i++)
    {
        gfInt uindex = uspan - m_u_degree + i;

        for (gfInt j = 0;j <= m_v_degree;j++)
        {
            gfInt vindex = vspan - m_v_degree + j;

            for (gfInt k = 0;k <= m_w_degree;k++)
            {
                gfInt windex = wspan - m_w_degree + k;

                gfDouble aux = ubasis.at(0, i) * vbasis.at(0, j) * wbasis.at(0, k) * GetWeight(uindex, vindex, windex);
                ax += aux * GetControlPoint(uindex, vindex, windex)->X();
                ay += aux * GetControlPoint(uindex, vindex, windex)->Y();
                az += aux * GetControlPoint(uindex, vindex, windex)->Z();
                aw += aux;

                gfDouble aux1 = ubasis.at(0, i) * vbasis.at(1, j) * wbasis.at(0, k) * GetWeight(uindex, vindex, windex);
                dvx += aux1 * GetControlPoint(uindex, vindex, windex)->X();
                dvy += aux1 * GetControlPoint(uindex, vindex, windex)->Y();
                dvz += aux1 * GetControlPoint(uindex, vindex, windex)->Z();
                dvw += aux1;
            }
        }
    }

    gfDouble fdvx = (dvx * aw - ax * dvw) / aw / aw;
    gfDouble fdvy = (dvy * aw - ay * dvw) / aw / aw;
    gfDouble fdvz = (dvz * aw - az * dvw) / aw / aw;

    gfVector3* result = gfCreateVector3();
    result->Set(fdvx, fdvy, fdvz);

    return result;
}

gfVector3* gfObjNURBSBox::GetPatchWDer(gfDouble u, gfDouble v, gfDouble w)
{
    gfInt n = GetNumUControlPoints()  - 1;
    gfInt m = GetNumVControlPoints()  - 1;
    gfInt o = GetNumWControlPoints()  - 1;

    // current u basis functions
    gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
    gfAux2DArray<gfDouble> ubasis(1, m_u_degree + 1);
    gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 0, &ubasis);

    // current v basis functions    
    gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
    gfAux2DArray<gfDouble> vbasis(1, m_v_degree + 1);
    gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 0, &vbasis);

    // current w basis functions    
    gfInt wspan = gfGetBSplineKnotSpan(o, m_w_degree, &m_w_knot_vector, w);
    gfAux2DArray<gfDouble> wbasis(2, m_w_degree + 1);
    gfGetBSplineAllDers(m_w_degree, &m_w_knot_vector, wspan, w, 1, &wbasis);

    gfDouble ax = 0.0;
    gfDouble ay = 0.0;
    gfDouble az = 0.0;
    gfDouble aw = 0.0;

    gfDouble dwx = 0.0;
    gfDouble dwy = 0.0;
    gfDouble dwz = 0.0;
    gfDouble dww = 0.0;

    for (gfInt i = 0;i <= m_u_degree;i++)
    {
        gfInt uindex = uspan - m_u_degree + i;

        for (gfInt j = 0;j <= m_v_degree;j++)
        {
            gfInt vindex = vspan - m_v_degree + j;

            for (gfInt k = 0;k <= m_w_degree;k++)
            {
                gfInt windex = wspan - m_w_degree + k;

                gfDouble aux = ubasis.at(0, i) * vbasis.at(0, j) * wbasis.at(0, k) * GetWeight(uindex, vindex, windex);
                ax += aux * GetControlPoint(uindex, vindex, windex)->X();
                ay += aux * GetControlPoint(uindex, vindex, windex)->Y();
                az += aux * GetControlPoint(uindex, vindex, windex)->Z();
                aw += aux;

                gfDouble aux1 = ubasis.at(0, i) * vbasis.at(0, j) * wbasis.at(1, k) * GetWeight(uindex, vindex, windex);
                dwx += aux1 * GetControlPoint(uindex, vindex, windex)->X();
                dwy += aux1 * GetControlPoint(uindex, vindex, windex)->Y();
                dwz += aux1 * GetControlPoint(uindex, vindex, windex)->Z();
                dww += aux1;
            }
        }
    }

    gfDouble fdwx = (dwx * aw - ax * dww) / aw / aw;
    gfDouble fdwy = (dwy * aw - ay * dww) / aw / aw;
    gfDouble fdwz = (dwz * aw - az * dww) / aw / aw;

    gfVector3* result = gfCreateVector3();
    result->Set(fdwx, fdwy, fdwz);

    return result;
}

gfVector3* gfObjNURBSBox::GetNormal(gfDouble u, gfDouble v, gfDouble w)
{
    gfInt n = GetNumUControlPoints() - 1;
    gfInt m = GetNumVControlPoints() - 1;
    gfInt o = GetNumWControlPoints() - 1;

    // current u basis
    gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
    gfAux2DArray<gfDouble> ubasis(2, m_u_degree + 1);
    gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 1, &ubasis);

    // current v basis     
    gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
    gfAux2DArray<gfDouble> vbasis(2, m_v_degree + 1);
    gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 1, &vbasis);

    // current w basis     
    gfInt wspan = gfGetBSplineKnotSpan(o, m_w_degree, &m_w_knot_vector, w);
    gfAux2DArray<gfDouble> wbasis(2, m_w_degree + 1);
    gfGetBSplineAllDers(m_w_degree, &m_w_knot_vector, wspan, w, 1, &wbasis);

    gfDouble ax = 0.0;
    gfDouble ay = 0.0;
    gfDouble az = 0.0;
    gfDouble aw = 0.0;

    gfDouble dux = 0.0;
    gfDouble duy = 0.0;
    gfDouble duz = 0.0;
    gfDouble duw = 0.0;

    gfDouble dvx = 0.0;
    gfDouble dvy = 0.0;
    gfDouble dvz = 0.0;
    gfDouble dvw = 0.0;

    gfDouble dwx = 0.0;
    gfDouble dwy = 0.0;
    gfDouble dwz = 0.0;
    gfDouble dww = 0.0;

    for (gfInt i = 0;i <= m_u_degree;i++)
    {
        gfInt uindex = uspan - m_u_degree + i;

        for (gfInt j = 0;j <= m_v_degree;j++)
        {
            gfInt vindex = vspan - m_v_degree + j;

            for (gfInt k = 0;k <= m_w_degree;k++)
            {
                gfInt windex = wspan - m_w_degree + k;

                gfDouble aux = ubasis.at(0, i) * vbasis.at(0, j) * wbasis.at(0, k) * GetWeight(uindex, vindex, windex);
                ax += aux * GetControlPoint(uindex, vindex, windex)->X();
                ay += aux * GetControlPoint(uindex, vindex, windex)->Y();
                az += aux * GetControlPoint(uindex, vindex, windex)->Z();
                aw += aux;

                gfDouble aux1 = ubasis.at(1, i) * vbasis.at(0, j) * wbasis.at(0, k) * GetWeight(uindex, vindex, windex);
                dux += aux1 * GetControlPoint(uindex, vindex, windex)->X();
                duy += aux1 * GetControlPoint(uindex, vindex, windex)->Y();
                duz += aux1 * GetControlPoint(uindex, vindex, windex)->Z();
                duw += aux1;

                gfDouble aux2 = ubasis.at(0, i) * vbasis.at(1, j) * wbasis.at(0, k) * GetWeight(uindex, vindex, windex);
                dvx += aux2 * GetControlPoint(uindex, vindex, windex)->X();
                dvy += aux2 * GetControlPoint(uindex, vindex, windex)->Y();
                dvz += aux2 * GetControlPoint(uindex, vindex, windex)->Z();
                dvw += aux2;

                gfDouble aux3 = ubasis.at(0, i) * vbasis.at(0, j) * wbasis.at(1, k) * GetWeight(uindex, vindex, windex);
                dwx += aux3 * GetControlPoint(uindex, vindex, windex)->X();
                dwy += aux3 * GetControlPoint(uindex, vindex, windex)->Y();
                dwz += aux3 * GetControlPoint(uindex, vindex, windex)->Z();
                dww += aux3;
            }
        }
    }

    gfDouble fdux = (dux * aw - ax * duw) / aw / aw;
    gfDouble fduy = (duy * aw - ay * duw) / aw / aw;
    gfDouble fduz = (duz * aw - az * duw) / aw / aw;

    gfDouble fdvx = (dvx * aw - ax * dvw) / aw / aw;
    gfDouble fdvy = (dvy * aw - ay * dvw) / aw / aw;
    gfDouble fdvz = (dvz * aw - az * dvw) / aw / aw;

    gfDouble fdwx = (dwx * aw - ax * dww) / aw / aw;
    gfDouble fdwy = (dwy * aw - ay * dww) / aw / aw;
    gfDouble fdwz = (dwz * aw - az * dww) / aw / aw;

    gfVector3* resultu = gfCreateVector3();
    gfVector3* resultv = gfCreateVector3();
    gfVector3* resultw = gfCreateVector3();
    gfDouble a, b, c, d, e, f;
    GetDomain(&a, &b, &c, &d, &e, &f);

    resultu->Set((fdwy*fdvz-fdwz*fdvy),
                 (-fdwx*fdvz+fdwz*fdvx),
                 (fdwx*fdvy-fdwy*fdvx));
    resultv->Set((fdwy*fduz-fdwz*fduy),
                 (-fdwx*fduz+fdwz*fdux),
                 (fdwx*fduy-fdwy*fdux));
    resultw->Set((fdvy*fduz-fdvz*fduy),
                 (-fdvx*fduz+fdvz*fdux),
                 (fdvx*fduy-fdvy*fdux));

    gfDouble jacob = fdux*fdvy*fdwz+fduy*fdvz*fdwx+fduz*fdvx*fdwy-
                     fduz*fdvy*fdwx-fduy*fdvx*fdwz-fdux*fdvz*fdwy;

    gfVector3* result = gfCreateVector3();
    if (gfIsZero(u-a) || gfIsZero(u-b))
        result->Set(resultu->X(), resultu->Y(), resultu->Z());
    else if (gfIsZero(v-c) || gfIsZero(v-d))
        result->Set(resultv->X(), resultv->Y(), resultv->Z());
    else if (gfIsZero(w-e) || gfIsZero(w-f))
        result->Set(resultw->X(), resultw->Y(), resultw->Z());
    else if (gfIsZero(jacob))
    {
        if (!gfIsZero(resultu->Length()))
            result->Set(resultu->X(), resultu->Y(), resultu->Z());
        else if (!gfIsZero(resultv->Length()))
            result->Set(resultv->X(), resultv->Y(), resultv->Z());
        else if (!gfIsZero(resultw->Length()))
            result->Set(resultw->X(), resultw->Y(), resultw->Z());
        else
            result->Set(0.0, 0.0, 0.0);
    }
    else
    {
        // TODO: normal is linear combination of boundary normals
        result->Set(0.0, 0.0, 0.0);
    }

    gfDeleteVector3(resultu);
    gfDeleteVector3(resultv);
    gfDeleteVector3(resultw);

    result->Normalize();

    return result;
}

GFRESULT gfObjNURBSBox::DecomposeToBezierSegments(gfAux3DArray<gfObjNURBSBox*>* segments)
{
    gfInt n = GetNumUControlPoints() - 1;
    gfInt m = GetNumVControlPoints() - 1;
    gfInt o = GetNumWControlPoints() - 1;

    // first decompose in u direction
    gfInt a = m_u_degree;
    gfInt b = m_u_degree + 1;
    gfInt nb = 0;
    gfInt i, j, k;
    gfAux4DArray<gfPoint4> Qw;
    Qw.resize(n, m_u_degree + 1, m+1, o+1);
    std::vector<gfDouble> alphas;
    alphas.resize(m_u_degree + 1);

    for (i = 0;i <= m_u_degree;i++)
    {
        for (gfInt row = 0;row <= m;row++)
        {
            for (gfInt dep = 0;dep <= o;dep++)
            {
                gfPoint3* pt = GetControlPoint(i, row, dep);
                gfDouble w = GetWeight(i, row, dep);
                Qw.Set(nb, i, row, dep, gfPoint4(w*pt->X(), w*pt->Y(), w*pt->Z(), w));
            }
        }
    }

    while (b < n + m_u_degree + 1)
    {
        i = b;
        while (b < (n + m_u_degree + 1) && m_u_knot_vector.at(b+1) == m_u_knot_vector.at(b))
            b++;
        gfInt mult = b-i+1;
        if (mult < m_u_degree)
        {
            gfDouble numer = m_u_knot_vector.at(b) - m_u_knot_vector.at(a);
            for (j = m_u_degree;j > mult;j--)
                alphas[j-mult-1] = numer/(m_u_knot_vector.at(a+j) - m_u_knot_vector.at(a));
            gfInt r = m_u_degree - mult;
            for (j = 1;j < r;j++)
            {
                gfInt save = r-j;
                gfInt s = mult+j;
                for  (k = m_u_degree;k >= s;k--)
                {
                    gfDouble alpha = alphas.at(k-s);
                    for (gfInt row = 0;row <= m;row++)
                        for (gfInt dep = 0;dep <= o;dep++)
                            Qw.Set(nb, k, row, dep, gfPoint4(alpha*Qw.at(nb, k, row, dep).X()+(1.0-alpha)*Qw.at(nb, k-1, row, dep).X(),
                                                        alpha*Qw.at(nb, k, row, dep).Y()+(1.0-alpha)*Qw.at(nb, k-1, row, dep).Y(),
                                                        alpha*Qw.at(nb, k, row, dep).Z()+(1.0-alpha)*Qw.at(nb, k-1, row, dep).Z(),
                                                        alpha*Qw.at(nb, k, row, dep).W()+(1.0-alpha)*Qw.at(nb, k-1, row, dep).W()));
                }

                if (b < (n + m_u_degree + 1))
                {
                    for (gfInt row = 0;row <= m;row++)
                        for (gfInt dep = 0;dep <= o;dep++)
                            Qw.Set(nb + 1, save, row, dep, gfPoint4(Qw.at(nb, m_u_degree, row, dep).X(), 
                                                                    Qw.at(nb, m_u_degree, row, dep).Y(), 
                                                                    Qw.at(nb, m_u_degree, row, dep).Z(), 
                                                                    Qw.at(nb, m_u_degree, row, dep).W()));
                                                        }
            }
        }

        nb++;
        if (b < (n + m_u_degree + 1)) 
        {
            for (i = m_u_degree-mult;i <= m_u_degree;i++)
            {
                for (gfInt row = 0;row <= m;row++)
                    for (gfInt dep = 0;dep <= o;dep++)
                    {
                        gfPoint3* pt = GetControlPoint(b-m_u_degree+i, row, dep);
                        gfDouble w = GetWeight(b-m_u_degree+i, row, dep);
                        Qw.Set(nb, i, row, dep, gfPoint4(w*pt->X(), w*pt->Y(), w*pt->Z(), w));
                    }
            }
            a = b;
            b++;
        }
    }

    // copy points from segments to new array of control points
    gfAux3DArray<gfPoint4> Pw;
    gfInt num_u_seg = nb;
    Pw.resize(m_u_degree*num_u_seg+1, m+1, o+1);
    for (gfInt row = 0;row <= m;row++)
        for (gfInt dep = 0;dep <= o;dep++)
        {
            for (gfInt seg = 0;seg < num_u_seg;seg++)
            {
                for (i = 0;i < m_u_degree;i++)
                {
                    gfPoint4 pt = Qw.at(seg, i, row, dep);
                    Pw.Set(i+seg*m_u_degree, row, dep, pt);
                }
            }
            gfPoint4 pt = Qw.at(num_u_seg-1, m_u_degree, row, dep);
            Pw.Set(num_u_seg*m_u_degree, row, dep, pt);
        }

    // now decompose in v direction
    nb = 0;
    a = m_v_degree;
    b = m_v_degree + 1;
    n = m_u_degree*num_u_seg+1;
    Qw.resize(m, n+1, m_v_degree + 1, o+1);
    alphas.resize(m_v_degree + 1);

    for (i = 0;i <= m_v_degree;i++)
    {
        for (gfInt col = 0;col <= n;col++)
            for (gfInt dep = 0;dep <= o;dep++)
            {
                gfPoint4 pt = Pw.at(col, i, dep);
                Qw.Set(nb, col, i, dep, gfPoint4(pt.X(), pt.Y(), pt.Z(), pt.W()));
            }
    }

    while (b < (m + m_v_degree + 1))
    {
        i = b;
        while (b < (m + m_v_degree + 1) && m_v_knot_vector.at(b+1) == m_v_knot_vector.at(b))
            b++;
        gfInt mult = b-i+1;
        if (mult < m_v_degree)
        {
            gfDouble numer = m_v_knot_vector.at(b) - m_v_knot_vector.at(a);
            for (j = m_v_degree;j > mult;j--)
                alphas[j-mult-1] = numer/(m_v_knot_vector.at(a+j) - m_v_knot_vector.at(a));
            gfInt r = m_v_degree - mult;
            for (j = 1;j < r;j++)
            {
                gfInt save = r-j;
                gfInt s = mult+j;
                for  (k = m_v_degree;k >= s;k--)
                {
                    gfDouble alpha = alphas.at(k-s);
                    for (gfInt col = 0;col <= n;col++)
                        for (gfInt dep = 0;dep <= o;dep++)
                            Qw.Set(nb, col, k, dep, gfPoint4(alpha*Qw.at(nb, col, k, dep).X()+(1.0-alpha)*Qw.at(nb, col, k-1, dep).X(),
                                                            alpha*Qw.at(nb, col, k, dep).Y()+(1.0-alpha)*Qw.at(nb, col, k-1, dep).Y(),
                                                            alpha*Qw.at(nb, col, k, dep).Z()+(1.0-alpha)*Qw.at(nb, col, k-1, dep).Z(),
                                                            alpha*Qw.at(nb, col, k, dep).W()+(1.0-alpha)*Qw.at(nb, col, k-1, dep).W()));
                }

                if (b < (m + m_v_degree + 1))
                {
                    for (gfInt col = 0;col <= n;col++)
                        for (gfInt dep = 0;dep <= o;dep++)
                            Qw.Set(nb + 1, col, save, dep, gfPoint4(Qw.at(nb, col, m_v_degree, dep).X(), 
                                                                    Qw.at(nb, col, m_v_degree, dep).Y(), 
                                                                    Qw.at(nb, col, m_v_degree, dep).Z(), 
                                                                    Qw.at(nb, col, m_v_degree, dep).W()));
                }
            }
        }

        nb++;
        if (b < (m + m_v_degree + 1)) 
        {
            for (i = m_v_degree-mult;i <= m_v_degree;i++)
            {
                for (gfInt col = 0;col <= n;col++)
                    for (gfInt dep = 0;dep <= o;dep++)
                    {
                        gfPoint4 pt = Pw.at(col, b-m_v_degree+i, dep);
                        Qw.Set(nb, col, i, dep, gfPoint4(pt.X(), pt.Y(), pt.Z(), pt.W()));
                    }
            }
            a = b;
            b++;
        }
    }

    // copy points from segments to new array of control points
    gfInt num_v_seg = nb;
    Pw.resize(n+1, m_v_degree*num_v_seg+1, o+1);
    for (gfInt col = 0;col <= n;col++)
        for (gfInt dep = 0;dep <= o;dep++)
        {
            for (gfInt seg = 0;seg < num_v_seg;seg++)
            {
                for (i = 0;i < m_v_degree;i++)
                {
                    gfPoint4 pt = Qw.at(seg, col, i, dep);
                    Pw.Set(col, i+seg*m_v_degree, dep, pt);
                }
            }
            gfPoint4 pt = Qw.at(num_v_seg-1, col, m_v_degree, dep);
            Pw.Set(col, num_v_seg*m_v_degree, dep, pt);
        }

    // now decompose in w direction
    nb = 0;
    a = m_w_degree;
    b = m_w_degree + 1;
    m = m_v_degree*num_v_seg+1;
    Qw.resize(o, n+1, m+1, m_w_degree + 1);
    alphas.resize(m_w_degree + 1);

    for (i = 0;i <= m_w_degree;i++)
    {
        for (gfInt col = 0;col <= n;col++)
            for (gfInt row = 0;row <= m;row++)
            {
                gfPoint4 pt = Pw.at(col, row, i);
                Qw.Set(nb, col, row, i, gfPoint4(pt.X(), pt.Y(), pt.Z(), pt.W()));
            }
    }

    while (b < (o + m_w_degree + 1))
    {
        i = b;
        while (b < (o + m_w_degree + 1) && m_w_knot_vector.at(b+1) == m_w_knot_vector.at(b))
            b++;
        gfInt mult = b-i+1;
        if (mult < m_w_degree)
        {
            gfDouble numer = m_w_knot_vector.at(b) - m_w_knot_vector.at(a);
            for (j = m_w_degree;j > mult;j--)
                alphas[j-mult-1] = numer/(m_w_knot_vector.at(a+j) - m_w_knot_vector.at(a));
            gfInt r = m_w_degree - mult;
            for (j = 1;j < r;j++)
            {
                gfInt save = r-j;
                gfInt s = mult+j;
                for  (k = m_w_degree;k >= s;k--)
                {
                    gfDouble alpha = alphas.at(k-s);
                    for (gfInt col = 0;col <= n;col++)
                        for (gfInt row = 0;row <= m;row++)
                            Qw.Set(nb, col, row, k, gfPoint4(alpha*Qw.at(nb, col, row, k).X()+(1.0-alpha)*Qw.at(nb, col, row, k-1).X(),
                                                             alpha*Qw.at(nb, col, row, k).Y()+(1.0-alpha)*Qw.at(nb, col, row, k-1).Y(),
                                                             alpha*Qw.at(nb, col, row, k).Z()+(1.0-alpha)*Qw.at(nb, col, row, k-1).Z(),
                                                             alpha*Qw.at(nb, col, row, k).W()+(1.0-alpha)*Qw.at(nb, col, row, k-1).W()));
                }

                if (b < (o + m_w_degree + 1))
                {
                    for (gfInt col = 0;col <= n;col++)
                        for (gfInt row = 0;row <= m;row++)
                            Qw.Set(nb + 1, col, row, save, gfPoint4(Qw.at(nb, col, row, m_w_degree).X(), 
                                                                    Qw.at(nb, col, row, m_w_degree).Y(), 
                                                                    Qw.at(nb, col, row, m_w_degree).Z(), 
                                                                    Qw.at(nb, col, row, m_w_degree).W()));
                }
            }
        }

        nb++;
        if (b < (o + m_w_degree + 1)) 
        {
            for (i = m_w_degree-mult;i <= m_w_degree;i++)
            {
                for (gfInt col = 0;col <= n;col++)
                    for (gfInt row = 0;row <= m;row++)
                    {
                        gfPoint4 pt = Pw.at(col, row, b-m_w_degree+i);
                        Qw.Set(nb, col, row, i, gfPoint4(pt.X(), pt.Y(), pt.Z(), pt.W()));
                    }
            }
            a = b;
            b++;
        }
    }

    // get all different u knots
    std::vector<gfDouble> dif_u_knots;
    gfInt aa = Pw.GetISize();
    gfInt bb = m_u_knot_vector.size();
    for (i = m_u_degree;i <= m_num_u_control_points;i++)
    {
        gfInt num = (gfInt)dif_u_knots.size();
        gfBool diff = true;
        for (gfInt j = 0;j < num;j++)
        {
            if (m_u_knot_vector.at(i) == dif_u_knots.at(j))
                diff = false;
        }
        if (diff)
            dif_u_knots.push_back(m_u_knot_vector.at(i));
    }
    std::vector<gfDouble> dif_v_knots;
    for (i = m_v_degree;i <= m_num_v_control_points;i++)
    {
        gfInt num = (gfInt)dif_v_knots.size();
        gfBool diff = true;
        for (gfInt j = 0;j < num;j++)
        {
            if (m_v_knot_vector.at(i) == dif_v_knots.at(j))
                diff = false;
        }
        if (diff)
            dif_v_knots.push_back(m_v_knot_vector.at(i));
    }
    std::vector<gfDouble> dif_w_knots;
    for (i = m_w_degree;i <= m_num_w_control_points;i++)
    {
        gfInt num = (gfInt)dif_w_knots.size();
        gfBool diff = true;
        for (gfInt j = 0;j < num;j++)
        {
            if (m_w_knot_vector.at(i) == dif_w_knots.at(j))
                diff = false;
        }
        if (diff)
            dif_w_knots.push_back(m_w_knot_vector.at(i));
    }

    // fill array with Bezier segments
    gfInt num_w_seg = nb;
    segments->resize(num_u_seg, num_v_seg, num_w_seg);
    for (i = 0;i < num_u_seg;i++)
    {
        for (j = 0;j < num_v_seg;j++)
        {
            for (k = 0;k < num_w_seg;k++)
            {
                // create NURBS box as Bezier box
                gfObjGeomObject* obj = gfCreateGeomObject(GF_NURBS_BOX);
                gfObjNURBSBox* box = (gfObjNURBSBox*) obj;
                box->InitBezier(m_u_degree + 1, m_v_degree + 1, m_w_degree + 1, 
                                  dif_u_knots.at(i), dif_u_knots.at(i+1), 
                                  dif_v_knots.at(j), dif_v_knots.at(j+1),
                                  dif_w_knots.at(k), dif_w_knots.at(k+1));

                for (gfInt ii = 0;ii <= m_u_degree;ii++)
                {
                    for (gfInt jj = 0;jj <= m_v_degree;jj++)
                    {
                        for (gfInt kk = 0;kk <= m_w_degree;kk++)
                        {
                            gfPoint3* pt = box->GetControlPoint(ii, jj, kk);
                            gfDouble ww = Qw.at(k, ii+i*m_u_degree, jj+j*m_v_degree, kk).W();
                            pt->Set(Qw.at(k, ii+i*m_u_degree, jj+j*m_v_degree, kk).X() / ww, 
                                    Qw.at(k, ii+i*m_u_degree, jj+j*m_v_degree, kk).Y() / ww, 
                                    Qw.at(k, ii+i*m_u_degree, jj+j*m_v_degree, kk).Z() / ww);

                            box->SetWeight(ii, jj, kk, Qw.at(k, ii+i*m_u_degree, jj+j*m_v_degree, kk).W());
                        }
                    }
                }

                segments->Set(i, j, k, box);
            }
        }
    }

    gfLog("NURBS decomposed to (%d x %d x %d) Bezier segments.\n", num_u_seg, num_v_seg, num_w_seg);

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::DecomposeUDeCasteljau(gfDouble u, gfObjNURBSBox* box1, gfObjNURBSBox* box2)
{
    gfDouble u_min, u_max, v_min, v_max, w_min, w_max;
    GetDomain(&u_min, &u_max, &v_min, &v_max, &w_min, &w_max);

    if (u < u_min || u > u_max || !IsBezier())
        return GF_E_FAIL;

    // initialize curves
    box1->InitBezier(m_u_degree + 1, m_v_degree + 1, m_w_degree + 1, u_min, u, v_min, v_max, w_min, w_max);
    box2->InitBezier(m_u_degree + 1, m_v_degree + 1, m_w_degree + 1, u, u_max, v_min, v_max, w_min, w_max);

    gfDouble true_u = (u-u_min) / (u_max-u_min);

    for (gfInt row = 0;row <= m_v_degree;row++)
    {
        for (gfInt dep = 0;dep <= m_w_degree;dep++)
        {
            gfAux2DArray<gfPoint4> aux;
            aux.resize(m_u_degree + 1, m_u_degree + 1);
            for (gfInt i = 0;i <= m_u_degree;i++)
            {
                gfPoint3* pt = GetControlPoint(i, row, dep);
                aux.Set(0, i, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(i, row, dep)));
            }

            // decompose using deCasteljau algorithm
            for (gfInt i = 1;i <= m_u_degree;i++)
            {
                for (gfInt j = 0;j < m_u_degree - i + 1;j++)
                {
                    gfDouble ax = (1.0-true_u)*aux.at(i-1, j).W()*aux.at(i-1, j).X()+true_u*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).X();
                    gfDouble ay = (1.0-true_u)*aux.at(i-1, j).W()*aux.at(i-1, j).Y()+true_u*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Y();
                    gfDouble az = (1.0-true_u)*aux.at(i-1, j).W()*aux.at(i-1, j).Z()+true_u*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Z();
                    gfDouble aw = (1.0-true_u)*aux.at(i-1, j).W()+true_u*aux.at(i-1, j+1).W();
                    aux.Set(i, j, gfPoint4(ax/aw, ay/aw, az/aw, aw));
                }
            }

            // fill curves
            for (gfInt i = 0;i <= m_u_degree;i++)
            {
                gfPoint3* pt1 = box1->GetControlPoint(i, row, dep);
                pt1->Set(aux.at(i, 0).X(), aux.at(i, 0).Y(), aux.at(i, 0).Z());
                box1->SetWeight(i, row, dep, aux.at(i, 0).W());

                gfPoint3* pt2 = box2->GetControlPoint(i, row, dep);
                pt2->Set(aux.at(m_u_degree-i, i).X(), aux.at(m_u_degree-i, i).Y(), aux.at(m_u_degree-i, i).Z());
                box2->SetWeight(i, row, dep, aux.at(m_u_degree-i, i).W());
            }
        }
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::DecomposeVDeCasteljau(gfDouble v, gfObjNURBSBox* box1, gfObjNURBSBox* box2)
{
    gfDouble u_min, u_max, v_min, v_max, w_min, w_max;
    GetDomain(&u_min, &u_max, &v_min, &v_max, &w_min, &w_max);

    if (v < v_min || v > v_max || !IsBezier())
        return GF_E_FAIL;

    // initialize curves
    box1->InitBezier(m_u_degree + 1, m_v_degree + 1, m_w_degree + 1, u_min, u_max, v_min, v, w_min, w_max);
    box2->InitBezier(m_u_degree + 1, m_v_degree + 1, m_w_degree + 1, u_min, u_max, v, v_max, w_min, w_max);

    gfDouble true_v = (v-v_min) / (v_max-v_min);

    for (gfInt col = 0;col <= m_u_degree;col++)
    {
        for (gfInt dep = 0;dep <= m_w_degree;dep++)
        {
            gfAux2DArray<gfPoint4> aux;
            aux.resize(m_v_degree + 1, m_v_degree + 1);
            for (gfInt i = 0;i <= m_v_degree;i++)
            {
                gfPoint3* pt = GetControlPoint(col, i, dep);
                aux.Set(0, i, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(col, i, dep)));
            }

            // decompose using deCasteljau algorithm
            for (gfInt i = 1;i <= m_v_degree;i++)
            {
                for (gfInt j = 0;j < m_v_degree - i + 1;j++)
                {
                    gfDouble ax = (1.0-true_v)*aux.at(i-1, j).W()*aux.at(i-1, j).X()+true_v*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).X();
                    gfDouble ay = (1.0-true_v)*aux.at(i-1, j).W()*aux.at(i-1, j).Y()+true_v*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Y();
                    gfDouble az = (1.0-true_v)*aux.at(i-1, j).W()*aux.at(i-1, j).Z()+true_v*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Z();
                    gfDouble aw = (1.0-true_v)*aux.at(i-1, j).W()+true_v*aux.at(i-1, j+1).W();
                    aux.Set(i, j, gfPoint4(ax/aw, ay/aw, az/aw, aw));
                }
            }

            // fill curves
            for (gfInt i = 0;i <= m_v_degree;i++)
            {
                gfPoint3* pt1 = box1->GetControlPoint(col, i, dep);
                pt1->Set(aux.at(i, 0).X(), aux.at(i, 0).Y(), aux.at(i, 0).Z());
                box1->SetWeight(col, i, dep, aux.at(i, 0).W());

                gfPoint3* pt2 = box2->GetControlPoint(col, i, dep);
                pt2->Set(aux.at(m_v_degree-i, i).X(), aux.at(m_v_degree-i, i).Y(), aux.at(m_v_degree-i, i).Z());
                box2->SetWeight(col, i, dep, aux.at(m_v_degree-i, i).W());
            }
        }
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::DecomposeWDeCasteljau(gfDouble w, gfObjNURBSBox* box1, gfObjNURBSBox* box2)
{
    gfDouble u_min, u_max, v_min, v_max, w_min, w_max;
    GetDomain(&u_min, &u_max, &v_min, &v_max, &w_min, &w_max);

    if (w < w_min || w > w_max || !IsBezier())
        return GF_E_FAIL;

    // initialize curves
    box1->InitBezier(m_u_degree + 1, m_v_degree + 1, m_w_degree + 1, u_min, u_max, v_min, v_max, w_min, w);
    box2->InitBezier(m_u_degree + 1, m_v_degree + 1, m_w_degree + 1, u_min, u_max, v_min, v_max, w, w_max);

    gfDouble true_w = (w-w_min) / (w_max-w_min);

    for (gfInt col = 0;col <= m_u_degree;col++)
    {
        for (gfInt row = 0;row <= m_v_degree;row++)
        {
            gfAux2DArray<gfPoint4> aux;
            aux.resize(m_w_degree + 1, m_w_degree + 1);
            for (gfInt i = 0;i <= m_w_degree;i++)
            {
                gfPoint3* pt = GetControlPoint(col, row, i);
                aux.Set(0, i, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(col, row, i)));
            }

            // decompose using deCasteljau algorithm
            for (gfInt i = 1;i <= m_w_degree;i++)
            {
                for (gfInt j = 0;j < m_w_degree - i + 1;j++)
                {
                    gfDouble ax = (1.0-true_w)*aux.at(i-1, j).W()*aux.at(i-1, j).X()+true_w*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).X();
                    gfDouble ay = (1.0-true_w)*aux.at(i-1, j).W()*aux.at(i-1, j).Y()+true_w*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Y();
                    gfDouble az = (1.0-true_w)*aux.at(i-1, j).W()*aux.at(i-1, j).Z()+true_w*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Z();
                    gfDouble aw = (1.0-true_w)*aux.at(i-1, j).W()+true_w*aux.at(i-1, j+1).W();
                    aux.Set(i, j, gfPoint4(ax/aw, ay/aw, az/aw, aw));
                }
            }

            // fill curves
            for (gfInt i = 0;i <= m_w_degree;i++)
            {
                gfPoint3* pt1 = box1->GetControlPoint(col, row, i);
                pt1->Set(aux.at(i, 0).X(), aux.at(i, 0).Y(), aux.at(i, 0).Z());
                box1->SetWeight(col, row, i, aux.at(i, 0).W());

                gfPoint3* pt2 = box2->GetControlPoint(col, row, i);
                pt2->Set(aux.at(m_w_degree-i, i).X(), aux.at(m_w_degree-i, i).Y(), aux.at(m_w_degree-i, i).Z());
                box2->SetWeight(col, row, i, aux.at(m_w_degree-i, i).W());
            }
        }
    }

    return GF_S_OK;
}

gfObjNURBSPatch* gfObjNURBSBox::GetUWIsopatch(gfDouble v_param)
{
    gfDouble a = GetVKnot(m_v_degree);
    gfDouble b = GetVKnot(GetNumVControlPoints());
    if (gfIsZero(b - a))
        return NULL;
    if (v_param >= b || v_param < a)
        return NULL;

    gfObjNURBSPatch* ret = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
    ret->SetParams(m_u_degree, m_w_degree, m_num_u_control_points, m_num_w_control_points);

    // fill knots
    gfInt num_u_knots = GetNumUKnots();
    for (gfInt i = 0;i < num_u_knots;i++)
    {
        ret->SetUKnot(i, GetUKnot(i));
    }
    gfInt num_w_knots = GetNumWKnots();
    for (gfInt i = 0;i < num_w_knots;i++)
    {
        ret->SetVKnot(i, GetWKnot(i));
    }

    // prepare basis
    gfInt m = GetNumVControlPoints() - 1;
    gfInt span = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v_param);
    std::vector<gfDouble> basis;
    gfGetBSplineAllValues(m_v_degree, &m_v_knot_vector, span, v_param, &basis);

    for (gfInt i = 0;i < m_num_u_control_points;i++)
        for (gfInt j = 0;j < m_num_w_control_points;j++)
        {
            gfPoint3 * pt = ret->GetControlPoint(i, j);

            gfDouble ax = 0;
            gfDouble ay = 0;
            gfDouble az = 0;
            gfDouble aw = 0;
            for (gfInt k = 0;k <= m_v_degree;k++)
            {
                gfInt index = span - m_v_degree + k;
                ax += basis.at(k) * GetWeight(i, index, j) * GetControlPoint(i, index, j)->X();
                ay += basis.at(k) * GetWeight(i, index, j) * GetControlPoint(i, index, j)->Y();
                az += basis.at(k) * GetWeight(i, index, j) * GetControlPoint(i, index, j)->Z();
                aw += basis.at(k) * GetWeight(i, index, j);
            }

            pt->Set(ax / aw, ay / aw, az / aw);
            ret->SetWeight(i, j, aw);
        }

    return ret;
}

gfObjNURBSPatch* gfObjNURBSBox::GetVWIsopatch(gfDouble u_param)
{
    gfDouble a = GetUKnot(m_u_degree);
    gfDouble b = GetUKnot(GetNumUControlPoints());
    if (gfIsZero(b - a))
        return NULL;
    if (u_param >= b || u_param < a)
        return NULL;

    gfObjNURBSPatch* ret = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
    ret->SetParams(m_v_degree, m_w_degree, m_num_v_control_points, m_num_w_control_points);

    // fill knots
    gfInt num_v_knots = GetNumVKnots();
    for (gfInt i = 0;i < num_v_knots;i++)
    {
        ret->SetUKnot(i, GetVKnot(i));
    }
    gfInt num_w_knots = GetNumWKnots();
    for (gfInt i = 0;i < num_w_knots;i++)
    {
        ret->SetVKnot(i, GetWKnot(i));
    }

    // prepare basis
    gfInt n = GetNumUControlPoints()  - 1;
    gfInt span = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u_param);
    std::vector<gfDouble> basis;
    gfGetBSplineAllValues(m_u_degree, &m_u_knot_vector, span, u_param, &basis);

    for (gfInt i = 0;i < m_num_v_control_points;i++)
        for (gfInt j = 0;j < m_num_w_control_points;j++)
        {
            gfPoint3 * pt = ret->GetControlPoint(i, j);

            gfDouble ax = 0;
            gfDouble ay = 0;
            gfDouble az = 0;
            gfDouble aw = 0;
            for (gfInt k = 0;k <= m_u_degree;k++)
            {
                gfInt index = span - m_u_degree + k;
                ax += basis.at(k) * GetWeight(index, i, j) * GetControlPoint(index, i, j)->X();
                ay += basis.at(k) * GetWeight(index, i, j) * GetControlPoint(index, i, j)->Y();
                az += basis.at(k) * GetWeight(index, i, j) * GetControlPoint(index, i, j)->Z();
                aw += basis.at(k) * GetWeight(index, i, j);
            }

            pt->Set(ax / aw, ay / aw, az / aw);
            ret->SetWeight(i, j, aw);
        }

    return ret;
}

gfAuxBoundingVolume* gfObjNURBSBox::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    // create auxiliary array of mesh points
    gfInt num_points = (gfInt)m_control_points.size();
    if (num_points < 1)
        return NULL;

    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfInt i = 0; i < num_points; i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        curr_point->Set(pt->X(), pt->Y(), pt->Z());
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    //delete [] points;
    return result;
}

GFRESULT gfObjNURBSBox::CreateVACache(gfInt approx)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    DeleteVACache();

    if (m_time_stamp == 0)
        return GF_S_OK;

    // interval of definition
    gfDouble a, b, c, d, e, f;
    GetDomain(&a, &b, &c, &d, &e, &f);
    if (gfIsZero(b - a) || gfIsZero(d - c) || gfIsZero(f - e))
        return GF_S_OK;

    gfInt n = GetNumUControlPoints() - 1;
    gfInt m = GetNumVControlPoints() - 1;
    gfInt o = GetNumWControlPoints() - 1;

    gfPoint3* aux_points = new gfPoint3[approx * approx * approx];
    gfVector3* aux_normals = new gfVector3[approx * approx * approx];
    // TODO: normals needs to be evaluated

    for (gfInt i = 0;i < approx;i++)
    {
        // current u parameter
        gfDouble u = a + i * (b - a) / (approx - 1);
        gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
        gfAux2DArray<gfDouble> ubasis(2, m_u_degree+1);
        gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 1, &ubasis);

        for (gfInt j = 0;j < approx;j++)
        {
            // current v parameter     
            gfDouble v = c + j * (d - c) / (approx - 1);
            gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
            gfAux2DArray<gfDouble> vbasis(2, m_v_degree+1);
            gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 1, &vbasis);

            for (gfInt k = 0;k < approx;k++)
            {
                // current w parameter     
                gfDouble w = e + k * (f - e) / (approx - 1);
                gfInt wspan = gfGetBSplineKnotSpan(o, m_w_degree, &m_w_knot_vector, w);
                gfAux2DArray<gfDouble> wbasis(2, m_w_degree+1);
                gfGetBSplineAllDers(m_w_degree, &m_w_knot_vector, wspan, w, 1, &wbasis);

                gfDouble ax = 0;
                gfDouble ay = 0;
                gfDouble az = 0;
                gfDouble aw = 0;

                gfDouble dux = 0.0;
                gfDouble duy = 0.0;
                gfDouble duz = 0.0;
                gfDouble duw = 0.0;

                gfDouble dvx = 0.0;
                gfDouble dvy = 0.0;
                gfDouble dvz = 0.0;
                gfDouble dvw = 0.0;

                gfDouble dwx = 0.0;
                gfDouble dwy = 0.0;
                gfDouble dwz = 0.0;
                gfDouble dww = 0.0;

                for (gfInt iter1 = 0;iter1 <= m_u_degree;iter1++)
                {
                    gfInt uindex = uspan - m_u_degree + iter1;

                    for (gfInt iter2 = 0;iter2 <= m_v_degree;iter2++)
                    {
                        gfInt vindex = vspan - m_v_degree + iter2;

                        for (gfInt iter3 = 0;iter3 <= m_w_degree;iter3++)
                        {
                            gfInt windex = wspan - m_w_degree + iter3;

                            gfDouble aux = ubasis.at(0, iter1) * vbasis.at(0, iter2) * wbasis.at(0, iter3)* GetWeight(uindex, vindex, windex);
                            ax += aux * GetControlPoint(uindex, vindex, windex)->X();
                            ay += aux * GetControlPoint(uindex, vindex, windex)->Y();
                            az += aux * GetControlPoint(uindex, vindex, windex)->Z();
                            aw += aux;

                            gfDouble aux1 = ubasis.at(1, iter1) * vbasis.at(0, iter2) * wbasis.at(0, iter3) * GetWeight(uindex, vindex, windex);
                            dux += aux1 * GetControlPoint(uindex, vindex, windex)->X();
                            duy += aux1 * GetControlPoint(uindex, vindex, windex)->Y();
                            duz += aux1 * GetControlPoint(uindex, vindex, windex)->Z();
                            duw += aux1;

                            gfDouble aux2 = ubasis.at(0, iter1) * vbasis.at(1, iter2) * wbasis.at(0, iter3) * GetWeight(uindex, vindex, windex);
                            dvx += aux2 * GetControlPoint(uindex, vindex, windex)->X();
                            dvy += aux2 * GetControlPoint(uindex, vindex, windex)->Y();
                            dvz += aux2 * GetControlPoint(uindex, vindex, windex)->Z();
                            dvw += aux2;

                            gfDouble aux3 = ubasis.at(0, iter1) * vbasis.at(0, iter2) * wbasis.at(1, iter3) * GetWeight(uindex, vindex, windex);
                            dwx += aux3 * GetControlPoint(uindex, vindex, windex)->X();
                            dwy += aux3 * GetControlPoint(uindex, vindex, windex)->Y();
                            dwz += aux3 * GetControlPoint(uindex, vindex, windex)->Z();
                            dww += aux3;
                        }
                    }
                }

                // set computed point to array
                aux_points[i + j * approx + k * approx * approx].Set(ax / aw, ay / aw, az / aw);

                // fill normal
                gfDouble fdux = (dux * aw - ax * duw) / aw / aw;
                gfDouble fduy = (duy * aw - ay * duw) / aw / aw;
                gfDouble fduz = (duz * aw - az * duw) / aw / aw;

                gfDouble fdvx = (dvx * aw - ax * dvw) / aw / aw;
                gfDouble fdvy = (dvy * aw - ay * dvw) / aw / aw;
                gfDouble fdvz = (dvz * aw - az * dvw) / aw / aw;

                gfDouble fdwx = (dwx * aw - ax * dww) / aw / aw;
                gfDouble fdwy = (dwy * aw - ay * dww) / aw / aw;
                gfDouble fdwz = (dwz * aw - az * dww) / aw / aw;

                gfVector3* resultu = gfCreateVector3();
                gfVector3* resultv = gfCreateVector3();
                gfVector3* resultw = gfCreateVector3();
                gfDouble a, b, c, d, e, f;
                GetDomain(&a, &b, &c, &d, &e, &f);

                resultu->Set((fdwy*fdvz-fdwz*fdvy),
                    (-fdwx*fdvz+fdwz*fdvx),
                    (fdwx*fdvy-fdwy*fdvx));
                resultv->Set((fdwy*fduz-fdwz*fduy),
                    (-fdwx*fduz+fdwz*fdux),
                    (fdwx*fduy-fdwy*fdux));
                resultw->Set((fdvy*fduz-fdvz*fduy),
                    (-fdvx*fduz+fdvz*fdux),
                    (fdvx*fduy-fdvy*fdux));

                gfDouble jacob = fdux*fdvy*fdwz+fduy*fdvz*fdwx+fduz*fdvx*fdwy-
                                 fduz*fdvy*fdwx-fduy*fdvx*fdwz-fdux*fdvz*fdwy;

                if (gfIsZero(u-a) || gfIsZero(u-b))
                    aux_normals[i + j * approx + k * approx * approx].Set(resultu->X(), resultu->Y(), resultu->Z());
                else if (gfIsZero(v-c) || gfIsZero(v-d))
                    aux_normals[i + j * approx + k * approx * approx].Set(resultv->X(), resultv->Y(), resultv->Z());
                else if (gfIsZero(w-e) || gfIsZero(w-f))
                    aux_normals[i + j * approx + k * approx * approx].Set(resultw->X(), resultw->Y(), resultw->Z());
                else if (gfIsZero(jacob))
                {
                    if (!gfIsZero(resultu->Length()))
                        aux_normals[i + j * approx + k * approx * approx].Set(resultu->X(), resultu->Y(), resultu->Z());
                    else if (!gfIsZero(resultv->Length()))
                        aux_normals[i + j * approx + k * approx * approx].Set(resultv->X(), resultv->Y(), resultv->Z());
                    else if (!gfIsZero(resultw->Length()))
                        aux_normals[i + j * approx + k * approx * approx].Set(resultw->X(), resultw->Y(), resultw->Z());
                    else
                        aux_normals[i + j * approx + k * approx * approx].Set(0.0, 0.0, 0.0);
                }
                else
                {
                    // TODO: normal is linear combination of boundary normals
                    aux_normals[i + j * approx + k * approx * approx].Set(0.0, 0.0, 0.0);
                }
                aux_normals[i + j * approx + k * approx * approx].Normalize();
                gfDeleteVector3(resultu);
                gfDeleteVector3(resultv);
                gfDeleteVector3(resultw);
            }
        }
    }

    // get number of triangles
    m_num_cached_quads = 3*approx*(approx-1)*(approx-1);
    m_num_cached_wire_lines = 3*(approx-1)*approx*approx;

    // allocate streams
    m_cache_vertices = new gfDouble[12*m_num_cached_quads];
    m_cache_normals = new gfDouble[12*m_num_cached_quads];
    m_cache_wire_vertices = new gfDouble[6*m_num_cached_wire_lines];

    // fill quads
    gfInt ii = 0;
    gfInt jj = 0;
    for (gfInt i = 0;i < approx;i++)
    {
        for (gfInt j = 0;j < approx;j++)
        {
            for (gfInt k = 0;k < approx;k++)
            { 
                if (i < approx-1 && j < approx-1)
                {
                    m_cache_vertices[ii] = aux_points[i + j * approx + k * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + j * approx + k * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + k * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + j * approx + k * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + k * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + j * approx + k * approx * approx].Z();
                    ii++;

                    m_cache_vertices[ii] = aux_points[i + (j + 1) * approx + k * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + (j + 1) * approx + k * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + (j + 1) * approx + k * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + (j + 1) * approx + k * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + (j + 1) * approx + k * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + (j + 1) * approx + k * approx * approx].Z();
                    ii++;

                    m_cache_vertices[ii] = aux_points[i + 1 + (j + 1) * approx + k * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + 1 + (j + 1) * approx + k * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + 1 + (j + 1) * approx + k * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + 1 + (j + 1) * approx + k * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + 1 + (j + 1) * approx + k * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + 1 + (j + 1) * approx + k * approx * approx].Z();
                    ii++;

                    m_cache_vertices[ii] = aux_points[i + 1 + j * approx + k * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + 1 + j * approx + k * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + 1 + j * approx + k * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + 1 + j * approx + k * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + 1 + j * approx + k * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + 1 + j * approx + k * approx * approx].Z();
                    ii++;
                }
                if (i < approx-1 && k < approx-1)
                {
                    m_cache_vertices[ii] = aux_points[i + j * approx + k * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + j * approx + k * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + k * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + j * approx + k * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + k * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + j * approx + k * approx * approx].Z();
                    ii++;

                    m_cache_vertices[ii] = aux_points[i + j * approx + (k + 1) * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + j * approx + (k + 1) * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + (k + 1) * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + j * approx + (k + 1) * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + (k + 1) * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + j * approx + (k + 1) * approx * approx].Z();
                    ii++;

                    m_cache_vertices[ii] = aux_points[i + 1 + j * approx + (k + 1) * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + 1 + j * approx + (k + 1) * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + 1 + j * approx + (k + 1) * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + 1 + j * approx + (k + 1) * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + 1 + j * approx + (k + 1) * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + 1 + j * approx + (k + 1) * approx * approx].Z();
                    ii++;

                    m_cache_vertices[ii] = aux_points[i + 1 + j * approx + k * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + 1 + j * approx + k * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + 1 + j * approx + k * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + 1 + j * approx + k * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + 1 + j * approx + k * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + 1 + j * approx + k * approx * approx].Z();
                    ii++;
                }
                if (j < approx-1 && k < approx-1)
                {
                    m_cache_vertices[ii] = aux_points[i + j * approx + k * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + j * approx + k * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + k * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + j * approx + k * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + k * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + j * approx + k * approx * approx].Z();
                    ii++;

                    m_cache_vertices[ii] = aux_points[i + j * approx + (k + 1) * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + j * approx + (k + 1) * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + (k + 1) * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + j * approx + (k + 1) * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + j * approx + (k + 1) * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + j * approx + (k + 1) * approx * approx].Z();
                    ii++;

                    m_cache_vertices[ii] = aux_points[i + (j + 1) * approx + (k + 1) * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + (j + 1) * approx + (k + 1) * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + (j + 1) * approx + (k + 1) * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + (j + 1) * approx + (k + 1) * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + (j + 1) * approx + (k + 1) * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + (j + 1) * approx + (k + 1) * approx * approx].Z();
                    ii++;

                    m_cache_vertices[ii] = aux_points[i + (j + 1) * approx + k * approx * approx].X();
                    m_cache_normals[ii] = aux_normals[i + (j + 1) * approx + k * approx * approx].X();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + (j + 1) * approx + k * approx * approx].Y();
                    m_cache_normals[ii] = aux_normals[i + (j + 1) * approx + k * approx * approx].Y();
                    ii++;
                    m_cache_vertices[ii] = aux_points[i + (j + 1) * approx + k * approx * approx].Z();
                    m_cache_normals[ii] = aux_normals[i + (j + 1) * approx + k * approx * approx].Z();
                    ii++;
                }
            }
        }
    }

    for (gfInt i = 0;i < approx;i++)
    {
        for (gfInt j = 0;j < approx;j++)
        {
            for (gfInt k = 0;k < approx;k++)
            {
                if (i < approx - 1)
                {
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + k * approx * approx].X();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + k * approx * approx].Y();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + k * approx * approx].Z();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + 1 + j * approx + k * approx * approx].X();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + 1 + j * approx + k * approx * approx].Y();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + 1 + j * approx + k * approx * approx].Z();
                    jj++;
                }
                if (j < approx - 1)
                {
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + k * approx * approx].X();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + k * approx * approx].Y();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + k * approx * approx].Z();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + (j + 1) * approx + k * approx * approx].X();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + (j + 1) * approx + k * approx * approx].Y();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + (j + 1) * approx + k * approx * approx].Z();
                    jj++;
                }
                if (k < approx - 1)
                {
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + k  * approx * approx].X();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + k * approx * approx].Y();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + k * approx * approx].Z();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + (k + 1) * approx * approx].X();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + (k + 1) * approx * approx].Y();
                    jj++;
                    m_cache_wire_vertices[jj] = aux_points[i + j * approx + (k + 1) * approx * approx].Z();
                    jj++;
                }
            }
        }
    }

    delete [] aux_points;
    delete [] aux_normals;

    gfLog("Computed VertexArray cache for NURBS box, time = %d ms.\n", count.EndMSCount());

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::DeleteVACache(void)
{
    if (m_cache_vertices)
        delete [] m_cache_vertices;
    m_cache_vertices = NULL;
    if (m_cache_normals)
        delete [] m_cache_normals;
    m_cache_normals = NULL;
    if (m_cache_wire_vertices)
        delete [] m_cache_wire_vertices;
    m_cache_wire_vertices = NULL;
    m_num_cached_quads = 0;
    m_num_cached_wire_lines = 0;

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::DrawVACache(gfLong flags)
{
    if (GF_FLG_3D_OPACITY & flags)
    {
        glDisable(GL_DEPTH_TEST);
        glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);
    }
    
    if (flags & GF_FLG_SOLID)
    {
        glEnableClientState(GL_VERTEX_ARRAY);
        if (flags & GF_FLG_NORMALS)
            glEnableClientState(GL_NORMAL_ARRAY);

        if (m_cache_vertices)
            glVertexPointer(3, GL_DOUBLE, 0, m_cache_vertices);
        if (m_cache_normals && (flags & GF_FLG_NORMALS))
            glNormalPointer(GL_DOUBLE, 0, m_cache_normals);

        glDrawArrays(GL_QUADS, 0, 4*m_num_cached_quads);

        if (flags & GF_FLG_NORMALS)
            glDisableClientState(GL_NORMAL_ARRAY);
        glDisableClientState(GL_VERTEX_ARRAY);
    }

    if (flags & GF_FLG_WIRE)
    {
        glNormal3f(0.0, 0.0, 0.0);
        glEnableClientState(GL_VERTEX_ARRAY);

        if (m_cache_wire_vertices)
            glVertexPointer(3, GL_DOUBLE, 0, m_cache_wire_vertices);

        glDrawArrays(GL_LINES, 0, 2*m_num_cached_wire_lines);

        glDisableClientState(GL_VERTEX_ARRAY);
    }

    if (GF_FLG_3D_OPACITY & flags)
    {
        glEnable(GL_DEPTH_TEST);
        glDisable(GL_BLEND);
    }
    
    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::DeleteAABBTree(gfBezier3AABBTreeNode* tree_root)
{
    if (tree_root->aabb)
        gfDeleteBoundVolume(tree_root->aabb);
    if (tree_root->left)
        DeleteAABBTree(tree_root->left);
    if (tree_root->right)
        DeleteAABBTree(tree_root->right);
    if (tree_root->box)
        gfDeleteGeomObject(tree_root->box);
    delete tree_root;

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::DeleteWholeAABBTree(void)
{
    gfInt num_trees = (gfInt)m_bezier_aabb_trees.size();
    for (gfInt i = 0;i < num_trees;i++)
    {
        gfBezier3AABBTreeNode* root = m_bezier_aabb_trees.at(i);
        DeleteAABBTree(root);
    }
    m_bezier_aabb_trees.clear();

    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::ComputeWholeAABBTree(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    DeleteWholeAABBTree();

    m_num_nodes = 0;

    // decompose to Bezier patches
    gfAux3DArray<gfObjNURBSBox*> beziers;
    DecomposeToBezierSegments(&beziers);

    gfInt num_i_trees = beziers.GetISize();
    gfInt num_j_trees = beziers.GetJSize();
    gfInt num_k_trees = beziers.GetKSize();
    for (gfInt i = 0;i < num_i_trees;i++)
        for (gfInt j = 0;j < num_j_trees;j++)
            for (gfInt k = 0;k < num_k_trees;k++)
            {
                gfObjNURBSBox* bezier = beziers.at(i, j, k);
                gfBezier3AABBTreeNode* root = CreateAABBTree(0, bezier);
                m_bezier_aabb_trees.push_back(root);
            }

    gfLog("Computed AABB tree of Bezier boxes for NURBS box, nodes = %d, time = %d ms.\n", m_num_nodes, count.EndMSCount());

    return GF_S_OK;
}

gfBezier3AABBTreeNode* gfObjNURBSBox::CreateAABBTree(gfDouble depth, gfObjNURBSBox* box)
{
    if (!box)
        return NULL;

    gfBezier3AABBTreeNode* root = new gfBezier3AABBTreeNode;
    root->left = root->right = NULL;
    root->aabb = (gfAuxAABB*) box->GetBoundVolume(GF_BOUND_VOL_AABB, NULL);
    root->box = NULL;
    root->min_dir = 0;
    
    m_num_nodes++;

    // compute bounds of jacobi determinant
    gfInt n = box->GetNumUControlPoints() - 1;
    gfInt m = box->GetNumVControlPoints() - 1;
    gfInt o = box->GetNumWControlPoints() - 1;
    
    // u-derivation
    std::vector<gfVector3> vectors1;
    vectors1.resize(n*(m+1)*(o+1));
    for (gfInt i = 0;i < n;i++)
        for (gfInt j = 0;j <= m;j++)
            for (gfInt k = 0;k <= o;k++)
            {
                gfVector3 sub;
                gfSubstractPoints3(box->GetControlPoint(i+1, j, k), box->GetControlPoint(i, j, k), &sub);
                vectors1[i+j*n+k*n*(m+1)] = sub;                            
            }
    // create cone from vectors
    gfAuxBoundCone ucone;
    ucone.CreateFromVectors(&vectors1);

    // v-derivation
    std::vector<gfVector3> vectors2;
    vectors2.resize((n+1)*m*(o+1));
    for (gfInt i = 0;i <= n;i++)
        for (gfInt j = 0;j < m;j++)
            for (gfInt k = 0;k <= o;k++)
            {
                gfVector3 sub;
                gfSubstractPoints3(box->GetControlPoint(i, j+1, k), box->GetControlPoint(i, j, k), &sub);
                vectors2[i+j*(n+1)+k*(n+1)*m] = sub;                            
            }
    // create cone from vectors
    gfAuxBoundCone vcone;
    vcone.CreateFromVectors(&vectors2);

    // w-derivation
    std::vector<gfVector3> vectors3;
    vectors3.resize((n+1)*(m+1)*o);
    for (gfInt i = 0;i <= n;i++)
    {
        for (gfInt j = 0;j <= m;j++)
            for (gfInt k = 0;k < o;k++)
            {
                gfVector3 sub;
                gfSubstractPoints3(box->GetControlPoint(i, j, k+1), box->GetControlPoint(i, j, k), &sub);
                vectors3[i+j*(n+1)+k*(n+1)*(m+1)] = sub;                            
            }
    }
    // create cone from vectors
    gfAuxBoundCone wcone;
    wcone.CreateFromVectors(&vectors3);

    gfAuxBoundCone cross;
    cross.CrossProduct(vcone, wcone);
    ucone.ScalarProduct(cross, &root->jacobi_min, &root->jacobi_max);

    if (root->jacobi_min * root->jacobi_max > 0)
    {
        return root;
    }

    gfDouble umin, umax, vmin, vmax, wmin, wmax;
    GetDomain(&umin, &umax, &vmin, &vmax, &wmin, &wmax);

    // compute also OBB
    /*gfAuxOBB* obb = (gfAuxOBB*) box->GetBoundVolume(GF_BOUND_VOL_OBB, NULL);

    // check if AABB needs to be splitted
    gfDouble x_dir = 2.0*gfSquareRoot(obb->m_axis1_x*obb->m_axis1_x+
        obb->m_axis1_y*obb->m_axis1_y+
        obb->m_axis1_z*obb->m_axis1_z);
    gfDouble y_dir = 2.0*gfSquareRoot(obb->m_axis2_x*obb->m_axis2_x+
        obb->m_axis2_y*obb->m_axis2_y+
        obb->m_axis2_z*obb->m_axis2_z);
    gfDouble z_dir = 2.0*gfSquareRoot(obb->m_axis3_x*obb->m_axis3_x+
        obb->m_axis3_y*obb->m_axis3_y+
        obb->m_axis3_z*obb->m_axis3_z);*/

    gfDouble x_dir = umax-umin;
    gfDouble y_dir = vmax-vmin;
    gfDouble z_dir = wmax-wmin;

    // find shortest dimension
    gfDouble min;
    gfDouble space;
    if (x_dir < y_dir)
    {
        if (x_dir < z_dir)
        {
            min = x_dir;
            root->min_dir = 1;
            space = y_dir*z_dir;
        }
        else
        {
            min = z_dir;
            root->min_dir = 3;
            space = y_dir*x_dir;
        }
    }
    else
    {
        if (y_dir < z_dir)
        {
            min = y_dir;
            root->min_dir = 2;
            space = x_dir*z_dir;
        }
        else
        {
            min = z_dir;
            root->min_dir = 3;
            space = y_dir*x_dir;
        }
    }

    // delete OBB
    //gfDeleteBoundVolume(obb);

    // check if OBB is flat enough
    if ((min < 0.001 && space < 0.001) || depth > GF_MAX_BEZIER_AABB_TREE_DEPTH)
    {
        root->box = box;
    }
    else
    {
        // split box
        gfDouble u_min, u_max, v_min, v_max, w_min, w_max;
        box->GetDomain(&u_min, &u_max, &v_min, &v_max, &w_min, &w_max);

        gfObjNURBSBox* box1 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
        gfObjNURBSBox* box2 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);

        if (((u_max-u_min) > (v_max-v_min)) && ((u_max-u_min) > (w_max-w_min)))
            box->DecomposeUDeCasteljau((u_max+u_min)/2.0, box1, box2);
        else if (((v_max-v_min) > (u_max-u_min)) && ((v_max-v_min) > (w_max-w_min)))
            box->DecomposeVDeCasteljau((v_max+v_min)/2.0, box1, box2);
        else
            box->DecomposeWDeCasteljau((w_max+w_min)/2.0, box1, box2);

        gfDeleteGeomObject(box);

        root->left = CreateAABBTree(depth+1, box1);
        root->right = CreateAABBTree(depth+1, box2);
    }

    return root;
}

gfRayInterInfo* gfObjNURBSBox::RayAABBIntersection(gfBezier3AABBTreeNode* root, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    if (!root)
        return NULL;
    gfDouble tmin, tmax;
    if (GF_FAIL(root->aabb->RayIntersection(ray_begin, ray_dir, &tmin, &tmax)))
        return NULL;

    if (root->jacobi_min * root->jacobi_max > 0)
        return NULL;

    if (root->left && root->right)
    {
        RayAABBIntersection(root->left, ray_begin, ray_dir);
        RayAABBIntersection(root->right, ray_begin, ray_dir);
    }
    else
    {
        gfBezier3AABBQueueUnit unit;
        unit.tmin = tmin;
        unit.tmax = tmax;
        unit.node = root;
        m_aabb_queue.push(unit);
    }

    return NULL;

    /*gfRayInterInfo* info = NULL;
    if (root->left && root->right)
    {
        // get intersection from childs
        gfRayInterInfo* info1 = RayAABBIntersection(root->left, ray_begin, ray_dir);
        gfRayInterInfo* info2 = RayAABBIntersection(root->right, ray_begin, ray_dir);
        if (info1 && info2)
        {
            if (info1->m_inters.at(0) < info2->m_inters.at(0))
            {
                info = info1;
                for (gfInt i = 0;i < info2->m_inters.size();i++)
                    info->m_inters.push_back(info2->m_inters.at(i));
                gfDeleteRayInterInfo(info2);
            }
            else
            {
                info = info2;
                for (gfInt i = 0;i < info1->m_inters.size();i++)
                    info->m_inters.push_back(info1->m_inters.at(i));
                gfDeleteRayInterInfo(info1);
            }
        }
        else if (info1)
        {
            info = info1;
        }
        else if (info2)
        {
            info = info2;
        }
    }
    else
    {
        if (!root->box || !root->aabb)
            return NULL;

        // compute exact intersection using bezier clipping
        // use bezier clipping
        gfVector3 up;
        up.Set(-ray_dir->Y()*ray_dir->Z(), ray_dir->X()*ray_dir->Z()/2.0, ray_dir->X()*ray_dir->Y()/2.0);
        gfVector3 right;
        gfVectorProduct3(ray_dir, &up, &right);
        right.Normalize();
        up.Normalize();
        gfDouble d1 = -(up.X()*ray_begin->X()+up.Y()*ray_begin->Y()+up.Z()*ray_begin->Z());
        gfDouble d2 = -(right.X()*ray_begin->X()+right.Y()*ray_begin->Y()+right.Z()*ray_begin->Z());

        gfRayBezierBoxInfo* res = root->box->BezierClipping(NULL, 0, 
                                                    up.X(), up.Y(), up.Z(), d1, right.X(), right.Y(), right.Z(), d2);
        if (res)
            {
                gfInt size = res->m_inters.size();
                if (size > 0)
                    info = gfCreateRayInterInfo();
                for (gfInt i = 0;i < size;i++)
                {
                    gfPoint3* pt = root->box->GetVolumePoint(res->m_inters.at(i).X(), res->m_inters.at(i).Y(), res->m_inters.at(i).Z());
                    gfVector3* nm = root->box->GetNormal(res->m_inters.at(i).X(), res->m_inters.at(i).Y(), res->m_inters.at(i).Z());
                    gfVector3 vec(pt->X()-ray_begin->X(), pt->Y()-ray_begin->Y(), pt->Z()-ray_begin->Z());

                    info->m_inters.push_back(gfScalarProduct3(ray_dir, &vec));
                    info->m_normals.push_back(gfVector3(nm->X(), nm->Y(), nm->Z()));
                    info->m_uvw.push_back(gfPoint3(res->m_inters.at(i).X(), res->m_inters.at(i).Y(), res->m_inters.at(i).Z()));

                    gfDeletePoint3(pt);
                    gfDeleteVector3(nm);
                }

                delete res;
            }

        if (info && info->m_inters.size() == 0)
        {
            gfDeleteRayInterInfo(info);
            info = NULL;
        }
    }

    // sort intersections
    if (info)
    {
        gfSort1DArray(&info->m_inters);
    }

    return info;*/
}

gfRayBezierBoxInfo* gfObjNURBSBox::BezierClipping(gfMatrix4* obj_matrix, gfInt depth, 
                                                      gfDouble a1, gfDouble b1, gfDouble c1, gfDouble d1,
                                                      gfDouble a2, gfDouble b2, gfDouble c2, gfDouble d2)
{
    if (!IsBezier())
    {
        // convert to Bezier segments
        gfAux3DArray<gfObjNURBSBox*> segments;
        DecomposeToBezierSegments(&segments);
        
        gfRayBezierBoxInfo* info = new gfRayBezierBoxInfo;
        for (gfInt i = 0;i < segments.GetISize();i++)
        {
            for (gfInt j = 0;j < segments.GetJSize();j++)
            {
                for (gfInt k = 0;k < segments.GetKSize();k++)
                {
                    gfRayBezierBoxInfo* info_aux = segments.at(i, j, k)->BezierClipping(obj_matrix, depth, a1, b1, c1, d1, a2, b2, c2, d2);
                    if (info_aux)
                    {
                        for (gfInt k = 0;k < (gfInt)info_aux->m_inters.size();k++)
                            info->m_inters.push_back(info_aux->m_inters.at(k));
                    }
                    delete info_aux;
                    gfDeleteGeomObject(segments.at(i, j, k));
                }
            }
        }

        if (info->m_inters.size() > 0)
        {
            return info;
        }
        else
        {
            delete info;
            return NULL;
        }
    }
    else
    {
        gfInt n = GetNumUControlPoints()-1;
        gfInt m = GetNumVControlPoints()-1;
        gfInt o = GetNumWControlPoints()-1;
        gfDouble umin, umax, vmin, vmax, wmin, wmax;
        GetDomain(&umin, &umax, &vmin, &vmax, &wmin, &wmax);

        // create auxiliary points
        gfAux3DArray<gfPoint2> auxP;
        auxP.resize(n+1, m+1, o+1);
        for (gfInt i = 0;i <= n;i++)
            for (gfInt j = 0;j <= m;j++)
                for (gfInt k = 0;k <= o;k++)
                {
                    gfPoint3* pt = GetControlPoint(i, j, k);
                    auxP.Set(i, j, k, gfPoint2(GetWeight(i, j, k)*(a1*pt->X()+b1*pt->Y()+c1*pt->Z()+d1), 
                                        GetWeight(i, j, k)*(a2*pt->X()+b2*pt->Y()+c2*pt->Z()+d2)));
                }

        // find line to project these points
        gfDouble a, b;
        gfPoint2 pt1 = auxP.at(0, 0, 0);
        gfPoint2 pt2 = auxP.at(n, 0, 0);
        gfPoint2 pt3 = auxP.at(0, m, 0);
        gfPoint2 pt4 = auxP.at(n, m, 0);
        gfPoint2 pt5 = auxP.at(0, 0, o);
        gfPoint2 pt6 = auxP.at(n, 0, o);
        gfPoint2 pt7 = auxP.at(0, m, o);
        gfPoint2 pt8 = auxP.at(n, m, o);

        gfDouble size1 = umax-umin;
        gfDouble size2 = vmax-vmin;
        gfDouble size3 = wmax-wmin;

        // find best direction
        gfInt dir = 1;
        if (size1>size2 && size1>size3)
        {
            // u direction
            dir = 1;
        }
        else if (size2>size1 && size2>size3)
        {
            // v direction
            dir = 2;
        }
        else
        {
            // w direction
            dir = 3;
        }

        // compute vector
        if (dir == 1)
        {
            a = -(pt2.Y()-pt1.Y()+pt4.Y()-pt3.Y()+pt6.Y()-pt5.Y()+pt8.Y()-pt7.Y());
            b = +(pt2.X()-pt1.X()+pt4.X()-pt3.X()+pt6.X()-pt5.X()+pt8.X()-pt7.X());
        }
        else if (dir == 2)
        {
            a = -(pt3.Y()-pt1.Y()+pt4.Y()-pt2.Y()+pt7.Y()-pt5.Y()+pt8.Y()-pt6.Y());
            b = +(pt3.X()-pt1.X()+pt4.X()-pt2.X()+pt7.X()-pt5.X()+pt8.X()-pt6.X());
        }
        else if (dir == 3)
        {
            a = -(pt5.Y()-pt1.Y()+pt6.Y()-pt2.Y()+pt7.Y()-pt3.Y()+pt8.Y()-pt4.Y());
            b = +(pt5.X()-pt1.X()+pt6.X()-pt2.X()+pt7.X()-pt3.X()+pt8.X()-pt4.X());
        }

        // project points
        gfAux3DArray<gfDouble> auxD;
        auxD.resize(n+1, m+1, o+1);
        for (gfInt i = 0;i <= n;i++)
            for (gfInt j = 0;j <= m;j++)
                for (gfInt k = 0;k <= o;k++)
                {
                    gfPoint2 pt = auxP.at(i, j, k);
                    auxD.Set(i, j, k, a*pt.X()+b*pt.Y());
                }

        // do clipping for significant direction
        std::vector<gfPoint2> aux;
        aux.resize((n+1)*(m+1)*(o+1));
        if (dir == 1)
        {
            // clipping in u direction
            for (gfInt i = 0;i <= n;i++)
                for (gfInt j = 0;j <= m;j++)
                    for (gfInt k = 0;k <= o;k++)
                    {
                        aux[i+j*(n+1)+k*(n+1)*(m+1)] = gfPoint2(umin+(umax-umin)*((gfDouble) i)/((gfDouble) n), auxD.at(i, j, k));
                    }

            // find better approximation of intersection region
            gfDouble uclipmin, uclipmax;
            if (gfAxisXConvexHullInter2D(&aux, &uclipmin, &uclipmax) != GF_S_OK)
                return NULL;
            if (uclipmin < umin || uclipmax > umax)
                return NULL;

            // check if region is small
            if (((uclipmax-uclipmin) < GF_BEZIER_CLIPPING_EPS && (vmax-vmin) < GF_BEZIER_CLIPPING_EPS && (wmax-wmin) < GF_BEZIER_CLIPPING_EPS) 
                || depth > GF_BEZIER_CLIPPING_MAX_DEPTH/1)
            {
                gfRayBezierBoxInfo* result = new gfRayBezierBoxInfo;
                //result->m_inters.push_back(gfPoint3((uclipmax+uclipmin)/2.0, (vmax+vmin)/2.0, (wmax+wmin)/2.0));
                result->m_inters.push_back(gfPoint3(uclipmin, (vmax+vmin)/2.0, (wmax+wmin)/2.0));
                result->m_inters.push_back(gfPoint3(uclipmax, (vmax+vmin)/2.0, (wmax+wmin)/2.0));
                return result;
            }

            // check if no significant change in interval happend
            if ((uclipmax-uclipmin)/(umax-umin) > GF_BEZIER_CLIPPING_DIVIDE_TRESHOLD)
            {
                gfObjNURBSBox* box1 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
                gfObjNURBSBox* box2 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
                DecomposeUDeCasteljau((umax+umin)/2.0, box1, box2);
                gfRayBezierBoxInfo* res1 = box1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfRayBezierBoxInfo* res2 = box2->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfDeleteGeomObject(box1);
                gfDeleteGeomObject(box2);

                // combine two results
                if (res1 && res2)
                {
                    for (gfInt i = 0;i < (gfInt)res2->m_inters.size();i++)
                        res1->m_inters.push_back(res2->m_inters.at(i));
                    delete res2;
                    return res1;
                }
                if (res1 && !res2)
                    return res1;
                if (res2 && !res1)
                    return res2;
                return NULL;
            }

            // divide patch using deCasteljau
            gfObjNURBSBox* box1 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
            gfObjNURBSBox* box2 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
            gfObjNURBSBox* box3 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
            DecomposeUDeCasteljau(uclipmin, box1, box2);
            box2->DecomposeUDeCasteljau(uclipmax, box1, box3);

            gfDeleteGeomObject(box2);
            gfDeleteGeomObject(box3);

            gfRayBezierBoxInfo* res = box1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);

            gfDeleteGeomObject(box1);

            return res;
        }
        else if (dir == 2)
        {
            // clipping in v direction
            for (gfInt i = 0;i <= n;i++)
                for (gfInt j = 0;j <= m;j++)
                    for (gfInt k = 0;k <= o;k++)
                    {
                        aux[i+j*(n+1)+k*(n+1)*(m+1)] = gfPoint2(vmin+(vmax-vmin)*((gfDouble) j)/((gfDouble) m), auxD.at(i, j, k));
                    }

            // find better approximation of intersection region
            gfDouble vclipmin, vclipmax;
            if (gfAxisXConvexHullInter2D(&aux, &vclipmin, &vclipmax) != GF_S_OK)
                return NULL;
            if (vclipmin < vmin || vclipmax > vmax)
                return NULL;

            // check if region is small
            if (((vclipmax-vclipmin) < GF_BEZIER_CLIPPING_EPS && (umax-umin) < GF_BEZIER_CLIPPING_EPS && (wmax-wmin) < GF_BEZIER_CLIPPING_EPS) 
                || depth > GF_BEZIER_CLIPPING_MAX_DEPTH/2)
            {
                gfRayBezierBoxInfo* result = new gfRayBezierBoxInfo;
                //result->m_inters.push_back(gfPoint3((umax+umin)/2.0, (vclipmax+vclipmin)/2.0, (wmax+wmin)/2.0));
                result->m_inters.push_back(gfPoint3((umax+umin)/2.0, vclipmin, (wmax+wmin)/2.0));
                result->m_inters.push_back(gfPoint3((umax+umin)/2.0, vclipmax, (wmax+wmin)/2.0));
                return result;
            }

            // check if no significant change in interval happend
            if ((vclipmax-vclipmin)/(vmax-vmin) > GF_BEZIER_CLIPPING_DIVIDE_TRESHOLD)
            {
                gfObjNURBSBox* box1 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
                gfObjNURBSBox* box2 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
                DecomposeVDeCasteljau((vmax+vmin)/2.0, box1, box2);
                gfRayBezierBoxInfo* res1 = box1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfRayBezierBoxInfo* res2 = box2->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfDeleteGeomObject(box1);
                gfDeleteGeomObject(box2);

                // combine two results
                if (res1 && res2)
                {
                    for (gfInt i = 0;i < (gfInt)res2->m_inters.size();i++)
                        res1->m_inters.push_back(res2->m_inters.at(i));
                    delete res2;
                    return res1;
                }
                if (res1 && !res2)
                    return res1;
                if (res2 && !res1)
                    return res2;
                return NULL;
            }

            // divide patch using deCasteljau
            gfObjNURBSBox* box1 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
            gfObjNURBSBox* box2 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
            gfObjNURBSBox* box3 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
            DecomposeVDeCasteljau(vclipmin, box1, box2);
            box2->DecomposeVDeCasteljau(vclipmax, box1, box3);

            gfDeleteGeomObject(box2);
            gfDeleteGeomObject(box3);

            gfRayBezierBoxInfo* res = box1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);

            gfDeleteGeomObject(box1);

            return res;
        }
        else if (dir == 3)
        {
            // clipping in v direction
            for (gfInt i = 0;i <= n;i++)
                for (gfInt j = 0;j <= m;j++)
                    for (gfInt k = 0;k <= o;k++)
                    {
                        aux[i+j*(n+1)+k*(n+1)*(m+1)] = gfPoint2(wmin+(wmax-wmin)*((gfDouble) k)/((gfDouble) o), auxD.at(i, j, k));
                    }

            // find better approximation of intersection region
            gfDouble wclipmin, wclipmax;
            if (gfAxisXConvexHullInter2D(&aux, &wclipmin, &wclipmax) != GF_S_OK)
                return NULL;
            if (wclipmin < wmin || wclipmax > wmax)
                return NULL;

            // check if region is small
            if (((wclipmax-wclipmin) < GF_BEZIER_CLIPPING_EPS && (umax-umin) < GF_BEZIER_CLIPPING_EPS && (vmax-vmin) < GF_BEZIER_CLIPPING_EPS) 
                || depth > GF_BEZIER_CLIPPING_MAX_DEPTH/2)
            {
                gfRayBezierBoxInfo* result = new gfRayBezierBoxInfo;
                //result->m_inters.push_back(gfPoint3((umax+umin)/2.0, (vmax+vmin)/2.0, (wclipmax+wclipmin)/2.0));
                result->m_inters.push_back(gfPoint3((umax+umin)/2.0, (vmax+vmin)/2.0, wclipmin));
                result->m_inters.push_back(gfPoint3((umax+umin)/2.0, (vmax+vmin)/2.0, wclipmax));
                return result;
            }

            // check if no significant change in interval happend
            if ((wclipmax-wclipmin)/(wmax-wmin) > GF_BEZIER_CLIPPING_DIVIDE_TRESHOLD)
            {
                gfObjNURBSBox* box1 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
                gfObjNURBSBox* box2 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
                DecomposeWDeCasteljau((wmax+wmin)/2.0, box1, box2);
                gfRayBezierBoxInfo* res1 = box1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfRayBezierBoxInfo* res2 = box2->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfDeleteGeomObject(box1);
                gfDeleteGeomObject(box2);

                // combine two results
                if (res1 && res2)
                {
                    for (gfInt i = 0;i < (gfInt)res2->m_inters.size();i++)
                        res1->m_inters.push_back(res2->m_inters.at(i));
                    delete res2;
                    return res1;
                }
                if (res1 && !res2)
                    return res1;
                if (res2 && !res1)
                    return res2;
                return NULL;
            }

            // divide patch using deCasteljau
            gfObjNURBSBox* box1 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
            gfObjNURBSBox* box2 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
            gfObjNURBSBox* box3 = (gfObjNURBSBox*) gfCreateGeomObject(GF_NURBS_BOX);
            DecomposeWDeCasteljau(wclipmin, box1, box2);
            box2->DecomposeWDeCasteljau(wclipmax, box1, box3);

            gfDeleteGeomObject(box2);
            gfDeleteGeomObject(box3);

            gfRayBezierBoxInfo* res = box1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);

            gfDeleteGeomObject(box1);

            return res;
        }

    }

    return NULL;
}

gfObjNURBSBox* gfObjNURBSBox::AddContinuousBoxU(gfInt order)
{
    // now just for bezier
    if (!IsBezier())
        return NULL;

    gfObjNURBSBox* box = (gfObjNURBSBox*)gfCreateGeomObject(GF_NURBS_BOX);

    box->InitBezier(m_num_u_control_points, m_num_v_control_points, m_num_w_control_points, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0);

    for (gfInt j = 0;j < m_num_v_control_points;j++)
    {
        for (gfInt k = 0;k < m_num_w_control_points;k++)
        {
            // do continuity criteria for bezier curves
            for (gfInt i = 0;i < m_num_u_control_points;i++)
            {
                gfPoint3* pt = box->GetControlPoint(i, j, k);

                gfDouble x = 0;
                gfDouble y = 0;
                gfDouble z = 0;
                gfDouble w = 0;

                for (gfInt l = 0;l <= i;l++)
                {
                    gfPoint3* pt2 = GetControlPoint(m_num_u_control_points-1-i+l,j,k);
                    gfDouble weight = GetWeight(m_num_u_control_points-1-i+l,j,k);
                    gfDouble poly = gfBernsteinPoly(i, l, 2);

                    x += weight * pt2->X() * poly;
                    y += weight * pt2->Y() * poly;
                    z += weight * pt2->Z() * poly;
                    w += weight * poly;
                }

                if (i > order)
                {
                    x += gfRand(-0.01, 0.01);
                    y += gfRand(-0.01, 0.01);
                    z += gfRand(-0.01, 0.01);
                }

                pt->Set(x / w, y / w, z / w);
                box->SetWeight(i, j, k, w);
            }
        }
    }

    return box;
}

GFRESULT gfObjNURBSBox::RaiseDegreeU()
{
    if (!IsBezier())
        return GF_E_FAIL;

    gfObjNURBSBox* new_bezier = (gfObjNURBSBox*)gfCreateGeomObject(GF_NURBS_BOX);
    new_bezier->InitBezier(m_num_u_control_points+1, m_num_v_control_points, m_num_w_control_points);
    gfInt n = m_num_u_control_points-1;
    gfDouble inv = 1.0 / (n+1);

    for (gfInt j = 0;j < m_num_v_control_points;j++)
    {
        for (gfInt k = 0;k < m_num_w_control_points;k++)
        {
            for (gfInt i = 0;i < m_num_u_control_points+1;i++)
            {
                gfPoint3* pt1,* pt2;
                gfDouble w1, w2;
                if (i > 0)
                {
                    pt1 = GetControlPoint(i-1,j,k);
                    w1 = GetWeight(i-1,j,k);
                }
                else
                {
                    pt1 = GetControlPoint(i,j,k);
                    w1 = GetWeight(i,j,k);
                }
                if (i <= n)
                {
                    pt2 = GetControlPoint(i,j,k);
                    w2 = GetWeight(i,j,k);
                }
                else
                {
                    pt2 = GetControlPoint(i-1,j,k);
                    w2 = GetWeight(i-1,j,k);
                }

                gfDouble x = w1 * pt1->X() * i * inv + w2 * pt2->X() * (n + 1 - i) * inv;
                gfDouble y = w1 * pt1->Y() * i * inv + w2 * pt2->Y() * (n + 1 - i) * inv;
                gfDouble z = w1 * pt1->Z() * i * inv + w2 * pt2->Z() * (n + 1 - i) * inv;
                gfDouble w = w1 * i * inv + w2 * (n + 1 - i) * inv;

                new_bezier->GetControlPoint(i, j, k)->Set(x / w, y / w, z / w);
                new_bezier->SetWeight(i, j, k, w);
            }
        }
    }

    // replace 
    InitBezier(m_num_u_control_points+1, m_num_v_control_points, m_num_w_control_points);
    for (gfInt j = 0;j < m_num_v_control_points;j++)
    {
        for (gfInt k = 0;k < m_num_w_control_points;k++)
        {
            for (gfInt i = 0;i < m_num_u_control_points;i++)
            {
                gfPoint3* pt = new_bezier->GetControlPoint(i, j, k);
                GetControlPoint(i, j, k)->Set(pt->X(), pt->Y(), pt->Z());
                SetWeight(i, j, k, new_bezier->GetWeight(i, j, k));
            }
        }
    }

    gfDeleteGeomObject(new_bezier);
    m_time_stamp++;
    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::RaiseDegreeV()
{
    if (!IsBezier())
        return GF_E_FAIL;

    gfObjNURBSBox* new_bezier = (gfObjNURBSBox*)gfCreateGeomObject(GF_NURBS_BOX);
    new_bezier->InitBezier(m_num_u_control_points, m_num_v_control_points+1, m_num_w_control_points);
    gfInt n = m_num_v_control_points-1;
    gfDouble inv = 1.0 / (n+1);

    for (gfInt i = 0;i < m_num_u_control_points;i++)
    {
        for (gfInt k = 0;k < m_num_w_control_points;k++)
        {
            for (gfInt j = 0;j < m_num_v_control_points+1;j++)
            {
                gfPoint3* pt1,* pt2;
                gfDouble w1, w2;
                if (j > 0)
                {
                    pt1 = GetControlPoint(i,j-1,k);
                    w1 = GetWeight(i,j-1,k);
                }
                else
                {
                    pt1 = GetControlPoint(i,j,k);
                    w1 = GetWeight(i,j,k);
                }
                if (j <= n)
                {
                    pt2 = GetControlPoint(i,j,k);
                    w2 = GetWeight(i,j,k);
                }
                else
                {
                    pt2 = GetControlPoint(i,j-1,k);
                    w2 = GetWeight(i,j-1,k);
                }

                gfDouble x = w1 * pt1->X() * j * inv + w2 * pt2->X() * (n + 1 - j) * inv;
                gfDouble y = w1 * pt1->Y() * j * inv + w2 * pt2->Y() * (n + 1 - j) * inv;
                gfDouble z = w1 * pt1->Z() * j * inv + w2 * pt2->Z() * (n + 1 - j) * inv;
                gfDouble w = w1 * j * inv + w2 * (n + 1 - j) * inv;

                new_bezier->GetControlPoint(i, j, k)->Set(x / w, y / w, z / w);
                new_bezier->SetWeight(i, j, k, w);
            }
        }
    }

    // replace 
    InitBezier(m_num_u_control_points, m_num_v_control_points+1, m_num_w_control_points);
    for (gfInt j = 0;j < m_num_v_control_points;j++)
    {
        for (gfInt k = 0;k < m_num_w_control_points;k++)
        {
            for (gfInt i = 0;i < m_num_u_control_points;i++)
            {
                gfPoint3* pt = new_bezier->GetControlPoint(i, j, k);
                GetControlPoint(i, j, k)->Set(pt->X(), pt->Y(), pt->Z());
                SetWeight(i, j, k, new_bezier->GetWeight(i, j, k));
            }
        }
    }

    gfDeleteGeomObject(new_bezier);
    m_time_stamp++;
    return GF_S_OK;
}

GFRESULT gfObjNURBSBox::RaiseDegreeW()
{
    if (!IsBezier())
        return GF_E_FAIL;

    gfObjNURBSBox* new_bezier = (gfObjNURBSBox*)gfCreateGeomObject(GF_NURBS_BOX);
    new_bezier->InitBezier(m_num_u_control_points, m_num_v_control_points, m_num_w_control_points+1);
    gfInt n = m_num_w_control_points-1;
    gfDouble inv = 1.0 / (n+1);

    for (gfInt j = 0;j < m_num_v_control_points;j++)
    {
        for (gfInt i = 0;i < m_num_u_control_points;i++)
        {
            for (gfInt k = 0;k < m_num_w_control_points+1;k++)
            {
                gfPoint3* pt1,* pt2;
                gfDouble w1, w2;
                if (k > 0)
                {
                    pt1 = GetControlPoint(i,j,k-1);
                    w1 = GetWeight(i,j,k-1);
                }
                else
                {
                    pt1 = GetControlPoint(i,j,k);
                    w1 = GetWeight(i,j,k);
                }
                if (k <= n)
                {
                    pt2 = GetControlPoint(i,j,k);
                    w2 = GetWeight(i,j,k);
                }
                else
                {
                    pt2 = GetControlPoint(i,j,k-1);
                    w2 = GetWeight(i,j,k-1);
                }

                gfDouble x = w1 * pt1->X() * k * inv + w2 * pt2->X() * (n + 1 - k) * inv;
                gfDouble y = w1 * pt1->Y() * k * inv + w2 * pt2->Y() * (n + 1 - k) * inv;
                gfDouble z = w1 * pt1->Z() * k * inv + w2 * pt2->Z() * (n + 1 - k) * inv;
                gfDouble w = w1 * k * inv + w2 * (n + 1 - k) * inv;

                new_bezier->GetControlPoint(i, j, k)->Set(x / w, y / w, z / w);
                new_bezier->SetWeight(i, j, k, w);
            }
        }
    }

    // replace 
    InitBezier(m_num_u_control_points, m_num_v_control_points, m_num_w_control_points+1);
    for (gfInt j = 0;j < m_num_v_control_points;j++)
    {
        for (gfInt k = 0;k < m_num_w_control_points;k++)
        {
            for (gfInt i = 0;i < m_num_u_control_points;i++)
            {
                gfPoint3* pt = new_bezier->GetControlPoint(i, j, k);
                GetControlPoint(i, j, k)->Set(pt->X(), pt->Y(), pt->Z());
                SetWeight(i, j, k, new_bezier->GetWeight(i, j, k));
            }
        }
    }

    gfDeleteGeomObject(new_bezier);
    m_time_stamp++;
    return GF_S_OK;
}