//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjNURBSBox.h                                      //
//   Description: NURBS box class                               //
//                                                              //
//////////////////////////////////////////////////////////////////


#ifndef _GF_C_NURBS_BOX_H_
#define _GF_C_NURBS_BOX_H_


#include "gfDefines.h"
#include <vector>
#include "gfAux3DArray.h"
#include "gfPoint.h"
#include "gfVector.h"
#include "gfMatrix.h"
#include "gfObjGeomObject.h"
#include "gfObjNURBSCurve.h"
#include "gfObjNURBSPatch.h"
#include "gfObjHalfEdgeMesh.h"


class gfObjNURBSBox;

struct gfBezier3AABBTreeNode
{
    gfAuxAABB* aabb;
    gfBezier3AABBTreeNode* left;
    gfBezier3AABBTreeNode* right;
    gfObjNURBSBox* box;
    gfInt min_dir;
    gfDouble jacobi_min;
    gfDouble jacobi_max;
};

struct gfBezier3AABBQueueUnit
{
    gfBezier3AABBTreeNode* node;
    gfDouble tmin;
    gfDouble tmax;
};

struct gfRayBezierBoxInfo
{
    std::vector<gfPoint3> m_inters;
};


class gfObjNURBSBox : public gfObjGeomObject
{
private:

    std::vector<gfDouble> m_u_knot_vector;
    std::vector<gfDouble> m_v_knot_vector;
    std::vector<gfDouble> m_w_knot_vector;
    std::vector<gfDouble> m_weights;
    std::vector<gfPoint3 *> m_control_points;
    gfInt m_num_u_control_points;
    gfInt m_num_v_control_points;
    gfInt m_num_w_control_points;
    gfInt m_u_degree;
    gfInt m_v_degree;
    gfInt m_w_degree;

    // vertex array cache
    gfDouble* m_cache_vertices;
    gfDouble* m_cache_normals;
    gfDouble* m_cache_wire_vertices;
    gfInt m_num_cached_quads;
    gfInt m_num_cached_wire_lines;
    GFRESULT CreateVACache(gfInt approx);
    GFRESULT DeleteVACache(void);
    GFRESULT DrawVACache(gfLong flags);

    // Bezier AABB tree functions
    struct compare_queue_unit
    {
        bool operator() (const gfBezier3AABBQueueUnit & x, const gfBezier3AABBQueueUnit & y)
        {
            return x.tmin > y.tmin;
        }
    };
    std::priority_queue<gfBezier3AABBQueueUnit, std::vector<gfBezier3AABBQueueUnit>, compare_queue_unit> m_aabb_queue;
    gfObjNURBSPatch* m_boundary_patches[6];
    gfBool m_computed_boundaries;

    std::vector<gfBezier3AABBTreeNode*> m_bezier_aabb_trees;
    gfRayInterInfo* RayAABBIntersection(gfBezier3AABBTreeNode* root, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfBezier3AABBTreeNode* CreateAABBTree(gfDouble depth, gfObjNURBSBox* patch);
    GFRESULT ComputeWholeAABBTree(void);
    GFRESULT DeleteWholeAABBTree(void);
    GFRESULT DeleteAABBTree(gfBezier3AABBTreeNode* tree_root);
    gfInt m_num_nodes;

public:

    gfObjNURBSBox(void);
    ~gfObjNURBSBox(void);

    // gfObjGeomObject functions
    GFRESULT OGLDraw(gfLong flags);
    gfRayInterInfo* RayIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfGeomObjID GetID(void);
    GFRESULT Transform(gfMatrix4* matrix);
    gfObjGeomObject* Clone(void);
    gfAuxBoundingVolume* GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix);

    // initialize NURBS box
    GFRESULT Clear(void);
    GFRESULT InitBezier(gfInt num_u_points, gfInt num_v_points, gfInt num_w_points);
    GFRESULT InitBezier(gfInt num_u_points, gfInt num_v_points, gfInt num_w_points,
        gfDouble u_min, gfDouble u_max, gfDouble v_min, gfDouble v_max, gfDouble w_min, gfDouble w_max);
    GFRESULT InitBox(gfPoint3* center, gfDouble tx, gfDouble ty, gfDouble tz);
    GFRESULT InitSphere(gfPoint3* center, gfDouble radius);
    GFRESULT InitTorus(gfPoint3* center, gfDouble radius1, gfDouble radius2);
    GFRESULT InitCone(gfPoint3* center, gfDouble radius1, gfDouble radius2, gfDouble height);
    GFRESULT InitRandom(gfPoint3* center, gfDouble tx, gfDouble ty, gfDouble tz);
    GFRESULT InitTransitional(gfObjNURBSCurve* nurbs_curve, gfObjNURBSPatch* nurbs_patch);
    GFRESULT InitRotational(gfPoint3* center, gfVector3* axis, gfObjNURBSPatch* nurbs_patch);
    GFRESULT InitRotational(gfPoint3* center, gfVector3* axis, gfObjNURBSCurve* nurbs_curve);
    GFRESULT InitRotational(gfPoint3* center, gfVector3* axis, gfDouble angle, gfObjNURBSPatch* nurbs_patch);
    GFRESULT InitRotational(gfPoint3* center, gfVector3* axis, gfDouble angle, gfObjNURBSCurve* nurbs_curve);
    GFRESULT InitProfile(gfDouble xmin, gfDouble xmax, gfInt num_u_points, 
                         gfDouble ymin, gfDouble ymax, gfInt num_v_points,
                         gfDouble zmin, gfDouble zmax, gfInt num_w_points);
    GFRESULT InitRuledVolume(gfObjNURBSPatch* nurbs_patch1, gfObjNURBSPatch* nurbs_patch2);

    // setting a retrieving parameters of NURBS box
    gfInt SetParams(gfInt u_degree, gfInt v_degree, gfInt w_degree, gfInt num_u_control_points, gfInt num_v_control_points, gfInt num_w_control_points);
    gfInt GetUDegree(void);
    gfInt GetVDegree(void);
    gfInt GetWDegree(void);
    gfInt GetNumUKnots(void);
    gfInt GetNumVKnots(void);
    gfInt GetNumWKnots(void);
    gfDouble GetUKnot(gfInt index);
    gfDouble GetVKnot(gfInt index);
    gfDouble GetWKnot(gfInt index);
    GFRESULT SetUKnot(gfInt index, gfDouble knot);
    GFRESULT SetVKnot(gfInt index, gfDouble knot);
    GFRESULT SetWKnot(gfInt index, gfDouble knot);
    gfInt GetNumUControlPoints(void);
    gfInt GetNumVControlPoints(void);
    gfInt GetNumWControlPoints(void);
    gfPoint3 * GetControlPoint(gfInt u_index, gfInt v_index, gfInt w_index);
    gfDouble GetWeight(gfInt u_index, gfInt v_index, gfInt w_index);
    GFRESULT SetWeight(gfInt u_index, gfInt v_index, gfInt w_index, gfDouble weight);
    GFRESULT GetDomain(gfDouble* u_min, gfDouble* u_max, gfDouble* v_min, gfDouble* v_max, gfDouble* w_min, gfDouble* w_max);

    // modelling functions
    gfBool IsBezier(void);
    gfPoint3* GetVolumePoint(gfDouble u, gfDouble v, gfDouble w);
    gfVector3* GetPatchUDer(gfDouble u, gfDouble v, gfDouble w);
    gfVector3* GetPatchVDer(gfDouble u, gfDouble v, gfDouble w);
    gfVector3* GetPatchWDer(gfDouble u, gfDouble v, gfDouble w);
    gfVector3* GetNormal(gfDouble u, gfDouble v, gfDouble w);
    GFRESULT DecomposeToBezierSegments(gfAux3DArray<gfObjNURBSBox*>* segments);
    GFRESULT DecomposeUDeCasteljau(gfDouble u, gfObjNURBSBox* box1, gfObjNURBSBox* box2);
    GFRESULT DecomposeVDeCasteljau(gfDouble v, gfObjNURBSBox* box1, gfObjNURBSBox* box2);
    GFRESULT DecomposeWDeCasteljau(gfDouble w, gfObjNURBSBox* box1, gfObjNURBSBox* box2);
    gfObjNURBSPatch* GetUVIsopatch(gfDouble w_param);
    gfObjNURBSPatch* GetUWIsopatch(gfDouble v_param);
    gfObjNURBSPatch* GetVWIsopatch(gfDouble u_param);
    gfObjNURBSBox* AddContinuousBoxU(gfInt order);
    GFRESULT RaiseDegreeU();
    GFRESULT RaiseDegreeV();
    GFRESULT RaiseDegreeW();

    // Bezier clipping
    gfRayBezierBoxInfo* BezierClipping(gfMatrix4* obj_matrix, gfInt depth,
                                         gfDouble a1, gfDouble b1, gfDouble c1, gfDouble d1,
                                         gfDouble a2, gfDouble b2, gfDouble c2, gfDouble d2);
} ;


#endif