//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjNURBSCurve.cpp                                  //
//   Description: NURBS curve class                             //
//                From "Piegl, Tiller: The NURBS Book"          //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfObjNURBSCurve.h"
#include "gfObjectFactory.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfAuxItersectionUtils.h"
#include "gfAuxLogging.h"
#include "gfWinFPSCounter.h"

#include "./OGL/glut.h"

gfObjNURBSCurve::gfObjNURBSCurve(void)
{
    m_time_stamp = 0;
    m_weights.resize(0);
    m_control_points.resize(0);
    m_knot_vector.resize(0);
    m_degree = 0;
}

gfObjNURBSCurve::~gfObjNURBSCurve(void)
{
    DeleteWholeAABBTree();
    SetParams(0, 0);
}


GFRESULT gfObjNURBSCurve::ElevateDegree(void)
{
    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::OGLDraw(gfLong flags)
{
    if (m_time_stamp == 0)
        return GF_S_OK;

    // interval of definition
    gfDouble a = GetKnot(m_degree);
    gfDouble b = GetKnot(GetNumControlPoints());
    if (gfIsZero(b - a))
        return GF_S_OK;

    gfInt n = GetNumControlPoints()  - 1;
    gfInt degree = m_degree;

    if ((flags & GF_FLG_WIRE) || (flags & GF_FLG_SOLID))
    {
        glLineWidth(5.0);
        glBegin(GL_LINE_STRIP);
        for (gfInt i = 0;i < 10*GF_VIS_PREC;i++)
        {
            // current parameter
            gfDouble u = a + i * (b - a) / 10.0 / GF_VIS_PREC;

            gfInt span = gfGetBSplineKnotSpan(n, degree, &m_knot_vector, u);

            std::vector<gfDouble> basis;
            gfGetBSplineAllValues(degree, &m_knot_vector, span, u, &basis);

            gfDouble ax = 0;
            gfDouble ay = 0;
            gfDouble az = 0;
            gfDouble aw = 0;

            for (gfInt j = 0;j <= degree;j++)
            {
                gfInt index = span - degree + j;
                ax += basis.at(j) * m_weights.at(index) * m_control_points.at(index)->X();
                ay += basis.at(j) * m_weights.at(index) * m_control_points.at(index)->Y();
                az += basis.at(j) * m_weights.at(index) * m_control_points.at(index)->Z();
                aw += basis.at(j) * m_weights.at(index);
            }

            glVertex3f(ax / aw, ay / aw, az / aw);
        }
        glEnd();
        glLineWidth(1.0);
    }

    if (flags & GF_FLG_CONTROL_NET)
    {
        glLineWidth(2.0);
        glColor3f(0.0, 0.0, 0.0);
        glBegin(GL_LINE_STRIP);
        for (gfInt i = 0;i < (gfInt)m_control_points.size();i++)
        {
            gfPoint3* pt = GetControlPoint(i);
            glVertex3f(pt->X(), pt->Y(), pt->Z());
        }
        glEnd();
        glLineWidth(1.0);

        glColor3f(0.0, 0.0, 0.0);
        glPointSize(8.0);
        glBegin(GL_POINTS);
        for (gfInt i = 0;i < (gfInt)m_control_points.size();i++)
        {
            gfPoint3* pt = GetControlPoint(i);
            glVertex3f(pt->X(), pt->Y(), pt->Z());
        }
        glEnd();
        glPointSize(1.0);
    }

    return GF_S_OK;
}

gfRayInterInfo* gfObjNURBSCurve::RayIntersection(gfMatrix4* obj_matrix, gfPoint3 * ray_begin, gfVector3* ray_dir)
{
    gfInt num_trees = (gfInt)m_bezier_aabb_trees.size();
    if (num_trees == 0)
        ComputeWholeAABBTree();
    num_trees = (gfInt)m_bezier_aabb_trees.size();

    gfDisableLogging();

    gfRayInterInfo* info = NULL;
    for (gfInt i = 0;i < num_trees;i++)
    {
        gfBezier1AABBTreeNode* root = m_bezier_aabb_trees.at(i);
        gfRayInterInfo* info_temp = RayAABBIntersection(root, ray_begin, ray_dir);
        if (!info)
            info = info_temp;
        else if (info_temp)
        {
            // copy info
            for (gfInt i = 0;i < (gfInt)info_temp->m_inters.size();i++)
            {
                info->m_inters.push_back(info_temp->m_inters.at(i));
                info->m_normals.push_back(info_temp->m_normals.at(i));
            }
            gfDeleteRayInterInfo(info_temp);
            gfSort1DArray(&info->m_inters);
        }
    }

    gfEnableLogging();

    return info;
}

gfGeomObjID gfObjNURBSCurve::GetID(void)
{
    return GF_NURBS_CURVE;
}

gfObjGeomObject* gfObjNURBSCurve::Clone(void)
{
    // create new curve
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());
    gfObjNURBSCurve* curve = (gfObjNURBSCurve*) clone;

    // set parameters of curve
    gfInt num_control_points = GetNumControlPoints();
    curve->SetParams(m_degree, num_control_points);

    // copy control points and weights
    for (gfInt i = 0;i < num_control_points;i++)
    {
        gfPoint3 * new_point = curve->GetControlPoint(i);
        gfPoint3 * old_point = GetControlPoint(i);
        new_point->Set(old_point->X(), old_point->Y(), old_point->Z());
        curve->SetWeight(i, GetWeight(i));
    }

    // copy control points
    gfInt num_knots = GetNumKnots();
    for (gfInt i = 0;i < num_knots;i++)
    {
        curve->SetKnot(i, GetKnot(i));
    }

    return clone;
}

GFRESULT gfObjNURBSCurve::Transform(gfMatrix4* matrix)
{
    // delete AABB tree
    DeleteWholeAABBTree();

    // all we need is transform control points
    gfInt size = (gfInt)m_control_points.size();
    for (gfInt i = 0; i < size; i++)
    {
        gfPoint3* point = m_control_points.at(i);
        if (point)
        {
            gfTransformPoint3(point, matrix);
        }
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::InitHalfCircle(gfPoint3* center, gfPoint3* start, gfVector3* normal)
{
    SetParams(2, 5);

    SetKnot(0, 0.0);SetKnot(1, 0.0);SetKnot(2, 0.0);
    SetKnot(3, 0.5);SetKnot(4, 0.5);
    SetKnot(5, 1.0);SetKnot(6, 1.0);SetKnot(7, 1.0);

    gfVector3* base1 = gfSubstractPoints3(start, center);
    gfVector3* base2 = gfVectorProduct3(normal, base1);
    base2->SetLength(base1->Length());

    gfVector3* aux_vec = gfCreateVector3();
    gfDouble sqrt2 = gfSquareRoot(2.0) / 2.0;

    gfPoint3* aux_point;

    gfCombineVectors3(1.0, base1, 0.0, base2, aux_vec);
    aux_point = GetControlPoint(0);
    gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
    SetWeight(0, 1.0);

    gfCombineVectors3(1.0, base1, 1.0, base2, aux_vec);
    aux_point = GetControlPoint(1);
    gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
    SetWeight(1, sqrt2);

    gfCombineVectors3(0.0, base1, 1.0, base2, aux_vec);
    aux_point = GetControlPoint(2);
    gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
    SetWeight(2, 1.0);

    gfCombineVectors3(-1.0, base1, 1.0, base2, aux_vec);
    aux_point = GetControlPoint(3);
    gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
    SetWeight(3, sqrt2);

    gfCombineVectors3(-1.0, base1, 0.0, base2, aux_vec);
    aux_point = GetControlPoint(4);
    gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
    SetWeight(4, 1.0);

    gfDeleteVector3(base1);
    gfDeleteVector3(base2);
    gfDeleteVector3(aux_vec);

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::InitCircle(gfPoint3 * center, gfPoint3 * start, gfVector3* normal)
{
    SetParams(2, 9);

    SetKnot(0, 0.0);SetKnot(1, 0.0);SetKnot(2, 0.0);
    SetKnot(3, 0.25);SetKnot(4, 0.25);
    SetKnot(5, 0.5);SetKnot(6, 0.5);
    SetKnot(7, 0.75);SetKnot(8, 0.75);
    SetKnot(9, 1.0);SetKnot(10, 1.0);SetKnot(11, 1.0);

    gfVector3 base1;
    gfSubstractPoints3(start, center, &base1);
    gfVector3 base2;
    gfVectorProduct3(normal, &base1, &base2);
    base2.SetLength(base1.Length());

    gfVector3 aux_vec;
    gfDouble sqrt2 = gfSquareRoot(2.0) / 2.0;
    gfPoint3* aux_point;

    gfCombineVectors3(1.0, &base1, 0.0, &base2, &aux_vec);
    aux_point = GetControlPoint(0);
    gfCombinePointVector3(center, 1.0, &aux_vec, aux_point);
    SetWeight(0, 1.0);

    gfCombineVectors3(1.0, &base1, 1.0, &base2, &aux_vec);
    aux_point = GetControlPoint(1);
    gfCombinePointVector3(center, 1.0, &aux_vec, aux_point);
    SetWeight(1, sqrt2);

    gfCombineVectors3(0.0, &base1, 1.0, &base2, &aux_vec);
    aux_point = GetControlPoint(2);
    gfCombinePointVector3(center, 1.0, &aux_vec, aux_point);
    SetWeight(2, 1.0);

    gfCombineVectors3(-1.0, &base1, 1.0, &base2, &aux_vec);
    aux_point = GetControlPoint(3);
    gfCombinePointVector3(center, 1.0, &aux_vec, aux_point);
    SetWeight(3, sqrt2);

    gfCombineVectors3(-1.0, &base1, 0.0, &base2, &aux_vec);
    aux_point = GetControlPoint(4);
    gfCombinePointVector3(center, 1.0, &aux_vec, aux_point);
    SetWeight(4, 1.0);

    gfCombineVectors3(-1.0, &base1, -1.0, &base2, &aux_vec);
    aux_point = GetControlPoint(5);
    gfCombinePointVector3(center, 1.0, &aux_vec, aux_point);
    SetWeight(5, sqrt2);

    gfCombineVectors3(0.0, &base1, -1.0, &base2, &aux_vec);
    aux_point = GetControlPoint(6);
    gfCombinePointVector3(center, 1.0, &aux_vec, aux_point);
    SetWeight(6, 1.0);

    gfCombineVectors3(1.0, &base1, -1.0, &base2, &aux_vec);
    aux_point = GetControlPoint(7);
    gfCombinePointVector3(center, 1.0, &aux_vec, aux_point);
    SetWeight(7, sqrt2);

    gfCombineVectors3(1.0, &base1, 0.0, &base2, &aux_vec);
    aux_point = GetControlPoint(8);
    gfCombinePointVector3(center, 1.0, &aux_vec, aux_point);
    SetWeight(8, 1.0);

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::InitCircularArc(gfPoint3* center, gfPoint3* start, gfVector3* normal, gfDouble angle)
{
    gfInt num_segs = gfInt(2.0 * angle / GF_PI);
    if (num_segs == 0)
        num_segs = 1;
    gfDouble seg_angle = angle / gfDouble(num_segs);
    gfInt num_control_points = 2 * num_segs + 1;

    gfVector3 base1;
    gfSubstractPoints3(start, center, &base1);
    gfVector3 base2;
    gfVectorProduct3(normal, &base1, &base2);
    base2.SetLength(base1.Length());

    // set arc parameters
    SetParams(2, num_control_points);

    gfDouble weight = gfCos(seg_angle/2.0);

    // set control points
    for (gfInt i = 0;i < num_control_points;i++)
    {
        gfPoint3* ctrl_point = GetControlPoint(i);

        if (i % 2 == 0)
        {
            gfDouble aux_sin = gfSin(seg_angle*i/2.0);
            gfDouble aux_cos = gfCos(seg_angle*i/2.0);

            ctrl_point->SetX(center->X() + aux_cos * base1.X() + aux_sin * base2.X());
            ctrl_point->SetY(center->Y() + aux_cos * base1.Y() + aux_sin * base2.Y());
            ctrl_point->SetZ(center->Z() + aux_cos * base1.Z() + aux_sin * base2.Z());

            SetWeight(i, 1.0);
        }
        else
        {
            gfDouble aux_sin1 = gfSin(seg_angle*(i - 1)/2.0);
            gfDouble aux_cos1 = gfCos(seg_angle*(i - 1)/2.0);
            gfDouble aux_sin2 = gfSin(seg_angle*(i + 1)/2.0);
            gfDouble aux_cos2 = gfCos(seg_angle*(i + 1)/2.0);
            gfPoint3 v1, v2;
            gfVector3 t1, t2;

            v1.Set(center->X() + aux_cos1 * base1.X() + aux_sin1 * base2.X(),
                    center->Y() + aux_cos1 * base1.Y() + aux_sin1 * base2.Y(),
                    center->Z() + aux_cos1 * base1.Z() + aux_sin1 * base2.Z());

            v2.Set(center->X() + aux_cos2 * base1.X() + aux_sin2 * base2.X(),
                    center->Y() + aux_cos2 * base1.Y() + aux_sin2 * base2.Y(),
                    center->Z() + aux_cos2 * base1.Z() + aux_sin2 * base2.Z());

            t1.Set(-aux_sin1 * base1.X() + aux_cos1 * base2.X(),
                    -aux_sin1 * base1.Y() + aux_cos1 * base2.Y(),
                    -aux_sin1 * base1.Z() + aux_cos1 * base2.Z());

            t2.Set(-aux_sin2 * base1.X() + aux_cos2 * base2.X(),
                    -aux_sin2 * base1.Y() + aux_cos2 * base2.Y(),
                    -aux_sin2 * base1.Z() + aux_cos2 * base2.Z());

            gfLinesIntersection(&v1, &t1, &v2, &t2, ctrl_point);
            SetWeight(i, weight);
        }
    }

    // set knots of arc
    switch (num_segs)
    {
    case 1:
        SetKnot(0, 0.0);SetKnot(1, 0.0);SetKnot(2, 0.0);
        SetKnot(3, 1.0);SetKnot(4, 1.0);SetKnot(5, 1.0);
        break;
    case 2:
        SetKnot(0, 0.0);SetKnot(1, 0.0);SetKnot(2, 0.0);
        SetKnot(3, 0.5);SetKnot(4, 0.5);
        SetKnot(5, 1.0);SetKnot(6, 1.0);SetKnot(7, 1.0);
        break;
    case 3:
        SetKnot(0, 0.0);SetKnot(1, 0.0);SetKnot(2, 0.0);
        SetKnot(3, 0.33333);SetKnot(4, 0.33333);
        SetKnot(5, 0.66666);SetKnot(6, 0.66666);
        SetKnot(7, 1.0);SetKnot(8, 1.0);SetKnot(9, 1.0);
        break;
    case 4:
        SetKnot(0, 0.0);SetKnot(1, 0.0);SetKnot(2, 0.0);
        SetKnot(3, 0.25);SetKnot(4, 0.25);
        SetKnot(5, 0.5);SetKnot(6, 0.5);
        SetKnot(7, 0.75);SetKnot(8, 0.75);
        SetKnot(9, 1.0);SetKnot(10, 1.0);SetKnot(11, 1.0);
        break;
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::InitSpiral(gfPoint3* center, gfPoint3* start, gfVector3* normal, gfInt num_twists, gfDouble twist_elevation)
{
    SetParams(2, 8*num_twists+1);

    gfVector3* base1 = gfSubstractPoints3(start, center);
    gfVector3* base2 = gfVectorProduct3(normal, base1);
    base2->SetLength(base1->Length());

    gfVector3* aux_vec = gfCreateVector3();
    gfDouble sqrt2 = gfSquareRoot(2.0) / 2.0;

    gfInt num_knots = GetNumKnots();
    SetKnot(0, 0.0);SetKnot(1, 0.0);SetKnot(2, 0.0);
    SetKnot(num_knots - 3, 1.0);SetKnot(num_knots - 2, 1.0);SetKnot(num_knots - 1, 1.0);
    for (gfInt i = 0;i < (num_knots - 2)/2;i++)
    {
        SetKnot(2*i+1, 2.0*i/(num_knots - 4));
        SetKnot(2*i+2, 2.0*i/(num_knots - 4));
    }

    for (gfInt i = 0;i < num_twists;i++)
    {
        gfPoint3* aux_point;

        gfCombineVectors3(1.0, base1, 0.0, base2, aux_vec);
        gfCombineVectors3(1.0, aux_vec, twist_elevation*(i+0/8.0), normal, aux_vec);
        aux_point = GetControlPoint(8*i+0);
        gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
        SetWeight(8*i+0, 1.0);

        gfCombineVectors3(1.0, base1, 1.0, base2, aux_vec);
        gfCombineVectors3(1.0, aux_vec, twist_elevation*(i+1/8.0), normal, aux_vec);
        aux_point = GetControlPoint(8*i+1);
        gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
        SetWeight(8*i+1, sqrt2);

        gfCombineVectors3(0.0, base1, 1.0, base2, aux_vec);
        gfCombineVectors3(1.0, aux_vec, twist_elevation*(i+2/8.0), normal, aux_vec);
        aux_point = GetControlPoint(8*i+2);
        gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
        SetWeight(8*i+2, 1.0);

        gfCombineVectors3(-1.0, base1, 1.0, base2, aux_vec);
        gfCombineVectors3(1.0, aux_vec, twist_elevation*(i+3/8.0), normal, aux_vec);
        aux_point = GetControlPoint(8*i+3);
        gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
        SetWeight(8*i+3, sqrt2);

        gfCombineVectors3(-1.0, base1, 0.0, base2, aux_vec);
        gfCombineVectors3(1.0, aux_vec, twist_elevation*(i+4/8.0), normal, aux_vec);
        aux_point = GetControlPoint(8*i+4);
        gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
        SetWeight(8*i+4, 1.0);

        gfCombineVectors3(-1.0, base1, -1.0, base2, aux_vec);
        gfCombineVectors3(1.0, aux_vec, twist_elevation*(i+5/8.0), normal, aux_vec);
        aux_point = GetControlPoint(8*i+5);
        gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
        SetWeight(8*i+5, sqrt2);

        gfCombineVectors3(0.0, base1, -1.0, base2, aux_vec);
        gfCombineVectors3(1.0, aux_vec, twist_elevation*(i+6/8.0), normal, aux_vec);
        aux_point = GetControlPoint(8*i+6);
        gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
        SetWeight(8*i+6, 1.0);

        gfCombineVectors3(1.0, base1, -1.0, base2, aux_vec);
        gfCombineVectors3(1.0, aux_vec, twist_elevation*(i+7/8.0), normal, aux_vec);
        aux_point = GetControlPoint(8*i+7);
        gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
        SetWeight(8*i+7, sqrt2);

        if (i == (num_twists - 1))
        {
            gfCombineVectors3(1.0, base1, 0.0, base2, aux_vec);
            gfCombineVectors3(1.0, aux_vec, twist_elevation*(i+8/8.0), normal, aux_vec);
            aux_point = GetControlPoint(8*i+8);
            gfCombinePointVector3(center, 1.0, aux_vec, aux_point);
            SetWeight(8*i+8, 1.0);
        }
    }

    gfDeleteVector3(base1);
    gfDeleteVector3(base2);
    gfDeleteVector3(aux_vec);

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::InitBezier(gfInt num_points)
{
    gfInt degree = num_points - 1;
    SetParams(degree, num_points);

    gfInt num_knots = GetNumKnots();
    for (gfInt i = 0;i < num_knots;i++)
    {
        SetKnot(i, (i < num_knots / 2) ? 0.0 : 1.0);
    }

    for (gfInt i = 0;i < num_points;i++)
    {
        SetWeight(i, 1.0);
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::InitBezier(gfInt num_points, gfDouble u_min, gfDouble u_max)
{
    gfInt degree = num_points - 1;
    SetParams(degree, num_points);

    gfInt num_knots = GetNumKnots();
    for (gfInt i = 0;i < num_knots;i++)
    {
        SetKnot(i, (i < num_knots / 2) ? u_min : u_max);
    }

    for (gfInt i = 0;i < num_points;i++)
    {
        SetWeight(i, 1.0);
    }

    return GF_S_OK;
}


GFRESULT gfObjNURBSCurve::InitPiecewiseBezier(gfInt degree, gfInt num_segments)
{
    gfInt num_points = degree * num_segments + 1;
    SetParams(degree, num_points);

    for (gfInt i = 0;i <= num_segments;i++)
    {
        for (gfInt j = 0;j < degree;j++)
        {
            SetKnot(i*degree + j + 1, gfDouble(i) / gfDouble(num_segments));
        }
    }
    SetKnot(0, 0.0);
    SetKnot(num_segments*degree + degree + 1, 1.0);

    for (gfInt i = 0;i < num_points;i++)
    {
        SetWeight(i, 1.0);
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::InitPolyline(gfInt num_points)
{
    SetParams(1, num_points);
    gfInt num_knots = GetNumKnots();

    for (gfInt i = 0;i < num_knots;i++)
    {
        SetKnot(i, (gfDouble) i);
    }

    for (gfInt i = 0;i < num_points;i++)
    {
        SetWeight(i, 1.0);
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::InitRandom(gfInt degree, gfInt num_points, gfDouble xmin, gfDouble xmax)
{
    SetParams(degree, num_points);

    gfInt num_knots = GetNumKnots();
    for (gfInt i = 0;i <= degree;i++)
    {
        SetKnot(i, 0);
    }
    for (gfInt i = degree+1;i < num_knots-degree-1;i++)
    {
        SetKnot(i, (gfDouble) i-degree);
    }
    for (gfInt i = num_knots-degree-1;i < num_knots;i++)
    {
        SetKnot(i, (gfDouble) num_knots-degree-2);
    }

    for (gfInt i = 0;i < num_points;i++)
    {
        gfPoint3* pt = GetControlPoint(i);
        pt->Set(xmin+(xmax-xmin)*i/(num_points-1), gfRand(-1.0,1.0), 0.0);
        SetWeight(i, 1.0);
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::SetParams(gfInt degree, gfInt num_control_points)
{
    m_knot_vector.resize(degree + num_control_points + 1);
    m_weights.resize(num_control_points);
    m_degree = degree;

    gfInt old_num = (gfInt)m_control_points.size();

    if (old_num > num_control_points)
    {
        for (gfInt i = num_control_points;i < old_num;i++)
        {
            gfPoint3 * old_pt = m_control_points.at(i);
            gfDeletePoint3(old_pt);
        }
    }

    m_control_points.resize(num_control_points);

    if (old_num < num_control_points)
    {
        for (gfInt i = old_num;i < num_control_points;i++)
        {
            gfPoint3* new_pt = gfCreatePoint3();
            m_control_points[i] = new_pt;
        }
    }

    // set eqidistant knots
    gfInt num_knots = GetNumKnots();
    for (gfInt i = 0;i < num_knots;i++)
    {
        SetKnot(i, (gfDouble) i);
    }

    m_time_stamp++;

    return GF_S_OK;
}

gfInt gfObjNURBSCurve::GetDegree(void)
{
    return m_degree;
}

gfInt gfObjNURBSCurve::GetNumKnots(void)
{
    return (gfInt)m_knot_vector.size();
}

gfDouble gfObjNURBSCurve::GetKnot(gfInt index)
{
    return m_knot_vector.at(index);
}

GFRESULT gfObjNURBSCurve::SetKnot(gfInt index, gfDouble knot)
{
    m_time_stamp++;
    m_knot_vector[index] = knot;
    return GF_S_OK;
}

gfInt gfObjNURBSCurve::GetNumControlPoints(void)
{
    return (gfInt)m_control_points.size();
}

gfPoint3 * gfObjNURBSCurve::GetControlPoint(gfInt index)
{
    return m_control_points.at(index);
}

gfDouble gfObjNURBSCurve::GetWeight(gfInt index)
{
    return m_weights.at(index);
}

GFRESULT gfObjNURBSCurve::SetWeight(gfInt index, gfDouble weight)
{
    m_time_stamp++;
    return m_weights[index] = weight;
}

gfPoint3* gfObjNURBSCurve::GetCurvePoint(gfDouble u)
{
    gfInt n = GetNumControlPoints()-1;
    gfInt span = gfGetBSplineKnotSpan(n, m_degree, &m_knot_vector, u);

    std::vector<gfDouble> basis;
    gfGetBSplineAllValues(m_degree, &m_knot_vector, span, u, &basis);

    gfDouble ax = 0;
    gfDouble ay = 0;
    gfDouble az = 0;
    gfDouble aw = 0;

    for (gfInt j = 0;j <= m_degree;j++)
    {
        gfInt index = span - m_degree + j;
        ax += basis.at(j) * m_weights.at(index) * m_control_points.at(index)->X();
        ay += basis.at(j) * m_weights.at(index) * m_control_points.at(index)->Y();
        az += basis.at(j) * m_weights.at(index) * m_control_points.at(index)->Z();
        aw += basis.at(j) * m_weights.at(index);
    }


    gfPoint3* result = gfCreatePoint3();
    if (!gfIsZero(aw))
        result->Set(ax/aw, ay/aw, az/aw);
    return result;
}

gfObjPolyLine* gfObjNURBSCurve::GetControlPolyline(void)
{
    gfInt num_points = (gfInt)m_control_points.size();
    gfObjPolyLine* poly = (gfObjPolyLine*) gfCreateGeomObject(GF_POLYLINE);
    poly->Init((gfInt)m_control_points.size(), false);

    for (gfInt i = 0;i < num_points;i++)
    {
        poly->GetPoint(i)->Set(m_control_points.at(i)->X(),m_control_points.at(i)->Y(),m_control_points.at(i)->Z());
    }

    return poly;
}

gfObjPolyLine* gfObjNURBSCurve::GetApproxPolyline(gfInt approx)
{
    gfInt num_points = (gfInt)m_control_points.size();
    gfObjPolyLine* poly = (gfObjPolyLine*) gfCreateGeomObject(GF_POLYLINE);
    poly->Init(approx, false);

    gfDouble a = GetKnot(m_degree);
    gfDouble b = GetKnot(GetNumControlPoints());
    if (gfIsZero(b - a))
        return NULL;
    gfInt n = GetNumControlPoints()  - 1;
    gfInt degree = m_degree;

    for (gfInt i = 0;i < approx;i++)
    {
        // current parameter
        gfDouble u = a + i * (b - a) / (approx - 1);

        gfInt span = gfGetBSplineKnotSpan(n, degree, &m_knot_vector, u);

        std::vector<gfDouble> basis;
        gfGetBSplineAllValues(degree, &m_knot_vector, span, u, &basis);

        gfDouble ax = 0;
        gfDouble ay = 0;
        gfDouble az = 0;
        gfDouble aw = 0;

        for (gfInt j = 0;j <= degree;j++)
        {
            gfInt index = span - degree + j;
            ax += basis.at(j) * m_weights.at(index) * m_control_points.at(index)->X();
            ay += basis.at(j) * m_weights.at(index) * m_control_points.at(index)->Y();
            az += basis.at(j) * m_weights.at(index) * m_control_points.at(index)->Z();
            aw += basis.at(j) * m_weights.at(index);
        }

        gfPoint3 * pt = poly->GetPoint(i);
        pt->Set(ax / aw, ay / aw, az / aw);
    }

    return poly;
}

GFRESULT gfObjNURBSCurve::GetDomain(gfDouble* u_min, gfDouble* u_max)
{
    *u_min = GetKnot(m_degree);
    *u_max = GetKnot(GetNumControlPoints());
    return GF_S_OK;
}

gfBool gfObjNURBSCurve::IsBezier(void)
{
    if (GetDegree() != (GetNumControlPoints() - 1))
        return false;

    gfInt num_knots = GetNumKnots();
    if (num_knots == 0)
        return true;

    for (gfInt i = 0;i < num_knots / 2;i++)
    {
        if (GetKnot(i) != GetKnot(0))
            return false;
    }
    for (gfInt i = num_knots / 2;i < num_knots ;i++)
    {
        if (GetKnot(i) != GetKnot(num_knots - 1))
            return false;
    }

    return true;
}

GFRESULT gfObjNURBSCurve::DecomposeDeCasteljau(gfDouble u, gfObjNURBSCurve* curve1, gfObjNURBSCurve* curve2)
{
    gfDouble u_min = GetKnot(m_degree);
    gfDouble u_max = GetKnot(GetNumControlPoints());

    if (u < u_min || u > u_max || !IsBezier() || gfIsZero(u_max - u_min))
        return GF_E_FAIL;

    // initialize curves
    curve1->InitBezier(m_degree + 1, u_min, u);
    curve2->InitBezier(m_degree + 1, u, u_max);

    gfDouble true_u = (u-u_min) / (u_max-u_min);

    gfAux2DArray<gfPoint4> aux;
    aux.resize(m_degree + 1, m_degree + 1);
    for (gfInt i = 0;i <= m_degree;i++)
    {
        gfPoint3* pt = GetControlPoint(i);
        aux.Set(0, i, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(i)));
    }

    // decompose using deCasteljau algorithm
    for (gfInt i = 1;i <= m_degree;i++)
    {
        for (gfInt j = 0;j < m_degree - i + 1;j++)
        {
            gfDouble ax = (1.0-true_u)*aux.at(i-1, j).W()*aux.at(i-1, j).X()+true_u*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).X();
            gfDouble ay = (1.0-true_u)*aux.at(i-1, j).W()*aux.at(i-1, j).Y()+true_u*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Y();
            gfDouble az = (1.0-true_u)*aux.at(i-1, j).W()*aux.at(i-1, j).Z()+true_u*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Z();
            gfDouble aw = (1.0-true_u)*aux.at(i-1, j).W()+true_u*aux.at(i-1, j+1).W();
            aux.Set(i, j, gfPoint4(ax/aw, ay/aw, az/aw, aw));
        }
    }

    // fill curves
    for (gfInt i = 0;i <= m_degree;i++)
    {
        gfPoint3* pt1 = curve1->GetControlPoint(i);
        pt1->Set(aux.at(i, 0).X(), aux.at(i, 0).Y(), aux.at(i, 0).Z());
        curve1->SetWeight(i, aux.at(i, 0).W());

        gfPoint3* pt2 = curve2->GetControlPoint(i);
        pt2->Set(aux.at(m_degree-i, i).X(), aux.at(m_degree-i, i).Y(), aux.at(m_degree-i, i).Z());
        curve2->SetWeight(i, aux.at(m_degree-i, i).W());
    }

    return GF_S_OK;
}

gfAuxBoundingVolume* gfObjNURBSCurve::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    // create auxiliary array of mesh points
    gfInt num_points = (gfInt)m_control_points.size();
    if (num_points < 1)
        return NULL;

    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfInt i = 0; i < num_points; i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        curr_point->Set(pt->X(), pt->Y(), pt->Z());
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    delete [] points;
    return result;
}

GFRESULT gfObjNURBSCurve::InsertKnot(gfInt mult, gfDouble u)
{
    gfInt np = GetNumControlPoints() - 1;
    gfInt k = gfGetBSplineKnotSpan(np, m_degree, &m_knot_vector, u);
    gfInt s = gfGetBSplineKnotMultiplicity(np, m_degree, &m_knot_vector, u);
    gfInt mp = np+m_degree+1;
    gfInt nq = np+mult;
    gfInt i,j;
    gfPoint4* Rw = new gfPoint4[m_degree + 1];
    gfPoint3* new_ctrl_pts = new gfPoint3[nq+1];
    gfDouble* new_weights = new gfDouble[nq+1];

    // set new knot vector
    std::vector<gfDouble> new_knot_vector;
    new_knot_vector.resize(mp+1);
    for (i = 0;i <= k;i++)
        new_knot_vector[i] = m_knot_vector.at(i);
    for (i = 1;i <= mult;i++)
        new_knot_vector[k+i] = u;
    for (i = k+1;i <= mp;i++)
        new_knot_vector[i+mult] = m_knot_vector.at(i);

    // fill new control points
    for (i = 0;i <= k-m_degree;i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        new_ctrl_pts[i].Set(pt->X(), pt->Y(), pt->Z());
        new_weights[i] = m_weights.at(i);
    }
    for (i = k-s;i <= np;i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        new_ctrl_pts[i+mult].Set(pt->X(), pt->Y(), pt->Z());
        new_weights[i+mult] = m_weights.at(i);
    }
    for (i = 0;i <= m_degree-s;i++)
    {
        gfPoint3* pt = m_control_points.at(k-m_degree+i);
        Rw[i].Set(pt->X(), pt->Y(), pt->Z(), m_weights.at(k-m_degree+i));
    }

    // insert the knot
    gfInt L;
    for (j = 1;j <= mult;j++)
    {
        L = k-m_degree+j;
        for (i = 0;i <= m_degree-j-s;i++)
        {
            gfDouble alpha = (u-m_knot_vector.at(L+i))/(m_knot_vector.at(i+k+1)-m_knot_vector.at(L+i));
            Rw[i].Set(alpha*Rw[i+1].X()+(1.0-alpha)*Rw[i].X(),
                      alpha*Rw[i+1].Y()+(1.0-alpha)*Rw[i].Y(),
                      alpha*Rw[i+1].Z()+(1.0-alpha)*Rw[i].Z(),
                      alpha*Rw[i+1].W()+(1.0-alpha)*Rw[i].W());
        }

        new_ctrl_pts[L].Set(Rw[0].X(), Rw[0].Y(), Rw[0].Z());
        new_weights[L] = Rw[0].W();
        new_ctrl_pts[k+mult-j-s].Set(Rw[m_degree-j-s].X(), Rw[m_degree-j-s].Y(), Rw[m_degree-j-s].Z());
        new_weights[k+mult-j-s] = Rw[m_degree-j-s].W();
    }

    for (i = L+1;i < k-s;i++)
    {
        new_ctrl_pts[i].Set(Rw[i-L].X(), Rw[i-L].Y(), Rw[i-L].Z());
        new_weights[i] = Rw[i-L].W();
    }

    // copy newly created arrays
    SetParams(m_degree, nq + 1);
    for (i = 0;i <= mp+mult;i++)
        SetKnot(i, new_knot_vector.at(i));
    for (i = 0;i <= nq;i++)
    {
        GetControlPoint(i)->Set(new_ctrl_pts[i].X(), new_ctrl_pts[i].Y(), new_ctrl_pts[i].Z());
        SetWeight(i, new_weights[i]);
    }

    delete [] Rw;
    delete [] new_ctrl_pts;
    delete [] new_weights;
    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::ConvertToPiecewiseBezier(void)
{
    // insert knots so each knot has multiplicity (m_degree + 1)
    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::DecomposeToBezierSegments(std::vector<gfObjNURBSCurve*>* segments)
{
    gfInt n = GetNumControlPoints()-1;
    gfInt m = n + 2;
    gfInt a = m_degree;
    gfInt b = m_degree + 1;
    gfInt nb = 0;
    gfInt i, j, k;
    gfAux2DArray<gfPoint4> Qw;
    Qw.resize(n, m_degree + 1);
    std::vector<gfDouble> alphas;
    alphas.resize(m_degree + 1);

    for (i = 0;i <= m_degree;i++)
    {
        gfPoint3* pt = GetControlPoint(i);
        Qw.Set(nb, i, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(i)));
    }

    while (b < m)
    {
        i = b;
        while ((b < m) && (m_knot_vector.at(b+1) == m_knot_vector.at(b)))
            b++;
        gfInt mult = b-i+1;
        if (mult < m_degree)
        {
            gfDouble numer = m_knot_vector.at(b) - m_knot_vector.at(a);
            for (j = m_degree;j > mult;j--)
                alphas[j-mult-1] = numer/(m_knot_vector.at(a+j) - m_knot_vector.at(a));
            gfInt r = m_degree - mult;
            for (j = 1;j <= r;j++)
            {
                gfInt save = r-j;
                gfInt s = mult+j;
                for  (k = m_degree;k >= s;k--)
                {
                    gfDouble alpha = alphas.at(k-s);
                    Qw.Set(nb, k, gfPoint4(alpha*Qw.at(nb,k).X()+(1.0-alpha)*Qw.at(nb, k-1).X(),
                                           alpha*Qw.at(nb,k).Y()+(1.0-alpha)*Qw.at(nb, k-1).Y(),
                                           alpha*Qw.at(nb,k).Z()+(1.0-alpha)*Qw.at(nb, k-1).Z(),
                                           alpha*Qw.at(nb,k).W()+(1.0-alpha)*Qw.at(nb, k-1).W()));
                }

                if (b < m)
                {
                    Qw.Set(nb + 1, save, gfPoint4(Qw.at(nb,m_degree).X(), Qw.at(nb,m_degree).Y(), Qw.at(nb,m_degree).Z(), Qw.at(nb,m_degree).W()));
                }
            }
        }

        nb++;
        if (b < m) 
        {
            for (i = m_degree-mult;i <= m_degree;i++)
            {
                gfPoint3* pt = GetControlPoint(b-m_degree+i);
                Qw.Set(nb, i, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(b-m_degree+i)));
            }
            a = b;
            b++;
        }
    }

    // create array of different knots
    std::vector<gfDouble> knots;
    for (i = 0;i < GetNumKnots();i++)
    {
        if (knots.size() == 0)
        {
            knots.push_back(GetKnot(i));
        }
        else
        {
            if (knots.at(knots.size()-1) < GetKnot(i))
                knots.push_back(GetKnot(i));
        }
    }

    for (i = 0;i < nb;i++)
    {
        // create NURBS curve as Bezier curve
        gfObjGeomObject* obj = gfCreateGeomObject(GF_NURBS_CURVE);
        gfObjNURBSCurve* curve = (gfObjNURBSCurve*) obj;
        curve->InitBezier(m_degree + 1, knots.at(i), knots.at(i+1));

        for (j = 0;j <= m_degree;j++)
        {
            gfPoint3* pt = curve->GetControlPoint(j);
            pt->Set(Qw.at(i, j).X(), Qw.at(i, j).Y(), Qw.at(i, j).Z());
            curve->SetWeight(j, Qw.at(i, j).W());
        }

        segments->push_back(curve);
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::DeleteAABBTree(gfBezier1AABBTreeNode* tree_root)
{
    if (tree_root->aabb)
        gfDeleteBoundVolume(tree_root->aabb);
    if (tree_root->left)
        DeleteAABBTree(tree_root->left);
    if (tree_root->right)
        DeleteAABBTree(tree_root->right);
    if (tree_root->curve)
        gfDeleteGeomObject(tree_root->curve);

    delete tree_root;
    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::DeleteWholeAABBTree(void)
{
    gfInt num_trees = (gfInt)m_bezier_aabb_trees.size();
    for (gfInt i = 0;i < num_trees;i++)
    {
        gfBezier1AABBTreeNode* root = m_bezier_aabb_trees.at(i);
        DeleteAABBTree(root);
    }
    m_bezier_aabb_trees.clear();

    return GF_S_OK;
}

GFRESULT gfObjNURBSCurve::ComputeWholeAABBTree(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    DeleteWholeAABBTree();

    m_num_nodes = 0;

    // decompose to Bezier patches
    std::vector<gfObjNURBSCurve*> beziers;
    DecomposeToBezierSegments(&beziers);

    gfInt num_trees = (gfInt)beziers.size();
    for (gfInt i = 0;i < num_trees;i++)
    {
        gfObjNURBSCurve* bezier = beziers.at(i);
        gfBezier1AABBTreeNode* root = CreateAABBTree(0, bezier);
        m_bezier_aabb_trees.push_back(root);
    }

    gfLog("Computed AABB tree of Bezier curves for NURBS curve, nodes = %d, time = %d ms.\n", m_num_nodes, count.EndMSCount());

    return GF_S_OK;
}

gfBezier1AABBTreeNode* gfObjNURBSCurve::CreateAABBTree(gfDouble depth, gfObjNURBSCurve* curve)
{
    gfBezier1AABBTreeNode* root = new gfBezier1AABBTreeNode;
    root->left = root->right = NULL;
    root->aabb = (gfAuxAABB*) curve->GetBoundVolume(GF_BOUND_VOL_AABB, NULL);
    root->curve = NULL;
    root->min_dir = 0;
    m_num_nodes++;

    // compute also OBB
    gfAuxOBB* obb = (gfAuxOBB*) curve->GetBoundVolume(GF_BOUND_VOL_OBB, NULL);

    // check if AABB needs to be splitted
    gfDouble x_dir = 2.0*gfSquareRoot(obb->m_axis1_x*obb->m_axis1_x+
        obb->m_axis1_y*obb->m_axis1_y+
        obb->m_axis1_z*obb->m_axis1_z);
    gfDouble y_dir = 2.0*gfSquareRoot(obb->m_axis2_x*obb->m_axis2_x+
        obb->m_axis2_y*obb->m_axis2_y+
        obb->m_axis2_z*obb->m_axis2_z);
    gfDouble z_dir = 2.0*gfSquareRoot(obb->m_axis3_x*obb->m_axis3_x+
        obb->m_axis3_y*obb->m_axis3_y+
        obb->m_axis3_z*obb->m_axis3_z);

    // find shortest dimension
    gfDouble min;
    gfDouble space;
    if (x_dir < y_dir)
    {
        if (x_dir < z_dir)
        {
            min = x_dir;
            root->min_dir = 1;
            space = y_dir*z_dir;
        }
        else
        {
            min = z_dir;
            root->min_dir = 3;
            space = y_dir*x_dir;
        }
    }
    else
    {
        if (y_dir < z_dir)
        {
            min = y_dir;
            root->min_dir = 2;
            space = x_dir*z_dir;
        }
        else
        {
            min = z_dir;
            root->min_dir = 3;
            space = y_dir*x_dir;
        }
    }

    // delete OBB
    gfDeleteBoundVolume(obb);

    // check if OBB is flat enough
    if ((min < 0.01 && space < 0.0001) || depth > GF_MAX_BEZIER_AABB_TREE_DEPTH/2)
    {
        root->curve = curve;
    }
    else
    {
        // split patch
        gfDouble u_min, u_max;
        curve->GetDomain(&u_min, &u_max);

        gfObjNURBSCurve* curve1 = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
        gfObjNURBSCurve* curve2 = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
        curve->DecomposeDeCasteljau((u_max+u_min)/2.0, curve1, curve2);

        gfDeleteGeomObject(curve);

        root->left = CreateAABBTree(depth+1, curve1);
        root->right = CreateAABBTree(depth+1, curve2);
    }

    return root;
}

gfRayInterInfo* gfObjNURBSCurve::RayAABBIntersection(gfBezier1AABBTreeNode* root, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    if (!root)
        return NULL;
    gfDouble tmin, tmax;
    if (GF_FAIL(root->aabb->RayIntersection(ray_begin, ray_dir, &tmin, &tmax)))
        return NULL;

    gfRayInterInfo* info = NULL;
    if (root->left && root->right)
    {
        // get intersection from childs
        gfRayInterInfo* info1 = RayAABBIntersection(root->left, ray_begin, ray_dir);
        gfRayInterInfo* info2 = RayAABBIntersection(root->right, ray_begin, ray_dir);
        if (info1 && info2)
        {
            if (info1->m_inters.at(0) < info2->m_inters.at(0))
            {
                info = info1;
                for (gfInt i = 0;i < (gfInt)info2->m_inters.size();i++)
                    info->m_inters.push_back(info2->m_inters.at(i));
                gfDeleteRayInterInfo(info2);
            }
            else
            {
                info = info2;
                for (gfInt i = 0;i < (gfInt)info1->m_inters.size();i++)
                    info->m_inters.push_back(info1->m_inters.at(i));
                gfDeleteRayInterInfo(info1);
            }
        }
        else if (info1)
        {
            info = info1;
        }
        else if (info2)
        {
            info = info2;
        }
    }
    else
    {
        if (!root->curve || !root->aabb)
            return NULL;

        // return intersection of ray and control polygon
        gfObjPolyLine* control = root->curve->GetControlPolyline();
        info = control->RayIntersection(NULL, ray_begin, ray_dir);

        if (info && info->m_inters.size() == 0)
        {
            gfDeleteRayInterInfo(info);
            info = NULL;
        }
    }

    // sort intersections
    if (info)
    {
        gfSort1DArray(&info->m_inters);
    }

    return info;
}

gfRayInterInfo* gfObjNURBSCurve::LineIntersection2D(gfDouble a, gfDouble b, gfDouble c)
{
    gfDouble min, max;
    GetDomain(&min, &max);

    if (IsBezier())
    {
        // use Bezier clipping
        std::vector<gfPoint2> points;
        points.resize(m_degree+1);
        for (gfInt i = 0;i <= m_degree;i++)
        {
            gfPoint2 pt(min+(max-min)*((gfDouble)i)/((gfDouble)m_degree), 
                GetWeight(i)*(a*GetControlPoint(i)->X()+b*GetControlPoint(i)->Y()+c));
            points[i] = pt;
        }

        // get intersection of convex hull and x axis
        gfDouble minu, maxu;
        if (GF_SUCC(gfAxisXConvexHullInter2D(&points, &minu, &maxu)))
        {
            // check for end of recursion
            if ((maxu-minu) < GF_BEZIER_CLIPPING_EPS)
            {
                gfRayInterInfo* info = gfCreateRayInterInfo();
                info->m_inters.push_back((maxu+minu)/2.0);
                return info;
            }

            // check for small diference
            if ((maxu-minu)/(max-min) > GF_BEZIER_CLIPPING_DIVIDE_TRESHOLD)
            {
                // divide curve into two
                gfObjNURBSCurve* curve1 = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
                gfObjNURBSCurve* curve2 = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
                DecomposeDeCasteljau((min+max)/2.0, curve1, curve2);
                gfRayInterInfo* info1 = curve1->LineIntersection2D(a, b, c);
                gfRayInterInfo* info2 = curve2->LineIntersection2D(a, b, c);
                gfDeleteGeomObject(curve1);
                gfDeleteGeomObject(curve2);

                // get together infos
                if (info1 && info2)
                {
                    for (gfInt i = 0;i < (gfInt)info2->m_inters.size();i++)
                        info1->m_inters.push_back(info2->m_inters.at(i));
                    gfDeleteRayInterInfo(info2);
                    gfSortRayInterInfo(info1);
                    return info1;
                }
                if (info1 && !info2)
                {
                    return info1;
                }
                if (info2 && !info1)
                {
                    return info2;
                }
                return NULL;      
            }

            // divide curve to three sub-curves <min, minu>,<minu, maxu>,<maxu, max>
            gfObjNURBSCurve* curve1 = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
            gfObjNURBSCurve* curve2 = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
            gfObjNURBSCurve* curve3 = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
            DecomposeDeCasteljau(minu, curve1, curve2);
            curve2->DecomposeDeCasteljau(maxu, curve1, curve3);

            gfDeleteGeomObject(curve2);
            gfDeleteGeomObject(curve3);

            // curve1 holds better approximation of curve for finding intersections
            gfRayInterInfo* info = curve1->LineIntersection2D(a, b, c);
            gfDeleteGeomObject(curve1);

            return info;
        }
        else
            return NULL;
    }
    else
    {
        //decompose to bezier curves
        std::vector<gfObjNURBSCurve*> segments;
        DecomposeToBezierSegments(&segments);

        // compute intersection for each segment
        gfRayInterInfo* info = gfCreateRayInterInfo();
        for (gfInt i = 0;i < (gfInt)segments.size();i++)
        {
            gfRayInterInfo* info_aux = segments.at(i)->LineIntersection2D(a, b, c);
            if (info_aux)
            {
                for (gfInt i = 0;i < (gfInt)info_aux->m_inters.size();i++)
                    info->m_inters.push_back(info_aux->m_inters.at(i));
                gfDeleteRayInterInfo(info_aux);
            }
            gfDeleteGeomObject(segments.at(i));
        }
        gfSort1DArray(&info->m_inters);
        return info;
    }

    return NULL;
}