//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjNURBSCurve.h                                    //
//   Description: NURBS curve class                             //
//                                                              //
//////////////////////////////////////////////////////////////////


#ifndef _GF_C_NURBS_CURVE_H_
#define _GF_C_NURBS_CURVE_H_


#include "gfDefines.h"
#include <vector>
#include "gfPoint.h"
#include "gfMatrix.h"
#include "gfObjGeomObject.h"
#include "gfObjPolyLine.h"


class gfObjNURBSCurve;

// node of AABB tree for faster ray intersection and collision
struct gfBezier1AABBTreeNode
{
    gfAuxAABB* aabb;
    gfBezier1AABBTreeNode* left;
    gfBezier1AABBTreeNode* right;
    gfObjNURBSCurve* curve;
    gfInt min_dir;
};


class gfObjNURBSCurve : public gfObjGeomObject
{
private:

    std::vector<gfDouble> m_knot_vector;
    std::vector<gfDouble> m_weights;
    std::vector<gfPoint3*> m_control_points;
    gfInt m_degree;

    // Bezier AABB tree functions
    std::vector<gfBezier1AABBTreeNode*> m_bezier_aabb_trees;
    gfRayInterInfo* RayAABBIntersection(gfBezier1AABBTreeNode* root, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfBezier1AABBTreeNode* CreateAABBTree(gfDouble depth, gfObjNURBSCurve* patch);
    GFRESULT ComputeWholeAABBTree(void);
    GFRESULT DeleteWholeAABBTree(void);
    GFRESULT DeleteAABBTree(gfBezier1AABBTreeNode* tree_root);
    gfInt m_num_nodes;

public:

    gfObjNURBSCurve(void);
    ~gfObjNURBSCurve(void);

    // gfObjGeomObject functions
    GFRESULT OGLDraw(gfLong flags);
    gfRayInterInfo* RayIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfGeomObjID GetID(void);
    GFRESULT Transform(gfMatrix4* matrix);
    gfObjGeomObject* Clone(void);
    gfAuxBoundingVolume* GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix);

    // initialize NURBS curve
    GFRESULT InitHalfCircle(gfPoint3* center, gfPoint3* start, gfVector3* normal);
    GFRESULT InitCircle(gfPoint3* center, gfPoint3* start, gfVector3* normal);
    GFRESULT InitCircularArc(gfPoint3* center, gfPoint3* start, gfVector3* normal, gfDouble angle);
    GFRESULT InitBezier(gfInt num_points);
    GFRESULT InitBezier(gfInt num_points, gfDouble u_min, gfDouble u_max);
    GFRESULT InitPiecewiseBezier(gfInt degree, gfInt num_segments);
    GFRESULT InitPolyline(gfInt num_points);
    GFRESULT InitSpiral(gfPoint3* center, gfPoint3* start, gfVector3* normal, gfInt num_twists, gfDouble twist_elevation);
    GFRESULT InitRandom(gfInt degree, gfInt num_points, gfDouble xmin, gfDouble xmax);

    // parameters of curve
    gfInt SetParams(gfInt degree, gfInt num_control_points);
    gfInt GetDegree(void);
    gfInt GetNumKnots(void);
    gfDouble GetKnot(int index);
    GFRESULT SetKnot(int index, gfDouble knot);
    gfInt GetNumControlPoints(void);
    gfPoint3* GetControlPoint(int index);
    gfDouble GetWeight(int index);
    GFRESULT SetWeight(int index, gfDouble weight);
    gfObjPolyLine* GetControlPolyline(void);
    gfObjPolyLine* GetApproxPolyline(gfInt approx);
    GFRESULT GetDomain(gfDouble* u_min, gfDouble* u_max);
    gfPoint3* GetCurvePoint(gfDouble u);

    // modeling functions
    gfBool IsBezier(void);
    GFRESULT DecomposeDeCasteljau(gfDouble u, gfObjNURBSCurve* curve1, gfObjNURBSCurve* curve2);
    GFRESULT ElevateDegree(void);
    GFRESULT InsertKnot(gfInt mult, gfDouble u);
    GFRESULT ConvertToPiecewiseBezier(void);
    GFRESULT DecomposeToBezierSegments(std::vector<gfObjNURBSCurve*>* segments);
    gfRayInterInfo* LineIntersection2D(gfDouble a, gfDouble b, gfDouble c);

};


#endif