//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjNURBSPatch.cpp                                  //
//   Description: NURBS patch class                             //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfAuxBSplineBasis.h"
#include "gfObjNURBSPatch.h"
#include "gfObjectFactory.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfAuxItersectionUtils.h"
#include "gfAuxLogging.h"
#include "gfAux3DArray.h"
#include "gfWinFPSCounter.h"

#include "./OGL/glut.h"

gfObjNURBSPatch::gfObjNURBSPatch(void)
{
    m_time_stamp = 0;
    m_weights.resize(0);
    m_control_points.resize(0);
    m_u_knot_vector.resize(0);
    m_v_knot_vector.resize(0);
    m_num_u_control_points = 0;
    m_num_v_control_points = 0;
    m_u_degree = 0;
    m_v_degree = 0;

    m_cache_vertices = NULL;
    m_cache_normals = NULL;
    m_cache_wire_vertices = NULL;

    m_bezier_clipping = true;
}

gfObjNURBSPatch::~gfObjNURBSPatch(void)
{
    Clear();
}

GFRESULT gfObjNURBSPatch::OGLDraw(gfLong flags)
{
    gfInt approx = 2 * GF_VIS_PREC;
    if (m_time_stamp == 0)
        return GF_S_OK;

    // draw possibly control points, it isn't cached
    if (flags & GF_FLG_CONTROL_NET)
    {
        glColor3f(0.0, 0.0, 0.0);
        for (gfInt i = 0;i < m_num_u_control_points;i++)
        {
            glBegin(GL_LINES);
            for (gfInt j = 0;j < m_num_v_control_points;j++)
            {
                gfPoint3* pt1 = GetControlPoint(i, j);

                if (i < m_num_u_control_points - 1)
                {
                    gfPoint3* pt2 = GetControlPoint(i + 1, j);
                    glVertex3f(pt1->X(), pt1->Y(), pt1->Z());
                    glVertex3f(pt2->X(), pt2->Y(), pt2->Z());
                }

                if (j < m_num_v_control_points - 1)
                {
                    gfPoint3* pt3 = GetControlPoint(i, j + 1);
                    glVertex3f(pt1->X(), pt1->Y(), pt1->Z());
                    glVertex3f(pt3->X(), pt3->Y(), pt3->Z());
                }
            }
            glEnd();
        }

        glPointSize(8.0);
        glBegin(GL_POINTS);
        for (gfInt i = 0;i < m_num_u_control_points;i++)
        {
            for (gfInt j = 0;j < m_num_v_control_points;j++)
            {
                gfPoint3* pt = GetControlPoint(i, j);
                glVertex3f(pt->X(), pt->Y(), pt->Z());
            }
        }
        glEnd();
        glPointSize(1.0);
    }

    // check if we are caching
    if (flags & GF_FLG_VERTEX_ARRAY_CACHE)
    {
        if (!m_cache_vertices)
            CreateVACache(approx);
        return DrawVACache(flags);
    }
    DeleteVACache();

    // Interval of definition
    gfDouble a = GetUKnot(m_u_degree);
    gfDouble b = GetUKnot(GetNumUControlPoints());
    if (gfIsZero(b - a))
        return GF_S_OK;

    // gfInterval of definition
    gfDouble c = GetVKnot(m_v_degree);
    gfDouble d = GetVKnot(GetNumVControlPoints());
    if (gfIsZero(d - c))
        return GF_S_OK;

    gfInt n = GetNumUControlPoints() - 1;
    gfInt m = GetNumVControlPoints() - 1;

    gfBool isTrimmed = (GetNumTrimCurves() > 0);

    // array for approximation points
    gfPoint3* aux_points = new gfPoint3[approx * approx];
    gfVector3* aux_normals = NULL;
    if (flags & GF_FLG_NORMALS)
        aux_normals = new gfVector3[approx * approx];

    // get UV space tesselation based on trim curves
    gfUVPoint* trimmed_points = NULL;
    if (isTrimmed)
        trimmed_points = TesselateTrimmedUVSpace(approx);

    for (gfInt i = 0;i < approx;i++)
    {
        // current u parameter
        gfDouble u = a + i * (b - a) / (approx - 1);
        gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
        //std::vector<gfDouble> ubasis;
        //gfGetBSplineAllValues(n, m_u_degree, &m_u_knot_vector, uspan, u, &ubasis);
        gfAux2DArray<gfDouble> ubasis(2, 100);
        gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 1, &ubasis);

        for (gfInt j = 0;j < approx;j++)
        {
            // current v parameter     
            gfDouble v = c + j * (d - c) / (approx - 1);
            gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
            //std::vector<gfDouble> vbasis;
            //gfGetBSplineAllValues(m, m_v_degree, &m_v_knot_vector, vspan, v, &vbasis);
            gfAux2DArray<gfDouble> vbasis(2, 100);
            gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 1, &vbasis);

            gfDouble ax = 0.0;
            gfDouble ay = 0.0;
            gfDouble az = 0.0;
            gfDouble aw = 0.0;

            gfDouble dux = 0.0;
            gfDouble duy = 0.0;
            gfDouble duz = 0.0;
            gfDouble duw = 0.0;
            gfDouble dvx = 0.0;
            gfDouble dvy = 0.0;
            gfDouble dvz = 0.0;
            gfDouble dvw = 0.0;

            for (gfInt k = 0;k <= m_u_degree;k++)
            {
                gfInt uindex = uspan - m_u_degree + k;

                for (gfInt l = 0;l <= m_v_degree;l++)
                {
                    gfInt vindex = vspan - m_v_degree + l;
                    //gfDouble aux = ubasis.at(k) * vbasis.at(l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                    gfDouble aux = ubasis.at(0, k) * vbasis.at(0, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                    ax += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
                    ay += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
                    az += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
                    aw += aux;

                    if (flags & GF_FLG_NORMALS)
                    {
                        gfDouble aux1 = ubasis.at(1, k) * vbasis.at(0, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                        dux += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
                        duy += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
                        duz += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
                        duw += aux1;
                        gfDouble aux2 = ubasis.at(0, k) * vbasis.at(1, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                        dvx += aux2 * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
                        dvy += aux2 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
                        dvz += aux2 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
                        dvw += aux2;
                    }
                }
            }

            if (isTrimmed && trimmed_points[i + j * approx].boundary)
            {
                gfPoint3* patch_pt = GetPatchPoint(u + trimmed_points[i + j * approx].du, v + trimmed_points[i + j * approx].dv);
                if (patch_pt)
                {
                    aux_points[i + j * approx].Set(patch_pt->X(), patch_pt->Y(), patch_pt->Z());
                    gfDeletePoint3(patch_pt);
                    if (flags & GF_FLG_NORMALS)
                    {
                        gfVector3* normal = GetPatchNormal(u + trimmed_points[i + j * approx].du, v + trimmed_points[i + j * approx].dv);
                        aux_normals[i + j * approx].Set(normal->X(), normal->Y(), normal->Z());
                        gfDeleteVector3(normal);
                    }
                }
                else
                {
                    patch_pt = GetPatchPoint(u, v);
                    aux_points[i + j * approx].Set(patch_pt->X(), patch_pt->Y(), patch_pt->Z());
                    gfDeletePoint3(patch_pt);
                    if (flags & GF_FLG_NORMALS)
                    {
                        gfVector3* normal = GetPatchNormal(u, v);
                        aux_normals[i + j * approx].Set(normal->X(), normal->Y(), normal->Z());
                        gfDeleteVector3(normal);
                    }
                }
            }
            else
            {
                aux_points[i + j * approx].Set(ax / aw, ay / aw, az / aw);
                if (flags & GF_FLG_NORMALS)
                {
                    gfDouble fdux = (dux * aw - ax * duw) / aw / aw;
                    gfDouble fduy = (duy * aw - ay * duw) / aw / aw;
                    gfDouble fduz = (duz * aw - az * duw) / aw / aw;

                    gfDouble fdvx = (dvx * aw - ax * dvw) / aw / aw;
                    gfDouble fdvy = (dvy * aw - ay * dvw) / aw / aw;
                    gfDouble fdvz = (dvz * aw - az * dvw) / aw / aw;

                    aux_normals[i + j * approx].Set((fduy*fdvz-fduz*fdvy),
                        (-fdux*fdvz+fduz*fdvx),
                        (fdux*fdvy-fduy*fdvx));
                    aux_normals[i + j * approx].Normalize();
                }
            }

            //point->curvature = trimmed_points[i + j * approx].winding_number;
        }
    }

    // fill triangle loops
    if (flags & GF_FLG_SOLID)
    {
        glBegin(GL_QUADS);
        for (gfInt i = 0;i < approx - 1;i++)
        {
            /*glBegin(GL_TRIANGLE_STRIP);
            for (gfInt j = 0;j < GF_VIS_PREC;j++)
            {
            glColor3f( 0.25 + i / (2.0 * approx), 0.25 + j / (2.0 * approx), 0.25);
            glVertex3f(aux_points[i + j * approx].X(), aux_points[i + j * approx].Y(), aux_points[i + j * approx].Z());
            glVertex3f(aux_points[i + 1 + j * approx].X(), aux_points[i + 1 + j * approx].Y(), aux_points[i + 1 + j * approx].Z());
            }
            glEnd();*/

            for (gfInt j = 0;j < approx - 1;j++)
            {
                gfInt wind1 = 0;
                gfInt wind2 = 0;
                gfInt wind3 = 0;
                gfInt wind4 = 0;
                if (isTrimmed)
                {
                    wind1 = (gfInt) trimmed_points[i + 1 + approx * j].winding_number;
                    wind2 = (gfInt) trimmed_points[i + 1 + approx * (j + 1)].winding_number;
                    wind3 = (gfInt) trimmed_points[i + approx * (j + 1)].winding_number;
                    wind4 = (gfInt) trimmed_points[i + approx * j].winding_number;
                }

                if (wind1 % 2 == 1 || wind2 % 2 == 1 || wind3 % 2 == 1 || wind4 % 2 == 1)
                {
                    // face is trimmed
                }
                else
                {
                    // draw face
                    //glColor3f( 0.25 + i / (2.0 * approx), 0.25 + j / (2.0 * approx), 0.25);

                    if (flags & GF_FLG_NORMALS)
                        glNormal3f(aux_normals[i + j * approx].X(), aux_normals[i + j * approx].Y(), aux_normals[i + j * approx].Z());
                    glVertex3d(aux_points[i + j * approx].X(), aux_points[i + j * approx].Y(), aux_points[i + j * approx].Z());

                    if (flags & GF_FLG_NORMALS)
                        glNormal3f(aux_normals[i + (j + 1) * approx].X(), aux_normals[i + (j + 1) * approx].Y(), aux_normals[i + (j + 1) * approx].Z());
                    glVertex3d(aux_points[i + (j + 1) * approx].X(), aux_points[i + (j + 1) * approx].Y(), aux_points[i + (j + 1) * approx].Z());

                    if (flags & GF_FLG_NORMALS)
                        glNormal3f(aux_normals[i + 1 + (j + 1) * approx].X(), aux_normals[i + 1 + (j + 1) * approx].Y(), aux_normals[i + 1 + (j + 1) * approx].Z());
                    glVertex3d(aux_points[i + 1 + (j + 1) * approx].X(), aux_points[i + 1 + (j + 1) * approx].Y(), aux_points[i + 1 + (j + 1) * approx].Z());

                    if (flags & GF_FLG_NORMALS)
                        glNormal3f(aux_normals[i + 1 + j * approx].X(), aux_normals[i + 1 + j * approx].Y(), aux_normals[i + 1 + j * approx].Z());
                    glVertex3d(aux_points[i + 1 + j * approx].X(), aux_points[i + 1 + j * approx].Y(), aux_points[i + 1 + j * approx].Z());
                }
            }
        }
        glEnd();
    }

    glNormal3f(0.0, 0.0, 0.0);
    if (flags & GF_FLG_WIRE)
    {
        for (gfInt i = 0;i < approx - 1;i++)
        {
            glBegin(GL_LINES);
            for (gfInt j = 0;j < approx;j++)
            {
                gfInt wind1 = 0;
                gfInt wind2 = 0;
                gfInt wind3 = 0;
                gfInt wind4 = 0;
                if (isTrimmed)
                {
                    wind1 = (gfInt) trimmed_points[i + 1 + approx * j].winding_number;
                    wind2 = (gfInt) trimmed_points[i + 1 + approx * (j + 1)].winding_number;
                    wind3 = (gfInt) trimmed_points[i + approx * (j + 1)].winding_number;
                    wind4 = (gfInt) trimmed_points[i + approx * j].winding_number;
                }

                if (!(wind1 % 2 == 1 || wind2 % 2 == 1))
                {
                    glVertex3f(aux_points[i + j * approx].X(), aux_points[i + j * approx].Y(), aux_points[i + j * approx].Z());
                    glVertex3f(aux_points[i + 1 + j * approx].X(), aux_points[i + 1 + j * approx].Y(), aux_points[i + 1 + j * approx].Z());
                }
                if (j < approx - 1 && !(wind3 % 2 == 1 || wind4 % 2 == 1))
                {
                    glVertex3f(aux_points[i + j * approx].X(), aux_points[i + j * approx].Y(), aux_points[i + j * approx].Z());
                    glVertex3f(aux_points[i + (j + 1) * approx].X(), aux_points[i + (j + 1) * approx].Y(), aux_points[i + (j + 1) * approx].Z());
                }
            }
            glEnd();
        }
    }

    delete [] aux_points;
    if (aux_normals)
        delete [] aux_normals;
    delete [] trimmed_points;

    return GF_S_OK;
}

gfRayInterInfo* gfObjNURBSPatch::RayIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    m_bezier_clipping = false;
    if (m_bezier_clipping)
    {
        gfDisableLogging();

        // for now, use bezier clipping
        gfVector3 up;
        if (gfAbs(ray_dir->X()) > gfAbs(ray_dir->Y()) && gfAbs(ray_dir->X()) > gfAbs(ray_dir->Z()))
            up.Set(ray_dir->Y(), -ray_dir->X(), 0.0);
        else
            up.Set(0.0, ray_dir->Z(), -ray_dir->Y());
        //up.Set(-ray_dir->Y()*ray_dir->Z(), ray_dir->X()*ray_dir->Z()/2.0, ray_dir->X()*ray_dir->Y()/2.0);
        gfVector3 right;
        gfVectorProduct3(ray_dir, &up, &right);
        right.Normalize();
        up.Normalize();
        gfDouble d1 = -(up.X()*ray_begin->X()+up.Y()*ray_begin->Y()+up.Z()*ray_begin->Z());
        gfDouble d2 = -(right.X()*ray_begin->X()+right.Y()*ray_begin->Y()+right.Z()*ray_begin->Z());

        gfRayBezierPatchInfo* res = BezierClipping(NULL, 0, up.X(), up.Y(), up.Z(), d1, right.X(), right.Y(), right.Z(), d2);

        gfEnableLogging();

        if (res)
        {
            // find first intersection on ray, also inside trimmed domain
            std::vector<gfDouble> tarray;
            gfInt size = (gfInt)res->m_inters.size();
            tarray.resize(size);

            for (gfInt i = 0;i < size;i++)
            {
                // compute point on surface
                gfPoint3* pt = GetPatchPoint(res->m_inters.at(i).X(), res->m_inters.at(i).Y());
                // project point on ray
                gfVector3 aux;
                gfSubstractPoints3(pt, ray_begin, &aux);
                gfDouble t = gfScalarProduct3(&aux, ray_dir);
                tarray[i] = t;

                gfDeletePoint3(pt);
            }

            // sort tarray
            gfBool done  = false;
            gfInt curr = (gfInt)tarray.size() - 1;
            while (!done)
            {
                done = true;
                for (gfInt i = 0;i < curr;i++)
                {
                    if (tarray.at(i) > tarray.at(i+1))
                    {
                        done = false;

                        // tarray.Swap(i, i+1);
                        gfDouble tmp = tarray[i];
                        tarray[i] = tarray[i+1];
                        tarray[i+1] = tmp;

                        //res->m_inters.Swap(i, i+1);
                        gfPoint2 tmp2 = res->m_inters[i];
                        res->m_inters[i] = res->m_inters[i+1];
                        res->m_inters[i+1] = tmp2;
                    }
                }
                curr--;
            }

            // find first inside domain
            for (gfInt i = 0;i < size;i++)
            {
                // check domain
                if (PointInsideDomain(res->m_inters.at(i).X(), res->m_inters.at(i).Y()) == GF_S_OK)
                {
                    // compute normal
                    gfVector3* normal = GetPatchNormal(res->m_inters.at(i).X(), res->m_inters.at(i).Y());
                    gfRayInterInfo* info = gfCreateRayInterInfo();

                    info->m_normals.push_back(gfVector3(normal->X(), normal->Y(), normal->Z()));
                    info->m_inters.push_back(tarray.at(i));

                    gfDeleteVector3(normal);

                    return info;
                }
            }

            return NULL;
        }
        else
        {
            return NULL;
        }
    }

    // no bezier clipping
    // using AABB tree of nested patches
    gfInt num_trees = (gfInt)m_bezier_aabb_trees.size();
    if (num_trees == 0)
        ComputeWholeAABBTree();
    num_trees = (gfInt)m_bezier_aabb_trees.size();

    gfDisableLogging();

    gfRayInterInfo* info = NULL;
    for (gfInt i = 0;i < num_trees;i++)
    {
        gfBezier2AABBTreeNode* root = m_bezier_aabb_trees.at(i);
        gfRayInterInfo* info_temp = RayAABBIntersection(root, ray_begin, ray_dir);

        if (!info)
        {
            info = gfCreateRayInterInfo();
        }

        if (info_temp)
        {
            
            // copy normal if necessary
            for (gfInt i = 0;i < (gfInt)info_temp->m_inters.size();i++)
            {
                if (m_bezier_clipping)
                {
                    if (!GF_SUCC(PointInsideDomain(info_temp->m_uvw.at(i).X(), info_temp->m_uvw.at(i).Y())))
                        continue;
                }

                info->m_inters.push_back(info_temp->m_inters.at(i));
                info->m_normals.push_back(info_temp->m_normals.at(i));
            }
            gfDeleteRayInterInfo(info_temp);
            gfSortRayInterInfo(info);
        }
    }

    gfEnableLogging();
    if (info && (info->m_inters.size() == 0))
    {
        gfDeleteRayInterInfo(info);
        info = NULL;
    }

    return info;
}

gfGeomObjID gfObjNURBSPatch::GetID(void)
{
    return GF_NURBS_PATCH;
}

gfObjGeomObject* gfObjNURBSPatch::Clone(void)
{
    // create new curve
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());
    gfObjNURBSPatch* patch = (gfObjNURBSPatch*) clone;

    // set parameters of patch
    gfInt num_u_control_points = GetNumUControlPoints();
    gfInt num_v_control_points = GetNumVControlPoints();
    patch->SetParams(m_u_degree, m_v_degree, num_u_control_points, num_v_control_points);

    // copy control points and weights
    for (gfInt i = 0;i < num_u_control_points;i++)
        for (gfInt j = 0;j < num_v_control_points;j++)
        {
            gfPoint3 * new_point = patch->GetControlPoint(i, j);
            gfPoint3 * old_point = GetControlPoint(i, j);
            new_point->Set(old_point->X(), old_point->Y(), old_point->Z());
            patch->SetWeight(i, j, GetWeight(i, j));
        }

        // copy knot vectors
        gfInt num_u_knots = GetNumUKnots();
        for (gfInt i = 0;i < num_u_knots;i++)
        {
            patch->SetUKnot(i, GetUKnot(i));
        }
        gfInt num_v_knots = GetNumVKnots();
        for (gfInt i = 0;i < num_v_knots;i++)
        {
            patch->SetVKnot(i, GetVKnot(i));
        }

        // copy trim curves
        gfInt num_trim_curves = GetNumTrimCurves();
        for (gfInt i = 0;i < num_trim_curves;i++)
        {
            gfObjGeomObject* trim_clone = GetTrimCurve(i)->Clone();

            patch->AddTrimCurve((gfObjNURBSCurve*) trim_clone);
        }

        return clone;
}

GFRESULT gfObjNURBSPatch::Transform(gfMatrix4* matrix)
{
    // delete cache
    DeleteVACache();

    // delete AABB tree
    DeleteWholeAABBTree();

    // transform control points
    for (gfInt i = 0;i < m_num_u_control_points;i++)
    {
        for (gfInt j = 0;j < m_num_v_control_points;j++)
        {
            gfPoint3* pt = GetControlPoint(i, j);
            gfTransformPoint3(pt, matrix);
        }
    }
    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::Clear(void)
{
    SetParams(0 ,0, 0, 0);
    RemoveAllTrimCurves();
    DeleteWholeAABBTree();
    DeleteVACache();

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitPlane(gfDouble xmin, gfDouble xmax, gfDouble ymin, gfDouble ymax)
{
    InitBezier(2, 2, xmin, xmax, ymin, ymax);

    GetControlPoint(0, 0)->Set(xmin, ymin, 0.0);
    GetControlPoint(1, 0)->Set(xmax, ymin, 0.0);
    GetControlPoint(0, 1)->Set(xmin, ymax, 0.0);
    GetControlPoint(1, 1)->Set(xmax, ymax, 0.0);

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitSphere(gfPoint3 * center, gfPoint3 * start)
{
    Clear();

    gfVector3* axis = gfSubstractPoints3(start, center);
    axis->Normalize();
    gfVector3* normal = gfCreateVector3();
    normal->Set(axis->Z(), 0, -axis->X());
    normal->Normalize();

    gfObjNURBSCurve* aux_curve = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
    aux_curve->InitHalfCircle(center, start, normal);
    InitRotational(center, axis, aux_curve);

    gfDeleteGeomObject((gfObjGeomObject*) aux_curve);
    gfDeleteVector3(axis);
    gfDeleteVector3(normal);

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitTorus(gfPoint3* center, gfPoint3* start, gfPoint3* start2)
{
    Clear();

    gfVector3* axis = gfSubstractPoints3(start, center);
    axis->Normalize();
    gfVector3* normal = gfCreateVector3();
    normal->Set(axis->Z(), 0, -axis->X());
    normal->Normalize();

    gfVectorProduct3(axis, normal, axis);

    gfObjNURBSCurve* aux_curve = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
    aux_curve->InitCircle(start, start2, normal);

    InitRotational(center, axis, aux_curve);

    gfDeleteGeomObject((gfObjGeomObject*) aux_curve);
    gfDeleteVector3(axis);
    gfDeleteVector3(normal);

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitCone(gfPoint3* center, gfDouble radius1, gfDouble radius2, gfDouble height)
{
    Clear();

    gfObjNURBSCurve* aux_curve = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
    aux_curve->InitPolyline(4);

    aux_curve->GetControlPoint(0)->Set(center->X(), center->Y(), center->Z() + height);
    aux_curve->GetControlPoint(1)->Set(center->X(), center->Y() + radius2, center->Z() + height);
    aux_curve->GetControlPoint(2)->Set(center->X(), center->Y() + radius1, center->Z());
    aux_curve->GetControlPoint(3)->Set(center->X(), center->Y(), center->Z());

    gfVector3 axis(0.0, 0.0, 1.0);
    InitRotational(center, &axis, aux_curve);

    gfDeleteGeomObject((gfObjGeomObject*) aux_curve);

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitRotational(gfPoint3 * center, gfVector3* axis, gfObjNURBSCurve* nurbs_curve)
{
    Clear();

    // retrieve data from given curve
    gfInt u_degree = nurbs_curve->GetDegree();
    gfInt num_u_points = nurbs_curve->GetNumControlPoints();
    gfInt num_u_knots = nurbs_curve->GetNumKnots();

    // set parameters of patch
    SetParams(u_degree, 2, num_u_points, 9);

    // set u and v knots of patch
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, nurbs_curve->GetKnot(i));

    SetVKnot(0, 0.0);SetVKnot(1, 0.0);SetVKnot(2, 0.0);
    SetVKnot(3, 0.25);SetVKnot(4, 0.25);
    SetVKnot(5, 0.5);SetVKnot(6, 0.5);
    SetVKnot(7, 0.75);SetVKnot(8, 0.75);
    SetVKnot(9, 1.0);SetVKnot(10, 1.0);SetVKnot(11, 1.0);

    gfPoint3* curr_center = gfCreatePoint3();
    gfVector3* diff = gfCreateVector3();

    for (gfInt i = 0;i < num_u_points;i++)
    {
        // here find projection of current point on axis given by point and vector
        gfPoint3* curr_point = nurbs_curve->GetControlPoint(i);

        gfSubstractPoints3(curr_point, center, diff);
        gfDouble scl_prod = gfScalarProduct3(axis, diff);
        gfCombinePointVector3(center, scl_prod, axis, curr_center);

        // init auxiliary curve as circle and get its data
        gfObjNURBSCurve* aux_curve = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
        aux_curve->InitCircle(curr_center, curr_point, axis);

        // fill patch with data from auxiliary curve
        for (gfInt j = 0; j < 9;j++)
        {
            gfPoint3* pt = GetControlPoint(i, j);
            gfPoint3* pt2 = aux_curve->GetControlPoint(j);
            pt->Set(pt2->X(), pt2->Y(), pt2->Z());
            SetWeight(i, j, aux_curve->GetWeight(j) * nurbs_curve->GetWeight(i));
        }

        gfDeleteGeomObject((gfObjGeomObject*) aux_curve);
    }

    gfDeleteVector3(diff);
    gfDeletePoint3(curr_center);

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitRotational(gfPoint3 * center, gfVector3* axis, gfDouble angle, gfObjNURBSCurve* nurbs_curve)
{
    Clear();

    // retrieve data from given curve
    gfInt u_degree = nurbs_curve->GetDegree();
    gfInt num_u_points = nurbs_curve->GetNumControlPoints();
    gfInt num_u_knots = nurbs_curve->GetNumKnots();

    gfObjNURBSCurve* aux_curve = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
    gfPoint3 start(1.0, 232.0, 41.3);
    aux_curve->InitCircularArc(center, &start, axis, angle);
    gfInt v_degree = aux_curve->GetDegree();
    gfInt num_v_points = aux_curve->GetNumControlPoints();
    gfInt num_v_knots = aux_curve->GetNumKnots();

    // set parameters of patch
    SetParams(u_degree, v_degree, num_u_points, num_v_points);

    // set u and v knots of patch
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, nurbs_curve->GetKnot(i));
    for (gfInt i = 0;i < num_v_knots;i++)
        SetVKnot(i, aux_curve->GetKnot(i));

    gfPoint3* curr_center = gfCreatePoint3();
    gfVector3* diff = gfCreateVector3();

    for (gfInt i = 0;i < num_u_points;i++)
    {
        // here find projection of current point on axis given by point and vector
        gfPoint3* curr_point = nurbs_curve->GetControlPoint(i);

        gfSubstractPoints3(curr_point, center, diff);
        gfDouble scl_prod = gfScalarProduct3(axis, diff);
        gfCombinePointVector3(center, scl_prod, axis, curr_center);

        // init auxiliary curve as circle and get its data
        aux_curve->InitCircularArc(curr_center, curr_point, axis, angle);

        // fill patch with data from auxiliary curve
        for (gfInt j = 0; j < num_v_points;j++)
        {
            gfPoint3* pt = GetControlPoint(i, j);
            gfPoint3* pt2 = aux_curve->GetControlPoint(j);
            pt->Set(pt2->X(), pt2->Y(), pt2->Z());
            SetWeight(i, j, aux_curve->GetWeight(j) * nurbs_curve->GetWeight(i));
        }
    }

    gfDeleteGeomObject((gfObjGeomObject*) aux_curve);
    gfDeleteVector3(diff);
    gfDeletePoint3(curr_center);

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitBezier(gfInt num_u_points, gfInt num_v_points)
{
    Clear();

    gfInt u_degree = num_u_points - 1;
    gfInt v_degree = num_v_points - 1;
    SetParams(u_degree, v_degree, num_u_points, num_v_points);

    gfInt num_u_knots = GetNumUKnots();
    gfInt num_v_knots = GetNumVKnots();
    for (gfInt i = 0;i < num_u_knots;i++)
    {
        SetUKnot(i, (i < num_u_knots / 2) ? 0.0 : 1.0);
    }
    for (gfInt i = 0;i < num_v_knots;i++)
    {
        SetVKnot(i, (i < num_v_knots / 2) ? 0.0 : 1.0);
    }

    for (gfInt i = 0;i < num_u_points;i++)
        for (gfInt j = 0;j < num_v_points;j++)
        {
            SetWeight(i, j, 1.0);
        }

        return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitBezier(gfInt num_u_points, gfInt num_v_points, gfDouble u_min, gfDouble u_max, gfDouble v_min, gfDouble v_max)
{
    Clear();

    gfInt u_degree = num_u_points - 1;
    gfInt v_degree = num_v_points - 1;
    SetParams(u_degree, v_degree, num_u_points, num_v_points);

    gfInt num_u_knots = GetNumUKnots();
    gfInt num_v_knots = GetNumVKnots();
    for (gfInt i = 0;i < num_u_knots;i++)
    {
        SetUKnot(i, (i < num_u_knots / 2) ? u_min : u_max);
    }
    for (gfInt i = 0;i < num_v_knots;i++)
    {
        SetVKnot(i, (i < num_v_knots / 2) ? v_min : v_max);
    }

    for (gfInt i = 0;i < num_u_points;i++)
        for (gfInt j = 0;j < num_v_points;j++)
        {
            SetWeight(i, j, 1.0);
        }

        return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitRandom(gfPoint3* center, gfDouble tx, gfDouble ty)
{
    Clear();
    InitBezier(3, 3);

    for (gfInt i = 0;i < 3;i++)
        for (gfInt j = 0;j < 3;j++)
        {
            gfDouble ax = center->X() - tx + 2 * tx * i / 2.0;
            gfDouble ay = center->Y() - ty + 2 * ty * j / 2.0;
            gfDouble az = center->Z() + gfRand(2.0) - 1.0;
            GetControlPoint(i, j)->Set(center->X() + ax, center->Y() + ay, center->Z() + az);
            SetWeight(i, j, 1.0);
        }

        return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitTransitional(gfObjNURBSCurve* curve1, gfObjNURBSCurve* curve2)
{
    Clear();

    // retrieve data from given curves
    gfInt u_degree = curve1->GetDegree();
    gfInt num_u_points = curve1->GetNumControlPoints();
    gfInt num_u_knots = curve1->GetNumKnots();

    gfInt v_degree = curve2->GetDegree();
    gfInt num_v_points = curve2->GetNumControlPoints();
    gfInt num_v_knots = curve2->GetNumKnots();

    // set parameters of patch
    SetParams(u_degree, v_degree, num_u_points, num_v_points);

    // set u and v knots of patch
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, curve1->GetKnot(i));
    for (gfInt i = 0;i < num_v_knots;i++)
        SetVKnot(i, curve2->GetKnot(i));

    gfPoint3* pivot = curve2->GetControlPoint(0);
    if (!pivot)
        return GF_E_FAIL;

    // set control points and weights
    for (gfInt i = 0;i < num_u_points;i++)
        for (gfInt j = 0;j < num_v_points;j++)
        {
            gfPoint3* point1 = curve1->GetControlPoint(i);
            gfPoint3* point2 = curve2->GetControlPoint(j);

            gfDouble px = point1->X()+point2->X()-pivot->X();
            gfDouble py = point1->Y()+point2->Y()-pivot->Y();
            gfDouble pz = point1->Z()+point2->Z()-pivot->Z();

            gfPoint3* pt = GetControlPoint(i, j);
            pt->Set(px, py, pz);
            SetWeight(i, j, curve1->GetWeight(i) * curve2->GetWeight(j));
        }

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitNormalTransitional(gfObjNURBSCurve* curve1, gfObjNURBSCurve* curve2)
{
    Clear();

    // retrieve data from given curves
    gfInt u_degree = curve1->GetDegree();
    gfInt num_u_points = curve1->GetNumControlPoints();
    gfInt num_u_knots = curve1->GetNumKnots();

    gfInt v_degree = curve2->GetDegree();
    gfInt num_v_points = curve2->GetNumControlPoints();
    gfInt num_v_knots = curve2->GetNumKnots();

    // set parameters of patch
    SetParams(u_degree, v_degree, num_u_points, num_v_points);

    // set u and v knots of patch
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, curve1->GetKnot(i));
    for (gfInt i = 0;i < num_v_knots;i++)
        SetVKnot(i, curve2->GetKnot(i));

    gfPoint3* pivot = curve1->GetControlPoint(1);
    if (!pivot)
        return GF_E_FAIL;

    // get newell normals and center of curves
    gfDouble n2x, n2y, n2z;
    n2x = n2y = n2z = 0.0;
    gfDouble cx, cy, cz;
    cx = cy = cz = 0.0;

    for (gfInt i = 0;i < num_v_points;i++)
    {
        gfPoint3* point1 = curve2->GetControlPoint(i);
        gfPoint3* point2 = curve2->GetControlPoint((i+1)%num_v_points);

        cx += point1->X()/num_v_points;
        cy += point1->Y()/num_v_points;
        cz += point1->Z()/num_v_points;

        n2x += (point2->Y()-point1->Y())*(point2->Z()+point1->Z());
        n2y += (point2->Z()-point1->Z())*(point2->X()+point1->X());
        n2z += (point2->X()-point1->X())*(point2->Y()+point1->Y());
    }

    gfVector3 normal(n2x, n2y, n2z);
    normal.Normalize();
    gfPoint3 center(cx, cy, cz);

    // set control points and weights
    for (gfInt i = 0;i < num_u_points;i++)
    {
        gfPoint3* point = curve1->GetControlPoint(i);

        // find quasi direction vector of control points
        gfVector3 vec1(0.0, 0.0, 0.0);
        if (i > 0)
        {
            gfPoint3* point_prev = curve1->GetControlPoint(i-1);
            vec1.Set(point->X()-point_prev->X(), point->Y()-point_prev->Y(), point->Z()-point_prev->Z());
        }
        gfVector3 vec2(0.0, 0.0, 0.0);
        if (i < num_u_points-1)
        {
            gfPoint3* point_next = curve1->GetControlPoint(i+1);
            vec2.Set(point_next->X()-point->X(), point_next->Y()-point->Y(), point_next->Z()-point->Z());
        }
        gfVector3 dir;
        gfCombineVectors3(1.0, &vec1, 1.0, &vec2, &dir);
        dir.Normalize();

        gfDouble angle = gfVectorsAngle3(&dir, &normal);

        gfVector3 up;
        gfVectorProduct3(&dir, &normal, &up);
        up.Normalize();
        if (angle < 0.1)
            angle = 0.0;
        if (angle > 3.0)
            up.Set(0.0, 0.0, 1.0);

        for (gfInt j = 0;j < num_v_points;j++)
        {
            gfPoint3* point2 = curve2->GetControlPoint(j);
            gfPoint3 pt_aux(point2->X(), point2->Y(), point2->Z());

            //if (angle > 0.14 || angle < 3.0)
                gfRotatePoint3(&center, &up, angle, &pt_aux);

            gfDouble px = point->X()+pt_aux.X()-center.X();
            gfDouble py = point->Y()+pt_aux.Y()-center.Y();
            gfDouble pz = point->Z()+pt_aux.Z()-center.Z();

            gfPoint3* pt = GetControlPoint(i, j);
            pt->Set(px, py, pz);
            SetWeight(i, j, curve1->GetWeight(i) * curve2->GetWeight(j));
        }
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitRuledCurveToLine(gfObjNURBSCurve* curve, gfPoint3* line_pt, gfVector3* line_dir)
{
    Clear();

    // retrieve data from given curves
    gfInt u_degree = curve->GetDegree();
    gfInt num_u_points = curve->GetNumControlPoints();
    gfInt num_u_knots = curve->GetNumKnots();

    gfInt v_degree = 1;
    gfInt num_v_points = 2;

    // set parameters of patch
    SetParams(u_degree, v_degree, num_u_points, num_v_points);
    gfInt num_v_knots = GetNumVKnots();

    // copy u knots from curve
    for (gfInt i = 0;i < num_u_knots;i++)
        SetUKnot(i, curve->GetKnot(i));
    SetVKnot(0, 0.0);SetVKnot(1, 0.0);
    SetVKnot(2, 1.0);SetVKnot(3, 1.0);

    gfDouble denom = line_dir->X()*line_dir->X() + line_dir->Y()*line_dir->Y() + line_dir->Z()*line_dir->Z();
    if (gfIsZero(denom))
        return GF_E_FAIL;
    gfDouble aux = line_pt->X()*line_dir->X() + line_pt->Y()*line_dir->Y() + line_pt->Z()*line_dir->Z();

    for (gfInt i = 0;i < num_u_points;i++)
    {
        gfPoint3* curve_pt = curve->GetControlPoint(i);

        gfPoint3* ctrl_point = GetControlPoint(i, 0);
        ctrl_point->Set(curve_pt->X(), curve_pt->Y(), curve_pt->Z());

        ctrl_point = GetControlPoint(i, 1);
        gfDouble t = (curve_pt->X()*line_dir->X()+curve_pt->Y()*line_dir->Y()+curve_pt->Z()*line_dir->Z()-aux) / denom;
        ctrl_point->Set(line_pt->X() + t * line_dir->X(),
                        line_pt->Y() + t * line_dir->Y(),
                        line_pt->Z() + t * line_dir->Z());

    }
        
    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InitProfile(gfDouble xmin, gfDouble xmax, gfInt num_u_points, gfDouble ymin, gfDouble ymax, gfInt num_v_points)
{
    SetParams(2, 2, num_u_points, num_v_points);

    gfInt num_u_knots = GetNumUKnots();
    gfInt num_v_knots = GetNumVKnots();

    for (gfInt i = 0;i <= m_u_degree;i++)
    {
        SetUKnot(i, 0);
    }
    for (gfInt i = m_u_degree+1;i < num_u_knots-m_u_degree-1;i++)
    {
        SetUKnot(i, (gfDouble) i-m_u_degree);
    }
    for (gfInt i = num_u_knots-m_u_degree-1;i < num_u_knots;i++)
    {
        SetUKnot(i, (gfDouble) num_u_knots-m_u_degree-2);
    }

    for (gfInt i = 0;i <= m_v_degree;i++)
    {
        SetVKnot(i, 0);
    }
    for (gfInt i = m_v_degree+1;i < num_v_knots-m_v_degree-1;i++)
    {
        SetVKnot(i, (gfDouble) i-m_v_degree);
    }
    for (gfInt i = num_v_knots-m_v_degree-1;i < num_v_knots;i++)
    {
        SetVKnot(i, (gfDouble) num_v_knots-m_v_degree-2);
    }

    for (gfInt i = 0;i < num_u_points;i++)
        for (gfInt j = 0;j < num_v_points;j++)
        {
            gfPoint3* pt = GetControlPoint(i, j);

            pt->Set(xmin+(xmax-xmin)*i/(num_u_points-1),
                    ymin+(ymax-ymin)*j/(num_v_points-1),
                    gfRand(-1.0,1.0));

            SetWeight(i, j, 1.0);
        }

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::SetParams(gfInt u_degree, gfInt v_degree, gfInt num_u_control_points, gfInt num_v_control_points)
{
    DeleteWholeAABBTree();
    DeleteVACache();

    // create auxiliary arrays and fill them with old 
    std::vector<gfDouble> aux_weights;
    std::vector<gfPoint3 *> aux_control_points;
    aux_weights.resize(num_u_control_points * num_v_control_points);
    aux_control_points.resize(num_u_control_points * num_v_control_points);

    m_u_knot_vector.resize(u_degree + num_u_control_points + 1);
    m_v_knot_vector.resize(v_degree + num_v_control_points + 1);
    m_u_degree = u_degree;
    m_v_degree = v_degree;

    gfInt umax = (m_num_u_control_points > num_u_control_points) ? m_num_u_control_points : num_u_control_points;
    gfInt vmax = (m_num_v_control_points > num_v_control_points) ? m_num_v_control_points : num_v_control_points;

    for (gfInt i = 0;i < umax;i++)
    {
        for (gfInt j = 0;j < vmax;j++)
        {
            // point in new & old
            if ((i < num_u_control_points && j < num_v_control_points) && (i < m_num_u_control_points && j < m_num_v_control_points))
            {
                aux_weights[i + num_u_control_points * j] = GetWeight(i, j);
                aux_control_points[i + num_u_control_points * j] = GetControlPoint(i, j);
            }

            // point in new & not in old
            if ((i < num_u_control_points && j < num_v_control_points) && !(i < m_num_u_control_points && j < m_num_v_control_points))
            {
                aux_weights[i + num_u_control_points * j] = 1.0;
                gfPoint3* new_pt = gfCreatePoint3();
                aux_control_points[i + num_u_control_points * j] = new_pt;
            }

            // point in old & not in new
            if (!(i < num_u_control_points && j < num_v_control_points) && (i < m_num_u_control_points && j < m_num_v_control_points))
            {
                gfPoint3* old_pt = GetControlPoint(i, j);
                gfDeletePoint3(old_pt);
            }
        }
    }

    m_num_u_control_points = num_u_control_points;
    m_num_v_control_points = num_v_control_points;

    m_control_points.resize(m_num_u_control_points * m_num_v_control_points);
    m_weights.resize(m_num_u_control_points * m_num_v_control_points);

    // copy from auxiliary arrays
    for (gfInt i = 0;i < m_num_u_control_points * m_num_v_control_points;i++)
    {
        m_control_points[i] = aux_control_points.at(i);
        m_weights[i] = aux_weights.at(i);
    }

    m_time_stamp++;

    return GF_S_OK;
}

gfInt gfObjNURBSPatch::GetUDegree(void)
{
    return m_u_degree;
}

gfInt gfObjNURBSPatch::GetVDegree(void)
{
    return m_v_degree;
}

gfInt gfObjNURBSPatch::GetNumUKnots(void)
{
    return (gfInt)m_u_knot_vector.size();
}

gfInt gfObjNURBSPatch::GetNumVKnots(void)
{
    return (gfInt)m_v_knot_vector.size();
}

gfDouble gfObjNURBSPatch::GetUKnot(gfInt index)
{
    return m_u_knot_vector.at(index);
}

gfDouble gfObjNURBSPatch::GetVKnot(gfInt index)
{
    return m_v_knot_vector.at(index);
}

GFRESULT gfObjNURBSPatch::SetUKnot(gfInt index, gfDouble knot)
{
    m_time_stamp++;
    m_u_knot_vector[index] = knot;
    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::SetVKnot(gfInt index, gfDouble knot)
{
    m_time_stamp++;
    m_v_knot_vector[index] = knot;
    return GF_S_OK;
}

gfInt gfObjNURBSPatch::GetNumUControlPoints(void)
{
    return m_num_u_control_points;
}

gfInt gfObjNURBSPatch::GetNumVControlPoints(void)
{
    return m_num_v_control_points;
}

gfPoint3 * gfObjNURBSPatch::GetControlPoint(gfInt u_index, gfInt v_index)
{
    gfInt u = u_index;
    if (u > (m_num_u_control_points-1)) u = m_num_u_control_points-1;
    if (u < 0) u = 0;

    gfInt v = v_index;
    if (v > (m_num_v_control_points-1)) v = m_num_v_control_points-1;
    if (v < 0) v = 0;

    return m_control_points.at(u + v * m_num_u_control_points);
}

gfDouble gfObjNURBSPatch::GetWeight(gfInt u_index, gfInt v_index)
{
    gfInt u = u_index;
    if (u > (m_num_u_control_points-1)) u = m_num_u_control_points-1;
    if (u < 0) u = 0;

    gfInt v = v_index;
    if (v > (m_num_v_control_points-1)) v = m_num_v_control_points-1;
    if (v < 0) v = 0;

    return m_weights.at(u + v * m_num_u_control_points);
}

GFRESULT gfObjNURBSPatch::SetWeight(gfInt u_index, gfInt v_index, gfDouble weight)
{
    m_time_stamp++;
    m_weights[u_index + v_index * m_num_u_control_points] = weight;
    return GF_S_OK;
}

gfPoint3* gfObjNURBSPatch::GetPatchPoint(gfDouble u, gfDouble v)
{
    gfInt n = GetNumUControlPoints() - 1;
    gfInt m = GetNumVControlPoints() - 1;

    // current u parameter
    gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
    std::vector<gfDouble> ubasis;
    gfGetBSplineAllValues(m_u_degree, &m_u_knot_vector, uspan, u, &ubasis);

    // current v parameter     
    gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
    std::vector<gfDouble> vbasis;
    gfGetBSplineAllValues(m_v_degree, &m_v_knot_vector, vspan, v, &vbasis);

    gfDouble ax = 0;
    gfDouble ay = 0;
    gfDouble az = 0;
    gfDouble aw = 0;

    for (gfInt k = 0;k <= m_u_degree;k++)
    {
        gfInt uindex = uspan - m_u_degree + k;

        for (gfInt l = 0;l <= m_v_degree;l++)
        {
            gfInt vindex = vspan - m_v_degree + l;
            gfDouble aux = ubasis.at(k) * vbasis.at(l) * m_weights.at(uindex + m_num_u_control_points * vindex);
            ax += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
            ay += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
            az += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
            aw += aux;
        }
    }

    gfPoint3* result = gfCreatePoint3();
    result->Set(ax / aw, ay / aw, az / aw);

    return result;
}

gfVector3* gfObjNURBSPatch::GetPatchUDer(gfDouble u, gfDouble v)
{
    gfInt n = GetNumUControlPoints()  - 1;
    gfInt m = GetNumVControlPoints()  - 1;

    // current u parameter
    gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
    gfAux2DArray<gfDouble> ubasis(30, 30);
    gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 1, &ubasis);

    // current v parameter     
    gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
    gfAux2DArray<gfDouble> vbasis(30, 30);
    gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 0, &vbasis);

    gfDouble ax = 0.0;
    gfDouble ay = 0.0;
    gfDouble az = 0.0;
    gfDouble aw = 0.0;

    gfDouble dux = 0.0;
    gfDouble duy = 0.0;
    gfDouble duz = 0.0;
    gfDouble duw = 0.0;

    for (gfInt k = 0;k <= m_u_degree;k++)
    {
        gfInt uindex = uspan - m_u_degree + k;

        for (gfInt l = 0;l <= m_v_degree;l++)
        {
            gfInt vindex = vspan - m_v_degree + l;

            gfDouble aux = ubasis.at(0, k) * vbasis.at(0, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
            ax += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
            ay += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
            az += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
            aw += aux;

            gfDouble aux1 = ubasis.at(1, k) * vbasis.at(0, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
            dux += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
            duy += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
            duz += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
            duw += aux1;
        }
    }

    gfDouble fdux = (dux * aw - ax * duw) / aw / aw;
    gfDouble fduy = (duy * aw - ay * duw) / aw / aw;
    gfDouble fduz = (duz * aw - az * duw) / aw / aw;

    gfVector3* result = gfCreateVector3();
    result->Set(fdux, fduy, fduz);

    return result;
}

gfVector3* gfObjNURBSPatch::GetPatchVDer(gfDouble u, gfDouble v)
{
    gfInt n = GetNumUControlPoints()  - 1;
    gfInt m = GetNumVControlPoints()  - 1;

    // current u parameter
    gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
    gfAux2DArray<gfDouble> ubasis(30, 30);
    gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 0, &ubasis);

    // current v parameter     
    gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
    gfAux2DArray<gfDouble> vbasis(30, 30);
    gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 1, &vbasis);

    gfDouble ax = 0.0;
    gfDouble ay = 0.0;
    gfDouble az = 0.0;
    gfDouble aw = 0.0;

    gfDouble dvx = 0.0;
    gfDouble dvy = 0.0;
    gfDouble dvz = 0.0;
    gfDouble dvw = 0.0;

    for (gfInt k = 0;k <= m_u_degree;k++)
    {
        gfInt uindex = uspan - m_u_degree + k;

        for (gfInt l = 0;l <= m_v_degree;l++)
        {
            gfInt vindex = vspan - m_v_degree + l;

            gfDouble aux = ubasis.at(0, k) * vbasis.at(0, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
            ax += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
            ay += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
            az += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
            aw += aux;

            gfDouble aux1 = ubasis.at(0, k) * vbasis.at(1, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
            dvx += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
            dvy += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
            dvz += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
            dvw += aux1;
        }
    }

    gfDouble fdvx = (dvx * aw - ax * dvw) / aw / aw;
    gfDouble fdvy = (dvy * aw - ay * dvw) / aw / aw;
    gfDouble fdvz = (dvz * aw - az * dvw) / aw / aw;

    gfVector3* result = gfCreateVector3();
    result->Set(fdvx, fdvy, fdvz);

    return result;
}

gfVector3* gfObjNURBSPatch::GetPatchNormal(gfDouble u, gfDouble v)
{
    gfInt n = GetNumUControlPoints()  - 1;
    gfInt m = GetNumVControlPoints()  - 1;

    // current u parameter
    gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
    gfAux2DArray<gfDouble> ubasis(30, 30);
    gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 1, &ubasis);

    // current v parameter     
    gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
    gfAux2DArray<gfDouble> vbasis(30, 30);
    gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 1, &vbasis);

    gfDouble ax = 0.0;
    gfDouble ay = 0.0;
    gfDouble az = 0.0;
    gfDouble aw = 0.0;

    gfDouble dux = 0.0;
    gfDouble duy = 0.0;
    gfDouble duz = 0.0;
    gfDouble duw = 0.0;

    gfDouble dvx = 0.0;
    gfDouble dvy = 0.0;
    gfDouble dvz = 0.0;
    gfDouble dvw = 0.0;

    for (gfInt k = 0;k <= m_u_degree;k++)
    {
        gfInt uindex = uspan - m_u_degree + k;

        for (gfInt l = 0;l <= m_v_degree;l++)
        {
            gfInt vindex = vspan - m_v_degree + l;

            gfDouble aux = ubasis.at(0, k) * vbasis.at(0, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
            ax += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
            ay += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
            az += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
            aw += aux;

            gfDouble aux1 = ubasis.at(1, k) * vbasis.at(0, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
            dux += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
            duy += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
            duz += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
            duw += aux1;

            gfDouble aux2 = ubasis.at(0, k) * vbasis.at(1, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
            dvx += aux2 * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
            dvy += aux2 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
            dvz += aux2 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
            dvw += aux2;
        }
    }

    gfDouble fdux = (dux * aw - ax * duw) / aw / aw;
    gfDouble fduy = (duy * aw - ay * duw) / aw / aw;
    gfDouble fduz = (duz * aw - az * duw) / aw / aw;

    gfDouble fdvx = (dvx * aw - ax * dvw) / aw / aw;
    gfDouble fdvy = (dvy * aw - ay * dvw) / aw / aw;
    gfDouble fdvz = (dvz * aw - az * dvw) / aw / aw;

    gfVector3* result = gfCreateVector3();
    result->Set((fduy*fdvz-fduz*fdvy),
        (-fdux*fdvz+fduz*fdvx),
        (fdux*fdvy-fduy*fdvx));
    result->Normalize();

    return result;
}

gfObjHalfEdgeMesh* gfObjNURBSPatch::GetControlMesh(void)
{
    gfObjHalfEdgeMesh* ret = (gfObjHalfEdgeMesh*) gfCreateGeomObject(GF_HALF_EDGE_MESH);

    gfHalfEdge** edges = new gfHalfEdge*[4 * (m_num_u_control_points - 1) * (m_num_v_control_points - 1)];
    gfHEPoint** points = new gfHEPoint*[m_num_u_control_points * m_num_v_control_points];

    // create vertices
    for (gfInt i = 0;i < m_num_u_control_points;i++)
    {
        for (gfInt j = 0;j < m_num_v_control_points;j++)
        {
            gfHEPoint* point = gfCreateHEPoint();

            gfPoint3 * pt = GetControlPoint(i, j);
            point->x = pt->X();
            point->y = pt->Y();
            point->z = pt->Z();
            point->edge = NULL;
            points[i + m_num_u_control_points * j] = point;
            ret->AddHEPoint(point);
        }
    }

    // create faces and edges
    for (gfInt i = 0;i < m_num_u_control_points - 1;i++)
    {
        for (gfInt j = 0;j < m_num_v_control_points - 1;j++)
        {
            gfHEFace* face = gfCreateHEFace();
            gfHalfEdge* edge1 = gfCreateHalfEdge();
            gfHalfEdge* edge2 = gfCreateHalfEdge();
            gfHalfEdge* edge3 = gfCreateHalfEdge();
            gfHalfEdge* edge4 = gfCreateHalfEdge();

            edge1->end = points[i + 1 + m_num_u_control_points * j];
            if (edge1->end->edge == NULL)
                edge1->end->edge = edge1;
            edge1->face = face;
            edge1->next = edge2;

            if (j > 0)
            {
                edges[4 * (i + (m_num_u_control_points - 1) * (j - 1)) + 2]->opp = edge1;
                edge1->opp = edges[4 * (i + (m_num_u_control_points - 1) * (j - 1)) + 2];
            }
            else
            {
                edge1->opp = NULL;
            }
            edges[4 * (i + (m_num_u_control_points - 1) * j) + 0] = edge1;

            edge2->end = points[i + 1 + m_num_u_control_points * (j + 1)];
            if (edge2->end->edge == NULL)
                edge2->end->edge = edge2;
            edge2->face = face;
            edge2->next = edge3;
            edge2->opp = NULL;
            edges[4 * (i + (m_num_u_control_points - 1) * j) + 1] = edge2;

            edge3->end = points[i + m_num_u_control_points * (j + 1)];
            if (edge3->end->edge == NULL)
                edge3->end->edge = edge3;
            edge3->face = face;
            edge3->next = edge4;
            edge3->opp = NULL;
            edges[4 * (i + (m_num_u_control_points - 1) * j) + 2] = edge3;

            edge4->end = points[i + m_num_u_control_points * j];
            if (edge4->end->edge == NULL)
                edge4->end->edge = edge4;
            edge4->face = face;
            edge4->next = edge1;
            if (i > 0)
            {
                edges[4 * (i - 1 + (m_num_u_control_points - 1) * j) + 1]->opp = edge4;
                edge4->opp = edges[4 * (i - 1 + (m_num_u_control_points - 1) * j) + 1];
            }
            else
            {
                edge4->opp = NULL;
            }
            edges[4 * (i + (m_num_u_control_points - 1) * j) + 3] = edge4;

            face->edge = edge1;

            // add all to mesh
            ret->AddHEFace(face);
            ret->AddHalfEdge(edge1);
            ret->AddHalfEdge(edge2);
            ret->AddHalfEdge(edge3);
            ret->AddHalfEdge(edge4);
        }
    }

    delete [] edges;
    delete [] points;

    return ret;
}

gfObjHalfEdgeMesh* gfObjNURBSPatch::GetApproxMesh(gfInt approx)
{
    // Interval of definition
    gfDouble a = GetUKnot(m_u_degree);
    gfDouble b = GetUKnot(GetNumUControlPoints());
    if (gfIsZero(b - a))
        return NULL;

    // gfInterval of definition
    gfDouble c = GetVKnot(m_v_degree);
    gfDouble d = GetVKnot(GetNumVControlPoints());
    if (gfIsZero(d - c))
        return NULL;

    gfInt n = GetNumUControlPoints()  - 1;
    gfInt m = GetNumVControlPoints()  - 1;

    gfHalfEdge** edges = new gfHalfEdge*[4 * (approx - 1) * (approx - 1)];
    gfHEPoint** points = new gfHEPoint*[approx * approx];
    gfObjHalfEdgeMesh* ret = (gfObjHalfEdgeMesh*) gfCreateGeomObject(GF_HALF_EDGE_MESH);

    // create vertices
    for (gfInt i = 0;i < approx;i++)
    {
        // current u parameter
        gfDouble u = a + i * (b - a) / (approx - 1);
        gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
        std::vector<gfDouble> ubasis;
        gfGetBSplineAllValues(m_u_degree, &m_u_knot_vector, uspan, u, &ubasis);

        for (gfInt j = 0;j < approx;j++)
        {
            // current v parameter     
            gfDouble v = c + j * (d - c) / (approx - 1);
            gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
            std::vector<gfDouble> vbasis;
            gfGetBSplineAllValues(m_v_degree, &m_v_knot_vector, vspan, v, &vbasis);

            gfDouble ax = 0;
            gfDouble ay = 0;
            gfDouble az = 0;
            gfDouble aw = 0;

            for (gfInt k = 0;k <= m_u_degree;k++)
            {
                gfInt uindex = uspan - m_u_degree + k;

                for (gfInt l = 0;l <= m_v_degree;l++)
                {
                    gfInt vindex = vspan - m_v_degree + l;
                    gfDouble aux = ubasis.at(k) * vbasis.at(l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                    ax += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
                    ay += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
                    az += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
                    aw += aux;
                }
            }

            // set computed point
            gfHEPoint* point = gfCreateHEPoint();

            gfPoint3 * pt = GetControlPoint(i, j);
            point->x = ax / aw;
            point->y = ay / aw;
            point->z = az / aw;
            point->edge = NULL;
            points[i + approx * j] = point;
            ret->AddHEPoint(point);
        }
    }

    // create faces and edges
    for (gfInt i = 0;i < approx - 1;i++)
    {
        for (gfInt j = 0;j < approx - 1;j++)
        {
            gfHEFace* face = gfCreateHEFace();
            gfHalfEdge* edge1 = gfCreateHalfEdge();
            gfHalfEdge* edge2 = gfCreateHalfEdge();
            gfHalfEdge* edge3 = gfCreateHalfEdge();
            gfHalfEdge* edge4 = gfCreateHalfEdge();

            edge1->end = points[i + 1 + approx * j];
            if (edge1->end->edge == NULL)
                edge1->end->edge = edge1;
            edge1->face = face;
            edge1->next = edge2;
            if (j > 0)
            {
                edges[4 * (i + (approx - 1) * (j - 1)) + 2]->opp = edge1;
                edge1->opp = edges[4 * (i + (approx - 1) * (j - 1)) + 2];
            }
            else
            {
                edge1->opp = NULL;
            }
            edges[4 * (i + (approx - 1) * j) + 0] = edge1;

            edge2->end = points[i + 1 + approx * (j + 1)];
            if (edge2->end->edge == NULL)
                edge2->end->edge = edge2;
            edge2->face = face;
            edge2->next = edge3;
            edge2->opp = NULL;
            edges[4 * (i + (approx - 1) * j) + 1] = edge2;

            edge3->end = points[i + approx * (j + 1)];
            if (edge3->end->edge == NULL)
                edge3->end->edge = edge3;
            edge3->face = face;
            edge3->next = edge4;
            edge3->opp = NULL;
            edges[4 * (i + (approx - 1) * j) + 2] = edge3;

            edge4->end = points[i + approx * j];
            if (edge4->end->edge == NULL)
                edge4->end->edge = edge4;
            edge4->face = face;
            edge4->next = edge1;
            if (i > 0)
            {
                edges[4 * (i - 1 + (approx - 1) * j) + 1]->opp = edge4;
                edge4->opp = edges[4 * (i - 1 + (approx - 1) * j) + 1];
            }
            else
            {
                edge4->opp = NULL;
            }
            edges[4 * (i + (approx - 1) * j) + 3] = edge4;

            face->edge = edge1;

            // add all to mesh
            ret->AddHEFace(face);
            ret->AddHalfEdge(edge1);
            ret->AddHalfEdge(edge2);
            ret->AddHalfEdge(edge3);
            ret->AddHalfEdge(edge4);
        }
    }

    delete [] edges;
    delete [] points;

    return ret;
}

gfObjHalfEdgeMesh* gfObjNURBSPatch::GetTrimmedApproxMesh(gfInt approx)
{
    // Interval of definition
    gfDouble a = GetUKnot(m_u_degree);
    gfDouble b = GetUKnot(GetNumUControlPoints());
    if (gfIsZero(b - a))
        return NULL;

    // gfInterval of definition
    gfDouble c = GetUKnot(m_v_degree);
    gfDouble d = GetUKnot(GetNumVControlPoints());
    if (gfIsZero(d - c))
        return NULL;

    gfInt n = GetNumUControlPoints()  - 1;
    gfInt m = GetNumVControlPoints()  - 1;

    gfHalfEdge** edges = new gfHalfEdge*[4 * (approx - 1) * (approx - 1)];
    gfHEPoint** points = new gfHEPoint*[approx * approx];
    gfObjHalfEdgeMesh* ret = (gfObjHalfEdgeMesh*) gfCreateGeomObject(GF_HALF_EDGE_MESH);


    // get UV space tesselation based on trim curves
    gfUVPoint* trimmed_points = TesselateTrimmedUVSpace(approx);

    // create vertices
    for (gfInt i = 0;i < approx;i++)
    {
        // current u parameter
        gfDouble u = a + i * (b - a) / (approx - 1);
        gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
        std::vector<gfDouble> ubasis;
        gfGetBSplineAllValues(m_u_degree, &m_u_knot_vector, uspan, u, &ubasis);

        for (gfInt j = 0;j < approx;j++)
        {
            // current v parameter     
            gfDouble v = c + j * (d - c) / (approx - 1);
            gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
            std::vector<gfDouble> vbasis;
            gfGetBSplineAllValues(m_v_degree, &m_v_knot_vector, vspan, v, &vbasis);

            gfDouble ax = 0;
            gfDouble ay = 0;
            gfDouble az = 0;
            gfDouble aw = 0;

            for (gfInt k = 0;k <= m_u_degree;k++)
            {
                gfInt uindex = uspan - m_u_degree + k;

                for (gfInt l = 0;l <= m_v_degree;l++)
                {
                    gfInt vindex = vspan - m_v_degree + l;
                    gfDouble aux = ubasis.at(k) * vbasis.at(l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                    ax += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
                    ay += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
                    az += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
                    aw += aux;
                }
            }

            // set computed point
            gfHEPoint* point = gfCreateHEPoint();

            if (trimmed_points[i + j * approx].boundary)
            {
                gfPoint3 * patch_pt = GetPatchPoint(u + trimmed_points[i + j * approx].du, v + trimmed_points[i + j * approx].dv);
                point->x = patch_pt->X();
                point->y = patch_pt->Y();
                point->z = patch_pt->Z();
            }
            else
            {
                point->x = ax / aw;
                point->y = ay / aw;
                point->z = az / aw;
            }

            point->curvature = trimmed_points[i + j * approx].winding_number;
            point->edge = NULL;
            points[i + approx * j] = point;
            ret->AddHEPoint(point);
        }
    }

    std::vector<gfHEFace*> faces_to_trim;

    // create faces and edges
    for (gfInt i = 0;i < approx - 1;i++)
    {
        for (gfInt j = 0;j < approx - 1;j++)
        {
            gfHEFace* face = gfCreateHEFace();
            gfHalfEdge* edge1 = gfCreateHalfEdge();
            gfHalfEdge* edge2 = gfCreateHalfEdge();
            gfHalfEdge* edge3 = gfCreateHalfEdge();
            gfHalfEdge* edge4 = gfCreateHalfEdge();

            edge1->end = points[i + 1 + approx * j];
            if (edge1->end->edge == NULL)
                edge1->end->edge = edge1;
            edge1->face = face;
            edge1->next = edge2;
            if (j > 0)
            {
                edges[4 * (i + (approx - 1) * (j - 1)) + 2]->opp = edge1;
                edge1->opp = edges[4 * (i + (approx - 1) * (j - 1)) + 2];
            }
            else
            {
                edge1->opp = NULL;
            }
            edges[4 * (i + (approx - 1) * j) + 0] = edge1;

            edge2->end = points[i + 1 + approx * (j + 1)];
            if (edge2->end->edge == NULL)
                edge2->end->edge = edge2;
            edge2->face = face;
            edge2->next = edge3;
            edge2->opp = NULL;
            edges[4 * (i + (approx - 1) * j) + 1] = edge2;

            edge3->end = points[i + approx * (j + 1)];
            if (edge3->end->edge == NULL)
                edge3->end->edge = edge3;
            edge3->face = face;
            edge3->next = edge4;
            edge3->opp = NULL;
            edges[4 * (i + (approx - 1) * j) + 2] = edge3;

            edge4->end = points[i + approx * j];
            if (edge4->end->edge == NULL)
                edge4->end->edge = edge4;
            edge4->face = face;
            edge4->next = edge1;
            if (i > 0)
            {
                edges[4 * (i - 1 + (approx - 1) * j) + 1]->opp = edge4;
                edge4->opp = edges[4 * (i - 1 + (approx - 1) * j) + 1];
            }
            else
            {
                edge4->opp = NULL;
            }
            edges[4 * (i + (approx - 1) * j) + 3] = edge4;

            face->edge = edge1;

            // add all to mesh
            ret->AddHalfEdge(edge1);
            ret->AddHalfEdge(edge2);
            ret->AddHalfEdge(edge3);
            ret->AddHalfEdge(edge4);
            ret->AddHEFace(face);

            // is face trimmed ?
            gfInt wind1 = (gfInt) points[i + 1 + approx * j]->curvature;
            gfInt wind2 = (gfInt) points[i + 1 + approx * (j + 1)]->curvature;
            gfInt wind3 = (gfInt) points[i + approx * (j + 1)]->curvature;
            gfInt wind4 = (gfInt) points[i + approx * j]->curvature;

            if (wind1 % 2 == 1 || wind2 % 2 == 1 || wind3 % 2 == 1 || wind4 % 2 == 1)
                faces_to_trim.push_back(face);
        }
    }

    // remove trimmed faces and edges
    gfInt num_trimmed_faces = (gfInt)faces_to_trim.size();
    for (gfInt i = 0;i < num_trimmed_faces;i++)
    {
        gfHEFace* face = faces_to_trim.at(i);

        ret->RemoveHEFace(face);
    }

    delete [] edges;
    delete [] points;
    delete [] trimmed_points;

    return ret;
}

GFRESULT gfObjNURBSPatch::GetDomain(gfDouble* u_min, gfDouble* u_max, gfDouble* v_min, gfDouble* v_max)
{
    *u_min = GetUKnot(m_u_degree);
    *u_max = GetUKnot(GetNumUControlPoints());
    *v_min = GetVKnot(m_v_degree);
    *v_max = GetVKnot(GetNumVControlPoints());
    return GF_S_OK;
}

gfObjNURBSCurve* gfObjNURBSPatch::GetUIsocurve(gfDouble v_param)
{
    gfObjNURBSCurve* ret = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);

    ret->SetParams(m_u_degree, m_num_u_control_points);

    gfDouble a = GetVKnot(m_v_degree);
    gfDouble b = GetVKnot(GetNumVControlPoints());
    if (gfIsZero(b - a))
        return NULL;
    if (v_param >= b || v_param < a)
        return NULL;

    // fill knots
    gfInt num_u_knots = GetNumUKnots();
    for (gfInt i = 0;i < num_u_knots;i++)
    {
        ret->SetKnot(i, GetUKnot(i));
    }

    // prepare basis
    gfInt m = GetNumVControlPoints()  - 1;
    gfInt span = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v_param);
    std::vector<gfDouble> basis;
    gfGetBSplineAllValues(m_v_degree, &m_v_knot_vector, span, v_param, &basis);

    for (gfInt i = 0;i < m_num_u_control_points;i++)
    {
        gfPoint3 * pt = ret->GetControlPoint(i);

        gfDouble ax = 0;
        gfDouble ay = 0;
        gfDouble az = 0;
        gfDouble aw = 0;
        for (gfInt j = 0;j <= m_v_degree;j++)
        {
            gfInt index = span - m_v_degree + j;
            ax += basis.at(j) * GetWeight(i, index) * GetControlPoint(i, index)->X();
            ay += basis.at(j) * GetWeight(i, index) * GetControlPoint(i, index)->Y();
            az += basis.at(j) * GetWeight(i, index) * GetControlPoint(i, index)->Z();
            aw += basis.at(j) * GetWeight(i, index);
        }

        pt->Set(ax, ay, az);
        ret->SetWeight(i, aw);
    }

    return ret;
}

gfObjNURBSCurve* gfObjNURBSPatch::GetVIsocurve(gfDouble u_param)
{
    // create curve
    gfObjNURBSCurve* ret = (gfObjNURBSCurve*) gfCreateGeomObject(GF_NURBS_CURVE);
    ret->SetParams(m_v_degree, m_num_v_control_points);

    gfDouble a = GetUKnot(m_u_degree);
    gfDouble b = GetUKnot(GetNumUControlPoints());
    if (gfIsZero(b - a))
        return NULL;
    if (u_param >= b || u_param < a)
        return NULL;

    // fill knots
    gfInt num_v_knots = GetNumVKnots();
    for (gfInt i = 0;i < num_v_knots;i++)
    {
        ret->SetKnot(i, GetVKnot(i));
    }

    // prepare basis
    gfInt n = GetNumUControlPoints()  - 1;
    gfInt span = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u_param);
    std::vector<gfDouble> basis;
    gfGetBSplineAllValues(m_u_degree, &m_u_knot_vector, span, u_param, &basis);

    for (gfInt i = 0;i < m_num_v_control_points;i++)
    {
        gfPoint3 * pt = ret->GetControlPoint(i);

        gfDouble ax = 0;
        gfDouble ay = 0;
        gfDouble az = 0;
        gfDouble aw = 0;
        for (gfInt j = 0;j <= m_u_degree;j++)
        {
            gfInt index = span - m_u_degree + j;
            ax += basis.at(j) * GetWeight(index, i) * GetControlPoint(index, i)->X();
            ay += basis.at(j) * GetWeight(index, i) * GetControlPoint(index, i)->Y();
            az += basis.at(j) * GetWeight(index, i) * GetControlPoint(index, i)->Z();
            aw += basis.at(j) * GetWeight(index, i);
        }

        pt->Set(ax, ay, az);
        ret->SetWeight(i, aw);
    }

    return ret;
}

gfBool gfObjNURBSPatch::IsBezier(void)
{
    // check parameters of patch
    if (GetUDegree() != (GetNumUControlPoints() - 1))
        return false;
    if (GetVDegree() != (GetNumVControlPoints() - 1))
        return false;

    // get num of knots
    gfInt num_u_knots = GetNumUKnots();
    gfInt num_v_knots = GetNumVKnots();
    if (num_u_knots == 0)
        return true;
    if (num_v_knots == 0)
        return true;

    // check u knots
    for (gfInt i = 0;i < num_u_knots / 2;i++)
    {
        if (GetUKnot(i) != GetUKnot(0))
            return false;
    }
    for (gfInt i = num_u_knots / 2;i < num_u_knots ;i++)
    {
        if (GetUKnot(i) != GetUKnot(num_u_knots - 1))
            return false;
    }

    // check v knots
    for (gfInt i = 0;i < num_v_knots / 2;i++)
    {
        if (GetVKnot(i) != GetVKnot(0))
            return false;
    }
    for (gfInt i = num_v_knots / 2;i < num_v_knots ;i++)
    {
        if (GetVKnot(i) != GetVKnot(num_v_knots - 1))
            return false;
    }

    return true;
}

GFRESULT gfObjNURBSPatch::DecomposeToBezierSegments(gfAux2DArray<gfObjNURBSPatch*>* segments)
{
    gfInt n = GetNumUControlPoints() - 1;
    gfInt m = GetNumVControlPoints() - 1;

    // first decompose in u direction
    gfInt max = n + 2;
    gfInt a = m_u_degree;
    gfInt b = m_u_degree + 1;
    gfInt nb = 0;
    gfInt i, j, k;
    gfAux3DArray<gfPoint4> Qw;
    Qw.resize(n, m_u_degree + 1, m+1);
    std::vector<gfDouble> alphas;
    alphas.resize(m_u_degree + 1);

    for (i = 0;i <= m_u_degree;i++)
    {
        for (gfInt row = 0;row <= m;row++)
        {
            gfPoint3* pt = GetControlPoint(i, row);
            Qw.Set(nb, i, row, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(i, row)));
        }
    }

    while (b < max)
    {
        i = b;
        while (b < (max) && (m_u_knot_vector.at(b+1) == m_u_knot_vector.at(b)))
            b++;
        gfInt mult = b-i+1;
        if (mult < m_u_degree)
        {
            gfDouble numer = m_u_knot_vector.at(b) - m_u_knot_vector.at(a);
            for (j = m_u_degree;j > mult;j--)
                alphas[j-mult-1] = numer/(m_u_knot_vector.at(a+j) - m_u_knot_vector.at(a));
            gfInt r = m_u_degree - mult;
            for (j = 1;j <= r;j++)
            {
                gfInt save = r-j;
                gfInt s = mult+j;
                for  (k = m_u_degree;k >= s;k--)
                {
                    gfDouble alpha = alphas.at(k-s);
                    for (gfInt row = 0;row <= m;row++)
                        Qw.Set(nb, k, row, gfPoint4(alpha*Qw.at(nb, k, row).X()+(1.0-alpha)*Qw.at(nb, k-1, row).X(),
                        alpha*Qw.at(nb, k, row).Y()+(1.0-alpha)*Qw.at(nb, k-1, row).Y(),
                        alpha*Qw.at(nb, k, row).Z()+(1.0-alpha)*Qw.at(nb, k-1, row).Z(),
                        alpha*Qw.at(nb, k, row).W()+(1.0-alpha)*Qw.at(nb, k-1, row).W()));
                }

                if (b < max)
                {
                    for (gfInt row = 0;row <= m;row++)
                        Qw.Set(nb + 1, save, row, gfPoint4(Qw.at(nb, m_u_degree, row).X(), 
                        Qw.at(nb, m_u_degree, row).Y(), 
                        Qw.at(nb, m_u_degree, row).Z(), 
                        Qw.at(nb, m_u_degree, row).W()));
                }
            }
        }

        nb++;
        if (b < max) 
        {
            for (i = m_u_degree-mult;i <= m_u_degree;i++)
            {
                for (gfInt row = 0;row <= m;row++)
                {
                    gfPoint3* pt = GetControlPoint(b-m_u_degree+i, row);
                    Qw.Set(nb, i, row, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(b-m_u_degree+i, row)));
                }
            }
            a = b;
            b++;
        }
    }

    // copy points from segments to new array of control points
    gfAux2DArray<gfPoint4> Pw;
    gfInt num_u_seg = nb;
    Pw.resize(m_u_degree*num_u_seg+1, m+1);
    for (gfInt row = 0;row <= m;row++)
    {
        for (gfInt seg = 0;seg < num_u_seg;seg++)
        {
            for (i = 0;i < m_u_degree;i++)
            {
                gfPoint4 pt = Qw.at(seg, i, row);
                Pw.Set(i+seg*m_u_degree, row, pt);
            }
        }
        gfPoint4 pt = Qw.at(num_u_seg-1, m_u_degree, row);
        Pw.Set(num_u_seg*m_u_degree, row, pt);
    }

    // now decompose in v direction
    alphas.resize(m_v_degree + 1);
    nb = 0;
    a = m_v_degree;
    b = m_v_degree + 1;
    n = m_u_degree*num_u_seg;
    Qw.resize(m, n+1, m_v_degree + 1);
    max = m + 2;

    for (i = 0;i <= m_v_degree;i++)
    {
        for (gfInt col = 0;col <= n;col++)
        {
            gfPoint4 pt = Pw.at(col, i);
            Qw.Set(nb, col, i, gfPoint4(pt.X(), pt.Y(), pt.Z(), pt.W()));
        }
    }

    while (b < max)
    {
        i = b;
        while (b < (max) && m_v_knot_vector.at(b+1) == m_v_knot_vector.at(b))
            b++;
        gfInt mult = b-i+1;
        if (mult < m_v_degree)
        {
            gfDouble numer = m_v_knot_vector.at(b) - m_v_knot_vector.at(a);
            for (j = m_v_degree;j > mult;j--)
                alphas[j-mult-1] = numer/(m_v_knot_vector.at(a+j) - m_v_knot_vector.at(a));
            gfInt r = m_v_degree - mult;
            for (j = 1;j <= r;j++)
            {
                gfInt save = r-j;
                gfInt s = mult+j;
                for  (k = m_v_degree;k >= s;k--)
                {
                    gfDouble alpha = alphas.at(k-s);
                    for (gfInt col = 0;col <= n;col++)
                        Qw.Set(nb, col, k, gfPoint4(alpha*Qw.at(nb, col, k).X()+(1.0-alpha)*Qw.at(nb, col, k-1).X(),
                        alpha*Qw.at(nb, col, k).Y()+(1.0-alpha)*Qw.at(nb, col, k-1).Y(),
                        alpha*Qw.at(nb, col, k).Z()+(1.0-alpha)*Qw.at(nb, col, k-1).Z(),
                        alpha*Qw.at(nb, col, k).W()+(1.0-alpha)*Qw.at(nb, col, k-1).W()));
                }

                if (b < (max))
                {
                    for (gfInt col = 0;col <= n;col++)
                        Qw.Set(nb + 1, col, save, gfPoint4(Qw.at(nb, col, m_v_degree).X(), 
                        Qw.at(nb, col, m_v_degree).Y(), 
                        Qw.at(nb, col, m_v_degree).Z(), 
                        Qw.at(nb, col, m_v_degree).W()));
                }
            }
        }

        nb++;
        if (b < max) 
        {
            for (i = m_v_degree-mult;i <= m_v_degree;i++)
            {
                for (gfInt col = 0;col <= n;col++)
                {
                    gfPoint4 pt = Pw.at(col, b-m_v_degree+i);
                    Qw.Set(nb, col, i, gfPoint4(pt.X(), pt.Y(), pt.Z(), pt.W()));
                }
            }
            a = b;
            b++;
        }
    }

    // get all different u and v knots
    std::vector<gfDouble> dif_u_knots;
    for (i = m_u_degree;i <= GetNumUControlPoints()+1;i++)
    {
        gfInt num = (gfInt)dif_u_knots.size();
        gfBool diff = true;
        for (gfInt j = 0;j < num;j++)
        {
            if (m_u_knot_vector.at(i) == dif_u_knots.at(j))
                diff = false;
        }
        if (diff)
            dif_u_knots.push_back(m_u_knot_vector.at(i));
    }
    std::vector<gfDouble> dif_v_knots;
    for (i = m_v_degree;i <= GetNumVControlPoints()+1;i++)
    {
        gfInt num = (gfInt)dif_v_knots.size();
        gfBool diff = true;
        for (gfInt j = 0;j < num;j++)
        {
            if (m_v_knot_vector.at(i) == dif_v_knots.at(j))
                diff = false;
        }
        if (diff)
            dif_v_knots.push_back(m_v_knot_vector.at(i));
    }

    // fill array with Bezier segments
    gfInt num_v_seg = nb;
    segments->resize(num_u_seg, num_v_seg);
    gfInt num_trims = GetNumTrimCurves();
    for (i = 0;i < num_u_seg;i++)
    {
        for (j = 0;j < num_v_seg;j++)
        {
            // create NURBS curve as Bezier curve
            gfObjGeomObject* obj = gfCreateGeomObject(GF_NURBS_PATCH);
            gfObjNURBSPatch* patch = (gfObjNURBSPatch*) obj;
            patch->InitBezier(m_u_degree + 1, m_v_degree + 1, 
                dif_u_knots.at(i), dif_u_knots.at(i+1), dif_v_knots.at(j), dif_v_knots.at(j+1));

            for (gfInt ii = 0;ii <= m_u_degree;ii++)
            {
                for (gfInt jj = 0;jj <= m_v_degree;jj++)
                {
                    gfPoint3* pt = patch->GetControlPoint(ii, jj);
                    pt->Set(Qw.at(j, ii+i*m_u_degree, jj).X(), 
                        Qw.at(j, ii+i*m_u_degree, jj).Y(), 
                        Qw.at(j, ii+i*m_u_degree, jj).Z());

                    patch->SetWeight(ii, jj, Qw.at(j, ii+i*m_u_degree, jj).W());
                }
            }

            for (gfInt trim = 0;trim < num_trims;trim++)
                patch->AddTrimCurve((gfObjNURBSCurve*)GetTrimCurve(trim)->Clone());
            segments->Set(i, j, patch);
        }
    }

    gfLog("NURBS decomposed to (%d x %d) Bezier segments.\n", num_u_seg, num_v_seg);

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::DecomposeUDeCasteljau(gfDouble u, gfObjNURBSPatch* patch1, gfObjNURBSPatch* patch2)
{
    gfDouble u_min = GetUKnot(m_u_degree);
    gfDouble u_max = GetUKnot(GetNumUControlPoints());
    gfDouble v_min = GetVKnot(m_v_degree);
    gfDouble v_max = GetVKnot(GetNumVControlPoints());

    if (u < u_min || u > u_max || !IsBezier() || gfIsZero(u_max - u_min))
        return GF_E_FAIL;

    // initialize curves
    patch1->InitBezier(m_u_degree + 1, m_v_degree + 1, u_min, u, v_min, v_max);
    patch2->InitBezier(m_u_degree + 1, m_v_degree + 1, u, u_max, v_min, v_max);

    gfDouble true_u = (u-u_min) / (u_max-u_min);

    for (gfInt row = 0;row <= m_v_degree;row++)
    {
        gfAux2DArray<gfPoint4> aux;
        aux.resize(m_u_degree + 1, m_u_degree + 1);
        for (gfInt i = 0;i <= m_u_degree;i++)
        {
            gfPoint3* pt = GetControlPoint(i, row);
            aux.Set(0, i, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(i, row)));
        }

        // decompose using deCasteljau algorithm
        for (gfInt i = 1;i <= m_u_degree;i++)
        {
            for (gfInt j = 0;j < m_u_degree - i + 1;j++)
            {
                gfDouble ax = (1.0-true_u)*aux.at(i-1, j).W()*aux.at(i-1, j).X()+true_u*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).X();
                gfDouble ay = (1.0-true_u)*aux.at(i-1, j).W()*aux.at(i-1, j).Y()+true_u*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Y();
                gfDouble az = (1.0-true_u)*aux.at(i-1, j).W()*aux.at(i-1, j).Z()+true_u*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Z();
                gfDouble aw = (1.0-true_u)*aux.at(i-1, j).W()+true_u*aux.at(i-1, j+1).W();
                aux.Set(i, j, gfPoint4(ax/aw, ay/aw, az/aw, aw));
            }
        }

        // fill curves
        for (gfInt i = 0;i <= m_u_degree;i++)
        {
            gfPoint3* pt1 = patch1->GetControlPoint(i, row);
            pt1->Set(aux.at(i, 0).X(), aux.at(i, 0).Y(), aux.at(i, 0).Z());
            patch1->SetWeight(i, row, aux.at(i, 0).W());

            gfPoint3* pt2 = patch2->GetControlPoint(i, row);
            pt2->Set(aux.at(m_u_degree-i, i).X(), aux.at(m_u_degree-i, i).Y(), aux.at(m_u_degree-i, i).Z());
            patch2->SetWeight(i, row, aux.at(m_u_degree-i, i).W());
        }
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::DecomposeVDeCasteljau(gfDouble v, gfObjNURBSPatch* patch1, gfObjNURBSPatch* patch2)
{
    gfDouble u_min = GetUKnot(m_u_degree);
    gfDouble u_max = GetUKnot(GetNumUControlPoints());
    gfDouble v_min = GetVKnot(m_v_degree);
    gfDouble v_max = GetVKnot(GetNumVControlPoints());

    if (v < v_min || v > v_max || !IsBezier() || gfIsZero(v_max - v_min))
        return GF_E_FAIL;

    // initialize curves
    patch1->InitBezier(m_u_degree + 1, m_v_degree + 1, u_min, u_max, v_min, v);
    patch2->InitBezier(m_u_degree + 1, m_v_degree + 1, u_min, u_max, v, v_max);

    gfDouble true_v = (v-v_min) / (v_max-v_min);

    for (gfInt col = 0;col <= m_u_degree;col++)
    {
        gfAux2DArray<gfPoint4> aux;
        aux.resize(m_v_degree + 1, m_v_degree + 1);
        for (gfInt i = 0;i <= m_v_degree;i++)
        {
            gfPoint3* pt = GetControlPoint(col, i);
            aux.Set(0, i, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(col, i)));
        }

        // decompose using deCasteljau algorithm
        for (gfInt i = 1;i <= m_v_degree;i++)
        {
            for (gfInt j = 0;j < m_v_degree - i + 1;j++)
            {
                gfDouble ax = (1.0-true_v)*aux.at(i-1, j).W()*aux.at(i-1, j).X()+true_v*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).X();
                gfDouble ay = (1.0-true_v)*aux.at(i-1, j).W()*aux.at(i-1, j).Y()+true_v*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Y();
                gfDouble az = (1.0-true_v)*aux.at(i-1, j).W()*aux.at(i-1, j).Z()+true_v*aux.at(i-1, j+1).W()*aux.at(i-1, j+1).Z();
                gfDouble aw = (1.0-true_v)*aux.at(i-1, j).W()+true_v*aux.at(i-1, j+1).W();
                aux.Set(i, j, gfPoint4(ax/aw, ay/aw, az/aw, aw));
            }
        }

        // fill curves
        for (gfInt i = 0;i <= m_v_degree;i++)
        {
            gfPoint3* pt1 = patch1->GetControlPoint(col, i);
            pt1->Set(aux.at(i, 0).X(), aux.at(i, 0).Y(), aux.at(i, 0).Z());
            patch1->SetWeight(col, i, aux.at(i, 0).W());

            gfPoint3* pt2 = patch2->GetControlPoint(col, i);
            pt2->Set(aux.at(m_v_degree-i, i).X(), aux.at(m_v_degree-i, i).Y(), aux.at(m_v_degree-i, i).Z());
            patch2->SetWeight(col, i, aux.at(m_v_degree-i, i).W());
        }
    }

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::AddTrimCurve(gfObjNURBSCurve* curve)
{
    m_trim_curves.push_back(curve);
    return GF_S_OK;
}

gfObjNURBSCurve* gfObjNURBSPatch::GetTrimCurve(gfInt index)
{
    gfInt count = (gfInt)m_trim_curves.size();
    if (index < 0 || index >= count)
        return NULL;

    return m_trim_curves.at(index);
}

gfInt gfObjNURBSPatch::GetNumTrimCurves(void)
{
    return (gfInt)m_trim_curves.size();
}

GFRESULT gfObjNURBSPatch::RemoveTrimCurve(gfInt index)
{
    gfInt count = (gfInt)m_trim_curves.size();
    if (index < 0 || index >= count)
        return GF_E_BADARG;

    gfObjNURBSCurve* curve = m_trim_curves.at(index);
    gfDeleteGeomObject(curve);
    m_trim_curves.erase(m_trim_curves.begin() + index);

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::RemoveAllTrimCurves(void)
{
    gfInt count = (gfInt)m_trim_curves.size();
    for (gfInt i = 0;i < count;i++)
    {
        gfObjNURBSCurve* curve = m_trim_curves.at(i);
        gfDeleteGeomObject(curve);
    }

    m_trim_curves.clear();

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::PointInsideDomain(gfDouble u, gfDouble v)
{
    gfDouble umin, umax, vmin, vmax;
    GetDomain(&umin, &umax, &vmin, &vmax);
    if (u < umin || v < vmin || u > umax || v > vmax)
        return GF_E_FAIL;

    gfDouble a, b, c;
    //a = gfRand(1.0);
    a = 0.5444;
    b = gfSquareRoot(1.0 - a*a);
    c = -a*u-b*v;

    gfInt wind = 0;
    gfVector3 dir(b, -a, 0.0);

    gfInt count = (gfInt)m_trim_curves.size();
    for (gfInt i = 0;i < count;i++)
    {
        gfObjNURBSCurve* curve = m_trim_curves.at(i);
        gfRayInterInfo* info = curve->LineIntersection2D(a, b, c);

        for (gfInt j = 0;j < (gfInt)info->m_inters.size();j++)
        {
            // get point of intersection
            gfPoint3* pt = curve->GetCurvePoint(info->m_inters.at(j));
            gfVector3 auxvec;
            gfPoint3 auxpt(u, v, 0.0);
            gfSubstractPoints3(pt, &auxpt, &auxvec);
            gfDeletePoint3(pt);

            gfDouble t = gfScalarProduct3(&dir, &auxvec);
            if (t > 0.0)
                wind++;
        }

        gfDeleteRayInterInfo(info);
    }

    if (wind % 2 == 0)
        return GF_S_OK;
    else
        return GF_E_FAIL;
}

gfUVPoint* gfObjNURBSPatch::TesselateTrimmedUVSpace(gfInt approx)
{
    // domain rectangle
    gfDouble umin = GetUKnot(GetUDegree());
    gfDouble umax = GetUKnot(GetNumUControlPoints());
    gfDouble vmin = GetVKnot(GetVDegree());
    gfDouble vmax = GetVKnot(GetNumVControlPoints());

    // prepare resulting array
    gfUVPoint* result = new gfUVPoint[approx * approx];
    memset(result, 0, approx * approx * sizeof(gfUVPoint));
    gfInt num_trim_curves = GetNumTrimCurves();
    std::vector<gfObjPolyLine*> approx_polylines;
    gfInt curve_approx = 4 * approx;

    gfInt num_inter = 0;
    // go throught all trim curves and prepare winding numbers
    for (gfInt i = 0;i < num_trim_curves;i++)
    {
        // get approximation of trim curve
        gfObjPolyLine* polyline = GetTrimCurve(i)->GetApproxPolyline(curve_approx);
        approx_polylines.push_back(polyline);

        for (gfInt j = 0;j < curve_approx;j++)
        {
            // get one line of polyline
            gfPoint3 * pt1 = polyline->GetPoint(j);
            gfPoint3 * pt2 = polyline->GetPoint((j + 1) % curve_approx);

            // find all intersections of line with v-isocurves of UV space
            gfInt pt1intx = (gfInt) ((approx - 1) * (pt1->X() - umin) / (umax - umin));
            gfInt pt2intx = (gfInt) ((approx - 1) * (pt2->X() - umin) / (umax - umin));
            gfInt ptmin;
            gfInt ptmax;
            if (pt1intx <= pt2intx)
            {
                ptmin = pt1intx + 1;
                ptmax = pt2intx;
            }
            else
            {
                ptmin = pt2intx + 1;
                ptmax = pt1intx;
            }
            for (gfInt k = ptmin;k <= ptmax;k++)
            {
                gfDouble u0 = umin + k * (umax - umin) / (approx - 1);
                gfDouble v0 = pt1->Y() + (pt2->Y() - pt1->Y()) * (u0 - pt1->X()) / (pt2->X() - pt1->X());

                gfInt v0int = (gfInt) ((approx - 1) * (v0 - vmin) / (vmax - vmin));
                for (gfInt l = v0int;l < approx;l++)
                {
                    // increase winding number
                    if (l >= 0 && k >= 0 && k < approx)
                        result[k + l * approx].winding_number++;
                }
            }

            // set points around line to be closer to that line
            gfInt pt1inty = (gfInt) ((approx - 1) * (pt1->Y() - vmin) / (vmax - vmin));
            gfInt pt2inty = (gfInt) ((approx - 1) * (pt2->Y() - vmin) / (vmax - vmin));

            gfInt lmin,lmax, kmin, kmax;
            if (pt1intx <= pt2intx)
            {
                kmin = pt1intx - 1;kmax = pt2intx + 1;
            }
            else
            {
                kmin = pt2intx - 1;kmax = pt1intx + 1;
            }
            if (pt1inty <= pt2inty)
            {
                lmin = pt1inty - 1;lmax = pt2inty + 1;
            }
            else
            {
                lmin = pt2inty - 1;lmax = pt1inty + 1;
            }
            for (gfInt k = kmin;k <= kmax;k++)
                for (gfInt l = lmin;l <= lmax;l++)
                {
                    if (k >= 0 && k < approx && l >= 0 && l < approx)
                        result[k + l * approx].boundary = i + 1;
                }
        }
    }

    // go throught resulting points and find points near trim curves
    for (gfInt k = 0;k < approx;k++)
        for (gfInt l = 0;l < approx;l++)
        {
            // check if current point has 8-neighbour with different winding number
            gfInt winding = result[k + l * approx].winding_number;
            gfInt boundary = result[k + l * approx].boundary;

            // change point near the curve to point on that curve
            if (boundary > 0 && boundary <= num_trim_curves)
            {
                gfDouble u = umin + k * (umax - umin) / (approx - 1);
                gfDouble v = vmin + l * (vmax - vmin) / (approx - 1);
                gfObjPolyLine* polyline = approx_polylines.at(boundary - 1);
                gfInt min_index;
                gfDouble min_value = GF_MAX_DOUBLE;

                for (gfInt i = 0;i < curve_approx;i++)
                {
                    gfDouble upt = polyline->GetPoint(i)->X();
                    gfDouble vpt = polyline->GetPoint(i)->Y();
                    gfDouble dist = (u - upt)*(u - upt)+(v - vpt)*(v - vpt);
                    if (dist < min_value)
                    {
                        min_value = dist;
                        min_index = i;
                    }
                }

                // minimal distance found
                // check if new point is inside domain
                gfDouble nx = polyline->GetPoint(min_index)->X();
                gfDouble ny = polyline->GetPoint(min_index)->Y();
                if (umin > nx)
                {
                    //ny = ny + (v-ny)*(umin-nx)/(u-nx);
                    nx = umin;
                }
                if (umax < nx)
                {
                    //ny = ny + (v-ny)*(umax-nx)/(u-nx);
                    nx = umax;
                }
                if (vmin > ny)
                {
                    //nx = nx + (u-nx)*(vmin-ny)/(v-ny);
                    ny = vmin;
                }
                if (vmax < ny)
                {
                    //nx = nx + (u-nx)*(vmax-ny)/(v-ny);
                    ny = vmax;
                }

                if (!gfIsZero(u - umin) && !gfIsZero(u - umax))
                    result[k + l * approx].du = nx - u;
                else
                    result[k + l * approx].du= 0.0;
                if (!gfIsZero(v - vmin) && !gfIsZero(v - vmax))
                    result[k + l * approx].dv = ny - v;
                else
                    result[k + l * approx].dv = 0.0;
            }
        }

        // clear
        for (gfInt i = 0;i < num_trim_curves;i++)
        {
            // get approximation of trim curve
            gfObjPolyLine* polyline = approx_polylines.at(i);
            gfDeleteGeomObject(polyline);
        }
        approx_polylines.clear();

        return result;
}

gfAuxBoundingVolume* gfObjNURBSPatch::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    // create auxiliary array of mesh points
    gfInt num_points = (gfInt)m_control_points.size();
    if (num_points < 1)
        return NULL;

    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfInt i = 0; i < num_points; i++)
    {
        gfPoint3* pt = m_control_points.at(i);
        curr_point->Set(pt->X(), pt->Y(), pt->Z());
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    delete [] points;
    return result;
}

GFRESULT gfObjNURBSPatch::InsertUKnot(gfInt mult, gfDouble u)
{
    gfInt np = GetNumUControlPoints() - 1;
    gfInt mp = GetNumVControlPoints() - 1;
    gfInt k = gfGetBSplineKnotSpan(np, m_u_degree, &m_u_knot_vector, u);
    gfInt s = gfGetBSplineKnotMultiplicity(np, m_u_degree, &m_u_knot_vector, u);
    gfInt new_np = np+mult;
    gfInt new_mp = mp;
    gfInt old_u_knot_size = GetNumUKnots();
    gfInt new_u_knot_size = old_u_knot_size + mult;
    gfInt old_v_knot_size = GetNumVKnots();
    gfInt new_v_knot_size = old_v_knot_size;

    gfInt i,j;
    gfAux2DArray<gfPoint4> Qw;
    Qw.resize(new_np + 1, new_mp + 1);
    gfPoint4* Rw = new gfPoint4[m_u_degree + 1];

    // set new u - knot vector
    std::vector<gfDouble> new_u_knot_vector;
    new_u_knot_vector.resize(new_u_knot_size);
    for (i = 0;i <= k;i++)
        new_u_knot_vector[i] = m_u_knot_vector.at(i);
    for (i = 1;i <= mult;i++)
        new_u_knot_vector[k+i] = u;
    for (i = k+1;i < new_u_knot_size;i++)
        new_u_knot_vector[i+mult] = m_u_knot_vector.at(i);

    // copy old v - knot vector
    std::vector<gfDouble> new_v_knot_vector;
    new_v_knot_vector.resize(new_v_knot_size);
    for (i = 0;i < new_v_knot_size;i++)
        new_v_knot_vector[i] = m_v_knot_vector.at(i);

    // save the alphas
    gfInt L;
    gfAux2DArray<gfDouble> alpha;
    alpha.resize(m_u_degree + 1, mult + 1);
    for (j = 1;j <= mult;j++)
    {
        L = k-m_u_degree+j;
        for (i = 0;i <= m_u_degree-j-s;i++)
        {
            alpha.Set(i, j, (u-m_u_knot_vector.at(L+i))/(m_u_knot_vector.at(i+k+1)-m_u_knot_vector.at(L+i)));
        }
    }

    for (gfInt row = 0;row <= mp;row++)
    {
        // fill new control points
        for (i = 0;i <= k-m_u_degree;i++)
        {
            gfPoint3* pt = GetControlPoint(i, row);
            Qw.Set(i, row, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(i, row)));
        }
        for (i = k-s;i <= np;i++)
        {
            gfPoint3* pt = GetControlPoint(i, row);
            Qw.Set(i+mult, row, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(i, row)));
        }
        for (i = 0;i <= m_u_degree-s;i++)
        {
            gfPoint3* pt = GetControlPoint(k-m_u_degree+i, row);
            Rw[i].Set(pt->X(), pt->Y(), pt->Z(), GetWeight(k-m_u_degree+i, row));
        }

        // insert the knot
        gfInt L;
        for (j = 1;j <= mult;j++)
        {
            L = k-m_u_degree+j;
            for (i = 0;i <= m_u_degree-j-s;i++)
            {
                //gfDouble alpha = (u-m_u_knot_vector.at(L+i))/(m_u_knot_vector.at(i+k+1)-m_u_knot_vector.at(L+i));
                Rw[i].Set(alpha.at(i, j)*Rw[i+1].X()+(1.0-alpha.at(i, j))*Rw[i].X(),
                    alpha.at(i, j)*Rw[i+1].Y()+(1.0-alpha.at(i, j))*Rw[i].Y(),
                    alpha.at(i, j)*Rw[i+1].Z()+(1.0-alpha.at(i, j))*Rw[i].Z(),
                    alpha.at(i, j)*Rw[i+1].W()+(1.0-alpha.at(i, j))*Rw[i].W());
            }

            Qw.Set(L, row, gfPoint4(Rw[0].X(), Rw[0].Y(), Rw[0].Z(), Rw[0].W()));
            Qw.Set(k+mult-j-s, row, gfPoint4(Rw[m_u_degree-j-s].X(), Rw[m_u_degree-j-s].Y(), Rw[m_u_degree-j-s].Z(), Rw[m_u_degree-j-s].W()));
        }

        for (i = L+1;i < k-s;i++)
        {
            Qw.Set(i, row, gfPoint4(Rw[i-L].X(), Rw[i-L].Y(), Rw[i-L].Z(), Rw[i-L].W()));
        }
    }

    // copy newly created arrays
    SetParams(m_u_degree, m_v_degree, new_np + 1, new_mp + 1);
    for (i = 0;i < new_u_knot_size;i++)
        SetUKnot(i, new_u_knot_vector.at(i));
    for (i = 0;i < new_v_knot_size;i++)
        SetVKnot(i, new_v_knot_vector.at(i));
    for (i = 0;i <= new_np;i++)
        for (j = 0;j <= new_mp;j++)
        {
            GetControlPoint(i, j)->Set(Qw.at(i, j).X(), Qw.at(i, j).Y(), Qw.at(i, j).Z());
            SetWeight(i, j, Qw.at(i, j).W());
        }

        delete [] Rw;
        return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::InsertVKnot(gfInt mult, gfDouble v)
{
    gfInt np = GetNumUControlPoints() - 1;
    gfInt mp = GetNumVControlPoints() - 1;
    gfInt k = gfGetBSplineKnotSpan(mp, m_v_degree, &m_v_knot_vector, v);
    gfInt s = gfGetBSplineKnotMultiplicity(mp, m_v_degree, &m_v_knot_vector, v);
    gfInt new_np = np;
    gfInt new_mp = mp+mult;
    gfInt old_u_knot_size = GetNumUKnots();
    gfInt new_u_knot_size = old_u_knot_size;
    gfInt old_v_knot_size = GetNumVKnots();
    gfInt new_v_knot_size = old_v_knot_size + mult;

    gfInt i,j;
    gfAux2DArray<gfPoint4> Qw;
    Qw.resize(new_np + 1, new_mp + 1);
    gfPoint4* Rw = new gfPoint4[m_v_degree + 1];

    // set new v - knot vector
    std::vector<gfDouble> new_v_knot_vector;
    new_v_knot_vector.resize(new_v_knot_size);
    for (i = 0;i <= k;i++)
        new_v_knot_vector[i] = m_v_knot_vector.at(i);
    for (i = 1;i <= mult;i++)
        new_v_knot_vector[k+i] = v;
    for (i = k+1;i < new_v_knot_size;i++)
        new_v_knot_vector[i+mult] = m_v_knot_vector.at(i);

    // copy old u - knot vector
    std::vector<gfDouble> new_u_knot_vector;
    new_u_knot_vector.resize(new_u_knot_size);
    for (i = 0;i < new_u_knot_size;i++)
        new_u_knot_vector[i] = m_u_knot_vector.at(i);

    // save the alphas
    gfInt L;
    gfAux2DArray<gfDouble> alpha;
    alpha.resize(m_v_degree + 1, mult + 1);
    for (j = 1;j <= mult;j++)
    {
        L = k-m_v_degree+j;
        for (i = 0;i <= m_v_degree-j-s;i++)
        {
            alpha.Set(i, j, (v-m_v_knot_vector.at(L+i))/(m_v_knot_vector.at(i+k+1)-m_v_knot_vector.at(L+i)));
        }
    }

    for (gfInt col = 0;col <= mp;col++)
    {
        // fill new control points
        for (i = 0;i <= k-m_v_degree;i++)
        {
            gfPoint3* pt = GetControlPoint(col, i);
            Qw.Set(col, i, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(col, i)));
        }
        for (i = k-s;i <= np;i++)
        {
            gfPoint3* pt = GetControlPoint(col, i);
            Qw.Set(col, i+mult, gfPoint4(pt->X(), pt->Y(), pt->Z(), GetWeight(col, i)));
        }
        for (i = 0;i <= m_v_degree-s;i++)
        {
            gfPoint3* pt = GetControlPoint(col, k-m_v_degree+i);
            Rw[i].Set(pt->X(), pt->Y(), pt->Z(), GetWeight(col, k-m_v_degree+i));
        }

        // insert the knot
        gfInt L;
        for (j = 1;j <= mult;j++)
        {
            L = k-m_v_degree+j;
            for (i = 0;i <= m_v_degree-j-s;i++)
            {
                //gfDouble alpha = (u-m_u_knot_vector.at(L+i))/(m_u_knot_vector.at(i+k+1)-m_u_knot_vector.at(L+i));
                Rw[i].Set(alpha.at(i, j)*Rw[i+1].X()+(1.0-alpha.at(i, j))*Rw[i].X(),
                    alpha.at(i, j)*Rw[i+1].Y()+(1.0-alpha.at(i, j))*Rw[i].Y(),
                    alpha.at(i, j)*Rw[i+1].Z()+(1.0-alpha.at(i, j))*Rw[i].Z(),
                    alpha.at(i, j)*Rw[i+1].W()+(1.0-alpha.at(i, j))*Rw[i].W());
            }

            Qw.Set(col, L, gfPoint4(Rw[0].X(), Rw[0].Y(), Rw[0].Z(), Rw[0].W()));
            Qw.Set(col, k+mult-j-s, gfPoint4(Rw[m_v_degree-j-s].X(), Rw[m_v_degree-j-s].Y(), Rw[m_v_degree-j-s].Z(), Rw[m_v_degree-j-s].W()));
        }

        for (i = L+1;i < k-s;i++)
        {
            Qw.Set(col, i, gfPoint4(Rw[i-L].X(), Rw[i-L].Y(), Rw[i-L].Z(), Rw[i-L].W()));
        }
    }

    // copy newly created arrays
    SetParams(m_u_degree, m_v_degree, new_np + 1, new_mp + 1);
    for (i = 0;i < new_u_knot_size;i++)
        SetUKnot(i, new_u_knot_vector.at(i));
    for (i = 0;i < new_v_knot_size;i++)
        SetVKnot(i, new_v_knot_vector.at(i));
    for (i = 0;i <= new_np;i++)
        for (j = 0;j <= new_mp;j++)
        {
            GetControlPoint(i, j)->Set(Qw.at(i, j).X(), Qw.at(i, j).Y(), Qw.at(i, j).Z());
            SetWeight(i, j, Qw.at(i, j).W());
        }

        delete [] Rw;
        return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::DeleteAABBTree(gfBezier2AABBTreeNode* tree_root)
{
    if (tree_root->aabb)
        gfDeleteBoundVolume(tree_root->aabb);
    if (tree_root->left)
        DeleteAABBTree(tree_root->left);
    if (tree_root->right)
        DeleteAABBTree(tree_root->right);
    if (tree_root->patch)
        gfDeleteGeomObject(tree_root->patch);
    if (tree_root->control_mesh)
        gfDeleteGeomObject(tree_root->control_mesh);

    delete tree_root;
    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::DeleteWholeAABBTree(void)
{
    gfInt num_trees = (gfInt)m_bezier_aabb_trees.size();
    for (gfInt i = 0;i < num_trees;i++)
    {
        gfBezier2AABBTreeNode* root = m_bezier_aabb_trees.at(i);
        DeleteAABBTree(root);
    }
    m_bezier_aabb_trees.clear();

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::ComputeWholeAABBTree(void)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    DeleteWholeAABBTree();

    m_num_nodes = 0;

    // decompose to Bezier patches
    gfAux2DArray<gfObjNURBSPatch*> beziers;
    DecomposeToBezierSegments(&beziers);

    gfInt num_i_trees = beziers.GetISize();
    gfInt num_j_trees = beziers.GetJSize();
    for (gfInt i = 0;i < num_i_trees;i++)
        for (gfInt j = 0;j < num_j_trees;j++)
        {
            gfObjNURBSPatch* bezier = beziers.at(i, j);
            gfBezier2AABBTreeNode* root = CreateAABBTree(0, bezier);
            m_bezier_aabb_trees.push_back(root);
        }

    gfLog("Computed AABB tree of Bezier patches for NURBS patch, nodes = %d, time = %d ms.\n", m_num_nodes, count.EndMSCount());

    return GF_S_OK;
}

gfBezier2AABBTreeNode* gfObjNURBSPatch::CreateAABBTree(gfDouble depth, gfObjNURBSPatch* patch)
{
    gfBezier2AABBTreeNode* root = new gfBezier2AABBTreeNode;
    root->left = root->right = NULL;
    root->aabb = (gfAuxAABB*) patch->GetBoundVolume(GF_BOUND_VOL_AABB, NULL);
    root->patch = NULL;
    root->control_mesh = NULL;
    root->min_dir = 0;
    m_num_nodes++;

    // compute also OBB
    gfAuxOBB* obb = (gfAuxOBB*) patch->GetBoundVolume(GF_BOUND_VOL_OBB, NULL);

    // check if AABB needs to be splitted
    gfDouble x_dir = 2.0*gfSquareRoot(obb->m_axis1_x*obb->m_axis1_x+
        obb->m_axis1_y*obb->m_axis1_y+
        obb->m_axis1_z*obb->m_axis1_z);
    gfDouble y_dir = 2.0*gfSquareRoot(obb->m_axis2_x*obb->m_axis2_x+
        obb->m_axis2_y*obb->m_axis2_y+
        obb->m_axis2_z*obb->m_axis2_z);
    gfDouble z_dir = 2.0*gfSquareRoot(obb->m_axis3_x*obb->m_axis3_x+
        obb->m_axis3_y*obb->m_axis3_y+
        obb->m_axis3_z*obb->m_axis3_z);

    // find shortest dimension
    gfDouble min;
    gfDouble space;
    if (x_dir < y_dir)
    {
        if (x_dir < z_dir)
        {
            min = x_dir;
            root->min_dir = 1;
            space = y_dir*z_dir;
        }
        else
        {
            min = z_dir;
            root->min_dir = 3;
            space = y_dir*x_dir;
        }
    }
    else
    {
        if (y_dir < z_dir)
        {
            min = y_dir;
            root->min_dir = 2;
            space = x_dir*z_dir;
        }
        else
        {
            min = z_dir;
            root->min_dir = 3;
            space = y_dir*x_dir;
        }
    }

    // delete OBB
    gfDeleteBoundVolume(obb);

    // check if OBB is flat enough
    if ((min < 0.01 && space < 0.001) || depth > GF_MAX_BEZIER_AABB_TREE_DEPTH)
    {
        root->patch = patch;
        //root->control_mesh = patch->GetControlMesh();
        //root->control_mesh->ComputeSmoothNormals();
    }
    else
    {
        // split patch
        gfDouble u_min, u_max, v_min, v_max;
        patch->GetDomain(&u_min, &u_max, &v_min, &v_max);

        gfObjNURBSPatch* patch1 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
        gfObjNURBSPatch* patch2 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
        if ((u_max-u_min) > (v_max-v_min))
            patch->DecomposeUDeCasteljau((u_max+u_min)/2.0, patch1, patch2);
        else
            patch->DecomposeVDeCasteljau((v_max+v_min)/2.0, patch1, patch2);

        gfDeleteGeomObject(patch);

        root->left = CreateAABBTree(depth+1, patch1);
        root->right = CreateAABBTree(depth+1, patch2);
    }

    return root;
}

gfRayInterInfo* gfObjNURBSPatch::RayAABBIntersection(gfBezier2AABBTreeNode* root, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    if (!root)
        return NULL;
    gfDouble tmin, tmax;
    if (GF_FAIL(root->aabb->RayIntersection(ray_begin, ray_dir, &tmin, &tmax)))
        return NULL;

    gfRayInterInfo* info = NULL;
    if (root->left && root->right)
    {
        // get intersection from childs
        gfRayInterInfo* info1 = RayAABBIntersection(root->left, ray_begin, ray_dir);
        gfRayInterInfo* info2 = RayAABBIntersection(root->right, ray_begin, ray_dir);
        if (info1 && info2)
        {
            info = info1;
            for (gfInt i = 0;i < (gfInt)info2->m_inters.size();i++)
            {
                info->m_inters.push_back(info2->m_inters.at(i));
                info->m_normals.push_back(info2->m_normals.at(i));
                info->m_uvw.push_back(info2->m_uvw.at(i));
            }
            gfDeleteRayInterInfo(info2);
            /*if (info1->m_inters.at(0) < info2->m_inters.at(0))
            {
                info = info1;
                for (gfInt i = 0;i < info2->m_inters.size();i++)
                    info->m_inters.push_back(info2->m_inters.at(i));
                gfDeleteRayInterInfo(info2);
            }
            else
            {
                info = info2;
                for (gfInt i = 0;i < info1->m_inters.size();i++)
                    info->m_inters.push_back(info1->m_inters.at(i));
                gfDeleteRayInterInfo(info1);
            }*/
        }
        else if (info1)
        {
            info = info1;
        }
        else if (info2)
        {
            info = info2;
        }
    }
    else
    {
        if (!root->patch || !root->aabb)
            return NULL;

        if (!m_bezier_clipping)
        {
            // return intersection of ray and control mesh

            //info = root->control_mesh->RayIntersection(NULL, ray_begin, ray_dir);
            gfObjHalfEdgeMesh* control_mesh = root->patch->GetControlMesh();
            control_mesh->ComputeSmoothNormals();
            info = control_mesh->RayIntersection(NULL, ray_begin, ray_dir);
            gfDeleteGeomObject(control_mesh);

            if (info)
            {
                // add dummy uvw coordinates
                for (gfInt i = 0;i < (gfInt)info->m_inters.size();i++)
                    info->m_uvw.push_back(gfPoint3(0.0, 0.0, 0.0));
            }
        }
        else
        {
            // use bezier clipping
            gfVector3 up;
            if (gfAbs(ray_dir->X()) > gfAbs(ray_dir->Y()) && gfAbs(ray_dir->X()) > gfAbs(ray_dir->Z()))
                up.Set(ray_dir->Y(), -ray_dir->X(), 0.0);
            else
                up.Set(0.0, ray_dir->Z(), -ray_dir->Y());
            gfVector3 right;
            gfVectorProduct3(ray_dir, &up, &right);
            right.Normalize();
            up.Normalize();
            gfDouble d1 = -(up.X()*ray_begin->X()+up.Y()*ray_begin->Y()+up.Z()*ray_begin->Z());
            gfDouble d2 = -(right.X()*ray_begin->X()+right.Y()*ray_begin->Y()+right.Z()*ray_begin->Z());

            gfRayBezierPatchInfo* res = root->patch->BezierClipping(NULL, 0, up.X(), up.Y(), up.Z(), d1, right.X(), right.Y(), right.Z(), d2);
            if (res)
            {
                gfInt size = (gfInt)res->m_inters.size();
                if (size > 0)
                    info = gfCreateRayInterInfo();
                for (gfInt i = 0;i < size;i++)
                {
                    gfPoint3* pt = root->patch->GetPatchPoint(res->m_inters.at(i).X(), res->m_inters.at(i).Y());
                    gfVector3* nm = root->patch->GetPatchNormal(res->m_inters.at(i).X(), res->m_inters.at(i).Y());
                    gfVector3 vec(pt->X()-ray_begin->X(), pt->Y()-ray_begin->Y(), pt->Z()-ray_begin->Z());

                    info->m_inters.push_back(gfScalarProduct3(ray_dir, &vec));
                    info->m_normals.push_back(gfVector3(nm->X(), nm->Y(), nm->Z()));
                    info->m_uvw.push_back(gfPoint3(res->m_inters.at(i).X(), res->m_inters.at(i).Y(), 0.0));

                    gfDeletePoint3(pt);
                    gfDeleteVector3(nm);
                }

                delete res;
            }
        }
    }

    // sort intersections
    if (info)
    {
        gfSortRayInterInfo(info);
    }

    return info;
}

GFRESULT gfObjNURBSPatch::CreateVACache(gfInt approx)
{
    gfWinFPSCounter count;
    count.BeginMSCount();

    DeleteVACache();

    if (m_time_stamp == 0)
        return GF_S_OK;

    // Interval of definition
    gfDouble a = GetUKnot(m_u_degree);
    gfDouble b = GetUKnot(GetNumUControlPoints());
    if (gfIsZero(b - a))
        return GF_S_OK;

    // gfInterval of definition
    gfDouble c = GetVKnot(m_v_degree);
    gfDouble d = GetVKnot(GetNumVControlPoints());
    if (gfIsZero(d - c))
        return GF_S_OK;

    gfInt n = GetNumUControlPoints() - 1;
    gfInt m = GetNumVControlPoints() - 1;

    gfBool isTrimmed = (GetNumTrimCurves() > 0);

    // array for approximation points
    gfPoint3* aux_points = new gfPoint3[approx * approx];
    gfVector3* aux_normals = NULL;
    aux_normals = new gfVector3[approx * approx];

    // get UV space tesselation based on trim curves
    gfUVPoint* trimmed_points = NULL;
    if (isTrimmed)
        trimmed_points = TesselateTrimmedUVSpace(approx);

    for (gfInt i = 0;i < approx;i++)
    {
        // current u parameter
        gfDouble u = a + i * (b - a) / (approx - 1);
        gfInt uspan = gfGetBSplineKnotSpan(n, m_u_degree, &m_u_knot_vector, u);
        //std::vector<gfDouble> ubasis;
        //gfGetBSplineAllValues(n, m_u_degree, &m_u_knot_vector, uspan, u, &ubasis);
        gfAux2DArray<gfDouble> ubasis(2, 100);
        gfGetBSplineAllDers(m_u_degree, &m_u_knot_vector, uspan, u, 1, &ubasis);

        for (gfInt j = 0;j < approx;j++)
        {
            // current v parameter     
            gfDouble v = c + j * (d - c) / (approx - 1);
            gfInt vspan = gfGetBSplineKnotSpan(m, m_v_degree, &m_v_knot_vector, v);
            //std::vector<gfDouble> vbasis;
            //gfGetBSplineAllValues(m, m_v_degree, &m_v_knot_vector, vspan, v, &vbasis);
            gfAux2DArray<gfDouble> vbasis(2, 100);
            gfGetBSplineAllDers(m_v_degree, &m_v_knot_vector, vspan, v, 1, &vbasis);

            gfDouble ax = 0.0;
            gfDouble ay = 0.0;
            gfDouble az = 0.0;
            gfDouble aw = 0.0;

            gfDouble dux = 0.0;
            gfDouble duy = 0.0;
            gfDouble duz = 0.0;
            gfDouble duw = 0.0;
            gfDouble dvx = 0.0;
            gfDouble dvy = 0.0;
            gfDouble dvz = 0.0;
            gfDouble dvw = 0.0;

            for (gfInt k = 0;k <= m_u_degree;k++)
            {
                gfInt uindex = uspan - m_u_degree + k;

                for (gfInt l = 0;l <= m_v_degree;l++)
                {
                    gfInt vindex = vspan - m_v_degree + l;
                    //gfDouble aux = ubasis.at(k) * vbasis.at(l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                    gfDouble aux = ubasis.at(0, k) * vbasis.at(0, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                    ax += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
                    ay += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
                    az += aux * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
                    aw += aux;

                    gfDouble aux1 = ubasis.at(1, k) * vbasis.at(0, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                    dux += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
                    duy += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
                    duz += aux1 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
                    duw += aux1;
                    gfDouble aux2 = ubasis.at(0, k) * vbasis.at(1, l) * m_weights.at(uindex + m_num_u_control_points * vindex);
                    dvx += aux2 * m_control_points.at(uindex + m_num_u_control_points * vindex)->X();
                    dvy += aux2 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Y();
                    dvz += aux2 * m_control_points.at(uindex + m_num_u_control_points * vindex)->Z();
                    dvw += aux2;
                }
            }

            if (isTrimmed && trimmed_points[i + j * approx].boundary)
            {
                gfPoint3* patch_pt = GetPatchPoint(u + trimmed_points[i + j * approx].du, v + trimmed_points[i + j * approx].dv);
                if (patch_pt)
                {
                    aux_points[i + j * approx].Set(patch_pt->X(), patch_pt->Y(), patch_pt->Z());
                    gfDeletePoint3(patch_pt);
                    gfVector3* normal = GetPatchNormal(u + trimmed_points[i + j * approx].du, v + trimmed_points[i + j * approx].dv);
                    aux_normals[i + j * approx].Set(normal->X(), normal->Y(), normal->Z());
                    gfDeleteVector3(normal);
                }
                else
                {
                    patch_pt = GetPatchPoint(u, v);
                    aux_points[i + j * approx].Set(patch_pt->X(), patch_pt->Y(), patch_pt->Z());
                    gfDeletePoint3(patch_pt);
                    gfVector3* normal = GetPatchNormal(u, v);
                    aux_normals[i + j * approx].Set(normal->X(), normal->Y(), normal->Z());
                    gfDeleteVector3(normal);
                }
            }
            else
            {
                aux_points[i + j * approx].Set(ax / aw, ay / aw, az / aw);
                gfDouble fdux = (dux * aw - ax * duw) / aw / aw;
                gfDouble fduy = (duy * aw - ay * duw) / aw / aw;
                gfDouble fduz = (duz * aw - az * duw) / aw / aw;

                gfDouble fdvx = (dvx * aw - ax * dvw) / aw / aw;
                gfDouble fdvy = (dvy * aw - ay * dvw) / aw / aw;
                gfDouble fdvz = (dvz * aw - az * dvw) / aw / aw;

                aux_normals[i + j * approx].Set((fduy*fdvz-fduz*fdvy),
                    (-fdux*fdvz+fduz*fdvx),
                    (fdux*fdvy-fduy*fdvx));
                aux_normals[i + j * approx].Normalize();
            }

            //point->curvature = trimmed_points[i + j * approx].winding_number;
        }
    }

    // get number of triangles
    m_num_cached_quads = (approx-1)*(approx-1);
    m_num_cached_wire_lines = 2*(approx-1)*approx;

    // allocate streams
    m_cache_vertices = new gfDouble[12*m_num_cached_quads];
    m_cache_normals = new gfDouble[12*m_num_cached_quads];
    m_cache_wire_vertices = new gfDouble[6*m_num_cached_wire_lines];

    // fill quads
    gfInt ii = 0;
    gfInt jj = 0;
    for (gfInt i = 0;i < approx - 1;i++)
    {
        for (gfInt j = 0;j < approx - 1;j++)
        {
            gfInt wind1 = 0;
            gfInt wind2 = 0;
            gfInt wind3 = 0;
            gfInt wind4 = 0;
            if (isTrimmed)
            {
                wind1 = (gfInt) trimmed_points[i + 1 + approx * j].winding_number;
                wind2 = (gfInt) trimmed_points[i + 1 + approx * (j + 1)].winding_number;
                wind3 = (gfInt) trimmed_points[i + approx * (j + 1)].winding_number;
                wind4 = (gfInt) trimmed_points[i + approx * j].winding_number;
            }

            if (wind1 % 2 == 1 || wind2 % 2 == 1 || wind3 % 2 == 1 || wind4 % 2 == 1)
            {
                // face is trimmed
            }
            else
            {
                m_cache_vertices[ii] = aux_points[i + j * approx].X();
                m_cache_normals[ii] = aux_normals[i + j * approx].X();
                ii++;
                m_cache_vertices[ii] = aux_points[i + j * approx].Y();
                m_cache_normals[ii] = aux_normals[i + j * approx].Y();
                ii++;
                m_cache_vertices[ii] = aux_points[i + j * approx].Z();
                m_cache_normals[ii] = aux_normals[i + j * approx].Z();
                ii++;

                m_cache_vertices[ii] = aux_points[i + (j + 1) * approx].X();
                m_cache_normals[ii] = aux_normals[i + (j + 1) * approx].X();
                ii++;
                m_cache_vertices[ii] = aux_points[i + (j + 1) * approx].Y();
                m_cache_normals[ii] = aux_normals[i + (j + 1) * approx].Y();
                ii++;
                m_cache_vertices[ii] = aux_points[i + (j + 1) * approx].Z();
                m_cache_normals[ii] = aux_normals[i + (j + 1) * approx].Z();
                ii++;

                m_cache_vertices[ii] = aux_points[i + 1 + (j + 1) * approx].X();
                m_cache_normals[ii] = aux_normals[i + 1 + (j + 1) * approx].X();
                ii++;
                m_cache_vertices[ii] = aux_points[i + 1 + (j + 1) * approx].Y();
                m_cache_normals[ii] = aux_normals[i + 1 + (j + 1) * approx].Y();
                ii++;
                m_cache_vertices[ii] = aux_points[i + 1 + (j + 1) * approx].Z();
                m_cache_normals[ii] = aux_normals[i + 1 + (j + 1) * approx].Z();
                ii++;

                m_cache_vertices[ii] = aux_points[i + 1 + j * approx].X();
                m_cache_normals[ii] = aux_normals[i + 1 + j * approx].X();
                ii++;
                m_cache_vertices[ii] = aux_points[i + 1 + j * approx].Y();
                m_cache_normals[ii] = aux_normals[i + 1 + j * approx].Y();
                ii++;
                m_cache_vertices[ii] = aux_points[i + 1 + j * approx].Z();
                m_cache_normals[ii] = aux_normals[i + 1 + j * approx].Z();
                ii++;
            }
        }
    }

    for (gfInt i = 0;i < approx;i++)
    {
        for (gfInt j = 0;j < approx;j++)
        {
            gfInt wind1 = 0;
            gfInt wind2 = 0;
            gfInt wind3 = 0;
            gfInt wind4 = 0;
            if (isTrimmed)
            {
                wind1 = (gfInt) trimmed_points[i + 1 + approx * j].winding_number;
                wind2 = (gfInt) trimmed_points[i + 1 + approx * (j + 1)].winding_number;
                wind3 = (gfInt) trimmed_points[i + approx * (j + 1)].winding_number;
                wind4 = (gfInt) trimmed_points[i + approx * j].winding_number;
            }

            if (i < approx - 1 && !(wind1 % 2 == 1 || wind2 % 2 == 1))
            {
                m_cache_wire_vertices[jj] = aux_points[i + j * approx].X();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + j * approx].Y();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + j * approx].Z();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + 1 + j * approx].X();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + 1 + j * approx].Y();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + 1 + j * approx].Z();
                jj++;
            }
            if (j < approx - 1 && !(wind3 % 2 == 1 || wind4 % 2 == 1))
            {
                m_cache_wire_vertices[jj] = aux_points[i + j * approx].X();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + j * approx].Y();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + j * approx].Z();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + (j + 1) * approx].X();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + (j + 1) * approx].Y();
                jj++;
                m_cache_wire_vertices[jj] = aux_points[i + (j + 1) * approx].Z();
                jj++;
            }
        }
    }

    delete [] aux_points;
    delete [] aux_normals;
    delete [] trimmed_points;

    gfLog("Computed VertexArray cache for NURBS patch, time = %d ms.\n", count.EndMSCount());

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::DeleteVACache(void)
{
    if (m_cache_vertices)
        delete [] m_cache_vertices;
    m_cache_vertices = NULL;
    if (m_cache_normals)
        delete [] m_cache_normals;
    m_cache_normals = NULL;
    if (m_cache_wire_vertices)
        delete [] m_cache_wire_vertices;
    m_cache_wire_vertices = NULL;
    m_num_cached_quads = 0;
    m_num_cached_wire_lines = 0;

    return GF_S_OK;
}

GFRESULT gfObjNURBSPatch::DrawVACache(gfLong flags)
{
    if (flags & GF_FLG_SOLID)
    {
        glEnableClientState(GL_VERTEX_ARRAY);
        if (flags & GF_FLG_NORMALS)
            glEnableClientState(GL_NORMAL_ARRAY);

        if (m_cache_vertices)
            glVertexPointer(3, GL_DOUBLE, 0, m_cache_vertices);
        if (m_cache_normals && (flags & GF_FLG_NORMALS))
            glNormalPointer(GL_DOUBLE, 0, m_cache_normals);

        glDrawArrays(GL_QUADS, 0, 4*m_num_cached_quads);

        if (flags & GF_FLG_NORMALS)
            glDisableClientState(GL_NORMAL_ARRAY);
        glDisableClientState(GL_VERTEX_ARRAY);
    }

    if (flags & GF_FLG_WIRE)
    {
        glNormal3f(0.0, 0.0, 0.0);
        glEnableClientState(GL_VERTEX_ARRAY);

        if (m_cache_wire_vertices)
            glVertexPointer(3, GL_DOUBLE, 0, m_cache_wire_vertices);

        glDrawArrays(GL_LINES, 0, 2*m_num_cached_wire_lines);

        glDisableClientState(GL_VERTEX_ARRAY);
    }

    return GF_S_OK;
}

gfRayBezierPatchInfo* gfObjNURBSPatch::BezierClipping(gfMatrix4* obj_matrix, gfInt depth, 
                                                      gfDouble a1, gfDouble b1, gfDouble c1, gfDouble d1,
                                                      gfDouble a2, gfDouble b2, gfDouble c2, gfDouble d2)
{
    if (!IsBezier())
    {
        // convert to Bezier segments
        gfAux2DArray<gfObjNURBSPatch*> segments;
        DecomposeToBezierSegments(&segments);
        
        gfRayBezierPatchInfo* info = new gfRayBezierPatchInfo;
        for (gfInt i = 0;i < segments.GetISize();i++)
        {
            for (gfInt j = 0;j < segments.GetJSize();j++)
            {
                gfRayBezierPatchInfo* info_aux = segments.at(i, j)->BezierClipping(obj_matrix, depth, a1, b1, c1, d1, a2, b2, c2, d2);
                if (info_aux)
                {
                    for (gfInt k = 0;k < (gfInt)info_aux->m_inters.size();k++)
                        info->m_inters.push_back(info_aux->m_inters.at(k));
                }
                delete info_aux;
                gfDeleteGeomObject(segments.at(i, j));
            }
        }

        if (info->m_inters.size() > 0)
        {
            return info;
        }
        else
        {
            delete info;
            return NULL;
        }
    }
    else
    {
        gfInt n = GetNumUControlPoints()-1;
        gfInt m = GetNumVControlPoints()-1;
        gfDouble umin, umax, vmin, vmax;
        GetDomain(&umin, &umax, &vmin, &vmax);

        // create auxiliary points
        gfAux2DArray<gfPoint2> auxP;
        auxP.resize(n+1, m+1);
        for (gfInt i = 0;i <= n;i++)
            for (gfInt j = 0;j <= m;j++)
            {
                gfPoint3* pt = GetControlPoint(i, j);
                auxP.Set(i,j, gfPoint2(GetWeight(i, j)*(a1*pt->X()+b1*pt->Y()+c1*pt->Z()+d1), 
                                    GetWeight(i, j)*(a2*pt->X()+b2*pt->Y()+c2*pt->Z()+d2)));
            }

        // find line to project these points
        gfDouble a, b;
        gfPoint2 pt1 = auxP.at(0, 0);
        gfPoint2 pt2 = auxP.at(n, 0);
        gfPoint2 pt3 = auxP.at(0, m);
        gfPoint2 pt4 = auxP.at(n, m);

        gfVector2 dif1;
        dif1.Set(pt2.X()-pt1.X()+pt4.X()-pt3.X(),
                pt2.Y()-pt1.Y()+pt4.Y()-pt3.Y());
        gfVector2 dif2;
        dif2.Set(pt3.X()-pt1.X()+pt4.X()-pt2.X(),
                pt3.Y()-pt1.Y()+pt4.Y()-pt2.Y());

        // this can be changed to lenght of domain
        //gfDouble size1 = dif2.Length();
        //gfDouble size2 = dif1.Length();
        gfDouble size1 = umax-umin;
        gfDouble size2 = vmax-vmin;

        //if (size1>size2)
        if (depth % 2 == 0)
        {
            // u direction
            dif1.Normalize();
            a = -dif2.Y();
            b = dif2.X();
        }
        else
        {
            // v direction
            dif2.Normalize();
            a = -dif1.Y();
            b = dif1.X();
        }

        // project points
        gfAux2DArray<gfDouble> auxD;
        auxD.resize(n+1, m+1);
        for (gfInt i = 0;i <= n;i++)
            for (gfInt j = 0;j <= m;j++)
            {
                gfPoint2 pt = auxP.at(i, j);
                gfDouble d = a*pt.X()+b*pt.Y();
                auxD.Set(i, j, a*pt.X()+b*pt.Y());
            }

        // do clipping for significant direction
        std::vector<gfPoint2> aux;
        aux.resize((n+1)*(m+1));
        if (depth % 2 == 0)
        {
            // clipping in u direction
            for (gfInt i = 0;i <= n;i++)
                for (gfInt j = 0;j <= m;j++)
                {
                    aux[i+j*(n+1)] = gfPoint2(umin+(umax-umin)*((gfDouble) i)/((gfDouble) n), auxD.at(i, j));
                }

            // find better approximation of intersection region
            gfDouble uclipmin, uclipmax;
            if (gfAxisXConvexHullInter2D(&aux, &uclipmin, &uclipmax) != GF_S_OK)
                return NULL;
            if (uclipmin < umin || uclipmax > umax)
                return NULL;

            // check if region is small
            if (((uclipmax-uclipmin) < GF_BEZIER_CLIPPING_EPS && (vmax-vmin) < GF_BEZIER_CLIPPING_EPS) || depth > GF_BEZIER_CLIPPING_MAX_DEPTH)
            {
                gfRayBezierPatchInfo* result = new gfRayBezierPatchInfo;
                result->m_inters.push_back(gfPoint2((uclipmax+uclipmin)/2.0, (vmax+vmin)/2.0));
                return result;
            }

            // check if no significant change in interval happend
            if ((uclipmax-uclipmin)/(umax-umin) > GF_BEZIER_CLIPPING_DIVIDE_TRESHOLD)
            {
                gfObjNURBSPatch* patch1 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
                gfObjNURBSPatch* patch2 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
                DecomposeUDeCasteljau((umax+umin)/2.0, patch1, patch2);
                gfRayBezierPatchInfo* res1 = patch1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfRayBezierPatchInfo* res2 = patch2->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfDeleteGeomObject(patch1);
                gfDeleteGeomObject(patch2);

                // combine two results
                if (res1 && res2)
                {
                    for (gfInt i = 0;i < (gfInt)res2->m_inters.size();i++)
                        res1->m_inters.push_back(res2->m_inters.at(i));
                    delete res2;
                    return res1;
                }
                if (res1 && !res2)
                    return res1;
                if (res2 && !res1)
                    return res2;
                return NULL;
            }

            // divide patch using deCasteljau
            gfObjNURBSPatch* patch1 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
            gfObjNURBSPatch* patch2 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
            gfObjNURBSPatch* patch3 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
            DecomposeUDeCasteljau(uclipmin, patch1, patch2);
            patch2->DecomposeUDeCasteljau(uclipmax, patch1, patch3);

            gfDeleteGeomObject(patch2);
            gfDeleteGeomObject(patch3);

            gfRayBezierPatchInfo* res = patch1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);

            gfDeleteGeomObject(patch1);

            return res;
        }
        else
        {
            // clipping in v direction
            for (gfInt i = 0;i <= n;i++)
                for (gfInt j = 0;j <= m;j++)
                {
                    aux[i+j*(n+1)] = gfPoint2(vmin+(vmax-vmin)*((gfDouble) j)/((gfDouble) m), auxD.at(i, j));
                }

            // find better approximation of intersection region
            gfDouble vclipmin, vclipmax;
            if (gfAxisXConvexHullInter2D(&aux, &vclipmin, &vclipmax) != GF_S_OK)
                return NULL;
            if (vclipmin < vmin || vclipmax > vmax)
                return NULL;

            // check if region is small
            if (((vclipmax-vclipmin) < GF_BEZIER_CLIPPING_EPS && (umax-umin) < GF_BEZIER_CLIPPING_EPS) || depth > GF_BEZIER_CLIPPING_MAX_DEPTH)
            {
                gfRayBezierPatchInfo* result = new gfRayBezierPatchInfo;
                result->m_inters.push_back(gfPoint2((umax+umin)/2.0, (vclipmax+vclipmin)/2.0));
                return result;
            }

            // check if no significant change in intterval happend
            if ((vclipmax-vclipmin)/(vmax-vmin) > GF_BEZIER_CLIPPING_DIVIDE_TRESHOLD)
            {
                gfObjNURBSPatch* patch1 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
                gfObjNURBSPatch* patch2 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
                DecomposeVDeCasteljau((vmax+vmin)/2.0, patch1, patch2);
                gfRayBezierPatchInfo* res1 = patch1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfRayBezierPatchInfo* res2 = patch2->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);
                gfDeleteGeomObject(patch1);
                gfDeleteGeomObject(patch2);

                // combine two results
                if (res1 && res2)
                {
                    for (gfInt i = 0;i < (gfInt)res2->m_inters.size();i++)
                        res1->m_inters.push_back(res2->m_inters.at(i));
                    delete res2;
                    return res1;
                }
                if (res1 && !res2)
                    return res1;
                if (res2 && !res1)
                    return res2;
                return NULL;
            }

            // divide patch using deCasteljau
            gfObjNURBSPatch* patch1 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
            gfObjNURBSPatch* patch2 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
            gfObjNURBSPatch* patch3 = (gfObjNURBSPatch*) gfCreateGeomObject(GF_NURBS_PATCH);
            DecomposeVDeCasteljau(vclipmin, patch1, patch2);
            patch2->DecomposeVDeCasteljau(vclipmax, patch1, patch3);

            gfDeleteGeomObject(patch2);
            gfDeleteGeomObject(patch3);

            gfRayBezierPatchInfo* res = patch1->BezierClipping(obj_matrix, depth+1, a1, b1, c1, d1, a2, b2, c2, d2);

            gfDeleteGeomObject(patch1);

            return res;
        }
    }

    return NULL;
}

GFRESULT gfObjNURBSPatch::EnableBezierClipping(gfBool enable)
{
    m_bezier_clipping = enable;
    return GF_S_OK;
}