//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjNURBSPatch.h                                    //
//   Description: NURBS patch class                             //
//                                                              //
//////////////////////////////////////////////////////////////////


#ifndef _GF_C_NURBS_PATCH_H_
#define _GF_C_NURBS_PATCH_H_


#include "gfDefines.h"
#include <vector>
#include "gfAux2DArray.h"
#include "gfPoint.h"
#include "gfVector.h"
#include "gfMatrix.h"
#include "gfObjGeomObject.h"
#include "gfObjNURBSCurve.h"
#include "gfObjHalfEdgeMesh.h"

class gfObjNURBSPatch;

// structure for tesselating uv parameter space
typedef struct tag_gfUVPoint
{
    gfInt winding_number;
    gfInt boundary;
    gfDouble du, dv;
} gfUVPoint;

// node of AABB tree for faster ray intersection and collision
struct gfBezier2AABBTreeNode
{
    gfAuxAABB* aabb;
    gfBezier2AABBTreeNode* left;
    gfBezier2AABBTreeNode* right;
    gfObjNURBSPatch* patch;
    gfObjHalfEdgeMesh* control_mesh;
    gfInt min_dir;
};

struct gfRayBezierPatchInfo
{
    std::vector<gfPoint2> m_inters;
};

class gfObjNURBSPatch : public gfObjGeomObject
{
private:
    // basic parameters
    std::vector<gfDouble> m_u_knot_vector;
    std::vector<gfDouble> m_v_knot_vector;
    std::vector<gfDouble> m_weights;
    std::vector<gfPoint3*> m_control_points;
    gfInt m_num_u_control_points;
    gfInt m_num_v_control_points;
    gfInt m_u_degree;
    gfInt m_v_degree;

    // trimming curves
    std::vector<gfObjNURBSCurve*> m_trim_curves;
    gfUVPoint* TesselateTrimmedUVSpace(gfInt approx);

    // vertex array cache
    gfDouble* m_cache_vertices;
    gfDouble* m_cache_normals;
    gfDouble* m_cache_wire_vertices;
    gfInt m_num_cached_quads;
    gfInt m_num_cached_wire_lines;
    GFRESULT CreateVACache(gfInt approx);
    GFRESULT DeleteVACache(void);
    GFRESULT DrawVACache(gfLong flags);

    // Bezier AABB tree functions
    std::vector<gfBezier2AABBTreeNode*> m_bezier_aabb_trees;
    gfRayInterInfo* RayAABBIntersection(gfBezier2AABBTreeNode* root, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfBezier2AABBTreeNode* CreateAABBTree(gfDouble depth, gfObjNURBSPatch* patch);
    GFRESULT ComputeWholeAABBTree(void);
    GFRESULT DeleteWholeAABBTree(void);
    GFRESULT DeleteAABBTree(gfBezier2AABBTreeNode* tree_root);
    gfInt m_num_nodes;

    // Bezier clipping
    gfBool m_bezier_clipping;

public:

    gfObjNURBSPatch(void);
    ~gfObjNURBSPatch(void);

    // gfObjGeomObject functions
    GFRESULT OGLDraw(gfLong flags);
    gfRayInterInfo* RayIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir);
    gfGeomObjID GetID(void);
    GFRESULT Transform(gfMatrix4* matrix);
    gfObjGeomObject* Clone(void);
    gfAuxBoundingVolume* GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix);

    // initialize NURBS patch
    GFRESULT Clear(void);
    GFRESULT InitPlane(gfDouble xmin, gfDouble xmax, gfDouble ymin, gfDouble ymax);
    GFRESULT InitSphere(gfPoint3* center, gfPoint3* start);
    GFRESULT InitTorus(gfPoint3* center, gfPoint3* start, gfPoint3 * start2);
    GFRESULT InitCone(gfPoint3* center, gfDouble radius1, gfDouble radius2, gfDouble height);
    GFRESULT InitRotational(gfPoint3 * center, gfVector3* axis, gfObjNURBSCurve* nurbs_curve);
    GFRESULT InitRotational(gfPoint3 * center, gfVector3* axis, gfDouble angle, gfObjNURBSCurve* nurbs_curve);
    GFRESULT InitBezier(gfInt num_u_points, gfInt num_v_points);
    GFRESULT InitBezier(gfInt num_u_points, gfInt num_v_points, gfDouble u_min, gfDouble u_max, gfDouble v_min, gfDouble v_max);
    GFRESULT InitRandom(gfPoint3* center, gfDouble tx, gfDouble ty);
    GFRESULT InitTransitional(gfObjNURBSCurve* curve1, gfObjNURBSCurve* curve2);
    GFRESULT InitNormalTransitional(gfObjNURBSCurve* curve1, gfObjNURBSCurve* curve2);
    GFRESULT InitRuledCurveToLine(gfObjNURBSCurve* curve, gfPoint3* line_pt, gfVector3* line_dir);
    GFRESULT InitProfile(gfDouble xmin, gfDouble xmax, gfInt num_u_points, gfDouble ymin, gfDouble ymax, gfInt num_v_points);

    // working with parameters of patch
    gfInt SetParams(gfInt u_degree, gfInt v_degree, gfInt num_u_control_points, gfInt num_v_control_points);
    gfInt GetUDegree(void);
    gfInt GetVDegree(void);
    gfInt GetNumUKnots(void);
    gfInt GetNumVKnots(void);
    gfDouble GetUKnot(int index);
    gfDouble GetVKnot(int index);
    GFRESULT SetUKnot(int index, gfDouble knot);
    GFRESULT SetVKnot(int index, gfDouble knot);
    gfInt GetNumUControlPoints(void);
    gfInt GetNumVControlPoints(void);
    gfPoint3* GetControlPoint(gfInt u_index, gfInt v_index);
    gfDouble GetWeight(gfInt u_index, gfInt v_index);
    GFRESULT SetWeight(gfInt u_index, gfInt v_index, gfDouble weight);
    gfObjHalfEdgeMesh* GetControlMesh(void);
    gfObjHalfEdgeMesh* GetApproxMesh(gfInt approx);
    gfObjHalfEdgeMesh* GetTrimmedApproxMesh(gfInt approx);
    GFRESULT GetDomain(gfDouble* u_min, gfDouble* u_max, gfDouble* v_min, gfDouble* v_max);

    // modeling functions
    gfPoint3* GetPatchPoint(gfDouble u, gfDouble v);
    gfVector3* GetPatchUDer(gfDouble u, gfDouble v);
    gfVector3* GetPatchVDer(gfDouble u, gfDouble v);
    gfVector3* GetPatchNormal(gfDouble u, gfDouble v);
    gfObjNURBSCurve* GetUIsocurve(gfDouble v_param);
    gfObjNURBSCurve* GetVIsocurve(gfDouble u_param);
    gfBool IsBezier(void);
    GFRESULT InsertUKnot(gfInt mult, gfDouble u);
    GFRESULT InsertVKnot(gfInt mult, gfDouble v);
    GFRESULT DecomposeToBezierSegments(gfAux2DArray<gfObjNURBSPatch*>* segments);
    GFRESULT DecomposeUDeCasteljau(gfDouble u, gfObjNURBSPatch* patch1, gfObjNURBSPatch* patch2);
    GFRESULT DecomposeVDeCasteljau(gfDouble v, gfObjNURBSPatch* patch1, gfObjNURBSPatch* patch2);

    // trimmed surface support
    GFRESULT AddTrimCurve(gfObjNURBSCurve* curve);
    gfObjNURBSCurve* GetTrimCurve(gfInt index);
    gfInt GetNumTrimCurves(void);
    GFRESULT RemoveTrimCurve(gfInt index);
    GFRESULT RemoveAllTrimCurves(void);
    GFRESULT PointInsideDomain(gfDouble u, gfDouble v);

    // Bezier clipping
    gfRayBezierPatchInfo* BezierClipping(gfMatrix4* obj_matrix, gfInt depth,
                                         gfDouble a1, gfDouble b1, gfDouble c1, gfDouble d1,
                                         gfDouble a2, gfDouble b2, gfDouble c2, gfDouble d2);
    GFRESULT EnableBezierClipping(gfBool enable);
} ;


#endif