//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjPolyLine.cpp                                    //
//   Description: Polyline class                                //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfObjectFactory.h"
#include "gfObjPolyLine.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"
#include "gfAuxItersectionUtils.h"

#include "./OGL/glut.h"


gfObjPolyLine::gfObjPolyLine(void)
{
    m_time_stamp = 0;
    m_points.resize(0);
    m_closed = false;
}

gfObjPolyLine::~gfObjPolyLine(void)
{
    SetNumPoints(0);
}

GFRESULT gfObjPolyLine::OGLDraw(gfLong flags)
{
    if (m_time_stamp == 0)
        return GF_S_OK;

    gfInt n = GetNumPoints();

    if ((flags & GF_FLG_WIRE) || (flags & GF_FLG_SOLID))
    {
        glLineWidth(5.0);

        if (m_closed)
            glBegin(GL_LINE_LOOP);
        else
            glBegin(GL_LINE_STRIP);
        for (gfInt i = 0;i < n;i++)
        {
            glVertex3f(m_points.at(i)->X(), m_points.at(i)->Y(), m_points.at(i)->Z());
        }
        glEnd();

        glLineWidth(1.0);
    }

    return GF_S_OK;
}

gfRayInterInfo* gfObjPolyLine::RayIntersection(gfMatrix4* obj_matrix, gfPoint3* ray_begin, gfVector3* ray_dir)
{
    gfInt num_points = GetNumPoints();

    // go throught all segments of polyline
    gfRayInterInfo* info = NULL;
    gfDouble nx = 0.0;
    gfDouble ny = 0.0;
    gfDouble nz = 0.0;

    gfInt diff = -1;
    if (m_closed)
        diff = 0;

    for (gfInt i = 0;i < num_points + diff;i++)
    {
        gfPoint3* point1 = GetPoint(i);
        gfPoint3* point2 = GetPoint((i+1)%num_points);

        nx += (point2->Y()-point1->Y())*(point2->Z()-point1->Z());
        ny += (point2->Z()-point1->Z())*(point2->X()-point1->X());
        nz += (point2->X()-point1->X())*(point2->Y()-point1->Y());

        gfDouble t;
        GFRESULT res = gfRaySegmentInter(ray_begin, ray_dir, point1, point2, &t);
        if (res == GF_S_OK)
        {
            if (!info)
            {
                info = gfCreateRayInterInfo();
            }

            info->m_inters.push_back(t);
            info->m_normals.push_back(gfVector3(nx, ny, nz));
            gfSort1DArray(&info->m_inters);
        }
    }

    return info;
}

gfGeomObjID gfObjPolyLine::GetID(void)
{
    return GF_POLYLINE;
}

gfObjGeomObject* gfObjPolyLine::Clone(void)
{
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());
    gfObjPolyLine* polyline = (gfObjPolyLine*) clone;

    // initialize
    gfInt num_points = GetNumPoints();
    polyline->Init(num_points, m_closed);

    // copy points
    for (gfInt i = 0;i < num_points;i++)
    {
        gfPoint3* point = polyline->GetPoint(i);
        gfPoint3* old_point = m_points.at(i);
        point->Set(old_point->X(), old_point->Y(), old_point->Z());
    }

    return clone;
}

GFRESULT gfObjPolyLine::Transform(gfMatrix4* matrix)
{
    gfInt n = GetNumPoints();

    for (gfInt i = 0;i < n;i++)
    {
        gfPoint3* pt = m_points.at(i);
        gfTransformPoint3(pt, matrix);
    }

    m_time_stamp++;

    return GF_S_OK;
}

GFRESULT gfObjPolyLine::Init(gfInt num_points, gfBool closed)
{
    m_closed = closed;

    m_points.clear();
    SetNumPoints(num_points);

    return GF_S_OK;
}

GFRESULT gfObjPolyLine::InitRandom(gfInt num_points, gfBool closed)
{
    m_closed = closed;

    m_points.clear();
    SetNumPoints(num_points);

    for (gfInt i = 0;i < num_points;i++)
    {
        gfPoint3* pt = GetPoint(i);
        pt->Set((i-num_points/2), gfRand(-2.0, 2.0), 0.0);
    }

    return GF_S_OK;
}

GFRESULT gfObjPolyLine::InitNgon(gfPoint3* center, gfVector3* up, gfInt num_points, gfDouble radius)
{
    Init(num_points, true);

    // get first base vector
    gfVector3 base1(-2*up->Y()-up->Z(), 2*up->X(), up->X());
    base1.Normalize();
    gfVector3 base2;
    gfVectorProduct3(&base1, up, &base2);
    base2.Normalize();

    for (gfInt i = 0;i < num_points;i++)
    {
        gfPoint3* pt = GetPoint(i);
        gfVector3 aux;
        gfCombineVectors3(gfCos(2.0*GF_PI*i/num_points), &base1, gfSin(2.0*GF_PI*i/num_points), &base2, &aux);
        gfCombinePointVector3(center, radius, &aux, pt);
    }

    return GF_S_OK;
}

GFRESULT gfObjPolyLine::SetNumPoints(gfInt num_points)
{
    gfInt old_num = (gfInt)m_points.size();

    if (old_num > num_points)
    {
        for (gfInt i = num_points;i < old_num;i++)
        {
            gfPoint3* old_pt = m_points.at(i);
            gfDeletePoint3(old_pt);
        }
    }

    m_points.resize(num_points);

    if (old_num < num_points)
    {
        for (gfInt i = old_num;i < num_points;i++)
        {
            gfPoint3* new_pt = gfCreatePoint3();
            m_points[i] = new_pt;
        }
    }

    m_time_stamp++;

    return GF_S_OK;
}

gfInt gfObjPolyLine::GetNumPoints(void)
{
    return (gfInt)m_points.size();
}

gfPoint3* gfObjPolyLine::GetPoint(gfInt index)
{
    return m_points.at(index);
}

gfBool gfObjPolyLine::IsClosed(void)
{
    return m_closed;
}

GFRESULT gfObjPolyLine::SetClosed(gfBool closed)
{
    m_closed = closed;
    return GF_S_OK;
}

gfObjPolyLine* gfObjPolyLine::ChainkinSubdivision(void)
{
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());
    gfObjPolyLine* polyline = (gfObjPolyLine*) clone;

    // initialize
    gfInt num_old_points = GetNumPoints();
    gfInt num_new_points;
    if (m_closed)
    {
        num_new_points = 2 * num_old_points;
    }
    else
    {
        num_new_points = 2 * num_old_points - 2;
    }
    polyline->Init(num_new_points, m_closed);

    // copy points
    if (m_closed)
    {
        for (gfInt i = 0;i < num_old_points;i++)
        {
            gfPoint3* new_point_1 = polyline->GetPoint(2 * i);
            gfPoint3* new_point_2 = polyline->GetPoint(2 * i + 1);

            // (i-1)=i+num_old_point modulo num_old_point
            gfPoint3* prev_point = m_points.at((i + num_old_points - 1) % num_old_points);
            gfPoint3* point = m_points.at(i);
            gfPoint3* next_point = m_points.at((i + 1) % num_old_points);

            gfDouble px, py, pz;
            px = 0.75 * point->X() + 0.25 * prev_point->X();
            py = 0.75 * point->Y() + 0.25 * prev_point->Y();
            pz = 0.75 * point->Z() + 0.25 * prev_point->Z();
            new_point_1->Set(px, py, pz);

            px = 0.25 * next_point->X() + 0.75 * point->X();
            py = 0.25 * next_point->Y() + 0.75 * point->Y();
            pz = 0.25 * next_point->Z() + 0.75 * point->Z();
            new_point_2->Set(px, py, pz);
        }
    }
    else
    {
        for (gfInt i = 1;i < num_old_points - 1;i++)
        {
            gfPoint3* new_point_1 = polyline->GetPoint(2 * i - 1);
            gfPoint3* new_point_2 = polyline->GetPoint(2 * i);

            gfPoint3* prev_point = m_points.at((i + num_old_points - 1) % num_old_points);
            gfPoint3* point = m_points.at(i);
            gfPoint3* next_point = m_points.at((i + 1) % num_old_points);

            gfDouble px, py, pz;
            px = 0.75 * point->X() + 0.25 * prev_point->X();
            py = 0.75 * point->Y() + 0.25 * prev_point->Y();
            pz = 0.75 * point->Z() + 0.25 * prev_point->Z();
            new_point_1->Set(px, py, pz);

            px = 0.25 * next_point->X() + 0.75 * point->X();
            py = 0.25 * next_point->Y() + 0.75 * point->Y();
            pz = 0.25 * next_point->Z() + 0.75 * point->Z();
            new_point_2->Set(px, py, pz);
        }

        // set first and last point
        gfPoint3* first_point = m_points.at(0);
        gfPoint3* new_first_point = polyline->GetPoint(0);
        new_first_point->Set(first_point->X(), first_point->Y(), first_point->Z());

        gfPoint3* last_point = m_points.at(num_old_points - 1);
        gfPoint3* new_last_point = polyline->GetPoint(num_new_points - 1);
        new_last_point->Set(last_point->X(), last_point->Y(), last_point->Z());
    }

    return polyline;
}

gfObjPolyLine* gfObjPolyLine::CatmulClarkSubdivision(void)
{
    gfObjGeomObject* clone = gfCreateGeomObject(GetID());
    gfObjPolyLine* polyline = (gfObjPolyLine*) clone;

    // initialize
    gfInt num_old_points = GetNumPoints();
    gfInt num_new_points;
    if (m_closed)
    {
        num_new_points = 2 * num_old_points;
    }
    else
    {
        num_new_points = 2 * num_old_points - 1;
    }
    polyline->Init(num_new_points, m_closed);

    // copy points
    if (!m_closed)
    {
        // curve isn't closed
        for (gfInt i = 1;i < num_new_points-1;i++)
        {
            if ((i % 2) == 1)
            {
                // new edge point
                gfPoint3 * new_point = polyline->GetPoint(i);
                gfInt old_index = i / 2;
                gfPoint3 * point = m_points.at(old_index);
                gfPoint3 * next_point = m_points.at(old_index + 1);

                gfDouble px, py, pz;
                px = 0.5 * point->X() + 0.5 * next_point->X();
                py = 0.5 * point->Y() + 0.5 * next_point->Y();
                pz = 0.5 * point->Z() + 0.5 * next_point->Z();
                new_point->Set(px, py, pz);
            }
            else
            {
                // new edge point
                gfPoint3 * new_point = polyline->GetPoint(i);
                gfInt old_index = i / 2;
                gfPoint3 * point = m_points.at(old_index);
                gfPoint3 * next_point = m_points.at(old_index + 1);
                gfPoint3 * prev_point = m_points.at(old_index - 1);

                gfDouble px, py, pz;
                px = 0.75 * point->X() + 0.125 * next_point->X() + 0.125 * prev_point->X();
                py = 0.75 * point->Y() + 0.125 * next_point->Y() + 0.125 * prev_point->Y();
                pz = 0.75 * point->Z() + 0.125 * next_point->Z() + 0.125 * prev_point->Z();
                new_point->Set(px, py, pz);
            }
        }

        // set first and last point
        gfPoint3* first_point = m_points.at(0);
        gfPoint3* new_first_point = polyline->GetPoint(0);
        new_first_point->Set(first_point->X(), first_point->Y(), first_point->Z());

        gfPoint3* last_point = m_points.at(num_old_points - 1);
        gfPoint3* new_last_point = polyline->GetPoint(num_new_points - 1);
        new_last_point->Set(last_point->X(), last_point->Y(), last_point->Z());
    }
    else
    {
        // curve is closed
        for (gfInt i = 0;i < num_new_points;i++)
        {
            if ((i % 2) == 1)
            {
                // new edge point
                gfPoint3 * new_point = polyline->GetPoint(i);
                gfInt old_index = i / 2;
                gfPoint3 * point = m_points.at(old_index);
                gfPoint3 * next_point = m_points.at((old_index + 1) % num_old_points);

                gfDouble px, py, pz;
                px = 0.5 * point->X() + 0.5 * next_point->X();
                py = 0.5 * point->Y() + 0.5 * next_point->Y();
                pz = 0.5 * point->Z() + 0.5 * next_point->Z();
                new_point->Set(px, py, pz);
            }
            else
            {
                // new point
                gfPoint3 * new_point = polyline->GetPoint(i);
                gfInt old_index = i / 2;
                gfPoint3 * point = m_points.at(old_index);
                gfPoint3 * next_point = m_points.at((old_index + 1) % num_old_points);
                gfPoint3 * prev_point = m_points.at((old_index - 1 + num_old_points) % num_old_points);

                gfDouble px, py, pz;
                px = 0.75 * point->X() + 0.125 * next_point->X() + 0.125 * prev_point->X();
                py = 0.75 * point->Y() + 0.125 * next_point->Y() + 0.125 * prev_point->Y();
                pz = 0.75 * point->Z() + 0.125 * next_point->Z() + 0.125 * prev_point->Z();
                new_point->Set(px, py, pz);
            }
        }
    }

    return polyline;
}

gfAuxBoundingVolume* gfObjPolyLine::GetBoundVolume(gfBoundVolID id, gfMatrix4* obj_matrix)
{
    // create auxiliary array of mesh points
    gfInt num_points = (gfInt)m_points.size();
    if (num_points < 1)
        return NULL;

    gfPoint3* points = new gfPoint3[num_points];
    gfPoint3* curr_point = points;

    for (gfInt i = 0; i < num_points; i++)
    {
        gfPoint3* pt = m_points.at(i);
        curr_point->Set(pt->X(), pt->Y(), pt->Z());
        if (obj_matrix)
            gfTransformPoint3(curr_point, obj_matrix);

        curr_point++;
    }

    gfAuxBoundingVolume* result;
    if (id == GF_BOUND_VOL_OBB)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_OBB);
        gfAuxOBB* obb = (gfAuxOBB*) result;
        obb->InitFromPoints(num_points, points);
    }
    else if (id == GF_BOUND_VOL_SPHERE)
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_SPHERE);
        gfAuxBoundingShere* sphere = (gfAuxBoundingShere*) result;
        sphere->InitFromPoints(num_points, points);
    }
    else
    {
        result = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
        gfAuxAABB* aabb = (gfAuxAABB*) result;
        aabb->InitFromPoints(num_points, points);
    }

    delete [] points;
    return result;
}
