//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfObjbjectFactory.cpp                                  //
//   Description: Factory class for working with objects        //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfObjectFactory.h"
#include "gfAuxLogging.h"


// geometry objects
#include "gfObjPolyLine.h"
#include "gfObjNURBSCurve.h"
#include "gfObjNURBSPatch.h"
#include "gfObjBezierTriangle.h"
#include "gfObjBezierSPatch.h"
#include "gfObjBezierTetrahedra.h"
#include "gfObjBezierSVolume.h"
#include "gfObjNURBSBox.h"
#include "gfObj3DBoxLattice.h"
#include "gfObjHalfEdgeMesh.h"


// --------------- geometric objects --------------------

gfObjGeomObject* gfCreateGeomObject(gfGeomObjID id)
{
    gfObjGeomObject* obj = NULL;

    if (id == GF_POLYLINE)
        obj = (gfObjGeomObject*) new gfObjPolyLine;
    else if (id == GF_NURBS_CURVE)
        obj = (gfObjGeomObject*) new gfObjNURBSCurve;
    else if (id == GF_NURBS_PATCH)
        obj = (gfObjGeomObject*) new gfObjNURBSPatch;
    else if (id == GF_BEZIER_TRIANGLE)
        obj = (gfObjGeomObject*) new gfObjBezierTriangle;
    else if (id == GF_BEZIER_SPATCH)
        obj = (gfObjGeomObject*) new gfObjBezierSPatch;
    else if (id == GF_BEZIER_TETRAHEDRA)
        obj = (gfObjGeomObject*) new gfObjBezierTetrahedra;
    else if (id == GF_BEZIER_SVOLUME)
        obj = (gfObjGeomObject*) new gfObjBezierSVolume;
    else if (id == GF_NURBS_BOX)
        obj = (gfObjGeomObject*) new gfObjNURBSBox;
    else if (id == GF_3D_BOX_LATTICE)
        obj = (gfObjGeomObject*) new gfObj3DBoxLattice;
    else if (id == GF_HALF_EDGE_MESH)
        obj = (gfObjGeomObject*) new gfObjHalfEdgeMesh;

    m_alloc_geom_objects++;

    return obj;
}

GFRESULT gfDeleteGeomObject(gfObjGeomObject* object)
{
    gfGeomObjID id = object->GetID();

    if (id == GF_POLYLINE)
        delete (gfObjPolyLine*) object;
    else if (id == GF_NURBS_CURVE)
        delete (gfObjNURBSCurve*) object;
    else if (id == GF_NURBS_PATCH)
        delete (gfObjNURBSPatch*) object;
    else if (id == GF_BEZIER_TRIANGLE)
        delete (gfObjBezierTriangle*) object;
    else if (id == GF_BEZIER_SPATCH)
        delete (gfObjBezierSPatch*) object;
    else if (id == GF_BEZIER_TETRAHEDRA)
        delete (gfObjBezierTetrahedra*) object;
    else if (id == GF_BEZIER_SVOLUME)
        delete (gfObjBezierSVolume*) object;
    else if (id == GF_NURBS_BOX)
        delete (gfObjNURBSBox*) object;
    else if (id == GF_3D_BOX_LATTICE)
        delete (gfObj3DBoxLattice*) object;
    else if (id == GF_HALF_EDGE_MESH)
        delete (gfObjHalfEdgeMesh*) object;
 
    m_alloc_geom_objects--;

    return GF_S_OK;
}

// --------------- matrices --------------------

gfMatrix2* gfCreateMatrix2(void)
{
    m_alloc_matrices++;
    gfMatrix2* mat = (gfMatrix2*) new gfMatrix2;
    return mat;
}

GFRESULT gfDeleteMatrix2(gfMatrix2* matrix)
{
    m_alloc_matrices--;
    delete matrix;
    return GF_S_OK;
}

gfMatrix3* gfCreateMatrix3(void)
{
    m_alloc_matrices++;
    gfMatrix3* mat = (gfMatrix3*) new gfMatrix3;
    return mat;
}

GFRESULT gfDeleteMatrix3(gfMatrix3* matrix)
{
    m_alloc_matrices--;
    delete matrix;
    return GF_S_OK;
}

gfMatrix4* gfCreateMatrix4(void)
{
    m_alloc_matrices++;
    gfMatrix4* mat = (gfMatrix4*) new gfMatrix4;
    return mat;
}

GFRESULT gfDeleteMatrix4(gfMatrix4* matrix)
{
    m_alloc_matrices--;
    delete matrix;
    return GF_S_OK;
}

// --------------- points --------------------

gfPoint2* gfCreatePoint2(void)
{
    m_alloc_points++;
    return new gfPoint2;
}

GFRESULT gfDeletePoint2(gfPoint2* point)
{
    m_alloc_points--;
    delete point;
    return GF_S_OK;
}

gfPoint3* gfCreatePoint3(void)
{
    m_alloc_points++;
    return new gfPoint3;
}

GFRESULT gfDeletePoint3(gfPoint3* point)
{
    m_alloc_points--;
    delete point;
    return GF_S_OK;
}

gfPoint4* gfCreatePoint4(void)
{
    m_alloc_points++;
    return new gfPoint4;
}

GFRESULT gfDeletePoint4(gfPoint4* point)
{
    m_alloc_points--;
    delete point;
    return GF_S_OK;
}

// --------------- vectors --------------------

gfVector2* gfCreateVector2(void)
{
    m_alloc_vectors++;
    return new gfVector2;
}

GFRESULT gfDeleteVector2(gfVector2* vector)
{
    m_alloc_vectors--;
    delete vector;
    return GF_S_OK;
}

gfVector3* gfCreateVector3(void)
{
    m_alloc_vectors++;
    gfVector3* new_vector = new gfVector3;
    return new_vector;
}

GFRESULT gfDeleteVector3(gfVector3* vector)
{
    m_alloc_vectors--;
    delete vector;
    return GF_S_OK;
}

gfVector4* gfCreateVector4(void)
{
    m_alloc_vectors++;
    return new gfVector4;
}

GFRESULT gfDeleteVector4(gfVector4* vector)
{
    m_alloc_vectors--;
    delete vector;
    return GF_S_OK;
}

// ------------------- bounding volumes --------------------------

gfAuxBoundingVolume* gfCreateBoundVolume(gfBoundVolID id)
{
    gfAuxBoundingVolume* vol = NULL;

    if (id == GF_BOUND_VOL_AABB)
        vol = (gfAuxAABB*) new gfAuxAABB;
    else if (id == GF_BOUND_VOL_OBB)
        vol = (gfAuxOBB*) new gfAuxOBB;
    else if (id == GF_BOUND_VOL_SPHERE)
        vol = (gfAuxBoundingShere*) new gfAuxBoundingShere;

    m_alloc_bound_volumes++;

    return vol;
}

GFRESULT gfDeleteBoundVolume(gfAuxBoundingVolume* vol)
{
    gfBoundVolID id = vol->GetID();

    if (id == GF_BOUND_VOL_AABB)
        delete (gfAuxAABB*) vol;
    else if (id == GF_BOUND_VOL_OBB)
        delete (gfAuxOBB*) vol;
    else if (id == GF_BOUND_VOL_SPHERE)
        delete (gfAuxBoundingShere*) vol;
 
    m_alloc_bound_volumes--;

    return GF_S_OK;
}

// ----------------------------scene objects ------------------------

gfScnNode* gfCreateSceneNode(void)
{
    m_alloc_scene_nodes++;
    return new gfScnNode;
}

GFRESULT gfDeleteSceneNode(gfScnNode* node)
{
    m_alloc_scene_nodes--;
    delete node;
    return GF_S_OK;
}

gfScnCamera* gfCreateCamera(void)
{
    m_alloc_cameras++;
    return new gfScnCamera;
}

GFRESULT gfDeleteCamera(gfScnCamera* camera)
{
    m_alloc_cameras--;
    delete camera;
    return GF_S_OK;
}

gfScnLight* gfCreateLight(void)
{
    m_alloc_lights++;
    return new gfScnLight;
}

GFRESULT gfDeleteLight(gfScnLight* camera)
{
    m_alloc_lights--;
    delete camera;
    return GF_S_OK;
}

gfScnMaterial* gfCreateMaterial(void)
{
    m_alloc_materials++;
    return new gfScnMaterial;
}

GFRESULT gfDeleteMaterial(gfScnMaterial* material)
{
    m_alloc_materials--;
    delete material;
    return GF_S_OK;
}

// info about intersection of ray with any object

gfRayInterInfo* gfCreateRayInterInfo(void)
{
    m_alloc_ray_infos++;
    gfRayInterInfo* res = new gfRayInterInfo;
    res->m_inters.resize(0);
    res->m_normals.resize(0);
    res->m_uvw.resize(0);

    return res;
}

GFRESULT gfDeleteRayInterInfo(gfRayInterInfo* info)
{
    m_alloc_ray_infos--;
    delete info;
    return GF_S_OK;
}

void gfPrintAllocationStatus(void)
{
    gfLog("==== Currently allocated: ====\n");
    gfLog("Geometric objects: %d\n", m_alloc_geom_objects);
    gfLog("Points: %d\n", m_alloc_points);
    gfLog("Vectors: %d\n", m_alloc_vectors);
    gfLog("Matrices: %d\n", m_alloc_matrices);
    gfLog("Strings: %d\n", m_alloc_strings);
    gfLog("Bounding volumes: %d\n", m_alloc_bound_volumes);
    gfLog("Scene nodes: %d\n", m_alloc_scene_nodes);
    gfLog("Cameras: %d\n", m_alloc_cameras);
    gfLog("Lights: %d\n", m_alloc_lights);
    gfLog("Materials: %d\n", m_alloc_materials);
    gfLog("Ray intersection infos: %d\n", m_alloc_ray_infos);
    gfLog("==============================\n");
}
