//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfScnCamera.h                                        //
//   Description: Scene camera manipulations                    //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfScnCamera.h"
#include "gfAuxNumUtils.h"
#include "gfAuxGeomUtils.h"

#include "OGL/glut.h"


gfScnCamera::gfScnCamera(void)
{
    InitDefaultPerspective();
}

// initialization functions
GFRESULT gfScnCamera::InitDefaultPerspective(void)
{
    m_type = GF_CAMERA_PERSPECTIVE_DEF;
    
    m_left = -0.4;
    m_bottom = -0.3;
    m_right = 0.4;
    m_top = 0.3;
    m_near = 0.5;
    m_far = 1000.0;

    m_target_x = 0.0;
    m_target_y = 0.0;
    m_target_z = 0.0;

    m_fi = GF_PI / 4.0;
    m_xi = GF_PI / 4.0;
    m_dist = 2.0f;

    gfDouble ex = m_dist * gfCos(m_fi) * gfCos(m_xi);
    gfDouble ey = m_dist * gfSin(m_fi) * gfCos(m_xi);
    gfDouble ez = m_dist * gfSin(m_xi);

    /*m_up_x = -ex;
    m_up_y = -ey;
    m_up_z = (ex * ex + ey * ey + ez * ez) / ez - ez;

    gfDouble length = gfSquareRoot(m_up_x * m_up_x + m_up_y * m_up_y + m_up_x * m_up_z);
    if (!gfIsZero(length))
    {
        m_up_x /= length;
        m_up_y /= length;
        m_up_z /= length;
    }*/
    m_up_x = 0.0;
    m_up_y = 0.0;
    m_up_z = 1.0;

    return GF_S_OK;
}

GFRESULT gfScnCamera::InitFrontView(void)
{
    m_type = GF_CAMERA_ORTHO_FRONT;
    
    m_left = m_bottom = -0.1;
    m_right = m_top = 0.1;
    m_near = -1000.0;
    m_far = 1000.0;

    m_target_x = 0.0;
    m_target_y = 0.0;
    m_target_z = 0.0;

    m_fi = 0.0;
    m_xi = 0.0;
    m_dist = 8.0f;

    m_up_x = 0.0;
    m_up_y = 0.0;
    m_up_z = 1.0;

    return GF_S_OK;
}

GFRESULT gfScnCamera::InitRightView(void)
{
    m_type = GF_CAMERA_ORTHO_RIGHT;
    
    m_left = m_bottom = -0.1;
    m_right = m_top = 0.1;
    m_near = -1000.0;
    m_far = 1000.0;

    m_target_x = 0.0;
    m_target_y = 0.0;
    m_target_z = 0.0;

    m_fi = GF_PI / 2.0;
    m_xi = 0.0;
    m_dist = 8.0f;

    m_up_x = 0.0;
    m_up_y = 0.0;
    m_up_z = 1.0;

    return GF_S_OK;
}

GFRESULT gfScnCamera::InitTopView(void)
{
    m_type = GF_CAMERA_ORTHO_TOP;
    
    m_left = m_bottom = -0.1;
    m_right = m_top = 0.1;
    m_near = -1000.0;
    m_far = 1000.0;

    m_target_x = 0.0;
    m_target_y = 0.0;
    m_target_z = 0.0;

    m_fi = 0.0;
    m_xi = GF_PI / 2.0;
    m_dist = 8.0f;

    m_up_x = 0.0;
    m_up_y = 1.0;
    m_up_z = 0.0;

    return GF_S_OK;
}

// OGL support functions
GFRESULT gfScnCamera::SetOGLProjectionMatrix(void)
{
    glMatrixMode(GL_PROJECTION);						
	glLoadIdentity();
    if (m_type == GF_CAMERA_PERSPECTIVE_DEF || m_type == GF_CAMERA_PERSPECTIVE_CUST)
	    glFrustum(m_left, m_right, m_bottom, m_top, m_near, m_far);
    else
        glOrtho(m_dist * m_left, m_dist * m_right, m_dist * m_bottom, m_dist * m_top, m_near, m_far);

    return GF_S_OK;
}

GFRESULT gfScnCamera::SetOGLViewMatrix(void)
{
    gfDouble ex = m_target_x + m_dist * gfCos(m_fi) * gfCos(m_xi);
    gfDouble ey = m_target_y + m_dist * gfSin(m_fi) * gfCos(m_xi);
    gfDouble ez = m_target_z + m_dist * gfSin(m_xi);

    glMatrixMode(GL_MODELVIEW);						
	glLoadIdentity();
    gluLookAt(ex, ey, ez, m_target_x, m_target_y, m_target_z, m_up_x, m_up_y, m_up_z);

    return GF_S_OK;
}

// manipulation
GFRESULT gfScnCamera::Zoom(gfDouble delta)
{
    m_dist += delta;
    return GF_S_OK;
}

GFRESULT gfScnCamera::Pan(gfDouble px, gfDouble py)
{
    gfDouble ex = m_dist * gfCos(m_fi) * gfCos(m_xi);
    gfDouble ey = m_dist * gfSin(m_fi) * gfCos(m_xi);
    gfDouble ez = m_dist * gfSin(m_xi);

    gfDouble tx = m_up_y * ez - m_up_z * ey;
    gfDouble ty = - m_up_x * ez + m_up_z * ex;
    gfDouble tz = m_up_x * ey - m_up_y * ex;

    gfDouble length = gfSquareRoot(tx * tx + ty * ty + tz * tz);
    if (!gfIsZero(length))
    {
        tx /= length;
        ty /= length;
        tz /= length;
    }

    m_target_x += px * tx + py * m_up_x;
    m_target_y += px * ty + py * m_up_y;
    m_target_z += px * tz + py * m_up_z;

    return GF_S_OK;
}

GFRESULT gfScnCamera::RotateHorizontal(gfDouble angle)
{
    m_fi += angle;
    if (m_fi > 2 * GF_PI)
        m_fi = m_fi - 2 * GF_PI;
    if (m_fi < 0.0)
        m_fi = m_fi + 2 * GF_PI;

    return GF_S_OK;
}

GFRESULT gfScnCamera::RotateVertical(gfDouble angle)
{
    m_xi += angle;

    if (m_xi > GF_PI / 2.0)
        m_xi = m_xi - GF_PI / 2.0;
    if (m_xi < -GF_PI / 2.0)
        m_xi = m_xi + GF_PI / 2.0;

    return GF_S_OK;
}

GFRESULT gfScnCamera::SetAspectRatio(gfDouble ratio)
{
    gfDouble width = ratio * (m_top - m_bottom);
    m_left = -width / 2.0;
    m_right = width / 2.0;

    return GF_S_OK;
}

GFRESULT gfScnCamera::GetCameraFrame(gfVector3* dir, gfVector3* up, gfVector3* right)
{
    // camera position
    gfDouble ex = m_target_x + m_dist * gfCos(m_fi) * gfCos(m_xi);
    gfDouble ey = m_target_y + m_dist * gfSin(m_fi) * gfCos(m_xi);
    gfDouble ez = m_target_z + m_dist * gfSin(m_xi);

    dir->Set(m_target_x - ex, m_target_y - ey, m_target_z - ez);
    up->Set(m_up_x, m_up_y, m_up_z);
    gfVectorProduct3(dir, up, right);
    gfVectorProduct3(right, dir, up);

    dir->Normalize();
    up->Normalize();
    right->Normalize();

    return GF_S_OK;
}

GFRESULT gfScnCamera::GetCameraPosition(gfPoint3* pos)
{
    gfDouble ex = m_target_x + m_dist * gfCos(m_fi) * gfCos(m_xi);
    gfDouble ey = m_target_y + m_dist * gfSin(m_fi) * gfCos(m_xi);
    gfDouble ez = m_target_z + m_dist * gfSin(m_xi);

    pos->Set(ex, ey, ez);
    return GF_S_OK;
}