//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfScnGraph.cpp                                       //
//   Description: Scene graph classes and utilities             //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfScnGraph.h"
#include "gfScnCamera.h"
#include "gfObjectFactory.h"
#include "gfObjHalfEdgeMesh.h"
#include "gfAuxNumUtils.h"

#include "OGL/glut.h"


gfScnNode::gfScnNode(void)
{
    m_id = GF_SCENE_EMPTY;
    m_node_matrix = NULL;
    m_bound_vol = NULL;
    m_name.clear();
    m_ogl_flags = GF_FLG_NONE;
    m_data = NULL;
    m_curr_child = 0;
}

gfScnNode::~gfScnNode(void)
{
    Clear();
}

gfScnNode* gfScnNode::GetFirstChild(void)
{
    m_curr_child = 0;
    gfInt size = (gfInt)m_childs.size();

    if (size < 1)
        return NULL;

    m_curr_child++;
    return m_childs.at(0);
}

gfScnNode* gfScnNode::GetNextChild(void)
{
    gfInt size = (gfInt)m_childs.size();
    if (size <= m_curr_child)
        return NULL;

    m_curr_child++;
    return m_childs.at(m_curr_child - 1);
}

gfScnNode* gfScnNode::GetNextChild(gfScnNode* prev)
{
    gfInt size = (gfInt)m_childs.size();
    if (prev == NULL && size > 0)
        return m_childs.at(0);

    for (gfInt i = 0;i < size - 1;i++)
    {
        if (prev == m_childs.at(i))
            return m_childs.at(i + 1);
    }

    return NULL;
}

GFRESULT gfScnNode::AddChild(gfScnNode* child)
{
    child->m_parent = this;
    m_childs.push_back(child);
    return GF_S_OK;
}

GFRESULT gfScnNode::RemoveChild(gfScnNode* child)
{
    gfInt size = (gfInt)m_childs.size();
    for (gfInt i = 0;i < size;i++)
    {
        if (child == m_childs.at(i))
        {
            m_childs.erase(m_childs.begin() + i);
            return GF_S_OK;
        }
    }

    child->Clear();
    gfDeleteSceneNode(child);

    return GF_E_BADARG;
}

GFRESULT gfScnNode::RemoveChildCons(gfScnNode* child)
{
    gfInt size = (gfInt)m_childs.size();
    for (gfInt i = 0;i < size;i++)
    {
        if (child == m_childs.at(i))
        {
            m_childs.erase(m_childs.begin() + i);
            return GF_S_OK;
        }
    }

    return GF_E_BADARG;
}

GFRESULT gfScnNode::RemoveChilds(std::vector<gfScnNode*>* childs)
{
    gfInt size = (gfInt)m_childs.size();
    gfInt sel_size = (gfInt)childs->size();
    for (gfInt i = size - 1;i >= 0;i--)
    {
        for (gfInt j = 0;j < sel_size;j++)
        {
            gfScnNode* child = childs->at(j);
            if (child == m_childs.at(i))
            {
                child->Clear();
                gfDeleteSceneNode(child);
                m_childs.erase(m_childs.begin() + i);
            }
        }
    }

    return GF_E_BADARG;
}

GFRESULT gfScnNode::RemoveChildsCons(std::vector<gfScnNode*>* childs)
{
    gfInt size = (gfInt)m_childs.size();
    gfInt sel_size = (gfInt)childs->size();
    for (gfInt i = size - 1;i >= 0;i--)
    {
        for (gfInt j = 0;j < sel_size;j++)
        {
            if (childs->at(j) == m_childs.at(i))
            {
                m_childs.erase(m_childs.begin() + i);
                break;
            }
        }
    }

    return GF_E_BADARG;
}

GFRESULT gfScnNode::RemoveAllChilds(void)
{
    gfInt size = (gfInt)m_childs.size();
    for (gfInt i = 0;i < size;i++)
    {
        gfScnNode* child = m_childs.at(i);
        if (child)
        {
            child->Clear();
            gfDeleteSceneNode(child);
        }
    }

    m_childs.clear();
    return GF_S_OK;
}

GFRESULT gfScnNode::RemoveAllChildsCons(void)
{
    m_childs.clear();
    return GF_S_OK;
}

GFRESULT gfScnNode::Clear(void)
{
    // delete associated matrix
    if (m_node_matrix)
        gfDeleteMatrix4(m_node_matrix);
    m_node_matrix = NULL;

    // delete associated bounding volume
    if (m_bound_vol)
        gfDeleteBoundVolume(m_bound_vol);
    m_bound_vol = NULL;

    // delete associated object
    if (m_data)
    {
        gfObjGeomObject* obj;
        gfScnCamera* cam;
        gfScnLight* light;
        gfScnMaterial* mat;

        switch (m_id)
        {
        case GF_SCENE_GEOM_OBJECT:
            obj = (gfObjGeomObject*) m_data;
            if (obj)
                gfDeleteGeomObject(obj);
            break;
        case GF_SCENE_CONTROL_OBJECT_GROUP:
            obj = (gfObjGeomObject*) m_data;
            if (obj)
                gfDeleteGeomObject(obj);
            break;
        case GF_SCENE_CAMERA:
            cam = (gfScnCamera*) m_data;
            if (cam)
                gfDeleteCamera(cam);
            break;
        case GF_SCENE_LIGHT:
            light = (gfScnLight*) m_data;
            if (light)
                gfDeleteLight(light);
            break;
        case GF_SCENE_MATERIAL:
            mat = (gfScnMaterial*) m_data;
            if (mat)
                gfDeleteMaterial(mat);
            break;
        }
    }
    m_data = NULL;

    //m_childs.clear();

    return GF_S_OK;
}

GFRESULT gfScnNode::OGLDraw(gfInt flags)
{
    if (m_id == GF_SCENE_GEOM_OBJECT)
    {
        gfObjGeomObject* obj = (gfObjGeomObject*) m_data;
        if (!obj)
            return GF_E_FAIL;

        //gfInt flg = m_ogl_flags & flags & ~GF_FLG_WIRE;
        //glColor3f(m_color_r, m_color_g, m_color_b);
        //obj->OGLDraw(m_ogl_flags & flags & ~GF_FLG_WIRE);

        // draw solid
        obj->OGLDraw(flags & ~(GF_FLG_WIRE | GF_FLG_CONTROL_NET));

        GLfloat color[4] = {0.0, 0.0, 0.0, 0.0};
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, color);
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, color);

        // draw lines
        obj->OGLDraw(flags & ~GF_FLG_SOLID);

        // draw bounding box
        if (((m_ogl_flags & GF_FLG_BOUND_VOLUME) || (flags & GF_FLG_BOUND_VOLUME)) && m_bound_vol)
        {
            glLineWidth(3.0);
            glNormal3f(0.0, 0.0, 0.0);
            m_bound_vol->OGLDraw(GF_FLG_WIRE);
            glLineWidth(1.0);
        }
        //glDisable(GL_COLOR_MATERIAL);
    }
    else if (m_id == GF_SCENE_CONTROL_OBJECT_GROUP || m_id == GF_SCENE_GROUP)
    {
        if (((m_ogl_flags & GF_FLG_BOUND_VOLUME) || (flags & GF_FLG_BOUND_VOLUME)) && m_bound_vol)
        {
            glEnable(GL_COLOR_MATERIAL);
            glColor3f(0.0f, 0.0f, 0.0f);
            glLineWidth(3.0);
            m_bound_vol->OGLDraw(GF_FLG_WIRE);
            glLineWidth(1.0);
            glDisable(GL_COLOR_MATERIAL);
        }
    }

    return GF_S_OK;
}

GFRESULT gfScnNode::UpdateBoundVolume()
{
    if (m_bound_vol)
        gfDeleteBoundVolume(m_bound_vol);
    m_bound_vol = NULL;

    gfObjGeomObject* obj = (gfObjGeomObject*) m_data;
    m_bound_vol = obj->GetBoundVolume(GF_BOUND_VOL_AABB, m_node_matrix);

    return GF_S_OK;
}

// ---------------- scene graph ---------------------------


gfScnGraph::gfScnGraph(void)
{
    // create root node
    m_root = gfCreateSceneNode();

    m_root->m_id = GF_SCENE;

    m_root->m_data = NULL;
    m_root->m_parent = NULL;

    m_root->m_bound_vol = gfCreateBoundVolume(GF_BOUND_VOL_AABB);
    m_root->m_ogl_flags = GF_FLG_NONE;
    m_root->m_name.append("Scene");

    m_root->m_node_matrix = gfCreateMatrix4();
    m_root->m_node_matrix->LoadIdentity();
}

gfScnGraph::~gfScnGraph(void)
{
    DeleteSubtree(m_root, true, true);
    gfDeleteSceneNode(m_root);
    m_root = NULL;
}

GFRESULT gfScnGraph::Clear(void)
{
    ClearObjectsSelection();
    return DeleteSubtree(m_root, true, true);
}

GFRESULT gfScnGraph::InitDefaultScene()
{
    // create four default cameras
    gfScnCamera* cam_perspective = gfCreateCamera();
    gfScnCamera* cam_ortho_front = gfCreateCamera();
    gfScnCamera* cam_ortho_right = gfCreateCamera();
    gfScnCamera* cam_ortho_top = gfCreateCamera();

    cam_perspective->InitDefaultPerspective();
    cam_ortho_front->InitFrontView();
    cam_ortho_right->InitRightView();
    cam_ortho_top->InitTopView();

    AddCamera(cam_perspective);
    AddCamera(cam_ortho_front);
    AddCamera(cam_ortho_right);
    AddCamera(cam_ortho_top);

    // create light
    gfScnLight* light = gfCreateLight();
    /*light->InitDefaultLight();
    light->m_pos_x = -3.0;
    light->m_pos_y = -3.0;
    light->m_pos_z = 0.0;
    AddLight(light);
    light = gfCreateLight();*/
    light->InitDefaultLight();
    light->m_pos_x = 3.0;
    light->m_pos_y = 3.0;
    light->m_pos_z = 3.0;
    AddLight(light);

    // set background color
    m_background_r = 0.6;
    m_background_g = 0.6;
    m_background_b = 0.6;

    // set global ambient color
    m_global_ambient_r = 0.0;
    m_global_ambient_g = 0.0;
    m_global_ambient_b = 0.0;

    return GF_S_OK;
}

GFRESULT gfScnGraph::DeleteSubtree(gfScnNode* node, gfBool delete_root, gfBool remove_con_to_parent)
{
    // delete child subtrees
    for (gfScnNode* child = node->GetFirstChild(); child; child = node->GetNextChild())
    {
        DeleteSubtree(child, true, false);
    }
    node->RemoveAllChildsCons();

    // delete root node
    if (delete_root)
    {
        if (node->m_parent && remove_con_to_parent)
            node->m_parent->RemoveChildCons(node);
        node->Clear();
        RemoveObjectFromSelection(node);
        if (node != m_root)
            gfDeleteSceneNode(node);
    }

    return GF_S_OK;
}

GFRESULT gfScnGraph::DrawScene(gfInt flags)
{
    glClearColor(m_background_r, m_background_g, m_background_b, 1.0);

    DrawTree(m_root, flags);

    return GF_S_OK;
}

GFRESULT gfScnGraph::DrawTree(gfScnNode* root, gfInt flags)
{
    // draw root node
    if (!root)
        return GF_E_BADARG;

    // set material    
    std::vector<gfScnMaterial*> materials;
    GetNodeMaterials(root, &materials);
    gfInt mat_num = (gfInt)materials.size();
    for (gfInt i = 0;i < mat_num;i++)
        materials.at(i)->SetAsOGLMaterial();

    

    //draw bounding volume around if object is in selection
    if (IsObjectInSelection(root))
    {
        root->OGLDraw(flags | GF_FLG_BOUND_VOLUME);
    }
    else
    {
        root->OGLDraw(flags);
    }

    for (gfScnNode* node = root->GetFirstChild(); node; node = root->GetNextChild())
    {
        DrawTree(node, flags);
    }

    return GF_S_OK;
}

gfScnNode* gfScnGraph::AddGeomObject(gfScnNode* parent, gfMatrix4* obj_matrix, gfObjGeomObject* object)
{
    if (!object)
        return NULL;
    gfScnNode* par = m_root;
    if (parent)
        par = parent;

    // add as par child
    gfScnNode* node = gfCreateSceneNode();

    node->m_id = GF_SCENE_GEOM_OBJECT;
    node->m_data = object;
    node->m_bound_vol = object->GetBoundVolume(GF_BOUND_VOL_AABB, obj_matrix);
    node->m_ogl_flags = GF_FLG_SOLID | GF_FLG_WIRE | GF_FLG_NORMALS;
    node->m_name.append("Object");
    node->m_node_matrix = obj_matrix;

    // create random material for this
    gfScnMaterial* mat = gfCreateMaterial();
    mat->InitDefaultMaterial();
    mat->m_ambient_r = mat->m_diffuse_r = gfRand(1.0);
    mat->m_ambient_g = mat->m_diffuse_g = gfRand(1.0);
    mat->m_ambient_b = mat->m_diffuse_b = gfRand(1.0);
    AddMaterial(node, mat);

    par->AddChild(node);

    return node;
}

gfScnNode* gfScnGraph::AddCamera(gfScnCamera* camera)
{ 
    if (!camera)
        return NULL;
    // add as root child
    gfScnNode* node = gfCreateSceneNode();

    node->m_id = GF_SCENE_CAMERA;
    node->m_data = camera;
    node->m_bound_vol = NULL;
    node->m_ogl_flags = GF_FLG_NONE;
    node->m_name.append("Camera");
    node->m_node_matrix = NULL;

    m_root->AddChild(node);

    return node;
}

gfScnNode* gfScnGraph::AddLight(gfScnLight* light)
{
    if (!light)
        return NULL;
    gfScnNode* node;

    // get num of lights
    gfInt num = 0;
    for (node = m_root->GetFirstChild(); node; node = m_root->GetNextChild())
    {
        if (node->m_id == GF_SCENE_LIGHT)
        {
            gfScnLight* light = (gfScnLight*) node->m_data;
            if (light)
                num++;
        }
    }

    if ((num + 1) > GF_MAX_NUM_LIGHTS)
    {
        gfDeleteLight(light);
        return NULL;
    }
    light->SetAsOGLLight(GL_LIGHT0 + num);

    // add as root child
    node = gfCreateSceneNode();

    node->m_id = GF_SCENE_LIGHT;
    node->m_data = light;
    node->m_bound_vol = NULL;
    node->m_ogl_flags = GF_FLG_NONE;
    node->m_name.append("Light");
    node->m_node_matrix = NULL;

    m_root->AddChild(node);

    return node;
}

gfScnNode* gfScnGraph::AddMaterial(gfScnNode* root_node, gfScnMaterial* material)
{
    if (!material)
        return NULL;

    // add as root child
    gfScnNode* node = gfCreateSceneNode();

    node->m_id = GF_SCENE_MATERIAL;
    node->m_data = material;
    node->m_bound_vol = NULL;
    node->m_ogl_flags = GF_FLG_NONE;
    node->m_name.append("Material");
    node->m_node_matrix = NULL;

    root_node->AddChild(node);

    return node;
}

GFRESULT gfScnGraph::RemoveAllMaterials(gfScnNode* root)
{
    std::vector<gfScnNode*> materials;
    for (gfScnNode* node = root->GetFirstChild(); node; node = root->GetNextChild())
    {
        if (node->m_id == GF_SCENE_MATERIAL)
        {
            materials.push_back(node);
        }
    }

    gfInt num = (gfInt)materials.size();
    for (gfInt i = 0;i < num;i++)
    {
        gfScnNode* node = materials.at(i);
        DeleteSubtree(node, true, true);
    }

    return GF_S_OK;
}

gfScnCamera* gfScnGraph::GetDefPerspectiveCam(void)
{
    // search in root childs
    for (gfScnNode* node = m_root->GetFirstChild(); node; node = m_root->GetNextChild())
    {
        if (node->m_id == GF_SCENE_CAMERA)
        {
            gfScnCamera* cam = (gfScnCamera*) node->m_data;
            if (cam && cam->m_type == GF_CAMERA_PERSPECTIVE_DEF)
                return cam;
        }
    }

    return NULL;
}

gfScnCamera* gfScnGraph::GetDefOrthoFrontCam(void)
{
    // search in root childs
    for (gfScnNode* node = m_root->GetFirstChild(); node; node = m_root->GetNextChild())
    {
        if (node->m_id == GF_SCENE_CAMERA)
        {
            gfScnCamera* cam = (gfScnCamera*) node->m_data;
            if (cam && cam->m_type == GF_CAMERA_ORTHO_FRONT)
                return cam;
        }
    }

    return NULL;
}

gfScnCamera* gfScnGraph::GetDefOrthoRightCam(void)
{
    // search in root childs
    for (gfScnNode* node = m_root->GetFirstChild(); node; node = m_root->GetNextChild())
    {
        if (node->m_id == GF_SCENE_CAMERA)
        {
            gfScnCamera* cam = (gfScnCamera*) node->m_data;
            if (cam && cam->m_type == GF_CAMERA_ORTHO_RIGHT)
                return cam;
        }
    }

    return NULL;
}

gfScnCamera* gfScnGraph::GetDefOrthoTopCam(void)
{
    // search in root childs
    for (gfScnNode* node = m_root->GetFirstChild(); node; node = m_root->GetNextChild())
    {
        if (node->m_id == GF_SCENE_CAMERA)
        {
            gfScnCamera* cam = (gfScnCamera*) node->m_data;
            if (cam && cam->m_type == GF_CAMERA_ORTHO_TOP)
                return cam;
        }
    }

    return NULL;
}

GFRESULT gfScnGraph::GetSceneLights(std::vector<gfScnLight*>* lights)
{
    lights->clear();

    // search in root childs
    for (gfScnNode* node = m_root->GetFirstChild(); node; node = m_root->GetNextChild())
    {
        if (node->m_id == GF_SCENE_LIGHT)
        {
            gfScnLight* light = (gfScnLight*) node->m_data;
            if (light)
                lights->push_back(light);
        }
    }

    return GF_S_OK;
}

GFRESULT gfScnGraph::GetAllVisibleLights(gfPoint3* pos, std::vector<gfScnLight*>* lights)
{
    lights->clear();

    // search in root childs
    for (gfScnNode* node = m_root->GetFirstChild(); node; node = m_root->GetNextChild(node))
    {
        if (node->m_id == GF_SCENE_LIGHT)
        {
            gfScnLight* light = (gfScnLight*) node->m_data;
            if (light)
            {
                // check visibility of light from pos
                gfVector3 dir(light->m_pos_x - pos->X(), light->m_pos_y - pos->Y(), light->m_pos_z - pos->Z());
                gfDouble t_light = dir.Length();
                dir.Normalize();
                gfDouble t;
                gfScnNode* pick = PickNodeByRay(pos, &dir, &t);
                if (!pick || (pick && (t > t_light)))
                    lights->push_back(light);
            }
        }
    }

    return GF_S_OK;
}

GFRESULT gfScnGraph::GetNodeMaterials(gfScnNode* node, std::vector<gfScnMaterial*>* materials)
{
    if (!node)
        return GF_E_BADARG;
    materials->clear();

    // search in node childs
    for (gfScnNode* child = node->GetFirstChild(); child; child = node->GetNextChild())
    {
        if (child->m_id == GF_SCENE_MATERIAL)
        {
            gfScnMaterial* material = (gfScnMaterial*) child->m_data;
            if (material)
                materials->push_back(material);
        }
    }

    return GF_S_OK;
}

GFRESULT gfScnGraph::ClearObjectsSelection(void)
{
    m_selected_objects.clear();
    return GF_S_OK;
}

GFRESULT gfScnGraph::AddObjectToSelection(gfScnNode* node)
{
    m_selected_objects.push_back(node);
    return GF_S_OK;
}

GFRESULT gfScnGraph::RemoveObjectFromSelection(gfScnNode* node)
{
    gfInt num_sel_objects = (gfInt)m_selected_objects.size();
    for (gfInt i = 0;i < num_sel_objects;i++)
    {
        if (m_selected_objects.at(i) == node)
        {
            m_selected_objects.erase(m_selected_objects.begin() + i);
            return GF_S_OK;
        }
    }

    return GF_E_BADARG;
}

gfInt gfScnGraph::GetNumSelectedObjects(void)
{
    return (gfInt)m_selected_objects.size();
}

gfScnNode* gfScnGraph::GetSelectedObject(gfInt index)
{
    gfInt num_sel_objects = (gfInt)m_selected_objects.size();
    if (index >= 0 && index < num_sel_objects)
        return m_selected_objects.at(index);

    return NULL;
}

gfBool gfScnGraph::IsObjectInSelection(gfScnNode* node)
{
    gfInt num_sel_objects = (gfInt)m_selected_objects.size();
    for (gfInt i = 0;i < num_sel_objects;i++)
    {
        if (m_selected_objects.at(i) == node)
        {
            return true;
        }
    }

    return false;
}

gfScnNode* gfScnGraph::PickNodeByRay(gfPoint3* ray_begin, gfVector3* ray_dir, gfDouble* t_min)
{
    std::vector<gfScnNode*> nodes;
    
    gfDouble min_t = GF_MAX_DOUBLE;
    gfScnNode* nearest_node = NULL;

    nodes.push_back(m_root);
    gfInt size = (gfInt)nodes.size();
    while (size != 0)
    {
        gfScnNode* node = nodes.at(0);
        nodes.erase(nodes.begin());
        size = (gfInt)nodes.size();

        if (!node)
            continue;

        if (node->m_bound_vol)
        {
            gfDouble tmin, tmax;
            if (GF_SUCC(node->m_bound_vol->RayIntersection(ray_begin, ray_dir, &tmin, &tmax)))
            {
                if (node->m_id == GF_SCENE_GEOM_OBJECT)                        
                {
                    gfObjGeomObject* obj = (gfObjGeomObject*) node->m_data;
                    gfRayInterInfo* info = obj->RayIntersection(node->m_node_matrix, ray_begin, ray_dir);

                    if (info)
                    {
                        if (info->m_inters.at(0) < min_t)
                        {
                            min_t = info->m_inters.at(0);
                            nearest_node = node;
                        }

                        gfDeleteRayInterInfo(info);
                    }
                }
                else if (node->m_id == GF_SCENE_GROUP ||
                         node->m_id == GF_SCENE_CONTROL_OBJECT_GROUP)
                {
                    // find intersection of bounding volume and ray
                    // for now, group object is picked only if there is no object before him
                    if (min_t > 10.0)
                    {
                        min_t = 10.0;
                        nearest_node = node;
                    }
                }
            }
        }

        // add child to queue
        if (node->m_id != GF_SCENE_GROUP && node->m_id != GF_SCENE_CONTROL_OBJECT_GROUP)
        {
            for (gfScnNode* child = node->GetFirstChild(); child; child = node->GetNextChild())
            {
                nodes.push_back(child);
            }
        }

        size = (gfInt)nodes.size();
    }

    // find object with nearest intersected point
    *t_min = min_t;
    return nearest_node;
}

GFRESULT gfScnGraph::TransformNode(gfScnNode* node, gfMatrix4* matrix)
{
    if (!node)
        return GF_E_BADARG;

    // transform
    if (node->m_id == GF_SCENE_GEOM_OBJECT || node->m_id == GF_SCENE_CONTROL_OBJECT_GROUP)
    {
        gfObjGeomObject* obj = (gfObjGeomObject*) node->m_data;
        obj->Transform(matrix);

        // update bounding box
        if (node->m_bound_vol)
        {
            if (node->m_bound_vol->GetID() == GF_BOUND_VOL_AABB)
            {
                // for AABB, must be recomputed
                gfDeleteBoundVolume(node->m_bound_vol);
                node->m_bound_vol = obj->GetBoundVolume(GF_BOUND_VOL_AABB, NULL);
            }
            else
                node->m_bound_vol->Transform(matrix);
        }
    }

    // transform also childs
    for (gfScnNode* child = node->GetFirstChild(); child; child = node->GetNextChild())
    {
        TransformNode(child, matrix);
    }

    // TODO: update bounding volumes in parent nodes

    return GF_S_OK;
}