//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfScnGraph.h                                         //
//   Description: Scene graph classes and utilities             //
//                                                              //
//////////////////////////////////////////////////////////////////


#ifndef _GF_SCN_GRAPH_H_
#define _GF_SCN_GRAPH_H_


#include "gfDefines.h"
#include <vector>
#include "gfMatrix.h"
#include "gfObjGeomObject.h"
#include <string>
#include "gfScnCamera.h"
#include "gfScnLight.h"
#include "gfScnMaterial.h"


class gfScnNode
{
private:
    std::vector<gfScnNode*> m_childs;
    gfInt m_curr_child;

public:
    gfSceneNodeID m_id;
    gfMatrix4* m_node_matrix;
    gfAuxBoundingVolume* m_bound_vol;
    std::string m_name;
    gfLong m_ogl_flags;
    void* m_data;
    gfScnNode* m_parent;

    gfScnNode(void);
    ~gfScnNode(void);

    gfScnNode* GetFirstChild(void);
    gfScnNode* GetNextChild(void);
    gfScnNode* GetNextChild(gfScnNode* prev);
    GFRESULT AddChild(gfScnNode* child);
    GFRESULT RemoveChild(gfScnNode* child);
    GFRESULT RemoveChildCons(gfScnNode* child);
    GFRESULT RemoveChilds(std::vector<gfScnNode*>* childs);
    GFRESULT RemoveChildsCons(std::vector<gfScnNode*>* childs);
    GFRESULT RemoveAllChilds(void);
    GFRESULT RemoveAllChildsCons(void);

    GFRESULT Clear(void);

    GFRESULT UpdateBoundVolume();
    GFRESULT OGLDraw(gfInt flags);
};


class gfScnGraph
{
private:
    std::vector<gfScnNode*> m_selected_objects;

public:

    gfScnNode* m_root;
    gfDouble m_background_r;
    gfDouble m_background_g;
    gfDouble m_background_b;

    gfDouble m_global_ambient_r;
    gfDouble m_global_ambient_g;
    gfDouble m_global_ambient_b;

    gfScnGraph(void);
    ~gfScnGraph(void);

    GFRESULT Clear(void);
    GFRESULT InitDefaultScene();
    GFRESULT DeleteSubtree(gfScnNode* node, gfBool delete_root, gfBool remove_con_to_parent);

    // using OGL
    GFRESULT SetupToOGL(void);
    GFRESULT DrawScene(gfInt flags);
    GFRESULT DrawTree(gfScnNode* root, gfInt flags);

    // adding nodes to graph
    gfScnNode* AddGeomObject(gfScnNode* parent, gfMatrix4* obj_matrix, gfObjGeomObject* object);
    gfScnNode* AddCamera(gfScnCamera* camera);
    gfScnNode* AddLight(gfScnLight* light);
    gfScnNode* AddMaterial(gfScnNode* root_node, gfScnMaterial* material);

    GFRESULT RemoveAllMaterials(gfScnNode* root);

    // searching utilities
    gfBool IsNodeInGraph(gfScnNode* node);
    gfScnCamera* GetDefPerspectiveCam(void);
    gfScnCamera* GetDefOrthoFrontCam(void);
    gfScnCamera* GetDefOrthoRightCam(void);
    gfScnCamera* GetDefOrthoTopCam(void);
    GFRESULT GetSceneLights(std::vector<gfScnLight*>* lights);
    GFRESULT GetAllVisibleLights(gfPoint3* pos, std::vector<gfScnLight*>* lights);
    GFRESULT GetNodeMaterials(gfScnNode* node, std::vector<gfScnMaterial*>* materials);

    // work with objects selection
    GFRESULT ClearObjectsSelection(void);
    GFRESULT AddObjectToSelection(gfScnNode* node);
    GFRESULT RemoveObjectFromSelection(gfScnNode* node);
    gfInt GetNumSelectedObjects(void);
    gfScnNode* GetSelectedObject(gfInt index);
    gfBool IsObjectInSelection(gfScnNode* node);

    gfScnNode* PickNodeByRay(gfPoint3* ray_begin, gfVector3* ray_dir, gfDouble* t_min);
    GFRESULT TransformNode(gfScnNode* node, gfMatrix4* matrix);
};

#endif