//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfScnLight.h                                         //
//   Description: Scene camera manipulations                    //
//                                                              //
//////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "gfScnLight.h"
#include "gfAuxNumUtils.h"

#include "OGL/glut.h"


gfScnLight::gfScnLight(void)
{
    InitDefaultLight();
}

GFRESULT gfScnLight::InitDefaultLight(void)
{
    m_pos_x = -3.0;
    m_pos_y = -3.0;
    m_pos_z = 3.0;
    m_pos_w = 0.0;

    m_ambient_r = 0.4;
    m_ambient_g = 0.4;
    m_ambient_b = 0.4;
    m_ambient_a = 1.0;

    m_diffuse_r = 0.8;
    m_diffuse_g = 0.8;
    m_diffuse_b = 0.8;
    m_diffuse_a = 1.0;

    m_specular_r = 1.0;
    m_specular_g = 1.0;
    m_specular_b = 1.0;
    m_specular_a = 1.0;
    
    m_spot_dir_x = 0.0;
    m_spot_dir_y = 0.0;
    m_spot_dir_z = -1.0;
    m_spot_exp = 0.0;
    m_spot_cutoff = GF_PI;

    m_att_const = 1.0;
    m_att_lin = 0.0;
    m_att_quad = 0.0;

    return GF_S_OK;
}

GFRESULT gfScnLight::SetAmbient(gfDouble r, gfDouble g, gfDouble b, gfDouble a)
{
    m_ambient_r = r;
    m_ambient_g = g;
    m_ambient_b = b;
    m_ambient_a = a;
    return GF_S_OK;
}

GFRESULT gfScnLight::SetDiffuse(gfDouble r, gfDouble g, gfDouble b, gfDouble a)
{
    m_diffuse_r = r;
    m_diffuse_g = g;
    m_diffuse_b = b;
    m_diffuse_a = a;
    return GF_S_OK;
}

GFRESULT gfScnLight::SetSpecular(gfDouble r, gfDouble g, gfDouble b, gfDouble a)
{
    m_specular_r = r;
    m_specular_g = g;
    m_specular_b = b;
    m_specular_a = a;
    return GF_S_OK;
}
GFRESULT gfScnLight::SetPosition(gfDouble x, gfDouble y, gfDouble z, gfDouble w)
{
    m_pos_x = x;
    m_pos_y = y;
    m_pos_z = z;
    m_pos_w = w;
    return GF_S_OK;
}
GFRESULT gfScnLight::SetSpotLight(gfDouble dir_x, gfDouble dir_y, gfDouble dir_z, gfDouble exp, gfDouble cutoff)
{
    m_spot_dir_x = dir_x;
    m_spot_dir_y = dir_y;
    m_spot_dir_z = dir_z;
    m_spot_exp = exp;
    m_spot_cutoff = cutoff;
    return GF_S_OK;
}

GFRESULT gfScnLight::SetAttenuation(gfDouble constant, gfDouble linear, gfDouble quadratic)
{
    m_att_const = constant;
    m_att_lin = linear;
    m_att_quad = quadratic;
    return GF_S_OK;
}

GFRESULT gfScnLight::Enable(gfBool state)
{
    m_state = state;
    if (state)
        glEnable(m_ogl_id);
    else
        glDisable(m_ogl_id);
    return GF_S_OK;
}

GFRESULT gfScnLight::SetAsOGLLight(gfInt light_id)
{
    m_ogl_id = light_id;
    GLfloat aux_flt[4];

    // ambient color
    aux_flt[0] = (GLfloat) m_ambient_r;
    aux_flt[1] = (GLfloat) m_ambient_g;
    aux_flt[2] = (GLfloat) m_ambient_b;
    aux_flt[3] = (GLfloat) m_ambient_a;
    glLightfv(light_id, GL_AMBIENT, (GLfloat*)aux_flt);

    // diffuse color
    aux_flt[0] = (GLfloat) m_diffuse_r;
    aux_flt[1] = (GLfloat) m_diffuse_g;
    aux_flt[2] = (GLfloat) m_diffuse_b;
    aux_flt[3] = (GLfloat) m_diffuse_a;
    glLightfv(light_id, GL_DIFFUSE, (GLfloat*)aux_flt);

    // specular color
    aux_flt[0] = (GLfloat) m_specular_r;
    aux_flt[1] = (GLfloat) m_specular_g;
    aux_flt[2] = (GLfloat) m_specular_b;
    aux_flt[3] = (GLfloat) m_specular_a;
    glLightfv(light_id, GL_SPECULAR, (GLfloat*)aux_flt);

    // light position
    aux_flt[0] = (GLfloat) m_pos_x;
    aux_flt[1] = (GLfloat) m_pos_y;
    aux_flt[2] = (GLfloat) m_pos_z;
    aux_flt[3] = (GLfloat) m_pos_w;
    glLightfv(light_id, GL_POSITION, (GLfloat*)aux_flt);

    // spot light
    aux_flt[0] = (GLfloat) m_spot_dir_x;
    aux_flt[1] = (GLfloat) m_spot_dir_y;
    aux_flt[2] = (GLfloat) m_spot_dir_z;
    glLightfv(light_id, GL_SPOT_DIRECTION, (GLfloat*)aux_flt);
    glLightf(light_id, GL_SPOT_EXPONENT, (GLfloat) m_spot_exp);
    glLightf(light_id, GL_SPOT_CUTOFF, (GLfloat) m_spot_cutoff);

    // attenuation
    glLightf(light_id, GL_CONSTANT_ATTENUATION, (GLfloat) m_att_const);
    glLightf(light_id, GL_LINEAR_ATTENUATION, (GLfloat) m_att_lin);
    glLightf(light_id, GL_QUADRATIC_ATTENUATION, (GLfloat) m_att_quad);

    // enable this light
    Enable(true);

    return GF_S_OK;
}

GFRESULT gfScnLight::SetPosToOGL(void)
{
    GLfloat aux_flt[4];

    aux_flt[0] = (GLfloat) m_pos_x;
    aux_flt[1] = (GLfloat) m_pos_y;
    aux_flt[2] = (GLfloat) m_pos_z;
    aux_flt[3] = (GLfloat) m_pos_w;
    glLightfv(m_ogl_id, GL_POSITION, (GLfloat*)aux_flt);

    return GF_S_OK;
}

gfInt gfScnLight::GetOGLLightId(void)
{
    return m_ogl_id;
}