//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfVector.cpp                                        //
//   Description: Class for work with 3D homogenous vectors     //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "gfVector.h"
#include "gfAuxNumUtils.h"
#include "gfObjectFactory.h"


// ---------------------- 2D vector --------------------------------
gfVector2::gfVector2()
{
    m_data[0] = 0.0;
    m_data[1] = 0.0;
}

gfVector2::gfVector2(gfDouble x, gfDouble y)
{
    m_data[0] = x;
    m_data[1] = y;
}

gfVector2::~gfVector2()
{
}

gfVector2* gfVector2::Clone()
{
    gfVector2* new_vector = gfCreateVector2();
    new_vector->Set(m_data[0], m_data[1]);
    return new_vector;
}

GFRESULT gfVector2::Set(gfDouble x, gfDouble y)
{
    m_data[0] = x;
    m_data[1] = y;

    return GF_S_OK;
}

GFRESULT gfVector2::Set(gfInt index, gfDouble value)
{
    if (index < 1 || index > 3)
        return GF_E_BADARG;

    m_data[index - 1] = value;

    return GF_S_OK;
}

gfDouble gfVector2::at(gfInt index)
{
    if (index < 1 || index > 2)
        return 0.0;

    return m_data[index - 1];
}

gfDouble gfVector2::X(void)
{
    return m_data[0];
}

gfDouble gfVector2::Y(void)
{
    return m_data[1];
}

GFRESULT gfVector2::SetX(gfDouble x)
{
    m_data[0] = x;
    return GF_S_OK;
}

GFRESULT gfVector2::SetY(gfDouble y)
{
    m_data[1] = y;
    return GF_S_OK;
}

gfDouble* gfVector2::GetData(void)
{
    return (gfDouble*) m_data;
}

gfDouble gfVector2::Length(void)
{
    return gfSquareRoot(m_data[0] * m_data[0] + m_data[1] * m_data[1]);
}

GFRESULT gfVector2::SetLength(gfDouble new_length)
{
    if (gfIsZero(new_length))
        return GF_E_BADARG;

    gfDouble d = Length();
    if (gfIsZero(d))
        return GF_S_FALSE;

    m_data[0] *= new_length / d;
    m_data[1] *= new_length / d;

    return GF_S_OK;
}

GFRESULT gfVector2::Normalize(void)
{
    gfDouble d = Length();
    if (!gfIsZero(d))
    {
        m_data[0] /= d;
        m_data[1] /= d;
    }

    return GF_S_OK;
}

GFRESULT gfVector2::Rotate(gfDouble alpha)
{
    return GF_E_NOTIMPL;
}

GFRESULT gfVector2::Translate(gfDouble tx, gfDouble ty)
{
    m_data[0] += tx;
    m_data[1] += ty;

    return GF_S_OK;
}

// ---------------------- 3D vector --------------------------------
gfVector3::gfVector3()
{
    m_data[0] = 0.0;
    m_data[1] = 0.0;
    m_data[2] = 0.0;
}

gfVector3::gfVector3(gfDouble x, gfDouble y, gfDouble z)
{
    m_data[0] = x;
    m_data[1] = y;
    m_data[2] = z;
}

gfVector3::~gfVector3(void)
{
}

gfVector3* gfVector3::Clone(void)
{
    gfVector3* new_vector = gfCreateVector3();
    new_vector->Set(m_data[0], m_data[1], m_data[2]);
    return new_vector;
}

GFRESULT gfVector3::Set(gfDouble x, gfDouble y, gfDouble z)
{
    m_data[0] = x;
    m_data[1] = y;
    m_data[2] = z;

    return GF_S_OK;
}

GFRESULT gfVector3::Set(gfInt index, gfDouble value)
{
    if (index < 1 || index > 3)
        return GF_E_BADARG;

    m_data[index - 1] = value;

    return GF_S_OK;
}

gfDouble gfVector3::at(gfInt index)
{
    if (index < 1 || index > 3)
        return 0.0;

    return m_data[index - 1];
}

gfDouble gfVector3::X(void)
{
    return m_data[0];
}

gfDouble gfVector3::Y(void)
{
    return m_data[1];
}

gfDouble gfVector3::Z(void)
{
    return m_data[2];
}

GFRESULT gfVector3::SetX(gfDouble x)
{
    m_data[0] = x;
    return GF_S_OK;
}

GFRESULT gfVector3::SetY(gfDouble y)
{
    m_data[1] = y;
    return GF_S_OK;
}

GFRESULT gfVector3::SetZ(gfDouble z)
{
    m_data[2] = z;
    return GF_S_OK;
}

gfDouble* gfVector3::GetData(void)
{
    return (gfDouble*) m_data;
}

gfDouble gfVector3::Length(void)
{
    return gfSquareRoot(m_data[0] * m_data[0] + m_data[1] * m_data[1] + m_data[2] * m_data[2]);
}

GFRESULT gfVector3::SetLength(gfDouble new_length)
{
    if (gfIsZero(new_length))
        return GF_E_BADARG;

    gfDouble d = Length();
    if (gfIsZero(d))
        return GF_S_FALSE;

    m_data[0] *= new_length / d;
    m_data[1] *= new_length / d;
    m_data[2] *= new_length / d;

    return GF_S_OK;
}

GFRESULT gfVector3::Normalize(void)
{
    gfDouble d = Length();
    if (!gfIsZero(d) && !gfIsZero(d-1.0))
    {
        gfDouble inv = 1/d;
        m_data[0] *= inv;
        m_data[1] *= inv;
        m_data[2] *= inv;
    }

    return GF_S_OK;
}

GFRESULT gfVector3::Rotate(gfVector3*axis, gfDouble alpha)
{
    return GF_E_NOTIMPL;
}

GFRESULT gfVector3::Translate(gfDouble tx, gfDouble ty, gfDouble tz)
{
    m_data[0] += tx;
    m_data[1] += ty;
    m_data[2] += tz;

    return GF_S_OK;
}

GFRESULT gfVector3::Flip(void)
{
    m_data[0] *= -1;
    m_data[1] *= -1;
    m_data[2] *= -1;

    return GF_S_OK;
}


// ---------------------- 4D vector --------------------------------
gfVector4::gfVector4()
{
    m_data[0] = 0.0;
    m_data[1] = 0.0;
    m_data[2] = 0.0;
    m_data[3] = 0.0;
}

gfVector4::gfVector4(gfDouble x, gfDouble y, gfDouble z, gfDouble w)
{
    m_data[0] = x;
    m_data[1] = y;
    m_data[2] = z;
    m_data[3] = w;
}

gfVector4::~gfVector4()
{
}

gfVector4* gfVector4::Clone()
{
    gfVector4* new_vector = gfCreateVector4();
    new_vector->Set(m_data[0], m_data[1], m_data[2], m_data[3]);
    return new_vector;
}

GFRESULT gfVector4::Set(gfDouble x, gfDouble y, gfDouble z, gfDouble w)
{
    m_data[0] = x;
    m_data[1] = y;
    m_data[2] = z;
    m_data[3] = w;

    return GF_S_OK;
}

GFRESULT gfVector4::Set(gfInt index, gfDouble value)
{
    if (index < 1 || index > 4)
        return GF_E_BADARG;

    m_data[index - 1] = value;

    return GF_S_OK;
}

gfDouble gfVector4::at(gfInt index)
{
    if (index < 1 || index > 4)
        return 0.0;

    return m_data[index - 1];
}

gfDouble gfVector4::X(void)
{
    return m_data[0];
}

gfDouble gfVector4::Y(void)
{
    return m_data[1];
}

gfDouble gfVector4::Z(void)
{
    return m_data[2];
}

gfDouble gfVector4::W(void)
{
    return m_data[3];
}

GFRESULT gfVector4::SetX(gfDouble x)
{
    m_data[0] = x;
    return GF_S_OK;
}

GFRESULT gfVector4::SetY(gfDouble y)
{
    m_data[1] = y;
    return GF_S_OK;
}

GFRESULT gfVector4::SetZ(gfDouble z)
{
    m_data[2] = z;
    return GF_S_OK;
}

GFRESULT gfVector4::SetW(gfDouble w)
{
    m_data[3] = w;
    return GF_S_OK;
}

gfDouble* gfVector4::GetData(void)
{
    return (gfDouble*) m_data;
}

gfDouble gfVector4::Length(void)
{
    return gfSquareRoot(m_data[0] * m_data[0] + m_data[1] * m_data[1] + m_data[2] * m_data[2] + m_data[3] * m_data[3]);
}

GFRESULT gfVector4::SetLength(gfDouble new_length)
{
    if (gfIsZero(new_length))
        return GF_E_BADARG;

    gfDouble d = Length();
    if (gfIsZero(d))
        return GF_S_FALSE;

    m_data[0] *= new_length / d;
    m_data[1] *= new_length / d;
    m_data[2] *= new_length / d;
    m_data[3] *= new_length / d;

    return GF_S_OK;
}

GFRESULT gfVector4::Normalize(void)
{
    gfDouble d = Length();
    if (!gfIsZero(d))
    {
        m_data[0] /= d;
        m_data[1] /= d;
        m_data[2] /= d;
        m_data[3] /= d;
    }

    return GF_S_OK;
}

GFRESULT gfVector4::Translate(gfDouble tx, gfDouble ty, gfDouble tz, gfDouble tw)
{
    m_data[0] += tx;
    m_data[1] += ty;
    m_data[2] += tz;
    m_data[3] += tw;

    return GF_S_OK;
}

