//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfWinOGLExtensions.cpp                               //
//   Description: Definitions for working with OGL extensions   //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "gfWinOGLExtensions.h"
#include "gfAuxLogging.h"

PFNGLCREATEPROGRAMOBJECTARBPROC          glCreateProgramObjectARB        = NULL;
PFNGLDELETEOBJECTARBPROC                 glDeleteObjectARB               = NULL;
PFNGLDETACHOBJECTARBPROC                 glDetachObjectARB               = NULL;
PFNGLUSEPROGRAMOBJECTARBPROC             glUseProgramObjectARB           = NULL;
PFNGLCREATESHADEROBJECTARBPROC           glCreateShaderObjectARB         = NULL;
PFNGLSHADERSOURCEARBPROC                 glShaderSourceARB               = NULL;
PFNGLCOMPILESHADERARBPROC                glCompileShaderARB              = NULL;
PFNGLGETOBJECTPARAMETERIVARBPROC         glGetObjectParameterivARB       = NULL;
PFNGLATTACHOBJECTARBPROC                 glAttachObjectARB               = NULL;
PFNGLGETINFOLOGARBPROC                   glGetInfoLogARB                 = NULL;
PFNGLLINKPROGRAMARBPROC                  glLinkProgramARB                = NULL;
PFNGLGETUNIFORMLOCATIONARBPROC           glGetUniformLocationARB         = NULL;
PFNGLUNIFORM4FARBPROC                    glUniform4fARB                  = NULL;
PFNGLUNIFORM3FARBPROC                    glUniform3fARB                  = NULL;
PFNGLUNIFORM2FARBPROC                    glUniform2fARB                  = NULL;
PFNGLUNIFORM1FARBPROC                    glUniform1fARB                  = NULL;
PFNGLUNIFORM4IARBPROC                    glUniform4iARB                  = NULL;
PFNGLUNIFORM3IARBPROC                    glUniform3iARB                  = NULL;
PFNGLUNIFORM2IARBPROC                    glUniform2iARB                  = NULL;
PFNGLUNIFORM1IARBPROC                    glUniform1iARB                  = NULL;

PFNGLGETATTRIBLOCATIONARBPROC            glGetAttribLocationARB          = NULL;

PFNWGLGETEXTENSIONSSTRINGARBPROC            wglGetExtensionsStringARB       = NULL;
PFNWGLGETPIXELFORMATATTRIBIVARBPROC         wglGetPixelFormatAttribivARB    = NULL;
PFNWGLGETPIXELFORMATATTRIBFVARBPROC         wglGetPixelFormatAttribfvARB    = NULL;
PFNWGLCHOOSEPIXELFORMATARBPROC              wglChoosePixelFormatARB         = NULL;
PFNWGLCREATEPBUFFERARBPROC                  wglCreatePbufferARB             = NULL;
PFNWGLGETPBUFFERDCARBPROC                   wglGetPbufferDCARB              = NULL;
PFNWGLRELEASEPBUFFERDCARBPROC               wglReleasePbufferDCARB          = NULL;
PFNWGLDESTROYPBUFFERARBPROC                 wglDestroyPbufferARB            = NULL;
PFNWGLQUERYPBUFFERARBPROC                   wglQueryPbufferARB              = NULL;
PFNWGLBINDTEXIMAGEARBPROC			        wglBindTexImageARB              = NULL;
PFNWGLRELEASETEXIMAGEARBPROC			    wglReleaseTexImageARB           = NULL;
PFNWGLSETPBUFFERATTRIBARBPROC		        wglSetPbufferAttribARB          = NULL;

PFNGLGENBUFFERSARBPROC                      glGenBuffersARB                 = NULL;
PFNGLBINDBUFFERARBPROC                      glBindBufferARB                 = NULL;
PFNGLBUFFERDATAARBPROC                      glBufferDataARB                 = NULL;
PFNGLDELETEBUFFERSARBPROC                   glDeleteBuffersARB              = NULL;

PFNGLLOCKARRAYSEXTPROC                      glLockArraysEXT                 = NULL;
PFNGLUNLOCKARRAYSEXTPROC                    glUnlockArraysEXT               = NULL;


unsigned char *gfReadFile(const char *file_name)
{
    FILE *file = fopen(file_name, "r");

    if( file == NULL )
    {
        gfLog("Cannot open file %s.\n", file_name);
		return NULL;
    }

    struct _stat file_stats;
    if(_stat(file_name, &file_stats) != 0 )
    {
        gfLog("Cannot get file stats for file %s.\n", file_name);
        return NULL;
    }

    unsigned char *buffer = new unsigned char[file_stats.st_size];

	gfInt bytes = (gfInt) fread(buffer, 1, file_stats.st_size, file);
    buffer[bytes] = 0;
	fclose(file);

	return buffer;
}

GFRESULT gfCheckExtension(const char *ext_name)
{
    char *exts = (char*)glGetString(GL_EXTENSIONS);
    if(strstr(exts, ext_name) == NULL)
    {
        gfLog("Extension %s isn't supported.\n", ext_name);
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfCheckWGLExtension(const char *ext_name)
{
    if (!wglGetExtensionsStringARB)
    {
        gfLog("WGL Extension string isn't supported or initialized.\n");
        return GF_E_FAIL;
    }

    char *exts = (char*)wglGetExtensionsStringARB(wglGetCurrentDC());
    if(strstr(exts, ext_name) == NULL)
    {
        gfLog("Extension %s isn't supported.\n", ext_name);
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

gfOGLShaderInfo* gfCreateShader(const char *vertex_shader_file, const char *fragment_shader_file)
{
    gfOGLShaderInfo* shader = new gfOGLShaderInfo;
    ZeroMemory(shader, sizeof(gfOGLShaderInfo));

    const char *vertex_shader_strings[1];
    const char *fragment_shader_strings[1];
    GLint bVertCompiled;
    GLint bFragCompiled;
    GLint bLinked;
    char str[4096];

	// Create the vertex shader...
    unsigned char *vertex_shader_assembly = gfReadFile(vertex_shader_file);
    if (!vertex_shader_assembly)
    {
        gfLog("Couldn't open vertex shader file. No shader generated...\n");

        delete shader;

        return NULL;
    }
    shader->vertex_shader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);

    vertex_shader_strings[0] = (char*)vertex_shader_assembly;
    glShaderSourceARB(shader->vertex_shader, 1, vertex_shader_strings, NULL);
    glCompileShaderARB(shader->vertex_shader);
    delete vertex_shader_assembly;

    // check for compilation errors in vertex shader
    glGetObjectParameterivARB(shader->vertex_shader, GL_OBJECT_COMPILE_STATUS_ARB, &bVertCompiled);
    if(bVertCompiled  == false)
	{
		glGetInfoLogARB(shader->vertex_shader, sizeof(str), NULL, str);
        gfLog("============== Vertex shader compile error ==============\n");
        gfLog("File: %s\n", vertex_shader_file);
        gfLog("Errors: %s", str);
        gfLog("=========================================================\n");

        delete shader;

        return NULL;
	}

	// Create the fragment shader...
    unsigned char *fragment_shader_assembly = gfReadFile(fragment_shader_file);
    if (!fragment_shader_assembly)
    {
        gfLog("Couldn't open fragment shader file. No shader generated...\n");

        glDeleteObjectARB(shader->vertex_shader);
        delete shader;

        return NULL;
    }
    shader->fragment_shader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);

    fragment_shader_strings[0] = (char*)fragment_shader_assembly;
    glShaderSourceARB(shader->fragment_shader, 1, fragment_shader_strings, NULL);
    glCompileShaderARB(shader->fragment_shader);
    delete fragment_shader_assembly;

    glGetObjectParameterivARB(shader->fragment_shader, GL_OBJECT_COMPILE_STATUS_ARB, &bFragCompiled);
    if(bFragCompiled == false)
	{
		glGetInfoLogARB(shader->fragment_shader, sizeof(str), NULL, str);
        gfLog("============== Fragment shader compile error ==============\n");
        gfLog("File: %s\n", fragment_shader_file);
        gfLog("Errors: %s", str);
        gfLog("=========================================================\n");

        glDeleteObjectARB(shader->vertex_shader);
        glDeleteObjectARB(shader->fragment_shader);
        delete shader;

        return NULL;
	}

    // Create a program object for outline and attach the two compiled shaders...
    shader->program_obj = glCreateProgramObjectARB();
    glAttachObjectARB(shader->program_obj, shader->vertex_shader);
    glAttachObjectARB(shader->program_obj, shader->fragment_shader);
    glLinkProgramARB(shader->program_obj);
    glGetObjectParameterivARB(shader->program_obj, GL_OBJECT_LINK_STATUS_ARB, &bLinked);
    if( bLinked == false )
	{
		glGetInfoLogARB(shader->fragment_shader, sizeof(str), NULL, str);
        gfLog("============== Shader linking error =====================\n");
        gfLog("Errors: %s", str);
        gfLog("=========================================================\n");

        delete shader;
        glDeleteObjectARB(shader->vertex_shader);
        glDeleteObjectARB(shader->fragment_shader);
        glDeleteObjectARB(shader->program_obj);

        return NULL;
	}

    glGetInfoLogARB(shader->fragment_shader, sizeof(str), NULL, str);
    gfLog("============== Shader prepared successfully =============\n");
    gfLog("Vertex shader file: %s\n", vertex_shader_file);
    gfLog("Fragment shader file: %s\n", fragment_shader_file);
    gfLog("=========================================================\n");

    return shader;
}

void gfReleaseShader(gfOGLShaderInfo* shader)
{
    glDetachObjectARB(shader->program_obj, shader->vertex_shader);
    glDetachObjectARB(shader->program_obj, shader->fragment_shader);

    glDeleteObjectARB(shader->vertex_shader);
    glDeleteObjectARB(shader->fragment_shader);
    glDeleteObjectARB(shader->program_obj);

    delete shader;
}

void gfSetActiveShader(gfOGLShaderInfo* shader)
{
    if (shader)
        glUseProgramObjectARB(shader->program_obj);
    else
        glUseProgramObjectARB(NULL);

}

void gfSetShaderUniform4f(gfOGLShaderInfo* shader, const char* uniform, GLfloat f1, GLfloat f2, GLfloat f3, GLfloat f4)
{
    if (!glGetUniformLocationARB || !glUniform4fARB)
    {
        gfLog("Extension GL_ARB_shader_objects not initialized.\n");
        return;
    }

    GLuint location_light = glGetUniformLocationARB(shader->program_obj, uniform);
    glUniform4fARB(location_light, f1, f2, f3, f4);
}

void gfSetShaderUniform3f(gfOGLShaderInfo* shader, const char* uniform, GLfloat f1, GLfloat f2, GLfloat f3)
{
    if (!glGetUniformLocationARB || !glUniform3fARB)
    {
        gfLog("Extension GL_ARB_shader_objects not initialized.\n");
        return;
    }

    GLuint location_light = glGetUniformLocationARB(shader->program_obj, uniform);
    glUniform3fARB(location_light, f1, f2, f3);
}

void gfSetShaderUniform2f(gfOGLShaderInfo* shader, const char* uniform, GLfloat f1, GLfloat f2)
{
    if (!glGetUniformLocationARB || !glUniform2fARB)
    {
        gfLog("Extension GL_ARB_shader_objects not initialized.\n");
        return;
    }

    GLuint location_light = glGetUniformLocationARB(shader->program_obj, uniform);
    glUniform2fARB(location_light, f1, f2);
}

void gfSetShaderUniform1f(gfOGLShaderInfo* shader, const char* uniform, GLfloat f1)
{
    if (!glGetUniformLocationARB || !glUniform1fARB)
    {
        gfLog("Extension GL_ARB_shader_objects not initialized.\n");
        return;
    }

    GLuint location_light = glGetUniformLocationARB(shader->program_obj, uniform);
    glUniform1fARB(location_light, f1);
}

void gfSetShaderUniform4i(gfOGLShaderInfo* shader, const char* uniform, GLint i1, GLint i2, GLint i3, GLint i4)
{
    if (!glGetUniformLocationARB || !glUniform4iARB)
    {
        gfLog("Extension GL_ARB_shader_objects not initialized.\n");
        return;
    }

    GLuint location_light = glGetUniformLocationARB(shader->program_obj, uniform);
    glUniform4iARB(location_light, i1, i2, i3, i4);
}

void gfSetShaderUniform3i(gfOGLShaderInfo* shader, const char* uniform, GLint i1, GLint i2, GLint i3)
{
    if (!glGetUniformLocationARB || !glUniform3fARB)
    {
        gfLog("Extension GL_ARB_shader_objects not initialized.\n");
        return;
    }

    GLuint location_light = glGetUniformLocationARB(shader->program_obj, uniform);
    glUniform3iARB(location_light, i1, i2, i3);
}

void gfSetShaderUniform2i(gfOGLShaderInfo* shader, const char* uniform, GLint i1, GLint i2)
{
    if (!glGetUniformLocationARB || !glUniform2iARB)
    {
        gfLog("Extension GL_ARB_shader_objects not initialized.\n");
        return;
    }

    GLuint location_light = glGetUniformLocationARB(shader->program_obj, uniform);
    glUniform2iARB(location_light, i1, i2);
}

void gfSetShaderUniform1i(gfOGLShaderInfo* shader, const char* uniform, GLint i1)
{
    if (!glGetUniformLocationARB || !glUniform1fARB)
    {
        gfLog("Extension GL_ARB_shader_objects not initialized.\n");
        return;
    }

    GLuint location_light = glGetUniformLocationARB(shader->program_obj, uniform);
    glUniform1iARB(location_light, i1);
}

GFRESULT gfSetup_GL_ARB_shading_language_100(void)
{
    if (GF_FAIL(gfCheckExtension("GL_ARB_shading_language_100")))
        return GF_E_FAIL;

    gfLog("Setting up extension GL_ARB_shading_language_100.\n");

    return GF_S_OK;
}

GFRESULT gfSetup_GL_ARB_shader_objects(void)
{
    if (GF_FAIL(gfCheckExtension("GL_ARB_shader_objects")))
        return GF_E_FAIL;

    gfLog("Setting up extension GL_ARB_shader_objects.\n");

    glCreateProgramObjectARB  = (PFNGLCREATEPROGRAMOBJECTARBPROC)wglGetProcAddress("glCreateProgramObjectARB");
    glDeleteObjectARB         = (PFNGLDELETEOBJECTARBPROC)wglGetProcAddress("glDeleteObjectARB");
    glDetachObjectARB         = (PFNGLDETACHOBJECTARBPROC)wglGetProcAddress("glDetachObjectARB");
    glUseProgramObjectARB     = (PFNGLUSEPROGRAMOBJECTARBPROC)wglGetProcAddress("glUseProgramObjectARB");
    glCreateShaderObjectARB   = (PFNGLCREATESHADEROBJECTARBPROC)wglGetProcAddress("glCreateShaderObjectARB");
    glShaderSourceARB         = (PFNGLSHADERSOURCEARBPROC)wglGetProcAddress("glShaderSourceARB");
    glCompileShaderARB        = (PFNGLCOMPILESHADERARBPROC)wglGetProcAddress("glCompileShaderARB");
    glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)wglGetProcAddress("glGetObjectParameterivARB");
    glAttachObjectARB         = (PFNGLATTACHOBJECTARBPROC)wglGetProcAddress("glAttachObjectARB");
    glGetInfoLogARB           = (PFNGLGETINFOLOGARBPROC)wglGetProcAddress("glGetInfoLogARB");
    glLinkProgramARB          = (PFNGLLINKPROGRAMARBPROC)wglGetProcAddress("glLinkProgramARB");
    glGetUniformLocationARB   = (PFNGLGETUNIFORMLOCATIONARBPROC)wglGetProcAddress("glGetUniformLocationARB");
    glUniform4fARB            = (PFNGLUNIFORM4FARBPROC)wglGetProcAddress("glUniform4fARB");
    glUniform3fARB            = (PFNGLUNIFORM3FARBPROC)wglGetProcAddress("glUniform3fARB");
    glUniform2fARB            = (PFNGLUNIFORM2FARBPROC)wglGetProcAddress("glUniform2fARB");
    glUniform1fARB            = (PFNGLUNIFORM1FARBPROC)wglGetProcAddress("glUniform1fARB");
    glUniform4iARB            = (PFNGLUNIFORM4IARBPROC)wglGetProcAddress("glUniform4iARB");
    glUniform3iARB            = (PFNGLUNIFORM3IARBPROC)wglGetProcAddress("glUniform3iARB");
    glUniform2iARB            = (PFNGLUNIFORM2IARBPROC)wglGetProcAddress("glUniform2iARB");
    glUniform1iARB            = (PFNGLUNIFORM1IARBPROC)wglGetProcAddress("glUniform1iARB");

    if(!glCreateProgramObjectARB || !glDeleteObjectARB || !glUseProgramObjectARB ||
        !glCreateShaderObjectARB || !glCreateShaderObjectARB || !glCompileShaderARB || 
        !glGetObjectParameterivARB || !glAttachObjectARB || !glGetInfoLogARB || 
        !glLinkProgramARB || !glGetUniformLocationARB || !glUniform4fARB ||
        !glUniform1iARB)
    {
        gfLog("One or more GL_ARB_shader_objects functions were not found.\n");
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfSetup_GL_ARB_vertex_shader(void)
{
    if (GF_FAIL(gfCheckExtension("GL_ARB_vertex_shader")))
        return GF_E_FAIL;

    gfLog("Setting up extension GL_ARB_vertex_shader.\n");

    glGetAttribLocationARB = (PFNGLGETATTRIBLOCATIONARBPROC) wglGetProcAddress("glGetAttribLocationARB");
    
    if(!glGetAttribLocationARB)
    {
        gfLog("One or more GL_ARB_vertex_shader functions were not found.\n");
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfSetup_GL_ARB_fragment_shader(void)
{
    if (GF_FAIL(gfCheckExtension("GL_ARB_fragment_shader")))
        return GF_E_FAIL;

    gfLog("Setting up extension GL_ARB_fragment_shader.\n");

    return GF_S_OK;
}

GFRESULT gfSetup_WGL_ARB_extensions_string(void)
{
    gfLog("Setting up extension WGL_ARB_extensions_string.\n");

    wglGetExtensionsStringARB = (PFNWGLGETEXTENSIONSSTRINGARBPROC) wglGetProcAddress("wglGetExtensionsStringARB");

    if(!wglGetExtensionsStringARB)
    {
        gfLog("One or more WGL_ARB_extensions_string functions were not found.\n");
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfSetup_WGL_ARB_pixel_format(void)
{
    if (GF_FAIL(gfCheckWGLExtension("WGL_ARB_pixel_format")))
        return GF_E_FAIL;

    gfLog("Setting up extension WGL_ARB_pixel_format.\n");

    wglGetPixelFormatAttribivARB = (PFNWGLGETPIXELFORMATATTRIBIVARBPROC) wglGetProcAddress("wglGetPixelFormatAttribivARB");
	wglGetPixelFormatAttribfvARB = (PFNWGLGETPIXELFORMATATTRIBFVARBPROC) wglGetProcAddress("wglGetPixelFormatAttribfvARB");
	wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC) wglGetProcAddress("wglChoosePixelFormatARB");

    if (!wglGetPixelFormatAttribivARB ||
        !wglGetPixelFormatAttribfvARB ||
        !wglChoosePixelFormatARB)
    {
        gfLog("One or more WGL_ARB_pixel_format functions were not found.\n");
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfSetup_WGL_ARB_pbuffer(void)
{
    if (GF_FAIL(gfCheckWGLExtension("WGL_ARB_pbuffer")))
        return GF_E_FAIL;

    gfLog("Setting up extension WGL_ARB_pbuffer.\n");

    wglCreatePbufferARB = (PFNWGLCREATEPBUFFERARBPROC) wglGetProcAddress("wglCreatePbufferARB");
	wglGetPbufferDCARB = (PFNWGLGETPBUFFERDCARBPROC) wglGetProcAddress("wglGetPbufferDCARB");
	wglReleasePbufferDCARB = (PFNWGLRELEASEPBUFFERDCARBPROC) wglGetProcAddress("wglReleasePbufferDCARB");
	wglDestroyPbufferARB = (PFNWGLDESTROYPBUFFERARBPROC) wglGetProcAddress("wglDestroyPbufferARB");
	wglQueryPbufferARB = (PFNWGLQUERYPBUFFERARBPROC) wglGetProcAddress("wglQueryPbufferARB");

    if (!wglCreatePbufferARB ||
	    !wglGetPbufferDCARB ||
	    !wglReleasePbufferDCARB ||
	    !wglDestroyPbufferARB ||
	    !wglQueryPbufferARB)
    {
        gfLog("One or more WGL_ARB_pbuffer functions were not found.\n");
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfSetup_WGL_ARB_render_texture(void)
{
    if (GF_FAIL(gfCheckWGLExtension("WGL_ARB_render_texture")))
        return GF_E_FAIL;

    gfLog("Setting up extension WGL_ARB_render_texture.\n");

    wglBindTexImageARB = (PFNWGLBINDTEXIMAGEARBPROC) wglGetProcAddress("wglBindTexImageARB");
	wglReleaseTexImageARB = (PFNWGLRELEASETEXIMAGEARBPROC) wglGetProcAddress("wglReleaseTexImageARB");
	wglSetPbufferAttribARB = (PFNWGLSETPBUFFERATTRIBARBPROC) wglGetProcAddress("wglSetPbufferAttribARB");

    if (!wglBindTexImageARB ||
	    !wglReleaseTexImageARB ||
	    !wglSetPbufferAttribARB)
    {
        gfLog("One or more WGL_ARB_pbuffer functions were not found.\n");
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfSetup_GL_ARB_vertex_buffer_object(void)
{
    if (GF_FAIL(gfCheckExtension("GL_ARB_vertex_buffer_object")))
        return GF_E_FAIL;

    gfLog("Setting up extension GL_ARB_vertex_buffer_object.\n");

    glGenBuffersARB = (PFNGLGENBUFFERSARBPROC) wglGetProcAddress("glGenBuffersARB");
    glBindBufferARB = (PFNGLBINDBUFFERARBPROC) wglGetProcAddress("glBindBufferARB");
    glBufferDataARB = (PFNGLBUFFERDATAARBPROC) wglGetProcAddress("glBufferDataARB");
    glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC) wglGetProcAddress("glDeleteBuffersARB");

    if (!glGenBuffersARB ||
        !glBindBufferARB ||
        !glBufferDataARB ||
        !glDeleteBuffersARB)
    {
        gfLog("One or more GL_ARB_vertex_buffer_object functions were not found.\n");
        return GF_E_FAIL;
    }

    return GF_S_OK;
}

GFRESULT gfSetup_GL_EXT_compiled_vertex_array(void)
{
    if (GF_FAIL(gfCheckExtension("GL_EXT_compiled_vertex_array")))
        return GF_E_FAIL;

    gfLog("Setting up extension GL_EXT_compiled_vertex_array.\n");

    glLockArraysEXT = (PFNGLLOCKARRAYSEXTPROC) wglGetProcAddress( "glLockArraysEXT" );
    glUnlockArraysEXT = (PFNGLUNLOCKARRAYSEXTPROC) wglGetProcAddress( "glUnlockArraysEXT" ); 

    if (!glLockArraysEXT ||
        !glUnlockArraysEXT)
    {
        gfLog("One or more GL_EXT_compiled_vertex_array functions were not found.\n");
        return GF_E_FAIL;
    }

    return GF_S_OK;
}
