//////////////////////////////////////////////////////////////////
//                                                              //
//   Project: GeomForge                                         //
//   Author: Martin Samuelcik                                   //
//   File: gfWinOGLPBuffer.cpp                                  //
//   Description: Wrapper for working with OpenGL pixel buffer  //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "gfWinOGLPBuffer.h"
#include "gfAuxLogging.h"


GFRESULT gfWinOGLPBuffer::CreatePBuffer(gfInt width, gfInt height, gfInt color_bits, gfInt depth_bits, gfInt stencil_bits, 
                       gfInt num_extra_iattribs, gfInt* extra_iattrib_list, gfInt* flags)
{
    //Check for pbuffer support
	if(false)
	{
		gfLog("WGL extensions required for pbuffer unsupported.\n");
		return GF_E_FAIL;
	}

	//Get the current device context
	HDC h_current_DC = wglGetCurrentDC();
	if(!h_current_DC)
	{
		gfLog("Unable to get current Device Context.\n");
		return GF_E_FAIL;
	}
	

	// create attributes for pixel format
	GLint pixel_format;
	const gfInt standard_iattrib_list[]={WGL_DRAW_TO_PBUFFER_ARB, 1,
										WGL_COLOR_BITS_ARB, color_bits,
										WGL_ALPHA_BITS_ARB, color_bits == 32 ? 8 : 0,
										WGL_DEPTH_BITS_ARB, depth_bits,
										WGL_STENCIL_BITS_ARB, stencil_bits};
	const float f_attrib_list[]={0};

	//add the extraIAttribList to the standardIAttribList
	gfInt* i_attrib_list = new gfInt[sizeof(standard_iattrib_list) / sizeof(gfInt) + num_extra_iattribs * 2 + 1];
	if(!i_attrib_list)
	{
		gfLog("Unable to allocate space for attribute list.\n");
		return GF_E_FAIL;
	}

	memcpy(i_attrib_list, standard_iattrib_list, sizeof(standard_iattrib_list));
	memcpy(i_attrib_list + sizeof(standard_iattrib_list) / sizeof(gfInt),
			extra_iattrib_list, num_extra_iattribs * 2 * sizeof(gfInt) + sizeof(gfInt));

	//Choose pixel format
	unsigned int num_formats;
	if(!wglChoosePixelFormatARB(h_current_DC, i_attrib_list, f_attrib_list, 1,
								&pixel_format, &num_formats))
	{
		gfLog("Unable to find a pixel format for the pbuffer.\n");
		return GF_E_FAIL;
	}

	//Create the pbuffer
	m_hBuffer = wglCreatePbufferARB(h_current_DC, pixel_format, width, height, flags);
	if(!m_hBuffer)
	{
		gfLog("Unable to create pbuffer.\n");
		return GF_E_FAIL;
	}

	//Get the pbuffer's device context
	m_hDC = wglGetPbufferDCARB(m_hBuffer);
	if(!m_hDC)
	{
		gfLog("Unable to get pbuffer's device context.\n");
		return GF_E_FAIL;
	}

	//Create a rendering context for the pbuffer
	m_hRC = wglCreateContext(m_hDC);
	if(!m_hRC)
	{
		gfLog("Unable to create pbuffer's rendering context.\n");
		return GF_E_FAIL;
	}

	//Set and output the actual pBuffer dimensions
    gfInt w, h;
	wglQueryPbufferARB(m_hBuffer, WGL_PBUFFER_WIDTH_ARB, &w);
	wglQueryPbufferARB(m_hBuffer, WGL_PBUFFER_HEIGHT_ARB, &h);
	gfLog("Pbuffer Created: (%d x %d).\n", w, h);
	
	return GF_S_OK;	
}

GFRESULT gfWinOGLPBuffer::ReleasePBuffer(void)
{
    if(m_hRC)
	{
		if(!wglDeleteContext(m_hRC))
		{
			gfLog("Release of Pbuffer Rendering Context Failed.\n");
		}
		
		m_hRC = NULL;
	}

	if(m_hDC && !wglReleasePbufferDCARB(m_hBuffer, m_hDC))
	{
		gfLog("Release of Pbuffer Device Context Failed.\n");
		m_hDC = NULL;
	}
	
	if(!wglDestroyPbufferARB(m_hBuffer))
	{
		gfLog("Unable to destroy pbuffer.\n");
	}

    gfLog("Release of pbuffer finished.\n");

    return GF_S_OK;
}

void gfWinOGLPBuffer::BindToTexture(GLuint texture)
{
    glBindTexture(GL_TEXTURE_2D, texture);
	wglBindTexImageARB(m_hBuffer, WGL_FRONT_LEFT_ARB);
}

void gfWinOGLPBuffer::ReleaseFromTexture(void)
{
    wglReleaseTexImageARB(m_hBuffer, WGL_FRONT_LEFT_ARB);
}

GFRESULT gfWinOGLPBuffer::MakeCurrent(void)
{
    if(!wglMakeCurrent(m_hDC, m_hRC))
	{
		gfLog("Unable to change current pbuffer context.\n");
		return GF_E_FAIL;
	}

	return GF_S_OK;
}